#!/usr/bin/env python
"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""

__author__ = 'david.fann@vanderbilt.edu'

import logging
import os.path
import glob
import sys

import ConformanceMain
if __name__ == '__main__':
    ConformanceMain.initializeLogger()

egg_path = os.path.join(ConformanceMain.getRootDir(),'face_conformance_app', 'stringtemplate3-3.1-py2.7.egg')
sys.path.append(egg_path)
import stringtemplate3

import TestConfiguration
import constants
import c_parser
import ada_parser
import commonTools
import set_distro_statement

logger = logging.getLogger(__name__)

def limit_profile(profile_list):
    """
    removes any package/include file that is limited to TSS/IOS communication
    """
    new_list = list(profile_list)
    limited_packages = ['apex-sampling_ports', 'apex-sampling_ports-data', 'apex-sampling_ports-extensions',
                        'apex-queuing_ports', 'apex-queuing_port_list']
    for l in limited_packages:
        if l in new_list:
            new_list.remove(l)
    return new_list


class Arinc_653_Conformance:

    arinc_653_part1_packages = ['apex', 'apex-timing', 'apex-processes', 'apex-partitions',
                                'apex-sampling_ports', 'apex-queuing_ports', 'apex-buffers',
                                'apex-blackboards', 'apex-semaphores', 'apex-events', 'apex-health_monitoring',
                                'apex-adaptor']

    arinc_653_part2_packages = ['apex-file_system', 'apex-sampling_ports-data',
                                'apex-multiple_module_schedules', 'apex-logbook_system',
                                'apex-sampling_ports-extensions', 'apex-sap_ports',
                                'apex-name_service', 'apex-memory_block',
                                'apex-health_monitoring_extensions', 'apex-queuing_port_list']

    arinc_653_include_to_package_map = {'APEX_BLACKBOARD': 'apex-blackboards',
                                        'APEX_BUFFER': 'apex-buffers',
                                        'APEX_ERROR': 'apex-health_monitoring',
                                        'APEX_EVENT': 'apex-events',
                                        'APEX_FILE_SYSTEM':'apex-file_system',
                                        'APEX_HEALTH_MONITORING_EXTENSIONS': 'apex-health_monitoring_extensions',
                                        'APEX_LOGBOOK_SYSTEM': 'apex-logbook_system',
                                        'APEX_MEMORY_BLOCK': 'apex-memory_block',
                                        'APEX_MULTIPLE_MODULE_SCHEDULES':'apex-multiple_module_schedules',
                                        'APEX_NAME_SERVICE':'apex-name_service',
                                        'APEX_PARTITION': 'apex-partitions',
                                        'APEX_PROCESS':'apex-processes',
                                        'APEX_QUEUING': 'apex-queuing_ports',
                                        'APEX_QUEUING_PORT_LIST': 'apex-queuing_port_list',
                                        'APEX_SAMPLING': 'apex-sampling_ports',
                                        'APEX_SAMPLING_PORT_EXTENSIONS':'apex-sampling_ports-extensions',
                                        'APEX_SAP_PORTS': 'apex-sap_ports',
                                        'APEX_SEMAPHORE': 'apex-semaphores',
                                        'APEX_SP_DATA': 'apex-sampling_ports-data',
                                        'APEX_TIME': 'apex-timing',
                                        'APEX_TYPES': 'apex',
                                        }

    arinc_653_package_to_include_map = dict((v, k) for k, v in arinc_653_include_to_package_map.items())

    arinc_653_security_profile = arinc_653_part1_packages + ['apex-memory_block']
    arinc_653_safety_profile = arinc_653_part1_packages + ['apex-file_system', 'apex-sampling_ports-extensions',
                                                           'apex-sampling_ports-data', 'apex-memory_block']
    arinc_653_safety_ext_profile = list(arinc_653_safety_profile) # Right now, they are the same

    # General Profile is Part 1 + Part2 except log books and multiple module schedules
    arinc_653_general_profile = arinc_653_part1_packages + arinc_653_part2_packages
    arinc_653_general_profile.remove('apex-logbook_system')
    arinc_653_general_profile.remove('apex-multiple_module_schedules')

    arinc_653_general_posix_profile = ['apex', 'apex-sampling_ports', 'apex-queuing_ports', 'apex-health_monitoring',
                                       'apex-sampling_ports-data','apex-sampling_ports-extensions', 'apex-processes',
                                       'apex-health_monitoring_extensions', 'apex-queuing_port_list']

    arinc_653_general_limited_profile       = limit_profile(arinc_653_general_profile)
    arinc_653_safety_limited_profile        = limit_profile(arinc_653_safety_profile)
    arinc_653_safety_ext_limited_profile    = limit_profile(arinc_653_safety_ext_profile)
    arinc_653_security_limited_profile      = limit_profile(arinc_653_security_profile)
    arinc_653_general_posix_limited_profile = limit_profile(arinc_653_general_posix_profile)

    # CR 199
    arinc_653_safety_posix_profile              = ['apex', 'apex-sampling_ports', 'apex-queuing_ports', 'apex-health_monitoring',
                                                   'apex-sampling_ports-data','apex-sampling_ports-extensions', 'apex-processes']
    arinc_653_safety_posix_limited_profile      = ['apex', 'apex-health_monitoring', 'apex-processes']
    arinc_653_safety_ext_posix_profile          = arinc_653_safety_posix_profile
    arinc_653_safety_ext_posix_limited_profile  = arinc_653_safety_posix_limited_profile
    arinc_653_security_posix_profile            = ['apex', 'apex-sampling_ports', 'apex-queuing_ports', 'apex-health_monitoring',
                                                   'apex-processes']
    arinc_653_security_posix_limited_profile    = arinc_653_safety_posix_limited_profile


    directory_profile_packages = {'general': arinc_653_general_profile,
                                  'safetyBase': arinc_653_safety_profile,
                                  'safetyExt': arinc_653_safety_ext_profile,
                                  'security': arinc_653_security_profile,
                                  'general_posix_allowed': arinc_653_general_posix_profile,
                                  'general_limited': arinc_653_general_limited_profile,
                                  'safetyBase_limited': arinc_653_safety_limited_profile,
                                  'safetyExt_limited': arinc_653_safety_ext_limited_profile,
                                  'security_limited': arinc_653_security_limited_profile,
                                  'general_posix_allowed_limited': arinc_653_general_posix_limited_profile,
                                  'safetyBase_posix_allowed': arinc_653_safety_posix_profile,
                                  'safetyBase_posix_allowed_limited': arinc_653_safety_posix_limited_profile,
                                  'safetyExt_posix_allowed': arinc_653_safety_ext_posix_profile,
                                  'safetyExt_posix_allowed_limited': arinc_653_safety_ext_posix_limited_profile,
                                  'security_posix_allowed': arinc_653_security_posix_profile,
                                  'security_posix_allowed_limited': arinc_653_security_posix_limited_profile
                                  }

    '''
    directory_profile_packages = {'general': arinc_653_general_profile,
                                  'safetyBase': arinc_653_safety_profile,
                                  'safetyExt': arinc_653_safety_ext_profile,
                                  'security': arinc_653_security_profile,
                                  'posix_allowed': arinc_653_general_posix_profile,
                                  'general_limited': arinc_653_general_limited_profile,
                                  'safetyBase_limited': arinc_653_safety_limited_profile,
                                  'safetyExt_limited': arinc_653_safety_ext_limited_profile,
                                  'security_limited': arinc_653_security_limited_profile,
                                  'posix_allowed_limited': arinc_653_general_posix_limited_profile
                                  }
    '''

    all_languages = TestConfiguration.Language.allLanguages
    all_profiles = directory_profile_packages.keys()

    def generate_gold_libraries(self, **params):
        root_dir=params.get("rootDir", ConformanceMain.getRootDir())
        languages=params.get("languages", self.all_languages)
        profiles=params.get("profiles", self.all_profiles)
        logger.debug(locals())

        if TestConfiguration.Language.C in languages:
            self.generate_c_gold_libraries(root_dir, profiles)

        if TestConfiguration.Language.CPP in languages:
            self.generate_cpp_gold_libraries(root_dir, profiles)

        if TestConfiguration.Language.Ada in languages:
            self.generate_ada_gold_libraries(root_dir, profiles)

    def generate_c_source_file(self, root_dir, include_filename):
        arinc_include_file = open(include_filename, 'r')
        c_code = arinc_include_file.readlines()
        arinc_include_file.close()
        c_code = '\n'.join(c_code)
        c_parser_inst = c_parser.CParser()
        c_parser_inst.parse_string(c_code)
        logger.debug('finished parsing ' + os.path.basename(include_filename))
        source_filename = os.path.splitext(include_filename)[0] \
            + TestConfiguration.Language.source_suffix[TestConfiguration.Language.C]
        logger.debug('source filename: ' + source_filename)
        language = TestConfiguration.Language.C

        template_file=os.path.join(root_dir, constants.DATAFILES_DIR, 'ARINC_653_C_GoldLibTemplates.stg')
        template = stringtemplate3.StringTemplateGroup(name='gold_lib', file=open(template_file, 'r'))

        function_declarations = list()
        for node in iter(c_parser_inst.c_tree):
            if not isinstance(node, c_parser.OperationNode):
                continue
            logger.debug(node.__class__.__name__ + ': ' + node.name)
            parameters = list()

            for n in node.children:
                if isinstance(n, c_parser.ParameterItemNode):
                    parameter_declaration = template.getInstanceOf('parameterDeclaration')
                    if n.direction is not None:
                        parameter_declaration['DIRECTION'] = n.direction
                    parameter_declaration['TYPE_NAME'] = n.type_name
                    parameter_declaration['NAME'] = n.name
                    last_element = n == node.children[-1]
                    if last_element:
                        parameter_declaration['LAST_ELEMENT'] = True
                    parameters.append(str(parameter_declaration))

            function_declaration = template.getInstanceOf('functionDeclaration')
            function_declaration['RETURN_TYPE'] = node.return_type
            function_declaration['FUNCTION_NAME'] = node.name
            function_declaration['PARAMETERS'] = '\n'.join(parameters)

            function_declarations.append(str(function_declaration))

        if len(function_declarations) > 0: # do we need to create a source file?
            include_files = template.getInstanceOf('includeFiles')
            include_filename = os.path.splitext(os.path.basename(source_filename))[0] \
                               + TestConfiguration.Language.header_suffix[language]
            include_files['FILENAME'] = include_filename

            gold_lib_source_file = template.getInstanceOf('goldLibSourceFile')
            gold_lib_source_file['HEADING'] = \
                set_distro_statement.get_distribution_statement(language, os.path.basename(source_filename))
            gold_lib_source_file['INCLUDE_FILES'] = str(include_files)
            gold_lib_source_file['FUNCTION_DECLARATIONS'] = '\n'.join(function_declarations)

            source_file = open(source_filename, 'w')
            source_file.write(str(gold_lib_source_file))
            source_file.close()

    def generate_c_source_files(self, root_dir, arinc_full_imple_path):
        include_filenames = glob.glob(os.path.join(arinc_full_imple_path, 'APEX*.h'))
        if len(include_filenames) > 0: # i.e. got the right directory
            # delete the old source files
            source_suffix = TestConfiguration.Language.source_suffix[TestConfiguration.Language.C]
            old_src_files = glob.glob(os.path.join(arinc_full_imple_path, '*' + source_suffix))
            for o in old_src_files:
                    os.remove(o)

        # Parse include files to generate source files
        for include_filename in include_filenames:
            logger.debug('include file: ' + os.path.basename(include_filename))
            self.generate_c_source_file(root_dir, include_filename)

    def generate_cpp_gold_libraries(self, root_dir, profiles):
        gold_lib_dir = os.path.join(root_dir, constants.CPP_ARINC_653_GOLD_LIB_DIR)

        #ARINC is a C Library, so just create the same file, but store it in a different location
        self.generate_c_gold_libraries(root_dir, profiles, goldLibDir=gold_lib_dir)

    def generate_c_gold_libraries(self, root_dir, profiles, **params):

        arinc_gold_lib_path = params.get("goldLibDir", os.path.join(root_dir, constants.C_ARINC_653_GOLD_LIB_DIR))
        arinc_full_imple_path = params.get(
            "fullImplePath", os.path.join(root_dir, constants.C_ARINC_653_GOLD_LIB_DIR, 'full_implementation'))

        self.generate_c_source_files(root_dir, arinc_full_imple_path)

        for profile in profiles:
            logger.debug('Generating: ' + profile)
            included_names = list()
            excluded_names = list()
            for base_name, package_name in self.arinc_653_include_to_package_map.iteritems():
                if package_name in self.directory_profile_packages[profile]:
                    included_names.append(base_name)

                else:
                    excluded_names.append(base_name)

            logger.debug('included: ' + ', '.join(included_names))
            logger.debug('excluded: ' + ', '.join(excluded_names))

            # Create profile directory if it doesn't already exist
            profile_path = os.path.join(arinc_gold_lib_path, profile)
            commonTools.mkdir_p(profile_path)

            # Delete any old files in the profile directory
            old_files = glob.glob(os.path.join(profile_path, '*'))
            for o in old_files:
                os.remove(o)

            # Creating ARINC653.h and ARINC653P2.h that has the allowed include files for the given profile
            arinc_part1_include = open(os.path.join(profile_path, constants.ARINC_PART1_HEADER_NAME), 'w')
            arinc_part2_include = open(os.path.join(profile_path, constants.ARINC_PART2_HEADER_NAME), 'w')

            arinc_part1_include.write('/* These are the APEX headers from ARINC 653 Part 1 allowed in the '
                                      + profile + ' profile */\n\n')
            arinc_part2_include.write('/* These are the APEX headers from ARINC 653 Part 2 allowed in the '
                                      + profile + ' profile */\n\n')

            # copy the relevant files from the full implementation directory to the profile directory
            # and add them to the ARINC653.h and ARINC653P2.h files
            for included_name in included_names:
                commonTools.copy_files(os.path.join(arinc_full_imple_path, included_name + '.*'), profile_path)
                if self.arinc_653_include_to_package_map[included_name] in self.arinc_653_part1_packages:
                    arinc_part1_include.write('#include "' + included_name + '.h"\n')
                else:
                    arinc_part2_include.write('#include "' + included_name + '.h"\n')

            arinc_part1_include.close()
            arinc_part2_include.close()

    def generate_ada_source_file(self, root_dir, spec_filename):

        language = TestConfiguration.Language.Ada # Using Ada

        arinc_spec_file = open(spec_filename, 'r')
        ada_code = arinc_spec_file.readlines()
        arinc_spec_file.close()
        ada_code = '\n'.join(ada_code)
        ada_parser_inst = ada_parser.AdaParser()
        ada_parser_inst.parse_string(ada_code)
        logger.debug('finished parsing ' + os.path.basename(spec_filename))
        source_filename = os.path.splitext(spec_filename)[0] + TestConfiguration.Language.source_suffix[language]
        logger.debug('source filename: ' + source_filename)

        template_file=os.path.join(root_dir, constants.DATAFILES_DIR, 'ARINC_653_Ada_GoldLibTemplates.stg')
        template = stringtemplate3.StringTemplateGroup(name='gold_lib', file=open(template_file, 'r'))

        gold_lib_source_file = template.getInstanceOf('goldLibSourceFile')
        procedure_bodies = ''
        for node in iter(ada_parser_inst.ada_tree):
            if isinstance(node, ada_parser.PackageNode):
                gold_lib_source_file['PACKAGE_NAME'] = node.name

            elif isinstance(node, ada_parser.ProcedureNode):
                logger.info('Procedure: ' + node.name)
                procedure_bodies += node.generate_gold_lib_body(template, language)
            else:
                pass # For now, don't do anything...later might want to do something

        if len(procedure_bodies) > 0: # Actually had some bodies
            gold_lib_source_file['BODIES'] = procedure_bodies
            ada_body_file = open(source_filename, 'w')
            ada_body_file.write(str(gold_lib_source_file))
            ada_body_file.close()

    def generate_ada_source_files(self, root_dir):

        arinc_full_imple_path = os.path.join(root_dir, constants.ADA_ARINC_653_GOLD_LIB_DIR, 'full_implementation')

        spec_filenames = glob.glob(os.path.join(arinc_full_imple_path, 'apex*.ads'))
        if len(spec_filenames) > 0: # i.e. got the right directory
            # delete the old source files
            source_suffix = TestConfiguration.Language.source_suffix[TestConfiguration.Language.Ada]
            old_src_files = glob.glob(os.path.join(arinc_full_imple_path, '*' + source_suffix))
            for o in old_src_files:
                    os.remove(o)

        # Parse include files to generate source files
        for spec_filename in spec_filenames:
            logger.debug('spec file: ' + os.path.basename(spec_filename))
            self.generate_ada_source_file(root_dir, spec_filename)

    def generate_ada_gold_libraries(self, root_dir, profiles):

        self.generate_ada_source_files(root_dir)

        arinc_gold_lib_path = os.path.join(root_dir, constants.ADA_ARINC_653_GOLD_LIB_DIR)
        arinc_full_imple_path = os.path.join(root_dir, constants.ADA_ARINC_653_GOLD_LIB_DIR, 'full_implementation')

        for profile in profiles:
            logger.debug('Generating: ' + profile)
            included_names = list()
            excluded_names = list()
            for base_name, package_name in self.arinc_653_include_to_package_map.iteritems():
                # don't use the base_name for building ada...
                if package_name in self.directory_profile_packages[profile]:
                    included_names.append(package_name)
                else:
                    excluded_names.append(package_name)

            logger.debug('included: ' + ', '.join(included_names))
            logger.debug('excluded: ' + ', '.join(excluded_names))

            # Create profile directory if it doesn't already exist
            profile_path = os.path.join(arinc_gold_lib_path, profile)
            commonTools.mkdir_p(profile_path)

            # Delete any old files in the profile directory
            old_files = glob.glob(os.path.join(profile_path, '*'))
            for o in old_files:
                os.remove(o)

            # copy the relevant files from the full implementation directory to the profile directory
            for included_name in included_names:
                commonTools.copy_files(os.path.join(arinc_full_imple_path, included_name + '.*'), profile_path)

    def generate_conformance_tests(self, **params):
        root_dir=params.get("rootDir", ConformanceMain.getRootDir())
        languages=params.get("languages", self.all_languages)
        profiles=params.get("profiles", self.all_profiles)
        logger.debug(locals())

        if TestConfiguration.Language.C in languages:
            self.generate_c_conformance_tests(root_dir)

        if TestConfiguration.Language.Ada in languages:
            self.generate_ada_conformance_tests(root_dir)

    def generate_c_include_file_conformance_tests(
            self, root_dir, include_filename, target_directory, added_assertion_text=''):

        template_file=os.path.join(root_dir, constants.DATAFILES_DIR, 'ARINC_653_C_ConformanceTemplates.stg')
        if os.path.isfile(template_file):
            template = stringtemplate3.StringTemplateGroup(name='conformance', file=open(template_file, 'r'))
        else:
            raise IOError('Could not find: ' + template_file)

        c_parser_inst = c_parser.CParser()
        include_file = open(include_filename, 'r')
        include_file_lines = include_file.readlines()
        include_file_parsed = c_parser_inst.parse_string('\n'.join(include_file_lines))

        if include_file_parsed:
            c_conformance_source_basename, extension = os.path.splitext(os.path.basename(include_filename))

            assertion_filename = os.path.join(target_directory, c_conformance_source_basename + '_assertions.txt')
            assertion_file = open(assertion_filename, 'w')
            all_tests_filename = os.path.join(target_directory, c_conformance_source_basename + '_all_tests.txt')
            all_tests_file = open(all_tests_filename, 'w')
            test_num = 1
            language = TestConfiguration.Language.C
            conformance_test_exists = False
            for node in iter(c_parser_inst.c_tree):

                if not isinstance(node, c_parser.OperationNode):
                    continue

                conformance_test_exists = True
                logger.debug(node.__class__.__name__ + ': ' + node.name)
                parameters = list()

                # CR 477: Requires declaring variables and invoking functions with the variables
                #         instead of declaring pointers to the functions.
                parameter_instantiation_list = list()

                for n in node.children:
                    """
                    if isinstance(n, c_parser.ParameterItemNode):
                        parameter_declaration = template.getInstanceOf('parameterDeclaration')
                        if n.direction is not None:
                            parameter_declaration['DIRECTION'] = n.direction
                        parameter_declaration['TYPE_NAME'] = n.type_name
                        parameter_declaration['NAME'] = n.name
                        last_element = n == node.children[-1]
                        if last_element:
                            parameter_declaration['LAST_ELEMENT'] = True
                        parameters.append(str(parameter_declaration))
                    """
                    if isinstance(n, c_parser.ParameterItemNode):
                        parameter_declaration = template.getInstanceOf('parameterDeclaration_2')
                        if n.direction is not None:
                            parameter_declaration['DIRECTION'] = n.direction
                        parameter_declaration['NAME'] = n.name
                        last_element = n == node.children[-1]
                        if last_element:
                            parameter_declaration['LAST_ELEMENT'] = True
                        parameters.append(str(parameter_declaration))
                        parameter_instantiation = template.getInstanceOf('parameterInstantiation')
                        parameter_instantiation['TYPE_NAME'] = n.type_name
                        parameter_instantiation['NAME'] = n.name
                        parameter_instantiation_list.append(str(parameter_instantiation))

                # CR 477
                """
                function_signature_test = template.getInstanceOf('functionSignatureTest')
                function_signature_test['RETURN_TYPE'] = node.return_type.replace('extern', '')
                function_signature_test['FUNCTION_NAME'] = node.name
                function_signature_test['PARAMETERS'] = '\n'.join(parameters)
                """
                function_signature_test = template.getInstanceOf('functionSignatureTest_2')
                function_signature_test['RETURN_TYPE'] = node.return_type.replace('extern', '')
                function_signature_test['FUNCTION_NAME'] = node.name
                function_signature_test['PARAMETERS'] = '\n'.join(parameters)
                function_signature_test['PARAMETERS_INSTANTIATION'] = '\n'.join(parameter_instantiation_list)


                included_filenames = ['ARINC653_CONFORMANCE_TEST.h']
                included_files = list()
                for i in included_filenames:
                    include_files_template = template.getInstanceOf('includeFiles')
                    include_files_template['FILENAME'] = i
                    included_files.append(str(include_files_template))

                conformance_source_file = template.getInstanceOf('conformanceSourceFile')
                conformance_source_file['HEADER'] = '\n'.join(included_files)

                conformance_source_file['TEST_CODE'] = str(function_signature_test)

                test_basename = 'test_' + c_conformance_source_basename + '_' + str(test_num)
                test_basename = test_basename.replace('-', '_')
                logger.debug('test basename: ' + test_basename)

                source_filename = test_basename + TestConfiguration.Language.source_suffix[language]
                source_file = open(os.path.join(target_directory, source_filename), 'w')
                source_file.write(set_distro_statement.get_distribution_statement(language, source_filename))
                source_file.write(str(conformance_source_file))
                source_file.close()

                all_tests_file.write(2*'\n' + TestConfiguration.Language.begin_single_line_comment[language] +
                                     80 * '#' + TestConfiguration.Language.end_single_line_comment[language] + '\n')

                all_tests_file.write(TestConfiguration.Language.begin_single_line_comment[language] + ' ' +
                                     test_basename + TestConfiguration.Language.source_suffix[language] +
                                     ' Test Code ' +
                                     TestConfiguration.Language.end_single_line_comment[language] + '\n')

                all_tests_file.write(str(conformance_source_file))

                assertion_file.write('Assertion' + str(test_num) + ' Testing ' +
                                        node.name + added_assertion_text +
                                     ' conformance\n')
                test_num += 1

        assertion_file.close()
        all_tests_file.close()


        if not conformance_test_exists: # didn't actually write a conformance test
            logger.warning('Include File: ' + include_filename + ' did not generate any conformance tests')
            os.remove(all_tests_filename)
            os.remove(assertion_filename)

        return include_file_parsed

    def generate_c_conformance_tests(self, root_dir):

        arinc_gold_full_imple_path = os.path.join(
            root_dir, constants.C_ARINC_653_GOLD_LIB_DIR, 'full_implementation')

        arinc_conformance_full_imple_path = os.path.join(
            root_dir, constants.C_ARINC_653_CONFORMANCE_DIR, 'full_implementation')

        if not os.path.isdir(arinc_conformance_full_imple_path):
            commonTools.mkdir_p(arinc_conformance_full_imple_path)

        include_filenames = glob.glob(os.path.join(arinc_gold_full_imple_path, 'APEX*.h'))
        if len(include_filenames) > 0: # i.e. got the right directory
            # Parse include files to generate source files
            for include_filename in include_filenames:
                logger.debug('include file: ' + os.path.basename(include_filename))
                self.generate_c_include_file_conformance_tests(
                    root_dir, include_filename, arinc_conformance_full_imple_path)

        conformance_test_base_directory =  os.path.join(root_dir, constants.C_ARINC_653_CONFORMANCE_DIR)
        for directory, allowed_packages in self.directory_profile_packages.iteritems():
            if directory.endswith('limited') > 0:  # Don't need to do a conformance test for the limited profiles
                continue

            target_directory = os.path.join(conformance_test_base_directory, directory)
            subdirectory_used = False
            for a in allowed_packages:
                if a == 'apex':
                    continue
                if a not in self.arinc_653_package_to_include_map:
                    continue

                c = self.arinc_653_package_to_include_map[a]

                files_to_copy = glob.glob(os.path.join(
                    arinc_conformance_full_imple_path, 'test_' + c.replace('-', '_') + '_*.c'))

                if len(files_to_copy) == 0:
                    continue

                subdirectory_used = True
                subdirectory = c.replace('APEX_', '')
                subdirectory = os.path.join(target_directory, subdirectory)
                commonTools.mkdir_p(subdirectory)
                old_c_files = glob.glob(os.path.join(subdirectory, '*.c'))
                for o in old_c_files:
                    os.remove(o)

                logger.debug('INCLUDE: ' + a + ' SUBDIRECTORY: ' + subdirectory)

                for f in files_to_copy:
                    new_filename = os.path.basename(f)
                    new_filename = new_filename.replace('_' + c + '_', '').capitalize()
                    commonTools.copyFile(f, os.path.join(subdirectory, new_filename))

                #commonTools.copy_files(os.path.join(arinc_conformance_full_imple_path,
                #                                    'test_' + c + '_*.c'),
                #                      subdirectory)
                commonTools.copyFile(os.path.join(arinc_conformance_full_imple_path,
                                                  c + '_assertions.txt'),
                                     os.path.join(subdirectory, 'assertions.txt'))
            if subdirectory_used:
                makefile = os.path.join(arinc_conformance_full_imple_path, 'Makefile')
                if os.path.isfile(makefile):
                    commonTools.copy(makefile, target_directory)

    def generate_ada_conformance_tests(self, root_dir):

        arinc_gold_full_imple_path = os.path.join(
            root_dir, constants.ADA_ARINC_653_GOLD_LIB_DIR, 'full_implementation')

        arinc_conformance_full_imple_path = os.path.join(
            root_dir, constants.ADA_ARINC_653_CONFORMANCE_DIR, 'full_implementation')

        spec_filenames = glob.glob(os.path.join(arinc_gold_full_imple_path, 'apex*.ads'))
        if len(spec_filenames) > 0: # i.e. got the right directory
            # Parse include files to generate source files
            for spec_filename in spec_filenames:
                logger.debug('spec file: ' + os.path.basename(spec_filename))
                self.generate_ada_spec_file_conformance_tests(root_dir, spec_filename, arinc_conformance_full_imple_path)

        conformance_test_base_directory =  os.path.join(root_dir, constants.ADA_ARINC_653_CONFORMANCE_DIR)
        for directory, allowed_packages in self.directory_profile_packages.iteritems():
            if directory.endswith('limited') > 0:  # Don't need to do a conformance test for the limited profiles
                continue

            target_directory = os.path.join(conformance_test_base_directory, directory)
            #if not os.path.isdir(target_directory):
            #    commonTools.mkdir_p(target_directory)
            subdirectory_used = False
            for a in allowed_packages:
                if a == 'apex':
                    continue

                files_to_copy = glob.glob(os.path.join(
                    arinc_conformance_full_imple_path, 'test_' + a.replace('-', '_') + '_*.ad*'))

                if len(files_to_copy) == 0:
                    continue

                subdirectory_used = True
                subdirectory = a.replace\
                    ('apex-', '').upper().replace('-','_')
                subdirectory = os.path.join(target_directory, subdirectory)
                commonTools.mkdir_p(subdirectory)
                old_ada_files = glob.glob(os.path.join(subdirectory, '*.ad*'))
                for o in old_ada_files:
                    os.remove(o)

                logger.debug('PACKAGE: ' + a + ' SUBDIRECTORY: ' + subdirectory)
                commonTools.copy_files(os.path.join(arinc_conformance_full_imple_path,
                                                    'test_' + a.replace('-', '_') + '_*.ad*'),
                                       subdirectory)
                commonTools.copyFile(os.path.join(arinc_conformance_full_imple_path,
                                                  a + '_assertions.txt'),
                                     os.path.join(subdirectory, 'assertions.txt'))
            if subdirectory_used:
                commonTools.copy(os.path.join(arinc_conformance_full_imple_path, 'Makefile'), target_directory)

    def generate_ada_spec_file_conformance_tests(
            self, root_dir, ada_spec_filename, target_directory, added_assertion_text=''):

        template_file=os.path.join(root_dir, constants.DATAFILES_DIR, 'ARINC_653_Ada_ConformanceTemplates.stg')
        if os.path.isfile(template_file):
            template = stringtemplate3.StringTemplateGroup(name='conformance', file=open(template_file, 'r'))
        else:
            raise IOError('Could not find: ' + template_file)

        ada_parser_inst = ada_parser.AdaParser()
        spec_file_parsed = ada_parser_inst.parse_ada_spec_file(ada_spec_filename)

        if spec_file_parsed:
            ada_conformance_body_basename, extension = os.path.splitext(os.path.basename(ada_spec_filename))

            assertion_filename = os.path.join(target_directory, ada_conformance_body_basename + '_assertions.txt')
            assertion_file = open(assertion_filename, 'w')
            all_tests_filename = os.path.join(target_directory, ada_conformance_body_basename + '_all_tests.txt')
            all_tests_file = open(all_tests_filename, 'w')
            test_num = 1
            language = TestConfiguration.Language.Ada
            with_modules = list()
            conformance_test_exists = False
            for current_node in iter(ada_parser_inst.ada_tree):
                if isinstance(current_node, ada_parser.WithNode):
                    with_modules.append(current_node.name)

                if current_node.conformance_test_node:
                    logger.debug('writing conformance test for: ' + current_node.toString())
                    conformance_test_exists = True
                    module_nodes = current_node.get_specific_node_type_parents([ada_parser.PackageNode])
                    module_names = with_modules
                    conformance_module_names = list()

                    for n in module_nodes:
                        full_module_name = n.get_fully_specified_name(language)
                        module_name = ''
                        for f in full_module_name.split('.'):
                            if len(module_name) > 0:
                                module_name += '.' + f
                            else:
                                module_name = f
                            module_names.append(module_name)

                    module_names += current_node.additional_modules()

                    test_basename = 'test_' + ada_conformance_body_basename + '_' + str(test_num)
                    test_basename = test_basename.replace('-', '_')
                    logger.debug('Module Names: ' + ', '.join(module_names))
                    logger.debug('test basename: ' + test_basename)

                    test_code = current_node.get_conformance_test_code(template, language)
                    code_template = template.getInstanceOf('conformanceSourceFile')
                    code_template['TEST_CODE_SECTION1'] = test_code[0]
                    code_template['TEST_CODE_SECTION2'] = test_code[1]
                    code_template['FACE_PACKAGES'] = ', '.join(set(module_names))
                    code_template['CONFORMANCE_TEST_FACE_PACKAGES'] = ', '.join(conformance_module_names)
                    code_template['TEST_NAME'] = test_basename

                    source_filename = test_basename + TestConfiguration.Language.source_suffix[language]
                    source_file = open(os.path.join(target_directory, source_filename), 'w')
                    source_file.write(set_distro_statement.get_distribution_statement(language, source_filename))
                    source_file.write(str(code_template))
                    source_file.close()

                    all_tests_file.write(2*'\n' + TestConfiguration.Language.begin_single_line_comment[language] +
                                         80 * '#' + TestConfiguration.Language.end_single_line_comment[language] + '\n')

                    all_tests_file.write(TestConfiguration.Language.begin_single_line_comment[language] + ' ' +
                                         test_basename + TestConfiguration.Language.source_suffix[language] +
                                         ' Test Code ' +
                                         TestConfiguration.Language.end_single_line_comment[language] + '\n')

                    all_tests_file.write(str(code_template))

                    assertion_file.write('Assertion' + str(test_num) + ' Testing ' +
                                         current_node.get_fully_specified_name(language) + added_assertion_text +
                                         ' is conformant\n')
                    test_num += 1
            assertion_file.close()
            all_tests_file.close()
            if not conformance_test_exists: # didn't actually write a conformance test
                logger.warning('Spec File: ' + ada_spec_filename + ' did not generate any conformance tests')
                os.remove(all_tests_filename)
                os.remove(assertion_filename)

        return spec_file_parsed




if __name__ == '__main__':
    logger = logging.getLogger(__name__)
    logger.info('main running from ' + __file__)
    logger.info('Used currently to generate for 1.0 conformance')
    arinc_conformance = Arinc_653_Conformance()

    root_dir = ConformanceMain.getRootDir()
    gold_dir = os.path.join(os.path.sep, 'home','dfann','FACE','ARINC_653_for_Conformance_1.0')
    imple_dir = os.path.join(gold_dir, 'full_implementation')
    if not os.path.isdir(root_dir):
        logger.error('Root dir does not exist: ' + root_dir)
        raise IOError

    if not os.path.isdir(gold_dir):
        logger.error('Gold dir does not exist: ' + gold_dir)
        raise IOError

    if not os.path.isdir(imple_dir):
        logger.error('Imple Dir does not exist: ' + imple_dir)
        raise IOError

    arinc_conformance.generate_c_gold_libraries(root_dir, arinc_conformance.directory_profile_packages.keys(),
                                                goldLibDir=gold_dir, fullImplePath=imple_dir)
