#!/usr/bin/env python
"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""

__author__ = 'david.fann@vanderbilt.edu'


import logging
import logging.config
import os


def initializeLogger(**kw):
    """
    initializes the standard python logger using either a configuration file or if
    no file is set, will default to a stream and file handler saving to conformanceTestSuite.log

    keywords:
    configFile - the configuration file to use for the logger.  Default is logger.ini
    logFilename - the log filename to use for the logger (ignored if the configFile exists)
    """

    configFile = kw.get('configFile', os.path.join(getRootDir(), 'face_conformance_app', 'logger.ini'))
    filename = kw.get('logFilename', 'conformanceTestSuite.log')
    if os.path.isfile(configFile):
        logging.config.fileConfig(configFile, None, True)

    else:  # no logger.ini file...therefore use default values
        fh = logging.FileHandler(filename)
        fh.setLevel(logging.DEBUG)
        ch = logging.StreamHandler()
        ch.setLevel(logging.WARNING)
        formatter = logging.Formatter('%(asctime)s %(levelname)8s: %(filename)s:%(lineno)d - %(message)s')
        fh.setFormatter(formatter)
        ch.setFormatter(formatter)

    conformanceStartString = 'Logging initialized'
    if os.path.isfile(configFile):
        conformanceStartString += ' using ' + configFile + ' for configuration'
    else:
        conformanceStartString += ' using default values, config file: ' + configFile + ' does not exist'
    logger = logging.getLogger(__name__)
    logger.info(conformanceStartString)

__rootDir = None

def getRootDir():
    # rootDir only needs to be determined once
    global __rootDir

    if __rootDir is None:
        face_conformance_app_path = os.path.dirname(os.path.realpath(__file__))
        __rootDir = os.path.abspath(os.path.join(face_conformance_app_path, os.pardir))

        """ OLD WAY
        cwd = os.getcwd()
        if os.path.isfile(os.path.join(cwd, 'runConformanceTest.sh')):
            __rootDir = cwd

        else:
            cwdSplit = cwd.rpartition(os.path.sep)
            while cwdSplit[0] != '':
                if os.path.isfile(os.path.join(cwdSplit[0], 'runConformanceTest.sh')):
                    __rootDir = cwdSplit[0]
                    break
                cwdSplit = cwdSplit[0].rpartition(os.path.sep)
            if __rootDir is None:
                raise RuntimeError('Cannot determine the test suite root directory.\n' +
                                   ' Must run test suite from its home directory')
        """
    return __rootDir


if __name__ == '__main__':
    initializeLogger()
    logger = logging.getLogger(__name__)
    logger.info('Running FACE Conformance Test from ConformanceMain')

    import constants
    import optparse
    import sys

    parser = optparse.OptionParser(usage="usage: %prog [options] [config_file1] [config_file2] ...")

    # TODO: add this option to the parser to run changeProjectDirectory before running test suite
    parser.add_option("-d", "--projectDir", dest="projectDir", metavar="new_project_directory",
                help="Run test suite using config file, but changing the project directory to new_project_directory. " +
                     "Changing project directory only affects segment related files, not build files. "
                     "A configuration file must be specified, and the original project directory must be specified in" +
                     " the configuration file.  Multiple configuration files may be specified.")

    parser.add_option("-m", "--model", action="store_true",
                      help="Test only the data model for conformance")

    parser.add_option("-o", "--logDir", dest="logDir", metavar="new_log_directory",
                help="Run test suite using config file, but changing the log directory to new_log_directory")

    parser.add_option("-g", "--gui", action="store_true",
                      help="Run test suite with GUI even if a configuration file is supplied")

    parser.add_option("-v", "--version", action="store_true",
                      help="Echo test suite version and FACE Standard version conformance is checked against")

    parser.add_option("-l", "--gsl", dest="genLibDir", metavar="generated_library_directory",
                      help="Generate Gold Standard Libraries needed by segments specified in the configuration file" +
                           " and store in the generated_library_directory. A configuration file must be specified," +
                           " but subsequent configuration files will be ignored")

    parser.description = "If config file(s) is(are) supplied, conformance test will run directly (no GUI)." \
        + "\nIf no config_file is supplied, the test suite GUI will launch for interactive control of conformance test."

    (options, args) = parser.parse_args()

    logger.info('Options: ' + str(options) + ' Args: ' + str(args))

    # args are the configuration files to process (if any)
    configFiles = args
    exitNormal = True

    if options.projectDir is not None:
        logger.info('Changing project directory')
        if len(args) < 1:
            errorMsg = 'Must pass new directory and configuration file to specify the project directory'
            logger.error(errorMsg)
            raise RuntimeError(errorMsg)

    # show versioning info only
    if options.version:
        import commonTools

        markingsXmlFile = os.path.join(getRootDir(), 'datafiles', 'FACEToolsMarkings.xml')
        toolMarkings = constants.FACEToolsMarkings(markingsXmlFile)
        toolMarkings.getToolInfo()
        FACEStandardVersion = '\n'.join(toolMarkings.FACEStandardVersion)
        toolsVersion = '\n'.join(toolMarkings.ToolsVersion)
        print 'FACE Standard Version: ' + FACEStandardVersion + '\nTest Suite Version: ' + toolsVersion

    # generate gold standard libraries
    elif options.genLibDir is not None:
        import ConformanceTest
        import commonTools
        commonTools.setInteractive(False)
        logger.info("Generating Gold Standard Libraries")

        if not os.path.isdir(options.genLibDir):
            errorMsg = 'Directory specified to store Gold Standard Libraries does not exists: ' + options.genLibDir
            logger.error(errorMsg)
            raise IOError(errorMsg)

        for c in configFiles:
            conformanceTest = ConformanceTest.getConformanceConfiguration(c)
            if options.projectDir is not None:
                conformanceTest.changeProjectDirectory(options.projectDir)
            if options.logDir is not None:
                conformanceTest.changeLogDirectory(options.logDir)
            conformanceTest.initializeConformanceTest()
            exitNormal = conformanceTest.buildGoldStandardLibraries(targetDir=options.genLibDir) and exitNormal

    # launch the GUI if there isn't a configuration file, or if the GUI option is specified
    elif len(configFiles) == 0 or options.gui == True:
        logger.info("Launching ConformanceTestGUI")
        import ConformanceTestGUI

        kw = dict()
        # use the first config file if it is specified
        if len(configFiles) > 0:
            kw['configFile'] = configFiles[0]

        conformanceTestGUI = ConformanceTestGUI.ConformanceTestGUI(**kw)
        if options.projectDir is not None:
            conformanceTestGUI.configuration.changeProjectDirectory(options.projectDir)
        if options.logDir is not None:
            conformanceTestGUI.configuration.changeLogDirectory(options.logDir)

        conformanceTestGUI.master.title(constants.FACE_TESTSUITE_NAME)
        conformanceTestGUI.mainloop()

    # otherwise run conformance test headless for all configuration files specified
    else:
        import ConformanceTest
        import commonTools

        commonTools.setInteractive(False)
        logger.info('Running Test Suite without GUI')
        exitNormal = True
        for configFilename in configFiles:
            conformanceTest = ConformanceTest.getConformanceConfiguration(configFilename)
            if options.model is not None:
                conformanceTest.testDataModelOnly()
            if options.projectDir is not None:
                conformanceTest.changeProjectDirectory(options.projectDir)
            if options.logDir is not None:
                conformanceTest.changeLogDirectory(options.logDir)
            testPassed, reportFilename = conformanceTest.runConformanceTest(interactive=False)
            exitNormal = exitNormal and testPassed

    if exitNormal:
        sys.exit(0)
    else:
        sys.exit(1)

