#!/usr/bin/env python
"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""

'''
---------------------------------------------------------------------------------
DISTRIBUTION STATEMENT A. Approved for public release; distribution is unlimited.

Product produced under DoD SENSIAC contract HC104705D4000 under the sponsorship of the Defense Technical Information Center, 
ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218.  
SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical Information Center.

HANDLING AND DESTRUCTION NOTICE - Comply with distribution statement and destroy by any method that will prevent disclosure 
of the contents or reconstruction of the document.

Copyright (c) Vanderbilt University, 2013

ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED

Developed under contract for the U.S. Department of Defense
by the Institute for Software Integrated Systems (ISIS), Vanderbilt University

The FACE Consortium will not currently accept any results from this release for certifying software components as FACE Conformant 
  

Vanderbilt University disclaims all warranties with regard to this software, 
including all implied warranties of merchantability and fitness.  In no event 
shall Vanderbilt University be liable for any special, indirect or 
consequential damages or any damages whatsoever resulting from loss of use, 
data or profits, whether in an action of contract, negligence or other tortious
action, arising out of or in connection with the use or performance of this 
software.

@author: david.fann@vanderbilt.edu
---------------------------------------------------------------------------------
'''

__author__ = 'david.fann@vanderbilt.edu'

import os
from time import ctime
from HTMLWriter import HTMLWriter
from commonTools import listToFilePath, listToString, safeOpen
from SourceCodeToHtml import codefileToHTML, addSourceCodeElement, stringListToHTML
import logging
import TestConfiguration
import commonTools
import JunitXMLWriter
import ConformanceMain
import JavaConformanceTest
import constants

logger = logging.getLogger(__name__)


class ConformanceReport():
    """
    Generates HTML report from results of conformance test
    """

    def __init__(self, conformanceTest, **kw):
        """
        Constructor
        """
        logger.info('kw: ' + str(kw))
        
        self.currentTime = ctime()
        self.conformanceTest = conformanceTest
    
        self.reportDirectory = kw.get('reportDirectory', None)
        self.reportFilename = kw.get('reportFile', None)

        if self.reportDirectory is None and self.reportFilename is None:
            self.reportDirectory = self.conformanceTest.baseConfig.logDirectory
            self.reportFilename = os.path.join(self.reportDirectory, 'ConformanceReport.html')
        
        elif self.reportFilename is None:
            self.reportFilename = os.path.join(self.reportDirectory, 'ConformanceReport.html')
        
        elif self.reportDirectory is None:
            self.reportDirectory = os.path.dirname(self.reportFilename)

        self.logSummaryFilename = os.path.join(self.reportDirectory, 'logsSummary.html')
    
        self.logFileToMenuID = dict()
        self.OSLogs = None
        
        markingsXmlFile = os.path.join(ConformanceMain.getRootDir(), 'datafiles', 'FACEToolsMarkings.xml')
        
        toolMarkings = constants.FACEToolsMarkings(markingsXmlFile)
        toolMarkings.getToolInfo()
        self.FACEStandardVersion = listToString(toolMarkings.FACEStandardVersion)
        self.toolsVersion = listToString(toolMarkings.ToolsVersion)
  
    def generate_junit_report(self):
        junit_xml_writer = JunitXMLWriter.JunitXMLWriter(self.conformanceTest, self.reportDirectory)
        junit_xml_writer.write_config_results()
    
    def generateReport(self):
        self.currentTime = ctime()
        
        reportFile = HTMLWriter("FACE Conformance Test Results")
        reportFile.addLink('', rel="stylesheet", type="text/css", href="ConformanceReport.css") 
        reportFile.addScript('', type="text/javascript", src="ConformanceReport.js")
        reportFile.addLink('', rel="stylesheet", type="text/css", href="styles/shCoreDefault.css")
        reportFile.addScript('', type="text/javascript", src="scripts/shCore.js")
        reportFile.addScript('', type="text/javascript", src="scripts/shBrushJScript.js")
        reportFile.addScript('', type="text/javascript", src="scripts/shBrushPlain.js")
        reportFile.addScript('', type="text/javascript", src="scripts/shBrushAda.js")
        reportFile.addScript('', type="text/javascript", src="scripts/shBrushCpp.js")
        reportFile.addScript('', type="text/javascript", src="scripts/shBrushJava.js")
        reportFile.addScript('SyntaxHighlighter.all();', type="text/javascript")
        metaAttr = { 'http-equiv':"Content-Type", 'content':"text/html; charset=UTF-8", 'superElement': reportFile.head }
        reportFile.addTag('meta', '', **metaAttr)

        mainHeadingElement = reportFile.addHeading(1, "FACE Conformance Test Results")

        reportFile.addButton('Expand All', cssClass = "headingButton", onClick='openAll()',superElement = mainHeadingElement)
        reportFile.addButton('Collapse All', cssClass = "headingButton", onClick='closeAll()', superElement = mainHeadingElement)

        #reportFile.addSpan(self.currentTime, cssClass="time", superElement = mainHeadingElement)
 
        menuElement, divElement = reportFile.addCollapsibleMenuItem("Test Configuration")
        reportFile.addHeading(3, 'Start Time/End Time: ' + self.conformanceTest.startTime + '/' + self.currentTime, superElement = divElement)
       
        reportFile.addHeading(3, 'Test Suite Information', superElement = divElement)
        testsuiteDivElement = reportFile.addDiv(superElement = divElement) 
       
        reportFile.addConfigElement('Test Suite Version', self.toolsVersion, superElement = testsuiteDivElement)
        reportFile.addConfigElement('FACE Technical Standard edition', self.FACEStandardVersion, superElement = testsuiteDivElement)
        if self.conformanceTest is not None:
            self.conformanceTest.toHTML(reportFile, divElement)


        testResult = True
        if self.conformanceTest.data_model_testing_required():
            testResult = self.addDataModelResultSummary(self.conformanceTest.datatypeConfig.dataModel, reportFile) and testResult

        if self.conformanceTest.testPC:
            testResult = self.addSegmentResultSummary(self.conformanceTest.pcConfig, reportFile) and testResult
            
        if self.conformanceTest.testTS:
            testResult = self.addSegmentResultSummary(self.conformanceTest.tsConfig, reportFile) and testResult
     
        if self.conformanceTest.testPSS:
            testResult = self.addSegmentResultSummary(self.conformanceTest.pssConfig, reportFile) and testResult

        if self.conformanceTest.testIOS:
            testResult = self.addSegmentResultSummary(self.conformanceTest.ioConfig, reportFile) and testResult
    
        if self.conformanceTest.testOS:
            testResult = self.addSegmentResultSummary(self.conformanceTest.osConfig, reportFile) and testResult

        reportFile.write(self.reportFilename)


        # make sure something was tested before returning a true
        somethingTested = (
            self.conformanceTest.testPC or
            self.conformanceTest.testTS or
            self.conformanceTest.testPSS or
            self.conformanceTest.testIOS or
            self.conformanceTest.testOS or
            self.conformanceTest.testDataModel
        )

        if not somethingTested:
            logger.warning('No segment specified to be tested in conformance test')
        
        testResult = testResult and somethingTested
        
        return testResult

    def updateJavaConformanceResult(self, current_value, new_value):
        """" Update value | Current value | Return Value
                FAIL      |     *         |  FAIL
                  *       |   FAIL        |  FAIL
                INSPECT   | PASS/INSPECT  |  INSPECT
                PASS      |      *        |  Current Value
        """

        if (current_value == JavaConformanceTest.JavaConformanceTestResult.FAIL or
            new_value == JavaConformanceTest.JavaConformanceTestResult.FAIL):
            return JavaConformanceTest.JavaConformanceTestResult.FAIL

        elif new_value == JavaConformanceTest.JavaConformanceTestResult.PASS:
            return current_value

        else: # new value must be inspect, and current value must not be fail...return inspect
            return JavaConformanceTest.JavaConformanceTestResult.INSPECT

    def addJavaResultText(self, result, superElement, htmlFile):
        if result == JavaConformanceTest.JavaConformanceTestResult.FAIL:
            htmlFile.addSpan(" FAILED", superElement=superElement, cssClass="failedTestResult")
        elif result == JavaConformanceTest.JavaConformanceTestResult.INSPECT:
            htmlFile.addSpan(" INSPECTION REQUIRED", superElement=superElement, cssClass="forkPresentResult")
        else:
            htmlFile.addSpan(" PASSED", superElement=superElement, cssClass="passedTestResult")


    def addJavaSubgroupResultSummary(self, java_subgroup, superElement, headingLevel, htmlFile):
        logger.info('Adding Java Subgroup: ' + java_subgroup.name)
        subgroup_conformant_result = JavaConformanceTest.JavaConformanceTestResult.PASS

        testDiv = htmlFile.addDiv(superElement=superElement)

        testMenu, testCollapsedElement = htmlFile.addCollapsibleMenuItem("Test: " + java_subgroup.name,
                                                                           headingLevel=headingLevel,
                                                                           superElement=testDiv)
        self.addRightArrow(htmlFile, testMenu)

        WarningErrorDict = { JavaConformanceTest.JavaConformanceTestResult.WARNING: 'Warning: ',
                             JavaConformanceTest.JavaConformanceTestResult.ERROR: 'Error: ',
                             JavaConformanceTest.JavaConformanceTestResult.INSPECT: 'Inspection Required: '}

        for r in java_subgroup.results:
            subgroup_conformant_result = self.updateJavaConformanceResult(subgroup_conformant_result, r.score)
            resultDiv = htmlFile.addDiv(superElement=testCollapsedElement)

            resultsMenu, resultsCollapsedElement = htmlFile.addCollapsibleMenuItem(r.name, headingLevel=headingLevel+1,
                                                                                   superElement=resultDiv)
            self.addRightArrow(htmlFile, resultsMenu)

            self.addJavaResultText(r.score, resultsMenu, htmlFile)

            htmlFile.addParagraph(r.description, superElement=resultsCollapsedElement)
            for l in r.logs:
                htmlFile.addParagraph(WarningErrorDict[l.level] + l.text, superElement=resultsCollapsedElement)

        for s in java_subgroup.subgroups:
            sub_subgroup_conformant = self.addJavaSubgroupResultSummary(s, testCollapsedElement, headingLevel+1, htmlFile)
            subgroup_conformant_result = self.updateJavaConformanceResult(subgroup_conformant_result, sub_subgroup_conformant)

        self.addJavaResultText(subgroup_conformant_result, testMenu, htmlFile)

        logger.debug('subgroup ' + java_subgroup.name + ' conformance is ' +
                     JavaConformanceTest.JavaConformanceTestResult.result_string(subgroup_conformant_result))

        return subgroup_conformant_result


    def addJavaPackageResultSummary(self, java_package, superElement, headingLevel, htmlFile):
        logger.info('Adding Java Package: ' + java_package.name)

        result_names = ['Attributes', 'Fields', 'Methods']
        if len(java_package.children) == 0:
            if java_package.subgroup is not None:
                if len(java_package.subgroup.subgroups) == 0:
                    return  # Special case for Attributes/Methods/Fields don't want to show them here

        testDiv = htmlFile.addDiv(superElement=superElement)

        testMenu, testCollapsedElement = htmlFile.addCollapsibleMenuItem(java_package.full_name,
                                                                         headingLevel=headingLevel,
                                                                         superElement=testDiv)
        self.addRightArrow(htmlFile, testMenu)

        WarningErrorDict = { JavaConformanceTest.JavaConformanceTestResult.WARNING: 'Warning: ',
                             JavaConformanceTest.JavaConformanceTestResult.ERROR: 'Error: ',
                             JavaConformanceTest.JavaConformanceTestResult.INSPECT: 'Inspect: '
        }
        package_conformant_result = JavaConformanceTest.JavaConformanceTestResult.PASS

        if java_package.subgroup is not None:
            for s in java_package.subgroup.subgroups:
                subgroupMenu, subgroupCollapsedElement = \
                    htmlFile.addCollapsibleMenuItem(s.name,
                                                    headingLevel=headingLevel+1,
                                                    superElement=testCollapsedElement)
                subgroup_conformant_result = JavaConformanceTest.JavaConformanceTestResult.PASS

                for r in s.results:
                    divElement = htmlFile.addDiv(superElement=subgroupCollapsedElement)
                    resultsMenu, resultsCollapsedElement = htmlFile.addCollapsibleMenuItem(r.name,
                                                                                         headingLevel=headingLevel+2,
                                                                                         superElement=divElement)
                    self.addRightArrow(htmlFile, resultsMenu)
                    subgroup_conformant_result = self.updateJavaConformanceResult(subgroup_conformant_result, r.score)
                    self.addJavaResultText(r.score, resultsMenu, htmlFile)

                    htmlFile.addParagraph(r.description, superElement=resultsCollapsedElement)
                    for l in r.logs:
                        htmlFile.addParagraph(WarningErrorDict[l.level] + l.text,
                                              superElement=resultsCollapsedElement)

                self.addRightArrow(htmlFile, subgroupMenu)
                self.addJavaResultText(subgroup_conformant_result, subgroupMenu, htmlFile)
                package_conformant_result = self.updateJavaConformanceResult(package_conformant_result, subgroup_conformant_result)


        for s in java_package.children:
            java_package_children_conformant = self.addJavaPackageResultSummary(s, testCollapsedElement, headingLevel+1, htmlFile)
            package_conformant_result = self.updateJavaConformanceResult(package_conformant_result, java_package_children_conformant)

        self.addJavaResultText(package_conformant_result, testMenu, htmlFile)

        logger.debug('package ' + java_package.name + ' conformance is ' +
                     JavaConformanceTest.JavaConformanceTestResult.result_string(package_conformant_result))

        return package_conformant_result



    def addJavaSegmentResultSummary(self, segment_config, htmlFile):

        menuElement, collapsibleElement = htmlFile.addCollapsibleMenuItem(segment_config.segmentName + " Tests")

        if segment_config.java_conformance_test_results is None:
            raise RuntimeError('Tried to display java results for a segment with no java results')
        segment_conformant = JavaConformanceTest.JavaConformanceTestResult.PASS

        for s in segment_config.java_conformance_test_results.subgroups:
            segment_conformant = self.updateJavaConformanceResult(
                segment_conformant,
                self.addJavaSubgroupResultSummary(s, collapsibleElement, 3, htmlFile)
            )
        self.addRightArrow(htmlFile, menuElement)

        self.addJavaResultText(segment_conformant, menuElement, htmlFile)

        return segment_conformant

    def addJavaOSSegmentResultSummary(self, segment_config, htmlFile):

        menuElement, collapsibleElement = htmlFile.addCollapsibleMenuItem(segment_config.segmentName + " Tests")

        if segment_config.java_conformance_test_results is None:
            raise RuntimeError('Tried to display java results for a segment with no java results')
        segment_conformant = JavaConformanceTest.JavaConformanceTestResult.PASS

        for s in segment_config.java_conformance_test_results.java_package_tree.root_node.children:
            if s.name == 'Interface Results':
                # JAVA OSS Test have a top level node named "Interface Results" that causes havoc to HTML output
                # So just ignore it
                continue
            segment_conformant = self.updateJavaConformanceResult(
                segment_conformant,
                self.addJavaPackageResultSummary(s, collapsibleElement, 3, htmlFile)
            )

        self.addRightArrow(htmlFile, menuElement)
        self.addJavaResultText(segment_conformant, menuElement, htmlFile)
        return segment_conformant


    def addDataModelResultSummary(self, dataModel, htmlFile):

        menuElement, collapsibleElement = htmlFile.addCollapsibleMenuItem("Data Model Tests")

        dataModelConformant = True
        for t in dataModel.testResults:
            if t.name == 'FACE XML Validation':
                continue # This isn't a valid test any more, but I still get a test result for it...
            dataModelConformant = t.score == dataModel.TEST_PASSED and dataModelConformant

            testDiv = htmlFile.addDiv(superElement = collapsibleElement)

            testMenu, testCollapsedElement = htmlFile.addCollapsibleMenuItem("Test: " + t.name,
                                                                               headingLevel=3, superElement=testDiv)
            self.addRightArrow(htmlFile, testMenu)

            if t.score == dataModel.TEST_PASSED:
                htmlFile.addSpan(" PASSED", superElement=testMenu, cssClass="passedTestResult")
            elif t.score == dataModel.TEST_FAILED:
                htmlFile.addSpan(" FAILED", superElement=testMenu, cssClass="failedTestResult")
            elif t.score == dataModel.TEST_NOT_PERFORMED:
                htmlFile.addSpan(" NOT PERFORMED", superElement=testMenu, cssClass="notPerformedTestResult")
            else:
                htmlFile.addSpan(" UNKNOWN RESULT", superElement=testMenu, cssClass="failedTestResult")

            htmlFile.addParagraph(t.description, superElement=testCollapsedElement)

            if t.score != dataModel.TEST_NOT_PERFORMED:
                self.addLogfile(htmlFile, 'Log File', dataModel.testLogFilenames[t], superElement=testCollapsedElement)

        self.addRightArrow(htmlFile, menuElement)
        if dataModelConformant:
            htmlFile.addSpan(" PASSED", superElement=menuElement, cssClass="passedTestResult")
        else:
            htmlFile.addSpan(" FAILED", superElement=menuElement, cssClass="failedTestResult")
        return dataModelConformant

    
    def addNoteOnWarningsToReport(self, reportFile, headLevel, owningElement):
        noteMenu, noteCollapsedElement = reportFile.addCollapsibleMenuItem("NOTE ABOUT TESTS MARKED \"INSPECT\"", headingLevel=headLevel, superElement=owningElement)
        noteDiv = reportFile.addDiv(superElement = noteCollapsedElement)
        noteText = "Tests marked \"INSPECT\" in this section contain code intended to produce compiler warnings (i.e. not errors) when a UoC does not conform to a specific part of the interface definition.   Each of these test's corresponding \"Test Log File\" should be examined to verify no compiler warnings are produced by the corresponding \"Test Code\"."
        noteText2 = "NOTE:  No warnings found in the \"Test Log File\" should be caused by the file name (i.e. test*.adb) referred to by the \"Test Code\".  Any such warnings should classify the test as failing. "
        reportFile.addParagraph(noteText, superElement=noteDiv)
        reportFile.addParagraph(noteText2, superElement=noteDiv)
    

    def addSegmentResultSummary(self, segmentConfig, htmlFile):
        if segmentConfig.segment == TestConfiguration.SegmentConfig.Segment.OSS:
            return self.addOsSegmentResultSummary(segmentConfig, htmlFile)

        if self.conformanceTest.baseConfig.language == TestConfiguration.Language.Java:
            return self.addJavaSegmentResultSummary(segmentConfig, htmlFile)

            
        menuElement, collapsibleElement = htmlFile.addCollapsibleMenuItem(segmentConfig.segmentName +  " Segment Tests")
        numErrors = 0
               
        forkTest = self.conformanceTest.baseConfig.osProfile == TestConfiguration.Profile.SafetyExt and \
            self.conformanceTest.baseConfig.language in [TestConfiguration.Language.C, TestConfiguration.Language.CPP]
        limitedCallTest = (segmentConfig.segment == TestConfiguration.SegmentConfig.Segment.PCS or
                              segmentConfig.segment == TestConfiguration.SegmentConfig.Segment.PSS)


        segmentConformant = True
        inspection_required = False
        
        ''' Certain Ada Tests can only be confirmed as non-conformant via inspection of compiler output for warnings '''
        if self.conformanceTest.baseConfig.language == TestConfiguration.Language.Ada:
            self.addNoteOnWarningsToReport(htmlFile, 3, collapsibleElement)

        for target in segmentConfig.targets:

            addReport = True
            logger.debug('target: ' + str(target))
            ''' Don't add report if target is a fork test, unless already shown conformance and fork is legal '''
            if (target.endswith('ForkConformanceTest')) and not (segmentConformant and forkTest):
                addReport = False

            if (target.endswith('LimitedCallTest')) and not (segmentConformant and limitedCallTest):
                addReport = False

            ''' Don't add report if target is register read callback unless register read callback is allowed '''

            if constants.getFACEStandardVersion() in [constants.FACE_STANDARD_VERSION_2_1, constants.FACE_STANDARD_VERSION_2_1_1]:
                if isinstance(segmentConfig, TestConfiguration.TsSegmentConfig) and \
                    segmentConfig.registerReadCallback == 0 and \
                    target in segmentConfig.testAssertion.keys() and \
                    segmentConfig.testAssertion[target].lower().find("register_callback") != -1:
                    addReport = False

            if addReport:
                self.addTestReport(target, segmentConfig, htmlFile, collapsibleElement)
        
                if target in segmentConfig.testResult and segmentConfig.testResult[target] != 0:
                    if not (target.endswith('ForkConformanceTest') or target.endswith('LimitedCallTest')):
                        segmentConformant = False
                    else:
                        inspection_required = True
                        
                if segmentConformant and self.conformanceTest.baseConfig.language == TestConfiguration.Language.Ada:
                    inspection_required = True
            
        self.addRightArrow(htmlFile, menuElement)
        if segmentConformant:
            if not inspection_required:
                htmlFile.addSpan(" PASSED", superElement = menuElement, cssClass = "passedTestResult")
            else:
                htmlFile.addSpan(" INSPECTION REQUIRED", superElement = menuElement, cssClass="forkPresentResult")

        else:
            htmlFile.addSpan(" FAILED", superElement = menuElement, cssClass="failedTestResult")
         
        return segmentConformant

    def addTestReport(self, target, segmentConfig, reportFile, superElement):

        if target in segmentConfig.testAssertion:
            assertion = segmentConfig.testAssertion[target]
        elif target.endswith('ForkConformanceTest'):
            assertion = 'Fork Call In Safety Profile Test'
        else:
            assertion = 'Unknown Test'
        testMenu, testCollapsedElement = reportFile.addCollapsibleMenuItem("Test: " + assertion,
                                                                           headingLevel=3, superElement=superElement)

        #heading = reportFile.addHeading(3, "Test: " + testName, superElement = superElement)
        self.addRightArrow(reportFile, testMenu)

        if target.endswith('ForkConformanceTest'):  # special case...test result may not be there
            if target in segmentConfig.testResult and segmentConfig.testResult[target] != 0:
                reportFile.addSpan(" Fork Present", superElement = testMenu, cssClass = "presentResult")
            else:
                reportFile.addSpan(" Fork Absent", superElement = testMenu, cssClass = "notPresentResult")

        elif target.endswith('LimitedCallTest'): # special case...test result may not be there
            if target in segmentConfig.testResult and segmentConfig.testResult[target] != 0:
                reportFile.addSpan(" Restricted Calls Present", superElement = testMenu, cssClass = "presentResult")
            else:
                reportFile.addSpan(" No restricted calls", superElement = testMenu, cssClass = "notPresentResult")
        else:
            if segmentConfig.testResult[target] != 0:
                reportFile.addSpan(" FAILED", superElement = testMenu, cssClass = "failedTestResult")
            else:
                need_inspect = False
                if target in segmentConfig.testCodeFile:
                    need_inspect = self.conditionCheckedByWarning(os.path.join(segmentConfig.conformanceTestDir, segmentConfig.testCodeFile[target]))

                if need_inspect == True:
                    reportFile.addSpan(" INSPECT", superElement=testMenu, cssClass="forkPresentResult")
                else:
                    reportFile.addSpan(" PASSED", superElement = testMenu, cssClass = "passedTestResult")

        testDiv = reportFile.addDiv(superElement = testCollapsedElement)
        if target in segmentConfig.testDetails:
            reportFile.addParagraph(segmentConfig.testDetails[target], superElement=testDiv)
        if target in segmentConfig.testCodeFile:
            self.addCodefile(reportFile, 'Test Code', os.path.join(segmentConfig.conformanceTestDir, segmentConfig.testCodeFile[target]), superElement = testDiv)
        if target in segmentConfig.targetLogfile:
            log = segmentConfig.targetLogfile[target]
            if type(log) == type(list()): # multiple logs for the test
                for l in log:
                    self.addLogfile(reportFile, os.path.basename(l), l, superElement = testDiv)
            else:  # must be one log for target
                self.addLogfile(reportFile, 'Test Log File', segmentConfig.targetLogfile[target], superElement = testDiv)


    def conditionCheckedByWarning(self, targetFile):
        if self.conformanceTest.baseConfig.language == TestConfiguration.Language.Ada:
            with open(targetFile, 'r') as file:
                data = file.read()
                if "The assignment on the following line will produce a warning if not conformant" in data:
                    return True
        
        return False


    def addOsSegmentResultSummary(self, segmentConfig, htmlFile):

        logger.info('Adding OS Segment Result Summary')

        if self.conformanceTest.baseConfig.language == TestConfiguration.Language.Java:
            return self.addJavaOSSegmentResultSummary(segmentConfig, htmlFile)
            #return self.addJavaSegmentResultSummary(segmentConfig, htmlFile)

        if self.conformanceTest.baseConfig.language == TestConfiguration.Language.Ada:
            return self.addAdaOSSegmentResultSummary(segmentConfig, htmlFile)

        segmentMenuElement, collapsibleElement = htmlFile.addCollapsibleMenuItem(segmentConfig.segmentName +  " Tests")
        self.addRightArrow(htmlFile, segmentMenuElement)

        numErrors = 0
        segmentConformant = True
        index = 3
        for subTestName, subTestTargets in segmentConfig.targets.iteritems():
            subTestElement, collapsibleSubTestElement = htmlFile.addCollapsibleMenuItem(subTestName + ' Conformance Tests',
                                                                    superElement=collapsibleElement, headingLevel=index)

            subTestConformant = True
            headerFiles = list()
            collapsibleHeaderElement = None
            headerElement = None
            headerPassed = True
            for target in sorted(subTestTargets):
                logger.debug('target: ' + str(target))
                targetSplit = target.split('/')

                includeTest = '/'.join(targetSplit[1:len(targetSplit)-2])
                if len(targetSplit) == 3: # Special case when there is only 3
                                      # profile/INCLUDE_DIR/testname
                    includeTest = targetSplit[1]

                if includeTest not in headerFiles:
                    # Finish last header test
                    if headerElement is not None:
                        self.addRightArrow(htmlFile, headerElement)
                        if headerPassed:
                            htmlFile.addSpan(" PASSED", superElement = headerElement, cssClass = "passedTestResult")
                        else:
                            htmlFile.addSpan(" FAILED", superElement = headerElement, cssClass = "failedTestResult")

                    headerFiles.append(includeTest)
                    headerPassed = True
                    headerElement, collapsibleHeaderElement = htmlFile.addCollapsibleMenuItem(includeTest + ' Tests',
                                                                    superElement=collapsibleSubTestElement, headingLevel=index + 1)

                self.addOSTestReport(subTestName, target, segmentConfig, htmlFile, collapsibleHeaderElement)
                if (segmentConfig.testResult[subTestName][target] != 0):
                    headerPassed = False
                    subTestConformant = False

            # Do the last header result
            if headerElement is not None:
                self.addRightArrow(htmlFile, headerElement)
                if headerPassed:
                    htmlFile.addSpan(" PASSED", superElement = headerElement, cssClass = "passedTestResult")
                else:
                    htmlFile.addSpan(" FAILED", superElement = headerElement, cssClass = "failedTestResult")


            self.addRightArrow(htmlFile, subTestElement)
            if subTestConformant:
                htmlFile.addSpan(" PASSED", superElement = subTestElement, cssClass = "passedTestResult")
            else:
                htmlFile.addSpan(" FAILED", superElement = subTestElement, cssClass = "failedTestResult")

            segmentConformant = segmentConformant and subTestConformant

        if segmentConformant:
            htmlFile.addSpan(" PASSED", superElement = segmentMenuElement, cssClass = "passedTestResult")
        else:
            htmlFile.addSpan(" FAILED", superElement = segmentMenuElement, cssClass = "failedTestResult")

        return segmentConformant

    def addAdaOSSegmentResultSummary(self, segmentConfig, htmlFile):
        segmentMenuElement, collapsibleElement = htmlFile.addCollapsibleMenuItem(segmentConfig.segmentName +  " Tests")
        self.addRightArrow(htmlFile, segmentMenuElement)

        numErrors = 0
        segmentConformant = True
        inspection_required = False
        index = 3
        for subTestName, subTestTargets in segmentConfig.targets.iteritems():
            subTestElement, collapsibleSubTestElement = htmlFile.addCollapsibleMenuItem(subTestName + ' Conformance Tests',
                                                                    superElement=collapsibleElement, headingLevel=index)


            if "ARINC" not in subTestName:
                self.addNoteOnWarningsToReport(htmlFile, 3, collapsibleSubTestElement)

            subTestConformant = True
            subTestRequiresInspect = False
            headerFiles = list()
            collapsibleHeaderElement = None
            headerElement = None
            headerPassed = True
            for target in sorted(subTestTargets):
                logger.debug('target: ' + str(target))
                targetSplit = target.split('/')

                includeTest = '/'.join(targetSplit[1:len(targetSplit)-2])
                if len(targetSplit) == 3: # Special case when there is only 3
                                      # profile/INCLUDE_DIR/testname
                    includeTest = targetSplit[1]

                if includeTest not in headerFiles:
                    # Finish last header test
                    if headerElement is not None:
                        self.addRightArrow(htmlFile, headerElement)
                        if headerPassed:
                            if "ARINC" not in subTestName: 
                                ''' ARINC 653 Tests are generated/look differently than other Ada interface Tests '''
                                htmlFile.addSpan(" INSPECTION REQUIRED", superElement = headerElement, cssClass = "forkPresentResult")
                                subTestRequiresInspect = True
                            else:
                                htmlFile.addSpan(" PASSED", superElement = headerElement, cssClass = "passedTestResult")
                        else:
                            htmlFile.addSpan(" FAILED", superElement = headerElement, cssClass = "failedTestResult")

                    headerFiles.append(includeTest)
                    headerPassed = True
                    
                    headerElement, collapsibleHeaderElement = htmlFile.addCollapsibleMenuItem(includeTest + ' Tests',
                                                                    superElement=collapsibleSubTestElement, headingLevel=index + 1)
                self.addOSTestReport(subTestName, target, segmentConfig, htmlFile, collapsibleHeaderElement)
                if (segmentConfig.testResult[subTestName][target] != 0):
                    headerPassed = False
                    subTestConformant = False

            # Do the last header result
            if headerElement is not None:
                self.addRightArrow(htmlFile, headerElement)
                if headerPassed:
                    if "ARINC" not in subTestName:
                        htmlFile.addSpan(" INSPECTION REQUIRED", superElement = headerElement, cssClass = "forkPresentResult")
                        subTestRequiresInspect = True
                    else:
                        htmlFile.addSpan(" PASSED", superElement = headerElement, cssClass = "passedTestResult")
                else:
                    htmlFile.addSpan(" FAILED", superElement = headerElement, cssClass = "failedTestResult")


            self.addRightArrow(htmlFile, subTestElement)
            if subTestConformant:
                if subTestRequiresInspect:
                    htmlFile.addSpan(" INSPECTION REQUIRED", superElement = subTestElement, cssClass = "forkPresentResult")
                else:
                    htmlFile.addSpan(" PASSED", superElement = subTestElement, cssClass = "passedTestResult")
            else:
                htmlFile.addSpan(" FAILED", superElement = subTestElement, cssClass = "failedTestResult")

            segmentConformant = segmentConformant and subTestConformant
            inspection_required = inspection_required or subTestRequiresInspect

        if segmentConformant:
            if not inspection_required:
                htmlFile.addSpan(" PASSED", superElement = segmentMenuElement, cssClass="passedTestResult")
            else:
                htmlFile.addSpan(" INSPECTION REQUIRED", superElement = segmentMenuElement, cssClass="forkPresentResult")
        else:
            htmlFile.addSpan(" FAILED", superElement = segmentMenuElement, cssClass = "failedTestResult")

        return segmentConformant


    def addOSTestReport(self, subTest, target, segmentConfig, reportFile, superElement):
        logger.info('subTest: ' + subTest + ' target: ' + target)
        if target in segmentConfig.testAssertions[subTest]:
            assertion = segmentConfig.testAssertions[subTest][target]
        else:
            assertion = 'Unknown Test'

        testMenu, testCollapsedElement = reportFile.addCollapsibleMenuItem(assertion,
                                                                           headingLevel=3, superElement=superElement)
        self.addRightArrow(reportFile, testMenu)

        if segmentConfig.testResult[subTest][target] != 0:
            reportFile.addSpan(" FAILED", superElement = testMenu, cssClass = "failedTestResult")
        else:
            if self.conformanceTest.baseConfig.language == TestConfiguration.Language.Ada and self.conditionCheckedByWarning(segmentConfig.testCodeFile[subTest][target]) == True:
                reportFile.addSpan(" INSPECT", superElement = testMenu, cssClass = "forkPresentResult")
            else:
                reportFile.addSpan(" PASSED", superElement = testMenu, cssClass = "passedTestResult")

        testDiv = reportFile.addDiv(superElement = testCollapsedElement)
        if target in segmentConfig.testDetails[subTest]:
            reportFile.addParagraph(segmentConfig.testDetails[subTest][target], superElement = testDiv)
        if target in segmentConfig.testCodeFile[subTest]:
            self.addCodefile(reportFile, 'Test Code', segmentConfig.testCodeFile[subTest][target], superElement = testDiv)
        if target in segmentConfig.targetLogfile[subTest]:
            log = segmentConfig.targetLogfile[subTest][target]
            if type(log) == type(list()): # multiple logs for the test
                for l in log:
                    self.addLogfile(reportFile, os.path.basename(l), l, superElement = testDiv)
            else:  # must be one log for target
                self.addLogfile(reportFile, 'Test Log File', segmentConfig.targetLogfile[subTest][target], superElement = testDiv)

    def addRightArrow(self, reportFile, superElement):
        return reportFile.addTag('em', u'\u2192', superElement = superElement)
  
  
    def addSegmentLogAndCodefiles(self, segmentConfig, htmlFile):
          
        htmlFile.addHeading(2, segmentConfig.segmentName +  " Segment Test Log files")

        if segmentConfig.segmentName == "OperatingSystem":
            self.addOSSegmentLogFiles(htmlFile)
            return
        
        divElement = htmlFile.addDiv()
                   
        for target in segmentConfig.targets:
            if target in segmentConfig.testCodeFile:
                self.addCodefile(htmlFile, 'testCode', os.path.join(segmentConfig.conformanceTestDir, segmentConfig.testCodeFile[target]), superElement = divElement)
            if target in segmentConfig.targetLogfile:
                if type(segmentConfig.targetLogfile[target]) == type(list()):
                    for log in segmentConfig.targetLogfile[target]:
                        self.addLogfile(htmlFile, log, superElement = divElement)
                else:
                    self.addLogfile(htmlFile, segmentConfig.targetLogfile[target], superElement = divElement)
                    
    def addSegmentLogfiles(self, segmentConfig, htmlFile):
          
        htmlFile.addHeading(2, segmentConfig.segmentName +  " Segment Test Log files")

        if segmentConfig.segmentName == "OperatingSystem":
            self.addOSSegmentLogFiles(htmlFile)
            return
        
        divElement = htmlFile.addDiv()
        
        if segmentConfig.logfiles is not None:            
            if len(segmentConfig.logfiles) > 0:
                logger.debug('logfiles type:' + str(type(segmentConfig.logfiles)))
                for logfile in sorted(segmentConfig.logfiles):
                    self.addLogfile(htmlFile, logfile, superElement = divElement)
        else:
            logger.error('Logfiles is None?')

    def addCodefile(self, htmlFile, title, codefile, **attributes):
        logger.info('codefile:' + codefile + ' attrDict: ' + str(attributes))
        
        menuElement, collapsibleElement = htmlFile.addCollapsibleMenuItem(title, headingLevel = 3, **attributes)
        
        self.logFileToMenuID[codefile] = htmlFile.menuIdLUT[menuElement]
        
        attributes['superElement']=collapsibleElement
        
        codefileToHTML(htmlFile, codefile, self.conformanceTest.baseConfig.language, **attributes)

    def addLogfile(self, htmlFile, title, logfile, **attributes):

        logger.info('logfile:' + logfile + ' attrDict: ' + str(attributes))
        menuElement, collapsibleElement = htmlFile.addCollapsibleMenuItem(title, headingLevel = 3, **attributes)

        self.logFileToMenuID[os.path.join(self.conformanceTest.baseConfig.logDirectory,logfile)] = htmlFile.menuIdLUT[menuElement]

        attributes['superElement']=collapsibleElement
        attributes['maxLineLength'] = 80

        codefileToHTML(htmlFile, os.path.join(self.conformanceTest.baseConfig.logDirectory, logfile),
                       self.conformanceTest.baseConfig.language, brush='text', **attributes)


    def addCodeFileViaStringList(self, htmlFile, codeFilename, codeFileStringList, **attributes):
        menuElement, collapsibleElement = htmlFile.addCollapsibleMenuItem(codeFilename, headingLevel = 3, **attributes)
        self.logFileToMenuID[codeFilename] = htmlFile.menuIdLUT[menuElement]
        attributes['superElement'] = collapsibleElement
        
        try:
            stringListToHTML(htmlFile, codeFileStringList, startingLineNumber='-1', **attributes)
        except Exception as x:
            logger.error('Could not add code file, ' + codeFilename + ': ' +  x.message + '\codeFileStringList:'+ str(codeFileStringList))


    def addOSSegmentLogFiles(self, htmlFile, **attributes):
        if self.OSLogs is None:
            self.parseOSLogSummaryFile()

        collapsibleElementLUT = dict()
        menuElementLUT = dict()
        #for logName, logLines in sorted(self.OSLogs.iteritems(), key=stringSplitByNumbers):
        for logName, logLines in sorted(self.OSLogs.iteritems()):
            splitLogName = logName.split('/')
            partialLogName = splitLogName[1]
            if partialLogName not in collapsibleElementLUT:
                menuElement, collapsibleElement = htmlFile.addCollapsibleMenuItem(partialLogName, headingLevel = 3, **attributes)
                collapsibleElementLUT[partialLogName] = collapsibleElement
                menuElementLUT[partialLogName] = menuElement
                        
            fullMenuID = htmlFile.menuIdLUT[menuElementLUT[partialLogName]] + '|' 
            for i in range(2, len(splitLogName)):
                lastName = partialLogName
                partialLogName += '/' + splitLogName[i]
                if partialLogName not in collapsibleElementLUT:
                    menuElement, collapsibleElement = htmlFile.addCollapsibleMenuItem(partialLogName, headingLevel = 2+i, superElement = collapsibleElementLUT[lastName])
                    collapsibleElementLUT[partialLogName] = collapsibleElement
                    menuElementLUT[partialLogName] = menuElement
                fullMenuID += htmlFile.menuIdLUT[menuElementLUT[partialLogName]] + '|'
                    
            self.logFileToMenuID[logName] = fullMenuID[:-1]
            print  'logfiletomenuid: ' +  self.logFileToMenuID[logName]   
            #self.logFileToMenuID[logName] = htmlFile.menuIdLUT[menuElement]
            #print 'fullMenuId: ' + fullMenuID + ' logfiletomenuid: ' +  self.logFileToMenuID[logName]   
            logString = ''
            for l in logLines:
                logString += l
            codeElement = addSourceCodeElement(htmlFile, logString, 'plain', 1, superElement=collapsibleElement)

    def parseOSLogSummaryFile(self):
        logFilename = os.path.join(self.conformanceTest.baseConfig.logDirectory, 'osConformanceTestResultLogs')
        
        logfile = safeOpen(logFilename, 'r')
        logLines = logfile.readlines()
        logHeader = logLines[0]
        self.OSLogs = dict()
        
        index = 1
        while index < len(logLines):
            line = logLines[index]
            print line
            
            if line == '-------------------\n':  # Separator between logfiles
                index += 1
                logName = logLines[index].rstrip()
                index += 1
                self.OSLogs[logName] = list()
                while index < len(logLines) and logLines[index] != '-------------------\n':
                    self.OSLogs[logName].append(logLines[index])
                    index += 1
            else:
                index += 1


    def writeCodeFiles(self):
        pass
    
        
    
