#!/usr/bin/env python
"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""

__author__ = 'david.fann@vanderbilt.edu'

# standard python imports
import copy
import gc
from glob import glob, iglob
import logging
from optparse import OptionParser
import os
from os.path import join, sep, isdir, isfile, pathsep, relpath
from shutil import copytree, copy
from pickle import load, UnpicklingError
import platform
from shutil import rmtree
import subprocess
import sys
from time import ctime

# conformance test suite specific imports
import commonTools
import constants
from ConformanceMain import getRootDir
from ConformanceReport import ConformanceReport
import Dialogs
import JavaConformanceTest
from TestConfiguration import BaseConfig, DatatypeConfig, SegmentConfig, TsSegmentConfig, IoSegmentConfig, \
    OsSegmentConfig
from SourceCodeToHtml import codefileToHTML
import SplashScreen
import TestConfiguration


logger = logging.getLogger(__name__)


def getDefaultConfigurationFile():
    platformName = ''
    if platform.system() == 'Windows':
        platformName = 'WINDOWS'
    else:
        platformName = 'LINUX'
    return join(getRootDir(), 'face_conformance_app') + sep + 'initialConfiguration' + platformName + '.cfg'


class ConformanceTest():
    ''' 
    ConformanceTest is the main class for the Conformance Test Suite.  
    
    This contains all configuration information as well as methods used to run the conformance test. 
    
    '''

    def __init__(self, **kw):

        logger.info('kw: ' + str(kw))
        self.NUM_POSIX_GOLD_LIBS = 16

        # Configuration information
        self.baseConfig = BaseConfig()
        self.datatypeConfig = DatatypeConfig()
        self.pcConfig = SegmentConfig(SegmentConfig.Segment.PCS)
        self.tsConfig = TsSegmentConfig()
        self.pssConfig = SegmentConfig(SegmentConfig.Segment.PSS)
        self.ioConfig = IoSegmentConfig()
        self.osConfig = OsSegmentConfig()

        self.configFile = None
        self.testPC = False
        self.testTS = False
        self.testPSS = False
        self.testIOS = False
        self.testOS = False
        self.testDataModel = False
        self.testCodeGenerated = False
        self.conformanceTestGUI = None
        self.goldLibrariesBuilt = list()
        self.startTime = 0

        self._splashScreen = None


    def addCompilerSpecificCode(self, htmlWriter, parentElement):

        allowedDefinitionDir = os.path.join(getRootDir(), 'compilerSpecific', self.baseConfig.language, 'allowedDefinitions')
        first_time = True
        for i in iglob(os.path.join(allowedDefinitionDir, '*')):
            if isfile(i) and \
                (i.endswith(TestConfiguration.Language.header_suffix[self.baseConfig.language]) or
                 i.endswith(TestConfiguration.Language.source_suffix[self.baseConfig.language])):

                if first_time:
                    menuElement, collapsibleElement = htmlWriter.addCollapsibleMenuItem(
                        "Compiler Specific Allowed Definitions files",
                        superElement=parentElement,
                        headingLevel=4)
                    first_time = False

                codefileToHTML(htmlWriter, i, self.baseConfig.language, superElement=collapsibleElement)

        systemLibraryDefinitionDir = os.path.join(getRootDir(), 'compilerSpecific', self.baseConfig.language, 'systemLibraryDefinitions')
        first_time = True
        for i in iglob(os.path.join(systemLibraryDefinitionDir, '*')):
            if isfile(i) and \
                (i.endswith(TestConfiguration.Language.header_suffix[self.baseConfig.language]) or
                 i.endswith(TestConfiguration.Language.source_suffix[self.baseConfig.language])):

                if first_time:
                    menuElement, collapsibleElement = htmlWriter.addCollapsibleMenuItem(
                        "Compiler Specific System Library Definition files",
                        superElement=parentElement,
                        headingLevel=4)
                    first_time = False

                codefileToHTML(htmlWriter, i, self.baseConfig.language, superElement=collapsibleElement)




    def toHTML(self, htmlWriter, parentElement):

        # configFile = kw.get('configFile', None)

        if self.configFile is not None:
            if parentElement is not None:
                htmlWriter.addHeading(3, 'Configuration File: ' + self.configFile, superElement=parentElement)
            else:
                htmlWriter.addHeading(3, 'Configuration File: ' + self.configFile)

        baseElement = self.baseConfig.toHTML(htmlWriter, parentElement)
        if self.testPC or self.testTS or self.testPSS:
            htmlWriter.addConfigElement('FACE Specified Type Interface',
                                        commonTools.trueFalseString(self.tsConfig.faceSpecifiedType),
                                        superElement=baseElement)
            htmlWriter.addConfigElement('UntypedInterface', commonTools.trueFalseString(self.tsConfig.untypedInterface),
                                        superElement=baseElement)
            htmlWriter.addConfigElement('Register Read Callback',
                                        commonTools.trueFalseString(self.tsConfig.registerReadCallback),
                                        superElement=baseElement)

        self.addCompilerSpecificCode(htmlWriter, baseElement)

        if self.testDataModel or self.testPC or self.testPSS or self.testTS:
            self.datatypeConfig.toHTML(htmlWriter, parentElement)
        if self.testPC:
            self.pcConfig.toHTML(htmlWriter, parentElement, self.baseConfig.language)
        if self.testTS:
            self.tsConfig.toHTML(htmlWriter, parentElement, self.baseConfig.language)
        if self.testPSS:
            self.pssConfig.toHTML(htmlWriter, parentElement, self.baseConfig.language)
        if self.testIOS:
            self.ioConfig.toHTML(htmlWriter, parentElement, self.baseConfig.language)
        if self.testOS:
            self.osConfig.toHTML(htmlWriter, parentElement, self.baseConfig.language)

    def __str__(self):
        result_string = ''
        for d in self.__dict__:
            if d != 'conformanceTestGUI':
                result_string += '[{0}]: [{1}]\n'.format(d, self.__dict__[d])
        return result_string

    def __eq__(self, other):

        if type(self) is not type(other):
            return False

        for d in self.__dict__:
            if d != 'conformanceTestGUI':
                if self.__dict__[d] != other.__dict__[d]:
                    pass
                    return False

        return True

        ''' OLD WAY...need to ignore conformanceTestGUI 
        if (type(self) is type(other)) and (self.__dict__ == other.__dict__)
        '''

    def __ne__(self, other):
        return not self == other

    def changeLogDirectory(self, newDir, overrideExceptions=False, exceptionList=list()):
        if not os.path.isdir(newDir):
            errorMsg = "Log Directory: " + newDir + " does not exist"
            if overrideExceptions:
                exceptionList.append(errorMsg)
                logger.warning(errorMsg)
            else:
                logger.error(errorMsg)
                raise IOError(errorMsg)

        self.baseConfig.logDirectory = newDir

    def checkNewProjectDirectory(self, newDir):
        exceptionMsgList = list()
        configuration = copy.deepcopy(self)
        configuration.changeProjectDirectory(newDir, True, exceptionMsgList)

        return exceptionMsgList

    def changeProjectDirectory(self, newDir, overrideExceptions=False, exceptionMsgList=list()):
        logger.debug('Change Project Directory Old Value: {0}'.format(self))

        if not os.path.isdir(newDir):
            errorMsg = "Project Directory: " + newDir + " does not exist"
            if overrideExceptions:
                exceptionMsgList.append(errorMsg)
            else:
                logger.error(errorMsg)
                raise IOError(errorMsg)

        oldDir = self.baseConfig.projectDirectory
        self.baseConfig.projectDirectory = newDir

        # update PCS Files
        self._updateDirectory(newDir, oldDir, self.pcConfig.libraries, overrideExceptions, exceptionMsgList,
                              'PCS Objects:\n')

        # update TSS Files/Dirs
        self._updateDirectory(newDir, oldDir, self.tsConfig.includePath, overrideExceptions, exceptionMsgList,
                              'TSS IncludePath:\n')
        if constants.getFACEStandardVersion() != constants.FACE_STANDARD_VERSION_2_0:
            self._updateDirectory(newDir, oldDir, self.tsConfig.headerFiles, overrideExceptions, exceptionMsgList,
                              'TSS Include Files:\n')
        self._updateDirectory(newDir, oldDir, self.tsConfig.libraries, overrideExceptions, exceptionMsgList,
                              'TSS Objects:\n')

        # update PSS Files
        self._updateDirectory(newDir, oldDir, self.pssConfig.libraries, overrideExceptions, exceptionMsgList,
                              'PSS Objects:\n')

        # update Data Model Files
        if self.datatypeConfig.dataModel is not None:
            dataModelList = [self.datatypeConfig.dataModel.sharedDataModelFilename,
                             self.datatypeConfig.dataModel.unitOfConformanceDataModelFilename]
            self._updateDirectory(newDir, oldDir, dataModelList, overrideExceptions, exceptionMsgList, 'Data Model:\n')
            self.datatypeConfig.dataModel.sharedDataModelFilename = dataModelList[0]
            self.datatypeConfig.dataModel.unitOfConformanceDataModelFilename = dataModelList[1]


        # update IOS Files/Dirs
        self._updateDirectory(newDir, oldDir, self.ioConfig.includePath, overrideExceptions, exceptionMsgList,
                              'IOS Include Path:\n')
        if constants.getFACEStandardVersion() != constants.FACE_STANDARD_VERSION_2_0:
            self._updateDirectory(newDir, oldDir, self.ioConfig.headerFiles, overrideExceptions, exceptionMsgList,
                              'IOS Include Files:\n')
        self._updateDirectory(newDir, oldDir, self.ioConfig.libraries, overrideExceptions, exceptionMsgList,
                              'IOS Objects:\n')

        # update OSS Files/Dirs
        for name, s in self.osConfig.subTests.iteritems():
            if name == self.osConfig.SubTests.ARINC_653:  # ARINC_653 can have header files
                self._updateDirectory(newDir, oldDir, s.headerFiles, overrideExceptions, exceptionMsgList,
                                      'OSS ' + name + 'Header Files:\n')
            self._updateDirectory(newDir, oldDir, s.includePath, overrideExceptions, exceptionMsgList,
                                  'OSS ' + name + 'Include Path:\n')
            self._updateDirectory(newDir, oldDir, s.objectFiles, overrideExceptions, exceptionMsgList,
                                  'OSS ' + name + 'Objects:\n')

        # update the log directory, if it is a subdirectory of the project directory
        if self.baseConfig.logDirectory is not None and self.baseConfig.logDirectory != '':
            relPath = os.path.relpath(self.baseConfig.logDirectory, oldDir)
            if not relPath.startswith('..'):
                self.changeLogDirectory(os.path.join(newDir, relPath), overrideExceptions, exceptionMsgList)

        logger.debug('Change Project Directory New Value: {0}'.format(self))



    def _updateDirectory(self, newDir, oldDir, pathList,
                         overrideExceptions=False, exceptionMsgList=list(), exceptionPrefix=''):
        logger.debug('NewDir: ' + newDir + ' OldDir: ' + oldDir + ' pathList: ' + ', '.join(pathList) +
                     'overrideExceptions: ' + commonTools.trueFalseString(overrideExceptions))
        if newDir == oldDir:
            return

        for i in range(0, len(pathList)):
            path = pathList[i]
            if path == '':
                continue

            if oldDir != '':
                relDir = os.path.relpath(path, oldDir)
                if relDir.startswith('..'):
                    errorMsg = exceptionPrefix + 'PathLocation:' + path + ' not in current project directory: ' + oldDir
                    if overrideExceptions:
                        exceptionMsgList.append(errorMsg)
                        if not os.path.isfile(pathList[i]) and not os.path.isdir(pathList[i]):
                            errorMsg = 'PathLocation: ' + pathList[i] + ' does not exist'
                            exceptionMsgList.append(errorMsg)
                        return  # Don't need to do anything else in this case
                    else:
                        logger.error(errorMsg)
                        raise IOError(errorMsg)

                pathList[i] = os.path.join(newDir, relDir)
                logger.debug('OldPath: ' + path + ' NewPath: ' + pathList[i])
                if not os.path.isfile(pathList[i]) and not os.path.isdir(pathList[i]):
                    errorMsg = exceptionPrefix + "New Project Directory does not contain: " + pathList[i]
                    if overrideExceptions:
                        exceptionMsgList.append(errorMsg)
                    else:
                        logger.error(errorMsg)
                        raise IOError(errorMsg)
            else:  # Setting a project directory...not changing anything, just checking to see if the files are in
                # the new project directory
                relDir = os.path.relpath(path, newDir)
                if relDir.startswith('..'):
                    errorMsg = exceptionPrefix + 'PathLocation:' + path + ' not in new project directory: ' + newDir
                    if overrideExceptions:
                        logger.warning(errorMsg)  # warning, because they can change it later, if they want (GUI)
                        exceptionMsgList.append(errorMsg)
                    else:
                        logger.error(errorMsg)  # error, since this is mandatory in this case
                        raise IOError(errorMsg)


    def compare(self, other):
        if self == other:
            logger.debug('configurations are the same')
        else:  # need to go through each category...maybe the configs are different, but not in a meaningful way
            logger.debug('Configuration Differences:')
            if self.baseConfig != other.baseConfig:
                logger.debug('base config different\nself:' + commonTools.toString(
                    self.baseConfig) + '\nother:' + commonTools.toString(other.baseConfig))
            if self.datatypeConfig != other.datatypeConfig:
                logger.debug('datatype config different\nself:' + commonTools.toString(
                    self.datatypeConfig) + '\nother:' + commonTools.toString(other.datatypeConfig))
            if self.pcConfig != other.pcConfig:
                logger.debug('pcConfig different\nself:' + commonTools.toString(
                    self.pcConfig) + '\nother:' + commonTools.toString(other.pcConfig))
            if self.tsConfig != other.tsConfig:
                logger.debug('tsConfig different')
            if self.pssConfig != other.pssConfig:
                logger.debug('pssConfig different')
            if self.ioConfig != other.ioConfig:
                logger.debug('ioConfig different')
            if self.osConfig != other.osConfig:
                logger.debug('osConfig different')
            if self.testPC != other.testPC:
                logger.debug('testPC different')
            if self.testTS != other.testTS:
                logger.debug('testTS different')
            if self.testPSS != other.testPSS:
                logger.debug('testPSS different')
            if self.testIOS != other.testIOS:
                logger.debug('testIOS different')
            if self.testOS != other.testOS:
                logger.debug('testOS different')
            if self.testCodeGenerated != other.testCodeGenerated:
                logger.debug('testCodeGenerated different')

            logger.debug('self:' + commonTools.toString(self) + '\nOther: ' + commonTools.toString(other))


    def verifyConfigurationDialog(self, **kw):
        ''' Displays all errors and warnings in the configuration, returns True only if there
            are no errors or warnings, or if there are only warnings and the user still wants
            to continue.  Returns False otherwise.
        '''
        self.verifyConfiguration()
        parent = kw.get('parent', None)
        numErrors = len(self.baseConfig.errors) + \
                    len(self.pcConfig.errors) + \
                    len(self.tsConfig.errors) + \
                    len(self.pssConfig.errors) + \
                    len(self.ioConfig.errors) + \
                    len(self.osConfig.errors) + \
                    len(self.datatypeConfig.errors)

        numWarnings = len(self.baseConfig.warnings) + \
                      len(self.pcConfig.warnings) + \
                      len(self.tsConfig.warnings) + \
                      len(self.pssConfig.warnings) + \
                      len(self.ioConfig.warnings) + \
                      len(self.osConfig.warnings) + \
                      len(self.datatypeConfig.warnings)

        dialogMsg = ''
        if numErrors > 0:
            dialogMsg += 'Configuration is incorrect, please fix the following '
            if numErrors == 1:
                dialogMsg += 'error\n'
            else:
                dialogMsg += str(numErrors) + ' errors\n'

            if len(self.baseConfig.errors) > 0:
                dialogMsg += 'Base Configuration Errors: \n' + '\n'.join(self.baseConfig.errors) + '\n'
            if len(self.pcConfig.errors) > 0:
                dialogMsg += 'Portable Components Configuration Errors: \n' + '\n'.join(self.pcConfig.errors) + '\n'
            if len(self.tsConfig.errors) > 0:
                dialogMsg += 'Transport Services Configuration Errors: \n' + '\n'.join(self.tsConfig.errors) + '\n'
            if len(self.pssConfig.errors) > 0:
                dialogMsg += 'Platform Specific Services Configuration Errors: \n' + '\n'.join(
                    self.pssConfig.errors) + '\n'
            if len(self.ioConfig.errors) > 0:
                dialogMsg += 'I/O Services Configuration Errors: \n' + '\n'.join(self.ioConfig.errors) + '\n'
            if len(self.osConfig.errors) > 0:
                dialogMsg += 'O/S Configuration Errors: \n' + '\n'.join(self.osConfig.errors) + '\n'
            if len(self.datatypeConfig.errors) > 0:
                dialogMsg += 'Data Model Configuration Errors: \n' + '\n'.join(self.datatypeConfig.errors) + '\n'

        if numWarnings > 0:
            dialogMsg += 'Configuration contains ' + str(numWarnings) + ' warning'
            if numWarnings > 1:
                dialogMsg += 's\n'
            else:
                dialogMsg += '\n'

            if len(self.baseConfig.warnings) > 0:
                dialogMsg += 'Base Configuration Warnings: \n' + '\n'.join(self.baseConfig.warnings) + '\n'
            if len(self.pcConfig.warnings) > 0:
                dialogMsg += 'Portable Components Configuration Warnings: \n' + '\n'.join(self.pcConfig.warnings) + '\n'
            if len(self.tsConfig.warnings) > 0:
                dialogMsg += 'Transport Services Configuration Warnings: \n' + '\n'.join(self.tsConfig.warnings) + '\n'
            if len(self.pssConfig.warnings) > 0:
                dialogMsg += 'Platform Specific Services Configuration Warnings: \n' + '\n'.join(
                    self.pssConfig.warnings) + '\n'
            if len(self.ioConfig.warnings) > 0:
                dialogMsg += 'I/O Services Configuration Warnings: \n' + '\n'.join(self.ioConfig.warnings) + '\n'
            if len(self.osConfig.warnings) > 0:
                dialogMsg += 'O/S Configuration Warnings: \n' + '\n'.join(self.osConfig.warnings) + '\n'
            if len(self.datatypeConfig.warnings) > 0:
                dialogMsg += 'Data Model Configuration Warnings: \n' + '\n'.join(self.datatypeConfig.warnings) + '\n'

        if numErrors > 0:
            Dialogs.ErrorMessageBox('Conformance Test Suite: Error in Configuration', dialogMsg, parent=parent)
            return False

        if numWarnings > 0:
            dialogMsg += '\nDo you wish to continue as is?'
            return Dialogs.YesNoMessageBox('Warning(s) in configuration', dialogMsg, parent=parent) == 'yes'

        return True

    def verifyConfiguration(self):
        ''' Makes sure that the configuration for the conformance test is valid (i.e. required fields are present, files referenced exist, etc) '''
        logger.info('')
        verified = self.baseConfig.verifyConfiguration()

        if self.baseConfig.language == TestConfiguration.Language.Java and not commonTools.verifyJavaAddonInstalled():
            err_message = 'Please download and install the Java add-on for your O/S before testing a Java app. ' + \
                          'See manual for more details'
            logger.error(err_message)
            verified = False
            self.baseConfig.errors.append(err_message)

        if self.testPC:
            verified = self.pcConfig.verifyConfiguration(self.baseConfig.language) and verified

        if self.testTS:
            verified = self.tsConfig.verifyConfiguration(self.baseConfig.language) and verified

        if self.testPSS:
            verified = self.pssConfig.verifyConfiguration(self.baseConfig.language) and verified

        if self.testIOS:
            verified = self.ioConfig.verifyConfiguration(self.baseConfig.language) and verified

        if self.testOS:
            verified = self.osConfig.verifyConfiguration(self.baseConfig.language) and verified

        if self.testTS or self.testPC or self.testPSS or \
                (self.datatypeConfig.dataModel is not None and \
                         (self.datatypeConfig.dataModel.unitOfConformanceDataModelFilename != '' or \
                                      self.datatypeConfig.dataModel.sharedDataModelFilename != '')):

            if constants.getFACEStandardVersion() == constants.FACE_STANDARD_VERSION_2_0 or \
                    self.testPC or self.testPSS or \
                    (self.testTS and self.tsConfig.interfaceUnderTest !=
                        TestConfiguration.TsSegmentConfig.InterfaceUnderTest.TYPE_ABSTRACTION_INTERFACE_UoP):
                # Assuming if a data model is specified it wants to be tested for conformance
                verified = self.datatypeConfig.verifyConfiguration(
                    self.testTS or self.testPC or self.testPSS) and verified

        return verified

    def data_model_testing_required(self):
        test_data_model = False

        if self.testDataModel:
            test_data_model = True
        elif self.testPC or self.testPSS or self.testTS:
            if constants.getFACEStandardVersion() in [constants.FACE_STANDARD_VERSION_2_1, constants.FACE_STANDARD_VERSION_2_1_1] and \
                self.tsConfig.interfaceUnderTest == TestConfiguration.TsSegmentConfig.InterfaceUnderTest.TYPE_ABSTRACTION_INTERFACE_UoP:
                    test_data_model = False
            else:
                test_data_model = True

        return test_data_model


    def runConformanceTest(self, **kw):
        ''' Main method to run the conformance test, keyword interactive flags whether to bring up messagebox if an error/warning occurs '''
        logger.info('kw: ' + str(kw))
        self.startTime = ctime()

        self.deleteTestResults()

        self.conformanceTestGUI = kw.get('conformanceTestGUI', None)

        self.initializeConformanceTest(**kw)

        if self.checkWithUserBeforeContinuing() == False: return False, None

        SplashScreen.show('\n\nPlease wait...running Conformance Tests')

        conformanceTestResult = True


        test_data_model = self.data_model_testing_required()
        if test_data_model:

            if commonTools.isInteractive() and constants.isSnapshot():
                # Helpful for when running tests on internal releases
                if Dialogs.YesNoMessageBox('Data Model Test?', 'You want to run Data Model test? ') == 'no':
                    test_data_model = False

            if test_data_model:
                SplashScreen.show("\n\nPlease wait...running Conformance Tests on Data Model")
                conformanceTestResult = self.conformanceTestDataModel() and conformanceTestResult

        if self.baseConfig.language == TestConfiguration.Language.Java:
            # Java is handled differently
            kw['profile'] = self.baseConfig.osProfile
            conformanceTestResult = JavaConformanceTest.runConformanceTest(self, **kw) and conformanceTestResult

        else:  # Not Java...use standard way...

            if self.buildGoldStandardLibraries(**kw) == False: return False, None

            if self.testIOS:
                conformanceTestResult = self.conformanceTestSegment(self.ioConfig) and conformanceTestResult

            if self.testPC:
                conformanceTestResult = self.conformanceTestSegment(self.pcConfig) and conformanceTestResult

            if self.testPSS:
                conformanceTestResult = self.conformanceTestSegment(self.pssConfig) and conformanceTestResult

            if self.testTS:
                conformanceTestResult = self.conformanceTestSegment(self.tsConfig) and conformanceTestResult

            if self.testOS:
                conformanceTestResult = self.conformanceTestSegment(self.osConfig) and conformanceTestResult

            '''
            TODO:  maybe have a different behavior if conformanceResult is false...right now returns false if the test
                   fails.  Maybe want to redo so that if returns false if there was an error in building test, not that the
                   test failed....not sure how easy it would be...maybe easier to throw an exception instead...
            '''

        SplashScreen.show('\n\nPlease wait...generating conformance report')

        conformanceReport = ConformanceReport(self)
        conformanceResult = conformanceReport.generateReport()
        conformanceFilename = conformanceReport.reportFilename

        conformanceReport.generate_junit_report()

        """ THIS MAY BE AN ISSUE...WE WILL SEE
        if self.baseConfig.language != TestConfiguration.Language.Java:
            if constants.getFACEStandardVersion() == constants.FACE_STANDARD_VERSION_2_0 or not self.testTS:
                conformanceReport.generate_junit_report()
        else:
            # TODO: Marian needs to add Java tests to the junit results and modify the TS for 2.1
            logger.warning('Currently do not generate Java junit results')
        """

        logger.info('Conformance Report returned: ' + commonTools.trueFalseString(conformanceResult))

        SplashScreen.hide()
        self.resetConformanceTest()

        return conformanceResult, conformanceFilename

    def initializeConformanceTest(self, **kw):

        SplashScreen.show('\n\nPlease wait, initializing conformance test')

        self.resetConformanceTest()  # Clear any results from previous test

        if self.baseConfig.logDirectory == '':
            self.baseConfig.logDirectory = getRootDir() + constants.LOG_DIR

        if not isdir(self.baseConfig.logDirectory):
            logger.debug('Log directory did not exist, creating: [' + self.baseConfig.logDirectory + ']')
            commonTools.mkdir_p(self.baseConfig.logDirectory)

        if not isdir(self.baseConfig.logDirectory):
            logger.info('Creating log directory: ' + self.baseConfig.logDirectory)
            commonTools.mkdir_p(self.baseConfig.logDirectory)

        # Scripts and styles contain javascript necessary to show source code in the html files
        if not isdir(self.baseConfig.logDirectory + sep + 'scripts'):
            copytree(getRootDir() + sep + 'scripts', self.baseConfig.logDirectory + sep + 'scripts')
        if not isdir(self.baseConfig.logDirectory + sep + 'styles'):
            copytree(getRootDir() + sep + 'styles', self.baseConfig.logDirectory + sep + 'styles')
        commonTools.copyFile(getRootDir() + sep + constants.APP_DIR + sep + 'ConformanceReport.css',
                             self.baseConfig.logDirectory)
        commonTools.copyFile(getRootDir() + sep + constants.APP_DIR + sep + 'ConformanceReport.js',
                             self.baseConfig.logDirectory)

        needDataModelGeneratedCode = self.testPSS or self.testPC or self.testTS

        if constants.getFACEStandardVersion() in [constants.FACE_STANDARD_VERSION_2_1, constants.FACE_STANDARD_VERSION_2_1_1]:
            # If type abstraction interface...no datatype specific code is needed
            needDataModelGeneratedCode = self.testPSS or self.testPC or (self.testTS and
                                                                         self.tsConfig.interfaceUnderTest != self.tsConfig.InterfaceUnderTest.TYPE_ABSTRACTION_INTERFACE_UoP)

        if needDataModelGeneratedCode:
            SplashScreen.show('\n\nPlease wait...Generating Data Model based code')
            self.generateDatatypeSpecificCode()

    def resetConformanceTest(self):
        ''' Needed so that a conformance test can be run multiple times without having duplicate errors/logs/etc '''
        logger.debug('Resetting conformance test')
        self.baseConfig.resetTest()
        self.tsConfig.resetTest()
        self.pcConfig.resetTest()
        self.pssConfig.resetTest()
        self.ioConfig.resetTest()
        self.osConfig.resetTest()
        gc.collect()  # force a garbage collection since I just deleted lots o stuff


    def testDataModelOnly(self):
        """
        Destructively sets the test flags so that only the data model is tested.  Used to test datamodel from command line
        :return:None
        """
        self.testIOS = False
        self.testOS = False
        self.testPC = False
        self.testPSS = False
        self.testTS = False
        self.testDataModel = True
        if self.datatypeConfig.dataModel is None:
            err_msg = "Tried to test data model, but dataModel was set to None"
            logger.error(err_msg)
            raise RuntimeError(err_msg)




    def priorTestExists(self):
        if not isdir(self.baseConfig.logDirectory):
            return False

        oldLogfiles = glob(self.baseConfig.logDirectory + sep + '*.log')
        if len(oldLogfiles) > 0:
            return True

        if isfile(join(self.baseConfig.logDirectory, 'ConformanceReport.html')):
            return True

        if isfile(join(self.baseConfig.logDirectory, 'osConformanceTestResultLogs')):
            return True

        return False

    def deleteTestResults(self):
        if not isdir(self.baseConfig.logDirectory):
            logger.error('Log directory: ' + str(self.baseConfig.logDirectory) + ' does not exist')
            return

        oldLogfiles = glob(self.baseConfig.logDirectory + sep + '*.log')
        for o in oldLogfiles:
            os.remove(o)

        htmlReport = join(self.baseConfig.logDirectory, 'ConformanceReport.html')
        if os.path.isfile(htmlReport):
            os.remove(htmlReport)


    def generateDatatypeSpecificCode(self):
        kw = dict()
        kw['goldLibraryCode'] = self.testPC or self.testPSS

        kw['conformanceTestCode'] = self.testTS

        if constants.getFACEStandardVersion() in [constants.FACE_STANDARD_VERSION_2_1, constants.FACE_STANDARD_VERSION_2_1_1]:
            kw['tssSubtest'] = self.tsConfig.interfaceUnderTest
            kw['conformanceTestCode'] = self.testTS and \
                self.tsConfig.interfaceUnderTest != self.tsConfig.InterfaceUnderTest.TYPE_ABSTRACTION_INTERFACE_UoP

        return self.datatypeConfig.initializeTest(self.baseConfig, **kw)


    def getLimitedGoldLibraryIncludePaths(self, segmentsUnderTest, relDir=None):
        requiredEnvPaths, allRequiredPaths = self.getRequiredGoldLibraryIncludePaths(segmentsUnderTest, relDir)

        limitedEnvPaths = list()
        for p in requiredEnvPaths:
            if p.find('ARINC') >= 0 or p.find('POSIX') >= 0:
                limitedEnvPaths.append(p + '_limited')
            else:
                limitedEnvPaths.append(p)

        limitedReqPaths = list()
        for p in allRequiredPaths:
            if p.find('ARINC') >= 0 or p.find('POSIX') >= 0:
                limitedReqPaths.append(p + '_limited')
            else:
                limitedReqPaths.append(p)

        return limitedEnvPaths, limitedReqPaths

    def getRequiredGoldLibraryIncludePaths(self, segmentsUnderTest, relDir=None):

        PCS = SegmentConfig.Segment.PCS in segmentsUnderTest
        PSS = SegmentConfig.Segment.PSS in segmentsUnderTest
        IOS = SegmentConfig.Segment.IOS in segmentsUnderTest
        TSS = SegmentConfig.Segment.TSS in segmentsUnderTest
        OSS = SegmentConfig.Segment.OSS in segmentsUnderTest  # OSS doesn't need any GSLs

        includeDir = 'include'
        if self.baseConfig.language == TestConfiguration.Language.Ada:
            includeDir = 'spec'

        goldLibRoot = self.baseConfig.getGoldLibDirectory()
        compilerSpecificDir = os.path.join(getRootDir(), constants.COMPILER_SPECIFIC_DIR, self.baseConfig.language)
        compilerSpecificTypesDir = os.path.join(compilerSpecificDir, constants.COMPILER_SPECIFIC_TYPES_DIR)
        compilerSpecificAllowedDir = os.path.join(compilerSpecificDir, constants.COMPILER_SPECIFIC_ALLOWED_DIR)

        requiredEnvPaths = dict()
        allRequiredPaths = list()

        # All tests need this, since it has the compiler specific type to FACE type conversion
        allRequiredPaths.append(compilerSpecificTypesDir)
        requiredEnvPaths['FACE_COMPILER_SPECIFIC_CFLAGS'] = \
            ' -I' + os.path.relpath(allRequiredPaths[-1], goldLibRoot).replace(sep, '/')

        needOSLibs = PCS or PSS or IOS or TSS

        if needOSLibs:
            if self.baseConfig.osPartition == BaseConfig.Partition.POSIX or \
                            self.baseConfig.osPartition == BaseConfig.Partition.POSIX_ARINC_653:
                if self.baseConfig.language == TestConfiguration.Language.CPP or \
                                self.baseConfig.language == TestConfiguration.Language.C:
                    allRequiredPaths.append(os.path.join(goldLibRoot, 'src', 'OSS', 'POSIX', self.baseConfig.osProfile))
                    requiredEnvPaths['FACE_POSIX_CFLAGS'] = \
                        ' -I' + os.path.relpath(allRequiredPaths[-1], goldLibRoot).replace(sep, '/')
                    if self.baseConfig.language == TestConfiguration.Language.CPP and self.baseConfig.osProfile == TestConfiguration.Profile.General:
                        allRequiredPaths.append(
                            os.path.join(goldLibRoot, 'src', 'OSS', 'libstdc++', self.baseConfig.osProfile))
                        requiredEnvPaths['FACE_POSIX_CFLAGS'] += ' -I' + os.path.relpath(allRequiredPaths[-1],
                                                                                         goldLibRoot).replace(sep, '/')
            '''
            if self.baseConfig.osPartition == BaseConfig.Partition.ARINC_653 or \
                            self.baseConfig.osPartition == BaseConfig.Partition.POSIX_ARINC_653:
                allRequiredPaths.append(os.path.join(goldLibRoot, 'src', 'OSS', 'ARINC-653', self.baseConfig.osProfile))
                requiredEnvPaths['FACE_ARINC653_CFLAGS'] = ' -I' + os.path.relpath(allRequiredPaths[-1],
                                                                                   goldLibRoot).replace(sep, '/')

            if self.baseConfig.osPartition == BaseConfig.Partition.POSIX:
                allRequiredPaths.append(os.path.join(goldLibRoot, 'src', 'OSS', 'ARINC-653', 'posix_allowed'))
                requiredEnvPaths['FACE_ARINC653_CFLAGS'] = ' -I' + os.path.relpath(allRequiredPaths[-1],
                                                                                   goldLibRoot).replace(sep, '/')
            '''

            if self.baseConfig.osPartition == BaseConfig.Partition.ARINC_653:
                allRequiredPaths.append(os.path.join(goldLibRoot, 'src', 'OSS', 'ARINC-653', self.baseConfig.osProfile))
                requiredEnvPaths['FACE_ARINC653_CFLAGS'] = ' -I' + os.path.relpath(allRequiredPaths[-1],
                                                                                   goldLibRoot).replace(sep, '/')

            if self.baseConfig.osPartition == BaseConfig.Partition.POSIX_ARINC_653:
                allRequiredPaths.append(os.path.join(goldLibRoot, 'src', 'OSS', 'ARINC-653', self.baseConfig.osProfile + '_posix_allowed'))
                requiredEnvPaths['FACE_ARINC653_CFLAGS'] = ' -I' + os.path.relpath(allRequiredPaths[-1],
                                                                                   goldLibRoot).replace(sep, '/')

            if self.baseConfig.language == TestConfiguration.Language.Ada:
                allRequiredPaths.append(os.path.join(goldLibRoot, includeDir, constants.getfaceOrFACE()))
            else:
                allRequiredPaths.append(os.path.join(goldLibRoot, includeDir))
            requiredEnvPaths['FACE_HMFM_CFLAGS'] = ' -I' + os.path.relpath(allRequiredPaths[-1], goldLibRoot).replace(
                sep, '/')

            if self.baseConfig.language == TestConfiguration.Language.C:
                allRequiredPaths.append(os.path.join(goldLibRoot, 'src', 'OSS',
                                                     TestConfiguration.Language.standardLib[self.baseConfig.language],
                                                     self.baseConfig.osProfile))
                requiredEnvPaths['FACE_CLIB_CFLAGS'] = \
                    ' -I' + os.path.relpath(allRequiredPaths[-1], goldLibRoot).replace(sep, '/')

                # DY: Added to fix HMFM GSL build error in ARINC partition for C language
                if self.baseConfig.osPartition == BaseConfig.Partition.ARINC_653:
                    requiredEnvPaths['FACE_HMFM_CFLAGS'] += ' -I' + os.path.relpath(allRequiredPaths[-1], goldLibRoot).replace(sep, '/')

            if self.baseConfig.language == TestConfiguration.Language.CPP:
                allRequiredPaths.append(os.path.join(goldLibRoot, 'src', 'OSS',
                                                     TestConfiguration.Language.standardLib[self.baseConfig.language],
                                                     self.baseConfig.osProfile))
                requiredEnvPaths['FACE_CPPLIB_CFLAGS'] = \
                    ' -I' + os.path.relpath(allRequiredPaths[-1], goldLibRoot).replace(sep, '/')

        needTSSLib = PCS or PSS
        needIOSLib = PSS
        needTSS_TypeAbstraction_UoP_Lib = constants.getFACEStandardVersion() in [constants.FACE_STANDARD_VERSION_2_1, constants.FACE_STANDARD_VERSION_2_1_1] and \
                                          self.tsConfig.interfaceUnderTest == TestConfiguration.TsSegmentConfig.InterfaceUnderTest.ABSTRACTION_LAYER

        # depending on standard, you have face/dm or FACE/DM
        face_or_FACE = constants.getfaceOrFACE()
        dm_or_DM = 'dm'
        if face_or_FACE == 'FACE':
            dm_or_DM = 'DM'

        if needTSSLib or needIOSLib or needTSS_TypeAbstraction_UoP_Lib:
            allRequiredPaths.append(os.path.join(goldLibRoot, includeDir, face_or_FACE))

        if needTSSLib and self.baseConfig.language != TestConfiguration.Language.Ada:
            allRequiredPaths.append(os.path.join(goldLibRoot, includeDir, face_or_FACE, dm_or_DM))
            if self.baseConfig.language == TestConfiguration.Language.CPP:
                allRequiredPaths.append(os.path.join(goldLibRoot, includeDir, constants.DM_CPP_HELPER_INCLUDE))

        if needTSSLib or needTSS_TypeAbstraction_UoP_Lib:
            allRelativePaths = list()
            for a in allRequiredPaths:
                allRelativePaths.append(os.path.relpath(a, goldLibRoot).replace(sep, '/'))

            if len(allRelativePaths) > 0:
                requiredEnvPaths['FACE_TSS_CFLAGS'] = commonTools.directoryListToIncludeOptionString(allRelativePaths)

        needIOSLib = PSS

        if needIOSLib:
            allRelativePaths = list()
            for a in allRequiredPaths:
                allRelativePaths.append(os.path.relpath(a, goldLibRoot).replace(sep, '/'))

            requiredEnvPaths['FACE_IOS_CFLAGS'] = commonTools.directoryListToIncludeOptionString(allRelativePaths)

        return requiredEnvPaths, allRequiredPaths

    def getAllPossiblyNeededGoldLibraryNames(self, segmentsUnderTest):

        limited_libraries = list()
        no_fork_libraries = list()

        required_libraries = self.getRequiredGoldLibraryNames(segmentsUnderTest)

        if (SegmentConfig.Segment.PCS in segmentsUnderTest) | (SegmentConfig.Segment.PSS in segmentsUnderTest):
            limited_libraries = self.getLimitedGoldLibraryNames(segmentsUnderTest, required_libraries)

        if self.baseConfig.osProfile == TestConfiguration.Profile.SafetyExt:
            no_fork_libraries = self.getForkGoldLibraryNames(segmentsUnderTest, required_libraries)

        return list(set(required_libraries) | set(limited_libraries) | set(no_fork_libraries))

    def getForkGoldLibraryNames(self, segmentsUnderTest, required_libraries=None):

        if required_libraries is None:
            required_libraries = self.getRequiredGoldLibraryNames(segmentsUnderTest)

        no_fork_libraries = list()
        for l in required_libraries:
            if l.startswith('posix'):
                no_fork_libraries.append(l.replace('GoldLib', 'NoForkGoldLib'))
            else:
                no_fork_libraries.append(l)
        return no_fork_libraries

    def getLimitedGoldLibraryNames(self, segmentsUnderTest, required_libraries=None):
        if required_libraries is None:
            required_libraries = self.getRequiredGoldLibraryNames(segmentsUnderTest)
        limited_libraries = list()
        for l in required_libraries:
            if l.startswith('arinc653') or l.startswith('posix'):
                limited_libraries.append(l.replace('GoldLib', 'LimitedGoldLib'))
            else:
                limited_libraries.append(l)
        return limited_libraries

    def getRequiredGoldLibraryNames(self, segmentsUnderTest):

        requiredGoldLibs = list()
        libraryEnding = 'GoldLib.a'
        PCS = SegmentConfig.Segment.PCS in segmentsUnderTest
        PSS = SegmentConfig.Segment.PSS in segmentsUnderTest
        IOS = SegmentConfig.Segment.IOS in segmentsUnderTest
        TSS = SegmentConfig.Segment.TSS in segmentsUnderTest
        OSS = SegmentConfig.Segment.OSS in segmentsUnderTest

        needOSLibs = PCS or PSS or IOS or TSS
        needCompilerSpecificLibrary = PCS or PSS or IOS or TSS
        if needCompilerSpecificLibrary:
            compilerSpecificSearchPattern = os.path.join(getRootDir(), constants.COMPILER_SPECIFIC_DIR,
                                                         self.baseConfig.language,
                                                         constants.COMPILER_SPECIFIC_ALLOWED_DIR,
                                                         '*' + self.baseConfig.getLanguageSuffix())

            compilerSpecificSourceFiles = glob(compilerSpecificSearchPattern)
            logger.debug('compilerSpecificSourceFiles: ' + str(compilerSpecificSourceFiles) +
                         ', Search pattern: ' + compilerSpecificSearchPattern)

            # Only need if there are compilerSpecificSourceFiles
            needCompilerSpecificLibrary = len(compilerSpecificSourceFiles) > 0

        if needOSLibs:
            osProfileCapitalized = self.baseConfig.osProfile[0].capitalize() + self.baseConfig.osProfile[1:]

            if self.baseConfig.osPartition == BaseConfig.Partition.POSIX or \
                            self.baseConfig.osPartition == BaseConfig.Partition.POSIX_ARINC_653:
                if self.baseConfig.language != TestConfiguration.Language.Ada:  # Need to handle Ada differently
                    requiredGoldLibs.append('posix' + osProfileCapitalized + libraryEnding)
                if needCompilerSpecificLibrary:
                    requiredGoldLibs.append('compilerSpecific' + libraryEnding)

            '''
            if self.baseConfig.osPartition == BaseConfig.Partition.ARINC_653 or \
                            self.baseConfig.osPartition == BaseConfig.Partition.POSIX_ARINC_653:
                requiredGoldLibs.append('arinc653' + osProfileCapitalized + libraryEnding)
            '''
            if self.baseConfig.osPartition == BaseConfig.Partition.ARINC_653:
                requiredGoldLibs.append('arinc653' + osProfileCapitalized + libraryEnding)
            elif self.baseConfig.osPartition == BaseConfig.Partition.POSIX_ARINC_653:
                requiredGoldLibs.append('arinc653' + osProfileCapitalized + 'PosixAllowed' + libraryEnding)


            # In ARINC 653 partition, you have access to libc in C++ (very special case)
            if self.baseConfig.osPartition == BaseConfig.Partition.ARINC_653 and \
                     self.baseConfig.language == TestConfiguration.Language.CPP:
                requiredGoldLibs.append(TestConfiguration.Language.standardLib[TestConfiguration.Language.C]
                                        + osProfileCapitalized + libraryEnding)

            '''
            if self.baseConfig.osPartition == BaseConfig.Partition.POSIX:
                requiredGoldLibs.append('arinc653PosixAllowed' + libraryEnding)
            '''

            if self.baseConfig.openGLConfiguration == TestConfiguration.OpenGLConfiguration.OpenGL_ES_with_EGL:
                requiredGoldLibs.append('openGL_EGL12_' + libraryEnding)
                requiredGoldLibs.append('openGL_ES20_' + libraryEnding)

            elif self.baseConfig.openGLConfiguration == TestConfiguration.OpenGLConfiguration.OpenGL_SC:
                requiredGoldLibs.append('openGL_SC101_' + libraryEnding)
                if constants.getFACEStandardVersion() != constants.FACE_STANDARD_VERSION_2_0:
                    requiredGoldLibs.append('openGL_EGL12_' + libraryEnding)

            elif self.baseConfig.openGLConfiguration == TestConfiguration.OpenGLConfiguration.OpenGL_with_GLX:
                requiredGoldLibs.append('openGL_EGL12_' + libraryEnding)

            requiredGoldLibs.append('hmfm' + libraryEnding)

            if self.baseConfig.language == TestConfiguration.Language.C or \
                            self.baseConfig.language == TestConfiguration.Language.CPP:
                requiredGoldLibs.append(TestConfiguration.Language.standardLib[self.baseConfig.language]
                                        + osProfileCapitalized + libraryEnding)


        needTSSLib = PCS or PSS
        if needTSSLib:
            if constants.getFACEStandardVersion() == constants.FACE_STANDARD_VERSION_2_0:
                requiredGoldLibs.append('tss' + libraryEnding)
            else:
                requiredGoldLibs.append('tsInterfaceTSSUoPGoldLib.a')

        if TSS and constants.getFACEStandardVersion() in [constants.FACE_STANDARD_VERSION_2_1, constants.FACE_STANDARD_VERSION_2_1_1] and \
                        self.tsConfig.interfaceUnderTest == \
                        TestConfiguration.TsSegmentConfig.InterfaceUnderTest.ABSTRACTION_LAYER:
            requiredGoldLibs.append('tsTypeAbstractionTSSUoPGoldLib.a')

        needIOSLib = PSS
        if needIOSLib:
            requiredGoldLibs.append('ios' + libraryEnding)

        if needCompilerSpecificLibrary:
            requiredGoldLibs.append('compilerSpecific' + libraryEnding)

        return requiredGoldLibs

    def buildGoldStandardLibraries(self, **kw):
        segmentsUnderTest = list()
        targetDirectory = kw.get('targetDir', None)
        if self.testPC:
            segmentsUnderTest.append(SegmentConfig.Segment.PCS)
        if self.testPSS:
            segmentsUnderTest.append(SegmentConfig.Segment.PSS)
        if self.testTS:
            segmentsUnderTest.append(SegmentConfig.Segment.TSS)
        if self.testIOS:
            segmentsUnderTest.append(SegmentConfig.Segment.IOS)
        if self.testOS:
            segmentsUnderTest.append(SegmentConfig.Segment.OSS)

        targets = self.getAllPossiblyNeededGoldLibraryNames(segmentsUnderTest)
        goldStandardPath = os.path.join(getRootDir(), constants.GOLD_LIB_ROOT, self.baseConfig.language)
        requiredEnvironPathVars, allRequiredPaths = self.getRequiredGoldLibraryIncludePaths(segmentsUnderTest,
                                                                                            relDir=goldStandardPath)

        buildResult = self.buildGoldStandardLibraryTargets(targets, goldStandardPath=goldStandardPath,
                                                           environVariables=requiredEnvironPathVars)

        if targetDirectory is not None:
            if not os.path.isdir(targetDirectory):
                errorMsg = 'Directory to save Gold Standard Libraries does not exist: ' + targetDirectory
                logger.error(errorMsg)
                if commonTools.isInteractive():
                    Dialogs.ErrorMessageBox(errorMsg)
                raise IOError(errorMsg)

            if len(self.goldLibrariesBuilt) == 0:
                warningMessage = 'No gold standard libraries built.  Either none required or an error occurred during build'
                logger.warning(warningMessage)
                if commonTools.isInteractive():
                    Dialogs.WarningMessageBox(constants.FACE_TESTSUITE_NAME, warningMessage)

            else:

                goldStandardPath = self.baseConfig.getGoldLibDirectory()
                compilerSpecificPath = self.baseConfig.getCompilerSpecificDirectory()
                for g in self.goldLibrariesBuilt:
                    if g.startswith('compilerSpecific'):
                        commonTools.copyFile(os.path.join(compilerSpecificPath, g), targetDirectory)
                    else:
                        commonTools.copyFile(os.path.join(goldStandardPath, g), targetDirectory)

                # Need to write a read me file as well.
                goldReadMeFile = os.path.join(targetDirectory, 'README.txt')
                readme = open(goldReadMeFile, 'w')
                readmeHeader = os.path.join(getRootDir(), 'datafiles', 'goldLibReadmeHeader.txt')

                readme.write('File: ' + goldReadMeFile + '\n')
                readme.write('Generated: ' + ctime() + '\n')
                readme.writelines(open(readmeHeader, 'r').readlines())

                readme.write('\nBuild Configuration used:')
                readme.write('\nO/S Partition: ' + self.baseConfig.osPartition)
                readme.write('\nO/S Profile: ' + self.baseConfig.osProfile)
                readme.write('\nLanguage: ' + self.baseConfig.language)
                readme.write('\ncompiler: ' + self.baseConfig.compiler)
                readme.write('\ncompiler flags: ' + self.baseConfig.compilerFlags)
                readme.write('\nprocessor specific compiler flags: ' + self.baseConfig.processorSpecificCompilerFlags)
                readme.write('\nlinker: ' + self.baseConfig.linker)
                readme.write('\nlinker flags: ' + self.baseConfig.linkerFlags)

                if self.baseConfig.language == TestConfiguration.Language.Ada:
                    readme.write(
                        '\nFor Ada Segments, POSIX libraries are also available, but will need to be compiled\n'
                        + 'separately, using the Test Suite configured for the same partition/profile, but\n'
                        + 'with the relevant C compiler.  Additionally, an approved Ada runtime library is\n'
                        + 'also allowed, but not currently being built in the test suite.\n')

                for s in segmentsUnderTest:
                    readme.write('\n\nGold Standard Libraries allowed for ' + s + ' segment: \n\t' + '\n\t'.join(
                        self.getRequiredGoldLibraryNames([s])) + '\n')
                    segmentEnvVars, segmentRequiredPaths = self.getRequiredGoldLibraryIncludePaths([s],
                                                                                                   relDir=goldStandardPath)
                    readme.write('\nInclude Directories allowed for ' + s + ' segment: \n\t' + '\n\t'.join(
                        segmentRequiredPaths) + '\n')
                    if s in [SegmentConfig.Segment.PCS, SegmentConfig.Segment.PSS]:
                        readme.write('\nLimited Libraries do not contain POSIX/ARINC 653 functions limited to TSS, IOS,' \
                                     + ' and within a single UoP listed in section E.4 of the FACE Standard\n')
                        readme.write(
                            '\n\nGold Standard Libraries allowed for a limited ' + s + ' segment: \n\t' + '\n\t'.join(
                                self.getLimitedGoldLibraryNames([s])) + '\n')

                        segmentEnvVars, segmentRequiredPaths = self.getLimitedGoldLibraryIncludePaths([s],
                                                                                                      relDir=goldStandardPath)
                        readme.write(
                            '\nInclude Directories allowed for a limited ' + s + ' segment: \n\t' + '\n\t'.join(
                                segmentRequiredPaths) + '\n')

                readme.close()

                if commonTools.isInteractive():
                    SplashScreen.hide()
                    commonTools.displayFile(goldReadMeFile)

        return buildResult


    def checkExactTypesHeader(self, goldStandardPath):
        '''
        Build a very simple example object that uses exact types header info, if it compiles, most likely it is correct
        '''
        makeCmd = ''
        fullName = ''

        if self.baseConfig.language == TestConfiguration.Language.C:
            objectName = 'src/OSS/POSIX/general/inttypes.' + self.baseConfig.objectSuffix
            fullName = os.path.join(goldStandardPath, objectName)
            makeCmd = self.baseConfig.make + ' -f Makefile.POSIX ' + objectName

        elif self.baseConfig.language == TestConfiguration.Language.CPP:
            objectName = 'src/OSS/libstdc++/general/new.' + self.baseConfig.objectSuffix

            fullName = os.path.join(goldStandardPath, objectName)
            makeCmd = self.baseConfig.make + ' -f Makefile.libstdc++ ' + objectName

        if os.path.isfile(fullName):
            os.remove(fullName)

        result = commonTools.subprocessWithSpecifiedPath(makeCmd, self.baseConfig.addedPathDirs, goldStandardPath)
        logfile = os.path.join(self.baseConfig.logDirectory, 'exactTypesCheck.log')
        info = {'working directory': goldStandardPath, 'path directories': pathsep.join(self.baseConfig.addedPathDirs)}
        commonTools.logSubprocessResults(makeCmd, result, logfile, otherInfo=info)

        if result.returncode != 0:
            compilerSpecificDir = os.path.join(getRootDir(), constants.COMPILER_SPECIFIC_DIR, self.baseConfig.language,
                                               constants.COMPILER_SPECIFIC_TYPES_DIR)
            errorMessage = 'Either build configuration is incorrect or ' \
                           + os.path.join(compilerSpecificDir, constants.EXACT_TYPES_HEADER) \
                           + ' has not been modified correctly.  See ' + logfile + ' for more information'
            # Do the subprocess and check for return code here
            if commonTools.isInteractive():
                Dialogs.ErrorMessageBox(constants.FACE_TESTSUITE_NAME, errorMessage, parent=self.conformanceTestGUI)
                commonTools.displayFile(logfile)
            else:
                raise RuntimeError(errorMessage)

        return (result.returncode == 0)

    def buildGoldStandardLibraryTargets(self, targets, **kw):
        if targets is None or len(targets) == 0:
            logger.warning('No Gold Standard Libraries to Build')
            return True

        logger.info('targets: ' + ', '.join(targets))
        goldStandardPath = kw.get('goldStandardPath', self.baseConfig.getGoldLibDirectory())
        compilerSpecificPath = kw.get('compilerSpecificPath',
                                      os.path.join(getRootDir(), constants.COMPILER_SPECIFIC_DIR,
                                                   self.baseConfig.language))
        environVars = kw.get('environVariables', None)

        # Write the Makefile.env for the gold libs
        makeEnvFilename = join(goldStandardPath, 'Makefile.env')
        compilerSpecificMakeEnvFilename = join(compilerSpecificPath, 'Makefile.env')
        makeEnv = open(makeEnvFilename, 'w')
        compilerSpecificMakeEnv = open(compilerSpecificMakeEnvFilename, 'w')
        self.baseConfig.writeMakefileEnvironment(makeEnv, environVariables=environVars, relativeDir=goldStandardPath,
                                                 useGoldLibPath=False)
        self.baseConfig.writeMakefileEnvironment(compilerSpecificMakeEnv, environVariables=environVars,
                                                 relativeDir=goldStandardPath, useGoldLibPath=False)

        makeEnv.close()
        compilerSpecificMakeEnv.close()

        if not self.checkExactTypesHeader(goldStandardPath):
            return False

        logger.info('cmd:' + self.baseConfig.make + ' addedPathDir: ' + pathsep.join(
            self.baseConfig.addedPathDirs) + ' goldPath: ' + goldStandardPath)
        result = commonTools.subprocessWithSpecifiedPath(self.baseConfig.make + ' clean', self.baseConfig.addedPathDirs,
                                                         goldStandardPath)
        logger.debug('make clean: returncode = ' + str(
            result.returncode) + '\nstdout: ' + result.stdout + '\nstderr: ' + result.stderr)

        # CR-763 Clean Compiler Specific Directory
        result = commonTools.subprocessWithSpecifiedPath(self.baseConfig.make + ' clean', self.baseConfig.addedPathDirs,
                                                         os.path.join(getRootDir(), constants.COMPILER_SPECIFIC_DIR, self.baseConfig.language)
                                                         )
        logger.debug('make clean of compilerSpecificGoldLib.a: returncode = ' + str(
            result.returncode) + '\nstdout: ' + result.stdout + '\nstderr: ' + result.stderr)
        # CR-763 End

        buildSuccess = True

        self.goldLibrariesBuilt = list()
        for t in targets:
            SplashScreen.show('\n\nPlease wait...building ' + t + ' gold standard library')

            makeTargetCommand = ''
            workingDirectory = ''
            if t.startswith('compilerSpecific'):
                makeTargetCommand = self.baseConfig.make
                workingDirectory = compilerSpecificPath
            else:
                makeTargetCommand = self.baseConfig.make + ' ' + t
                workingDirectory = goldStandardPath

            result = commonTools.subprocessWithSpecifiedPath(makeTargetCommand, self.baseConfig.addedPathDirs,
                                                             workingDirectory)

            logger.debug('make ' + t + ': returncode = ' + str(
                result.returncode) + '\nstdout: ' + result.stdout + '\nstderr: ' + result.stderr)
            logfile = os.path.join(self.baseConfig.logDirectory, t[:-2] + '.log')
            info = {'working directory': workingDirectory,
                    'path directories': pathsep.join(self.baseConfig.addedPathDirs)}
            commonTools.logSubprocessResults(makeTargetCommand, result, logfile, otherInfo=info)
            if result.returncode != 0:
                buildSuccess = False
                errorMsg = 'Error in building gold lib: ' + t
                logger.error(errorMsg + ' see logfile: ' + logfile)
                if commonTools.isInteractive():
                    Dialogs.ErrorMessageBox(constants.FACE_TESTSUITE_NAME,
                                            errorMsg + '\nPress Ok to see log file: ' + logfile,
                                            parent=self.conformanceTestGUI)
                    commonTools.displayFile(logfile)
                else:
                    raise RuntimeError(errorMsg)
            else:
                self.goldLibrariesBuilt.append(t)

        return buildSuccess

    def checkWithUserBeforeContinuing(self):
        if not commonTools.isInteractive(): return True

        if (self.testPC or self.testPSS or self.testTS or self.testIOS) and \
                not self.baseConfig.language == TestConfiguration.Language.Java:
            question = 'To test a PCS/IOS/TSS/PSS Segment for conformance, the source code must be compiled and ' \
                       + 'linked using the gold standard include and/or spec files, not default system includes/specs.  Are the object files ready to test?'
            continueTest = Dialogs.YesNoMessageBox('Conformance Test Suite', question,
                                                   parent=self.conformanceTestGUI) == 'yes'
            if not continueTest:
                return False

        if self.testOS:
            question = '\nTesting an operating system for FACE conformance in no way implies POSIX, ARINC 653, Java, or OpenGL certification\n'
            Dialogs.InfoMessageBox('Conformance Test Suite', question, parent=self.conformanceTestGUI) == 'yes'

        return True


    def conformanceTestDataModel(self):
        self.datatypeConfig.dataModel.runConformanceTest()

        self.datatypeConfig.dataModel.testLogFilenames = dict()
        for t in self.datatypeConfig.dataModel.testResults:
            if t.score != self.datatypeConfig.dataModel.TEST_NOT_PERFORMED:
                #logFilename = 'DMConformanceTest_SHDM_' + os.path.basename(
                #    self.datatypeConfig.dataModel.sharedDataModelFilename) \
                #              + '_UOCDM_' + os.path.basename(
                #    self.datatypeConfig.dataModel.unitOfConformanceDataModelFilename) + '_' \
                #              + t.name + '.log'
                logFilename = t.name + '.log'  # The other was more descriptive, but could get VERY long filenames that caused problems on Windows
                logFilename = logFilename.replace(' ', '_')
                logFilename = os.path.join(self.baseConfig.logDirectory, logFilename)
                logFile = commonTools.safeOpen(logFilename, 'w', 'utf-8')
                logFile.writelines(t.log)
                logFile.close()
                self.datatypeConfig.dataModel.testLogFilenames[t] = logFilename

    def conformanceTestOSS(self, segment, **params):
        assert isinstance(segment, OsSegmentConfig)
        logger.debug('Testing an OS segment')

        segment.testResult = dict()
        execSuffix = self.baseConfig.execSuffix
        if not execSuffix.startswith('.') and execSuffix != '':
            execSuffix = '.' + execSuffix
        for subTestName, subTestTargets in segment.targets.iteritems():
            logger.info(
                'OSS SubTest: ' + subTestName + ', conformanceTestDir: ' + segment.conformanceSubTestDirs[subTestName])

            SplashScreen.show('\n\nPlease wait...testing ' + subTestName)
            SplashScreen.show_progress(len(subTestTargets), 5)

            conformancePassed = True
            segment.testResult[subTestName] = dict()
            for target in subTestTargets:
                SplashScreen.tick_progress()

                makeCmdLine = self.baseConfig.make + ' ' + target + execSuffix
                logger.debug('makeCmdLine: ' + makeCmdLine)
                result = commonTools.subprocessWithSpecifiedPath(makeCmdLine, self.baseConfig.addedPathDirs,
                                                                 segment.conformanceSubTestDirs[subTestName])
                logFilename = segment.targetLogfile[subTestName][target]
                logger.debug('logfilename: ' + logFilename)
                commonTools.logSubprocessResults(makeCmdLine, result, logFilename, otherinfo={
                    'working directory': segment.conformanceSubTestDirs[subTestName]})
                segment.testResult[subTestName][target] = result.returncode
                if result.returncode != 0:
                    errorMsg = 'Make returned with error when building ' + target + ' during the ' + segment.segmentName + ' conformance test: '
                    if result.stderr is not None:
                        errorMsg += ' Stderr: ' + result.stderr
                    if result.stdout is not None:
                        errorMsg += ' Std Out: ' + result.stdout
                    logger.debug(errorMsg)
                    conformancePassed = False
                else:
                    logger.info(target + ' built successfully during the ' + segment.segmentName + ' conformance test')

        return conformancePassed

    def conformanceTestSegment(self, segment, **params):

        logger.info('Testing: ' + segment.segmentName)

        conformanceTestDir = self.getConformanceTestDirectory(segment)

        environDict = dict()
        # if self.baseConfig.osProfile == TestConfiguration.Profile.SafetyExt:
        #    environDict['FACE_FORK_LIBS'] = getObjectNamesStringFromList(self.getLinkableLibraries(segment.segment, True))

        requiredGoldLibNames = self.getRequiredGoldLibraryNames([segment.segment])
        goldPath = os.path.relpath(self.baseConfig.getGoldLibDirectory(), conformanceTestDir)

        compilerSpecificPath = os.path.relpath(self.baseConfig.getCompilerSpecificDirectory(), conformanceTestDir)

        fullPathGoldLibNames = list()
        for g in requiredGoldLibNames:
            if g.startswith('compilerSpecific'):
                fullPathGoldLibNames.append(os.path.join(compilerSpecificPath, g))
            else:
                fullPathGoldLibNames.append(os.path.join(goldPath, g))

        forkTest = self.baseConfig.osProfile == TestConfiguration.Profile.SafetyExt
        limitedCallTest = (segment.segment == TestConfiguration.SegmentConfig.Segment.PCS or
                           segment.segment == TestConfiguration.SegmentConfig.Segment.PSS)

        if limitedCallTest:
            limitedGoldLibNames = self.getLimitedGoldLibraryNames([segment.segment], requiredGoldLibNames)
            fullPathLimitedGoldLibNames = list()
            for g in limitedGoldLibNames:
                if g.startswith('compilerSpecific'):
                    fullPathLimitedGoldLibNames.append(os.path.join(compilerSpecificPath, g))
                else:
                    fullPathLimitedGoldLibNames.append(os.path.join(goldPath, g))

            environDict['FACE_LIMITED_LIBS'] = commonTools.getObjectNamesStringFromList(fullPathLimitedGoldLibNames)

        if forkTest:
            forkGoldLibNames = self.getForkGoldLibraryNames([segment.segment], requiredGoldLibNames)
            fullPathNoForkGoldLibNames = list()
            for g in forkGoldLibNames:
                if g.startswith('compilerSpecific'):
                    fullPathNoForkGoldLibNames.append(os.path.join(compilerSpecificPath, g))
                else:
                    fullPathNoForkGoldLibNames.append(os.path.join(goldPath, g))

            environDict['FACE_FORK_LIBS'] = commonTools.getObjectNamesStringFromList(fullPathNoForkGoldLibNames)

        goldEnvironPathVars, goldLibPaths = self.getRequiredGoldLibraryIncludePaths([segment.segment],
                                                                                    conformanceTestDir)

        segment.generateConformanceTest(self.baseConfig, fullPathGoldLibNames, goldLibPaths, conformanceTestDir,
                                        environDict)

        if segment.segment == SegmentConfig.Segment.OSS:
            return self.conformanceTestOSS(segment, **params)

        ''' Handle the non-OS cases '''
        segmentPassedConformance = True

        execSuffix = self.baseConfig.execSuffix
        if not execSuffix.startswith('.') and execSuffix != '':
            execSuffix = '.' + execSuffix

        SplashScreen.show('\n\nPlease wait...testing ' + segment.segmentName)
        SplashScreen.show_progress(len(segment.targets), 5)

        for target in segment.targets:
            SplashScreen.tick_progress()

            specialTest = False  # Not a limited or fork test

            if target.endswith('LimitedCallTest'):
                specialTest = True
                if not (limitedCallTest and segmentPassedConformance):
                    continue  # don't do limited call test unless you pass the original test

            elif target.endswith('ForkConformanceTest'):
                specialTest = True
                if not (forkTest and segmentPassedConformance):
                    continue  # don't do fork test unless you want a forkTest and conformance passed for all the other tests

            makeCmdLine = self.baseConfig.make + ' ' + target + execSuffix
            logger.debug('makeCmdLine: ' + makeCmdLine)
            result = commonTools.subprocessWithSpecifiedPath(makeCmdLine, self.baseConfig.addedPathDirs,
                                                             conformanceTestDir)
            logFilename = relpath(conformanceTestDir, getRootDir())
            logFilename = '_'.join(logFilename.split(sep)) + '_' + target + '.log'
            logger.debug('logfilename: ' + logFilename)
            commonTools.logSubprocessResults(makeCmdLine, result, join(self.baseConfig.logDirectory, logFilename),
                                             otherInfo={'working directory': conformanceTestDir})
            segment.targetLogfile[target] = join(self.baseConfig.logDirectory, logFilename)
            segment.testResult[target] = result.returncode
            if result.returncode != 0:
                errorMsg = 'Make returned with error when building ' + target + ' during the ' + segment.segmentName + ' conformance test: '
                if result.stderr is not None:
                    errorMsg += ' Stderr: ' + result.stderr
                if result.stdout is not None:
                    errorMsg += ' Std Out: ' + result.stdout
                logger.error(errorMsg)
                if not specialTest:
                    segmentPassedConformance = False
            else:
                logger.info(target + ' built successfully during the ' + segment.segmentName + ' conformance test')

        return True

    def getConformanceTestDirectory(self, segment):

        conformanceTestDirectory = os.path.join(getRootDir(),
                                                constants.CONFORMANCE_TEST_ROOT,
                                                self.baseConfig.language,
                                                segment.segment)

        if constants.getFACEStandardVersion() in [constants.FACE_STANDARD_VERSION_2_1, constants.FACE_STANDARD_VERSION_2_1_1] and \
                isinstance(segment, TestConfiguration.TsSegmentConfig):
            conformanceTestDirectory = os.path.join(conformanceTestDirectory,
                                                    TestConfiguration.TsSegmentConfig.InterfaceUnderTest.interface_test_subdirectory[
                                                        segment.interfaceUnderTest])

        if not os.path.isdir(conformanceTestDirectory):
            errMessage = 'conformanceTestDirectory: ' + conformanceTestDirectory + ' does not exist'
            logger.error(errMessage)
            if commonTools.isInteractive():
                Dialogs.ErrorMessageBox(constants.FACE_TESTSUITE_NAME, errMessage, parent=self.conformanceTestGUI)
            raise IOError(errMessage)

        return conformanceTestDirectory


def checkConfig(correctConfig, configUnderTest):
    correct_config_vars = vars(correctConfig)
    test_config_vars = vars(configUnderTest)

    missingVars = list()
    for var_name, var_value in correct_config_vars.iteritems():
        logger.debug('Checking ' + var_name)
        if var_name not in test_config_vars.keys():
            missingVars.append(var_name)
        else:
            try:
                if var_value.__dict__ > 0:
                    missingVars.extend(checkConfig(var_value, test_config_vars[var_name]))

            except:
                pass  # if no __dict__ then don't have to continue...just assign the values
                correct_config_vars[var_name] = test_config_vars[var_name]

    return missingVars


def getConformanceConfiguration(configFilename):
    if not isfile(configFilename):
        rootDir = getRootDir()
        if isfile(join(rootDir, configFilename)):
            configFilename = join(rootDir, configFilename)
        elif isfile(join(join(rootDir, 'config'), configFilename)):
            configFilename = join(join(rootDir, 'config'), configFilename)
        else:
            errMsg = 'Configuration file: [' + configFilename + '] does not exist'
            logger.error(errMsg)
            raise IOError(errMsg)

    try:
        configFile = open(configFilename, 'r')
        loadedConformanceTest = load(configFile)
        conformanceTest = ConformanceTest()
        missingVars = checkConfig(conformanceTest, loadedConformanceTest)
        if len(missingVars) > 0:
            warningMessage = 'You are reading a configuration file from a previous version of the test suite ' + \
                             'with missing values: \n' + '\n'.join(
                missingVars) + '\nYou will want to save the configuration ' + \
                             'file before continuing in the new configuration file format'
            logger.warning(warningMessage)
            if commonTools.isInteractive():
                Dialogs.WarningMessageBox(constants.FACE_TESTSUITE_NAME, warningMessage)

        return conformanceTest

    except IOError as (errno, strerror):
        errMsg = 'I/O error({0}): {1}'.format(errno, strerror)
        logger.error(errMsg)
        parser.print_help()
        raise IOError(errMsg)

    except UnpicklingError as u:
        logger.error('Error in loading configuration file: ' + configFilename)
        parser.print_help()
        raise u

    except Exception as e:
        logger.error('Unexpected error in loading configuration file: ' + configFilename + str(sys.exc_info()[0]))
        parser.print_help()
        raise e


def NoninteractiveConformanceTest(configFiles):
    commonTools.setInteractive(False)
    for configFilename in configFiles:
        conformanceTest = getConformanceConfiguration(configFilename)
        conformanceTest.runConformanceTest(interactive=False)


if __name__ == '__main__':

    parser = OptionParser(usage="usage: %prog config_file1 config_file2 ...")
    parser.description = 'Runs conformance test non-interactively (no GUI).  All tests in configuration are run.' \
                         + ' Multiple configuration files may be run'



    (options, args) = parser.parse_args()


    if len(args) == 0:
        print parser.print_help()
        sys.exit(0)

    print str(args)

    import ConformanceMain

    ConformanceMain.initializeLogger()
    logger = logging.getLogger(__name__)
    logger.info('In ConformanceTest main')

    NoninteractiveConformanceTest(args)

    '''
    import Tkinter

    ROOT = Tkinter.Tk()
    ROOT.withdraw()
 
    ps = ProgressSplash(None, text='Please wait...building gold standard libraries')
    ps.after(1000, ps.destroy)
    
    ROOT.mainloop()
    '''
    ''' Run ConformanceTest from command line '''
    if NoninteractiveConformanceTest() == True:
        sys.exit(0)
    else:
        sys.exit(1)
