#!/usr/bin/env python
"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""


__author__ = 'david.fann@vanderbilt.edu'

import ConformanceMain

if __name__ == '__main__':
    ConformanceMain.initializeLogger()

import ConformanceTest
import commonTools
from ConfigurationGUI import ConfigurationGUI
import SplashScreen
import gc
import constants
import logging
import pickle
from Tkinter import *
from os.path import isfile, join, sep, isdir
import os.path
import Dialogs
import tkFont
from optparse import OptionParser
import TestConfiguration
import subprocess


logger = logging.getLogger(__name__)

class ConformanceTestGUI(Frame):
    '''
    classdocs
    '''

    def __init__(self, **params):
        '''
        Constructor
        '''

        logger.info('params: ' + str(params))    
        master = params.get('master', None)
        Frame.__init__(self, master)

        commonTools.pushParentGUI(self)
        self.rootDir = params.get('rootDir', ConformanceMain.getRootDir())
        self._storedTestOptions = None
        self.configuration = None 
        self.configurationFilename = params.get('configFile', ConformanceTest.getDefaultConfigurationFile())
        self.configSaved = True
        self.loadConfiguration()
             
        self.createWidgets()
        self.grid()            
        self.x_offset = 150
        self.y_offset = 150
        geometry_text = "+%d+%d" % (self.x_offset, self.y_offset)
        self.master.geometry(geometry_text)

        commonTools.setInteractive(True)
        SplashScreen.FACE_ReleaseSplashScreen(os.path.join(ConformanceMain.getRootDir(), constants.TOOLS_MARKING_XML_FILE), parent=self, x_offset=self.x_offset+50, y_offset=self.y_offset+450)


    def loadConfiguration(self):
        logger.info('configurationFile: ' + str(self.configurationFilename))
        
        defaultConfig = False
        if self.configurationFilename == ConformanceTest.getDefaultConfigurationFile():
            defaultConfig = True
            
        if self.configurationFilename is None:
            return False
                    
        if not isfile(self.configurationFilename):
            if isfile(join(self.rootDir + sep + 'config' + sep, self.configurationFilename)):
                self.configurationFilename = join(self.rootDir + sep + 'config' + sep, self.configurationFilename)
            elif isfile(join(self.rootDir, self.configurationFilename)):
                self.configurationFilename = join(self.rootDir, self.configurationFilename)
            else:
                self.configurationFilename = None
                return False

        try:
            self.configuration = pickle.load( open(self.configurationFilename, "r"))
            self.configuration.configFile = self.configurationFilename # may have changed location, since default is from development machine it was checked in on.
        except Exception as x:
            logger.error('Could not load configuration file: [' + self.configurationFilename + ']:' + x.message)
            return False
            
        self.configuration.baseConfig.checkMakeValue()
        #self.configuration.checkPosixCode()
        
        return True
    
    def createWidgets(self):
        ''' Creates main menu widgets '''
        logger.info('')
        titleFont = tkFont.Font(family="Helvetica", size=25, weight="bold")
        subTitleFont = tkFont.Font(family="Helvetica", size=12)
        buttonFont = tkFont.Font(family="Helvetica", size=20)
        testSuiteInfo = constants.getVersionedFACETestSuiteName()
        self.titleLabel = Label( self, text=testSuiteInfo.name, font=titleFont )
        self.titleLabel.grid()
        subTitle = Label(self, text='Supporting FACE Standard ' + testSuiteInfo.faceStandardVersion,
                         font=subTitleFont)
        subTitle.grid()
        subTitle = Label(self, text='Tool Version: ' + testSuiteInfo.toolVersion)
        subTitle.grid()

        self.configuration_label = Label(self, text='No configuration file selected')
        self.configuration_label.grid()

        self.buttons = list()
        self.buttons.append(Button (self, text='Configure Test Suite', font=buttonFont, command=self.configurationMenu))
        
        self.testAllButton = Button (self, text='Test All Segments', font=buttonFont, command=self.testAll)
        self.buttons.append(self.testAllButton)

        self.dataModelButton = Button(self, text='Test Data Model', font=buttonFont, command=self.testDataModel)
        self.buttons.append(self.dataModelButton)

        self.pcsButton = Button(self, text='Test Portable Components Segment', font=buttonFont, command=self.testPortableComponents)
        self.buttons.append(self.pcsButton)
        
        self.tssButton = Button (self, text='Test Transport Services Segment', font=buttonFont, command=self.testTransportServices)
        self.buttons.append(self.tssButton)
        
        self.pssButton = Button (self, text='Test Platform Specific Services Segment', font=buttonFont, command=self.testPlatformSpecificServices)
        self.buttons.append(self.pssButton)
        
        self.iosButton = Button (self, text='Test I/O Services Segment', font=buttonFont, command=self.testIOServices)
        self.buttons.append(self.iosButton)
        
        self.osButton = Button (self, text='Test Operating System Segment', font=buttonFont, command=self.testOperatingSystem)
        self.buttons.append(self.osButton)

        self.generateGoldLibrariesButton = Button (self, text='Generate Gold Standard Libraries', font=buttonFont, command=self.generateGoldLibraries)
        self.buttons.append(self.generateGoldLibrariesButton)

        self.cleanupButton = Button(self, text='Delete Prior Test Results', font=buttonFont, command=self.cleanupTestResults)
        self.buttons.append(self.cleanupButton)
        
        self.buttons.append(Button ( self, text='Quit', font=buttonFont,  command=self.quitTest ) )       
        
        maxLength = 0
        for b in self.buttons:
            if len(b.cget ( "text" )) > maxLength:
                maxLength = len(b.cget ( "text" ))
        
        for b in self.buttons:
            b.configure(width = maxLength)
        
        for b in self.buttons:
            b.grid()
        
        self.updateMainMenu()

    def configurationMenu(self):
        ''' calls the configuration window '''
            
        logger.info('config: ' + commonTools.toString(self.configuration))
        cw = ConfigurationGUI(self, config=self.configuration, configFile = self.configurationFilename)
        if cw.config != None:  # Actually updated the configuration
            self.configuration = cw.config
            self.configurationFilename = cw.configurationFilename
            self.updateMainMenu()
        
     
    def updateMainMenu(self):
        ''' called when returning from configuration menu to update main menu based on new configuration '''
        logger.info('')
        if self.configuration is None:
            self.generateGoldLibrariesButton.configure(state=DISABLED)
            self.testAllButton.configure(state=DISABLED)
            self.pcsButton.configure(state=DISABLED)
            self.tssButton.configure(state=DISABLED)
            self.pssButton.configure(state=DISABLED)
            self.iosButton.configure(state=DISABLED)
            self.osButton.configure(state=DISABLED)
            self.configuration_label.configure(text='No configuration file selected')
        else:
            testingSomething = False
            self.configuration.conformanceTestGUI = self
            if self.configurationFilename.find('initialConfiguration') == -1:
                configText = 'Configuration file: ' + os.path.basename(self.configurationFilename)
                if not self.configSaved:
                    configText += '*'
                self.configuration_label.configure(text=configText)

            if self.configuration.testPC:
                self.pcsButton.configure(state=NORMAL)
                testingSomething = True
            else:
                self.pcsButton.configure(state=DISABLED)
                
            if self.configuration.testTS:
                self.tssButton.configure(state=NORMAL)
                testingSomething = True
            else:
                self.tssButton.configure(state=DISABLED)

            if self.configuration.testPSS:
                self.pssButton.configure(state=NORMAL)
                testingSomething = True
            else:
                self.pssButton.configure(state=DISABLED)

            if self.configuration.testIOS:
                self.iosButton.configure(state=NORMAL)
                testingSomething = True
            else:
                self.iosButton.configure(state=DISABLED)

            if testingSomething and (self.configuration.baseConfig.language != TestConfiguration.Language.Java):
                # if testing any of the above segments, will need at least some GSLs
                self.generateGoldLibrariesButton.configure(state=NORMAL)
            else:
                self.generateGoldLibrariesButton.configure(state=DISABLED)

            if self.configuration.testOS:
                self.osButton.configure(state=NORMAL)
                testingSomething = True
            else:
                self.osButton.configure(state=DISABLED)


            if self.configuration.datatypeConfig.dataModel is not None:
                self.dataModelButton.configure(state=NORMAL)
                testingSomething = True
            else:
                self.dataModelButton.configure(state=DISABLED)


            if testingSomething:
                self.testAllButton.configure(state=NORMAL)
            else:
                self.testAllButton.configure(state=DISABLED)

            if self.configuration.priorTestExists():
                self.cleanupButton.configure(state=NORMAL)
            else:
                self.cleanupButton.configure(state=DISABLED)
            


    def quitTest(self):
        logger.info('User quit conformance test')
        commonTools.popParentGUI()
        sys.exit(0)
    
    def testAll(self):
        ''' calls the TestSuite module to run all appropriate conformance tests based on current configuration '''
        logger.info('') 
        conformancePassed, reportFilename = self.configuration.runConformanceTest(interactive=True, conformanceTestGUI=self)

        if reportFilename is None: # No report to display
            return

        logger.debug('Displaying Conformance Report')

        if self.configuration.baseConfig.browser is None:
            self.configuration.baseConfig.browser = commonTools.getDefaultBrowser()

        if self.configuration.baseConfig.browser is not None:
            result = subprocess.Popen([self.configuration.baseConfig.browser, reportFilename])
            logger.debug('browser call: ' + self.configuration.baseConfig.browser + ' ' + reportFilename +
                         ' returned a ' + str(result.returncode))
        else:
            logger.error('No browser specified')

        
    def generateGoldLibraries(self):
        ''' Calls TestSuite Module to generate all appropriate test code based on current configuration '''
        logger.info('')
        initialDir = os.path.join(ConformanceMain.getRootDir(), 'sample')
        if self.configuration is not None and self.configuration.baseConfig.projectDirectory != '':
            initialDir = self.configuration.baseConfig.projectDirectory
        options = { 'title':'What directory do you wish to save the Gold Standard Libraries',
                    'parent':self,
                    'initialdir':initialDir }

        goldLibDestDir = Dialogs.osAgnosticOpenDirectoryDialog(**options)
        logger.debug("Saving gold libs to " + goldLibDestDir)
        if goldLibDestDir != '':
            self.configuration.initializeConformanceTest()
            self.configuration.buildGoldStandardLibraries(targetDir=goldLibDestDir)
            logger.debug('Gold Libs built: ' + ', '.join(self.configuration.goldLibrariesBuilt))


    def _storeAndClearTestOptions(self):
        self._storedTestOptions = dict()
        self._storedTestOptions[TestConfiguration.SegmentConfig.Segment.IOS] = self.configuration.testIOS
        self.configuration.testIOS = False
        self._storedTestOptions[TestConfiguration.SegmentConfig.Segment.OSS] = self.configuration.testOS
        self.configuration.testOS = False
        self._storedTestOptions[TestConfiguration.SegmentConfig.Segment.PCS] = self.configuration.testPC
        self.configuration.testPC = False
        self._storedTestOptions[TestConfiguration.SegmentConfig.Segment.PSS] = self.configuration.testPSS
        self.configuration.testPSS = False
        self._storedTestOptions[TestConfiguration.SegmentConfig.Segment.TSS] = self.configuration.testTS
        self.configuration.testTS = False
        
    def _restoreTestOptions(self):
        if self._storedTestOptions is None:
            logger.error('Tried to restore test options without setting them')
            return
        
        self.configuration.testIOS = self._storedTestOptions[TestConfiguration.SegmentConfig.Segment.IOS]
        self.configuration.testOS = self._storedTestOptions[TestConfiguration.SegmentConfig.Segment.OSS]
        self.configuration.testPC = self._storedTestOptions[TestConfiguration.SegmentConfig.Segment.PCS]
        self.configuration.testPSS = self._storedTestOptions[TestConfiguration.SegmentConfig.Segment.PSS]
        self.configuration.testTS = self._storedTestOptions[TestConfiguration.SegmentConfig.Segment.TSS]

    def testDataModel(self):
        self._storeAndClearTestOptions()
        self.configuration.testDataModel = True
        self.testAll()
        self.configuration.testDataModel = False
        self._restoreTestOptions()


    def testIOServices(self):
        ''' Calls ConformanceTest Module to test the IO services segment based on current configuration '''
        ''' Keep copy of configuration...overwrite the test options...then restore options '''
        logger.info('')

        self._storeAndClearTestOptions()
        self.configuration.testIOS = True
        self.testAll()
        self._restoreTestOptions()

    def testPortableComponents(self):
        ''' Calls TestSuite Module to test the Portable Components segment based on current configuration '''
        logger.info('')

        self._storeAndClearTestOptions()
        self.configuration.testPC = True
        self.testAll()
        self._restoreTestOptions()


    def testTransportServices(self):
        ''' Calls TestSuite Module to test the Transport Services segment based on current configuration '''
 
        logger.info('')
        self._storeAndClearTestOptions()
        self.configuration.testTS = True
        self.testAll()
        self._restoreTestOptions()
        

    def testPlatformSpecificServices(self):
        ''' Calls TestSuite Module to test the Platform Specific Services segment based on current configuration '''
 
        logger.info('')
        self._storeAndClearTestOptions()
        self.configuration.testPSS = True
        self.testAll()
        self._restoreTestOptions()
        
    
    def testOperatingSystem(self):
        ''' Calls TestSuite Module to test the OS services segment based on current configuration '''
 
        logger.info('')
        self._storeAndClearTestOptions()
        self.configuration.testOS = True
        self.testAll()
        self._restoreTestOptions()
                
    def cleanupTestResults(self):
        ''' Deletes all previous test results '''
        
        logger.info('')
        self.configuration.deleteTestResults()
        self.updateMainMenu()
        

mainMenu = None

def ConformanceTestGUIMain(**kw):
    ''' Main Method to run the ConformanceTestGUI '''
    logger.info('Running ConformanceTestGUI')
    global mainMenu
    mainMenu = ConformanceTestGUI(**kw)                    
    mainMenu.master.title("Conformance Test Suite") 
    mainMenu.mainloop()                         
          
def getMainMenu():
    global mainMenu
    return mainMenu

if __name__ == '__main__':

    parser = OptionParser(description='FACE Conformance Test Suite GUI', usage="usage: %prog config_file")

    (options, args) = parser.parse_args()

    logger.info('Args: ' + str(args))
 
    kw = dict()
    
    if len(args) == 1:
        configFile = args[0]
        print 'configFile = ' + configFile
        kw['configFile'] = configFile
 
    ConformanceTestGUIMain(**kw)

        

        
        
