"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""

__author__ = 'david.fann@vanderbilt.edu'

import logging
import ConformanceMain

if __name__ == '__main__':
    ConformanceMain.initializeLogger()

import hashlib
import threading
import SocketServer
import os.path
import copy
import commonTools
import sys
import Dialogs

#import face_conformance_pb2
import constants

# Needed to use the protobuf without installing protobuf/setuptools on host computer, run them from the included eggs
egg_path = os.path.join(ConformanceMain.getRootDir(),'face_conformance_app', 'protobuf-2.5.0-py2.7.egg')
sys.path.append(egg_path)
egg_path = os.path.join(ConformanceMain.getRootDir(),'face_conformance_app', 'setuptools-0.6c11-py2.7!\\')
sys.path.append(egg_path)

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import descriptor_pb2

import datamodel_transfer_pb2
import DataModelTestResult_pb2
import collections
import protocol_buffer_socket_server


logger = logging.getLogger(__name__)

ConformanceResults = collections.namedtuple('ConformanceResults', ['name', 'description', 'score', 'log'])

class DataModel():

    def __init__(self, **params):
        self.unitOfConformanceDataModelFilename = params.get('uocDataModelFilename', None)
        self.sharedDataModelFilename = params.get('sharedDataModelFilename', None)
        self.jvmCommandLineOptions = params.get('jvmCommandLineOptions', '')
        self._uops = dict()

        self.validModel = None
        self.uocDMHashValue = None
        self.sharedDMHashValue = None
        self.testResults = list()
        self.TEST_PASSED = DataModelTestResult_pb2.DMTestResults.TestResult.PASS
        self.TEST_FAILED = DataModelTestResult_pb2.DMTestResults.TestResult.FAIL
        self.TEST_NOT_PERFORMED = DataModelTestResult_pb2.DMTestResults.TestResult.NOT_PERFORMED

        if self.unitOfConformanceDataModelFilename is not None:
            self.getDataModelInfo(self.unitOfConformanceDataModelFilename)

    def __eq__(self, other):
        try:
            # Did it this way to easily see what isn't equal while debugging...could make one statement
            # at some point
            equals = self.unitOfConformanceDataModelFilename == other.unitOfConformanceDataModelFilename
            equals = equals and self.sharedDataModelFilename == other.sharedDataModelFilename
            equals = equals and self._uops == other._uops
            equals = equals and self.validModel == other.validModel
            equals = equals and self.uocDMHashValue == other.uocDMHashValue
            equals = equals and self.sharedDMHashValue == other.sharedDMHashValue

            if equals is False:
                logger.debug('dataModel is different')
            return equals
        except Exception as e:
            logger.debug('dataModel comparison raised exception: ' + e.message + ' returning Not Equals')
            return False

    def isValid(self):
        return self.validModel

    def filename(self):
        return self.unitOfConformanceDataModelFilename

    def setHashValues(self):
        self.sharedDMHashValue = self.getHashValue(self.sharedDataModelFilename)
        self.uocDMHashValue = self.getHashValue(self.unitOfConformanceDataModelFilename)

    def getHashValue(self, dataModelFile):
        if dataModelFile is None or dataModelFile == "" or not os.path.isfile(dataModelFile):
            return None
        return hashlib.sha256(open(dataModelFile,'rb').read()).digest()

    def runConformanceTest(self):
        dm_test_results = DataModelTestResult_pb2.DMTestResults()
        protobuf_server = protocol_buffer_socket_server.ProtocolBufferSocketServer(dm_test_results)


        conformanceTestPath = os.path.join(ConformanceMain.getRootDir(), constants.DM_CONFORMANCE_TEST_SUBDIR)
        if constants.getFACEStandardVersion() == constants.FACE_STANDARD_VERSION_2_0:
          command = constants.getJavaExe() + " " + self.jvmCommandLineOptions + " -jar " + constants.getDMConformanceTestJar() + " -p" + \
                    str(protobuf_server.port) + ' "' + self.sharedDataModelFilename + '" "' \
                    + self.unitOfConformanceDataModelFilename + '" '
        else:
          command = constants.getJavaExe() + " " + self.jvmCommandLineOptions + " -jar " + constants.getDMConformanceTestJar() + " -p " + \
                    str(protobuf_server.port) + ' -sdm "' + self.sharedDataModelFilename + '" -usm "' \
                    + self.unitOfConformanceDataModelFilename + '" '
        pathList = list()

        returnValues = commonTools.subprocessWithSpecifiedPath(command, pathList, conformanceTestPath)
        logger.debug('return: ' + str(returnValues))
        if returnValues.returncode != 0:
            logger.error('DMConformanceTest returned an error. output: ' + returnValues.stdout + returnValues.stderr)
            # Server is shut down in socket handler...but doesn't get called if java program ends badly
            if not protobuf_server.handler_called:
                protobuf_server.finish()  # if handler not called by now...then DMConformanceTest must have died

            self.testResults = [ConformanceResults('DMConformanceTest Application Crashed',
                        'Java App that checks conformance has crashed.  Most likely UoP Data Model provided is invalid',
                        self.TEST_FAILED, returnValues.stdout + returnValues.stderr)]

            self.validModel = False # If DMConformanceTest crashed, good chance it failed conformance

        else:
            if protobuf_server.handler_called == True:
                self.uocDMHashValue = self.getHashValue(self.unitOfConformanceDataModelFilename)
                self.sharedDMHashValue = self.getHashValue(self.sharedDataModelFilename)
                self.validModel = True
                logger.debug("Waiting for socketServer handler to finish")
                protobuf_server.lock.acquire()
                self.parseDataModelTestResultsProtobuf(dm_test_results)

            else:
                self.validModel = False

        return self.validModel


    def getDataModelInfo(self, dataModelFilename):
        """
        parseDataModelFile sets up a socketserver, then calls the java program that parses
        data model files.  The values are passed back to the conformance test via the socket
        callback.  Once the java program ends, the socketserver may be closed
        """

        self.unitOfConformanceDataModelFilename = dataModelFilename

        if not os.path.isfile(dataModelFilename):
            raise IOError("No file exists: " + dataModelFilename)

        dm_transfer = datamodel_transfer_pb2.DataModel()
        protobuf_server = protocol_buffer_socket_server.ProtocolBufferSocketServer(dm_transfer)


        self._uops = dict()
        self.validModel = True

        if constants.getFACEStandardVersion() == constants.FACE_STANDARD_VERSION_2_0:
            codeGeneratorPath = os.path.join(ConformanceMain.getRootDir(), constants.DM_CONFORMANCE_CODE_GEN_SUBDIR)
            command = constants.getJavaExe() + ' ' + self.jvmCommandLineOptions + ' -jar ' + constants.getDMConformanceCodeGenJar() \
                      + ' -p' + str(protobuf_server.port) + \
                      ' -uop "' + dataModelFilename + '"'
        else:
            # DMConformanceTestJar can output uop info in 2.1
            codeGeneratorPath = os.path.join(ConformanceMain.getRootDir(), constants.DM_CONFORMANCE_TEST_SUBDIR)
            command = constants.getJavaExe() + ' ' + self.jvmCommandLineOptions + ' -jar ' + constants.getDMConformanceTestJar() \
                      + ' -p ' + str(protobuf_server.port) + \
                      ' -uop -usm "' + dataModelFilename + '"'


        pathList = list()
        returnValues = commonTools.subprocessWithSpecifiedPath(command, pathList, codeGeneratorPath)
        logger.debug('return: ' + str(returnValues))

        errMsg = ''
        if returnValues.returncode != 0:
            if constants.getFACEStandardVersion() == constants.FACE_STANDARD_VERSION_2_0:
                errMsg = 'DMConformanceCodeGenerator returned an error. output: ' + returnValues.stdout + returnValues.stderr
            else:
                errMsg = 'DMConformanceTest returned an error. output: ' + returnValues.stdout + returnValues.stderr        # CR-753: Fixed error message.
            self.validModel = False
        elif not protobuf_server.handler_called: # If DMConformanceCodeGenerator doesn't have a UoP, handler won't get called
            errMsg = 'UoC Data Model does not have a Unit of Portability defined'

        if len(errMsg) > 0:
            # Server is shut down in socket handler...but doesn't get called if java program ends badly
            protobuf_server.finish()
            logger.error(errMsg)
            if commonTools.isInteractive():
                Dialogs.ErrorMessageBox(constants.FACE_TESTSUITE_NAME, errMsg, parent=commonTools.getParentGUI())

        else: # Valid model
            self.uocDMHashValue = self.getHashValue(dataModelFilename)
            if protobuf_server.handler_called:
                # Waiting for the dataModel socketserver handler to finish parsing
                logger.debug("Waiting for socketServer handler to finish")
                protobuf_server.lock.acquire()
                self.parseDataModelTransferProtobuf(dm_transfer)

            else:
                errMsg = "Data Model processor did not return correctly, please exit test suite and try again\n" + \
                    "If you repeatedly get this message, you data model may be invalid"
                logger.error(errMsg)
                if commonTools.isInteractive():
                    Dialogs.ErrorMessageBox("Data Model Processing Error", errMsg, parent = commonTools.getParentGUI())
                else:
                    raise RuntimeError(errMsg)


    def getUoPNames(self):
        return sorted(self._uops.keys())

    def getUoPMessages(self, uopName):
        if uopName in self._uops:
            return sorted(set(self._uops[uopName]))  # removes duplicates
        else:
            return None

    def getUoPs(self):
       return copy.deepcopy(self._uops)

    def parseDataModelTestResultsProtobuf(self, dataModelTestResult):
        logger.info('dm_test_result protobuf:' + str(dataModelTestResult))
        try:
            self.testResults = list()

            for r in dataModelTestResult.result:
                self.testResults.append(ConformanceResults(r.name,
                                                           r.testDescription,
                                                           r.score, r.log))

            for t in self.testResults:
                logger.debug(str(t))

        except Exception as e:
            logger.error(40 * '!' + '\nParsing data failed: ' + e.message)
            raise e

    def parseDataModelTransferProtobuf(self, dataModelProtoBuf):
        logger.info('dm_transfer protobuf:' + str(dataModelProtoBuf))

        try:
            self._uops = dict()
            for uop in dataModelProtoBuf.unitOfPortability:
                self._uops[uop.name] = list()
                for m in uop.messageType:
                    self._uops[uop.name].append(m.name)

        except Exception as e:
            logger.error(40 * '!' + '\nParsing data failed: ' + e.message)
            raise e
