#!/usr/bin/env python
"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""

__author__ = 'david.fann@vanderbilt.edu'

# Python standard imports
from Tkinter import *
from platform import system
import tkFileDialog
import sys
import os
import tkFont
import logging
import SplashScreen

logger = logging.getLogger(__name__)

# Test suite imports
import constants
import commonTools


def insertHorizontalLine(master, **params):
    lineHeight = params.get('height', 1)
    spacerHeight = params.get('spacerHeight', 3)
    bg = params.get('bg', 'grey')
    relief = params.get('relief', SUNKEN)
    width = params.get('width', None)
    side = params.get('side', TOP)
    lineParams = {'height':lineHeight, 'bg':bg, 'relief':relief}

    if width is not None:
        lineParams['width'] = width

    mainFrame=Frame(master,height=spacerHeight)
    mainFrame.pack(side=TOP)
    f = Frame(mainFrame,height=spacerHeight)
    f.pack(side=TOP)

    f=Frame(mainFrame,**lineParams)
    if width is None:
        f.pack(fill=X, side=side)
    else:
        f.pack(side=side)

    f = Frame(mainFrame, height=spacerHeight)
    f.pack(side=TOP)
    return mainFrame



class Dialog(Toplevel):

    def __init__(self, parent, title=None, font=None):

        Toplevel.__init__(self, parent)
        self.transient(parent)

        if font is not None:
            self.font = font
        else:
            self.font = tkFont.nametofont('TkDefaultFont')

        if title:
            self.title(title)

        self.parent = parent
        body = Frame(self)
        self.initial_focus = self.body(body)
        body.pack(padx=5, pady=5)

        self.buttonbox()

        self.grab_set()

        if not self.initial_focus:
            self.initial_focus = self

        self.protocol("WM_DELETE_WINDOW", self.cancel)
        if parent is not None:
            self.geometry("+%d+%d" % (parent.winfo_rootx()+50,
                                      parent.winfo_rooty()+50))

        self.initial_focus.focus_set()

        self.wait_window(self)

    #
    # construction hooks

    def body(self, master):
        # create dialog body.  return widget that should have
        # initial focus.  this method should be overridden

        pass

    def buttonbox(self):
        # add standard button box. override if you don't want the
        # standard buttons
        
        box = Frame(self)

        w = Button(box, text="OK", width=10, command=self.ok, default=ACTIVE)
        w.pack(side=LEFT, padx=5, pady=5)
        w = Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=LEFT, padx=5, pady=5)

        self.bind("&lt;Return>", self.ok)
        self.bind("&lt;Escape>", self.cancel)

        box.pack()

    #
    # standard button semantics

    def ok(self, event=None):

        if not self.validate():
            self.initial_focus.focus_set() # put focus back
            return

        self.withdraw()
        self.update_idletasks()

        self.apply()

        self.cancel()

    def cancel(self, event=None):

        # put focus back to the parent window
        if self.parent is not None:
            self.parent.focus_set()
        self.destroy()

    #
    # command hooks

    def validate(self):

        return 1 # override

    def apply(self):
        pass # override


class ListDialog(Dialog):
    '''
    classdocs
    '''
    def FILE(self):
        @apply
        def fset(self, value):
            raise SyntaxError
        def fget(self):
            return 0xF113  # Random constant that looks sorta like file
        return property(**locals())

    def DIRECTORY(self):
        @apply
        def fset(self, value):
            raise SyntaxError
        def fget(self):
            return 0xD13 # Random constant that looks sorta like dir
        return property(**locals())

    def CLASS_PATH(self):
        @apply
        def fset(self, value):
            raise SyntaxError
        def fget(self):
            return 0xC1A33 # Random constant that looks sorta like class
        return property(**locals())


    def __init__(self, parent, items, itemType, **params):
        '''
        Constructor
        '''
        self.items = items
        self.itemType = itemType
        self.dialogWidth = params.get('dialogWidth', 660)
        self.dialogHeight = params.get('dialogHeight', 100)
        self.buttonSize = params.get('buttonSize', 5)
        self.stdWidth = params.get('stdWidth', 60)
        self.maxWidth = params.get('maxWidth', 100)
        self.prompt = params.get('prompt', None)
        self.filetypes = params.get('filetypes', None)
        self.__title = params.get('title', None)
        self._initialDir = params.get('initialdir', os.path.expanduser('~'))

        if type(self.items) != ListType:
            raise TypeError('items must be a list')

        all_item_types = [ListDialog.FILE, ListDialog.DIRECTORY, ListDialog.CLASS_PATH]

        if not self.itemType in all_item_types:
            raise TypeError('itemType must either be ListDialog.FILE or ListDialog.DIRECTORY or ListDialog.CLASS_PATH')

        if self.itemType == ListDialog.CLASS_PATH:
            self.stdWidth -= self.buttonSize

        if self._initialDir == None:
            self._initialDir = os.path.expanduser('~')
            if self.itemType == ListDialog.DIRECTORY:
                if len(self.items) > 0:
                    self._initialDir = self.items[-1]
            else:
                if len(self.items) > 0:
                    self._initialDir = os.path.dirname(self.items[-1])
                
        self.returnValue = None
        Dialog.__init__(self, parent, self.__title)
  
    def body(self, master):
        self.cwin = master
        if len(self.items) == 0 or self.items[-1] != '':
            self.items.append('')
        
        self.listLines = list()
        self.numLines = 0
        
        if self.prompt is not None and self.prompt != '':
            self.promptLabel = Message(self.cwin, text = self.prompt, width=self.dialogWidth)
            self.promptLabel.pack(side=TOP)
            insertHorizontalLine(self.cwin, width=0.9 * self.dialogWidth)

        self.listFrameMaster = VerticalScrolledFrame(self.cwin, width=self.dialogWidth, height=self.dialogHeight)
        self.listFrame = self.listFrameMaster.interior
        self.listFrameMaster.pack_propagate(0)
        self.listFrameMaster.pack(side=TOP)
        self.update()
              
    def apply(self):
        if len(self.items) > 1:
            self.returnValue = self.items[:-1]
        else:
            self.returnValue = None
    
    def deleteListItem(self, item):
        self.items.remove(item)
        self.update()

    def isDuplicate(self, item):
        if item in self.items:
            return True
        
        return False 
    
    def addListItems(self, items, prevItem = None, replaceItem = False):
        if prevItem is None:
            self.items.extend(items)
        else:
            index = self.items.index(prevItem)
            self.items = self.items[0:index] + items + self.items[index:]
            if replaceItem:
                self.items.remove(prevItem)
           
        self.update()
    
    def addListItem(self, item, prevItem = None):
        if prevItem is None:
            self.items.append(item)
        else:
            self.items.insert(self.items.index(prevItem), item)
        self.update()
 
    def replaceItem(self, item, prevItem):
        self.items[self.items.index(prevItem)] = item
        self.update()
        
    def moveListItem(self, item, direction):
        index = self.items.index(item)
        if direction == "UP":
            if index > 0:
                self.items[index] = self.items[index-1]
                self.items[index-1] = item
        if direction == "DOWN":
            if index < len(self.items):
                self.items[index] = self.items[index+1]
                self.items[index+1] = item
        self.update()
    
    def update(self):
        logger.debug('Num Items: ' + str(len(self.items)) + ' -- ' + str(self.items))

        scroll_to_bottom = False
        if len(self.items) == 0 or self.items[-1] != '':
            self.items.append('')
            scroll_to_bottom = True

        itemWidth = self.stdWidth

        listDif = len(self.listLines) - len(self.items)
        if listDif > 0:
            for i in range(listDif):
                self.listLines[-1].remove()
                self.listLines.pop()
                self.numLines -= 1
        if listDif < 0:
            if len(self.listLines) > 0:
                self.listLines[-1].downButton.config(state=NORMAL)
                self.listLines[-1].upButton.config(state=NORMAL)
            if len(self.listLines) > 1:
                self.listLines[-2].downButton.config(state=NORMAL)
            
            for i in range(-listDif):
                options = { 'row':self.numLines, 'initialdir':self._initialDir }
                if self.filetypes is not None:
                    options['filetypes'] = self.filetypes
                self.listLines.append(listLine(self.listFrame, self, self.itemType, **options))
                self.numLines += 1
                
        for i in range(0, len(self.items)):
            self.listLines[i].setItem(self.items[i])
            self.listLines[i].entry.config(width = itemWidth)
        
        if len(self.listLines) > 0:
            self.listLines[0].upButton.config(state=DISABLED)
            self.listLines[-1].downButton.config(state=DISABLED)
            self.listLines[-1].upButton.config(state=DISABLED)
        if len(self.listLines) > 1:
            self.listLines[-2].downButton.config(state=DISABLED)

        if scroll_to_bottom:
            self.listFrameMaster.canvas.yview_moveto(1)

        self.cwin.update_idletasks()
       
        
class listLine(object):
    
    def __init__(self, parentElement, parentListDialog, listType, **params):
       
        self.validate = params.get('validation', 'focusout')
        self.entryWidth = params.get('entryWidth', 60)
        self.row = params.get('row', None)
        self.filetypes = params.get('filetypes', [('all files', '*')])
        self.__title = params.get('title', None)
        self.skipValidation = False
        self.listVar = StringVar()
        self.frame = Frame(parentElement)
        self.parentDialog = parentListDialog
        self.lastValue = None

        columnNum = 0

        if listType == ListDialog.CLASS_PATH:
            validateCommand = self._validfileordirectory
            options = {'textvariable': self.listVar, 'validate':self.validate, 'validatecommand': validateCommand, 'width': self.entryWidth }
            self.entry = Entry(self.frame, options)
            self.entry.grid(row=0, column=columnNum)
            columnNum += 1

            buttonCommand = self._directoryDialog
            self.ellipseButton = Button(self.frame, text=' D ', command=buttonCommand)
            self.ellipseButton.grid(row=0, column=columnNum)
            columnNum += 1

            buttonCommand = self._fileDialog
            self.ellipseButton = Button(self.frame, text=' F ', command=buttonCommand)
            self.ellipseButton.grid(row=0, column = columnNum)
            columnNum += 1

        else:
            if listType == ListDialog.FILE:
                validateCommand = self._validfile
                buttonCommand = self._fileDialog

            elif listType == ListDialog.DIRECTORY:
                validateCommand = self._validDirectory
                buttonCommand = self._directoryDialog

            else:
                validateCommand = None
                buttonCommand = None

            options = {'textvariable': self.listVar, 'validate':self.validate, 'validatecommand': validateCommand, 'width': self.entryWidth }

            self.entry = Entry(self.frame, options)
            self.entry.grid(row=0, column = columnNum)
            columnNum += 1

            self.ellipseButton = Button(self.frame, text='...', command=buttonCommand)
            self.ellipseButton.grid(row=0, column = columnNum)
            columnNum += 1

        self.upButton = self._upButton()
        self.upButton.grid(row=0, column=columnNum)
        columnNum += 1
        self.downButton = self._downButton()
        self.downButton.grid(row=0, column=columnNum)
        columnNum += 1
        self.delButton = self._deleteListButton()
        self.delButton.grid(row=0, column=columnNum)
        columnNum += 1
        if self.row is not None:
            self.frame.grid(row=self.row)
        else:
            self.frame.grid()
    
    def setItem(self, value):
        self.lastValue = value
        self.skipValidation = True
        self.listVar.set(value)
        self.skipValidation = False
    
    def remove(self):
        self.frame.grid_remove()

    def _validfileordirectory(self):
        if self.skipValidation:
            return True

        entry_value = self.listVar.get()
        if (entry_value != '') and (not os.path.isfile(entry_value) and (not os.path.isdir(entry_value))):
            ErrorMessageBox('Error in file/directory selection', entry_value + ' is not a valid file or directory', parent=self.parentDialog)
            self.listVar.set(self.lastValue)
            return False

        self.parentDialog.replaceItem(self.listVar.get(), self.lastValue)
        self.lastValue = self.listVar.get()
        return True

    def _validfile(self):
        if self.skipValidation:
            return True
        
        if (self.listVar.get() != '') and (not os.path.isfile(self.listVar.get())):
            ErrorMessageBox('File does not exist',
                                   'File: ' + self.listVar.get() + ' is not a valid file', parent=self.parentDialog)
            self.listVar.set(self.lastValue)
            return False
               
        self.parentDialog.replaceItem(self.listVar.get(), self.lastValue)
        self.lastValue = self.listVar.get()
        return True
    
    def _validDirectory(self):
        if (self.listVar.get() != '') and (not os.path.isdir(self.listVar.get())):
            ErrorMessageBox('Directory does not exist',
                                   'Directory: ' + self.listVar.get() + ' is not a valid directory',
                                   parent=self.parentDialog)
            self.listVar.set(self.lastValue)
            return False
             
        self.parentDialog.replaceItem(self.listVar.get(), self.lastValue)
        self.lastValue = self.listVar.get()
        return True
   
    def _fileDialog(self):
        options = dict()
        originalValue = self.listVar.get()
        initialDir = self.parentDialog._initialDir
        initialFile = None
        
        if originalValue != '':
            initialDir, initialFile = os.path.split(originalValue)
        
        options['initialdir'] = initialDir

        # Don't want to select initialFile, since it will always return the initial file
        # as part of the returnFiles...not only if you select it.
        #if initialFile is not None:
        #    options['initialfile'] = initialFile

        if self.__title is None:
            options['title'] = 'Please pick a file(s)'
        else:
            options['title'] = self.__title
        options['filetypes'] = self.filetypes
        options['parent'] = self.parentDialog.cwin
        returnFiles = osAgnosticOpenFilenamesDialog(**options)
        
        if len(returnFiles) > 0 and ' '.join(returnFiles) != originalValue:

            ''' The windows version returns a unicode string, the linux version returns a tuple...need to handle either.''' 
            if type(returnFiles) == UnicodeType:
                returnFiles = returnFiles.split()
            elif type(returnFiles) == TupleType:             
                returnFiles = list(returnFiles)
            elif type(returnFiles) == ListType:
                pass # want it as a list

            newFiles = list()
            for r in returnFiles:
                if self.parentDialog.isDuplicate(r) and r != originalValue:
                    response = YesNoMessageBox(
                        'File already in list',
                        'File ' + r + ' is already in the list, are you sure you want duplicate copies?',
                        parent=self.parentDialog
                    )
                    if response == 'yes':
                        newFiles.append(r)

                else:
                    newFiles.append(r)

            if len(newFiles) > 0:
                self.parentDialog.addListItems(newFiles, originalValue, True)
                self.parentDialog._initialDir = os.path.dirname(newFiles[-1])


    def _directoryDialog(self):
        options = dict()
        initialDir = self.parentDialog._initialDir
        
        originalValue = self.listVar.get()
        if self.listVar.get() != '':
            initialDir = self.listVar.get()
        options['initialdir'] = initialDir
        
        
        if self.__title is None:
            options['title'] = 'Please pick a directory'
        else:
            options['title'] = self.__title
        options['parent'] = self.parentDialog.cwin 
        returnDirectory = osAgnosticOpenDirectoryDialog(**options)
        
        if len(returnDirectory) > 0 and returnDirectory != originalValue:
            if self.parentDialog.isDuplicate(returnDirectory):
                WarningMessageBox('Directory already in list',
                                         'Directory ' + returnDirectory +
                                         ' is already in the list, multiple copies not allowed',
                                         parent=self.parentDialog)
            else:
                self.parentDialog.replaceItem(returnDirectory, originalValue)
            self.parentDialog._initialDir = returnDirectory

    def _deleteListButton(self):
        return Button(self.frame, text=u'\u2718',command= self._deleteButtonCmd, foreground='red')

    def _deleteButtonCmd(self):
        self.parentDialog.deleteListItem(self.listVar.get())
    
    def _upButton(self):
        return Button(self.frame, text=u'\u21E7',command= self._upButtonCommand)
    
    def _upButtonCommand(self):
        self.parentDialog.moveListItem(self.listVar.get(), "UP")
    
    def _downButton(self):
        return Button(self.frame, text=u'\u21E9',command= self._downButtonCommand)
    
    def _downButtonCommand(self):
        self.parentDialog.moveListItem(self.listVar.get(), "DOWN")
    

class FancyDialog(Dialog):
    """
        Used to replace tkMessagebox dialogs
    """

    def __init__(self, text=None, **params):
        """  Base class for ErrorDialog, WarningDialog, and YesNoDialog
        Args:
            text (str): the text to display in the dialog window

        Kwargs:
            title (str): the title to use in the dialog window
            parent (Tkinter object): the parent widget to the dialog
            maxWidth (int): the maximum width in characters you wish to display on a line
            font (tkFont): the font you wish to use on the dialog box
        """
        self.titleText = params.get('title', constants.FACE_TESTSUITE_NAME)
        self.parent = params.get('parent', commonTools.getParentGUI())
        self.maxWidth = params.get('maxWidth', 100)
        self.text = commonTools.wraptext(text, maxLength=self.maxWidth)
        dialogFont = params.get('font', None)
        self.bodyIcon = None
        self.returnValue = False
        Dialog.__init__(self, self.parent, self.titleText, dialogFont)

    def body(self, master):
        if self.bodyIcon is not None:
            self.iconWidget = Label(master, image=self.bodyIcon)
            self.iconWidget.grid(row=0, column=0)

        self.textWidget = Label(master, text=self.text, padx=10, justify=LEFT, font=self.font)
        self.textWidget.grid(row=0, column=1)

    def okbuttonbox(self):
        # button box with only ok button. override if you don't want the
        # standard buttons

        box = Frame(self)

        w = Button(box, text="OK", width=10, command=self.ok, default=ACTIVE)
        w.pack(side=LEFT, padx=5, pady=5)

        self.bind("&lt;Return>", self.ok)

        box.pack()


    def yesnobuttonbox(self):
        # button box with yes/no options. override if you don't want the
        # standard buttons

        box = Frame(self)

        w = Button(box, text="Yes", width=10, command=self.yes, default=ACTIVE, font=self.font)
        w.pack(side=LEFT, padx=5, pady=5)
        w = Button(box, text="No", width=10, command=self.no, font=self.font)
        w.pack(side=LEFT, padx=5, pady=5)

        self.bind("&lt;Return>", self.ok)
        self.bind("&lt;Escape>", self.cancel)

        box.pack()

    def yes(self, event=None):
        self.returnValue = True
        Dialog.ok(self, event)

    def no(self, event=None):
        self.returnValue = False
        Dialog.cancel(self, event)

class InfoDialog(FancyDialog):
    def __init__(self, text=None, **params):
        FancyDialog.__init__(self, text, **params)

    def body(self, master):
        self.bodyIcon = PhotoImage(file=constants.INFO_GIF)
        FancyDialog.body(self, master)

    def buttonbox(self):
        FancyDialog.okbuttonbox(self)


class ErrorDialog(FancyDialog):
    def __init__(self, text=None, **params):
        FancyDialog.__init__(self, text, **params)

    def body(self, master):
        self.bodyIcon = PhotoImage(file=constants.ERROR_GIF)
        FancyDialog.body(self, master)

    def buttonbox(self):
        FancyDialog.okbuttonbox(self)


class WarningDialog(FancyDialog):
    def __init__(self, text=None, **params):
        FancyDialog.__init__(self, text, **params)

    def body(self, master):
        self.bodyIcon = PhotoImage(file=constants.WARNING_GIF)
        FancyDialog.body(self, master)

    def buttonbox(self):
        FancyDialog.okbuttonbox(self)


class YesNoDialog(FancyDialog):
    def __init__(self, text=None, **params):
        FancyDialog.__init__(self, text, **params)

    def body(self, master):
        self.bodyIcon = PhotoImage(file=constants.QUESTION_GIF)
        FancyDialog.body(self, master)

    def buttonbox(self):
        FancyDialog.yesnobuttonbox(self)


class VerticalScrolledFrame(Frame):
    """A pure Tkinter scrollable frame that actually works!

    * Use the 'interior' attribute to place widgets inside the scrollable frame
    * Construct and pack/place/grid normally
    * This frame only allows vertical scrolling

    """
    def __init__(self, parent, *args, **kw):
        Frame.__init__(self, parent, *args, **kw)

        # create a canvas object and a vertical scrollbar for scrolling it
        vscrollbar = Scrollbar(self, orient=VERTICAL)
        vscrollbar.pack(fill=Y, side=RIGHT, expand=FALSE)
        self.canvas = Canvas(self, bd=0, highlightthickness=0,
                        yscrollcommand=vscrollbar.set)
        self.canvas.pack(side=LEFT, fill=BOTH, expand=TRUE)
        vscrollbar.config(command=self.canvas.yview)

        # reset the view
        self.canvas.xview_moveto(0)
        self.canvas.yview_moveto(0)


        # create a frame inside the canvas which will be scrolled with it
        self.interior = interior = Frame(self.canvas)
        interior_id = self.canvas.create_window(0, 0, window=interior,
                                           anchor=NW)

        # track changes to the canvas and frame width and sync them,
        # also updating the scrollbar
        def _configure_interior(event):
            # update the scrollbars to match the size of the inner frame
            size = (interior.winfo_reqwidth(), interior.winfo_reqheight())
            self.canvas.config(scrollregion="0 0 %s %s" % size)
            if interior.winfo_reqwidth() != self.canvas.winfo_width():
                # update the canvas's width to fit the inner frame
                self.canvas.config(width=interior.winfo_reqwidth())
        interior.bind('<Configure>', _configure_interior)

        def _configure_canvas(event):
            if interior.winfo_reqwidth() != self.canvas.winfo_width():
                # update the inner frame's width to fill the canvas
                self.canvas.itemconfigure(interior_id, width=self.canvas.winfo_width())
        self.canvas.bind('<Configure>', _configure_canvas)

        # TODO: For some reason the on_mousewheel is not being called...need to figure this out at some point
        def _on_mousewheel(event):
            logger.debug('On Mousewheel event: ' + str(event))
            self.canvas.yview_scroll(-1*(event.delta/120), "units")

        # tried both canvas and interior...none have worked
        self.bind('<4>', _on_mousewheel)
        self.bind('<5>', _on_mousewheel)

        return


def YesNoMessageBox(title, text, **params):
    SplashScreen.hide()
    params['title'] = title
    yn = YesNoDialog(text, **params)
    if yn.returnValue:
        return 'yes'
    else:
        return 'no'

def ErrorMessageBox(title, text, **params):
    SplashScreen.hide()
    params['title'] = title
    ErrorDialog(text, **params)

def WarningMessageBox(title, text, **params):
    SplashScreen.hide()
    params['title'] = title
    WarningDialog(text, **params)

def InfoMessageBox(title, text, **params):
    SplashScreen.hide()
    params['title'] = title
    InfoDialog(text, **params)

# Used for the osAgnostic Dialogs to remember the last one used
lastDirectory = os.getcwd()

def osAgnosticSaveAsFilenameDialog(**kw):
    global lastDirectory
    logger.debug('Will bring up a file dialog here')

    kw = windowsSpecificFileDialogOptions(**kw)

    if ('initialdir' not in kw):
        kw['initialdir'] = lastDirectory

    returnFile = tkFileDialog.asksaveasfilename(**kw)
    logger.debug('Return File: ' + str(returnFile))
    if len(returnFile) != 0:
        lastDirectory = os.path.dirname(returnFile)
        if sys.platform == 'win32':
            returnFile = returnFile.replace('/', '\\') # tk is too smart for itself and replaces the \ with /
    else:
        returnFile = ''
    return returnFile


def osAgnosticOpenFilenameDialog(**kw):
    global lastDirectory
    logger.debug('Will bring up a file dialog here')
    if ('initialdir' not in kw):
        kw['initialdir'] = lastDirectory

    returnFile = tkFileDialog.askopenfilename(**kw)
    logger.debug('Return File: ' + str(returnFile))
    if len(returnFile) != 0:
        lastDirectory = os.path.dirname(returnFile)
        if sys.platform == 'win32':
            returnFile = returnFile.replace('/', '\\') # tk is too smart for itself and replaces the \ with /
    else:
        returnFile = ''
    return returnFile


def osAgnosticOpenFilenamesDialog(**kw):
    global lastDirectory
    if ('initialdir' not in kw):
        kw['initialdir'] = lastDirectory

    returnFiles = tkFileDialog.askopenfilenames(**kw)
    logger.debug('Return Files: ' + str(returnFiles) + 'type: ' + str(type(returnFiles)))
    if len(returnFiles) != 0:
    # Windows returns unicode sometimes...need to split it, Linux returns a tuple

        if type(returnFiles) == tuple:  #Linux and some versions of python on Windows returns a tuple...need a list
            returnFiles = list(returnFiles)
        else: # Some versions of python on Windows returns unicode which is a ROYAL PAIN
            # Must handle spaces in names...
            # tkFileDialog returns them as '{C:\Path with space\file.txt} C:\NoSpaceInPath\file.txt'
            tempReturnFiles = returnFiles.split() # split at the spaces
            returnFiles = list()
            tempFile = None
            # But reassemble the ones with spaces (known by the starting '{' and ending '}')
            # tempFile is used to store the interim values

            for f in tempReturnFiles:
                if tempFile is not None:
                    if f.endswith('}'):
                        returnFiles.append(tempFile + ' ' + f.strip('}'))
                        tempFile = None
                    else:
                        tempFile += ' ' + f
                elif not f.startswith('{'):
                    returnFiles.append(f)
                else:
                    tempFile = f.strip('{')

        if system() == 'Windows':
            windowsReturnFiles = list()
            for r in returnFiles:
                windowsReturnFiles.append(r.replace('/', '\\'))  # tk is too smart for itself and replaces the \ with /
            returnFiles = windowsReturnFiles

        lastDir = os.path.dirname(returnFiles[-1])  # get the directory of the last file in the list
        if os.path.isdir(lastDir):  # make sure it is a valid dir
            lastDirectory = lastDir
    else:
        returnFiles = ''

    return returnFiles


def osAgnosticOpenDirectoryDialog(**kw):
    global lastDirectory

    if ('initialdir' not in kw):
        kw['initialdir'] = lastDirectory

    parent = kw.get('parent', None)

    returnDir = tkFileDialog.askdirectory(**kw)

    if len(returnDir) != 0:
        if sys.platform == 'win32':
            returnDir = returnDir.replace('/', '\\') # tk is too smart for itself and replaces the \ with /
        if not os.path.isdir(returnDir):
            msg = returnDir + " doesn't exist, do you want to create it?"
            if YesNoMessageBox("Directory doesn't exist", msg, parent=parent) == 'yes':
                os.mkdir(returnDir)
            else:
                returnDir = ''

        if returnDir != '':
            lastDirectory = returnDir

    return returnDir


def directoryDialog(**kw):
    """
    Creates a tkinter directory dialog box, returns the directory selected

    keywords:
    initialDir - the initial directory (default is common working directory)
    titleString - the title of the dialog box
    """
    global lastDirectory
    logger.info('kw:' + str(kw))
    initialDir = kw.get('initialDir', lastDirectory)
    parent = kw.get('parent', None)
    title = kw.get('title', 'Please select a directory')
    returnDir = tkFileDialog.askdirectory(initialdir=initialDir, title=title, parent=parent)

    if (len(returnDir) > 0) and (not os.path.isdir(returnDir)):
        msg = returnDir + " doesn't exist, do you want to create it?"
        if YesNoMessageBox("Directory doesn't exist", msg, parent=parent) == 'yes':
            os.mkdir(returnDir)
        else:
            returnDir = ''

    logger.debug('Return Dir: ' + str(returnDir))
    return returnDir


def fileDialog(**kw):
    """
    Creates a tkinter file dialog box, returns the file(s) selected

    keywords:
    initialDir - the initial directory (default is common working directory)
    titleString - the title of the dialog box
    filetypes - a dict containing the filetypes to use
    """

    logger.info('kw:' + str(kw))
    global lastDirectory

    parent = kw.get('parent', None)
    initialDir = kw.get('initialDir', lastDirectory)
    titleString = kw.get('title', 'Please select file(s)')
    filetypes = kw.get('filetypes', None)

    logger.debug('Will bring up a file dialog here')
    returnFiles = tkFileDialog.askopenfilenames(initialdir=initialDir, title=titleString, filetypes=filetypes,
                                                parent=parent)
    logger.debug('Return Files: ' + str(returnFiles))
    return returnFiles


def windowsSpecificFileDialogOptions(**kw):
    """ So...windows version doesn't include the default extension unless you specifically add it, and
        windows picks the last item in the filetypes list as the default, so we must reverse it """
    if (system() == 'Windows') and ('filetypes' in kw):
        filetypesList = kw['filetypes']
        if len(filetypesList) > 0 and len(filetypesList[0]) == 2:
            defaultExtension = filetypesList[0][1]
            if 'defaultextension' not in kw:
                kw['defaultextension'] = defaultExtension
            filetypesList.reverse()
            kw['filetypes'] = filetypesList
    return kw






if __name__ == "__main__":

    testFrame = Frame()
    testFrame.grid()
    font = tkFont.Font(family="Helvetica", size=10)
    font = None
    dialogText = 'This is a test, this is only a test\nYes this is another test' \
        + 'This is a test, this is only a test...This is a test, this is only a test...This is a test, this is only a test\n'

    title = 'THis is the title'

    ErrorMessageBox(title, dialogText, parent=testFrame, font=font,  width=100)
    WarningMessageBox(title, dialogText, parent=testFrame, font=font, width=100)
    InfoMessageBox(title, dialogText, parent=testFrame, font=font, width=100)
    yn = YesNoMessageBox(title, dialogText, parent=testFrame, font=font, width=100)
    print 'ReturnValue: ' + str(yn)

    listItems = [ 'file1', 'file2', 'file3']
    l = ListDialog(testFrame, listItems, ListDialog.FILE, title='This is the title', prompt = 'This is a prompt')
    print 'Return Value: '
    print l.returnValue

    dialogList = [ ]
    ld1 = ListDialog(testFrame, dialogList, ListDialog.DIRECTORY, prompt='This is a prompt', title='This is a title', directoryPrompt='Pick your favorite directory')
    print 'Return Value: '
    print ld1.returnValue


