#!/usr/bin/env python
"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""

"""
---------------------------------------------------------------------------------
HTMLWriter - helper functions/classes to generate HTML files

"""
__author__ = 'david.fann@vanderbilt.edu'

#from xml.etree.ElementTree import ElementTree
import os
import copy
try:
#    # Python 2.5
    import xml.etree.ElementTree as etree
    print("running with ElementTree on Python 2.5+")
except ImportError:
    print("error importing elementtree")

from commonTools import listToString, safeOpen

import logging
logger = logging.getLogger(__name__)

class HTMLHeading():
    def __init__(self):
        self.text = ''
        self.subHeadings = list(HTMLHeading)
        self.superiorHeading = HTMLHeading()
 
class HeadingType:
     (Super, Peer, Sub, TopLevel, SpecifiedLevel) = range(-1,4)

class HeadingNode:
    def __init__(self):
        self.element = ''
        self.super = ''
        self.level = -1
      
 
class HTMLWriter():
    def __init__(self, titleText = "Generated by HTMLWriter"):
        htmlTags = {'xmlns': "http://www.w3.org/1999/xhtml", 'lang':"en"}
        self.root = etree.Element("html", **htmlTags)
        
        #self.root = etree.Element("html", xmlns="http://www.w3.org/1999/xhtml", xml:lang="en", lang="en")
        self.head = etree.SubElement(self.root, "head")
        self.title = etree.SubElement(self.head, "title")
        self.title.text = titleText
        self.body = etree.SubElement(self.root, "body")
        self.body.text = ' '
        self._UIDNumber = 0
        self.menuIdLUT = dict()
        
    def setTitle(self, titleText):
        self.title.text = titleText
    
    def getTitle(self):
        return self.title.text
    
    def addTag(self, tagName, tagText, **attributes):

        superElement = self.body
        if "superElement" in attributes:
            superElement = attributes["superElement"]
            del attributes["superElement"]
        
        if "cssClass" in attributes:
            attributes['class'] = attributes['cssClass']
            del attributes['cssClass']
                
        tagElement = etree.SubElement(superElement, tagName, **attributes)
        if tagText != '':
            tagElement.text = tagText
        return tagElement
    
    def addLink(self, text, **attributes):
        attributes['superElement']=self.head
        return self.addTag('link', text, **attributes)
    
    def addStyle(self, text, **attributes):
        attributes['superElement']=self.head
        return self.addTag('style', text, **attributes)
    
    def addScript(self, text, **attributes):
        if 'superElement' not in attributes:
            attributes['superElement'] = self.head
        if text == '':
            text = ' '  #scripts don't use the <script /> notation well, adding the space assures the <script> </script>
        return self.addTag('script', text, **attributes)
    
    def write(self, filename):
        #tree = etree.ElementTree(self.root)
        xmlFile = safeOpen(filename, 'w')
        xmlFile.write(etree.tostring(self.root, encoding='us-ascii'))  # would use html method...except earlier python versions do not support
        xmlFile.close()
        
            
    def indent(self, indentText = '', **attributes):
        self.addTag('BLOCKQUOTE', indentText, **attributes)
    
    def addAnchor(self, text, **attributes):
        return self.addTag('a', text, **attributes)
    
    def addParagraph(self,paragraphString, **attributes):
        return self.addTag('p', paragraphString, **attributes)
       
    def addDiv(self, **attributes):
        return self.addTag('div', '', **attributes)
    
    def addSpan(self, spanText, **attributes):  
        return self.addTag('span', spanText, **attributes)  
        
    def addHeading(self, headingLevel, headingText, **attributes ):
        return self.addTag('h' + str(headingLevel), headingText, **attributes)  

    def addUnorderedList(self, **attributes):
        return self.addTag('ul', '', **attributes)
    
    def addListItem(self, text, listElement, **attributes):
        attributes['superElement'] = listElement
        return self.addTag('li', text, **attributes)
    
    def addSpacer(self, **attributes):
        return self.addTag('spacer', '', **attributes)
    
    def addButton(self, text, **attributes):
        return self.addTag('button',text, **attributes)
    
    def addHyperLink(self, hrefToAdd, text, **attributes):
             
        name = attributes.get('name', '')    
        if name != '':
            hrefToAdd += '#' + name
        attributes['href'] = hrefToAdd
        self.addTag('a', text, **attributes)
        
    def addConfigListElement(self, name, valueList, **attributes):
        configAttributes = dict()
        for key, value in attributes.iteritems():
            configAttributes[key] = value
        
        menuElement, collapsibleElement = self.addCollapsibleMenuItem(name, **configAttributes)
        configAttributes = dict()
        configAttributes["cssClass"] = "configName"
        configAttributes["superElement"] = collapsibleElement
        for vl in valueList:
            self.addParagraph(vl, **configAttributes)
        
    def addConfigElement(self, name, value, **attributes):
        configAttributes = copy.copy(attributes)
        paragraph = self.addParagraph('', **configAttributes)
        configAttributes["superElement"] = paragraph
        configAttributes["cssClass"] = "configName"
        self.addSpan(name + ": ", **configAttributes)
        configAttributes["cssClass"] = "configValue"
        self.addSpan(value, **configAttributes)
        return paragraph
    
    def getMenuUID(self):
        self._UIDNumber += 1
        return "menu" + str(self._UIDNumber)
    
    def addCollapsibleMenuItem(self, menuText, **attributes):
        menuUID = self.getMenuUID()

        headingLevel = 2
        if "headingLevel" in attributes:
            headingLevel = attributes['headingLevel']
            del attributes['headingLevel']
        
        #menuAttributes['cssClass'] = "collapsibleMenu"
        #menuAttributes['onClick'] =        "openClose('" +menuUID + "')"
        menuElement = self.addHeading(headingLevel, '', cssClass='collapsibleMenu', onClick="openClose('" +menuUID + "')",  **attributes)
        self.addSpan('+ ', superElement = menuElement, id = "tick_" + menuUID, cssClass = 'tickMarker')
        self.addSpan(menuText, cssClass="collapsibleMenuText", superElement=menuElement)
        
        self.menuIdLUT[menuElement] = menuUID
        divElement = self.addDiv(id=menuUID, cssClass="collapsibleText", **attributes)
        return menuElement, divElement


def HTMLWriterMain():
    
    report = HTMLWriter("FACE Conformance Test Results")
    report.addLink('', rel="stylesheet", type="text/css", href="ConformanceReport.css") 
    report.addScript('', type="text/javascript", src="ConformanceReport.js")
    report.addHeading(1, "FACE Conformance Test Results")
    testConfigMenu = report.addHeading(2, '', cssClass='collapsibleMenu', onClick="openClose('testConfig')")
    report.addSpan('+ ', id="tick_testConfig", superElement=testConfigMenu)
    report.addSpan('Test Configuration', cssClass='collapsibleMenuText', superElement=testConfigMenu)
  
    div1 = report.addDiv(id="testConfig", cssClass="collapsibleText")
    report.addHeading(3, "Heading 1 Paragraph", superElement=div1)

    report.write('page.html')
     
     
if __name__ == '__main__':
    HTMLWriterMain()

