"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""
__author__ = 'david.fann@vanderbilt.edu'

import sys
import logging
import os
import platform
import collections
import itertools

import ConformanceMain
import TestConfiguration
import commonTools
import constants
import Dialogs

# Needed to use the protobuf without installing protobuf/setuptools on host computer, run them from the included eggs
egg_path = os.path.join(ConformanceMain.getRootDir(),'face_conformance_app', 'protobuf-2.5.0-py2.7.egg')
sys.path.append(egg_path)
egg_path = os.path.join(ConformanceMain.getRootDir(),'face_conformance_app', 'setuptools-0.6c11-py2.7!\\')
sys.path.append(egg_path)

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import descriptor_pb2

import TestResult_pb2
import JavaTypes_pb2
import protocol_buffer_socket_server
import SplashScreen
import face_specified_code_generator

logger = logging.getLogger(__name__)

ConformanceResults = collections.namedtuple('ConformanceResults', ['name', 'description', 'score', 'log'])

class JavaPackageTree:
    def __init__(self):
        self.root_node = JavaPackageNameNode(root=True)

    def __iter__(self):
        for n in iter(self.root_node):
            yield n

    def pretty_print(self):
        self.root_node.pretty_print()

    def to_string(self):
        return self.root_node.to_string()

    def add_node(self, node_name):
        parent_node = self.root_node
        for n in node_name.split('.'):
            parent_node = parent_node.get_child(n)
        return parent_node

    def add_subgroup_node(self, subgroup):
        new_node = self.add_node(subgroup.name)
        new_node.subgroup = subgroup

class JavaPackageNameNode:
    def __init__(self, **params):
        self.name = params.get('name', 'Java Conformance Test')
        self.full_name = params.get('fullname', '')
        self.subgroup = params.get('subgroup', None)
        self.root_flag = params.get('root', False)
        self.children = list()

    def __iter__(self):
        yield self
        for v in itertools.chain(*itertools.imap(iter, self.children)):
            yield v

    def to_string(self, level=0):
        return_string = ''
        offset = 0
        if not self.root_flag:
            return_string = level * '-' + self.name + '\n'
            offset = 1
        for c in self.children:
            return_string += c.to_string(level+offset)
        return return_string

    def pretty_print(self, level=0):
        offset = 0
        if not self.root_flag:
            print level * '-' + self.name
            offset = 1
        for c in self.children:
            c.pretty_print(level+offset)

    def get_child(self, child_name):
        # returns node with name = child_name if it exists, otherwise creates new child with child_name
        for c in self.children:
            if c.name == child_name:
                return c
        self.children.append(JavaPackageNameNode(name=child_name,
                                                 fullname=(self.full_name + '.' + child_name).lstrip('.')))
        return self.children[-1]

def build_tss_datatype_jar(test_configuration, jarfile, **kw):
    logger.info(locals())

    # CR 909: Remove then recreate src/ before generating datatypes to ensure no *.java leftover from the previous run
    goldTSSDMSrcDir = os.path.join(ConformanceMain.getRootDir(), constants.JAVA_GOLD_DIR, 'src')
    if (os.path.exists(goldTSSDMSrcDir)):
        commonTools.rmdir_p(goldTSSDMSrcDir)
    commonTools.mkdir_p(goldTSSDMSrcDir)

    face_specified_code_generator.generate_data_model_data_type_definitions(
        TestConfiguration.Language.Java,
        test_configuration.datatypeConfig.dataModel.unitOfConformanceDataModelFilename,
        test_configuration.baseConfig.logDirectory)

    goldStandardPath = os.path.join(ConformanceMain.getRootDir(), constants.JAVA_GOLD_DIR)

    # CR 455: Makefile.JavaGold got updated to put compiled TSS DM .class into /bin
    goldTSSDMBinDir = os.path.join(ConformanceMain.getRootDir(), constants.JAVA_GOLD_DIR, 'bin')
    if (os.path.exists(goldTSSDMBinDir)):
        commonTools.rmdir_p(goldTSSDMBinDir)
    commonTools.mkdir_p(goldTSSDMBinDir)

    environVars = kw.get('environ', dict())

    # Overwrite the autogenerated Makefile with the FACE one.
    commonTools.copyFile(
        os.path.join(goldStandardPath, 'Makefile.JavaGold'),
        os.path.join(goldStandardPath, 'Makefile')
    )

    # Write the Makefile.env for the gold libs
    makeEnvFilename = os.path.join(goldStandardPath, 'Makefile.env')
    makeEnv = open(makeEnvFilename, 'w')
    test_configuration.baseConfig.writeMakefileEnvironment(makeEnv, environVariables=environVars, relativeDir=goldStandardPath, useGoldLibPath=False)

    makeEnv.close()

    makeTargetCommand = test_configuration.baseConfig.make
    workingDirectory = os.path.join(ConformanceMain.getRootDir(), constants.JAVA_GOLD_DIR)
    result = commonTools.subprocessWithSpecifiedPath(makeTargetCommand,
                                                     [workingDirectory],
                                                     workingDirectory)
    jar_base_name = os.path.basename(jarfile)
    logger.debug(makeTargetCommand + ' ' + jar_base_name + ': returncode = ' + str(result.returncode) + '\nstdout: ' + result.stdout + '\nstderr: ' + result.stderr)
    logfile = os.path.join(test_configuration.baseConfig.logDirectory, jar_base_name + '.log')
    info = {'working directory':workingDirectory,
            'path directories':os.path.pathsep.join(test_configuration.baseConfig.addedPathDirs)}
    commonTools.logSubprocessResults(makeTargetCommand, result, logfile, otherInfo=info)
    if result.returncode != 0:
        buildSuccess = False
        errorMsg = 'Error in building gold lib: ' + jar_base_name
        logger.error(errorMsg + ' see logfile: ' + logfile)
        if commonTools.isInteractive():
            Dialogs.ErrorMessageBox(constants.FACE_TESTSUITE_NAME, errorMsg + '\nPress Ok to see log file: ' + logfile, parent=test_configuration.conformanceTestGUI)
            commonTools.displayFile(logfile)
        else:
            raise RuntimeError(errorMsg)


def GenerateJavaGoldCode(datatypes, filename, register_callback, interface_under_test=None):
    datatypes = list(set(datatypes)) # remove any duplicates...order doesn't matter for these
    logger.info('datatypes: ' + ', '. join(datatypes))

    java_types_proto = JavaTypes_pb2.JClasses()

    ts_interface = java_types_proto.classes.add()
    ts_interface.name = "LFACE/TS;"
    ts_interface.permission = JavaTypes_pb2.JP_PUBLIC
    ts_interface.parent = 'Ljava/lang/Object;'

    tss_methods_2_0 = {'Initialize': '(Ljava/lang/String;LFACE/RETURN_CODE_TYPEHolder;)V',
                       'create_connection': '(Ljava/lang/String;LFACE/MESSAGING_PATTERN_TYPE;Lus/opengroup/FACE/LongHolder;LFACE/CONNECTION_DIRECTION_TYPEHolder;Lus/opengroup/FACE/IntHolder;LFACE/RETURN_CODE_TYPEHolder;)V',
                       'destroy_connection': '(JLFACE/RETURN_CODE_TYPEHolder;)V',
                       'get_connection_parameters': '(Lus/opengroup/FACE/StringHolder;Lus/opengroup/FACE/LongHolder;LFACE/TRANSPORT_CONNECTION_STATUS_TYPEHolder;LFACE/RETURN_CODE_TYPEHolder;)V',
                   }

    tss_datatype_methods_2_0 = {'register_read_callback': ['(J[ZLFACE/Read_Callback_face_dm_',
                                                       ';ILFACE/RETURN_CODE_TYPEHolder;)V'],
                                'receive_message': ['(JJLus/opengroup/FACE/LongHolder;Lface/dm/',
                                                   'Holder;ILFACE/RETURN_CODE_TYPEHolder;)V'],
                                'send_message': ['(JLface/dm/',
                                                ';Lus/opengroup/FACE/IntHolder;JLus/opengroup/FACE/LongHolder;LFACE/RETURN_CODE_TYPEHolder;)V']
                            }


    tss_methods_2_1 = {'Initialize': '(Ljava/lang/String;LFACE/RETURN_CODE_TYPEHolder;)V',
                       'Create_Connection': '(Ljava/lang/String;LFACE/MESSAGING_PATTERN_TYPE;Lus/opengroup/FACE/LongHolder;LFACE/CONNECTION_DIRECTION_TYPEHolder;Lus/opengroup/FACE/IntHolder;JLFACE/RETURN_CODE_TYPEHolder;)V',
                       'Destroy_Connection': '(JLFACE/RETURN_CODE_TYPEHolder;)V',
                       'Get_Connection_Parameters': '(Lus/opengroup/FACE/StringHolder;Lus/opengroup/FACE/LongHolder;LFACE/TRANSPORT_CONNECTION_STATUS_TYPEHolder;LFACE/RETURN_CODE_TYPEHolder;)V'
                   }

    tss_datatype_methods_2_1 = {'Receive_Message': ['(JJLus/opengroup/FACE/LongHolder;LFACE/DM/',
                                                    'Holder;ILFACE/RETURN_CODE_TYPEHolder;)V'],
                                'Send_Message': ['(JJLus/opengroup/FACE/LongHolder;LFACE/DM/',
                                                 'Holder;Lus/opengroup/FACE/IntHolder;LFACE/RETURN_CODE_TYPEHolder;)V']
                            }


    tss_methods = {constants.FACE_STANDARD_VERSION_2_0: tss_methods_2_0,
                   constants.FACE_STANDARD_VERSION_2_1: tss_methods_2_1,
                   constants.FACE_STANDARD_VERSION_2_1_1: tss_methods_2_1}

    tss_datatype_methods = {constants.FACE_STANDARD_VERSION_2_0: tss_datatype_methods_2_0,
                            constants.FACE_STANDARD_VERSION_2_1: tss_datatype_methods_2_1,
                            constants.FACE_STANDARD_VERSION_2_1_1: tss_datatype_methods_2_1}


    version = constants.getFACEStandardVersion()
    if version == 'Version 2.1.1':
        version = 'Version 2.1'  # There is no difference in the Java Conformance Test for 2.1.1 from 2.1
    for name, descriptor in tss_methods[version].iteritems():
        method = ts_interface.methods.add()
        method.name = name
        method.descriptor_ = descriptor
        method.static = True

    for d in datatypes:
        for name, descriptor_list in tss_datatype_methods[version].iteritems():
            method = ts_interface.methods.add()
            method.name = name
            method.descriptor_ = descriptor_list[0] + d + descriptor_list[1]
            method.static = True

    # CR 501
    tss_register_callback_2_1 = {'Unregister_Callback': '(JLFACE/RETURN_CODE_TYPEHolder;)V',
                                         'Register_Callback': '(J[ZLFACE/Read_Callback_FACE_DM_%Holder;ILFACE/RETURN_CODE_TYPEHolder;)V'
                                        }

    tss_read_callback_2_1 = {'name':'LFACE/Read_Callback_FACE_DM_%;',
                             'methods': [('send_event', '(JLFACE/DM/%Holder;JI[ZLFACE/RETURN_CODE_TYPEHolder;)V')]
                            }

    tss_read_callback_holder_2_1 = {'name':'LFACE/Read_Callback_FACE_DM_%Holder;',
                                    'methods': [('<init>','()V'), ('<init>','(LFACE/Read_Callback_FACE_DM_%;)V')],
                                    'fields': [('value', 'LFACE/Read_Callback_FACE_DM_%;')]
                                   }

    #if 2.1 version and register_read_callback is True:
    if version == 'Version 2.1' and register_callback:
        unreg_method = ts_interface.methods.add()
        unreg_method.name = 'Unregister_Callback'
        unreg_method.descriptor_ = tss_register_callback_2_1['Unregister_Callback']
        unreg_method.static = True

        for d in datatypes:
            reg_method = ts_interface.methods.add()
            reg_method.name = 'Register_Callback'
            reg_method.descriptor_ = tss_register_callback_2_1['Register_Callback'].replace('%', d)
            reg_method.static = True

            # Read_CallbackDM
            read_callback = java_types_proto.classes.add()
            read_callback.name = tss_read_callback_2_1['name'].replace('%', d)
            read_callback.permission = JavaTypes_pb2.JP_PUBLIC
            read_callback.parent = 'Ljava/lang/Object;'
            read_callback.interface = True
            read_callback.abstract = True

            for m in tss_read_callback_2_1['methods']:
                method = read_callback.methods.add()
                method.name = m[0]
                method.descriptor_ = m[1].replace('%', d)
                method.permission = JavaTypes_pb2.JP_PUBLIC


            # Read_CallbackDMHolder
            read_callback_holder = java_types_proto.classes.add()
            read_callback_holder.name = tss_read_callback_holder_2_1['name'].replace('%', d)
            read_callback_holder.permission = JavaTypes_pb2.JP_PUBLIC
            read_callback_holder.parent = 'Ljava/lang/Object;'
            read_callback_holder.final = True
            read_callback_holder.abstract = False

            for f in tss_read_callback_holder_2_1['fields']:
                field = read_callback_holder.fields.add()
                field.name = f[0]
                field.descriptor_ = f[1].replace('%', d)
                field.permission = JavaTypes_pb2.JP_PUBLIC

            for m in tss_read_callback_holder_2_1['methods']:
                method = read_callback_holder.methods.add()
                method.name = m[0]
                method.descriptor_ = m[1].replace('%', d)
                method.permission = JavaTypes_pb2.JP_PUBLIC
    # end CR 501

    logger.debug('Writing TS Gold javatypes to file: ' + filename)
    logger.debug(str(java_types_proto))
    gold_file = open(filename, 'wb')
    gold_file.write(java_types_proto.SerializeToString())
    gold_file.close()

    gold_text_file = open(filename + 'text', 'w')
    gold_text_file.write(str(java_types_proto))
    gold_text_file.close()

    return filename


class JavaConformanceTestResult:
    PASS = TestResult_pb2.R_PASS
    FAIL = TestResult_pb2.R_FAIL
    WARNING = TestResult_pb2.L_WARNING
    INSPECT = TestResult_pb2.R_INSPECT
    ERROR = TestResult_pb2.L_ERROR

    @staticmethod
    def result_string(result):
        result_dict = {
            JavaConformanceTestResult.PASS: 'PASS',
            JavaConformanceTestResult.FAIL: 'FAIL',
            JavaConformanceTestResult.WARNING: 'WARNING',
            JavaConformanceTestResult.INSPECT: 'INSPECT',
            JavaConformanceTestResult.ERROR: 'ERROR'
        }
        return result_dict.get(result, 'UNKNOWN VALUE: {0}'.format(result))

    class LogEntry():
        def __init__(self):
            self.level = None
            self.text = None

    class TestResult():
        def __init__(self):
            self.name = None
            self.score = None
            self.logs = list()
            self.description = None

    class SubGroup():
        def __init__(self):
            self.name = None
            self.results = list()
            self.subgroups = list()

    def __init__(self):
        self.subgroups = list()
        self.java_package_tree = None
        self.dummy_tree_limit = 200

    def add_subgroup_to_package_tree(self, subgroup):
        self.java_package_tree.add_subgroup_node(subgroup)
        for s in subgroup.subgroups:
            self.add_subgroup_to_package_tree(s)

    def create_java_package_tree(self):
        self.java_package_tree = JavaPackageTree()

        for s in self.subgroups:
            self.add_subgroup_to_package_tree(s)

        self.java_package_tree.pretty_print()
        logger.debug(self.java_package_tree.to_string())

    def parse_protobuf(self, protobuf_result):
        for s in protobuf_result.subgroup:
            self.subgroups.append(self.parse_subgroup(s))

        self.create_java_package_tree()

    def parse_subgroup(self, protobuf_subgroup):
        subgroup = self.SubGroup()
        subgroup.name = protobuf_subgroup.name
        if len(protobuf_subgroup.results) == 0 and len(protobuf_subgroup.subgroup) == 0:
            result = self.TestResult()
            result.name = 'Java Conformance Test'
            result.score = self.PASS
            result.description = 'Subgroup ' + subgroup.name + ' is conformant.'
            subgroup.results.append(result)
        else:
            for r in protobuf_subgroup.results:
                result = self.TestResult()
                result.name = r.name
                result.score = r.score
                result.description = r.description
                for l in r.logs:
                    log = self.LogEntry()
                    log.level = l.level
                    log.text = l.text
                    result.logs.append(log)
                subgroup.results.append(result)

        for s in protobuf_subgroup.subgroup:
            subgroup.subgroups.append(self.parse_subgroup(s))

        return subgroup

    def fancy_print(self, indent_level=0):
        for s in self.subgroups:
            print indent_level * '-' + ' name: ' + s.name + '\n'
            self.fancy_print_results(s.results, indent_level)
            for ss in s.subgroups:
                ss.fancy_print(indent_level + 1)

    def fancy_print_results(self, results, indent_level):
        for r in results:
            print 30 * '-'
            print (indent_level + 1) * '-' + ' result name: ' + r.name
            print (indent_level + 1) * '-' + ' result score: ' + str(r.score)
            print (indent_level + 1) * '-' + ' result description: ' + r.description
            self.fancy_print_logs(r.logs, indent_level+1)

    def fancy_print_logs(self, logs, indent_level):
        for l in logs:
            print indent_level * '-' + ' level: ' + l.level
            print indent_level * '-' + ' text: ' + l.text


def _runSegmentTest(segment_config, segment_options, profile, logfile):

    splash_text = '\nPlease wait, running Java ' + segment_config.segmentName + ' Conformance Test\n\n'
    if segment_config.segment == TestConfiguration.SegmentConfig.Segment.OSS:
        splash_text += 'Java OSS Tests may take a while (up to 20 minutes depending on machine)'
    SplashScreen.show(splash_text)
    java_conformance_test_result = JavaConformanceTestResult()
    protobuf_java_test_results = TestResult_pb2.TestResultGroup()

    protobuf_server = protocol_buffer_socket_server.ProtocolBufferSocketServer(protobuf_java_test_results, save_buffer=True)

    java_conformance_test_dir = os.path.join(ConformanceMain.getRootDir(), constants.JAVA_CONFORMANCE_TEST_DIR)
    environ_dict = dict()
    path_list = list()
    if platform.system() == 'Windows':
        path_list.append(java_conformance_test_dir)
    else: # LINUX
        ld_library_path = ''
        if 'LD_LIBRARY_PATH' in os.environ:
            ld_library_path = os.environ['LD_LIBRARY_PATH']
        environ_dict['LD_LIBRARY_PATH'] = java_conformance_test_dir + os.path.pathsep + ld_library_path

    #TODO: make --general a profile specified option (general, safetyext, safetybase)
    cmd_line = os.path.join(java_conformance_test_dir, constants.JAVA_CONFORMANCE_APP) + ' --' + profile + ' --port ' + str(protobuf_server.port) + ' --nogui ' + segment_options
    java_conformance_test_dir
    logger.debug('Running: ' + cmd_line)
    #returnValues = commonTools.subprocessWithSpecifiedPath_noShell(cmd_line, path_list, java_conformance_test_dir, environ_dict)
    returnValues = commonTools.subprocessWithSpecifiedPath(cmd_line, path_list, java_conformance_test_dir,environ_dict)
    commonTools.logSubprocessResults(cmd_line, returnValues, logfile)

    if returnValues.returncode != 0:
        err_message = constants.JAVA_CONFORMANCE_APP + ' returned an error. output: ' + \
                      returnValues.stdout + returnValues.stderr
        logger.error(err_message)
        # Server is shut down in socket handler...but doesn't get called if java program ends badly
        if not protobuf_server.handler_called:
            protobuf_server.finish()  # if handler not called by now...then JavaConformance must have died

        if commonTools.isInteractive():
            Dialogs.ErrorMessageBox('Error in ' + constants.JAVA_CONFORMANCE_APP, err_message)
            return False
        else:
            raise RuntimeError(err_message)

    else:
        if protobuf_server.handler_called:
            # Waiting for the dataModel socketserver handler to finish parsing
            logger.debug("Waiting for socketServer handler to finish")
            protobuf_server.lock.acquire()
            logger.debug("SocketServer handler finished")
        else:
            err_message = constants.JAVA_CONFORMANCE_APP + ' did not return a value on the socket '
            logger.error(err_message)
            protobuf_server.finish()  # if handler not called by now...then JavaConformance must have died
            if commonTools.isInteractive():
                Dialogs.ErrorMessageBox('Error in ' + constants.JAVA_CONFORMANCE_APP, err_message)
                return False
            else:
                raise RuntimeError(err_message)

        try:
            logger.debug('Parsing protobuf')
            protobuf_java_test_results.ParseFromString(protobuf_server.data_buffer)
            logger.debug(str(protobuf_java_test_results))

            java_conformance_test_result.parse_protobuf(protobuf_java_test_results)
            logger.debug('converted protobuf results')

        except Exception as x:
            logger.error(x.message)

    segment_config.java_conformance_test_results = java_conformance_test_result
    return True


def runConformanceTest(test_configuration, **kw):

    profile = kw['profile']
    java_conformance_test_dir = os.path.join(ConformanceMain.getRootDir(), constants.JAVA_CONFORMANCE_TEST_DIR)
    tss_javagold = os.path.join(ConformanceMain.getRootDir(), constants.JAVA_GOLD_DIR, constants.JAVA_TS_GOLD)

    if not os.path.isfile(os.path.join(java_conformance_test_dir, constants.JAVA_CONFORMANCE_APP)):
        err_msg = "You must install the Java Conformance Test before testing Java segments.  See manual for more details"
        logger.error(err_msg)
        if commonTools.isInteractive():
            Dialogs.ErrorMessageBox('Error in Running Conformance Test', err_msg)
            return False
        else:
            raise RuntimeError(err_msg)

    test_configuration.tsConfig.object_suffix = test_configuration.baseConfig.objectSuffix
    test_configuration.ioConfig.object_suffix = test_configuration.baseConfig.objectSuffix
    test_configuration.pcConfig.object_suffix = test_configuration.baseConfig.objectSuffix
    test_configuration.pssConfig.object_suffix = test_configuration.baseConfig.objectSuffix
    test_configuration.osConfig.object_suffix = test_configuration.baseConfig.objectSuffix


    conformance_test_ran_successfully = True

    java_golds = list()

    # TSS needs datatypes if FACE 2.0 or if it is either Abstraction layer or TS Interface UoP in 2.1
    TS_needs_datatypes = test_configuration.testTS  and (
        constants.getFACEStandardVersion() == constants.FACE_STANDARD_VERSION_2_0 or
        test_configuration.tsConfig.interfaceUnderTest in [
            TestConfiguration.TsSegmentConfig.InterfaceUnderTest.TS_INTERFACE_UoP,
            TestConfiguration.TsSegmentConfig.InterfaceUnderTest.ABSTRACTION_LAYER])

    # CR 501: PCS and PSSS needs register callback in tss.javagold. TS depends on GUI register_callback checkbox
    TS_need_register_callback = True
    if test_configuration.testTS:
        TS_need_register_callback = test_configuration.tsConfig.registerReadCallback

    if TS_needs_datatypes or test_configuration.testPC or test_configuration.testPSS:
        data_types = list()

        for uop_name, test_uop in test_configuration.datatypeConfig.uopsToTest.iteritems():
            if test_uop == 1:
                uop_data_types = test_configuration.datatypeConfig.dataModel.getUoPMessages(uop_name)
                data_types.extend(uop_data_types)

        GenerateJavaGoldCode(data_types, tss_javagold, TS_need_register_callback)
        java_golds.append(tss_javagold)

        jarfile = os.path.join(ConformanceMain.getRootDir(), constants.JAVA_GOLD_DIR, 'tss_datatypes.jar')
        build_tss_datatype_jar(test_configuration, jarfile)
        java_golds.append(jarfile)

    if test_configuration.testIOS:
        segment_options = ' --testios --classpath "' + os.path.pathsep.join(test_configuration.ioConfig.libraries) + '"'

        # Java Approved Framework Additions - July 2019
        fmfiles = os.pathsep.join(test_configuration.ioConfig.frameworkFiles)
        if fmfiles:
            segment_options += ' --frameworkclasspath "' + fmfiles + '"'

        logfile = os.path.join(test_configuration.baseConfig.logDirectory, 'java_conformance_test_IOS.log')
        conformance_test_ran_successfully = _runSegmentTest(test_configuration.ioConfig, segment_options, profile, logfile) and \
                                            conformance_test_ran_successfully

    if test_configuration.testTS:
        if constants.getFACEStandardVersion() == constants.FACE_STANDARD_VERSION_2_0:
            segment_options = ' --testtss --classpath "' + os.path.pathsep.join(test_configuration.tsConfig.libraries) + \
                '" --tssgold "' + os.pathsep.join(java_golds) + '"'
            logfile = os.path.join(test_configuration.baseConfig.logDirectory, 'java_conformance_test_TSS.log')
            conformance_test_ran_successfully = _runSegmentTest(test_configuration.tsConfig, segment_options, profile, logfile) and \
                                                conformance_test_ran_successfully

        else:  # FACE STANDARD 2.1
            '''
            if test_configuration.tsConfig.interfaceUnderTest == \
                TestConfiguration.TsSegmentConfig.InterfaceUnderTest.TS_INTERFACE_UoP:

                segment_options = ' --testtss --classpath "' + os.path.pathsep.join(test_configuration.tsConfig.libraries) + \
                    '" --tssgold "' + os.pathsep.join(java_golds) + '"'
                logfile = os.path.join(test_configuration.baseConfig.logDirectory, 'java_conformance_test_TSS.log')
                conformance_test_ran_successfully = _runSegmentTest(test_configuration.tsConfig, segment_options, profile, logfile) and \
                                                    conformance_test_ran_successfully

            elif test_configuration.tsConfig.interfaceUnderTest == \
                TestConfiguration.TsSegmentConfig.InterfaceUnderTest.ABSTRACTION_LAYER:
                # TODO: fix the option once John finishes it in the JavaTest
                segment_options = ' --testtssabs --classpath "' + os.path.pathsep.join(test_configuration.tsConfig.libraries) + \
                    '" --tssgold "' + os.pathsep.join(java_golds) + '"'
                logfile = os.path.join(test_configuration.baseConfig.logDirectory, 'java_conformance_test_TSS.log')
                conformance_test_ran_successfully = _runSegmentTest(test_configuration.tsConfig, segment_options, profile, logfile) and \
                                                    conformance_test_ran_successfully

            elif test_configuration.tsConfig.interfaceUnderTest == \
                TestConfiguration.TsSegmentConfig.InterfaceUnderTest.TYPE_ABSTRACTION_INTERFACE_UoP:
                # TODO: fix the option once John finishes it in the JavaTest
                segment_options = ' --testtssabsiface --classpath "' + \
                                  os.path.pathsep.join(test_configuration.tsConfig.libraries) + '"'
                logfile = os.path.join(test_configuration.baseConfig.logDirectory, 'java_conformance_test_TSS.log')
                conformance_test_ran_successfully = _runSegmentTest(test_configuration.tsConfig, segment_options, profile, logfile) and \
                                                    conformance_test_ran_successfully
            '''

            segment_options = ''
            logfile = os.path.join(test_configuration.baseConfig.logDirectory, 'java_conformance_test_TSS.log')

            if test_configuration.tsConfig.interfaceUnderTest == \
                TestConfiguration.TsSegmentConfig.InterfaceUnderTest.TS_INTERFACE_UoP:

                segment_options = ' --testtss --tssgold "' + os.pathsep.join(java_golds) + '"'

            elif test_configuration.tsConfig.interfaceUnderTest == \
                TestConfiguration.TsSegmentConfig.InterfaceUnderTest.ABSTRACTION_LAYER:
                # TODO: fix the option once John finishes it in the JavaTest
                segment_options = ' --testtssabs --tssgold "' + os.pathsep.join(java_golds) + '"'

            elif test_configuration.tsConfig.interfaceUnderTest == \
                TestConfiguration.TsSegmentConfig.InterfaceUnderTest.TYPE_ABSTRACTION_INTERFACE_UoP:
                # TODO: fix the option once John finishes it in the JavaTest
                segment_options = ' --testtssabsiface'

            segment_options = segment_options + ' --classpath "' + os.path.pathsep.join(test_configuration.tsConfig.libraries) + '"'
            # Java Approved Framework Additions - July 2019
            fmfiles = os.pathsep.join(test_configuration.tsConfig.frameworkFiles)
            if fmfiles:
                segment_options = segment_options + ' --frameworkclasspath "' + fmfiles + '"'
            conformance_test_ran_successfully = _runSegmentTest(test_configuration.tsConfig, segment_options, profile, logfile) and \
                                                conformance_test_ran_successfully

    if test_configuration.testPC:
        segment_options = ' --testpcs --classpath "' + os.path.pathsep.join(test_configuration.pcConfig.libraries) + \
            '" --tssgold "' + os.pathsep.join(java_golds) + '"'

        # Java Approved Framework Additions - July 2019
        fmfiles = os.pathsep.join(test_configuration.pcConfig.frameworkFiles)
        if fmfiles:
            segment_options += ' --frameworkclasspath "' + fmfiles + '"'
        logfile = os.path.join(test_configuration.baseConfig.logDirectory, 'java_conformance_test_PCS.log')
        conformance_test_ran_successfully = _runSegmentTest(test_configuration.pcConfig, segment_options, profile, logfile) and \
                                            conformance_test_ran_successfully

    if test_configuration.testPSS:
        segment_options = ' --testpss --classpath "' + os.path.pathsep.join(test_configuration.pssConfig.libraries) + \
            '" --tssgold "' + os.pathsep.join(java_golds) + '"'

        # Java Approved Framework Additions - July 2019
        fmfiles = os.pathsep.join(test_configuration.pssConfig.frameworkFiles)
        if fmfiles:
            segment_options += ' --frameworkclasspath "' + fmfiles + '"'
        logfile = os.path.join(test_configuration.baseConfig.logDirectory, 'java_conformance_test_PSS.log')
        conformance_test_ran_successfully = _runSegmentTest(test_configuration.pssConfig, segment_options, profile, logfile) and \
                                            conformance_test_ran_successfully

    if test_configuration.testOS:
        javaOSSConfig = test_configuration.osConfig.subTests[TestConfiguration.OsSegmentConfig.SubTests.JAVA_GENERAL]
        if profile != TestConfiguration.Profile.General:
            javaOSSConfig = test_configuration.osConfig.subTests[TestConfiguration.OsSegmentConfig.SubTests.JAVA_SAFETY]
        if javaOSSConfig.underTest:
            segment_options = ' --testoss --classpath "' + os.path.pathsep.join(javaOSSConfig.objectFiles) + '"'
            logfile = os.path.join(test_configuration.baseConfig.logDirectory, 'java_conformance_test_OSS.log')
            conformance_test_ran_successfully = _runSegmentTest(test_configuration.osConfig, segment_options, profile, logfile) and \
                                                conformance_test_ran_successfully

    return conformance_test_ran_successfully

