#!/usr/bin/env python
"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""

'''
SourceCodeTOHtml - helper functions/classes to convert source code to user friendly html files (with error msgs/hyperlinks inserted into the code

'''
__author__ = 'david.fann@vanderbilt.edu'

from HTMLWriter import HTMLWriter
import os
import logging
from commonTools import safeOpen, wraptext
import xml.etree as etree

logger = logging.getLogger(__name__)


htmlPrefixFile = 'SourceCodePrefix.html'
htmlCodePostFix =']]></script>\n'


def strip_control_characters(inputString):
    if inputString:  
        logger.debug('removing illegal characters')   
        import re  
          
        # unicode invalid characters  
        RE_XML_ILLEGAL = u'([\u0000-\u0008\u000b-\u000c\u000e-\u001f\ufffe-\uffff])' + \
                         u'|' + \
                         u'([%s-%s][^%s-%s])|([^%s-%s][%s-%s])|([%s-%s]$)|(^[%s-%s])' % \
                         (unichr(0xd800),unichr(0xdbff),unichr(0xdc00),unichr(0xdfff), \
                          unichr(0xd800),unichr(0xdbff),unichr(0xdc00),unichr(0xdfff),  \
                           unichr(0xd800),unichr(0xdbff),unichr(0xdc00),unichr(0xdfff),  \
                           )  
        inputString = re.sub(RE_XML_ILLEGAL, "", inputString)  
                          
        # ascii control characters  
        inputString = re.sub(r"[\x01-\x1F\x7F]", "", inputString)  
              
    return inputString

def htmlCodePrefix(lineNumber = 1, brush = 'plain'):
    return '<script type="syntaxhighlighter" class="brush: ' + brush + '; first-line:' + str(lineNumber) + '"><![CDATA[\n'

def addSourceCodeElement(htmlWriter, sourceCode, brush='plain', lineNumber = 1, **attributes):
    if 'superElement' not in attributes:
        attributes['superElement'] = htmlWriter.body 
    if sourceCode.startswith('\r\n'):
        sourceCode = u'\u00A0' + sourceCode  # bug in Javascript that doesn't handle newlines at start of code
    
    #logger.debug('Brush: ' + str(brush))
    #logger.debug('Linenumber: ' + str(lineNumber))
    scriptElement = htmlWriter.addTag('pre', sourceCode, cssClass="brush: " + brush + "; first-line:" + str(lineNumber) , **attributes)
  
    return scriptElement

def getCodeBrush(language):
    if language == 'C' or language == 'C++':
        return 'cpp'

    if language == 'Ada':
        return 'ada' # first guess
    
    if language == 'Java':
        return 'Java'
    
    else:
        return 'plain'
    
def codefileToHTML(htmlWriter, codefilename, language, **attributes):

    brush = attributes.pop('brush', getCodeBrush(language))
    codeString = ''
    logger.info('codefilename: ' + codefilename)
    maxLineLength = attributes.pop('maxLineLength', None)
    if not os.path.isfile(codefilename):
        codeString = 'File: ' + codefilename + ' does not exist'
    
    else:
        codefile = safeOpen(codefilename, 'r', 'utf-8')
        codeString = codefilename + '\n\n'
        for line in codefile:
            if maxLineLength is not None:
                codeString += wraptext(line, maxLength=maxLineLength)
            else:
                codeString += line
        codefile.close()
    
        if codeString == '':
            codeString = 'Empty code file'
        
    startingLineNumber = attributes.get('startingLineNumber', -1)
    codeElement = addSourceCodeElement(htmlWriter, codeString, brush, startingLineNumber, **attributes)

    #return headingElement, codeElement
    return codeElement

def stringListToHTML(htmlWriter, stringList, **attributes):
    
    language = attributes.get('language', 'text')
    startingLineNumber = attributes.get('startingLineNumber', 1)
    brush = getCodeBrush(language)
    listStringified = ''
    listStringified = listStringified.join(stringList)
    codeElement = addSourceCodeElement(htmlWriter, listStringified, brush, startingLineNumber, **attributes)
    return codeElement

      

def logfileToHTML(htmlWriter, logfilename, **attributes):
    logString = ''
    
    if not os.path.isfile(logfilename):
        logString = 'No Logfile'
    
    else:
        logfile = safeOpen(logfilename, 'r', 'utf-8')
        logString = ''
        for line in logfile:
            logString += line
        logfile.close()
    
        if logString == '':
            logString = 'Empty Logfile'
        
  
    codeElement = addSourceCodeElement(htmlWriter, logString, 'plain', 1, **attributes)

    #return headingElement, codeElement
    return codeElement

    
class SourceCodeToHtml(object):
    '''
    classdocs
    '''


    def __init__(self, **kw):
        #sourceFilename = '', htmlFilename = '', logDirectory = '', rootDir = ''):
        '''
        Constructor
        '''
        self.prevError = None
        self.nextError = None
        self.sourceFilename = kw.get('sourceFilename','')
        self.htmlFilename = kw.get('htmlFilename', '')
        self.logDirectory = kw.get('logDirectory', '')
        self.rootDir = kw.get('rootDir', '')
        
        if self.sourceFilename == '' or self.htmlFilename == '' or self.logDirectory == '' or self.rootDir == '':
            logger.warning('at least one of sourceFilename, htmlFilename, logDirectory, and rootDir not defined source: ' + self.sourceFilename \
                           + ' html: ' + self.htmlFilename + ' log: ' + self.logDirectory + ' root: ' + self.rootDir)
            
        self.summaryPage = ''
        
        if self.sourceFilename != '' and self.htmlFilename != '':
            self.translate(self.sourceFilename, self.htmlFilename, self.logDirectory)

    def getHtmlFilename(self, sourceFilename):
        htmlFile = sourceFilename[sourceFilename.rfind('/')+1:]
        if htmlFile.endswith('.o'):
            htmlFile = htmlFile[:-2] + '_o.'
        htmlFile = htmlFile[:htmlFile.rfind('.')] + '.html'
        ''' Not putting logDirectory in the filename so that the report can be moved and the links still work
        if logDirectory != '':
            if not logDirectory.endswith('/'):
                logDirectory += '/'
            htmlFile = logDirectory + htmlFile 
        '''
        return htmlFile

    def getLinkName(self, face_error):
        
        return face_error.filename + ':' + str(face_error.lineNumber)

    def addErrorMsg(self, htmlWriter, error, prevError, nextError):
        errorDiv = htmlWriter.addDiv(cssClass="errorMsgCodeFile")
        htmlWriter.addHyperLink('#top', 'Top of File', superElement = errorDiv)

        if prevError is not None:
            linkPrefix = ''
            if prevError.filename != self.sourceFilename:
                linkPrefix = self.getHtmlFilename(prevError.filename)
            htmlWriter.addHyperLink(linkPrefix + '#' + self.getLinkName(prevError), 'previous error', superElement = errorDiv)

        htmlWriter.addSpan(error.errorDescription, superElement = errorDiv)

        if nextError != None:
            linkPrefix = ''
            if nextError.filename != self.sourceFilename:
                linkPrefix = self.getHtmlFilename(nextError.filename)
                
            htmlWriter.addHyperLink(linkPrefix + '#' + self.getLinkName(nextError), 'next error', superElement = errorDiv)
      
        logger.debug('logmessage: [' + error.logMessage + ']')
        htmlWriter.addParagraph(error.logMessage, superElement = errorDiv)

    def generateErrorOnlyFile(self, sourceFilename, htmlFilename, errors, currentTime = None):
        
        htmlWriter = HTMLWriter(htmlFilename)
 
        self.addCodeFileHeadingInfo(htmlWriter)
        headingElement = htmlWriter.addHeading(1, os.path.basename(sourceFilename))
        if currentTime is not None:
            htmlWriter.addSpan(currentTime, cssClass="time", superElement=headingElement)
        
        numError = 0
        lineNumbers = list()
        relevantErrors = dict()
        prevError = dict()
        nextError = dict()
        lastError = None
        
        for e in errors:  
            if e.filename == sourceFilename:
                numError += 1
                if e.lineNumber == -1: # means not associated with a line number
                    e.lineNumber = 'e' + str(numError)
                lineNumbers.append(e.lineNumber)
                relevantErrors[e.lineNumber] = e
                prevError[e.lineNumber] = lastError 
                nextError[e.lineNumber] = None
                if (lastError is not None) and (lastError.lineNumber in relevantErrors.keys()):
                    nextError[lastError.lineNumber] = e
            lastError = e

        for lineNumber in relevantErrors.keys():
            htmlWriter.addAnchor('', name = self.getLinkName(relevantErrors[lineNumber]))
            self.addErrorMsg(htmlWriter, relevantErrors[lineNumber], prevError[lineNumber], nextError[lineNumber])
 
        htmlWriter.write(os.path.join(self.logDirectory, htmlFilename))

    def addCodeFileHeadingInfo(self, htmlWriter):
        htmlWriter.addLink('', rel="stylesheet", type="text/css", href="ConformanceReport.css")
        htmlWriter.addScript('', type="text/javascript", src="ConformanceReport.js")
        htmlWriter.addLink('', rel="stylesheet", type="text/css", href="styles/shCoreDefault.css")
         
        htmlWriter.addScript('', type="text/javascript", src="scripts/shCore.js")
        htmlWriter.addScript('', type="text/javascript", src="scripts/shBrushJScript.js")
        htmlWriter.addScript('', type="text/javascript", src="scripts/shBrushPlain.js")
        htmlWriter.addScript('', type="text/javascript", src="scripts/shBrushCpp.js")
        htmlWriter.addScript('SyntaxHighlighter.all();', type="text/javascript")
        metaAttr = { 'http-equiv':"Content-Type", 'content':"text/html; charset=UTF-8", 'superElement': htmlWriter.head }
        htmlWriter.addTag('meta', '', **metaAttr)

    def generateErrorFile(self, sourceFilename, htmlFilename, errors, currentTime = None):
        logger.info('source: ' + sourceFilename + ' htmlfile: ' + htmlFilename)

        source = safeOpen(sourceFilename, 'r')
        htmlWriter = HTMLWriter(htmlFilename)

        self.addCodeFileHeadingInfo(htmlWriter)
          
        headingElement = htmlWriter.addHeading(1, os.path.basename(sourceFilename))
        if currentTime is not None:
            htmlWriter.addSpan(currentTime, cssClass="time", superElement=headingElement)
        
        relevantErrors = dict()
        lineNumbers = list()
        lastError = None
        prevError = dict()
        nextError = dict()
        
        for e in errors:  
            if e.filename == sourceFilename:
                if len(lineNumbers) == 0 or lineNumbers[-1] != e.lineNumber:  # Not equal to previos errors linenumber
                    lineNumbers.append(e.lineNumber)
                    relevantErrors[e.lineNumber] = e
                    prevError[e.lineNumber] = lastError
                else: # At least two errors on the same line
                    relevantErrors[e.lineNumber].errorDescription = 'Multiple Errors on line'
                    relevantErrors[e.lineNumber].logMessage += ' ' + e.logMessage
            
            if lastError != None and lastError.lineNumber in relevantErrors.keys():
                nextError[lastError.lineNumber] = e
            lastError = e
            
        errorLinks = htmlWriter.addParagraph("", cssClass="errorLinks")
    
    
        ''' Check for errors with linenumber = -1 (default value if linenumber not found) '''
        lineNumber = -1
        
        if lineNumber in lineNumbers: # There's an error about this line
                # put in the reference
                htmlWriter.addAnchor('', name = self.getLinkName(relevantErrors[lineNumber]))
            
                #then the links
                if not nextError.has_key(lineNumber):
                    nextError[lineNumber] = None
                self.addErrorMsg(htmlWriter, relevantErrors[lineNumber], prevError[lineNumber], nextError[lineNumber])
                
                htmlWriter.addHyperLink('#' + self.getLinkName(relevantErrors[lineNumber]), 'Error: line ' + str(lineNumber), superElement = errorLinks)
    
        sourceCode = ''
        startingLine = 1        
        lineNumber = 1

        for line in source:
   
             
            if lineNumber in lineNumbers: # There's an error about this line
                # put in the reference
                addSourceCodeElement(htmlWriter, sourceCode, startingLine)
                htmlWriter.addAnchor('', name = self.getLinkName(relevantErrors[lineNumber]))
            
                #then the links
                if not nextError.has_key(lineNumber):
                    nextError[lineNumber] = None
                self.addErrorMsg(htmlWriter, relevantErrors[lineNumber], prevError[lineNumber], nextError[lineNumber])
                
                htmlWriter.addHyperLink('#' + self.getLinkName(relevantErrors[lineNumber]), 'Error: line ' + str(lineNumber), superElement = errorLinks)

                startingLine = lineNumber
                sourceCode = ''        
 
            sourceCode += line  
            lineNumber += 1
 
        if sourceCode != '':
            addSourceCodeElement(htmlWriter, sourceCode, startingLine)
        
        htmlWriter.write(os.path.join(self.logDirectory, htmlFilename))
        source.close()

    def generateLogsHtmlFile(self, logFilenames, htmlFilename, logDirectory, summaryPage = ''):
        logger.info('log(s): ' + str(logFilenames) + ' htmlfile: ' + htmlFilename + ' summaryPage: ' + summaryPage)
        
        html = safeOpen(os.path.join(logDirectory, htmlFilename), 'w')
        htmlPrefix = safeOpen(os.path.join(self.rootDir, htmlPrefixFile), 'r')
        for line in htmlPrefix:
            html.write(line)

        head = '<head><title>Log File Results</title></head>\n'
        html.write(head)
        html.write('<body>\n<a name="top"/>\n')
        html.write('<a href="' + summaryPage + '">Test Summary</a> ')
        
        for lf in logFilenames:
            logger.info('log:' + lf)
            logFile = safeOpen(os.path.join(logDirectory, lf), 'r')
        
            lastSlash = lf.rfind('/')
            shortenedFilename = lf[lastSlash+1:]
            html.write('<h1>' + shortenedFilename + '</h1><a name="' + shortenedFilename + '"/>\n')
            html.write(htmlCodePrefix())
 
            for line in logFile:
                html.write(line)
 
            html.write(htmlCodePostFix)
            logFile.close()

        bodyEnd = '</body></html>\n'
        html.write(bodyEnd)
        html.close()
    
     
        
        
        

    def translate(self, sourceFilename, htmlFilename, logDirectory):
        self.sourceFilename = sourceFilename
        self.htmlFilename = htmlFilename
        self.logDirectory = logDirectory
        
        if not os.path.isfile(sourceFilename):
            logger.error('SourceCodeToHtml: ' + sourceFilename + ' does not exist')
            raise Exception.IOError
  
        source = safeOpen(sourceFilename, 'r')
        html = safeOpen(htmlFilename, 'w')

        htmlPrefix = safeOpen(self.rootDir + '/' + htmlPrefixFile, 'r')
        
        for line in htmlPrefix:
            html.write(line)
        
        head = '<head><title>' + sourceFilename + '</title></head>'
        html.write(head)
        
        lastSlash = sourceFilename.rfind('/')
        shortenedSourceFilename = sourceFilename[lastSlash+1:]
        bodyStart = '<body>\n<h1>' + shortenedSourceFilename + '</h1>'
            
        html.write(bodyStart)
        html.write(htmlCodePrefix())
        for line in source:
            html.write(line)
 
        html.write(htmlCodePostFix)

        bodyEnd = '</body></html>'
        html.write(bodyEnd)
        
from time import ctime

def sourceCodeToHtmlMain():
    
    scth = SourceCodeToHtml()
    scth.rootDir = '/home/dfann-local/DEVEL'
    logFilenames = ['ioInterfaceConformanceCompile.log', 'ioInterfaceConformanceLink.log']
    htmlFilename = '../logfiles/dummylog.html'
    logDirectory = '/tmp/FACE'
    summaryPage = 'ConformanceReport.html'
    #scth.generateLogsHtmlFile(logFilenames, htmlFilename, logDirectory, summaryPage)
    configuration = None
    scth.generateLogSummaryFile(htmlFilename, configuration, ctime())

    
    
    ''' OLD TEST
    scth = SourceCodeToHtml(sourceFilename='/home/dfann-local/FACEConformanceTestSuite/conformanceInterfaceTest/IO/ioInterfaceConformanceMain.cpp', \
                            htmlFilename='dummy.html', logDirectory = '/tmp', rootDir = '/home/dfann-local/FACEConformanceTestSuite')
    
    errors = list()
    error = ErrorReportGenerator.FACE_Error()
    error.filename = '/home/dfann-local/FACEConformanceTestSuite/conformanceInterfaceTest/IO/ioInterfaceConformanceMain.cpp'
    error.lineNumber = 20
    error.errorDescription = 'Error 1 Description'
    error.logMessage = 'Error 1 Log Message'
    error.errorType = 'errorType goes here'
    errors.append(error)
    
    error2 = ErrorReportGenerator.FACE_Error()
    error2.filename = '/home/dfann-local/FACEConformanceTestSuite/conformanceInterfaceTest/IO/ioInterfaceConformanceMain.cpp'
    error2.lineNumber = 40
    error2.errorDescription = 'Error 2 Description'
    error2.logMessage = 'Error 2 Log Message'
    errors.append(error2)
    
    error3 = ErrorReportGenerator.FACE_Error()
    error3.filename = '/home/dfann-local/FACEConformanceTestSuite/conformanceInterfaceTest/IO/ioInterfaceConformanceMain.cpp'
    error3.lineNumber = 60
    error3.errorDescription = 'Error 3 Description'
    error3.logMessage = 'Error 3 Log Message'
    errors.append(error3)
    
    scth.generateErrorFile('/home/dfann-local/FACEConformanceTestSuite/conformanceInterfaceTest/IO/ioInterfaceConformanceMain.cpp', 'dummy2.html', errors)
    '''


if __name__ == '__main__':
    import ConformanceMain
    ConformanceMain.initializeLogger()
    logger = logging.getLogger(__name__)
    sourceCodeToHtmlMain()
