#!/usr/bin/env python
"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""

__author__ = 'david.fann@vanderbilt.edu'

import Tkinter
from commonTools import listToString, isInteractive, getParentGUI
import constants
import logging

logger = logging.getLogger(__name__)


# Private variable used to keep up with the current splash screen (if one exists)
_splashScreen = None


def show(text_to_show):
    """
    showSplash shows a splash screen if the conformance test is running interactively (from commonTools.IsInteractive)
    It shows it from the parentGUI's location using the SPLASH_X_OFFSET and SPLASH_Y_OFFSET
    :param text_to_show: text to show on the splash screen
    :return:None
    """
    global _splashScreen
    if isInteractive():
        if _splashScreen is not None:
            #self._splashScreen.destroy()
            _splashScreen.setPrompt(text_to_show)

        else:
            parentGUI = getParentGUI()
            _splashScreen = _Splash(parent=parentGUI, text=text_to_show,
                                   bg='#cccccc',
                                   x_offset=parentGUI.x_offset + constants.SPLASH_X_OFFSET,
                                   y_offset=parentGUI.y_offset + constants.SPLASH_Y_OFFSET)

def show_progress(total_ticks, update_percentage):
    """
    Used in coordination with show_splash_screen and tick_progress_on_splash_screen,
    to show a progress percentage on the splash screen.
    Splash screen must already be visible by a call to show_splash_screen (shown only if running interactively)
    :param total_ticks: How many ticks make 100%
    :param update_percentage: How often do you want to update the progress (percentage)
    :return:None
    """
    if isInteractive():
        if _splashScreen is not None:
            _splashScreen.showProgress(total_ticks, update_percentage)

        else:
            logger.warning('Tried to show progress but _splashScreen was None')

def tick_progress():
    """
    Used in coordination show_splash_screen and show_progress_on_splash_screen.
    Both need to be called prior to calling this method.
    :return:None
    """
    if isInteractive():
        global _splashScreen
        if _splashScreen is not None:
            _splashScreen.tick()

        else:
            logger.warning('Tried to tick progress but _splashScreen was None')


def hide():
    """
    Hides a splash screen if it is currently being shown
    :return:None
    """
    global _splashScreen
    if _splashScreen is not None:
        _splashScreen.destroy()
        _splashScreen = None


class _Splash(Tkinter.Toplevel):
    def __init__(self, **params):
        """
        _Splash (Tkinter.Toplevel) is the private class used to display a splash screen
        :param parent: (default: None) parent window to the splash screen,
        :param bg: (default: None) background color for splash screen
        :param timeDelay: (default: None) how long to display splash (if None, will display till destroyed
        :param max_width: (default: 500)
        :param max_height: (default: 200)
        :param width: (default: calculated based on max_width and screen width)
        :param height: (default: calculated based on max_height and screen height)
        :param x_offset: (default: calculated based on max_width and screen width)
        :param y_offset: (default: calculated based on max_height and screen height)

        :return:None
        """

        parent = params.get('parent', None)
        bg = params.get('bg', None)
        topLevelParams = {'bd':'3', 'relief':Tkinter.RAISED}
        widgetParams = dict()
        if bg is not None:
            topLevelParams['bg'] = bg
            widgetParams['bg'] = bg
        Tkinter.Toplevel.__init__(self, parent, **topLevelParams)

        self.next_update = -1
        self.attributes("-topmost", True)  # This puts this on top of other windows
            
        self.timeDelay = params.get('timeDelay', None)
        self.text = params.get('text', 'Please wait...')
        
        # No window borders and decoration
        self.overrideredirect(True)
        
        # Get the size of the windows
        self.max_width = params.get('max_width', 500)
        self.max_height = params.get('max_height', 200)         
        
        default_width = min(int(self.winfo_screenwidth()/2), self.max_width)
        default_height = min(int(self.winfo_screenheight()/2), self.max_height)
        default_xOffset = int((self.winfo_screenwidth()  - default_width)/2)
        default_yOffset = int((self.winfo_screenheight() - default_height)/2)
         
        self.window_width = params.get('width', default_width)
        self.window_height = params.get('height', default_height)
        self.x_offset = params.get('x_offset', default_xOffset)
        self.y_offset = params.get('y_offset', default_yOffset)
        
        
        #Set window size
        geometry_text = "%dx%d+%d+%d" % (self.window_width, self.window_height, self.x_offset, self.y_offset)
        self.geometry(geometry_text)
        
        self.promptFrame = Tkinter.Frame(self, borderwidth=5, height=300, **widgetParams)
        self.promptFrame.pack(fill='both')
        self.label = Tkinter.Label(self.promptFrame, text =self.text, width=50, justify=Tkinter.CENTER, **widgetParams)
        self.label.pack(fill='x')
        self.configure(cursor='watch')
        
        if self.timeDelay is not None:
            self.after(self.timeDelay, self.destroy)
        
        self.update()

    def setPrompt(self, newPrompt):
        self.label.configure(text = newPrompt)
        self.text = newPrompt
        self.update()

    def showProgress(self, total_ticks, update_percentage):
        self.progress_base_prompt = self.text + ' '
        self.current_tick = 0
        self.total_ticks = total_ticks
        self.update_percentage = update_percentage
        self.tick_step  = update_percentage * 0.01 * total_ticks

        if self.tick_step == 0:  # i.e. nothing to progress
            self.tick_step = 1
        if self.tick_step < 1: # original percentage tick is smaller than one tick
            self.tick_step = 1
            self.update_percentage = 100/total_ticks

        self.tick_multiplier = 1
        self.next_update = int(self.tick_step * self.tick_multiplier)
        self.setPrompt(self.progress_base_prompt + '0% Complete')

    def tick(self):
        if self.next_update == -1: # called tick without calling show progress
            return

        self.current_tick += 1
        if self.current_tick > self.next_update:
            self.setPrompt(self.progress_base_prompt + str(self.tick_multiplier * self.update_percentage) + '% Complete')
            self.tick_multiplier += 1
            self.next_update = int(self.tick_multiplier * self.tick_step)




def FACE_ReleaseSplashScreen(ToolMarkingXmlFile, **params):
    """
    Shows the Release Information for the current version of the Test Suite as a splash screen
    :param ToolMarkingXmlFile:XML file containing FACE Tools Markings to use
    :param params: timeDelay (optional): how long to show splash screen
    :return:
    """
    faceMarkings = constants.FACEToolsMarkings(ToolMarkingXmlFile)
    faceMarkings.getToolInfo(maxLineLength=60)
    text = '\n\n' + listToString(faceMarkings.productMarking, '\n')
    height = (len(faceMarkings.productMarking) + 4) * 15

    defaultTimeDelay = constants.VERSION_SPLASH_SCREEN_TIME
    if 'snapshot' in ' '.join(faceMarkings.ToolsVersion).lower():
        defaultTimeDelay = constants.SNAPSHOT_SPLASH_SCREEN_TIME
        # This is not a released version, just an engineering release...don't need a long flash screen

    timeDelay = params.get('timeDelay', defaultTimeDelay)
    params['timeDelay']=timeDelay
    sp = _Splash(text=text, width = 500, height = height, **params)



if __name__ == '__main__':
    # Example FACE_ReleaseSplashScreen

    import ConformanceMain
    from commonTools import listToFilePath

    ROOT = Tkinter.Tk()
    ROOT.withdraw()
    
   
    x_offset = 150
    y_offset = 150
    
    rootDir = ConformanceMain.getRootDir()
    FACE_ReleaseSplashScreen(listToFilePath([rootDir, 'datafiles', 'FACEToolsMarkings.xml'], False), x_offset=x_offset+50, y_offset=y_offset+150, timeDelay=1000)

    ROOT.mainloop()
