#!/usr/bin/env python
"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""

__author__ = 'david.fann@vanderbilt.edu'

from ConformanceMain import getRootDir

if __name__ == '__main__':
    import ConformanceMain
    import logging
    ConformanceMain.initializeLogger()

from commonTools import isInteractive, safeOpen, getParentGUI
import commonTools
import constants
import Dialogs
import logging
import sys
import os
from os import sep, chmod, listdir, walk
from os.path import dirname, join, isdir, isfile, relpath, basename
from datetime import datetime
from glob import glob
from re import match
from platform import system


logger = logging.getLogger(__name__)

class Language:
    '''Constants for programming language types'''
    C="C"
    CPP="C++"
    Ada="Ada"
    Java="Java"
    Make="Make" # Not exactly a programming language...but it is used sometimes

    allLanguages = [C, CPP, Ada, Java]
    standardLib = {C:'libc', CPP:'libstdc++', Ada:'Ada'}
    delimiter = {C: '_', CPP: '::', Ada: '.', Java: '.'}
    source_suffix = {C: '.c', CPP: '.cpp', Ada: '.adb', Java: '.java', Make: '.mk'}
    header_suffix = {C: '.h', CPP: '.hpp', Ada: '.ads'}

    @staticmethod
    def format_case(language, text):
        if language not in Language.allLanguages:
            raise TypeError('language: ' + str(language) + ' should be one of ' +
                            str(Language.allLanguages))
        if not isinstance(text, basestring):
            raise TypeError('text: ' + str(text) + ' should be a string')

        if language == Language.Ada:
            '''
            we were making it standard Ada style, but changed our minds. Leaving it
            commented here, just in case the consortium wants it the other way
            return text.title()
            '''
            return text
        else:
            return text

    begin_single_line_comment = {C: '/*', CPP: '//', Ada: '--', Java: '//', Make: '#'}
    end_single_line_comment = {C: '*/', CPP: '', Ada: '', Java: ''}

    direction_keyword_dict = {C: {'in': '/* in */', 'out': '/* out */', 'inout': '/* inout */'},
                              CPP: {'in': '/* in */', 'out': '/* out */', 'inout': '/* inout */'},
                              Ada: {'in': 'in', 'out':'out', 'inout': 'in out'}}
    type_name_dict = {C: {'boolean'      : 'FACE_boolean',
                          'long'         : 'FACE_long',
                          'long long'    : 'FACE_long_long',
                          'short'        : 'FACE_short',
                          'string'       : 'void ',
                          'unsigned long': 'FACE_unsigned_long'
                          },
                      CPP: {'boolean'      : 'FACE::Boolean',
                            'long'         : 'FACE::Long',
                            'long long'    : 'FACE::LongLong',
                            'short'        : 'FACE::Short',
                            'string'       : 'void ',
                            'unsigned long': 'FACE::UnsignedLong'
                            },
                      Ada: {'boolean'       : 'FACE_Basic_Types.Boolean',
                            'char'          : 'FACE_Basic_Types.Char',
                            'float'         : 'FACE_Basic_Types.Float',
                            'long'          : 'FACE_Basic_Types.Long',
                            'long long'     : 'FACE_Basic_Types.Long_Long',
                            'short'         : 'FACE_Basic_Types.Short',
                            'unsigned short': 'FACE_Basic_Types.Unsigned_Short',
                            'unsigned long' : 'FACE_Basic_Types.Unsigned_Long',
                            'double'        : 'FACE_Basic_Types.Double',
                            'octet'         : 'FACE_Basic_Types.Octet',
                            'string'        : 'System.Address'
                      }
    }


import set_distro_statement

class Profile:
        '''Constants for profile types'''
        General="general"
        SafetyBase="safetyBase"
        SafetyExt="safetyExt"
        Security="security"
        allProfiles = [General, SafetyBase, SafetyExt, Security]

class OpenGLConfiguration:
    OpenGL_not_used ="OpenGL not used"
    OpenGL_with_GLX = "OpenGL with GLX (Direct/Indirect Rendering)"
    OpenGL_ES_with_EGL = "OpenGL ES with EGL (Direct Rendering)"
    OpenGL_SC = "OpenGL SC (Safety Critical)"

    class ConformanceTests:
        OPEN_GL_ES_20 = "OpenGL ES 2.0"
        KHRONOS_GROUP_EGL_12 = "Khronos Group EGL 1.2"
        OPEN_GL_SC_101 = "OpenGL SC 1.0.1"

        all_tests = [OPEN_GL_ES_20, KHRONOS_GROUP_EGL_12, OPEN_GL_SC_101]
        testDir = { OPEN_GL_ES_20: 'OpenGLES20',
                    OPEN_GL_SC_101: 'OpenGLSC101',
                    KHRONOS_GROUP_EGL_12: 'EGL12'}

        validTests = { Profile.General: [OPEN_GL_ES_20, KHRONOS_GROUP_EGL_12],
                       Profile.SafetyBase: [OPEN_GL_SC_101],
                       Profile.SafetyExt: [OPEN_GL_SC_101],
                       Profile.Security: []}

        if constants.getFACEStandardVersion() != constants.FACE_STANDARD_VERSION_2_0:
            validTests[Profile.SafetyBase].append(KHRONOS_GROUP_EGL_12)
            validTests[Profile.SafetyExt].append(KHRONOS_GROUP_EGL_12)


class SegmentInfo():
    def __init__(self):
        self.underTest = False
        self.compilerFlags = ''
        self.linkerFlags = ''
        self.binderFlags = ''
        self.includePath = list()
        self.headerFiles = list()
        self.objectFiles = list()
        self.frameworkfiles = list()

class BaseConfig():
    ''' Class containing the basic configuration information, i.e. compiler, compiler flags, linker, etc... '''
    def __init__(self):
        self.make = ''
        self.compiler = ''
        self.compilerFlags = ''
        self.processorSpecificCompilerFlags = ''
        self.testSpecificCompilerFlags = ''
        self.linker = ''
        self.linkerFlags = ''
        self.binder = ''
        self.binderFlags = ''
        self.archiver = ''
        self.archiverFlags = ''
        self.objectSuffix = ''
        self.execSuffix = ''
        self.architecture = ''
        self.osProfile = ''
        self.osPartition = ''
        self.language = ''
        self.openGLConfiguration = OpenGLConfiguration.OpenGL_not_used
        self.logDirectory = ''
        self.errors = list()
        self.warnings = list()
        self.projectDirectory = ''
        self.specificEnviron = dict()
        self.addedPathDirs = list()
        self.browser = None
        self.dataModelJVMOptions = ''

        self.notes = ''

    class Partition:
        '''Constants for partition types'''
        POSIX="POSIX"
        POSIX_ARINC_653="POSIX+ARINC_653"
        ARINC_653="ARINC_653"
        allPartitions = [POSIX, POSIX_ARINC_653, ARINC_653]

    def __eq__(self, other):
        if (type(self) != type(other)):
            return False

        for d in self.__dict__.keys():
            if d not in other.__dict__:
                logger.debug(d + ' not in both baseconfigs')
                return False
            if self.__dict__[d] != other.__dict__[d]:
                logger.debug(d + ': ' + str(self.__dict__[d]) + ' != ' + str(other.__dict__[d]))
                if d != 'make':
                    return False
                if not('make.py' in self.make and 'make.py' in other.make):  # both using pymake, but if saved from a different root directory, this would cause a false not equal
                    return False

        return True


    def __ne__(self, other):
        return not self == other
 
    def resetTest(self):
        # are these ever used?
        self.errors = list()
        self.warnings = list()
 
    def toHTML(self, htmlWriter, parentElement):
        attributes = dict()
        if parentElement is not None:
            attributes['superElement'] = parentElement
       
        '''    
        menuElement, collapsibleElement = htmlWriter.addCollapsibleMenuItem('Base Configuration Info', headingLevel = 3, **attributes)
                            
        divElement = htmlWriter.addDiv(superElement = collapsibleElement) 
        '''
             
        htmlWriter.addHeading(3, 'Base Configuration Info', **attributes)
        divElement = htmlWriter.addDiv(**attributes) 

        if (self.notes != '') and (self.notes.count('\n') != len(self.notes)):  # i.e. it has more than just \n in it...
            htmlWriter.addConfigElement('Configuration Notes', self.notes, superElement = divElement)
        
        htmlWriter.addConfigElement('O/S partition', self.osPartition, superElement=divElement)
        htmlWriter.addConfigElement('O/S profile', self.osProfile, superElement=divElement)
        htmlWriter.addConfigElement('OpenGL Configuration: ', self.openGLConfiguration, superElement=divElement)
        htmlWriter.addConfigElement('language', self.language, superElement = divElement)
        htmlWriter.addConfigElement('compiler', str(self.compiler), superElement = divElement)
        htmlWriter.addConfigElement( "compiler flags", str(self.compilerFlags), superElement = divElement)
        htmlWriter.addConfigElement( "processor specific compiler flags", str(self.processorSpecificCompilerFlags), superElement = divElement)
        htmlWriter.addConfigElement( "linker", str(self.linker), superElement = divElement)
        htmlWriter.addConfigElement( "linker flags", str(self.linkerFlags), superElement = divElement)
        if self.language == Language.Ada:
            htmlWriter.addConfigElement( "binder", str(self.binder), superElement = divElement)
            htmlWriter.addConfigElement( "binder flags", str(self.binderFlags), superElement = divElement)

        htmlWriter.addConfigElement( "archiver", str(self.archiver), superElement = divElement)
        htmlWriter.addConfigElement( "archiver flags", str(self.archiverFlags), superElement = divElement)
        htmlWriter.addConfigElement( "object file ext", str(self.objectSuffix), superElement = divElement)
        htmlWriter.addConfigElement( "executable file ext", str(self.execSuffix), superElement = divElement)
        htmlWriter.addConfigElement( "build O/S architecture", str(self.architecture), superElement = divElement)
        htmlWriter.addConfigElement( "log directory", str(self.logDirectory), superElement = divElement)
        return divElement

    def getCompilerSpecificDirectory(self, **kw):
        relDir = kw.get('relDir', None)
        compilerSpecificPath = os.path.join(getRootDir(), constants.COMPILER_SPECIFIC_DIR, self.language)
        if not isdir(compilerSpecificPath):
            errorMsg = 'Compiler specific path does not exist: ' + compilerSpecificPath
            if len(self.language) == 0:
                errorMsg += ' language not set'
            logger.error(errorMsg)
            if isInteractive():
                Dialogs.ErrorMessageBox(constants.FACE_TESTSUITE_NAME, errorMsg, parent=getParentGUI())
            raise IOError(errorMsg)

        if relDir is not None:
            compilerSpecificPath = os.path.relpath(compilerSpecificPath,relDir)

        return compilerSpecificPath

    def getGoldLibDirectory(self, **kw):
        relDir = kw.get('relDir', None)
        goldLibPath = os.path.join(getRootDir(), constants.GOLD_LIB_ROOT, self.language)
        if not isdir(goldLibPath):
            errorMsg = 'Gold library path does not exist: ' + goldLibPath
            if len(self.language) == 0:
                errorMsg += ' language not set'
            logger.error(errorMsg)
            if isInteractive():
                Dialogs.ErrorMessageBox(constants.FACE_TESTSUITE_NAME, errorMsg, parent=getParentGUI())
            raise IOError(errorMsg)

        if relDir is not None:
            goldLibPath = os.path.relpath(goldLibPath,relDir)

        return goldLibPath

    def addCompilerSpecificTypesKey(self, environDict, relDir=None):
        specificTypesDir = os.path.join(getRootDir(), constants.COMPILER_SPECIFIC_DIR, self.language, constants.COMPILER_SPECIFIC_TYPES_DIR)
        if relDir is not None:
            specificTypesDir = os.path.relpath(specificTypesDir, relDir)
        environDict['FACE_COMPILER_SPECIFIC_CFLAGS'] = ' -I' + specificTypesDir.replace(sep, '/')

    def goldLibOSIncludePaths(self, relativeDir=None):
        includePaths = list()

        ''' for ARINC, needs to be C or Ada '''
        langDir = Language.C #default is C
        if self.language == Language.Ada:
            langDir = Language.Ada

        ossRootPath = os.path.join(getRootDir(), constants.GOLD_LIB_ROOT, self.language, 'src', 'OSS')

        # Get language standard librariesc
        if self.language == Language.C:
            includePaths.append(os.path.join(ossRootPath, 'libc', self.osProfile))
        elif self.language == Language.CPP:
            includePaths.append(os.path.join(ossRootPath, 'libstdc++', self.osProfile))

        if self.osPartition == self.Partition.POSIX_ARINC_653:
            includePaths.append(os.path.join(ossRootPath, 'POSIX', self.osProfile))
            #includePaths.append(os.path.join(ossRootPath, 'ARINC-653', self.osProfile))
            includePaths.append(os.path.join(ossRootPath, 'ARINC-653', self.osProfile + '_posix_allowed'))

        elif self.osPartition == self.Partition.POSIX:
            includePaths.append(os.path.join(ossRootPath, 'POSIX', self.osProfile))
            #includePaths.append(os.path.join(ossRootPath, 'ARINC-653', 'posix_allowed'))

        elif self.osPartition == self.Partition.ARINC_653:
            includePaths.append(os.path.join(ossRootPath, 'ARINC-653', self.osProfile))
        else:
            logger.error("Don't know what to do with os partition: " + self.osPartition)
            return None

        # Any gold library may need the Compiler Specific paths
        includePaths.append(os.path.join(getRootDir(), constants.COMPILER_SPECIFIC_DIR, self.language, constants.COMPILER_SPECIFIC_TYPES_DIR))

        if relativeDir is not None and os.path.isdir(relativeDir):
            return [os.path.relpath(i, relativeDir) for i in includePaths]

        return includePaths


    def getConformanceTestPath(self):
        conformanceTestPath = sep.join([getRootDir(), 'conformanceInterfaceTests', self.language])
        if not isdir(conformanceTestPath):
            errorMsg = 'Conformance test path does not exist: ' + conformanceTestPath
            if len(getRootDir()) == 0:
                errorMsg += ' Root directory not set'
            if len(self.language) == 0:
                errorMsg += ' language not set'
            raise IOError(errorMsg)

        return conformanceTestPath

    def getConformanceTestSegmentPath(self, segment):
        if segment in [self.Segment.allSegments]:
            return sep.join(self.getConformanceTestPath(), segment)
        else:
            raise TypeError('Segment: ' + segment + ' does not have an associated conformance test')

    def checkMakeValue(self):
        if 'make.py' in self.make:
            self.make = os.path.join(getRootDir(), constants.APP_DIR, 'make.py')

    def writeMakefileEnvironment(self, makeEnv, **kw):
        moreCompilerFlags = kw.get('compilerFlags', '')
        moreLinkerFlags = kw.get('linkerFlags', '')
        moreBinderFlags = kw.get('binderFlags', '')
        moreArchiverFlags = kw.get('archiverFlags', '')
        moreEnvironmentVariables = kw.get('environVariables', dict())
        datatypeConfig = kw.get('datatypeConfig', None)
        useGoldLibPath = kw.get('useGoldLibPath', True)
        relativeDir = kw.get('relativeDir', None)
        goldIncludeFlags = ''
        
        if useGoldLibPath:
            for i in self.goldLibOSIncludePaths(relativeDir):
                goldIncludeFlags += ' -I' + commonTools.getPathSafeObjectName(i)

        makeEnv.write('FACE_CC = ' + commonTools.getPathSafeObjectName(self.compiler)+ '\n')
        faceCflags = self.compilerFlags + ' ' + self.processorSpecificCompilerFlags + ' ' + moreCompilerFlags + ' ' + \
                     goldIncludeFlags + ' ' + self.testSpecificCompilerFlags

        # Handle the '.' in front of object and exe suffixes...but don't put a '.' if there is not suffix
        objectSuffix = self.objectSuffix
        if len(objectSuffix) > 0 and not objectSuffix.startswith('.'):
            objectSuffix = '.' + objectSuffix

        execSuffix = self.execSuffix
        if len(execSuffix) > 0 and not execSuffix.startswith('.'):
            execSuffix = '.' + execSuffix

        faceMake = commonTools.getPathSafeObjectName(self.make)
        if self.make.startswith('python '):
            # if we don't handle this, it will do "python\ pymakePathLocation" escaping the space
            faceMake = 'python ' + commonTools.getPathSafeObjectName(self.make[self.make.find(' ')+1:])

        makeEnv.write('FACE_MAKE = ' + faceMake + '\n')
        makeEnv.write('FACE_CFLAGS = ' + faceCflags + '\n')
        makeEnv.write('FACE_LD = ' + commonTools.getPathSafeObjectName(self.linker) + '\n')
        faceLdFlags = self.linkerFlags + ' ' + moreLinkerFlags
        makeEnv.write('FACE_LD_FLAGS = ' + faceLdFlags + '\n')
        if self.language == Language.Ada:
            makeEnv.write('FACE_BIND = ' + commonTools.getPathSafeObjectName(self.binder) + '\n')
            makeEnv.write('FACE_BIND_FLAGS = ' + self.binderFlags + ' ' + moreBinderFlags +'\n')

        makeEnv.write('FACE_AR = ' + commonTools.getPathSafeObjectName(self.archiver) + '\n')
        faceArFlags =  self.archiverFlags + ' ' + moreArchiverFlags
        makeEnv.write('FACE_AR_FLAGS = ' + faceArFlags + '\n')
        logdir = self.logDirectory
        if relativeDir is not None:
            logdir = os.path.relpath(logdir, relativeDir)
        makeEnv.write('FACE_LOG_DIR = ' + commonTools.getPathSafeObjectName(logdir) + '\n') # log dir needs to end with sep
        makeEnv.write('FACE_LANG_SUFFIX = ' + self.getLanguageSuffix() + '\n')
        makeEnv.write('FACE_OBJ_SUFFIX = ' + objectSuffix + '\n' )
        makeEnv.write('FACE_EXEC_SUFFIX = ' + execSuffix + '\n')
        for key, value in moreEnvironmentVariables.iteritems():
            makeEnv.write(key + ' = ' + value + '\n')
 

    def printSelf(self):
        logger.debug( "Base Configuration Information:")
        logger.debug( "compiler: " + str(self.compiler))
        logger.debug( "compiler flags: " + str(self.compilerFlags))
        logger.debug( "linker: " + str(self.linker))
        logger.debug( "linker flags: " + str(self.linkerFlags))
        if self.language == Language.Ada:
            logger.debug( "binder: " + str(self.binder))
            logger.debug( "binder flags: " + str(self.binderFlags))

        logger.debug( "archiver: " + str(self.archiver))
        logger.debug( "archiver flags: " + str(self.archiverFlags))
        logger.debug( "object suffix: " + str(self.objectSuffix))
        logger.debug( "log directory: " + str(self.logDirectory))


    def getLanguageSuffix(self):
        ''' returns the standard file extension used for source files for the current language '''
        languageSuffix = '.languageSuffixNotSet'
        if self.language == Language.C:
            languageSuffix = '.c'
        elif self.language == Language.CPP:
            languageSuffix = '.cpp'
        elif self.language == Language.Ada:
            languageSuffix = '.adb'
        elif self.language == Language.Java:
            languageSuffix = '.java'
        return languageSuffix
    
    def getLanguageHeaderSuffix(self):
        ''' returns the standard file extension used for source files for the current language '''
        languageSuffix = '.languageHeaderSuffixNotSet'
        if self.language == Language.C:
            languageSuffix = '.h'
        elif self.language == Language.CPP:
            languageSuffix = '.hpp'
        elif self.language == Language.Ada:
            languageSuffix = '.ads'
        return languageSuffix

    def verifyConfiguration(self):
        ''' make sure required fields are present and files/directories exist '''
        logger.info('')
        self.errors = list()
        self.warnings = list()
        
        if self.compiler == '':
            self.errors.append("Must specify the compiler to use")
        elif self.language == Language.Java and self.compiler.find('javac') == -1:
            self.warnings.append('Typically the java compiler is named javac, may want to double check compiler')
        elif self.language == Language.CPP and self.compiler.find('gcc') != -1:
            self.warnings.append(
                'Typically the GNU C++ compiler is named g++ not gcc, may want to double check compiler')

        if self.linker == '' and self.language != Language.Java:
            self.errors.append("Must specify the linker to use")

        if self.compilerFlags.find('-O') >= 0 or self.compilerFlags.find('optimize') >= 0:
            self.warnings.append(
                'Using optimization during conformance testing can cause errors.  Please make sure there are no optimization flags'
            )

        # Until the binder works better, not going to suggest using it.
        #if self.language == Language.Ada:
        #    if self.binder == '':
        #        self.warnings.append('May wish to specify a binder')

        if self.archiver == '':
            self.errors.append("Must specify the archiver to use")
        elif self.language == Language.Java and self.archiver.find('jar') == -1:
            self.warnings.append('Typically the java archiver is named jar, may want to double check archiver')

        if self.logDirectory == '':
            self.errors.append("Must specify a log directory")
        elif not isdir(self.logDirectory):
            self.errors.append("Log directory does not exist");
            
        if self.browser == '':
            self.errors.append("Must specify a browser")
        elif not isfile(self.browser):
            self.warnings.append('Browser not fully specified: ' + self.browser + ', please double-check, if its directory is in path, browser specification may be fine')
        
        if (system() == 'Windows') and (len(self.addedPathDirs) == 0):
            self.warnings.append('No added path directories...need to make sure MinGW directories are in the system path or specified here')
                
        if self.objectSuffix == '':
            self.errors.append('Must specify the object file extension')

        if self.objectSuffix.startswith('.'):
            self.warnings.append('Unless objects end with .' + self.objectSuffix + ' please remove the "." from the object suffix')

        if self.language == Language.Java and self.objectSuffix != 'class':
            self.warnings.append('Typically the java object suffix is "class" not "' + self.objectSuffix + '"')

        if self.execSuffix.startswith('.'):
            self.warnings.append('Unless executables end with .' + self.execSuffix + ' please remove the "." from the exec suffix')


        if self.openGLConfiguration != OpenGLConfiguration.OpenGL_not_used:
            systemLibraryDefintionsDir = os.path.join(getRootDir(),
                                                      constants.COMPILER_SPECIFIC_DIR,
                                                      self.language,
                                                      constants.COMPILER_SPECIFIC_TYPES_DIR)
            required_definition_files = [os.path.join(systemLibraryDefintionsDir, 'EGL', 'eglplatform.h'),
                                         os.path.join(systemLibraryDefintionsDir, 'GLES2', 'gl2platform.h'),
                                         os.path.join(systemLibraryDefintionsDir, 'KHR', 'khrplatform.h')]
            missing_files = list()
            for r in required_definition_files:
                if not os.path.isfile(r):
                    missing_files.append(r)
            if len(missing_files) > 0:
                self.errors.append('To use OpenGL, you must supply the following missing files:\n' +
                                   '\n'.join(missing_files) +
                                   '\nPlease see the installation portion of the manual for more details.')


        if len(self.errors) == 0:
            return True
        else:
            return False     

class SegmentConfig():
    ''' Segment configuration base class, i.e. include files, libraries, etc... '''
    class Segment():
        ''' Constants for segment names '''
        OSS = 'OSS'
        TSS = 'TSS'
        IOS = 'IOS'
        PCS = 'PCS'
        PSS = 'PSS'
        allSegments = [OSS, TSS, IOS, PCS, PSS]
        segmentNames = {OSS: "Operating System Segment",
                        TSS: "Transport Services Segment",
                        PCS: "Portable Components Segment",
                        PSS: "Platform-Specific Segment",
                        IOS: "I/O Services Segment"}


    def __init__(self, segment):
        assert segment in self.Segment.segmentNames
        self.segment = segment
        self.segmentName = self.Segment.segmentNames[segment]
        self.logfiles = list()
        self.standardHeader = "# This is a standard header,\n# this is only a standard header\n\n"
        self.includePath = list()
        self.headerFiles = list()
        self.libraries = list()
        self.errors = list()
        self.warnings = list()
        self.targets = list()
        self.java_conformance_test_results = None # used for Java Conformance Tests (very different from other languages)
        self.testResult = dict()
        self.testAssertion = dict()
        self.targetLogfile = dict()
        self.testDetails = dict()
        self.testCodeFile = dict()
        self.recursive_library_search = True
        self.object_suffix = None

        # Java Approved Framework Additions - July 2019
        self.frameworkFiles = list()

    def __str__(self):
        result_string = ''
        for d in self.__dict__:
            result_string += '[{0}]: [{1}]\n'.format(d, self.__dict__[d])
        return result_string

    def __eq__(self, other):
        return (type(self) is type(other)) and (self.__dict__ == other.__dict__)
    
    def __ne__(self, other):
        return not self == other
    
    def resetTest(self):
        self.logfiles = list()
        self.warnings = list()
        self.errors = list()
        self.targets = list()
        self.targetLogfile = dict()
        self.testResult = dict()
        self.testAssertion = dict()
        self.testDetails = dict()
        self.testCodeFile = dict()
        self.java_conformance_test_results = None
     
    def toHTML(self, htmlWriter, parentElement, language):
        attributes = dict()
        if parentElement is not None:
            attributes['superElement'] = parentElement
            
        htmlWriter.addHeading(3, self.segmentName, **attributes)
        divElement = htmlWriter.addDiv(**attributes)

        if isinstance(self, TsSegmentConfig) and constants.getFACEStandardVersion() in [constants.FACE_STANDARD_VERSION_2_1, constants.FACE_STANDARD_VERSION_2_1_1]:
            htmlWriter.addConfigElement("TS Interface tested", self.interfaceUnderTest,
                                        superElement=divElement)

        if (len(self.includePath) > 0) and (self.includePath[0] != ''):
            htmlWriter.addConfigListElement("include path", self.includePath, headingLevel=3, superElement=divElement)
            
        if (len(self.headerFiles) > 0) and (self.headerFiles[0] != ''):
            htmlWriter.addConfigElement("header files", ', '.join(self.headerFiles), superElement = divElement)
        if (len(self.libraries) > 0) and (self.libraries[0] != ''):
            if language != Language.Java:
                htmlWriter.addConfigElement("libraries/objects",
                                        commonTools.getObjectNamesFromLibraryList(
                                            self.libraries, self.object_suffix, self.recursive_library_search, '', os.path.pathsep),
                                        superElement = divElement)
            else:
                htmlWriter.addConfigElement("class path", os.path.pathsep.join(self.libraries), superElement=divElement)
                htmlWriter.addConfigElement("approved framework", os.path.pathsep.join(self.frameworkFiles), superElement=divElement)

    def verifyConfiguration(self, language):
        ''' make sure required fields are present and files/directories exist '''
        logger.info( 'VerifyConfiguration for ' + self.segmentName)
        self.errors = list()
        self.warnings = list()

        # PCS and PSS, we don't need to look at the include information, just Objects
        verifyIncludes = self.segment in [self.Segment.TSS, self.Segment.IOS, self.Segment.OSS] \
                         and language != Language.Java

        if verifyIncludes:
            if len(self.includePath) == 0 or (len(self.includePath) == 1 and self.includePath[0] == ''):
                    self.warnings.append("No directory listed in the include path")

            for i in self.includePath:
                if i != '' and not os.path.isdir(i):
                    self.errors.append('Include Path Directory ' + i + ' does not exist')

            if constants.getFACEStandardVersion() == constants.FACE_STANDARD_VERSION_2_0:

                if self.segment in [self.Segment.TSS, self.Segment.IOS]:
                    # there must be face/ios.h(hpp,ads) or face/tss.h(hpp,ads) in a user supplied include path
                    header_found = False
                    for i in self.includePath:
                        if i != '' and os.path.isfile(os.path.join(i, self.headerFiles[0])):
                            header_found = True
                            break
                    if not header_found:
                        self.errors.append(self.headerFiles[0] + ' must be found in a specified include path')

            else:
                for h in self.headerFiles:
                    if not os.path.isfile(h):
                        self.errors.append('File ' + h + ' does not exist')

        if (len(self.libraries) == 0) or (len(self.libraries) == 1 and self.libraries[0] == ''):
            self.errors.append("No library/object/class path listed to link against")
        else:
            for lib in self.libraries:
                if not (isfile(lib) or isdir(lib)):
                    self.errors.append("library [" + lib + "] does not exist")
                    # TODO: make sure there is at least one object/library file
                    # if there is a directory, there may not be a object/library file
        return len(self.errors) == 0


    def printSelf(self):
        logger.debug( "include path: " + str(self.includePath))
        logger.debug( "header files: " + str(self.headerFiles))
        logger.debug( "libraries: " + str(self.libraries))

  
    def generateConformanceTest(self, baseConfig, goldLibs, goldLibPaths, conformanceTestDir, environDict = dict()):
        self.conformanceTestDir = conformanceTestDir

        ''' Generate a Makefile.env based on conformance test '''
        compilerFlags = ''
        for p in self.includePath:
            if p != "":
                compilerFlags += ' -I' + commonTools.getPathSafeObjectName(p)
        if 'FACE_CONFORMANCE_TEST_OBJS' not in environDict:
            environDict['FACE_CONFORMANCE_TEST_OBJS'] = ''  # default is no test objs (PSS and PC)
        binderFlags = ''
        if baseConfig.language == Language.Ada:
            binderFlags = compilerFlags

        self.object_suffix = baseConfig.objectSuffix
        environDict['FACE_SEGMENT_OBJS'] = \
            commonTools.getObjectNamesFromLibraryList(self.libraries, self.object_suffix,
                                                      self.recursive_library_search)

        environDict['FACE_GOLD_LIBS'] = commonTools.getObjectNamesStringFromList(goldLibs)


        if self.segment == SegmentConfig.Segment.PCS or self.segment == SegmentConfig.Segment.PSS:
            self.targets = list()
            self.logfiles = list()
            target = self.segmentName.replace(' ', '') + 'ConformanceTest'
            self.targets.append(target)
            self.logfiles.append(join(baseConfig.logDirectory, target + '.log'))
            self.targetLogfile[target] = self.logfiles[-1]
            self.testAssertion[target] = self.segmentName + ' only uses allowed FACE Segment APIs and Operating System Calls'
            self.testDetails[target] = 'Segment can only use external functions specified in the gold standard libraries: ' + ', '.join(goldLibs)
            environDict['FACE_CONFORMANCE_EXECUTABLE'] = target


            target = self.segmentName.replace(" ", "") + "LimitedCallTest"
            self.targets.append(target)
            self.logfiles.append(join(baseConfig.logDirectory, target + '.log'))
            self.targetLogfile[target] = self.logfiles[-1]
            self.testAssertion[target] = self.segmentName + \
                                         ' uses restricted function calls that can only be used in TSS/IOS and within a single UoP'

            self.testDetails[target] = 'If present, the segment uses restricted functions found in section E.4 of the standard, requiring a manual'\
                                       + ' inspection of code to assure that calls are used only for communication within the UoP'

            if baseConfig.osProfile == Profile.SafetyExt: # then need to generate the fork test
                target = self.segmentName.replace(' ', '') + 'ForkConformanceTest'
                self.targets.append(target)
                self.logfiles.append(join(baseConfig.logDirectory, target + '.log'))
                self.targetLogfile[target] = self.logfiles[-1]
                self.testAssertion[target] = self.segmentName + ' uses fork command'
                self.testDetails[target] = 'Must manually inspect code if fork is used in segment to determine if it is followed by exec'
                environDict['FACE_FORK_TEST_EXECUTABLE'] = target

        makeEnv = safeOpen(join(conformanceTestDir, 'Makefile.env'), 'w')
        baseConfig.writeMakefileEnvironment(makeEnv,
                                            compilerFlags = compilerFlags,
                                            binderFlags = binderFlags,
                                            relativeDir=conformanceTestDir,
                                            environVariables = environDict)
        makeEnv.close()
        
    def getSourceFiles(self, baseDir, sourceSuffix):
  
        sourceFiles = list()
        
        for rootdir, subdirs, files in walk(baseDir):
            if commonTools.isHiddenDir(rootdir):
                continue
            globFiles = glob(join(rootdir , '*' + sourceSuffix))
            relDir = relpath(rootdir, baseDir)
            if (relDir == './') or (relDir == '.'):
                relDir = ''
            for g in globFiles:
                sourceFiles.append(relDir + basename(g))                
       
        sourceFiles.sort()
        return sourceFiles
    
    def getTestAssertions(self, conformanceTestDir, **kw):

        assertionFilename = kw.get('assertionFilename', 'assertions.txt')
        assertPrefix = kw.get('assertPrefix', 'Assertion')
        targetPrefix = kw.get('targetPrefix', '')
        ossTarget = kw.get('ossTarget', False)
        ossZeroFill = kw.get('zeroFill', True)
        lowerCase = kw.get('lowerCase', False)

        testAssertions = kw.get('testAssertions', self.testAssertion)
        codeFiles = kw.get('codeFiles', None)
        assertionFile = safeOpen(join(conformanceTestDir, assertionFilename), 'r')
        preLength = len(assertPrefix)
        for line in assertionFile.readlines():
            firstSpace = line.find(' ')            
            if line.startswith(assertPrefix) and firstSpace != -1:
                target = line[preLength:firstSpace]
                if lowerCase:
                    target = target.lower()
                if ossTarget:
                    if ossZeroFill:
                        target = targetPrefix + '%06d' % int(target)
                    else:
                        target = targetPrefix + target
                    if lowerCase:
                        target = target.lower()

                    codeFile = target
                    target = target.replace(sep, '/')
                    if codeFiles is not None:
                        codeFiles[target] = codeFile
                else:
                    target = targetPrefix + target
                testAssertions[target] = line[firstSpace+1:]
                logger.debug('Target: ' + target + ' Assertion: ' + testAssertions[target])

    def createHeaderFile(self, headerFilename):
        ''' Creates a header file that contains includes to all headerfiles associated with the segment '''
        
       
        headerDefine = commonTools.getHeaderDefineConstant(headerFilename)
        
        headerfile = safeOpen(headerFilename, 'w')
        headerfile.write('#ifndef ' + headerDefine + '\n')
        headerfile.write('#define ' + headerDefine + 2*'\n')
        
        for hf in self.headerFiles:
            headerfile.write('#include \"' + hf + '\"\n')
            
        if len(self.headerFiles) == 0:
            headerfile.write('/* WARNING NO HEADER FILES SPECIFIED */')
        
        headerfile.write('\n#endif\n')
        headerfile.close()

class OsSegmentConfig(SegmentConfig):
    ''' contains OS specific methods/variables '''

    class SubTests:
        POSIX = 'POSIX'
        ARINC_653 = 'ARINC 653'
        C_STD_LIB = 'C Std Library'
        CPP_STD_LIB = 'C++ Std Library'
        HMFM = 'HMFM'
        JAVA_GENERAL = 'Java General'
        JAVA_SAFETY = 'Java Safety'
        allTests = [POSIX, ARINC_653, C_STD_LIB, CPP_STD_LIB, HMFM, JAVA_GENERAL, JAVA_SAFETY] + OpenGLConfiguration.ConformanceTests.all_tests
        testDir = {POSIX: 'POSIX', ARINC_653: 'ARINC-653', C_STD_LIB: 'libc', CPP_STD_LIB: 'libstdc++', HMFM: 'HMFM'}
        testDir.update(OpenGLConfiguration.ConformanceTests.testDir)

    def __init__(self):
        SegmentConfig.__init__(self, self.Segment.OSS)
        self.subTests = dict()
        for s in self.SubTests.allTests:
            self.subTests[s] = SegmentInfo()

    def toHTML(self, htmlWriter, parentElement, language):
        attributes = dict()
        if parentElement is not None:
            attributes['superElement'] = parentElement

        htmlWriter.addHeading(3, self.segmentName, **attributes)
        divElement = htmlWriter.addDiv(**attributes)

        for testName in self.subTests.keys():
            if self.subTests[testName].underTest:
                htmlWriter.addHeading(4, 'Testing ' + testName, superElement=divElement)
                subDivElement = htmlWriter.addDiv(superElement=divElement)
                htmlWriter.addConfigElement("additional compiler flags", self.subTests[testName].compilerFlags, superElement=subDivElement)
                htmlWriter.addConfigElement("additional linker flags", self.subTests[testName].linkerFlags, superElement = subDivElement)
                if testName in [OsSegmentConfig.SubTests.JAVA_GENERAL, OsSegmentConfig.SubTests.JAVA_SAFETY]:
                    htmlWriter.addConfigElement("class path", ', '.join(self.subTests[testName].objectFiles),
                                                    superElement = subDivElement)

                else:
                    htmlWriter.addConfigElement("include path", ', '.join(self.subTests[testName].includePath),
                                                superElement = subDivElement)
                    if self.subTests[testName].headerFiles is not None:
                        htmlWriter.addConfigElement("header files", ', '.join(self.subTests[testName].headerFiles),
                                                    superElement = subDivElement)
                    if language != Language.Java:
                        htmlWriter.addConfigElement("libraries/objects",
                                                    commonTools.getObjectNamesFromLibraryList(
                                                        self.subTests[testName].objectFiles,
                                                        self.object_suffix,
                                                        self.recursive_library_search,
                                                        '',
                                                        os.path.pathsep),
                                                    superElement=subDivElement)
                    else:
                        htmlWriter.addConfigElement("Java Classpath",
                                                    os.path.pathsep.join(self.subTests[testName].objectFiles),
                                                    superElement=subDivElement)


    def writeArinc653ConformanceHeaderFile(self, directory, includeFiles):
        headerConstant = 'ARINC653_CONFORMANCE_TEST'
        headerFilename = os.path.join(directory, headerConstant + '.h')
        headerFile = safeOpen(headerFilename, 'w')
        headerPrefix = '#ifndef ' + headerConstant + '_H\n#define ' + headerConstant + '_H\n\n'
        headerFile.write(headerPrefix)

        includedAFile = False
        for i in includeFiles:
            if i == '':
                continue
            includedAFile = True
            headerFile.write('#include "' + i + '"\n')

        if not includedAFile:
            headerFile.write('ERROR: Must specify ARINC-653 Header file(s) in Conformance Test Suite configuration\n')
        headerFile.write('#endif\n\n')

    def generateConformanceTest(self, baseConfig, goldLibs, goldLibPaths, conformanceTestDir, environDict = dict()):
        self.targets = dict()
        self.testAssertions = dict()
        self.logfiles = dict()
        self.targetLogfile = dict()
        self.testDetails = dict()
        self.testCodeFile = dict()
        self.conformanceSubTestDirs = dict()
        self.subTestSubDirs = dict()
        for testName, testValues in self.subTests.iteritems():
            if testValues.underTest == 1:
                conformanceSubTestDir = os.path.join(conformanceTestDir, self.SubTests.testDir[testName])
                self.conformanceSubTestDirs[testName] = conformanceSubTestDir
                logger.info('SubTest: ' + testName + ' Conformance SubDir: ' + conformanceSubTestDir)
                ossZeroFill = True
                lowerCaseTest = False

                if testName == self.SubTests.ARINC_653:
                    ossZeroFill = False # ARINC files don't need the 0000 padding to the test names

                    if baseConfig.language == Language.C:
                        self.writeArinc653ConformanceHeaderFile(conformanceSubTestDir, testValues.headerFiles)
                if baseConfig.language == Language.Ada:
                    if testName != self.SubTests.ARINC_653:
                        lowerCaseTest = True


                conformanceFiles = list()

                if testName == self.SubTests.HMFM:
                    conformanceFiles = commonTools.findFiles(os.path.join(conformanceSubTestDir), '*' + baseConfig.getLanguageSuffix())
                    ossZeroFill = False # HMFM files don't need the 0000 padding to the test names
                elif testName == self.SubTests.ARINC_653 and baseConfig.osPartition == baseConfig.Partition.POSIX_ARINC_653:
                    posix_arinc_itest_subfolder = baseConfig.osProfile + "_posix_allowed"
                    baseConformanceFiles = commonTools.findFiles(os.path.join(conformanceSubTestDir, posix_arinc_itest_subfolder), '*' + baseConfig.getLanguageSuffix())
                    for b in baseConformanceFiles:
                        conformanceFiles.append(os.path.join(posix_arinc_itest_subfolder, b))
                else:
                    baseConformanceFiles = commonTools.findFiles(os.path.join(conformanceSubTestDir, baseConfig.osProfile), '*' + baseConfig.getLanguageSuffix())
                    for b in baseConformanceFiles:
                        conformanceFiles.append(os.path.join(baseConfig.osProfile, b))

                logger.debug('ConformanceFiles: ' + '\n'.join(conformanceFiles))
                conformanceSubDirs = list()
                for c in conformanceFiles:
                    if os.path.dirname(c) not in conformanceSubDirs:
                        conformanceSubDirs.append(os.path.dirname(c))
                logger.debug('Sub dirs: ' + '\n'.join(conformanceSubDirs))
                self.subTestSubDirs[testName] = conformanceSubDirs
                self.testAssertions[testName] = dict()
                self.logfiles[testName] = list()
                self.targetLogfile[testName] = dict()
                self.testDetails[testName] = dict()
                self.testCodeFile[testName] = dict()
                for c in conformanceSubDirs:
                    logger.info('Root Dir: ' + conformanceSubTestDir + ' SubDir: ' + c)
                    fullPath = os.path.join(conformanceSubTestDir, c)
                    if isfile(os.path.join(fullPath, 'assertions.txt')):
                        logger.info('FullPath: ' + fullPath)
                        if baseConfig.language == Language.Ada:
                            if testName == self.SubTests.ARINC_653:
                                subSubDir = os.path.split(c)[-1]
                                targetPrefix = os.path.join(c, 'test_apex_' + subSubDir.lower() + '_')
                            else:
                                targetPrefix = c
                        else:
                            targetPrefix = os.path.join(c, 'Test')

                        self.getTestAssertions(fullPath, targetPrefix=targetPrefix, ossTarget=True, zeroFill=ossZeroFill,
                                               testAssertions=self.testAssertions[testName], codeFiles=self.testCodeFile[testName],
                                               lowerCase=lowerCaseTest)
                self.targets[testName] = self.testAssertions[testName].keys()

                for t in self.targets[testName]:
                    logfile = commonTools.removePathBadChars(testName.replace(' ', '') + '_' +t) + '.log'
                    logfile = os.path.join(baseConfig.logDirectory, logfile)
                    self.logfiles[testName].append(logfile)
                    self.targetLogfile[testName][t] = logfile
                    self.testDetails[testName][t] = 'Please see test code and log file for details'

                    self.testCodeFile[testName][t] = os.path.join(conformanceSubTestDir, self.testCodeFile[testName][t] + baseConfig.getLanguageSuffix())

                    logger.info('Target: ' + t + ' logfile: ' + logfile + ', testCodefile: ' + self.testCodeFile[testName][t])

                makeEnvFilename = os.path.join(conformanceSubTestDir, 'Makefile.env')
                logger.debug('Makeenv file: ' + makeEnvFilename)
                makeEnv = safeOpen(makeEnvFilename, 'w')
                compilerFlags = testValues.compilerFlags
                binderFlags = ""
                if testName == self.SubTests.ARINC_653:
                    compilerFlags += ' -I.' # need to include the conformance subdir in the include path
                                            # to find the Arinc 653 conformance header file

                #if len(testValues.includePath) > 0 and testValues.includePath != ['']:
                for i in testValues.includePath:
                    if i != '':
                        compilerFlags += ' -I' + commonTools.getPathSafeObjectName(i)
                        binderFlags += ' -I' + commonTools.getPathSafeObjectName(i)
                for g in goldLibPaths:
                    compilerFlags += ' -I' + commonTools.getPathSafeObjectName(g)

                if len(goldLibs) > 0:
                    # gold libs are one directory off if OSS testing, since OSS testing goes one more level down
                    # i.e. C++/TSS versus C++/OSS/POSIX
                    environDict['FACE_GOLD_LIBS'] = '../' + ' ../'.join(goldLibs).replace('\\', '/')
                linkerFlags = testValues.linkerFlags
                self.object_suffix = baseConfig.objectSuffix
                environDict['FACE_SEGMENT_OBJS'] = \
                    commonTools.getObjectNamesFromLibraryList(testValues.objectFiles, self.object_suffix,
                                                              self.recursive_library_search)

                options = {'compilerFlags': compilerFlags,
                           'linkerFlags': linkerFlags,
                           'useGoldLibPath': False,
                           'environVariables': environDict,
                           'binderFlags': binderFlags
                           }
                baseConfig.writeMakefileEnvironment(makeEnv, **options)
                makeEnv.close()



    def checkGoldLibLogs(self):
        ''' Make sure that the gold lib was created successfully '''
        noErrors = True
        for l in self.goldLogs:
            logger.info('checking goldLog:' + l)
            if not isfile(l):
                errorMsg = l + ' log file does not exist'
                logger.error(errorMsg)
                if isInteractive():
                    Dialogs.ErrorMessageBox('Error(s) in compiling ARINC 653 GoldLib', errorMsg, parent=getParentGUI())
                    noErrors = False
                else:
                    raise RuntimeError(errorMsg)
                
            for line in safeOpen(l, 'r'):
                if line.lower().find('error'):
                    errorMsg = 'There was an error when compiling the ARINC 653 gold standard library, check ' + l + ' for more information'
                    logger.error(errorMsg)
                    if isInteractive():
                        Dialogs.ErrorMessageBox('Error(s) in compiling ARINC 653 GoldLib', errorMsg, parent=getParentGUI())
                        commonTools.displayFile(l)
                        noErrors = False
                    else:
                        raise RuntimeError(errorMsg)
        return noErrors

    def verifyConfiguration(self, language):
        logger.info('')

        self.errors = list()
        self.warnings = list()

        testing_something = False

        for test, segment_info in self.subTests.iteritems():
            if segment_info.underTest:
                testing_something = True
                error_prefix = "In " + test + " test: "
                if segment_info.includePath is not None:
                    for i in segment_info.includePath:
                        if i == '':
                            continue
                        if not os.path.isdir(i):
                            self.errors.append(error_prefix + "include path [" + i + "] does not exist")

                if segment_info.headerFiles is not None:
                    for h in segment_info.headerFiles:
                        if h == '':
                            continue
                        if not os.path.isfile(h):
                            self.errors.append(error_prefix + "header file [" + h + "] does not exist")

                if not test in [self.SubTests.JAVA_GENERAL, self.SubTests.JAVA_SAFETY]:
                    for o in segment_info.objectFiles:
                        if o == '':
                            continue
                        if not (os.path.isfile(o) or os.path.isdir(o)):
                            self.errors.append(error_prefix + "object file or directory [" + o + "] does not exist")
                else:
                    for o in segment_info.objectFiles:
                        if o == '':
                            continue
                        if not (os.path.isfile(o) or os.path.isdir(o)):
                            self.errors.append(error_prefix + "class path [" + o + "] does not exist")

        if not testing_something:
            self.errors.append("Testing Operating System, but no Operating System Test has been selected."
                               +" Either select a test, or deselect the Test Operating System Segment checkbox")

        return len(self.errors) == 0  # if no errors, then segment has been verified

    def __eq__(self, other):
        if type(self) is not type(other):
            logger.debug('Different types')
            return False

        try:
            for s in self.subTests.keys():
                if self.subTests[s].__dict__ != other.subTests[s].__dict__:
                    logger.debug('Subtest ' + s + ' not equal')
                    return False
        except:
            logger.debug('Comparison through exception, must not be equal')
            return False

        return True

    def __ne__(self, other):
        return not self == other
    
class IoSegmentConfig(SegmentConfig):
    ''' contains methods specific to the IO segment '''
    def __init__(self):
        SegmentConfig.__init__(self, self.Segment.IOS)

         
    def generateConformanceTest(self, baseConfig, goldLibs, goldLibPaths, conformanceTestDir, environDict = dict()):
        self.conformanceTestDir = conformanceTestDir
        
        #copy(join(self.conformanceTestDir, 'base_assertions.txt'), join(self.conformanceTestDir, 'assertions.txt'))  # assertions can get overwritten
        sourceFiles = self.getSourceFiles(conformanceTestDir, baseConfig.getLanguageSuffix())
        self.targets = list()
        self.logfiles = list()
        self.targetLogfile = dict()
        objs = list()
        suffixLength = len(baseConfig.getLanguageSuffix())

        forkTest = (baseConfig.osProfile == Profile.SafetyExt) and \
                   (baseConfig.language != Language.Ada) # then need to generate the fork test
        forkSourceFiles = list()

        for s in sourceFiles:
            target = s[:-suffixLength]
            if target.endswith('ForkConformanceTest') or target.endswith('forkconformancetest'):
                forkSourceFiles.append(s)
                continue # handling fork test at the end of the targets (and only if needed

            if baseConfig.language == Language.Ada and target.startswith('conformance_test-face'):
                logger.debug('ignoring ada target: ' + target)
                continue # Don't need to test the conformance_test-face... as targets

            self.targets.append(target)
            self.logfiles.append(target + 'IoCompile.log')
            self.logfiles.append(target + 'IoLink.log')
            self.targetLogfile[target] = [self.logfiles[-2], self.logfiles[-1]] 
            self.testDetails[target] = 'Please see test code and log file for details'
            self.testCodeFile[target] = s


        if forkTest:
            for s in forkSourceFiles:
                target = s[:-suffixLength]
                self.targets.append(target)
                self.logfiles.append(target + 'IoCompile.log')
                self.logfiles.append(target + 'IoLink.log')
                self.targetLogfile[target] = [self.logfiles[-2], self.logfiles[-1]]
                self.testDetails[target] = 'Test code to determine if fork was used which requires code inspection'\
                        + ' to assure exec immediately follows.  Please see test code and log file for details'
                self.testCodeFile[target] = s
        
        lowerCaseTest = False
        if baseConfig.language == Language.Ada:
            lowerCaseTest = True
        self.getTestAssertions(conformanceTestDir, lowerCase=lowerCaseTest)

        if baseConfig.language != Language.Ada:  # Ada doesn't need a user_includes header file (for now, at least)
            self.createHeaderFile(join(conformanceTestDir, 'user_includes' + baseConfig.getLanguageHeaderSuffix()) )


        environDict['FACE_SRCS'] = commonTools.getObjectNamesStringFromList(sourceFiles)
        SegmentConfig.generateConformanceTest(self, baseConfig, goldLibs, goldLibPaths, conformanceTestDir, environDict)

        self.createHeaderFile(join(conformanceTestDir, 'ioInterfaceConformanceTest' + baseConfig.getLanguageHeaderSuffix()) )
        
class TsSegmentConfig(SegmentConfig):
    ''' TSS specific data/methods '''

    # Used in 2.1 Standard
    class InterfaceUnderTest:
        TS_INTERFACE_UoP='TS Interface TSS UoP'
        ABSTRACTION_LAYER='TSS Type Abstraction'
        TYPE_ABSTRACTION_INTERFACE_UoP='TS Type Abstraction Interface TSS UoP'

        interface_test_subdirectory = {TS_INTERFACE_UoP: 'FACE_TS',
                                       ABSTRACTION_LAYER: 'FACE_TS',
                                       TYPE_ABSTRACTION_INTERFACE_UoP: 'TypeAbstraction_TS'}

    def __init__(self):
        SegmentConfig.__init__(self, self.Segment.TSS)
        self.idlFiles = list()
        self.idlIncludePath = list()
        self.basefiles = list()
        self.faceSpecifiedType = True  # if we have a face specified type interface
        self.untypedInterface = True   # if there is an untyped Interface
        self.registerReadCallback = 1  # if there are registerReadCallback method calls
        self.faceSpecifiedServer = False
        self.interfaceUnderTest = None # Only used in 2.1 Standard
        if constants.getFACEStandardVersion() in [constants.FACE_STANDARD_VERSION_2_1, constants.FACE_STANDARD_VERSION_2_1_1]:
            self.interfaceUnderTest = self.InterfaceUnderTest.TS_INTERFACE_UoP

    def __eq__(self, other):
        return (type(self) is type(other)) and (self.__dict__ == other.__dict__)
    
    def __ne__(self, other):
        return not self == other
    
    def verifyConfiguration(self, language):
        logger.info('')

        SegmentConfig.verifyConfiguration(self, language)

        return len(self.errors) == 0

    def getTestAssertions(self, conformanceTestDir, languageSuffix, lowerCaseTest=False):
        SegmentConfig.getTestAssertions(self, conformanceTestDir, lowerCase=lowerCaseTest)

        ''' Handle the Datatype specific tests here '''
        sourceFiles = glob(os.path.join(conformanceTestDir, '*' + languageSuffix))
        for s in sourceFiles:
            target = os.path.basename(s)[:-len(languageSuffix)]
            if target.startswith('Test_') or target.startswith('test_'):
                testName = target[len('Test_'):]
                if target not in self.testAssertion:
                    self.testAssertion[target] = 'Testing ' + testName + ' for conformance'

    def getTestSpecificCompilerFlags(self):
        testSpecificCompilerFlags = ''
        if self.untypedInterface:
            testSpecificCompilerFlags += ' -DUNTYPED_INTERFACE'
        if self.registerReadCallback:
            testSpecificCompilerFlags += ' -DREGISTER_READ_CALLBACK_DEFINED'
        if self.faceSpecifiedType:
            testSpecificCompilerFlags += ' -DFACE_SPECIFIED_TYPE'
        if self.faceSpecifiedServer:
            testSpecificCompilerFlags += ' -DFACE_SPECIFIED_INTERFACE_SERVER'

        return testSpecificCompilerFlags

    def generateConformanceTest(self, baseConfig, goldLibs, goldLibPaths, conformanceTestDir, environDict = dict()):

        self.conformanceTestDir = conformanceTestDir

        #copy(join(self.conformanceTestDir, 'base_assertions.txt'), join(self.conformanceTestDir, 'assertions.txt'))  # assertions can get overwritten
        sourceFiles = self.getSourceFiles(conformanceTestDir, baseConfig.getLanguageSuffix())
        self.targets = list()
        self.logfiles = list()
        self.targetLogfile = dict()
        objs = list()
        suffixLength = len(baseConfig.getLanguageSuffix())

        forkTest = (baseConfig.osProfile == Profile.SafetyExt) and \
                   (baseConfig.language != Language.Ada) # then need to generate the fork test
        forkSourceFiles = list()

        for s in sourceFiles:
            target = s[:-suffixLength]
            if target.endswith('ForkConformanceTest') or target.endswith('forkconformancetest'):
                forkSourceFiles.append(s)
                continue # handling fork test at the end of the targets (and only if needed

            if self.registerReadCallback != 1 and target.endswith('ReadCallback'):
                continue  # Don't need to test ReadCallback if not supported by TSS
            if baseConfig.language == Language.Ada and target.startswith('conformance_test-face'):
                logger.debug('ignoring ada target: ' + target)
                continue # Don't need to test the conformance_test-face... as targets

            self.targets.append(target)
            self.logfiles.append(target + 'TsCompile.log')
            self.logfiles.append(target + 'TsLink.log')
            self.targetLogfile[target] = [self.logfiles[-2], self.logfiles[-1]]
            self.testDetails[target] = 'Please see test code and log file for details'
            self.testCodeFile[target] = s

        if forkTest:
            for s in forkSourceFiles:
                target = s[:-suffixLength]
                self.targets.append(target)
                self.logfiles.append(target + 'TsCompile.log')
                self.logfiles.append(target + 'TsLink.log')
                self.targetLogfile[target] = [self.logfiles[-2], self.logfiles[-1]]
                self.testDetails[target] = 'Test code to determine if fork was used which requires code inspection'\
                        + ' to assure exec immediately follows.  Please see test code and log file for details'
                self.testCodeFile[target] = s

        if baseConfig.language == Language.Ada:
            lowerCaseTest = True
        else: 
            lowerCaseTest = False
        self.getTestAssertions(conformanceTestDir, baseConfig.getLanguageSuffix(), lowerCaseTest)
        environDict['FACE_SRCS'] = commonTools.getObjectNamesStringFromList(sourceFiles)

        SegmentConfig.generateConformanceTest(self, baseConfig, goldLibs, goldLibPaths, conformanceTestDir, environDict)

        if baseConfig.language != Language.Ada:  # Ada doesn't need a user_includes header file (for now, at least)
            self.createHeaderFile(join(conformanceTestDir, 'user_includes' + baseConfig.getLanguageHeaderSuffix()) )


    def createConformanceTestFiles(self, **kw):
        ''' Creates idl based TS conformance test files...i.e. header file containing 
            includes to all idl headers and a main that calls all idl conformance test methods  '''
        logger.info('kw:' + str(kw))

        
        baseconfig = kw.get('baseConfig', None)
        userIncludes = kw.get('userIncludes', None)
        idlBasenames = kw.get('idlBasenames', None)
        if (baseconfig is None) or (userIncludes is None) or (idlBasenames is None):
            logger.error('userIncludes or rootDir or idlBasenames is None')
            return
        
        cHeader = safeOpen(join(commonTools.listToFilePath([getRootDir(), 'conformanceInterfaceTest', 'TS']) , \
                                       'tss_conformance_test' + baseconfig.getLanguageHeaderSuffix()), 'w')
        cHeader.write('#ifndef TSS_CONFORMANCE_TEST_HEADER\n#define TSS_CONFORMANCE_TEST_HEADER\n' )
        cHeader.write(set_distro_statement.get_distribution_statement(baseconfig.language) + '\n\n')
        
        cHeader.write('\n/* Include files specified by user configuration */')
        for uI in userIncludes:
            cHeader.write('\n#include "' + uI + '"')

 
        cHeader.write('\n/* conformance test methods */')
        cHeader.write('\nvoid tsCommonInterfaceComformanceTest();')
        for iB in idlBasenames:
            cHeader.write('\nvoid ' + iB + '_conformanceTest();')
        
        cHeader.write('\n\n#endif /* TSS_CONFORMANCE_TEST_HEADER */\n')
        cHeader.close()
        
        
        conformanceMain = safeOpen(join(commonTools.listToFilePath([getRootDir(), 'conformanceInterfaceTest', 'TS']),\
                                    'tsInterfaceConformanceMain' + baseconfig.getLanguageSuffix()), 'w')
        conformanceMain.write(set_distro_statement.get_distribution_statement(baseconfig.language) + '\n\n')
        conformanceMain.write('#include "tss_conformance_test' + baseconfig.getLanguageHeaderSuffix() + '" \n\n')
        conformanceMain.write("int main(int argc, char *argv[])\n{\n")
        conformanceMain.write('\ttsCommonInterfaceComformanceTest();\n')
   
        for iB in idlBasenames:
            conformanceMain.write('\t' + iB + '_conformanceTest();\n')

        conformanceMain.write('}\n')
        conformanceMain.close()

import face_specified_code_generator

class DatatypeConfig():

    def __init__(self, **kw):
        self.dataModel = None # DataModel from DataModel.py
        self.uopsToTest = dict() # Dictionary

        # used in verifyConfiguration method
        self.errors = list()
        self.warnings = list()

    def __eq__(self, other):

        equals = self.dataModel == other.dataModel
        equals = equals and self.uopsToTest == other.uopsToTest

        return equals
        #return (self.dataModel == other.dataModel) and \
        #        (self.uopsToTest == other.uopsToTest)

    def __ne__(self, other):
        return not self == other

    def toHTML(self, htmlWriter, parentElement):
        attributes = dict()
        if parentElement is not None:
            attributes['superElement'] = parentElement

        '''
        menuElement, collapsibleElement = htmlWriter.addCollapsibleMenuItem('Base Configuration Info', headingLevel = 3, **attributes)

        divElement = htmlWriter.addDiv(superElement = collapsibleElement)
        '''
        htmlWriter.addHeading(3, 'Data Model Configuration Info', **attributes)
        divElement = htmlWriter.addDiv(**attributes)
        if self.dataModel is None:
            htmlWriter.addConfigElement('Data Model', 'NO DATA MODEL SPECIFIED', superElement=divElement)
            return divElement

        htmlWriter.addConfigElement('Shared Data Model Filename', self.dataModel.sharedDataModelFilename, superElement = divElement)
        htmlWriter.addConfigElement('UoC Data Model Filename', self.dataModel.unitOfConformanceDataModelFilename, superElement = divElement)
        testedUops = list()
        untestedUops = list()
        for uopName, underTest in self.uopsToTest.iteritems():
            if underTest == 1:
                testedUops.append(uopName)
            else:
                untestedUops.append(uopName)

        if len(testedUops) > 0:
            htmlWriter.addConfigElement('UOPs Tested:', ', '.join(testedUops), superElement= divElement)
        if len(untestedUops) > 0:
            htmlWriter.addConfigElement('UOPs Not Tested:', ', '.join(untestedUops), superElement= divElement)

        return divElement


    def verifyConfiguration(self, TestingSegment):
        logger.info('')
        self.errors = list()
        self.warnings = list()

        if self.dataModel is None:
            self.errors.append("Data Model must be specified if testing TSS, PCS, or PSS")

        else:
            if self.dataModel.sharedDataModelFilename == '':
                self.errors.append('Must have a Shared Data Model')
            elif not isfile(self.dataModel.sharedDataModelFilename):
                self.errors.append('Shared Data Model file must exist')

            if self.dataModel.unitOfConformanceDataModelFilename == '':
                self.errors.append('Must have a Unit of Portability Data Model')
            elif not isfile(self.dataModel.unitOfConformanceDataModelFilename):
                self.errors.append('Unit of Portability Data Model file must exist')

            if len(self.errors) == 0 and len(self.dataModel.getUoPNames()) == 0:
                self.errors.append('Data Model must contain at least one unit of portability')

            if len(self.errors) == 0 and TestingSegment:
                for underTest in self.uopsToTest.itervalues():
                    if underTest:
                        break
                if not underTest:
                    self.errors.append('At least one unit of portability must be under test')

        if len(self.errors) == 0:
            self.dataModel.setHashValues()
            return True
        else:
            return False

    def deleteDataModelGeneratedCode(self, baseConfig):
        includeDir = join(getRootDir(), constants.GOLD_LIB_ROOT, baseConfig.language, "include")
        dmIncludeDir = join(includeDir, "face", "dm")
        dmGoldLibHeaderFiles = glob(join(dmIncludeDir, "*" + baseConfig.getLanguageHeaderSuffix()))
        tsGoldLibHeaderFiles = glob(join(includeDir, "face", "*_TSS" + baseConfig.getLanguageHeaderSuffix()))
        tssDataTypeSpecificHeader = join(includeDir, "face", "tss_datatype_specific" + baseConfig.getLanguageHeaderSuffix())
        if isfile(tssDataTypeSpecificHeader):
            tsGoldLibHeaderFiles.append(tssDataTypeSpecificHeader)
        tsGoldLibSourceFiles = glob(join(getRootDir(), constants.GOLD_LIB_ROOT, baseConfig.language, "src", "TSS","*?TSSGoldLib"+baseConfig.getLanguageSuffix()))
        tsConformanceSourceFiles = glob(join(getRootDir(), constants.CONFORMANCE_TEST_ROOT, baseConfig.language, "TSS","Test_*"+baseConfig.getLanguageSuffix()))

        filesToDelete = dmGoldLibHeaderFiles + tsGoldLibHeaderFiles + tsGoldLibSourceFiles + tsConformanceSourceFiles

        logger.debug('files to delete: ' + str(filesToDelete))

        # Delete any generated code from last test
        for d in filesToDelete:
            os.remove(d)



    def initializeTest(self, baseConfig, **kw):

        generate_gold_lib = kw.get('goldLibraryCode', False)
        generate_conformance_test = kw.get('conformanceTestCode', False)
        interface_to_test = kw.get('tssSubtest', None)

        datatype_list = list()
        for uop_name, test_flag in self.uopsToTest.iteritems():
            if test_flag == 1:
                datatype_list.extend(self.dataModel.getUoPMessages(uop_name))

        datatype_list = sorted(list(set(datatype_list)))  # gets rid of duplicates and put them in alphabetical order

        if len(datatype_list) == 0:
            error_msg = "No datatypes in UoPs under test"
            logger.error(error_msg)
            if isInteractive():
                Dialogs.ErrorMessageBox(error_msg)
                return False
            else:
                RuntimeError(error_msg)

        if baseConfig.language != Language.Java:  # Java handled separately...
            if generate_gold_lib:
                face_specified_code_generator.generate_face_specified_gold_standard_library_code(
                    baseConfig.language, self.dataModel.unitOfConformanceDataModelFilename, datatype_list, baseConfig.logDirectory
                )

            if generate_conformance_test:
                face_specified_code_generator.generate_face_specified_conformance_test_code(
                    baseConfig.language, datatype_list, interface_to_test)


        return True


        '''
        self.deleteDataModelGeneratedCode(baseConfig)

        datatypeJarDir = join(getRootDir(), constants.DM_DATATYPE_GEN_SUBDIR)
        conformanceCodeJarDir = join(getRootDir(), constants.DM_CONFORMANCE_CODE_GEN_SUBDIR)
        includeDir = join(getRootDir(), constants.GOLD_LIB_ROOT, baseConfig.language, "include")
        if baseConfig.language == Language.Ada:
            includeDir = join(getRootDir(), constants.GOLD_LIB_ROOT, baseConfig.language, "spec")

        dmIncludeDir = join(includeDir, "face", "dm")
        if baseConfig.language == Language.Java:
            # just dump it in the main java directory...going to make a jar from it anyway
            includeDir = join(getRootDir(), constants.GOLD_LIB_ROOT, baseConfig.language)

        datatypeCommand = ''
        conformanceTestCommand = ''

        std_option = "2.0"
        if constants.getFACEStandardVersion() == constants.FACE_STANDARD_VERSION_2_1:
            std_option = "2.1"

        datatypeCommand = constants.getJavaExe() + ' -jar ' + constants.DM_DATATYPE_GEN_JAR

        if baseConfig.language == Language.C:
            datatypeCommand += ' -c -o "' + includeDir + '" "' + self.dataModel.unitOfConformanceDataModelFilename + '"'
            conformanceTestCommand = constants.getJavaExe() + ' -jar ' + constants.getDMConformanceCodeGenJar() + \
                                     ' "' + self.dataModel.unitOfConformanceDataModelFilename + '" "' + \
                                     getRootDir() + '"'
        elif baseConfig.language == Language.CPP:
            datatypeCommand += ' -p -o "' + includeDir + '" "' + self.dataModel.unitOfConformanceDataModelFilename + '"'
            conformanceTestCommand = constants.getJavaExe() + ' -jar ' + constants.getDMConformanceCodeGenJar() + ' "' + \
                                     self.dataModel.unitOfConformanceDataModelFilename + '" "' + getRootDir() + '"'
        elif baseConfig.language == Language.Ada:
            datatypeCommand += ' -a -o "' + dmIncludeDir + '" "' + self.dataModel.unitOfConformanceDataModelFilename + '"'
            conformanceTestCommand = constants.getJavaExe() + ' -jar ' + constants.getDMConformanceCodeGenJar() + ' "' \
                                     + self.dataModel.unitOfConformanceDataModelFilename + '" "' + getRootDir() + '"'
        elif baseConfig.language == Language.Java:
            datatypeCommand += ' -j -o "' + includeDir + '" "' + self.dataModel.unitOfConformanceDataModelFilename + '"'
            # conformance code gen is done in python
        else:
            logger.error("Don't support code generation with " + baseConfig.language + " yet")

        logger.info('datatype gen cmd: ' + datatypeCommand + '\nconformance gen cmd: ' + conformanceTestCommand)
        # generate datatype header files
        results = commonTools.subprocessWithSpecifiedPath(datatypeCommand, "", datatypeJarDir)
        commonTools.logSubprocessResults(datatypeCommand, results,
                            os.path.join(baseConfig.logDirectory, constants.DM_DATATYPE_GEN_JAR + '_results.log'))
        logger.info('command: ' + datatypeCommand + '\nreturned code: ' + str(results.returncode) + '\nstd out: ' + results.stdout + '\nstd err: ' + results.stderr)
        if results.returncode != 0:
            errorMsg = 'Error in DMDatatypeGenerator\nStd Out: ' + results.stdout + '\nStd Err: ' + results.stderr
            logger.error(errorMsg)
            if isInteractive():
                Dialogs.ErrorMessageBox(constants.FACE_TESTSUITE_NAME, errorMsg, parent=getParentGUI())

            raise RuntimeError(errorMsg)

        # generate datatype specific gold lib/conformance test code
        if baseConfig.language in [Language.C, Language.CPP]:
            results = commonTools.subprocessWithSpecifiedPath(conformanceTestCommand, "", conformanceCodeJarDir)
            commonTools.logSubprocessResults(conformanceTestCommand, results,
                        os.path.join(baseConfig.logDirectory, constants.getDMConformanceCodeGenJar() + "_results.log"))

            logger.info('command: ' + conformanceTestCommand + '\nreturned code: ' + str(results.returncode) + '\nstd out: ' + results.stdout + '\nstd err: ' + results.stderr)
            if results.returncode != 0:
                errorMsg = 'Error in DMConformanceCodeGenerator\nStd Out: ' + results.stdout + '\nStd Err: ' + results.stderr
                logger.error(errorMsg)
                if isInteractive():
                    Dialogs.ErrorMessageBox(constants.FACE_TESTSUITE_NAME, errorMsg, parent=getParentGUI())
                raise RuntimeError(errorMsg)

            if constants.getFACEStandardVersion() == constants.FACE_STANDARD_VERSION_2_1:
                tss_subtest = kw.get('tssSubtest', 'FACE_TS')
                src_folder = os.path.join(getRootDir(), constants.CONFORMANCE_TEST_ROOT, baseConfig.language, 'TSS')
                glob_value =  'Test_*' + Language.source_suffix[baseConfig.language]
                src_glob = os.path.join(src_folder, glob_value)
                dst_folder = os.path.join(src_folder,  tss_subtest)

                old_files = glob(os.path.join(dst_folder, glob_value))
                for o in old_files:
                    os.remove(o)

                commonTools.move_files(src_glob, dst_folder)



        elif baseConfig.language == Language.Ada:
            import idlParser
            pass
            for uop in self.uopsToTest:
                data_types = self.dataModel.getUoPMessages(uop)

                idl_Parser = idlParser.FaceIdlParser(filename=os.path.join(getRootDir(),
                                                                           constants.FACE_SPECIFIED_TSS_IDL))
                gen_sub_dir = os.path.join(getRootDir(), constants.CONFORMANCE_TEST_ROOT,
                                           baseConfig.language, 'TSS')
                gold_spec_dir = os.path.join(getRootDir(), constants.GOLD_LIB_ROOT,
                                            baseConfig.language, 'spec', 'face')

                gold_src_dir = os.path.join(getRootDir(), constants.GOLD_LIB_ROOT,
                                            baseConfig.language, 'src', 'TSS')


                if not os.path.isdir(gen_sub_dir):
                    commonTools.mkdir_p(gen_sub_dir)
                template_filename = os.path.join(getRootDir(), constants.DATAFILES_DIR,
                                                 "Ada_ConformanceTemplates.stg")

                conformance_test_generator = idlParser.ConformanceTestGenerator(idl_Parser.idl_tree)
                conformance_test_generator.generate_face_specified_conformance_tests(data_types, baseConfig.language,
                                                                                     gen_sub_dir, template_filename)

                template_filename = os.path.join(getRootDir(), constants.DATAFILES_DIR,
                                                 "Ada_GoldLibTemplates.stg")
                conformance_test_generator.generate_face_specified_gold_library(data_types, baseConfig.language,
                                                                                gold_spec_dir, gold_src_dir,
                                                                                template_filename)
        '''


if __name__ == '__main__':
    logger.warning('Running __main__ from TestConfiguration.py')

