#!/usr/bin/env python
"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""

__author__ = 'david.fann@vanderbilt.edu'


from commonTools import mkdir_p, listToFilePath, getPathSafeObjectName, subprocessWithSpecifiedPath, \
    safeOpen, logSubprocessResults, getParentGUI
from ConformanceMain import getRootDir
from idlToSourceCodeGenerator import idlToSourceCodeGenerator

import commonTools

import pickle
import copy
import logging
import sys
import os
from os.path import exists, join, isdir, isfile, sep
import Dialogs
import subprocess


logger = logging.getLogger(__name__)


class VerificationSuite(object):
    '''
    classdocs
    '''

    def __init__(self, configuration, params = dict()):
        '''
        Constructor
        '''

        self.configuration = copy.copy(configuration) # going to mess with the configuration sent, so make a full copy
        self.verificationBuildErrors = list()
        self.errorLogs = list()
        
    def generateVerficationConfigFiles(self):
        self.generatePCConfigFiles()
        pass
    
    def _generateConfigFiles(self, configDir, testDir, segmentConfig):

        profiles = dict([('general','general'), ('safety_base','safetyBase'), ('safety_ext', 'safetyExtended'), ('security', 'security')])
        partitions = dict([(self.configuration.baseConfig.Partition.POSIX_ARINC, 'PosixWArinc653Profile'), (self.configuration.baseConfig.Partition.POSIX, 'PosixOnlyProfile')])
        testTypes = ('Pass', 'Fail')

        if not exists(testDir):
            logger.error('verification Path: ' + testDir + " doesn't exist")
            raise IOError('verification Path: ' + testDir + " doesn't exist")
                     
        if not exists(configDir):
            logger.info('creating directory: ' + configDir)
            mkdir_p(configDir)
        
        for partition, partitionSubname in partitions.iteritems():
            for profile, profileSubname in profiles.iteritems():    
                self.configuration.baseConfig.osPartition = partition
                self.configuration.baseConfig.osProfile = profile
                
                if((partition == 'posix') and (profile != 'general')):
                    continue  # posix only exists for a general partition
                for testType in testTypes:
                    objectBaseName = profileSubname + partitionSubname 
                    segmentConfig.libraries = list()
                    segmentConfig.libraries.append((join(testDir, objectBaseName + testType + '.o')))
                    configFilename = join(configDir, objectBaseName + testType + '.cfg')
                    self.configuration.configFile = configFilename
                    pickle.dump( self.configuration, safeOpen(configFilename, 'w'))

    def testNoSegments(self):
        self.configuration.testIOS = False
        self.configuration.testOS = False
        self.configuration.testTS = False
        self.configuration.testPC = False
        self.configuration.testPSS = False
    
    def verifyConfiguration(self):
        if not self.configuration.verifyConfiguration():
            Dialogs.ErrorMessageBox('Conformance Test Suite: Invalid Configuration', 'Must address configuration problems before generating verification test', parent=getParentGUI())
            return False
        return True
    
    def generateOSSConfigFiles(self):
        pass
    
    def generateIOSConfigFiles(self):
        if not self.verifyConfiguration():
            return False
        
        verificationTestDir = listToFilePath([getRootDir(), 'verificationTests', 'IO', self.configuration.baseConfig.language])
        verificationConfigDir = listToFilePath([getRootDir(), 'config', 'verificationTests', 'IO', self.configuration.baseConfig.language])
        if not isdir(verificationConfigDir):
            mkdir_p(verificationConfigDir)


        self.testNoSegments()
        self.configuration.testIOS = True
        self.configuration.ioConfig.includePath = [verificationTestDir]
        self.configuration.ioConfig.headerFiles = [join(verificationTestDir, 'ioInterfaceGoldLib' + self.configuration.baseConfig.getLanguageHeaderSuffix())]
        self.configuration.ioConfig.libraries = [join(verificationTestDir, 'verifyIoInterfaceGoldLib.o')]
        configFilename = join(verificationConfigDir, 'ioInterfaceVerifyPass.cfg')
        self.configuration.configFile = configFilename
        pickle.dump(self.configuration, safeOpen(configFilename, 'w'))

        self.configuration.ioConfig.headerFiles = [join(verificationTestDir, 'ioInterfaceFailLib' + self.configuration.baseConfig.getLanguageHeaderSuffix())]
        self.configuration.ioConfig.libraries = [join(verificationTestDir, 'verifyIoInterfaceFailLib.o')]
        configFilename = join(verificationConfigDir, 'ioInterfaceVerifyFail.cfg')
        self.configuration.configFile = configFilename
        pickle.dump(self.configuration, safeOpen(configFilename, 'w'))

    def generatePCSConfigFiles(self):
        if not self.verifyConfiguration():
            return False

        verificationTestDir = listToFilePath([getRootDir(), 'verificationTests', 'PC', self.configuration.baseConfig.language])
        verificationConfigDir = listToFilePath([getRootDir(), 'config', 'verificationTests', 'PC', self.configuration.baseConfig.language])
        if not isdir(verificationConfigDir):
            mkdir_p(verificationConfigDir)
        
        self.testNoSegments()
        self.configuration.testPC = True
        self.configuration.pcConfig.libraries = [join(verificationTestDir, 'verifyPcInterfaceGoldMain.o')]
        configFilename = join(verificationConfigDir, 'pcInterfaceVerifyPass.cfg')
        self.configuration.configFile = configFilename
        pickle.dump(self.configuration, safeOpen(configFilename, 'w'))

        self.configuration.pcConfig.libraries = [join(verificationTestDir, 'verifyPcInterfaceFailMain.o')]
        configFilename = join(verificationConfigDir, 'pcInterfaceVerifyFail.cfg')
        self.configuration.configFile = configFilename
        pickle.dump(self.configuration, safeOpen(configFilename, 'w'))

    def generatePSSConfigFiles(self):
        if not self.verifyConfiguration():
            return False
        
        verificationTestDir = listToFilePath([getRootDir(), 'verificationTests', 'PS', self.configuration.baseConfig.language])
        verificationConfigDir = listToFilePath([getRootDir(), 'config', 'verificationTests', 'PS', self.configuration.baseConfig.language])
        if not isdir(verificationConfigDir):
            mkdir_p(verificationConfigDir)
        
        self.testNoSegments()
        self.configuration.testPSS = True
        self.configuration.pssConfig.libraries = [join(verificationTestDir, 'verifyPsInterfaceGoldMain.o')]
        configFilename = join(verificationConfigDir, 'psInterfaceVerifyPass.cfg')
        self.configuration.configFile = configFilename
        pickle.dump(self.configuration, safeOpen(configFilename, 'w'))

        self.configuration.pssConfig.libraries = [join(verificationTestDir, 'verifyPsInterfaceFailMain.o')]
        configFilename = join(verificationConfigDir, 'psInterfaceVerifyFail.cfg')
        self.configuration.configFile = configFilename
        pickle.dump(self.configuration, safeOpen(configFilename, 'w'))

    def generateTSSConfigFiles(self):
        if not self.verifyConfiguration():
            return False

        verificationTestDir = listToFilePath([getRootDir(), 'verificationTests', 'TS', self.configuration.baseConfig.language])
        verificationConfigDir = listToFilePath([getRootDir(), 'config', 'verificationTests', 'TS', self.configuration.baseConfig.language])
        if not isdir(verificationConfigDir):
            mkdir_p(verificationConfigDir)
        
        self.testNoSegments()
        self.configuration.testTS = True
        self.configuration.tsConfig.includePath = [verificationTestDir]
        self.configuration.tsConfig.headerFiles = [join(verificationTestDir, 'tsInterfaceGoldLib' + self.configuration.baseConfig.getLanguageHeaderSuffix())]

        self.configuration.tsConfig.libraries = list()
        self.configuration.tsConfig.libraries.append(join(verificationTestDir,'tsInterfaceGoldLib.a'))
        '''
        possibleObjectFiles = [join(verificationTestDir,'verifyUntypedTsInterfaceGoldLib.o'), \
                               join(verificationTestDir,'verifyTypedTsInterfaceGoldLib.o')]
        for p in possibleObjectFiles:
            if isfile(p):
                self.configuration.tsConfig.libraries.append(p)
        '''
        configFilename = join(verificationConfigDir, 'tsInterfaceVerifyPass.cfg')
        self.configuration.configFile = configFilename
        pickle.dump(self.configuration, safeOpen(configFilename, 'w'))

        self.configuration.tsConfig.headerFiles = [join(verificationTestDir, 'tsInterfaceFailLib' + self.configuration.baseConfig.getLanguageHeaderSuffix())]
        
        self.configuration.tsConfig.libraries = list()
        possibleObjectFiles = [join(verificationTestDir,'verifyUntypedTsInterfaceFailLib.o'), \
                               join(verificationTestDir,'verifyTypedTsInterfaceFailLib.o')]
        for p in possibleObjectFiles:
            if isfile(p):
                self.configuration.tsConfig.libraries.append(p)

        configFilename = join(verificationConfigDir, 'tsInterfaceVerifyFail.cfg')
        self.configuration.configFile = configFilename
        pickle.dump(self.configuration, safeOpen(configFilename, 'w'))
    
    def generatePCSConfigFilesPosixArinc(self):
        if not self.verifyConfiguration():
            return False
 
        verificationDir = listToFilePath([getRootDir(), 'verificationTests', 'PC', self.configuration.baseConfig.language])

        profiles = dict([('general','general'), ('safety_base','safetyBase'), ('safety_ext', 'safetyExtended'), ('security', 'security')])
        partitions = dict([(self.configuration.baseConfig.Partition.POSIX_ARINC, 'PosixWArinc653Profile'), (self.configuration.baseConfig.Partition.POSIX, 'PosixOnlyProfile')])
        testTypes = ('Pass', 'Fail')

        if not exists(verificationDir):
            logger.error('verification Path: ' + verificationDir + " doesn't exist")
            raise IOError('verification Path: ' + verificationDir + " doesn't exist")
            
        configPath = listToFilePath([getRootDir(), 'config', 'verificationTests', 'PC', self.configuration.baseConfig.language])
                     
        if not exists(configPath):
            logger.info('creating directory: ' + configPath)
            mkdir_p(configPath)
        
        self.configuration.testPC = True
        for partition, partitionSubname in partitions.iteritems():
            for profile, profileSubname in profiles.iteritems():    
                self.configuration.baseConfig.osPartition = partition
                self.configuration.baseConfig.osProfile = profile
                
                if((partition == 'posix') and (profile != 'general')):
                    continue  # posix only exists for a general partition
                for testType in testTypes:
                    objectBaseName = profileSubname + partitionSubname 
                    self.configuration.pcConfig.libraries = list()
                    self.configuration.pcConfig.libraries.append((join(verificationDir, objectBaseName + testType + '.o')))
                    configFilename = join(configPath, objectBaseName + testType + '.cfg')
                    self.configuration.configFile = configFilename
                    pickle.dump( self.configuration, safeOpen(configFilename, 'w'))
    
    def buildOSSVerificationObjects(self, **params): 
        posixVerificationDir = listToFilePath([getRootDir(), 'verificationTests', 'OS', 'POSIX', self.configuration.baseConfig.osProfile])
        self.buildVerificationMakefileEnv(posixVerificationDir, **params)
        posixBuildResults = self.buildVerificationObjectsDir(posixVerificationDir, **params)
        arincBuildResults = None
        
        if self.baseConfig.osPartition == self.baseConfig.Partition.POSIX_ARINC:
            arincVerificationDir = listToFilePath([getRootDir(), 'verificationTests', 'OS', 'ARINC-653', self.configuration.baseConfig.osProfile])
            self.buildVerificationMakefileEnv(posixVerificationDir, **params)
            arincBuildResults = self.buildVerificationObjectsDir(arincVerificationDir, **params)
            
        
        return posixBuildResults  # TODO:  Need to join posix and arinc
        
        ''' Need to put everything in the return??
        return self.buildVerificationObjectsDir(posixVerificationDir, **params)
        '''
         
    def buildIOSVerificationObjects(self, **params):
        verificationDir = listToFilePath([getRootDir(), 'verificationTests', 'IO', self.configuration.baseConfig.language])
        self.buildVerificationMakefileEnv(verificationDir, **params)
        return self.buildVerificationObjectsDir(verificationDir, **params)

    def buildPCSVerificationObjects(self, **params):
        verificationDir = listToFilePath([getRootDir(), 'verificationTests', 'PC', self.configuration.baseConfig.language])
        tsVerificationDir = listToFilePath([getRootDir(), 'verificationTests', 'TS', self.configuration.baseConfig.language])
        self.generateTSSVerificationFiles(tsVerificationDir)  # Need TS files to build PC
        params['includePath'] = [tsVerificationDir]
        self.buildVerificationMakefileEnv(verificationDir, **params)
        return self.buildVerificationObjectsDir(verificationDir, **params)

    def buildPSSVerificationObjects(self, **params):
        verificationDir = listToFilePath([getRootDir(), 'verificationTests', 'PS', self.configuration.baseConfig.language])
        tsVerificationDir = listToFilePath([getRootDir(), 'verificationTests', 'TS', self.configuration.baseConfig.language])
        self.generateTSSVerificationFiles(tsVerificationDir)  # Need TS files to build PC
        params['includePath'] = [tsVerificationDir]
        self.buildVerificationMakefileEnv(verificationDir, **params)
        return self.buildVerificationObjectsDir(verificationDir, **params)

    def buildTSSVerificationObjects(self, **params):
        verificationDir = listToFilePath([getRootDir(), 'verificationTests', 'TS', self.configuration.baseConfig.language])

        self.generateTSSVerificationFiles(verificationDir, **params)

        '''
        if self.configuration.datatypeConfig.faceSpecifiedType:
            self.buildIDLBasedObjects(verificationDir, **params)
        else:
            self.buildTSSnoIDLMakefileEnv(verificationDir, **params)
        '''
        return self.buildVerificationObjectsDir(verificationDir, **params)
    
    def generateTSSVerificationFiles(self, verificationDir, **params):
        self.configuration.datatypeConfig.initializeTest(self.configuration.baseConfig.language)
        idlGen = idlToSourceCodeGenerator(baseConfig = self.configuration.baseConfig, tsConfig = self.configuration.tsConfig, datatypeConfig = self.configuration.datatypeConfig)
        idlGen.generateSourceCodeFromIDL()
        idlGen.copyTSGoldStandardCode(verificationDir)
        idlGen.generateTSGoldLibMakefileEnv(verificationDir)
        idlGen.generateVerificationTestCasesFile(verificationDir)
        idlGen.generateTSDatatypesHeader(verificationDir)
        idlGen.generateTSGoldHeaderFile(verificationDir)
        self.generateTSFailHeaderFile(verificationDir)
        
    def generateTSFailHeaderFile(self,verificationDir):
        goldHeaderFilename = join(verificationDir, 'tsInterfaceGoldLib' + self.configuration.baseConfig.getLanguageHeaderSuffix())
        failHeaderFilename = join(verificationDir, 'tsInterfaceFailLib' + self.configuration.baseConfig.getLanguageHeaderSuffix())
        
        goldHeaderFile = safeOpen(goldHeaderFilename, 'r')
        failHeaderFile = safeOpen(failHeaderFilename, 'w')
        
        goldLines = goldHeaderFile.readlines()
        failSuffix = '_FailLib' + self.configuration.baseConfig.getLanguageHeaderSuffix() + '"\n'
        for line in goldLines:
            if 'tss_common' in line:
                failHeaderFile.write('#include "tss_common' + failSuffix)
            elif 'untyped_tss' in line:
                failHeaderFile.write('#include "untyped_tss' + failSuffix)
            else:
                failHeaderFile.write(line)
        
        goldHeaderFile.close()
        failHeaderFile.close()
        
        
    
    def buildTSSnoIDLMakefileEnv(self, verificationDir, **params):
        useGoldLibs = params.get('useGoldLibs', True)
        includePath = params.get('includePath', list())
        compilerFlags = ''
        makeEnv = safeOpen(join(verificationDir, 'Makefile.env'), 'w')
        for p in includePath:
            if p.rstrip() != '':
                compilerFlags += ' -I' + p
        
        if useGoldLibs:
            compilerFlags += ' -DUSING_FACE_GOLD_LIBS'
        envParams = { 'useGoldLibPath': useGoldLibs, 'compilerFlags':compilerFlags }
        
        self.configuration.baseConfig.writeMakefileEnvironment(makeEnv, **envParams)
        makeEnv.close()


    def buildVerificationMakefileEnv(self, verificationDir, **params):
        useGoldLibs = params.get('useGoldLibs', True)
        includePath = params.get('includePath', None)
        compilerFlags = ''

        if includePath is not None:
            for i in includePath:
                compilerFlags += ' -I' + getPathSafeObjectName(i)

        if useGoldLibs:
            compilerFlags += ' -DUSING_FACE_GOLD_LIBS'
        envParams = { 'useGoldLibPath': useGoldLibs, 'compilerFlags':compilerFlags, 'datatypeConfig':self.configuration.datatypeConfig, 'compilerFlags':compilerFlags }
        
        makeEnv = safeOpen(join(verificationDir, 'Makefile.env'), 'w')
        self.configuration.baseConfig.writeMakefileEnvironment(makeEnv, **envParams)
        makeEnv.close()


    
    def buildVerificationObjectsDir(self, verificationDir, **params):
        testCases = self.getTestCases(verificationDir)
        
        subprocessWithSpecifiedPath(self.configuration.baseConfig.make + ' clean', self.configuration.baseConfig.addedPathDirs, verificationDir)
        self.verificationBuildErrors = list()
        verificationObjectsBuiltCorrectly = True       
        for t in testCases:
            makeCmd = self.configuration.baseConfig.make + ' ' + t
            
            result = subprocessWithSpecifiedPath(makeCmd, self.configuration.baseConfig.addedPathDirs, verificationDir)
            logFilename = join(self.configuration.baseConfig.logDirectory, t[:-1] + 'log')
            logSubprocessResults(makeCmd, result, logFilename, otherInfo={'workingDir':verificationDir, 'pathList':self.configuration.baseConfig.addedPathDirs})

            logger.info('cmd: [' + makeCmd + '], stdout: [' + result.stdout + '] stderr: [' + result.stderr + ']')

            if result.returncode != 0:
                errorMsg = 'Error: ' + join(verificationDir, t) + ' failed to build: ' + result.stderr
                self.verificationBuildErrors.append(errorMsg)
                self.errorLogs.append(join(self.configuration.baseConfig.logDirectory, t[:-1] + 'log'))
                verificationObjectsBuiltCorrectly = False                

        return verificationObjectsBuiltCorrectly        
                
    def buildPosixArincVerificationObjects(self, **params):

        verificationDir = listToFilePath([getRootDir(), 'verificationTests', 'PC', self.configuration.baseConfig.language])
        useGoldLibs = params.get('useGoldLibs', True)
        
        makeEnv = safeOpen(join(verificationDir, 'Makefile.env'), 'w')
        compilerFlags = ''
        for p in self.configuration.pcConfig.includePath:
            if p.rstrip() != '':
                compilerFlags += ' -I' + p
        
        if useGoldLibs:
            compilerFlags += ' -DUSING_FACE_GOLD_LIBS'
        envParams = { 'useGoldLibPath': useGoldLibs, 'compilerFlags':compilerFlags }
        
        self.configuration.baseConfig.writeMakefileEnvironment(makeEnv, **envParams)
        makeEnv.close()
        
        testCases = self.getTestCases(verificationDir)
        
        subprocess.call(self.configuration.baseConfig.make + ' clean', cwd = verificationDir, shell = True)

        verificationObjectsBuiltCorrectly = True       
        for t in testCases:
            makeCmd = self.configuration.baseConfig.make + ' ' + t
            result = subprocess.Popen(makeCmd, cwd = verificationDir, shell = True, stderr=subprocess.PIPE)
            result.wait()
            stdout, stderr = result.communicate()
            if stdout is None:
                stdout = ''
            if stderr is None:
                stderr = ''
            logger.info('cmd: [' + makeCmd + '], stdout: [' + stdout + '] stderr: [' + stderr + ']')
                            
            failCase = (t.find('Fail') != -1)  # fail case if fail is in the name of the test
            passCase = not failCase

            if useGoldLibs and failCase and result.returncode == 0:
                self.verificationBuildErrors.append('Warning: ' + join(verificationDir, t) + ' did not fail to build')


            if result.returncode != 0:
                
                if passCase or (failCase and not useGoldLibs):  # in both these cases, the test should build
                    errorMsg = 'Error: ' + join(verificationDir, t) + ' failed to build\n'
                    self.verificationBuildErrors.append(errorMsg)
                    self.errorLogs.append(join(self.configuration.baseConfig.logDirectory, t[:-1] + 'log'))
                    verificationObjectsBuiltCorrectly = False
        return verificationObjectsBuiltCorrectly        

    def getTestCases(self, testDir):
        testCaseFile = safeOpen(join(testDir, 'testCases.txt'), 'r')
        testCases = list()
        testCaseLines = testCaseFile.readlines()
        for t in testCaseLines:
            t = t.rstrip()
            if t != '':
                testCases.append(t.rstrip())
        return testCases
        
if __name__ == '__main__':
    import ConformanceMain
    ConformanceMain.initializeLogger()
    logger = logging.getLogger(__name__)
    logger.warning('Running from VerficationSuite.py')

    rootDir = getRootDir()
    

    configFile = rootDir + sep + 'ConformanceTestSuiteApplication' + sep + 'verificationTestBaseConfig.cfg' 
    testConfiguration = pickle.load(safeOpen(configFile, "r"))

    testConfiguration.baseConfig.language = 'C++'
    testConfiguration.baseConfig.logDirectory = rootDir + sep + 'logfiles'
    vs = VerificationSuite(testConfiguration)
    
    vs.buildPCSVerificationObjects()
    vs.generatePCSConfigFiles()
    
    

        
        

        
