#!/usr/bin/env python
"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""

__author__ = 'david.fann@vanderbilt.edu'

import os.path
import ConformanceMain
import sys
egg_path = os.path.join(ConformanceMain.getRootDir(),'face_conformance_app', 'pyparsing-2.0.1-py2.7.egg')
sys.path.append(egg_path)
egg_path = os.path.join(ConformanceMain.getRootDir(),'face_conformance_app', 'stringtemplate3-3.1-py2.7.egg')
sys.path.append(egg_path)
egg_path = os.path.join(ConformanceMain.getRootDir(),'face_conformance_app', 'setuptools-0.6c11-py2.7!\\')
sys.path.append(egg_path)


from pyparsing import Literal, CaselessLiteral, Word, Upcase, OneOrMore, ZeroOrMore, \
        Forward, NotAny, delimitedList, oneOf, Group, Optional, Combine, alphas, nums, restOfLine, cStyleComment, \
        alphanums, printables, empty, quotedString, ParseException, ParseResults, Keyword, hexnums
import stringtemplate3

import logging
import os
import copy
import glob

import TestConfiguration
import commonTools
import constants
import ConformanceMain
import set_distro_statement


if __name__ == '__main__':
    ConformanceMain.initializeLogger()

logger = logging.getLogger(__name__)

import idlParser


class AdaTree(idlParser.IdlTree):
    def __init__(self):
        idlParser.IdlTree.__init__(self)
        pass


class PackageNode(idlParser.Node):
    def __init__(self, loc, tokens, **kw):
        idlParser.Node.__init__(self, loc=loc, **kw)

        if tokens[0] != 'package':
            raise RuntimeError("Unexpected tokens: " + str(tokens))

        self.name = kw.get('fully_specified_name', 'undefined package name')


class ProcedureNode(idlParser.OperationNode): # Procedure is an Operation in IDL
    def __init__(self, loc, tokens, **kw):
        # Must use unique __init__ since tokens will be different since we are parsing an Ada Spec
        # Not an IDL file...but all other methods in base class should be okay
        idlParser.Node.__init__(self, loc=loc, **kw)

        if tokens[0] != 'procedure':
            raise RuntimeError("Unexpected tokens: " + str(tokens))

        self.name = tokens[1]
        self.return_type = None
        self.conformance_test_node = True


    def additional_modules(self):
        module_names = list()

        if self.name.find('.') >= 0:  # specified a module in the name
            module_name = ''
            for f in self.name.split('.'):
                if len(module_name) > 0:
                    module_name += '.' + f
                else:
                    module_name = f
                module_names.append(module_name)

            module_names.pop()  # last one is the name of the procedure...don't need it

        for c in self.children:
            if isinstance(c, ParameterItemNode):
                module_names += c.additional_modules()

        return module_names


class ParameterItemNode(idlParser.ParameterNode):
    # Equivalent to the IDL Parameter Node, however, must have unique __init__ since the tokens
    # will be different in an Ada Spec file than an IDL file
    def __init__(self, loc, tokens, **kw):
        idlParser.Node.__init__(self, loc=loc, **kw)

        if not(tokens[1] == ":"):
            raise RuntimeError("Unexpected tokens: " + str(tokens))
        self.type_name = kw.get('fully_specified_name', 'undefined parameter item type name')
        self.name = tokens[0]
        self.direction = ' '.join(tokens[2])
        if self.direction == 'in out':
            self.direction = 'inout' # Put it in the IDL format
        logger.info('tokens: ' + str(tokens) + ' => ' + self.toString())

    def additional_modules(self):
        module_names = list()
        if self.type_name.find('.') >= 0: # specified a module in the type name
            module_name = ''
            for f in self.type_name.split('.'):
                if len(module_name) > 0:
                    module_name += '.' + f
                else:
                    module_name = f
                module_names.append(module_name)

            module_names.pop()  # last one is the name of the procedure...don't need it

        return module_names


class RecordNode(idlParser.Node):
    def __init__(self, loc, tokens, **kw):
        idlParser.Node.__init__(self, loc=loc, **kw)

        #if not((tokens[2] == 'is') or (tokens[4] == 'record')):
        #    raise RuntimeError("Unexpected tokens: " + str(tokens))
        self.discriminate = kw.get("discriminate", None)
        self.name = tokens[1]


class RecordItemNode(idlParser.Node):
    def __init__(self, loc, tokens, **kw):
        idlParser.Node.__init__(self, loc=loc, **kw)

        if not((tokens[1] == ':') or (tokens[4] == 'record')):
            raise RuntimeError("Unexpected tokens: " + str(tokens))

        self.name = tokens[0]
        self.type_name = kw.get('fully_specified_name', 'undefined record item type name')


class WithNode(idlParser.Node):
    def __init__(self, loc, tokens, **kw):
        idlParser.Node.__init__(self, loc=loc, **kw)

        self.name = kw.get('fully_specified_name', 'undefined with name')


class AdaParser():
    def __init__(self, **kw):
        # class variables
        self.bnf = None
        self.reset_variables()
        self.initialize_parser()

    def reset_variables(self):
        self.unhandled_definitions = 0
        self.ada_tree = AdaTree()
        self.tokens = None
        self.fully_specified_name_list = list()
        self.record_discriminate = dict()

    def parse_ada_spec_file(self, ada_spec_filename):
        if not os.path.isfile(ada_spec_filename):
            raise IOError('Ada Spec File does not exist: ' + ada_spec_filename)

        self.directory = os.path.dirname(ada_spec_filename)

        spec_file = open(ada_spec_filename, 'r')
        lines = '\n'.join(spec_file.readlines())
        # Need to reset values
        self.reset_variables()
        return self.parse_string(lines)

    def generate_conformance_tests(self, ada_spec_filename, target_directory, parse_spec_file = True,
                                   template_file=os.path.join(ConformanceMain.getRootDir(),
                                                      constants.DATAFILES_DIR, 'Ada_ConformanceTemplates.stg'),
                                   added_assertion_text=''):

        # TODO: Finish fleshing out this routine and run the dummy.ads test to verify that it works
        spec_file_parsed = True
        if parse_spec_file:
            spec_file_parsed = self.parse_ada_spec_file(ada_spec_filename)

        if spec_file_parsed:
            ada_conformance_body_basename, extension = os.path.splitext(os.path.basename(ada_spec_filename))

            template = stringtemplate3.StringTemplateGroup(name='conformance', file=open(template_file, 'r'))
            assertion_file = open(os.path.join(target_directory,
                                               ada_conformance_body_basename + '_assertions.txt'), 'w')
            all_tests_file = open(os.path.join(target_directory,
                                               ada_conformance_body_basename + '_all_tests.txt'), 'w')
            test_num = 1
            language = TestConfiguration.Language.Ada
            with_modules = list()
            for current_node in iter(self.ada_tree):
                if isinstance(current_node, WithNode):
                    with_modules.append(current_node.name)


                if current_node.conformance_test_node:
                    logger.debug('writing conformance test for: ' + current_node.toString())

                    module_nodes = current_node.get_specific_node_type_parents([PackageNode])
                    module_names = with_modules
                    conformance_module_names = list()

                    for n in module_nodes:
                        full_module_name = n.get_fully_specified_name(language)
                        module_name = ''
                        for f in full_module_name.split('.'):
                            if len(module_name) > 0:
                                module_name += '.' + f
                            else:
                                module_name = f
                            module_names.append(module_name)

                    module_names += current_node.additional_modules()

                    test_basename = 'test_' +ada_conformance_body_basename + '_' + str(test_num)
                    test_basename = test_basename.replace('-', '_')
                    logger.debug('Module Names: ' + ', '.join(module_names))
                    logger.debug('test basename: ' + test_basename)

                    test_code = current_node.get_conformance_test_code(template, language)
                    code_template = template.getInstanceOf('conformanceSourceFile')
                    code_template['TEST_CODE_SECTION1'] = test_code[0]
                    code_template['TEST_CODE_SECTION2'] = test_code[1]
                    code_template['FACE_PACKAGES'] = ', '.join(set(module_names))
                    code_template['CONFORMANCE_TEST_FACE_PACKAGES'] = ', '.join(conformance_module_names)
                    code_template['TEST_NAME'] = test_basename

                    source_filename = test_basename + TestConfiguration.Language.source_suffix[language]
                    source_file = open(os.path.join(target_directory, source_filename), 'w')
                    source_file.write(set_distro_statement.get_distribution_statement(
                        language, source_filename))
                    source_file.write(str(code_template))
                    source_file.close()

                    all_tests_file.write(2*'\n' + TestConfiguration.Language.begin_single_line_comment[language] +
                                         80 * '#' + TestConfiguration.Language.end_single_line_comment[language] + '\n')

                    all_tests_file.write(TestConfiguration.Language.begin_single_line_comment[language] + ' ' +
                                         test_basename + TestConfiguration.Language.source_suffix[language] +
                                         ' Test Code ' +
                                         TestConfiguration.Language.end_single_line_comment[language] + '\n')

                    all_tests_file.write(str(code_template))

                    assertion_file.write('Assertion' + str(test_num) + ' Testing ' +
                                         current_node.get_fully_specified_name(language) + added_assertion_text +
                                         ' is conformant\n')
                    test_num += 1

        return spec_file_parsed


    def initialize_parser(self):
        # punctuation
        colon  = Literal(":")
        dash   = Literal("-")
        lbrace = Literal("{")
        rbrace = Literal("}")
        lbrack = Literal("[")
        rbrack = Literal("]")
        lparen = Literal("(")
        rparen = Literal(")")
        equals = Literal("=")
        comma  = Literal(",")
        dot    = Literal(".")
        hashtag= Literal("#")
        slash  = Literal("/")
        bslash = Literal("\\")
        star   = Literal("*")
        semi   = Literal(";")
        langle = Literal("<")
        rangle = Literal(">")
        quote  = Literal('"')
        underscore = Literal("_")
        single_quote = Literal("'")

        # keywords
        is_        = Keyword("is")
        new_       = Keyword("new")
        range_     = Keyword("range")
        type_      = Keyword("type")
        end_       = Keyword("end")
        package_   = Keyword("package")
        procedure_ = Keyword("procedure")
        constant_  = Keyword("constant")
        record_    = Keyword("record")
        pragma_     = Keyword("pragma")
        convention_ = Keyword("Convention")
        private_   = Keyword("private")
        with_      = Keyword("with")
        subtype_   = Keyword("subtype")
        array_     = Keyword("array")
        of_        = Keyword("of")
        for_       = Keyword("for")
        use_       = Keyword("use")
        alignment_ = Keyword("Alignment")
        size_      = Keyword("Size")
        renames_   = Keyword("renames")
        any_       = Keyword("any")
        attribute_ = Keyword("attribute")
        boolean_   = Keyword("boolean")
        case_      = Keyword("case")
        char_      = Keyword("char")
        const_     = Keyword("const")
        context_   = Keyword("context")
        default_   = Keyword("default")
        double_    = Keyword("double")
        enum_      = Keyword("enum")
        exception_ = Keyword("exception")
        false_     = Keyword("FALSE")
        fixed_     = Keyword("fixed")
        float_     = Keyword("float")
        interface_ = Keyword("interface")
        in_        = Keyword("in")
        include_   = Keyword("include")
        long_      = Keyword("long")
        long_long_ = Keyword("long long")
        module_    = Keyword("module")
        object_    = Keyword("Object")
        octet_     = Keyword("octet")
        oneway_    = Keyword("oneway")
        out_       = Keyword("out")
        raises_    = Keyword("raises")
        readonly_  = Keyword("readonly")
        sequence_  = Keyword("sequence")
        short_     = Keyword("short")
        string_    = Keyword("string")
        struct_    = Keyword("struct")
        switch_    = Keyword("switch")
        true_      = Keyword("TRUE")
        typedef_   = Keyword("typedef")
        unsigned_  = Keyword("unsigned")
        union_     = Keyword("union")
        void_      = Keyword("void")
        wchar_     = Keyword("wchar")
        wstring_   = Keyword("wstring")

        reg_integer = ( Combine( CaselessLiteral("0x") + Word( nums+"abcdefABCDEF" ) ) |
                    Word( nums+"+-", nums ) ).setName("int")

        base_specified_integer =  Word(nums) + hashtag + Word( nums+"abcdefABCDEF_" ) + hashtag
        integer = Group(base_specified_integer | reg_integer)

        identifier = Word( alphas, alphanums + "_" ).setName("identifier")

        fully_specified_name = identifier + ZeroOrMore(dot + identifier)
        fully_specified_name.setParseAction(self.fully_specified_name_handler)
        alpha_numeric = Group(identifier | integer)


        singleLineComment = dash + dash + restOfLine
        singleLineComment.setParseAction(self.single_line_comment_handler)
        package_items = singleLineComment

        #subtype Blackboard_Name_Type is Name_Type;
        subtype_def = subtype_ + identifier + is_ + fully_specified_name + semi
        subtype_def.setParseAction(self.subtype_handler)
        package_items |= subtype_def

        enumeration_list = lparen + identifier + OneOrMore(comma + Optional(singleLineComment) + identifier) + rparen
        enumeration_list.setParseAction(self.enumeration_list_handler)

        enumeration_type_def = type_ + identifier + is_ + enumeration_list + semi
        enumeration_type_def.setParseAction(self.enumeration_type_handler)
        package_items |= enumeration_type_def

        type_or_subtype = Group(type_ | subtype_)

        # subtype File_Name_Type is String (1 .. Max_File_Name_Length);
        range_def1 = range_ + alpha_numeric + dot + dot + alpha_numeric
        range_def2 = lparen + alpha_numeric + dot + dot + alpha_numeric + rparen
        range_def3 = range_ + langle + rangle
        range_def = Group(range_def1 | range_def2 | range_def3)

        record_item = identifier + colon + fully_specified_name + Optional(range_def) + semi + Optional(singleLineComment)
        record_item.setParseAction(self.record_item_handler)
        record_parameter = identifier + colon + fully_specified_name
        record_parameters = lparen + record_parameter + ZeroOrMore(comma + record_parameter) + rparen
        record_name = identifier + Optional(record_parameters)
        record_name.setParseAction(self.record_name_handler)
        record_def = type_ + record_name + is_ + record_ + ZeroOrMore(record_item) + end_ + record_ + semi
        record_def.setParseAction(self.record_handler)

        #array (Error_Message_Size_Type) of APEX_Byte
        array_items = Group((lparen + fully_specified_name + Optional(range_def) + rparen) | range_def2)
        array_def = array_ + array_items + of_ + fully_specified_name
        array_def.setParseAction(self.array_def_handler)


        type_items = Group(array_def | fully_specified_name + Optional(range_def) | enumeration_list)
        type_def = type_or_subtype + identifier + is_ + Optional(new_) + type_items + semi
        type_def.setParseAction(self.type_def_handler)
        package_items |= type_def

        # type Blackboard_Id_Type is private;
        type_private_def = type_ + identifier + is_ + private_ + semi
        type_private_def.setParseAction(self.type_private_handler)
        package_items |= type_private_def

        direction = Group((in_ + out_) | in_ | out_ )

        parameter_item = identifier + colon + direction + fully_specified_name + Optional(semi)
        parameter_item.setParseAction(self.parameter_item_handler)
        procedure_items = Group(parameter_item | singleLineComment)
        procedure_def = procedure_ + identifier + Optional(lparen + ZeroOrMore(procedure_items) + rparen) + semi
        procedure_def.setParseAction(self.procedure_handler)
        package_items |= procedure_def

        # for Logbook_Status_Type'Alignment use 8;
        property_def = Group(alignment_ | size_) + use_ + integer
        for_property_def = for_ + fully_specified_name + single_quote + property_def + semi
        for_property_def.setParseAction(self.for_property_def_handler)
        package_items |= for_property_def

        hexadecimal = Word(hexnums) + hashtag
        for_item = identifier + equals + rangle + hexadecimal + hexadecimal
        for_items = for_item + ZeroOrMore(comma + for_item)
        for_def = for_ + fully_specified_name + use_ + lparen + for_items + rparen + semi
        for_def.setParseAction(self.for_def_handler)
        package_items |= for_def

        private_def = private_
        private_def.setParseAction(self.private_handler)
        package_items |= private_def

        simple_pragma = Group("Pure")

        pragma_def1 = pragma_ + simple_pragma + semi
        pragma_def1.setParseAction(self.pragma1_handler)

        pragma_item = identifier + comma + identifier + comma + quote + identifier + quote
        pragma_def3 = pragma_ + identifier + lparen + pragma_item + rparen + semi

        pragma_def2 = pragma_ + convention_ + lparen + identifier + comma + identifier + rparen + semi
        pragma_def2.setParseAction(self.pragma2_handler)
        pragma_def = Group(pragma_def1 | pragma_def2 | pragma_def3)
        package_items |= pragma_def

        package_items |= record_def

        special_case_constant = integer + star + star + integer
        constant_value = Group(fully_specified_name | special_case_constant | integer)
        # Max_Number_Of_Blackboards : constant := System_Limit_Number_Of_Blackboards;
        constant_def1 = identifier + colon + constant_ + colon + equals + constant_value + semi
        constant_def1.setParseAction(self.constant1_handler)
        package_items |= constant_def1


        #   Null_Blackboard_Id : constant Blackboard_Id_Type := 0;
        #   System_Limit_File_Size : constant := 2 ** 30; -- file system scope
        constant_def3 = identifier + colon + constant_ + fully_specified_name + colon + equals + constant_value + semi
        constant_def3.setParseAction(self.constant3_handler)
        package_items |= constant_def3 + Optional(singleLineComment)


        # Null_Blackboard_Id : constant Blackboard_Id_Type;
        constant_def2 = identifier + colon + constant_ + identifier + semi
        constant_def2.setParseAction(self.constant2_handler)
        package_items |= constant_def2

        package_item = Group(package_items)

        with_def = with_ + fully_specified_name + semi
        with_def.setParseAction(self.with_handler)
        use_def = use_ + fully_specified_name + semi;

        heading_items = Group(with_def | use_def | singleLineComment)

        rename_package = package_ + fully_specified_name + renames_ + fully_specified_name + semi
        self.package_def = package_ + fully_specified_name + is_ + ZeroOrMore(package_item) + end_ +\
                           fully_specified_name + semi
        self.package_def.setParseAction(self.package_handler)
        self.bnf = ZeroOrMore(heading_items) + OneOrMore(self.package_def | rename_package)


    def parse_string(self, ada_string):
        try:
            parse_result = self.bnf.parseString(ada_string)

            if self.tokens is not None:
                self.tokens += parse_result
            else:
                self.tokens = parse_result

            if self.unhandled_definitions > 0:
                logger.debug('Unhandled definitions: ' + str(self.unhandled_definitions))
            return True

        except ParseException as x:
            logger.error(str(x))
            return False

    def fully_specified_name_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.fully_specified_name_list.append(''.join(tokens))
        logger.debug("fully specified name: " + self.fully_specified_name_list[-1])

    def pop_fully_specified_name(self):
        if len(self.fully_specified_name_list) > 0:
            return self.fully_specified_name_list.pop()
        else:
            return "NO FULLY SPECIFIED NAME IN CUE"

    def package_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.ada_tree.add_parent(PackageNode(loc,tokens, fully_specified_name=self.pop_fully_specified_name()))

    def procedure_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.ada_tree.add_parent(ProcedureNode(loc, tokens, fully_specified_name=self.pop_fully_specified_name()))

    def parameter_item_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.ada_tree.add_sibling(ParameterItemNode(loc, tokens,
                                                    fully_specified_name=self.pop_fully_specified_name()))

    def record_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.ada_tree.add_parent(RecordNode(loc, tokens, discriminate=self.record_discriminate))
        self.record_discriminate = dict()

    def record_name_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        if len(tokens) == 6:
            self.record_discriminate[tokens[2]] = tokens[4]
        elif len(tokens) > 1:
            raise RuntimeError('Unexpected Tokens: ' + str(tokens))



    def record_item_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.ada_tree.add_sibling(RecordItemNode(loc, tokens,
                                                 fully_specified_name=self.pop_fully_specified_name()))

    def with_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.ada_tree.add_sibling(WithNode(loc, tokens, fully_specified_name=self.pop_fully_specified_name()))
        self.unhandled_definitions += 1

    def subtype_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def for_def_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def for_property_def_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def array_def_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def pragma1_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def pragma2_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def private_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def package_name_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def single_line_comment_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        # self.unhandled_definitions += 1  NOT GOING TO WORRY ABOUT NOT HANDLING COMMENTS FOR NOW

    def constant1_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def constant2_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def constant3_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def enumeration_type_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def enumeration_list_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def type_def_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def type_private_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.unhandled_definitions += 1


class Arinc_653_Ada_Conformance:


        arinc_653_part1_packages = ['apex', 'apex-timing', 'apex-processes', 'apex-partitions',
                                    'apex-sampling_ports', 'apex-queuing_ports', 'apex-buffers',
                                    'apex-blackboards', 'apex-semaphores', 'apex-events', 'apex-health_monitoring',
                                    'apex-adaptor']

        arinc_653_part2_packages = ['apex-file_system', 'apex-sampling_ports-data',
                                    'apex-multiple_module_schedules', 'apex-logbook_system',
                                    'apex-sampling_ports-extensions', 'apex-sap_ports',
                                    'apex-name_service', 'apex-memory_block',
                                    'apex-health_monitoring_extensions', 'apex-queuing_port_list']

        arinc_653_security_profile = arinc_653_part1_packages + ['apex-memory_block']
        arinc_653_safety_profile = arinc_653_part1_packages + \
                                   ['apex-file_system', 'apex-sampling_ports-extensions', 'apex-sampling_ports-data', 'apex-memory_block']
        arinc_653_safety_ext_profile = list(arinc_653_safety_profile) # Right now, they are the same

        # General Profile is Part 1 + Part2 except log books and multiple module schedules
        arinc_653_general_profile = arinc_653_part1_packages + arinc_653_part2_packages
        arinc_653_general_profile.remove('apex-logbook_system')
        arinc_653_general_profile.remove('apex-multiple_module_schedules')


        arinc_653_general_posix_profile = ['apex-sampling_ports', 'apex-queuing_ports', 'apex-health_monitoring',
                                           'apex-sampling_ports-data','apex-sampling_ports-extensions',
                                           'apex-health_monitoring_extensions', 'apex-queuing_port_list']

        directory_profile_packages = {'general': arinc_653_general_profile,
                                      'safety': arinc_653_safety_profile,
                                      'safety_ext': arinc_653_safety_ext_profile,
                                      'security': arinc_653_security_profile,
                                      'posix_allowed': arinc_653_general_posix_profile}

        arinc_653_part2_packages = ['apex-file_system', 'apex-sampling_ports-data',
                                    'apex-multiple_module_schedules', 'apex-logbook_system',
                                    'apex-sampling_ports-extensions', 'apex-sap_ports',
                                    'apex-name_service', 'apex-memory_block',
                                    'apex-health_monitoring_extensions', 'apex-queuing_port_list']



        def get_full_implementation_spec_files(self, full_implementation_directory):

            spec_files = glob.glob(os.path.join(full_implementation_directory, '*.ads'))

            all_arinc_653_spec_files = list()
            for a in self.arinc_653_part1_packages + self.arinc_653_part2_packages:
                all_arinc_653_spec_files.append(os.path.join(full_implementation_directory, a + '.ads'))

            set_difference = set(all_arinc_653_spec_files) - set(spec_files)
            if set_difference != set():
                err_msg =  "full implementation directory doesn't have ALL part1 and part2 ARINC 653 files" + \
                    " missing files: " + ', '.join(set_difference)
                logger.error(err_msg)
                raise RuntimeError(err_msg)

            set_difference = set(spec_files) - set(all_arinc_653_spec_files)
            if set_difference != set():
                err_msg =  "full implementation directory doesn't have ONLY part1 and part2 ARINC 653 files" + \
                    " additional files: " + ', '.join(set_difference)
                logger.error(err_msg)
                raise RuntimeError(err_msg)

            return all_arinc_653_spec_files

        def generate_conformance_test(self, root_dir=ConformanceMain.getRootDir()):

            full_implementation_gold_lib_directory = os.path.join(root_dir, constants.ADA_ARINC_653_GOLD_LIB_DIR, 'full_implementation')
            all_arinc_653_spec_files = self.get_full_implementation_spec_files(full_implementation_gold_lib_directory)

            conformance_test_base_directory = os.path.join(root_dir, constants.ADA_ARINC_653_CONFORMANCE_DIR)
            full_implementation_conformance_test_directory = os.path.join(conformance_test_base_directory,
                                                                          'full_implementation')
            parsed_files = list()
            parse_error_files = list()
            ada_parser = AdaParser()
            unhandled_definitions = dict()

            for spec_file in all_arinc_653_spec_files:
                base_name = os.path.basename(spec_file)
                logger.debug("SPEC FILE: " + base_name + 20 * '-')
                try:
                    if ada_parser.generate_conformance_tests(
                            spec_file, os.path.join(conformance_test_base_directory, 'full_implementation')):
                        parsed_files.append(base_name)
                        unhandled_definitions[base_name] = ada_parser.unhandled_definitions

                    else:
                        parse_error_files.append(base_name)

                except Exception as parser_exception:
                    logger.error('Parser died on spec file: ' + base_name + '\n' + str(parser_exception))
                    raise parser_exception

            logger.info('Successful parsed files: ' + ', '.join(parsed_files))
            if len(parse_error_files) > 0:
                err_msg = 'Error in parsing files: ' + ', '.join(parse_error_files)
                logger.error(err_msg)
                raise RuntimeError(err_msg)

            for spec_file, unhandled_definition in unhandled_definitions.iteritems():
                if unhandled_definition > 0:
                    logger.warning('spec file: ' + spec_file + ' unhandled definitions: ' \
                        + str(unhandled_definition) + '\n')

            for directory, allowed_packages in self.directory_profile_packages.iteritems():
                target_directory = os.path.join(conformance_test_base_directory, directory)
                if not os.path.isdir(target_directory):
                    raise IOError('target directory does not exist: ' + target_directory)

                for a in allowed_packages:
                    if a == 'apex':
                        continue
                    subdirectory = a.replace('apex-', '').upper().replace('-','_')
                    subdirectory = os.path.join(target_directory, subdirectory)
                    commonTools.mkdir_p(subdirectory)
                    old_ada_files = glob.glob(os.path.join(subdirectory, '*.ad*'))
                    for o in old_ada_files:
                        os.remove(o)

                    logger.debug('PACKAGE: ' + a + ' SUBDIRECTORY: ' + subdirectory)
                    commonTools.copy_files(os.path.join(full_implementation_conformance_test_directory,
                                                        'test_' + a.replace('-', '_') + '_*.ad*'),
                                           subdirectory)
                    commonTools.copyFile(os.path.join(full_implementation_conformance_test_directory,
                                                      a + '_assertions.txt'),
                                         os.path.join(subdirectory, 'assertions.txt'))





if __name__ == '__main__':
    # when using this in the make_devel script, the generate_conformance_test_files and generate_gold_lib_files
    # need to be set to True and all others set to False.  (The others are used to test individual aspects of the
    # parser functionality without generating all the files.
    logger = logging.getLogger(__name__)
    logger.info('main running from ada_parser.py')

    generate_conformance_test_files = False
    generate_conformance_test_files = True
    if generate_conformance_test_files:
        conformance_gen = Arinc_653_Ada_Conformance()
        conformance_gen.generate_conformance_test()

    generate_gold_lib_files = False
    generate_gold_lib_files = True
    if generate_gold_lib_files:
        conformance_gen = Arinc_653_Ada_Conformance()
        conformance_gen.generate_gold_lib_files()




    generate_dummy_conformance_test = False
    # generate_dummy_conformance_test = True
    if generate_dummy_conformance_test:
        ada_parser = AdaParser()

        spec_file = os.path.join(ConformanceMain.getRootDir(), 'dummy.ads')
        target_directory = os.path.join(ConformanceMain.getRootDir(),
                                        constants.ADA_ARINC_653_CONFORMANCE_DIR, 'full_implementation')
        if ada_parser.generate_conformance_tests(spec_file, target_directory):
            logger.info("dummy.ads -- SUCCESS")
        else:
            logger.error("dummy.ads -- FAIL")



    #spec_file = Dialogs.osAgnosticOpenFilenameDialog(initialdir=os.path.curdir)

    spec_directory = os.path.join(ConformanceMain.getRootDir(), constants.GOLD_LIB_ROOT,
                                  'Ada','src','OSS', 'ARINC-653', 'full_implementation')

    spec_file_only = False
    #spec_file_only = True
    if spec_file_only == True:
        ada_parser = AdaParser()

        spec_file = os.path.join(spec_directory, 'apex-sampling_ports-data.ads')
        logger.debug("SPEC FILE: " + os.path.basename(spec_file) + 20 * '-')
        if ada_parser.parse_ada_spec_file(spec_file):
            logger.deubg('\nspec file: ' + os.path.basename(spec_file) + ' unhandled definitions: ' \
                + str(ada_parser.unhandled_definitions) + '\n')
            ada_parser.ada_tree.print_tree()

            logger.info("SUCCESS")
        else:
            logger.error("FAIL")

