#!/usr/bin/env python
"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""

__author__ = 'david.fann@vanderbilt.edu'

import os.path
import ConformanceMain
import sys
egg_path = os.path.join(ConformanceMain.getRootDir(),'face_conformance_app', 'pyparsing-2.0.1-py2.7.egg')
sys.path.append(egg_path)
egg_path = os.path.join(ConformanceMain.getRootDir(),'face_conformance_app', 'stringtemplate3-3.1-py2.7.egg')
sys.path.append(egg_path)
egg_path = os.path.join(ConformanceMain.getRootDir(),'face_conformance_app', 'setuptools-0.6c11-py2.7!\\')
sys.path.append(egg_path)


from pyparsing import Literal, CaselessLiteral, Word, Upcase, OneOrMore, ZeroOrMore, \
        Forward, NotAny, delimitedList, oneOf, Group, Optional, Combine, alphas, nums, restOfLine, cStyleComment, \
        alphanums, printables, empty, quotedString, ParseException, ParseResults, Keyword, hexnums
import stringtemplate3

import logging
import os
import copy
import glob

import TestConfiguration
import commonTools
import constants
import ConformanceMain


if __name__ == '__main__':
    ConformanceMain.initializeLogger()

logger = logging.getLogger(__name__)

import idlParser

class CTree(idlParser.IdlTree):
    def __init__(self):
        idlParser.IdlTree.__init__(self)
        pass

class IfdefBlockNode(idlParser.Node):
    def __init__(self, loc, tokens, **kw):
        idlParser.Node.__init__(self, loc=loc, **kw)
        if tokens[0][0] != '#' and tokens[[0][1]] != 'ifndef':
            raise RuntimeError("Unexpected tokens: " + str(tokens))
        self.name = tokens[0][2]

class OperationNode(idlParser.Node):
    def __init__(self, loc, tokens, **kw):
        idlParser.Node.__init__(self, loc=loc, **kw)
        if len(tokens) < 2:
            raise RuntimeError("Unexpected tokens, expected [['OPERATION_RETURN_TYPE', 'OPERATION_NAME',...]]")

        self.return_type = tokens[0]
        name_loc = 1
        if tokens[0] == 'extern':
            self.return_type = tokens[0] + ' ' + tokens[1]
            name_loc += 1
        self.name = tokens[name_loc]

        self.conformance_test_node = True  # Need to generate a conformance test for this type of node

class ParameterItemNode(idlParser.ParameterNode):
    # Equivalent to the IDL Parameter Node, however, must have unique __init__ since the tokens
    # will be different in an C include file than an IDL file
    def __init__(self, loc, tokens, **kw):
        idlParser.Node.__init__(self, loc=loc, **kw)

        start_token = 0
        self.direction = None
        if tokens[0] == '/':
            if tokens[4] != '/':
                raise RuntimeError("Unexpected tokens, expected ['/', '*', 'in and/or out', '*', '/',"
                    + "TYPENAME..., got: "  + str(tokens))
            self.direction = tokens[2]
            start_token = 5  # ignoring the '/', '*', 'in/out', '*', '/' list items
        self.type_name = tokens[start_token]
        self.name = tokens[start_token + 1]
        if tokens[start_token + 1] == '*':
            self.type_name += ' *'
            self.name = tokens[start_token + 2]

        logger.info('tokens: ' + str(tokens) + ' => ' + self.toString())


class CParser():
    def __init__(self, **kw):
        self.bnf = None
        self.reset_variables()
        self.initialize_parser()

    def reset_variables(self):
        self.unhandled_definitions = 0
        self.c_tree = CTree()
        self.tokens = None
        self.fully_specified_name_list = list()
        self.record_discriminate = dict()

    def initialize_parser(self):
        # punctuation
      # punctuation
        colon  = Literal(":")
        lbrace = Literal("{")
        rbrace = Literal("}")
        lbrack = Literal("[")
        rbrack = Literal("]")
        lparen = Literal("(")
        rparen = Literal(")")
        equals = Literal("=")
        comma  = Literal(",")
        dot    = Literal(".")
        hashtag= Literal("#")
        slash  = Literal("/")
        bslash = Literal("\\")
        star   = Literal("*")
        semi   = Literal(";")
        langle = Literal("<")
        rangle = Literal(">")
        quote  = Literal('"')

        # keywords
        any_       = Keyword("any")
        attribute_ = Keyword("attribute")
        boolean_   = Keyword("boolean")
        case_      = Keyword("case")
        char_      = Keyword("char")
        const_     = Keyword("const")
        context_   = Keyword("context")
        default_   = Keyword("default")
        double_    = Keyword("double")
        enum_      = Keyword("enum")
        exception_ = Keyword("exception")
        false_     = Keyword("FALSE")
        float_     = Keyword("float")
        inout_     = Keyword("inout")
        extern_    = Keyword("extern")
        interface_ = Keyword("interface")
        in_        = Keyword("in")
        int_       = Keyword("int")
        include_   = Keyword("include")
        long_      = Keyword("long")
        long_long_ = Keyword("long long")
        module_    = Keyword("module")
        object_    = Keyword("Object")
        octet_     = Keyword("octet")
        oneway_    = Keyword("oneway")
        out_       = Keyword("out")
        raises_    = Keyword("raises")
        readonly_  = Keyword("readonly")
        sequence_  = Keyword("sequence")
        short_     = Keyword("short")
        struct_    = Keyword("struct")
        switch_    = Keyword("switch")
        true_      = Keyword("TRUE")
        typedef_   = Keyword("typedef")
        unsigned_  = Keyword("unsigned")
        union_     = Keyword("union")
        void_      = Keyword("void")
        ifdef_     = Keyword("ifdef")
        cpluscplus_ = Keyword("__cplusplus")
        extern_C_   = Keyword('extern "C"')
        endif_      = Keyword("endif")

        identifier = Word( alphas + "_", alphanums + "_" ).setName("identifier")

        real = Combine( Word(nums+"+-", nums) + dot + Optional( Word(nums) )
                        + Optional( CaselessLiteral("E") + Word(nums+"+-",nums) ) )
        integer = ( Combine( CaselessLiteral("0x") + Word( nums+"abcdefABCDEF" ) ) |
                    Word( nums+"+-", nums ) ).setName("int")

        # Typenames
        udTypeName = delimitedList( identifier, "::", combine=True ).setName("udType")
        # have to use longest match for type, in case a user-defined type name starts with a keyword type, like "stringSeq" or "longArray"

        typeName = ( Optional(unsigned_) + char_ ^ double_ ^
                    float_ ^ Optional(unsigned_) + long_long_ ^ Optional(unsigned_) + long_ ^ octet_ ^
                     Optional(unsigned_) + short_ + int_ ^ Optional(unsigned_) + short_ ^ Optional(unsigned_) + int_ ^
                     udTypeName ).setName("type")

        # Comments (either single line or cStyle comments
        single_line_comment = "//" + restOfLine
        single_line_comment.setParseAction(self.single_line_comment_handler)
        cStyleComment.setParseAction(self.cStyle_comment_handler)


        any_comment = (single_line_comment | cStyleComment)
        optional_comment = Optional(any_comment)
        pathname = identifier + slash
        include_def = hashtag + include_ + (quote | langle) + ZeroOrMore(pathname) + identifier + dot + Literal('h') + (quote | rangle)
        include_def.setParseAction(self.include_def_handler)
        code_blocks = include_def + optional_comment

        enumeration_item = identifier + Optional(equals + integer)
        enumeration_item.setParseAction(self.enumeration_item_handler)
        enumeration_def = typedef_ + enum_ + lbrace + \
            enumeration_item + ZeroOrMore(comma + optional_comment + enumeration_item) + optional_comment + rbrace + \
            identifier + semi
        enumeration_def.setParseAction(self.enumeration_def_handler)
        code_blocks |= enumeration_def + optional_comment

        typedef_def = typedef_ + typeName + Optional(star) + identifier + semi
        typedef_def.setParseAction(self.typedef_def_handler)
        code_blocks |= typedef_def + optional_comment

        typedef_array_def = typedef_ + typeName + identifier + lbrack + (identifier | integer) + rbrack + semi
        typedef_array_def.setParseAction(self.typedef_array_def_handler)
        code_blocks |= typedef_array_def

        define_def = Group(hashtag + Keyword('define') + identifier + Optional(integer | (identifier + Optional(star))))
        define_def.setParseAction(self.define_def_handler)
        code_blocks |= define_def + optional_comment

        # Structures
        array_def = lbrack + (identifier | integer) + rbrack
        bit_field_def = colon + integer
        struct_item = typeName + identifier + Optional(array_def | bit_field_def) + semi + optional_comment
        struct_item.setParseAction(self.struct_item_handler)
        struct_def = Group( typedef_ + struct_ + Optional(identifier) + lbrace + optional_comment +
                           OneOrMore(struct_item) + rbrace + identifier + semi)
        struct_def.setParseAction(self.struct_handler)
        code_blocks |= struct_def

        # typedef struct __GLsync *GLsync;
        typedef_struct_ptr = typedef_ + struct_ + identifier + star + identifier + semi
        code_blocks |= typedef_struct_ptr

        # Operations
        in_out_comment = slash + star + (in_ | out_ | inout_) + star + slash
        const_star = const_ + star

        parameter = Optional(in_out_comment) + Optional(const_) + typeName + ZeroOrMore(star | const_star) + identifier
        parameter.setParseAction(self.parameter_handler)
        paramlist = parameter + ZeroOrMore(comma + parameter) + optional_comment

        operation_def = Optional(extern_) + void_ + identifier + lparen + Optional( void_ | paramlist ) + rparen + semi
        operation_def.setParseAction(self.operation_handler)
        code_blocks |= operation_def

        typedef_function_ptr_def = typedef_ + typeName + lparen + star + identifier + rparen + \
                                   Optional(lparen + paramlist + rparen) + semi
        typedef_function_ptr_def.setParseAction(self.function_ptr_handler)
        code_blocks |= typedef_function_ptr_def



        # GL_APICALL void GL_APIENTRY glCullFace (GLenum mode);
        gl_api_call = "GL_APICALL" + Optional(const_) + identifier + Optional(star)+ "GL_APIENTRY" + identifier + lparen + Optional( void_ | paramlist) + rparen + semi
        code_blocks |= gl_api_call


        # typedef void (GL_APIENTRYP PFNGLLINEWIDTHPROC) (GLfloat width);

        gl_api_entry_p = typedef_ + Optional(const_) + identifier + Optional(star) + lparen + "GL_APIENTRYP" + identifier + rparen + lparen + Optional( void_ | paramlist) + rparen + semi
        code_blocks |= gl_api_entry_p



        ifndef_def = Group(hashtag + Keyword('ifndef') + identifier)
        ifndef_def.setParseAction(self.ifndef_def_handler)
        endif_def = Group(hashtag + Keyword('endif'))

        cplusplus_comment_ = slash + star + cpluscplus_ + star + slash

        ifdef_cplusplus = hashtag + ifdef_ + cpluscplus_
        ifdef_cplusplus_intro_block = ifdef_cplusplus + extern_C_ + lbrace + hashtag + endif_ + ZeroOrMore(cplusplus_comment_)
        ifdef_cplusplus_exit_block = ifdef_cplusplus + rbrace + hashtag + endif_

        code_blocks |= ifdef_cplusplus_intro_block
        code_blocks |= ifdef_cplusplus_exit_block

        ifdef_block = Forward()
        code_blocks |= ifdef_block
        ifdef_def_start = hashtag + ifdef_ + identifier
        ifdef_block << ifdef_def_start + ZeroOrMore(code_blocks | any_comment) + endif_def

        if_block = Forward()
        code_blocks |= if_block
        if_def_start = hashtag + Keyword("if") + identifier
        if_block << if_def_start + ZeroOrMore(code_blocks | any_comment) + endif_def


        ifndef_block = Forward()
        code_blocks |= ifndef_block

        ifndef_block << ifndef_def + ZeroOrMore(code_blocks | any_comment) + endif_def
        ifndef_block.setParseAction(self.ifndef_block_handler)
        self.bnf = OneOrMore(ZeroOrMore(any_comment) + ifndef_block | ifdef_block)

    def parse_string(self, c_string):
        try:
            parse_result = self.bnf.parseString(c_string)

            if self.tokens is not None:
                self.tokens += parse_result
            else:
                self.tokens = parse_result

            if self.unhandled_definitions > 0:
                logger.debug('Unhandled definitions: ' + str(self.unhandled_definitions))

            return True

        except ParseException as x:
            logger.error(str(x))
            raise x

    def cStyle_comment_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.c_tree.add_sibling(CStyleCommentNode(loc, tokens))

    def single_line_comment_handler(self, s, loc, tokens):
        logger.info('tokens: ' + str(tokens))
        self.c_tree.add_sibling(SingleLineCommentNode(loc, tokens))

    def operation_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.c_tree.add_parent(OperationNode(loc, tokens))

    def parameter_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.c_tree.add_sibling(ParameterItemNode(loc, tokens))

    def include_def_handler(self, s, loc, tokens):
        logger.info('tokens: ' + str(tokens))
        self.unhandled_definitions += 1



    def ifndef_block_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.c_tree.add_parent(IfdefBlockNode(loc, tokens))


    def ifndef_def_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def define_def_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def typedef_def_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def enumeration_def_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def enumeration_item_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def struct_item_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def struct_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.unhandled_definitions += 1


    def typedef_array_def_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def function_ptr_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.unhandled_definitions += 1



class CommentNode(idlParser.Node):
    def __init__(self, loc, **kw):
        idlParser.Node.__init__(self, loc=loc, **kw)
        self.name = "Comment"

    def get_conformance_test_code(self, template_group, language):
        return None # Comments do not need conformance tests


class CStyleCommentNode(CommentNode):
    def __init__(self, loc, tokens, **kw):
        CommentNode.__init__(self, loc=loc, **kw)

        if len(tokens) != 1 or not tokens[0].startswith('/*') or not tokens[0].endswith('*/'):
            raise RuntimeError("Unexpected tokens, expected ['/* COMMENT */'], got " + str(tokens))

        self.comment = tokens[0].strip('/*')


class SingleLineCommentNode(CommentNode):
    def __init__(self, loc, tokens, **kw):
        CommentNode.__init__(self, loc=loc, **kw)

        if len(tokens) != 2 or not tokens[0].startswith('//'):
            raise RuntimeError("Unexpected tokens, expected ['//', 'COMMENT...'], got " + str(tokens))

        self.comment = tokens[1].strip()



if __name__ == '__main__':
    logger = logging.getLogger(__name__)
    logger.info('main running from ' + __file__)

    parse_ARINC_653_P1 = True
    if parse_ARINC_653_P1:
        c_parser = CParser()

        # Will replace dummy.h with ARINC653.h once everything is working
        arinc_include_filename = os.path.join(ConformanceMain.getRootDir(), constants.C_ARINC_653_GOLD_LIB_DIR, 'dummy.h')
        logger.debug(arinc_include_filename)
        arinc_include_file = open(arinc_include_filename, 'r')
        c_code = arinc_include_file.readlines()
        c_code = '\n'.join(c_code)
        c_parser.parse_string(c_code)
        logger.debug('finished parsing')
        for node in iter(c_parser.c_tree):
            if isinstance(node, CommentNode):
                continue

            logger.debug(node.__class__.__name__ + ': ' + node.name)
