#!/usr/bin/env python
"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""

__author__ = 'david.fann@vanderbilt.edu'

# Search for 32-bit or 64-bit Java 7 as needed by Robert's DM gadget.
# This does NOT have any bearing on Java 6 as used in the standard and its testing.

import sys
import platform
import os.path

_NOTFOUND = (False, None)
_REG32J17QUERY = 'reg query "HKLM\\SOFTWARE\\JavaSoft\\Java Runtime Environment\\1.7" /v JavaHome /reg:32'
_REG64J17QUERY = 'reg query "HKLM\\SOFTWARE\\JavaSoft\\Java Runtime Environment\\1.7" /v JavaHome /reg:64'
_REG32J18QUERY = 'reg query "HKLM\\SOFTWARE\\JavaSoft\\Java Runtime Environment\\1.8" /v JavaHome /reg:32'
_REG64J18QUERY = 'reg query "HKLM\\SOFTWARE\\JavaSoft\\Java Runtime Environment\\1.8" /v JavaHome /reg:64'
_REG32J19QUERY = 'reg query "HKLM\\SOFTWARE\\JavaSoft\\Java Runtime Environment\\1.9" /v JavaHome /reg:32'
_REG64J19QUERY = 'reg query "HKLM\\SOFTWARE\\JavaSoft\\Java Runtime Environment\\1.9" /v JavaHome /reg:64'
_REGQUERIES = (_REG32J17QUERY, _REG32J17QUERY, _REG32J18QUERY, _REG32J18QUERY, _REG32J19QUERY, _REG32J19QUERY)
_JAVAEXE = 'java.exe' if platform.system() == 'Windows' else 'java'
_JAVAHOMEPATHCOMPLETION = os.path.join('bin', _JAVAEXE)

def _findJavaInPath():
    import commonTools
    result = commonTools.subprocessWithSpecifiedPath('java -version', list())

    print result.stdout

    if result.returncode != 0:
        print 'ERROR: Java is not in path or not installed'
        return _NOTFOUND
    else:
        versionLoc = result.stdout.find('version')
        if versionLoc != -1:
            versionStart = result.stdout[versionLoc:].find('"') + 1 + versionLoc
            versionEnd = result.stdout[versionStart:].find('"') + versionStart
            version =  result.stdout[versionStart:versionEnd]
            splitVersion = version.split('.')
            if int(splitVersion[1]) < 7:
                print 'Java version must be >= 1.7'
                return _NOTFOUND
        else:
            print 'Warning, could not parse java version number, please check that it is >= 1.7'

        return (True, "java")

def _findJavaInRegistry():
    """Windows only"""
    if platform.system() != 'Windows':
        return _NOTFOUND

    import commonTools

    for query in _REGQUERIES:
        result = commonTools.subprocessWithSpecifiedPath(query, list())
        if result.returncode != 0:
            continue

        regszloc = result.stdout.find('REG_SZ')
        if regszloc == -1:
            continue

        sizeofregsz = len('REG_SZ')

        javahome = result.stdout[regszloc + sizeofregsz:].strip()
        javaexepath = os.path.join(javahome, _JAVAHOMEPATHCOMPLETION)
        if os.path.isfile(javaexepath):
            return (True, javaexepath)

    return _NOTFOUND

def findJava():
    """
    Find Java's installation location. Returns a tuple whose first value
    is whether Java was found and whose second value is the path to the
    Java executable. (True, "path/to/java.exe") or (False, None)
    """
    searchFuncs = ( \
                    _findJavaInRegistry, \
                    _findJavaInPath, \
                  )

    for search in searchFuncs:
        result = search()
        if (result[0]):
            return result

    return _NOTFOUND

def main():
    """
    Search for Java, exiting 0 if found an non-zero otherwise.
    """
    result = findJava()
    if result[0]:
        print "Java found at " + result[1]
        sys.exit(0)
    else:
        print 'ERROR: Java 1.7 could not be found'
        sys.exit(1)

if __name__ == "__main__":
    main()

