#!/usr/bin/env python
"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""
'''
commonTools - classes/methods useful to multiple modules associated with the FACE Conformance Test Suite

This module contains two classes and XX Functions
classes:
    EntryWLabel - tkinter based widgets that contain an entry box and associated label
    EditFile - tkinter based window that allows for simple editing of files (used to edit a file within the testsuite)
functions:
    displayFile - tkinter based window that displays the content of a file
    modalDialog - causes the dialog window to be displayed as a modal dialog
    listToString - converts a list to a delimited string
    fileDialog - tkinter based file dialog window
    directoryDialog - tkinter based directory dialog window
    initializeLogger - initializes the standard python logging mechanism
   
'''

__author__ = 'david.fann@vanderbilt.edu'

import logging

import ConformanceMain

if __name__ == '__main__':
    ConformanceMain.initializeLogger()

from os import getcwd, makedirs, environ, access, X_OK, pathsep, remove, walk, sep
from os.path import isfile, isdir, split, join, basename, exists, dirname, relpath
import os
import fnmatch
import glob

from collections import namedtuple
import platform
import re
import ctypes
import Tkinter
import errno
import logging
from copy import deepcopy
from shutil import move, copy, rmtree
from glob import iglob
import codecs
import csv
import subprocess
import sys
import constants
import shlex   # added 6-5-19 to support subprocess.Popen with shell=False

if platform.system() == 'Windows':
    from _winreg import *

interactive = True
logger = logging.getLogger(__name__)

def setInteractive(interactiveBoolean):
    global interactive
    logger.debug('Setting Interactive to ' + trueFalseString(interactiveBoolean))
    interactive = interactiveBoolean
    
def isInteractive(): 
    global interactive
    return interactive

_parentGUIList = list()
def pushParentGUI(newParentGUI):
    global _parentGUIList
    _parentGUIList.append(newParentGUI)

def popParentGUI():
    global _parentGUIList
    _parentGUIList.pop()

def getParentGUI():
    global _parentGUIList
    if len(_parentGUIList) == 0:
        return None
    else:
        return _parentGUIList[-1]

def unicode_csv_reader(unicode_csv_data, dialect=csv.excel, **kwargs):
    # csv.py doesn't do Unicode; encode temporarily as UTF-8:
    csv_reader = csv.reader(utf_8_encoder(unicode_csv_data),
                            dialect=dialect, **kwargs)
    for row in csv_reader:
        # decode UTF-8 back to Unicode, cell by cell:
        yield [unicode(cell, 'utf-8') for cell in row]


def utf_8_encoder(unicode_csv_data):
    for line in unicode_csv_data:
        yield line.encode('utf-8')


class EntryWLabel():
    """ Class used to manage tkinter entry widget with an associated label widget """

    def __init__(self, master, **kw):
        """
        master - tkinter master window associated with the widget
        
        keywords:
        labelText - what text to use as a label
        stringVariable - if a string is desired to be used to get the entry value instead of the entryVariable (entryVariable is preferred)
        widgetList - if the label and entry widgets are to be appended to a list
        entryVariable - a StringVar that should be associated with the Entry widget
        command - the method that should be called when the entry is updated
        noEntryLabelGrid - if this is set to false, the widgets will be set to row 0, col 0,1 using the grid command, if True, no grid command will be called
        noFrameGrid - if True, no grid command will be called on the Frame used for the label/entry widget, if False, a simple grid command will be called on the Frame
        validation - when the command should be called.  Default is 'focus out'
        frame - Frame to add label/entry widgets to...default is to create a new Frame instance
        
        """

        self.master = master
        self.labelText = kw.get('label', None)
        self.stringVariable = kw.get('stringVariable', None)
        self.widgets = kw.get('widgetList', None)
        self.entryVariable = kw.get('entryVariable', Tkinter.StringVar())
        self.command = kw.get('command', None)
        self.noEntryLabelGrid = kw.get('noEntryLabelGrid', False)
        self.noFrameGrid = kw.get('noFrameGrid', True)
        self.validate = kw.get('validation', 'focusout')
        self.frame = kw.get('frame', None)
        self.entryWidth = kw.get('entryWidth', 20)
        self.inheritedFrame = False
        if self.frame is None:
            self.frame = Tkinter.Frame(self.master)
            self.inheritedFrame = True

        self.labelWidget = Tkinter.Label(self.frame, text=self.labelText)
        if self.stringVariable is not None:
            self.entryVariable.set(self.stringVariable)
        options = {'textvariable': self.entryVariable, 'validate': self.validate, 'validatecommand': self._command,
                   'width': self.entryWidth}
        self.entryWidget = Tkinter.Entry(self.frame, options)
        self.deleteWidget = deleteButton(self.frame, self.entryVariable)
        if not self.noEntryLabelGrid:
            self.labelWidget.grid(row=0, column=0, sticky='e')
            self.entryWidget.grid(row=0, column=1, sticky='w')
            self.deleteWidget.grid(row=0, column=2, sticky='w')
        elif self.widgets is not None:
            self.widgets.append(self.labelWidget)
            self.widgets.append(self.entryWidget)
            self.widgets.append(self.deleteWidget)
        if not self.noFrameGrid:
            self.frame.grid()
        elif self.widgets is not None and not self.inheritedFrame:
            self.widgets.append(self.frame)


    def _command(self):
        if self.stringVariable is not None:
            self.stringVariable = self.entryVariable.get()

        if self.command is not None:
            self.command()

    def setEntry(self, value):
        """ set the value for the entry widget (not a new StringVar to associate with the widget) """
        self.entryVariable.set(value)

    def getEntry(self):
        """get the value from the entry widget, not the StringVar associated with the widget"""
        return self.entryVariable.get()

    def setLabel(self, newLabel):
        """ Set the label text """
        self.labelText = newLabel
        self.labelWidget.configure(text=self.labelText)

    def getLabel(self):
        """ get the label text """
        return self.labelText


def deleteButton(buttonFrame, textToDelete):
    return Tkinter.Button(buttonFrame, text=u'\u2718', command=lambda: deleteButtonCmd(textToDelete), foreground='red')


def deleteButtonCmd(textToDelete):
    logger.info('deleting text')
    try:
        textToDelete.set('')
    except:
        logger.error("Tried to set variable passed to ''")


class EditFile():
    """ A tkinter based window that can be used for simple edits to a file """

    def __init__(self, filename, **kw):
        """
            filename - the filename to edit
            
            keywords:
            maxWidth - the maximum width of the window
            masterWin - the master tkinter associated with the window (default None)
            prompt - a prompt to put at the top of the window (underneath the title)
        """
        logger.info('Editing ' + filename)
        self.filename = filename
        maxWidth = kw.get('maxWidth', -1)
        masterWin = kw.get('master', None)
        prompt = kw.get('prompt', None)

        if isfile(filename):
            if masterWin is None:
                self.cwin = Tkinter.Tk()
            else:
                self.cwin = Tkinter.Toplevel(masterWin)
            self.cwin.title(filename)
            windowRow = 0

            if prompt is not None:
                promptFrame = Tkinter.Frame(self.cwin)
                promptFrame.grid(row=windowRow)
                windowRow += 1
                Tkinter.Label(promptFrame, text=prompt).grid()

            textFrame = Tkinter.Frame(self.cwin)
            textFrame.grid(row=windowRow)
            windowRow += 1
            yscrollbar = Tkinter.Scrollbar(textFrame)
            self.text = Tkinter.Text(textFrame)
            self.text.config(yscrollcommand=yscrollbar.set)
            yscrollbar.config(command=self.text.yview)
            self.text.grid(row=0, column=0)
            yscrollbar.grid(row=0, column=1)

            logger.debug('Filename: ' + filename)
            dfile = safeOpen(filename, 'r')

            maxFileWidth = 0
            for d in dfile:
                if len(d) > maxFileWidth:
                    maxFileWidth = len(d)
                self.text.insert(Tkinter.END, d)
            if maxWidth > 0:
                if maxFileWidth > maxWidth:
                    maxFileWidth = maxWidth
            self.text.config(width=maxFileWidth)

            buttonFrame = Tkinter.Frame(self.cwin)
            buttonFrame.grid(row=windowRow)
            windowRow += 1

            okayButton = Tkinter.Button(buttonFrame, text='Save', command=self._save)
            okayButton.grid(row=0, column=0)
            cancelButton = Tkinter.Button(buttonFrame, text='Cancel', command=self._cancel)
            cancelButton.grid(row=0, column=1)

            self.cwin.mainloop()

        else:
            logger.error(filename + " is not a file!")

    def _save(self):
        logger.info('Saving file: ' + self.filename)
        contents = self.text.get(1.0, Tkinter.END)
        logger.debug('Contents: ' + str(contents))
        efile = safeOpen(self.filename, 'w')
        for c in contents:
            efile.write(c)
        self.cwin.destroy()

    def _cancel(self):
        logger.info('Canceling edit')
        self.cwin.destroy()


def displayFile(filename, **kw):
    """
        Using tkinter, displays a text file in a window
        
        filename - the filename to display
        
        keywords:
        master - the tkinter master window associated with this window
        maxWidth - the maximum width (in characters) of this window
    """
    logger.info('Displaying ' + filename)
    masterWin = kw.get('master', None)
    maxWidth = kw.get('maxWidth', 120)


    if isfile(filename):
        if masterWin is None:
            cwin = Tkinter.Tk()
        else:
            cwin = Tkinter.Toplevel(masterWin)
        makeResizeable(cwin)
        cwin.title(filename + ' (Read-only)')
        #xscrollbar = Tkinter.Scrollbar(cwin)
        yscrollbar = Tkinter.Scrollbar(cwin)
        text = Tkinter.Text(cwin)
        text.config(background='white')
        text.config(yscrollcommand=yscrollbar.set)
        #text.config(xscrollcommand=xscrollbar.set)
        yscrollbar.config(command=text.yview)
        #xscrollbar.config(command=text.xview)
        text.grid(row=0, column=0, sticky="nsew")
        yscrollbar.grid(row=0, column=1, sticky="ns")
        #xscrollbar.grid(row=1,column=0, sticky="ew")

        dfile = safeOpen(filename, 'r')
        lines = dfile.readlines()
        dfile.close()

        if maxWidth > 0:
            if type(lines) == type(list()):
                lines = '\n'.join(lines)
            lines = wraptext(lines, maxLength=maxWidth)
            maxFileWidth = maxWidth + 1

        else:
            maxFileWidth = 0
            for l in lines:
                if len(l) > maxFileWidth:
                    maxFileWidth = len(l)
            if type(lines) == type(list()):
                lines = '\n'.join(lines)

        text.insert(Tkinter.END, lines)

        text.config(width=maxFileWidth)
        text.config(state=Tkinter.DISABLED)
        cwin.mainloop()


    else:
        logger.error(filename + " is not a file!")


def modalDialog(dialogWin, masterWin):
    """
    Creates a modal dialog for the tkinter dialog window
    
    dialogWin - the dialog window desired to be modal
    masterWin - the associated master tkinter window of the dialog
    """
    dialogWin.focus_set()
    dialogWin.grab_set()
    if masterWin is not None:
        dialogWin.transient(masterWin)
    dialogWin.wait_window(dialogWin)


def makeResizeable(window):
    """
    Make a window resizeable (still need to use the sticky attribute
    when gridding widgets)

    Source: http://infohost.nmt.edu/tcc/help/pubs/tkinter/web/root-resize.html

    window - the window to become resizeable
    """
    window.winfo_toplevel().rowconfigure(0, weight=1)
    window.winfo_toplevel().columnconfigure(0, weight=1)
    window.rowconfigure(0, weight=1)
    window.columnconfigure(0, weight=1)


def removeDuplicatePathsFromList(inputList):
    #inputList = list(set(inputList)) # Doesn't maintain ordering, which may be important
    logger.info('input list:' + str(inputList))
    seen = dict()
    result = list()
    for x in inputList:
        if x == '':
            continue  # Don't want to add a blank path to the list
        if not x.endswith(sep):
            x += sep
        if x not in seen:
            seen[x] = 1
            result.append(x)

    logger.info('outputList: ' + str(result))
    return result


def listToString(inputList, delimiter=pathsep):
    """
    Converts a list into a delimited string
    inputList - the list (assumed to be of string members
    delimiter - the delimiter to use (default is ';')
        used ; instead of : since often these are used with filenames and Windows has C:\FILE
    """

    returnString = ''
    if type(inputList) != type(list()):
        logger.error('List to string got a ' + str(type(inputList)) + ' instead of a list, value:' + str(inputList))
        return returnString
    for item in inputList:
        returnString += item + delimiter

    return returnString[:-len(delimiter)] # don't return the final delimiter

def directoryListToIncludeOptionString(directoryList):
    return listToStringWithPrefix(directoryList, ' -I')

def listToStringWithPrefix(inputList, prefix):
        if len(inputList) > 0:
            return prefix + prefix.join(inputList)
        else:
            return ''

def listToFilePath(inputList, addDelimiterToEnd=True):
    returnString = ''
    returnString += sep.join(inputList)
    if addDelimiterToEnd:
        return returnString + sep
    else:
        return returnString


def isHiddenDir(pathname):
    if platform.system() == 'Linux':
        return pathname.startswith('.')

    elif platform.system() == 'Windows':
        """ TODO: fix this for windows systems """
        try:
            # eclipse says GetFileAttributesW is undefined variable...but it is a valid python command
            attrs = ctypes.windll.kernel32.GetFileAttributesW(unicode(pathname))
            assert attrs != -1
            result = bool(attrs & 2)
        except (AttributeError, AssertionError):
            result = False
        return result

    else:  # For now if not Linux or Windows assume not hidden
        return False


def trueFalseString(boolValue):
    if boolValue == True:
        return 'True'
    if boolValue == False:
        return 'False'
    logger.warning('boolValue was not a boolean: ' + str(boolValue))
    return None

def launchXterm(initialDir):
    """
    Launches an xterm with initialDir as its current working directory
    """

    logger.info('initialDir: ' + initialDir)

    if isdir(initialDir):
        cmdLine = "xterm -e 'cd " + initialDir + " && /bin/bash' &"
        subprocess.call(cmdLine, shell=True)


def addFlagToString(obj, flag):
    """
    returns flag added to the string obj, stripping lead and following white spaces and only putting one space in.
    Does not add the flag if it already is there
    
    i.e. obj = '-flag1'; flag = ' -flag2 '
    robj = addFlagToString(obj, flag) => robj = '-flag1 -flag2'
    
    i.e. obj = '-flag1'; flag = '-flag1'
    robj = addFlagToString(obj, flag) => robj = '-flag1'
    """
    minFlag = flag.lstrip().rstrip()

    if obj.find(minFlag) == -1:
        obj += ' ' + minFlag
    return obj


def toString(obj):
    """ Converts obj instance into a string """
    logger.info('')
    if obj is None:
        return 'None'
    showList = sorted(set(obj.__dict__))
    return (str(type(obj).__name__) + "(%i):\n" % id(obj)) + "\n".join(
        [" %s: %s" % (key.rjust(8), obj.__dict__[key]) for key in showList])


def safeOpen(fullPathname, mode, encoding='utf-8'):
    return codecs.open(fullPathname, mode, encoding)


def moveFile(filename, srcDir, destDir, keepBackup=False):
    """
    moves filename from srcDir to destDir, if keepBackup is True then if filename exists in the destDir, the existing
    file is moved to filename.bkup# where the # is the first whole number that doesn't already have a file saved with the same name
    """
    logger.info(
        'filename: ' + filename + ' scrDir: ' + srcDir + ' destDir: ' + destDir + 'keepBackup: ' + str(keepBackup))

    if not destDir.endswith(sep):
        destDir += sep
    if not srcDir.endswith(sep):
        srcDir += sep

    dest = destDir + filename
    if isfile(dest):
        if not keepBackup:
            remove(dest)
        else:
            i = 1
            while isfile(dest + '.bkup' + str(i)):
                i += 1
            move(dest, dest + '.bkup' + str(i))

    src = srcDir + filename
    logger.debug("Move file\nsrc: " + src + "\ndest: " + dest)
    move(src, dest)


def copy_files(src_glob, dst_folder, **kw):
    logger.info('src_glob:' + str(src_glob) + ' dstFolder:' + str(dst_folder))
    if not isdir(dst_folder):
        mkdir_p(dst_folder)
    for fname in iglob(src_glob):
        copy(fname, dst_folder)

def move_files(src_glob, dst_folder):
    logger.info('src_glob:' + str(src_glob) + ' dstFolder:' + str(dst_folder))
    for fname in iglob(src_glob):
        move(fname, dst_folder)


def mkdir_p(path):
    try:
        makedirs(path)
    except OSError as exc: # Python >2.5
        if exc.errno == errno.EEXIST:
            pass
        else:
            raise

def rmdir_p(path):
    try:
        rmtree(path)
    except OSError as exc: # Python >2.5
        raise

def copyFile(srcFile, destination, **kw):
    """
    copies filename from srcDir to destDir, if overwrite is false, will not overwrite file if it exist
    if keepBackup is True then if filename exists in the destDir, the existing
    file is moved to filename.bkup# where the # is the first whole number that doesn't already have a
    file saved with the same name
    """
    keepBackup = kw.get('keepBackup', False)
    overwrite = kw.get('overwrite', True)

    logger.info(
        'srcFile:' + srcFile + ' dest: ' + destination + ' keepbackup: ' + str(keepBackup) + ' overwrite: ' + str(
            overwrite))

    if not isfile(srcFile):
        errorMsg = 'srcFile:' + srcFile + ' does not exist'
        logger.error(errorMsg)
        raise IOError(errorMsg)

    if isdir(destination):
        destination = join(destination, basename(srcFile))

    else:
        if not exists(dirname(destination)):
            makedirs(dirname(destination))

        dstPath, dstFile = split(destination)

        if dstFile == '':
            dstFile = basename(srcFile)

        destination = join(dstPath, dstFile)

    if not overwrite:
        if isfile(destination):
            logger.info('Not overwriting existing: ' + destination)
            return False

    if keepBackup:
        if isfile(destination):
            i = 1
            while isfile(destination + '.bkup' + str(i)):
                i += 1
            move(destination, destination + '.bkup' + str(i))
    logger.debug('Copying ' + srcFile + ' to ' + destination)
    copy(srcFile, destination)


def getPathSafeObjectName(objectName):
    #logger.info('ObjectName: ' + objectName)
    returnVal = objectName

    """ Replace all back slash path separators with forward slash separators """
    backSlashIndex = returnVal.find('\\')
    while(backSlashIndex != -1):
        """ Slash at the end of the string...change it and exit loop """
        if backSlashIndex + 1 == len(returnVal):
            returnVal = returnVal[:-1] + '/'
            break

            """ If slash is not followed by space, its a path separator...need to make it a forward slash instead """
        elif returnVal[backSlashIndex + 1] != ' ':
            returnVal = returnVal[:backSlashIndex] + '/' + returnVal[backSlashIndex+1:]
        
        """ find next back slash """
        backSlashIndex = returnVal.find('\\', backSlashIndex + 1)

    """ Replace all spaces with back slash + space """
    spaceIndex = returnVal.find(' ')
    while (spaceIndex != -1):

        """ if space is at beginning (not really good, but...)
            or if space is not already preceded by a \, add a \ in front of the space """
        if spaceIndex == 0 or returnVal[spaceIndex - 1] != '\\':
            returnVal = returnVal[:spaceIndex] + '\ ' + returnVal[spaceIndex + 1:]
            spaceIndex += 2  # go past the \ and the ' ' for the next search
        else:
            spaceIndex += 1 # go past the ' ' for the next search

        if spaceIndex < len(returnVal):
            spaceIndex = returnVal.find(' ', spaceIndex)
        else:
            break  # at the end of the string

    #logger.debug('ReturnVal: ' + returnVal)
    return returnVal

def verifyJavaAddonInstalled():
    java_conformance_app_filename = \
        os.path.join(ConformanceMain.getRootDir(),
                     constants.JAVA_CONFORMANCE_TEST_DIR,
                     constants.JAVA_CONFORMANCE_APP)

    return os.path.isfile(java_conformance_app_filename)


def convertDirectoriesToObjectsInList(objectList, object_suffix, recursive):
    new_object_list = list()
    for o in objectList:
        if isfile(o):
            new_object_list.append(o)

        if isdir(o):
            if recursive:
                for root, dir_names, file_names in os.walk(o):
                    for filename in fnmatch.filter(file_names, '*' + object_suffix):
                        new_object_list.append(os.path.join(root, filename))
            else:
              new_object_list.extend(glob.glob(os.path.join(o, '*' + object_suffix)))

    return new_object_list


def getObjectNamesFromLibraryList(objectList, object_suffix, recursive, prefix='', separator=' '):
    new_object_list = convertDirectoriesToObjectsInList(objectList, object_suffix, recursive)
    return getObjectNamesStringFromList(new_object_list, prefix, separator)

def getObjectNamesStringFromList(objectList, prefix='', separator=' '):
    objectString = ''

    for o in objectList:
        objectString += prefix + getPathSafeObjectName(o) + separator

    objectString = objectString[:-len(separator)]

    return objectString


def removePathBadChars(pathString):
    pathString = pathString.replace('C++', 'CPP')

    badChars = '~`!@#$%^&*()-+=/?.>,<|\\'
    for b in badChars:
        pathString = pathString.replace(b, '_')

    return pathString


def getHeaderDefineConstant(headerFilename, maxLength=60):
    headerDefine = headerFilename
    headerDefine = headerDefine.upper()
    headerDefine = removePathBadChars(headerDefine)

    if len(headerDefine) > maxLength:
        headerDefine = headerDefine[-maxLength:]
        if headerDefine[0].isdigit():
            headerDefine = '_' + headerDefine[-(maxLength - 1):]
    return headerDefine


def findFiles(baseDir, searchPhrase, **kw):
    searchHiddenDirs = kw.get('searchHiddenDirs', False)

    fileList = list()

    for rootdir, subdirs, files in walk(baseDir):

        if not searchHiddenDirs:
            hiddenDir = False
            rootSplit = rootdir.split(sep)
            for rs in rootSplit:
                if rs.startswith('.'):  # Make sure that we aren't looking at hidden directories
                    hiddenDir = True
                    break
            if hiddenDir:
                continue

        relDir = relpath(rootdir, baseDir)
        if relDir == './' or relDir == '.':
            relDir = ''

        for i in iglob(join(rootdir, searchPhrase)):
            fileList.append(join(relDir, basename(i)))

    return fileList


def stringSplitByNumbers(x):
    r = re.compile('(\d+)')
    l = r.split(x)
    return [int(y) if y.isdigit() else y for y in l]


def which(program):  # TODO:  very slow...may try to optimize

    if platform.system() == 'Windows':
        program += ".exe"

    def is_exe(fpath):
        return isfile(fpath) and access(fpath, X_OK)

    fpath, fname = split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        if platform.system() == 'Windows':
            if "ProgramFiles" in environ:
                for path in environ["ProgramFiles"].split(pathsep):
                    exe_file = join(path, program)
                    if is_exe(exe_file):
                        return exe_file
                    for rootdir, subdirs, files in walk(path):
                        if program in files:
                            return join(rootdir, program)

            if "ProgramFiles(x86)" in environ:
                for path in environ["ProgramFiles(x86)"].split(pathsep):
                    exe_file = join(path, program)
                    if is_exe(exe_file):
                        return exe_file
                    for rootdir, subdirs, files in walk(path):
                        if program in files:
                            return join(rootdir, program)
        if "PATH" in environ:
            for path in environ["PATH"].split(pathsep):
                exe_file = join(path, program)
                if is_exe(exe_file):
                    return exe_file

    return None


def findBrowser():
    possibleBrowsers = ('firefox', 'chromium-browser', 'chrome', 'konqueror', 'opera', 'epiphany', 'iexplore')

    for p in possibleBrowsers:
        browser = which(p)
        if browser is not None:
            return browser

    return None


def getDefaultBrowser():
    defaultBrowser = environ.get("BROWSER")
    if defaultBrowser is None or (not isfile(defaultBrowser)):
        logger.debug('No default browser')
        defaultBrowser = findBrowser()

    logger.info('Browser: ' + defaultBrowser)
    return defaultBrowser





def wraptext(tstring, **optArgs):
    regex = optArgs.get('regex', """[\\\\ ,/;:']""")
    maxLength = optArgs.get('maxLength', 120)
    nl = optArgs.get('newLine', '\n')
    continuationPrefix = optArgs.get('continuationPrefix', '')
    maxLength -= len(continuationPrefix)

    if len(tstring) < maxLength:
        return tstring

    lines = tstring.split('\n')

    returnLines = list()
    for l in lines:
        #print 12 * '0123456789'
        #print l
        lineLength = len(l)
        if lineLength < maxLength:
            returnLines.append(l) # line is shorter than maxlength...add it to the list of lines and continue
            continue

        match = re.finditer(regex, l)
        lineStartPt = 0
        lastBreak = 0
        for m in match:
            breakPoint = m.start()
            if breakPoint - lineStartPt > maxLength:
                prefix = continuationPrefix
                if lineStartPt == 0: # not a continuation of a line
                    prefix = ''
                if lastBreak == 0: # too long a line with no obvious break points...just make a long line
                    breakString = m.group(0)
                    #print 'line1: ' + str(lineStartPt) + ':' + str(breakPoint)
                    returnLines.append(prefix + l[lineStartPt:breakPoint] + breakString)
                    lineStartPt = breakPoint + len(breakString)
                else: # make a break at previous breakpoint
                    #print 'line2: ' + str(lineStartPt) + ':' + str(lastBreak)
                    returnLines.append(prefix + l[lineStartPt:lastBreak] + breakString)
                    lineStartPt = lastBreak + len(breakString)
                    lastBreak = 0

                if lineLength - lineStartPt < maxLength:  # Rest of line is less than max length
                    #print 'line3: ' + str(lineStartPt) + ':END'
                    returnLines.append(prefix + l[lineStartPt:])
                    break
            else:
                lastBreak = breakPoint
                breakString = m.group(0)

        if lastBreak != 0: # Case where the end of the line was too long, but no obvious break points just make long line (again)
            prefix = continuationPrefix
            if lineStartPt == 0:
                prefix = ""
            returnLines.append(prefix + l[lineStartPt:])
            #print 'line4: ' + str(lineStartPt) + ':END'


    return '\n'.join(returnLines)



def logSubprocessResults(command, resultsNamedTuple, logFilename, **optArgs):
    mode = optArgs.get('mode', 'w')
    encoding = optArgs.get('encoding', 'utf-8')
    otherInfo = optArgs.get('otherInfo', dict())


    logfile = safeOpen(logFilename, mode, encoding)
    logfile.write('command: \n' + str(command) + '\n')
    for key, value in otherInfo.iteritems():
        logfile.write(str(key) + ': ' + str(value) + '\n')

    logfile.write('return code: \n' + str(resultsNamedTuple.returncode) + '\n')
    if len(resultsNamedTuple.stderr) == 0:
        logfile.write('Output: \n' + resultsNamedTuple.stdout + '\n')
    else:
        logfile.write('stderr: \n' + resultsNamedTuple.stderr + '\n')
        logfile.write('stdout: \n' + resultsNamedTuple.stdout + '\n')
    logfile.close()


def subprocessWithSpecifiedPath(command, specifiedPathList, workingDir=getcwd(), otherEnvironmentDict=dict(), redirectStderr=True):
    logger.debug('command: [' + str(command) + '] pathlist: [' + str(specifiedPathList) + '] workingDir: [' + str(
        workingDir) + '] environ: [' + str(otherEnvironmentDict) + ']')

    environVar = deepcopy(environ.data)
    logger.debug('env: ' + str(environVar))
    pathString = pathsep.join(specifiedPathList).encode('ascii', 'ignore')
    #print 'PATH: ' + pathString
    environVar['PATH'] = pathString + pathsep + environVar['PATH']
    #print 'environPATH: ' + environVar['PATH']
    environVar.update(otherEnvironmentDict)

    #logger.debug('Path: ' + environVar['PATH'])
    ReturnValues = namedtuple('ReturnValues', ['returncode', 'stdout', 'stderr'])

    try:
        if redirectStderr:
            stderrResult = subprocess.STDOUT
        else:
            stderrResult = subprocess.PIPE

        result = subprocess.Popen(command, cwd=workingDir, shell=True, env=environVar, stdout=subprocess.PIPE,
                                  stderr=stderrResult)
        stdout, stderr = result.communicate()

        if stdout is None:
            stdout = ''
        if stderr is None:
            stderr = ''

        #if redirectStderr:
        #    stderr = None

        returnvalues = ReturnValues(result.returncode, stdout.decode('utf-8'), stderr.decode('utf-8'))

    except Exception as x:
        stderr = 'subprocess returned an exception, ' + x.message + ' command: ' + str(command) + ' workingDir: ' + str(
            workingDir) + 'Env: ' + str(environVar)
        returnvalues = ReturnValues(1, "", stderr)

    return returnvalues

# modified subprocessWithSpecifiedPath to run items without shell (shell=False)
# CHANGE: command needs to contain the fullpath to the executable
def subprocessWithSpecifiedPath_noShell(command, specifiedPathList, workingDir=getcwd(), otherEnvironmentDict=dict(), redirectStderr=True):
    # arg_str_list that gets formed from command string is really weird if " are not removed and causes errors in Java Conformance
    #command = command.replace('"', '')     // doesn't work if arguments contain spaces
    logger.debug('command: [' + str(command) + '] pathlist: [' + str(specifiedPathList) + '] workingDir: [' + str(
        workingDir) + '] environ: [' + str(otherEnvironmentDict) + ']')

    logger.info('subprocessWithSpecifiedPath()')

    environVar = deepcopy(environ.data)
    logger.debug('env: ' + str(environVar))
    pathString = pathsep.join(specifiedPathList).encode('ascii', 'ignore')
    #print 'PATH: ' + pathString
    environVar['PATH'] = pathString + pathsep + environVar['PATH']
    #print 'environPATH: ' + environVar['PATH']
    environVar.update(otherEnvironmentDict)

    #logger.debug('Path: ' + environVar['PATH'])
    ReturnValues = namedtuple('ReturnValues', ['returncode', 'stdout', 'stderr'])

    try:
        if redirectStderr:
            stderrResult = subprocess.STDOUT
        else:
            stderrResult = subprocess.PIPE

        # create a sting list of argument for Popen with shell=False
        arg_str_list = shlex.split(command, posix=False)
        # The first argument, the executable name, needs to be a full path
        # in this case, the path the JavaConformance.exe is also the workingDir
        logger.debug('arg_str_list for Popen():\n')
        #if len(prependExePath) > 0:
        #    logger.debug('prependExePath = ' + prependExePath)
        #    arg_str_list[0] = os.path.join(prependExePath, arg_str_list[0])
        logger.debug(str(arg_str_list))

        result = subprocess.Popen(arg_str_list, cwd=workingDir, shell=False, stdout=subprocess.PIPE, stderr=stderrResult)
        stdout, stderr = result.communicate()

        if stdout is None:
            stdout = ''
        if stderr is None:
            stderr = ''

        #if redirectStderr:
        #    stderr = None

        returnvalues = ReturnValues(result.returncode, stdout.decode('utf-8'), stderr.decode('utf-8'))

    except Exception as x:
        stderr = 'subprocess returned an exception, ' + x.message + ' command: ' + str(command) + ' workingDir: ' + str(workingDir) + 'Env: ' + str(environVar)
        returnvalues = ReturnValues(1, "", stderr)

    return returnvalues
def callback(value):
    print 'Returned a ' + str(value)

if __name__ == '__main__':
    import ConformanceMain

    ConformanceMain.initializeLogger()
    logger = logging.getLogger(__name__)

    baseTestString = "The quick brown fox jumps over the lazy dog."
    i = 1
    testString = str(i) + baseTestString
    i += 1
    testString += str(i) + baseTestString.replace(' ', '/')
    i += 1
    testString += str(i) + baseTestString.replace(' ', '\\') + '\n'
    i += 1
    testString += str(i) + baseTestString.replace(' ', ':')
    i += 1
    testString += str(i) + baseTestString.replace(' ', '') + '\n'
    i += 1
    testString += str(i) + baseTestString.replace(' ', '')
    i += 1
    testString += str(i) + baseTestString.replace(' ', ':')

    testString = \
"""make.py[0]: Entering directory 'C:\\Users\\dfann\\Documents\\FACE\\B194\\compilerSpecific\\C'
C:\\Users\\dfann\\Documents\\FACE\\B194\\compilerSpecific\\C\\Makefile:24:0$ gcc -std=c99 -fno-builtin -nostdinc      -I../../compilerSpecific/C/exactTypesDefinitions -o allowedDefinitions/CompilerSpecific.o -c allowedDefinitions/CompilerSpecific.c
C:\\Users\\dfann\\Documents\\FACE\\B194\\compilerSpecific\\C\\Makefile:27:0$ ar cr  compilerSpecificGoldLib.a allowedDefinitions/CompilerSpecific.o """

    #print testString
    print wraptext(testString)



