"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""

__author__ = 'david.fann@vanderbilt.edu'

import sys
from xml.etree import ElementTree as ET
import ConformanceMain

__author__ = 'dfann'

import os.path
import collections
import platform

FACE_TESTSUITE_NAME = "FACE Conformance Test Suite"
APP_DIR = "face_conformance_app"
SPLASH_X_OFFSET = 50
SPLASH_Y_OFFSET = 450
SNAPSHOT_SPLASH_SCREEN_TIME = 500
VERSION_SPLASH_SCREEN_TIME = 5000

FACE_CONFORMANCE_APP_DIR = os.path.join(ConformanceMain.getRootDir(), APP_DIR)
QUESTION_GIF = os.path.join(FACE_CONFORMANCE_APP_DIR, 'question.gif')
INFO_GIF = os.path.join(FACE_CONFORMANCE_APP_DIR, 'info.gif')
ERROR_GIF = os.path.join(FACE_CONFORMANCE_APP_DIR, 'error.gif')
WARNING_GIF = os.path.join(FACE_CONFORMANCE_APP_DIR, 'warning.gif')


GOLD_LIB_ROOT = "goldStandardLibraries"
ADA_GOLD_LIB_DIR = os.path.join(GOLD_LIB_ROOT, 'Ada')
C_GOLD_LIB_DIR = os.path.join(GOLD_LIB_ROOT, 'C')
CPP_GOLD_LIB_DIR = os.path.join(GOLD_LIB_ROOT, 'C++')
JAVA_GOLD_LIB_DIR = os.path.join(GOLD_LIB_ROOT, 'Java')

CONFORMANCE_TEST_ROOT = "conformanceInterfaceTests"
ADA_CONFORMANCE_TEST_DIR = os.path.join(CONFORMANCE_TEST_ROOT, 'Ada')
C_CONFORMANCE_TEST_DIR = os.path.join(CONFORMANCE_TEST_ROOT, 'C')
CPP_CONFORMANCE_TEST_DIR = os.path.join(CONFORMANCE_TEST_ROOT, 'C++')
JAVA_CONFORMANCE_TEST_DIR = os.path.join(CONFORMANCE_TEST_ROOT, 'Java')


COMPILER_SPECIFIC_DIR = "compilerSpecific"
COMPILER_SPECIFIC_TYPES_DIR = "systemLibraryDefinitions"
COMPILER_SPECIFIC_ALLOWED_DIR = "allowedDefinitions"

DATAFILES_DIR = 'datafiles'
LOG_DIR = "logFiles"
C_SAMPLE_DIR = os.path.join('sample', 'C')
CPP_SAMPLE_DIR = os.path.join('sample', 'C++')
ADA_SAMPLE_DIR = os.path.join('sample', 'Ada')
JAVA_SAMPLE_DIR = os.path.join('sample', 'Java')

PROTOBUF_EGG = os.path.join(FACE_CONFORMANCE_APP_DIR, 'protobuf-2.5.0-py2.7.egg')
DEFAULT_CONFIGURATION_FILENAME = 'FACE_ConformanceTest_Configuration.cfg'
TOOLS_MARKING_XML_FILE = os.path.join(DATAFILES_DIR, 'FACEToolsMarkings.xml')
DISTRIBUTION_FILE = os.path.join(DATAFILES_DIR, "distributionStatement.txt")
C_CPP_GENERATOR_LICENSE_FILE = os.path.join('C_C++_POSIX_Gen', 'FaceConformanceLibraryGenerator', 'License.h')
TestSuiteVersion = collections.namedtuple('TestSuiteVersion', 'name toolVersion faceStandardVersion')
RSYNC_BUILDSCRIPTS_FILENAME = os.path.join('buildscripts', 'rsync_test_suite_dist.sh')
SOCKET_SERVER_TIMEOUT_IN_SECONDS = 20

JAVA_APP_DIR = os.path.join(APP_DIR, 'java_apps')
JAVA_CONFORMANCE_TEST_DIR = 'Java_Conformance'
JAVA_GOLD_DIR = os.path.join(GOLD_LIB_ROOT, 'Java')

JAVA_CONFORMANCE_APP = 'JavaConformance'
if platform.system() == 'Windows':
    JAVA_CONFORMANCE_APP += '.exe'

JAVA_CONFORMANCE_EXTENSION_OS_LIST = ['CentOS/RHEL 6.x', 'Ubuntu 12.04', 'Windows 7', 'Windows 8']
JAVA_TS_GOLD = 'tss.javagold'

DM_CONFORMANCE_CODE_GEN_SUBDIR = os.path.join(JAVA_APP_DIR, 'DMConformanceCodeGenerator')
DM_CONFORMANCE_CODE_GEN_2_0_JAR = 'DMConformanceCodeGenerator-2.0.8.jar'
DM_CONFORMANCE_CODE_GEN_2_1_JAR = 'edu.vanderbilt.isis.face.v21.datamodel.conformance.codegen-2.1.0.jar'

DM_CONFORMANCE_TEST_SUBDIR = os.path.join(JAVA_APP_DIR, 'DMConformanceTest')
DM_CONFORMANCE_TEST_2_0_JAR = 'DMConformanceTest-2.0.10.jar'
DM_CONFORMANCE_TEST_2_1_JAR = 'DMConformanceTest-2023.8.1.jar'    #CR-993: 2023.8.1, CR-900: 2022.5.1, CR-753: 2021.2.1
DM_CPP_HELPER_INCLUDE = 'DM_Helper_Include'



DM_DATATYPE_GEN_SUBDIR = os.path.join( ConformanceMain.getRootDir(), JAVA_APP_DIR, 'DMDatatypeGenerator' )
def _find_DM_DTG_Jar():
    import os
    import glob
    DTG_GLOB = 'DataTypeGenerator-*.jar'
    ERROR_CODE = 'ERROR_CouldNotFindDataTypeGenerator.jar'
    try:
        for file in glob.glob( os.path.join(DM_DATATYPE_GEN_SUBDIR, DTG_GLOB) ):
            return file
        return ERROR_CODE
    except BaseException as e:
        return ERROR_CODE
DM_DATATYPE_GEN_JAR = _find_DM_DTG_Jar()

EXACT_TYPES_HEADER = "CONFORMANCE_TEST_FACE_EXACT_TYPES.h"
FACE_SPECIFIED_TSS_IDL = os.path.join(DATAFILES_DIR, "FACE_SPECIFIED_TYPE_TSS.idl")
ARINC_PART1_HEADER_NAME = 'ARINC653.h'
ARINC_PART2_HEADER_NAME = 'ARINC653P2.h'
C_ARINC_653_CONFORMANCE_DIR = os.path.join(CONFORMANCE_TEST_ROOT, 'C', 'OSS', 'ARINC-653')
C_ARINC_653_GOLD_LIB_DIR = os.path.join(GOLD_LIB_ROOT, 'C', 'src', 'OSS', 'ARINC-653')
CPP_ARINC_653_GOLD_LIB_DIR = os.path.join(GOLD_LIB_ROOT, 'C++', 'src', 'OSS', 'ARINC-653')

ADA_ARINC_653_GOLD_LIB_DIR = os.path.join(GOLD_LIB_ROOT, 'Ada', 'src', 'OSS', 'ARINC-653')
ADA_ARINC_653_CONFORMANCE_DIR = os.path.join(CONFORMANCE_TEST_ROOT, 'Ada', 'OSS', 'ARINC-653')
ADA_BASE_EXTENSION = '.typeless'
ADA_TSS_FILENAME = 'face-ts'
ADA_FACE_BASIC_TYPES_CONFORMANCE_TEST_FILE = os.path.join(DATAFILES_DIR, 'FACE_Basic_Types_Conformance_Test.adb')

FACE_STANDARD_VERSION_2_0 = 'Version 2.0.1'  # 2.0 is dead...don't really want to keep up with another version
FACE_STANDARD_VERSION_2_1 = 'Version 2.1'
FACE_STANDARD_VERSION_2_1_1 = 'Version 2.1.1'

C_CPP_LIBRARY_DIR = os.path.join('C_C++_POSIX_Gen', 'FaceConformanceLibrary')


def getfaceOrFACE():
    version = getFACEStandardVersion()
    if version == FACE_STANDARD_VERSION_2_0:
        return 'face'
    elif version in [FACE_STANDARD_VERSION_2_1, FACE_STANDARD_VERSION_2_1_1]:
        return 'FACE'
    else:
        return 'getfaceOrFACE: UNKNOWN FACE VERSION' + version


# This is the Java executable for the DM stuff, NOT the one used to compile things for testing.
_JAVA_EXE = None
def getJavaExe():
    global _JAVA_EXE

    if _JAVA_EXE is None:
        import check_java
        _JAVA_EXE = '\"' + check_java.findJava()[1] + '\"'

    return _JAVA_EXE

_faceMarkings = None

def getVersionedFACETestSuiteName(rootDir=ConformanceMain.getRootDir()):

    global _faceMarkings
    # Only want to parse xml file once per run...once it is done, just use stored values.  xml file will not change
    if _faceMarkings is None:
        _faceMarkings = FACEToolsMarkings(os.path.join(rootDir, TOOLS_MARKING_XML_FILE))
        _faceMarkings.getToolInfo(maxLineLength=60)

    result = TestSuiteVersion(FACE_TESTSUITE_NAME, '\n'.join(_faceMarkings.ToolsVersion), '\n'.join(_faceMarkings.FACEStandardVersion))
    return result


def isSnapshot():

    rootDir=ConformanceMain.getRootDir()

    global  _faceMarkings

    # Only want to parse xml file once per run...once it is done, just use stored values.  xml file will not change
    if _faceMarkings is None:
        _faceMarkings = FACEToolsMarkings(os.path.join(rootDir, TOOLS_MARKING_XML_FILE))
        _faceMarkings.getToolInfo(maxLineLength=60)

    # if snapshot in name, then it is a snapshot
    return 'snapshot' in ' '.join(_faceMarkings.ToolsVersion).lower()


def getFACEStandardVersion(rootDir=ConformanceMain.getRootDir()):
    global  _faceMarkings
    # Only want to parse xml file once per run...once it is done, just use stored values.  xml file will not change
    if _faceMarkings is None:
        _faceMarkings = FACEToolsMarkings(os.path.join(rootDir, TOOLS_MARKING_XML_FILE))
        _faceMarkings.getToolInfo(maxLineLength=60)

    return _faceMarkings.FACEStandardVersion[-1]

def getReleaseDirectory():
    version_info = getVersionedFACETestSuiteName()
    name = FACE_TESTSUITE_NAME.replace(' ', '')
    root_dir = ConformanceMain.getRootDir()

    release = version_info.toolVersion.strip('Version ')

    release_dir = os.path.join(root_dir, name + '-' + release)

    return release_dir


def getDMConformanceCodeGenJar():

    if getFACEStandardVersion() == FACE_STANDARD_VERSION_2_0:
        return DM_CONFORMANCE_CODE_GEN_2_0_JAR
    else:
        return DM_CONFORMANCE_CODE_GEN_2_1_JAR

def getDMConformanceTestJar():
    if getFACEStandardVersion() == FACE_STANDARD_VERSION_2_0:
        return DM_CONFORMANCE_TEST_2_0_JAR
    else:
        return DM_CONFORMANCE_TEST_2_1_JAR

class FACEToolsMarkings(object):

    def __init__(self, xmlFilename, **params):
        '''
        Constructor
        '''

        tree = ET.parse(xmlFilename)
        self.root = tree.getroot()

        self.language = params.get('language', 'text')

    #def getProductMarking(self, **params):

    def getToolInfo(self, **params):
        self.maxLineLength = params.get('maxLineLength', None)

        self.FACEStandardVersion = list()
        self.ToolsVersion = list()
        self.productMarking = list()
        if sys.version_info[:2] >= (2,7):
            for child in self.root.iter():
                if child.tag == 'ProductMarking':
                    for c in child.iter():
                        if c.tag == 'ProductMarking': continue
                        self.productMarking.append(c.text)

                elif child.tag == 'FACETechnicalStandardEdition':
                    for c in child.iter():
                        if c.tag == 'FACETechnicalStandardEdition': continue
                        self.FACEStandardVersion.append(c.text)

                elif child.tag == 'FACEToolsVersion':
                    for c in child.iter():
                        if c.tag == 'FACEToolsVersion': continue
                        self.ToolsVersion.append(c.text)

        else: # Old python version
            for child in self.root.getiterator():
                if child.tag == 'ProductMarking':
                    for c in child.getiterator():
                        if c.tag == 'ProductMarking': continue
                        self.productMarking.append(c.text)

                elif child.tag == 'FACETechnicalStandardEdition':
                    for c in child.getiterator():
                        if c.tag == 'FACETechnicalStandardEdition': continue
                        self.FACEStandardVersion.append(c.text)

                elif child.tag == 'FACEToolsVersion':
                    for c in child.getiterator():
                        if c.tag == 'FACEToolsVersion': continue
                        self.ToolsVersion.append(c.text)

        self.productMarking = self.formatMarking(self.productMarking)

    def formatMarking(self, marking):
        newMarks = list()
        if self.maxLineLength is not None:
            for m in marking:
                if len(m) < self.maxLineLength:
                    newMarks.append(m)
                else:
                    line = []
                    lineLength = 0
                    for word in m.split(' '):
                        lineLength += len(word) + 1
                        if lineLength > self.maxLineLength:
                            newMarks.append(' '.join(line))
                            line = [word]
                            lineLength = len(word) + 1
                        else:
                            line.append(word)
                    newMarks.append(' '.join(line))
                newMarks.append('')
        else:
            for m in marking:
                newMarks.append(m)
                newMarks.append('')

        maxlength = 0
        if self.language != 'text':
            for n in newMarks:
                if len(n) > maxlength:
                    maxlength = len(n)

        if self.language == 'C':
            marking = newMarks
            newMarks = list()
            newMarks.append('/**' + maxlength * '*' + '**/')

            for m in marking:
                newMarks.append('/* ' + m + (maxlength - len(m)) * ' ' + ' */')

            newMarks.append('/**' + maxlength * '*' + '**/')

        elif (self.language == 'C++'):
            marking = newMarks
            newMarks = list()
            newMarks.append('///' + maxlength * '/')

            for m in marking:
                newMarks.append('// ' + m)

            newMarks.append('///' + maxlength * '/')

        elif (self.language == 'Ada'):
            marking = newMarks
            newMarks = list()
            newMarks.append('---' + maxlength * '-')

            for m in marking:
                newMarks.append('-- ' + m)

            newMarks.append('---' + maxlength * '-')



        return newMarks

    def wrap(self, text, width=80):
        lines = []
        for paragraph in text.split('\n'):
            line = []
            len_line = 0
            for word in paragraph.split(' '):
                len_word = len(word)
                if len_line + len_word <= width:
                    line.append(word)
                    len_line += len_word + 1
                else:
                    lines.append(' '.join(line))
                    line = [word]
                    len_line = len_word + 1
            lines.append(' '.join(line))
        return '\n'.join(lines)


    def getReleaseStatement(self):
        pass

    def printAllMarkings(self):
        if sys.version_info[:2] >= (2,7):
            for child in self.root.iter():
                print child.tag + ':"' + child.text + '"'
        else:
            for child in self.root.getiterator():
                print child.tag + ':"' + child.text + '"'






if __name__ == '__main__':
    testSuiteInfo = getVersionedFACETestSuiteName()

    print 'Name: ' + testSuiteInfo.name + ' standard: ' + testSuiteInfo.faceStandardVersion + \
          ' tool: ' + testSuiteInfo.toolVersion

    print 'FACEStandardVersion: {0}'.format(getFACEStandardVersion())
    print 'FACE or face: {0}'.format(getfaceOrFACE())
