#!/usr/bin/env python
"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""

"""
---------------------------------------------------------------------------------
fancyentry - classes/methods useful to the GUI module associated with the FACE Conformance Test Suite

This module contains one class and zero functions
classes:
    FancyEntry - tkinter based widget containing a label, an entry box, a delete button, and a '...' button which opens a file browser
functions:
    None
"""
__author__ = 'johnwilliams@isis.vanderbilt.edu'

import Tkinter
from os.path import expanduser, dirname, isdir, isfile
from os import pathsep
import Dialogs
import constants


# '...' button editor type
NONE        = 0
FILES       = 1
DIRECTORIES = 2
FILE        = 3
DIRECTORY   = 4
CLASS_PATH  = 5

class FancyEntry(Tkinter.Frame):
    """
    The FancyEntry is a widget with a label, 'X' delete button, and '...' file browser button. Multiple
    entries are separated by os.pathsep. Retrieve the entry contents from self.var as a Tkinter.StringVar.
    """

    __initFileDir = expanduser('~')



    def __init__(self, master, **kw):
        """
        master - tkinter master window associated with the widget

        keywords:
        text - text to use as a label. Default is empty string
        selectionType - selection types: NONE, FILES, DIRECTORIES
        labelWidth - width of the label
        entryWidth - width of the entry
        fileTypes - types of files listed in the file selection dialog (dict)
        title - title of file/dir selection dialog
        """
        text                = kw.pop("text",            ""  )
        selectionType       = kw.pop("selectionType",   NONE)
        labelWidth          = kw.pop("labelWidth",      12  )
        entryWidth          = kw.pop("entryWidth",      20  )
        self.fileTypes      = kw.pop("fileTypes",       None)
        self.title          = kw.pop("title",           None)
        self.validateCommand= kw.pop("validatecommand", None)  # MUST return True/False, otherwise Tkinter screws up
        self.invalidCommand = kw.pop("invalidcommand",  None)  # What to do if there is an invalid entry
        validateMethod      = kw.pop("validate",  "focusout")  # Assume only check when entry loses focus
        self.__initDir      = kw.pop("initDir",         None)
        self.defaultValidate = kw.pop("defaultValidate",True)
        self.dialogPrompt    = kw.pop("dialogPrompt", "")
        self.lastValue = ""

        if self.__initDir is not None and not isdir(self.__initDir):
            self.__initDir = None
        self.overrideValidation = False

        Tkinter.Frame.__init__(self, master, **kw)

        self.var = Tkinter.StringVar()

        entryOptions = {"textvariable": self.var, "width": entryWidth }

        if self.validateCommand is not None:
            entryOptions["validate"] = validateMethod
            entryOptions["validatecommand"] = self.validateCommand
            if self.invalidCommand is not None:
                entryOptions["invalidcommand"] = self.invalidCommand

        ''' THIS ISN'T WORKING AS WELL AS I'D LIKE...FOR NOW, NOT GOING TO IMPLEMENT IT
        elif self.defaultValidate:  # Use default validation methods
            if selectionType == DIRECTORY:
                entryOptions["validate"] = validateMethod
                entryOptions["validatecommand"] = self.validateDirectory
            elif selectionType == DIRECTORIES:
                entryOptions["validate"] = validateMethod
                entryOptions["validatecommand"] = self.validateDirectories
            elif selectionType == FILE:
                entryOptions["validate"] = validateMethod
                entryOptions["validatecommand"] = self.validateFile
            elif selectionType == FILES:
                entryOptions["validate"] = validateMethod
                entryOptions["validatecommand"] = self.validateFiles
            elif selectionType == NONE:
                pass  # No validation if not a file/directory type
            else:
                raise TypeError("Unknown selectionType: " + str(selectionType))

            if self.invalidCommand is not None:
                entryOptions["invalidcommand"] = self.invalidCommand
        '''
        self.__labelVar = Tkinter.StringVar()
        self.__labelVar.set(text)
        self.__labelWidget = Tkinter.Label(self, textvariable=self.__labelVar, width=labelWidth, anchor=Tkinter.E)
        self.__entryWidget = Tkinter.Entry(self, entryOptions)
        self.__deleteWidget = Tkinter.Button(self, text=u'\u2718', command=self.__deleteCmd, foreground='red')

        self.__labelWidget.grid(row=0, column=0, sticky=Tkinter.E)
        self.__entryWidget.grid(row=0, column=1, sticky=Tkinter.W)
        Tkinter.Frame(self, width=2).grid(row=0, column=2)  # two pixel spacer
        self.__deleteWidget.grid(row=0, column=4, sticky=Tkinter.W)

        if selectionType == FILES:
            self.__browserButton = Tkinter.Button(self, text='...', command=self.__filesDialog)
            self.__browserButton.grid(row=0, column=3, sticky=Tkinter.W)
        elif selectionType == DIRECTORIES:
            self.__browserButton = Tkinter.Button(self, text='...', command=self.__directoriesDialog)
            self.__browserButton.grid(row=0, column=3, sticky=Tkinter.W)
        elif selectionType == DIRECTORY:
            self.__browserButton = Tkinter.Button(self, text='...', command=self.__directoryDialog)
            self.__browserButton.grid(row=0, column=3, sticky=Tkinter.W)
        elif selectionType == FILE:
            self.__browserButton = Tkinter.Button(self, text='...', command=self.__fileDialog)
            self.__browserButton.grid(row=0, column=3, sticky=Tkinter.W)
        elif selectionType == CLASS_PATH:
            self.__browserButton = Tkinter.Button(self, text='...', command=self.__classPathDialog)
            self.__browserButton.grid(row=0, column=3, sticky=Tkinter.W)

    def setLabel(self, text):
        self.__labelVar.set(text)

    def getLabel(self):
        return self.__labelVar.get()

    def revertMessage(self):
        print "REVERT!"

    def setInitDir(self, initDir):
        if initDir is not None and isdir(initDir):
            self.__initDir = initDir

    def setValue(self, newValue):
        self.__entryWidget.delete(0, Tkinter.END)
        self.__entryWidget.insert(0,newValue)

    def validateDirectory(self):
        if self.overrideValidation:
            return True

        returnValue = True
        varGet = self.var.get()

        if varGet != "" and not isdir(varGet):
            Dialogs.ErrorMessageBox(constants.FACE_TESTSUITE_NAME,
                                   "Directory: " + varGet + " is not a valid directory", parent=self)
            self.setValue(self.lastValue)
            returnValue = False
        return returnValue

    def validateDirectories(self):
        if self.overrideValidation:
            return True

        invalidDirs = list()
        for d in self.var.get().split(pathsep):
            if not isdir(d):
                invalidDirs.append(d)

        if len(invalidDirs) > 0 and self.var.get() != "":
            Dialogs.ErrorMessageBox(constants.FACE_TESTSUITE_NAME,
                                   "Invalid Directory(ies): \n" + "\n".join(invalidDirs), parent=self)
            self.setValue(self.lastValue)
        return True

    def validateFile(self):
        if self.overrideValidation:
            return True

        varGet = self.var.get()
        if varGet != "" and not isfile(varGet):
            Dialogs.ErrorMessageBox(constants.FACE_TESTSUITE_NAME,
                                   "File: " + varGet + " is not a valid filename", parent=self)
            self.setValue(self.lastValue)
        return True

    def validateFiles(self):
        if self.overrideValidation:
            return True

        invalidFiles = list()
        for f in self.var.get().split(pathsep):
            if not isfile(f):
                invalidFiles.append(f)

        if len(invalidFiles) > 0 and self.var.get() != "":
            Dialogs.ErrorMessageBox(constants.FACE_TESTSUITE_NAME,
                                   "Invalid File(s): \n" + "\n".join(invalidFiles), parent=self)
            self.setValue(self.lastValue)
        return True

    def __deleteCmd(self):
        self.var.set('')

    def __filesDialog(self):
        """ brings up a files dialog window, sets the StringVar self.var to the resulting value"""
        fileList = self.__stringToList(self.var.get())

        initDir = FancyEntry.__initFileDir
        if self.__initDir is not None:
            initDir = self.__initDir # Only overwrite default if it this instance was set specifically

        ld = Dialogs.ListDialog(self, fileList, Dialogs.ListDialog.FILE, title=self.title, initialdir=initDir)
        if ld.returnValue is not None:
            print "Files: " + str(ld.returnValue)
            self.lastValue = pathsep.join(ld.returnValue)
            self.var.set(self.lastValue)
            FancyEntry.__initFileDir = dirname(ld.returnValue[-1])
            self.__initDir = FancyEntry.__initFileDir

    def __fileDialog(self):
        openFilenameOptions = { 'parent': self}
        initDir = FancyEntry.__initFileDir
        if self.__initDir is not None:
            initDir = self.__initDir

        currentFile = self.var.get()
        if len(currentFile) > 0 and isdir(dirname(currentFile)):
            initDir = dirname(currentFile)

        openFilenameOptions['initialdir'] = initDir
        if self.title is not None:
            openFilenameOptions['title'] = self.title
        if self.fileTypes is not None:
            openFilenameOptions['filetypes'] = self.fileTypes
        
        
        returnFile = Dialogs.osAgnosticOpenFilenameDialog(**openFilenameOptions)
        print 'returnFile: [' + str(returnFile) + ']'
        
        if returnFile != '' and returnFile != '()':
            self.var.set(returnFile)
            self.lastValue = returnFile
            FancyEntry.__initFileDir = dirname(returnFile)
            self.__initDir = FancyEntry.__initFileDir

    def __directoryDialog(self):
        returnValue = True
        initDir = FancyEntry.__initFileDir
        if self.__initDir is not None:
            initDir = self.__initDir

        currentDir = self.var.get()
        if len(currentDir) > 0 and isdir(currentDir):
            initDir = currentDir

        returnDir = Dialogs.osAgnosticOpenDirectoryDialog(initialdir=initDir, title=self.title, parent=self)
        print 'returnDir: [' + str(returnDir) + ']' 
        if returnDir != '' and returnDir != '()':
            self.var.set(returnDir)
            gotReturnDir = self.var.get() # Maybe invalid value or user cancelled
            if gotReturnDir != returnDir:
                #self.var.set(gotReturnDir)
                pass # Maybe we don't need to set it now???
                self.setValue(gotReturnDir)
                returnValue = False

            FancyEntry.__initFileDir = gotReturnDir
            self.__initDir = gotReturnDir
            self.lastValue = gotReturnDir

        return returnValue

    def __classPathDialog(self):
        """
        brings up a directories dialog window, sets the StringVar self.var to the resulting value

        keywords:
        appendDirectoryToText - (True or False) flag whether to set dirEntryText to new directory (False) or append value (True)
        initialDir - initial directory to start the file dialog
        title - title for the file dialog window
        """
        dirList = self.__stringToList(self.var.get())

        params = dict()
        if self.title is not None:
            params["title"] = self.title

        initDir = FancyEntry.__initFileDir
        if self.__initDir is not None:
            initDir = self.__initDir

        params["initialdir"] = initDir
        params["prompt"] = self.dialogPrompt

        ld = Dialogs.ListDialog(self, dirList, Dialogs.ListDialog.CLASS_PATH, **params)
        if ld.returnValue is not None:
            self.lastValue = pathsep.join(ld.returnValue)
            self.var.set(self.lastValue)
            FancyEntry.__initFileDir = dirname(ld.returnValue[-1])
            self.__initDir = FancyEntry.__initFileDir




    def __directoriesDialog(self):
        """
        brings up a directories dialog window, sets the StringVar self.var to the resulting value

        keywords:
        appendDirectoryToText - (True or False) flag whether to set dirEntryText to new directory (False) or append value (True)
        initialDir - initial directory to start the file dialog
        title - title for the file dialog window
        """
        dirList = self.__stringToList(self.var.get())

        params = dict()
        if self.title is not None:
            params["title"] = self.title

        initDir = FancyEntry.__initFileDir
        if self.__initDir is not None:
            initDir = self.__initDir

        params["initialdir"] = initDir
        ld = Dialogs.ListDialog(self, dirList, Dialogs.ListDialog.DIRECTORY, **params)
        if ld.returnValue is not None:
            self.lastValue = pathsep.join(ld.returnValue)
            self.var.set(self.lastValue)
            FancyEntry.__initFileDir = ld.returnValue[-1]
            self.__initDir = FancyEntry.__initFileDir

    def __stringToList(self, s):
        dirList = list()
        for d in list(s.split(pathsep)):
            if d != '':
                dirList.append(d)
        return dirList

    def __listToString(self, l):
        """
        Converts a list into a os.pathsep delimited string
        l - the list of string members
        """

        returnString = ''
        for item in l:
            returnString += item + pathsep

        return returnString[:-len(pathsep)] # don't return the final delimiter

    def configure(self, cnf=None, **kw):
        state = kw.pop("state", Tkinter.NORMAL)
        Tkinter.Frame.configure(self, cnf, **kw)

        children = self.children.values()
        for child in children:
            # Not all widgets (such as Frame) have a "state" so try/except to be safe.
            try:
                child.configure(state=state)
            except Tkinter.TclError:
                pass


def verifyQuestion():
    global rootFrame
    yesOrNo = Dialogs.YesNoMessageBox('Yes or no', 'Yes or no question', parent=rootFrame)
    if yesOrNo == 'yes':
        return True
    else:
        return False


rootFrame = None

if __name__ == '__main__':


    fancyFrame = FancyEntry(None, text='This is some text', selectionType=CLASS_PATH, dialogPrompt="This is the dialog prompt")
    global rootFrame
    rootFrame = fancyFrame
    fancyFrame.grid()
    fancyFrame.mainloop()
