#!/usr/bin/env python
"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""

"""
---------------------------------------------------------------------------------
fancyradiogroup - classes/methods useful to the GUI module associated with the FACE Conformance Test Suite

This module contains one class and zero functions
classes:
    FancyRadioGroup - tkinter based widget containing a label and grid of radio buttons
functions:
    None

"""

__author__ = 'johnwilliams@isis.vanderbilt.edu'

import Tkinter


class FancyRadioGroup(Tkinter.Frame):
    """
    The FancyRadioGroup is a widget with a label and a grid of radio buttons. Two radio buttons per line is the
    default. Add radio buttons using add(). Retrieve the selection from self.var as a Tkinter.StringVar.
    """

    __coloffset = 1  # number of columns to offset (1 for the label)

    def __init__(self, master, **kw):
        """
        master - tkinter master window associated with the widget

        keywords:
        text - text to use as a label. Default is empty string
        labelWidth - width of the label
        radioWidth - width of the radio button
        cols - number of columns of radio buttons
        """
        text                = kw.pop("text",       "")
        labelWidth          = kw.pop("labelWidth", 12)
        self.__radioWidth   = kw.pop("radioWidth", 50)
        self.__cols         = kw.pop("cols",        2) + self.__coloffset  # add column offset

        self.__row          = 0
        self.__col          = self.__coloffset

        Tkinter.Frame.__init__(self, master, **kw)

        self.var = Tkinter.StringVar()

        self.__labelWidget = Tkinter.Label(self, text=text, width=labelWidth, anchor=Tkinter.E)
        self.__labelWidget.grid(row=0, column=0, sticky=Tkinter.E)

    def add(self, text, value, selected=False, **additionalParams):
        """Add a new radio button whose label has the text 'text' and self.var value 'value'"""
        radio = Tkinter.Radiobutton(self, text=text, value=value, variable=self.var, width=self.__radioWidth, anchor=Tkinter.W, **additionalParams)
        radio.grid(row=self.__row, column=self.__col, sticky=Tkinter.W)

        if selected:
            self.var.set(value)

        # Increment location of next radio button
        self.__col += 1
        if self.__col == self.__cols:
            self.__col = self.__coloffset
            self.__row += 1
        
        return radio

if __name__ == "__main__":
    root = Tkinter.Tk()
    radio = FancyRadioGroup(root, text="test radio group")
    radio.add(text="option 1", value="opt1")
    radio.add(text="option B", value="optB", selected=True)
    radio.add(text="option 3", value="opt3")
    radio.pack(side=Tkinter.TOP)
    root.mainloop()
