"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""
"""
 idlParser.py

 Modified from an example of using the parsing module to be able to process a subset of the CORBA IDL grammar

 Original (most of the FaceIdlParser.initialize_parser) was Copyrighted (c) 2003, Paul McGuire (MIT License)

"""

__author__ = 'david.fann@vanderbilt.edu'


import os
import sys
import ConformanceMain

egg_path = os.path.join(ConformanceMain.getRootDir(), 'face_conformance_app', 'pyparsing-2.0.1-py2.7.egg')
sys.path.append(egg_path)
egg_path = os.path.join(ConformanceMain.getRootDir(), 'face_conformance_app', 'stringtemplate3-3.1-py2.7.egg')
sys.path.append(egg_path)
egg_path = os.path.join(ConformanceMain.getRootDir(), 'face_conformance_app', 'setuptools-0.6c11-py2.7!\\')
sys.path.append(egg_path)


from pyparsing import Literal, CaselessLiteral, Word, Upcase, OneOrMore, ZeroOrMore, \
        Forward, NotAny, delimitedList, oneOf, Group, Optional, Combine, alphas, nums, restOfLine, cStyleComment, \
        alphanums, printables, empty, quotedString, ParseException, ParseResults, Keyword
import pprint
import collections
import logging
import stringtemplate3
import itertools
import glob

import TestConfiguration
import constants
import commonTools
import set_distro_statement
import ConformanceMain

if __name__ == "__main__":
    ConformanceMain.initializeLogger()


logger = logging.getLogger(__name__)


class IdlTree:

    def __init__(self):
        self.root_node = Node(name='ROOT')
        self.root_node.in_namespace = False
        self.face_specified_type_nodes = None
        self.current_node = self.root_node
        self.data_callback_node = None

    def __iter__(self):
        for v in iter(self.root_node):
            yield v
        yield v

    def add_sibling(self, new_node):
        parent = self.current_node.parent

        if self.current_node == self.root_node:
           parent = self.root_node

        parent.children.append(new_node)
        new_node.parent = parent

        self.current_node = new_node

    def add_child(self, new_node):
        self.current_node.children.append(new_node)
        new_node.parent = self.current_node
        self.current_node = new_node

    def add_parent(self, new_node):
        if self.current_node == self.root_node:
            new_node.parent = self.root_node
            self.root_node.children.append(new_node)

        else:
            new_node.parent = self.current_node.parent
            if new_node.parent.loc > new_node.loc:
                raise RuntimeError('new node came before its parent, new_node: ' + new_node.to_string())

            siblings = list()
            for siblingOrChild in self.current_node.parent.children:
                if siblingOrChild.loc < new_node.loc:
                    siblings.append(siblingOrChild)
                else:
                    new_node.children.append(siblingOrChild)
                    siblingOrChild.parent = new_node

            siblings.append(new_node)
            new_node.parent.children = siblings

        self.current_node = new_node

    # Hack to replace FACE_SPECIFIED_TYPE in the idl and in the register_read_callback
    def overwrite_face_specified_type_nodes(self, new_name):
        if self.face_specified_type_nodes is None:
            self.face_specified_type_nodes = list()
            for n in iter(self):
                if 'type_name' in n.__dict__.keys() and n.type_name == "FACE_SPECIFIED_TYPE":
                        self.face_specified_type_nodes.append(n)

                if n.name.lower() == 'data_callback':
                    self.data_callback_node = n

        for n in self.face_specified_type_nodes:
            n.type_name = 'face.dm.' + new_name
        if self.data_callback_node is not None:
            self.data_callback_node.type_name = 'FACE.TS.Read_Callback.Send_Event_face_dm_' + new_name + '_Access'

    def print_tree(self, indent_text = '-'):
        self.root_node._print_node_tree(0, indent_text)

    def in_tree(self, node):
        return self._search_tree(node, self.root_node)

    def _search_tree(self, node, tree_node):
        if node == tree_node:
            return True
        for n in tree_node.children:
            if self._search_tree(node, n):
                return True

        return False

    def combine_modules(self):
        module_nodes = dict()
        for v in self:
            if isinstance(v, ModuleNode):
                if v.get_fully_specified_name() in module_nodes:  # Same module
                    original_module_node = module_nodes[v.get_fully_specified_name()]
                    if original_module_node.parent != v.parent:
                        raise RuntimeError('modules with same fully specified name should have the same parent: '
                            + str(v.parent) + ' ' + str(original_module_node.parent))

                    # put children of duplicate as children of original
                    # and set parent of duplicate children to be original
                    for c in v.children:
                        original_module_node.children.append(c)
                        c.parent = original_module_node

                    # delete duplicate from it's parent children
                    v.parent.children.remove(v)
                else:
                    module_nodes[v.get_fully_specified_name()] = v

    def get_fully_specified_type_name(self, type_name, language):
        fully_specified_type_name = TestConfiguration.Language.type_name_dict[language].get(type_name, None)
        if fully_specified_type_name is not None:
            return fully_specified_type_name # it is a language specific type

        for t in self:
            if t.name == type_name:
                return t.get_fully_specified_name(language)

        return None


class Node:

    def __init__(self, **kw):
        self.name = kw.get('name', None)
        self.parent = kw.get('parent', None)
        self.children = kw.get('children', list())
        self.loc = kw.get('loc', None)
        self.in_namespace = True
        self.conformance_test_node = False
        self.gold_lib_definition_node = False
        self.gold_lib_source_node = False

    def __iter__(self):
        for v in itertools.chain(*itertools.imap(iter, self.children)):
            yield v
        yield self

    def get_node_description(self, delimiter=':'):
        if self.parent is not None:
            return self.parent.get_node_description() + delimiter + self.name + ' (' + self.__class__.__name__ + ')'
        else:
            return self.name

    def get_fully_specified_name(self, language=TestConfiguration.Language.C):
        parent_node = self.parent
        parent_name = ''

        while parent_node is not None:
            if parent_node.in_namespace:
                parent_name = parent_node.get_fully_specified_name(language)
                break
            else:
                parent_node = parent_node.parent

        if len(parent_name) > 0:
            parent_name += TestConfiguration.Language.delimiter[language]
            return  parent_name + TestConfiguration.Language.format_case(language, self.name)
        else:
            return TestConfiguration.Language.format_case(language, self.name)

    def _find_node_name(self, name):
        """ Look for node with name at this node, then at the node's children, and finally recursively at it's parent
        """
        if self.name == name:
            return self
        for c in self.children:
            if c.name == name:
                return c
        if self.parent is not None:
            return self.parent._find_node_name(name)
        else:
            return None

    def get_fully_specified_type_name(self, language):
        if 'type_name' in vars(self).keys():
            return_value = TestConfiguration.Language.type_name_dict[language].get(self.type_name, None)
            if return_value is None and self.parent is not None:
                # TODO: This should be looked up in the idlTree to get the fully specified typename
                type_name_node = self.parent._find_node_name(self.type_name)
                if type_name_node is not None:
                    return_value = type_name_node.get_fully_specified_name(language)
                else:
                    # Couldn't find it in the all the nodes...so my guess is just use the type name
                    return_value = self.type_name
            return return_value
        else:
            return None

    def _print_node_tree(self, indent=0, indent_text = '-'):
        print (indent * indent_text) + self.toString()
        indent += 1
        for c in self.children:
            c._print_node_tree(indent, indent_text)

    def toString(self):
        return_value = self.__class__.__name__ + ' '
        for variable, value in sorted(vars(self).iteritems()):
            if variable == 'children' or variable == 'parent':
                continue
            return_value += '[' + variable + ': ' + str(value) + ']'
        return return_value

    def get_conformance_test_code(self, template, language):
        warning_message = self.__class__.__name__ + ' conformance test code not implemented yet: ' + self.name
        logger.warning(warning_message)
        return None

    def get_specific_node_type_parents(self, node_types):
        parent = self.parent
        specific_parents = list()

        while parent is not None:
            for n in node_types:
                if isinstance(parent, n):
                    specific_parents.append(parent)
                    break
            parent = parent.parent

        return reversed(specific_parents)


class ModuleNode(Node):
    def __init__(self, loc, tokens, **kw):
        Node.__init__(self, loc=loc, **kw)

        if len(tokens) < 2 or tokens[0] != 'module':
            raise RuntimeError("Unexpected tokens, expected ['module','MODULE_NAME', ...]")

        self.name = tokens[1]


class EnumerationNode(Node):

    def __init__(self, loc, tokens, **kw):
        Node.__init__(self, loc=loc, **kw)
        if len(tokens) < 2 or tokens[0] != 'enum':
            raise RuntimeError("Unexpected tokens, expected ['enum','ENUM_NAME', ...]")
        self.name = tokens[1]
        self.conformance_test_node = True
        self.in_namespace = False

    def get_conformance_test_code(self, template_group, language):
        section1_template = template_group.getInstanceOf('conformanceTestCodeSection1Enumeration')

        section1_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)

        enumeration_item_code_list = list()
        for enumeration_item_node in self.children:

            enumeration_item_code = enumeration_item_node.get_conformance_test_code(language, template_group)
            if enumeration_item_code is not None:
                enumeration_item_code_list.append(enumeration_item_code)

        section2_template = template_group.getInstanceOf('conformanceTestCodeSection2Enumeration')
        section2_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)
        section2_template['ENUMERATED_ITEM_CODE'] = '\n'.join(enumeration_item_code_list)

        return [str(section1_template), str(section2_template)]


class EnumerationItemNode(Node):
    def __init__(self, loc, tokens, **kw):
        Node.__init__(self, loc=loc, **kw)
        if len(tokens) < 1:
            raise RuntimeError("Unexpected tokens, expected ['ENUM_ITEM_NAME', ...]")

        self.name = tokens[0]

    def get_conformance_test_code(self, template_group, language):
        section1_template = template_group.getInstanceOf('conformanceTestCodeEnumeratedItem')

        section1_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)

        return str(section1_template)


class TypedefNode(Node):
    def __init__(self, loc, tokens, **kw):
        Node.__init__(self, loc=loc, **kw)

        if len(tokens) != 1 or len(tokens[0]) < 4 or tokens[0][0] != 'typedef' or tokens[0][-1] != ';':
            raise RuntimeError("Unexpected tokens, expected [['typedef','TYPE_NAME', 'NAME', ';']] got " + str(tokens))

        self.type_name = ' '.join(tokens[0][1:-2])
        self.name = tokens[0][-2]
        self.conformance_test_node = True  # Need to generate a conformance test for this type of node

    def get_conformance_test_code(self, template_group, language):
        logger.info(locals())

        section1_template = template_group.getInstanceOf('conformanceTestCodeSection1Typedef')

        section1_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)
        section1_template['TYPE_NAME'] = self.get_fully_specified_type_name(language)
        if self.type_name in TestConfiguration.Language.type_name_dict[language]:
            section1_template['CHECK_FIRST_LAST'] = True

        section2_template = template_group.getInstanceOf('conformanceTestCodeSection2Typedef')
        section2_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)
        section2_template['TYPE_NAME'] = self.get_fully_specified_type_name(language)



        return [str(section1_template), str(section2_template)]


class TypedefStringSequenceNode(Node):
    def __init__(self, loc, tokens, **kw):
        Node.__init__(self, loc=loc, **kw)

        if len(tokens) != 1 or len(tokens[0]) < 6 or tokens[0][0] != 'typedef' or tokens[0][1] != 'string':
            raise RuntimeError("Unexpected tokens, expected [['typedef','string', '<' 'SIZE', '>', 'NAME', ';']] " +
                               "got " + str(tokens))

        self.size = tokens[0][3]
        self.name = tokens[0][5]
        self.type_name = 'string' # TODO make this right regardless of language
        self.conformance_test_node = True  # Need to generate a conformance test for this type of node

    def get_conformance_test_code(self, template_group, language):
        logger.info(locals())

        section1_template = template_group.getInstanceOf('conformanceTestCodeSection1TypedefStringSequence')

        section1_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)
        section1_template['TYPE_NAME'] = self.get_fully_specified_type_name(language)

        section2_template = template_group.getInstanceOf('conformanceTestCodeSection2TypedefStringSequence')
        section2_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)
        section2_template['TYPE_NAME'] = self.get_fully_specified_type_name(language)

        return [str(section1_template), str(section2_template)]


class TypedefArrayNode(Node):
    def __init__(self, loc, tokens, **kw):
        Node.__init__(self, loc=loc, **kw)

        if len(tokens) != 1 or len(tokens[0]) < 7 or tokens[0][0] != 'typedef':
            raise RuntimeError("Unexpected tokens, expected [['typedef','TYPE_NAME', 'NAME', '[' 'SIZE', ']',  ';']]" +
                               "got " + str(tokens))

        self.size = tokens[0][4]
        self.name = tokens[0][2]
        self.type_name = tokens[0][1]
        self.conformance_test_node = True  # Need to generate a conformance test for this type of node

    def get_conformance_test_code(self, template_group, language):
        section1_template = template_group.getInstanceOf('conformanceTestCodeSection1TypedefArray')
        section1_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)
        section1_template['TYPE_NAME'] = self.get_fully_specified_type_name(language)
        section1_template['SIZE'] = self.size

        section2_template = template_group.getInstanceOf('conformanceTestCodeSection2TypedefArray')
        section2_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)
        section2_template['TYPE_NAME'] = self.get_fully_specified_type_name(language)
        section2_template['SIZE'] = self.size
        return [str(section1_template), str(section2_template)]


class ConstantNode(Node):
    def __init__(self, loc, tokens, **kw):
        Node.__init__(self, loc=loc, **kw)

        if len(tokens) != 1 or len(tokens[0]) < 6 or tokens[0][0] != 'const' or tokens[0][5] != ';':
            raise RuntimeError("Unexpected tokens, expected [['const', TYPE_NAME', 'NAME', '=', 'VALUE', ';']]")

        self.type_name = tokens[0][1]
        self.name = tokens[0][2]
        self.value = tokens[0][4]
        self.conformance_test_node = True  # Need to generate a conformance test for this type of node

    def get_conformance_test_code(self, template_group, language):
        section1_template = template_group.getInstanceOf('conformanceTestCodeSection1Constant')

        constants_without_size = ['Fault_Message_Maximum_Size']

        section1_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)
        section1_template['TYPE_NAME'] = self.get_fully_specified_type_name(language)
        section1_template['VALUE'] = self.value
        if language == TestConfiguration.Language.Ada:
            interface_nodes = self.get_specific_node_type_parents([InterfaceNode])
            for i in interface_nodes:
                if i.name.lower() != 'hmfm':
                    section1_template['CHECK_SIZE'] = True

        section2_template = template_group.getInstanceOf('conformanceTestCodeSection2Constant')
        section2_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)
        section2_template['TYPE_NAME'] = self.get_fully_specified_type_name(language)
        section2_template['VALUE'] = self.value

        return [str(section1_template), str(section2_template)]


class StructNode(Node):
    def __init__(self, loc, tokens, **kw):
        Node.__init__(self, loc=loc, **kw)
        if len(tokens) != 1 or len(tokens[0]) < 2 or tokens[0][0] != 'struct':
            raise RuntimeError("Unexpected tokens, expected [['struct', 'STRUCT_NAME',...]]")

        self.name = tokens[0][1]
        self.conformance_test_node = True  # Need to generate a conformance test for this type of node

    def get_conformance_test_code(self, template_group, language):
        section1_template = template_group.getInstanceOf('conformanceTestCodeSection1Struct')

        section1_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)

        struct_item_code_dict = {0: list(), 1: list()}

        for struct_item_node in self.children:

            struct_item_code = struct_item_node.get_conformance_test_code(language, template_group)
            if struct_item_code is not None:
                struct_item_code_dict[0].append(struct_item_code[0])
                struct_item_code_dict[1].append(struct_item_code[1])

        section1_template['SECTION1_STRUCT_ITEM_CODE'] = '\n'.join(struct_item_code_dict[0])

        section2_template = template_group.getInstanceOf('conformanceTestCodeSection2Struct')
        section2_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)
        section2_template['SECTION2_STRUCT_ITEM_CODE'] = '\n'.join(struct_item_code_dict[1])

        return [str(section1_template), str(section2_template)]


class StructItemNode(Node):
    def __init__(self, loc, tokens, **kw):
        Node.__init__(self, loc=loc, **kw)
        if len(tokens) < 2:
            raise RuntimeError("Unexpected tokens, expected ['VARIABLE_TYPE', 'VARIABLE_NAME',...]")

        self.name = tokens[1]
        self.type_name = tokens[0]

    def get_conformance_test_code(self, template_group, language):
        section1_template = template_group.getInstanceOf('conformanceTestCodeSection1StructItem')
        section1_template['NAME'] = TestConfiguration.Language.format_case(language, self.name)
        section1_template['TYPE_NAME'] = self.get_fully_specified_type_name(language)


        section2_template = template_group.getInstanceOf('conformanceTestCodeSection2StructItem')
        section2_template['NAME'] = TestConfiguration.Language.format_case(language, self.name)
        section2_template['TYPE_NAME'] = self.get_fully_specified_type_name(language)

        return [str(section1_template), str(section2_template)]


class InterfaceNode(Node):
    def __init__(self, loc, tokens, **kw):
        Node.__init__(self, loc=loc, **kw)
        if len(tokens) != 1 or tokens[0][0] != 'interface':
            raise RuntimeError("Unexpected tokens, expected [['interface', 'INTERFACE_NAME',...]]")
        self.name = tokens[0][1]


class OperationNode(Node):
    def __init__(self, loc, tokens, **kw):
        Node.__init__(self, loc=loc, **kw)
        if len(tokens) < 2:
            raise RuntimeError("Unexpected tokens, expected [['OPERATION_RETURN_TYPE', 'OPERATION_NAME',...]]")
        self.name = tokens[1]
        self.return_type = tokens[0]
        self.conformance_test_node = True  # Need to generate a conformance test for this type of node

    def get_conformance_test_code(self, template_group, language):
        section1_template = template_group.getInstanceOf('conformanceTestCodeSection1Operation')
        section2_template = template_group.getInstanceOf('conformanceTestCodeSection2Operation')

        section1_template['NAME'] = self.name
        section1_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)
        if self.return_type is not None and len(self.return_type) > 0:
            section1_template['RETURN_TYPE'] = self.return_type
            section2_template['RETURN_TYPE'] = self.return_type

        section2_template['NAME'] = self.name
        section2_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)

        parameter_item_code_dict = {0: list(), 1: list(), 2: list(), 3: list()}

        last_parameter = None
        for child in reversed(self.children):
            if not isinstance(child, ParameterNode):
                continue
            last_parameter = child
            break

        for child in self.children:
            if not isinstance(child, ParameterNode):
                continue
            last_element = child == last_parameter
            parameter_item_code = child.get_conformance_test_code(template_group, language, last_element)
            if parameter_item_code is not None:
                parameter_item_code_dict[0].append(parameter_item_code[0])
                parameter_item_code_dict[1].append(parameter_item_code[1])
                parameter_item_code_dict[2].append(parameter_item_code[2])
                if language == TestConfiguration.Language.Ada:
                    parameter_item_code_dict[3].append(parameter_item_code[3])

        section1_template['SECTION1_PARAMETER_ITEM_CODE'] = '\n'.join(parameter_item_code_dict[0])
        section1_template['SECTION1A_PARAMETER_ITEM_CODE'] = '\n'.join(parameter_item_code_dict[1])
        section2_template['SECTION2_PARAMETER_ITEM_CODE'] = '\n'.join(parameter_item_code_dict[2])
        if language == TestConfiguration.Language.Ada:
            section2_template['SECTION2_PARAMETER_INIT_CODE'] = '\n'.join(parameter_item_code_dict[3])

        return [str(section1_template), str(section2_template)]

    def generate_gold_lib_definition(self, template_group, language):
        logger.info(locals())

        operation_signature_template = template_group.getInstanceOf('operationSignature')
        operation_definition_template = template_group.getInstanceOf('operationDefinition')

        operation_signature_template['OPERATION_NAME'] = TestConfiguration.Language.format_case(language, self.name)

        last_parameter = None
        for child in reversed(self.children):
            if not isinstance(child, ParameterNode):
                continue
            last_parameter = child
            break

        for child in self.children:
            if not isinstance(child, ParameterNode):
                continue
            last_element = child == last_parameter
            parameter_code = child.generate_gold_lib_definition(template_group, language, last_element)
            if child is not None:
                operation_signature_template['PARAMETER_SIGNATURES'] = parameter_code

        operation_definition_template['OPERATION_SIGNATURE'] = str(operation_signature_template)

        return str(operation_definition_template)

    def generate_gold_lib_body(self, template_group, language):
        logger.info(locals())

        operation_signature_template = template_group.getInstanceOf('operationSignature')
        operation_body_template = template_group.getInstanceOf('operationBody')

        operation_signature_template['OPERATION_NAME'] = TestConfiguration.Language.format_case(language, self.name)
        operation_body_template['OPERATION_NAME'] = TestConfiguration.Language.format_case(language, self.name)

        for parameter_node in self.children:
            last_element = parameter_node == self.children[-1]
            parameter_code = parameter_node.generate_gold_lib_definition(template_group, language, last_element)
            if parameter_code is not None:
                operation_signature_template['PARAMETER_SIGNATURES'] = parameter_code


        operation_body_template['OPERATION_SIGNATURE'] = str(operation_signature_template)

        return str(operation_body_template)



class ParameterNode(Node):
    def __init__(self, loc, tokens, **kw):
        Node.__init__(self, loc=loc, **kw)

        if len(tokens) != 1 or len(tokens[0]) < 3 or \
                (tokens[0][0] != 'in' and tokens[0][0] != 'out' and tokens[0][0] != 'inout'):
            raise RuntimeError("Unexpected tokens, expected [['in/out', 'PARAM_TYPE', 'PARAM_NAME', ...]], got " + str(tokens))

        self.direction = tokens[0][0]
        self.type_name = tokens[0][1]
        self.name = tokens[0][2]

    def get_conformance_test_code(self, template_group, language, last_element):
        section1_template = template_group.getInstanceOf('conformanceTestCodeSection1ParameterItem')
        section1_template['NAME'] = TestConfiguration.Language.format_case(language, self.name)
        section1_template['TYPE_NAME'] = self.get_fully_specified_type_name(language)
        section1_template['DIRECTION'] = TestConfiguration.Language.direction_keyword_dict[language][self.direction]
        if last_element:
            section1_template['LAST_ELEMENT'] = True

        section1a_template = template_group.getInstanceOf('conformanceTestCodeSection1AParameterItem')
        section1a_template['NAME'] = TestConfiguration.Language.format_case(language, self.name)
        section1a_template['TYPE_NAME'] = self.get_fully_specified_type_name(language)
        section1a_template['DIRECTION'] = TestConfiguration.Language.direction_keyword_dict[language][self.direction]
        if last_element:
            section1a_template['LAST_ELEMENT'] = True

        section2_template = template_group.getInstanceOf('conformanceTestCodeSection2ParameterItem')
        section2_template['NAME'] = TestConfiguration.Language.format_case(language, self.name)
        section2_template['TYPE_NAME'] = self.get_fully_specified_type_name(language)
        section2_template['DIRECTION'] = TestConfiguration.Language.direction_keyword_dict[language][self.direction]
        if last_element:
            section2_template['LAST_ELEMENT'] = True

        section2a_template = template_group.getInstanceOf('conformanceTestCodeSection2AParameterItem')
        section2a_template['NAME'] = TestConfiguration.Language.format_case(language, self.name)
        section2a_template['TYPE_NAME'] = self.get_fully_specified_type_name(language)
        section2a_template['DIRECTION'] = TestConfiguration.Language.direction_keyword_dict[language][self.direction]
        if last_element:
            section2a_template['LAST_ELEMENT'] = True
        
        return [str(section1_template), str(section1a_template), str(section2_template), str(section2a_template)]

    def generate_gold_lib_definition(self, template_group, language, last_element):
        parameter_signature_template = template_group.getInstanceOf('parameterSignature')
        parameter_signature_template['VARIABLE_NAME'] = \
            TestConfiguration.Language.format_case(language, self.name)
        parameter_signature_template['TYPE_NAME'] = self.get_fully_specified_type_name(language)
        parameter_signature_template['DIRECTION'] = TestConfiguration.Language.direction_keyword_dict[language][self.direction]
        if last_element:
            parameter_signature_template['LAST_ELEMENT'] = True

        return str(parameter_signature_template)


class CommentNode(Node):
    def __init__(self, loc, **kw):
        Node.__init__(self, loc=loc, **kw)
        self.name = "Comment"

    def get_conformance_test_code(self, template_group, language):
        return None # Comments do not need conformance tests


class CStyleCommentNode(CommentNode):
    def __init__(self, loc, tokens, **kw):
        CommentNode.__init__(self, loc=loc, **kw)

        if len(tokens) != 1 or not tokens[0].startswith('/*') or not tokens[0].endswith('*/'):
            raise RuntimeError("Unexpected tokens, expected ['/* COMMENT */'], got " + str(tokens))

        self.comment = tokens[0].strip('/*')


class SingleLineCommentNode(CommentNode):
    def __init__(self, loc, tokens, **kw):
        CommentNode.__init__(self, loc=loc, **kw)

        if len(tokens) != 2 or not tokens[0].startswith('//'):
            raise RuntimeError("Unexpected tokens, expected ['//', 'COMMENT...'], got " + str(tokens))

        self.comment = tokens[1].strip()


class FaceIdlParser():

    def __init__(self, **kw):
        # class variables
        self.bnf = None
        self.unhandled_definitions = 0
        self.idl_tree = IdlTree()
        self.tokens = None

        self.initialize_parser()

        idl_filename = kw.get('filename', None)

        if idl_filename is not None:
            self.parse_idl_file(idl_filename)

    def parse_idl_file(self, idl_filename):
        if not os.path.isfile(idl_filename):
            raise IOError('Idl File does not exist: ' + idl_filename)

        self.idl_directory = os.path.dirname(idl_filename)

        idl_file = open(idl_filename, 'r')
        lines = '\n'.join(idl_file.readlines())
        self.parse_string(lines)
        self.idl_tree.combine_modules()

    def initialize_parser(self):

        # punctuation
        colon  = Literal(":")
        lbrace = Literal("{")
        rbrace = Literal("}")
        lbrack = Literal("[")
        rbrack = Literal("]")
        lparen = Literal("(")
        rparen = Literal(")")
        equals = Literal("=")
        comma  = Literal(",")
        dot    = Literal(".")
        hashtag= Literal("#")
        slash  = Literal("/")
        bslash = Literal("\\")
        star   = Literal("*")
        semi   = Literal(";")
        langle = Literal("<")
        rangle = Literal(">")
        quote  = Literal('"')

        # keywords
        any_       = Keyword("any")
        attribute_ = Keyword("attribute")
        boolean_   = Keyword("boolean")
        case_      = Keyword("case")
        char_      = Keyword("char")
        const_     = Keyword("const")
        context_   = Keyword("context")
        default_   = Keyword("default")
        double_    = Keyword("double")
        enum_      = Keyword("enum")
        exception_ = Keyword("exception")
        false_     = Keyword("FALSE")
        fixed_     = Keyword("fixed")
        float_     = Keyword("float")
        inout_     = Keyword("inout")
        interface_ = Keyword("interface")
        in_        = Keyword("in")
        include_   = Keyword("include")
        long_      = Keyword("long")
        long_long_ = Keyword("long long")
        module_    = Keyword("module")
        object_    = Keyword("Object")
        octet_     = Keyword("octet")
        oneway_    = Keyword("oneway")
        out_       = Keyword("out")
        raises_    = Keyword("raises")
        readonly_  = Keyword("readonly")
        sequence_  = Keyword("sequence")
        short_     = Keyword("short")
        string_    = Keyword("string")
        struct_    = Keyword("struct")
        switch_    = Keyword("switch")
        true_      = Keyword("TRUE")
        typedef_   = Keyword("typedef")
        unsigned_  = Keyword("unsigned")
        union_     = Keyword("union")
        void_      = Keyword("void")
        wchar_     = Keyword("wchar")
        wstring_   = Keyword("wstring")

        identifier = Word( alphas, alphanums + "_" ).setName("identifier")

        real = Combine( Word(nums+"+-", nums) + dot + Optional( Word(nums) )
                        + Optional( CaselessLiteral("E") + Word(nums+"+-",nums) ) )
        integer = ( Combine( CaselessLiteral("0x") + Word( nums+"abcdefABCDEF" ) ) |
                    Word( nums+"+-", nums ) ).setName("int")

        # include definitions
        includeDef = ( Group( hashtag + include_ + langle + identifier + dot + 'idl' + rangle) |
                       Group( hashtag + include_ + quote + identifier + dot + 'idl' + quote) )
        includeDef.setParseAction(self.include_handler)

        # Comments (either single line or cStyle comments
        singleLineComment = "//" + restOfLine
        singleLineComment.setParseAction(self.single_line_comment_handler)
        cStyleComment.setParseAction(self.cStyle_comment_handler)

        anyComment = (singleLineComment | cStyleComment)
        optionalComment = ZeroOrMore(anyComment)

        # Typenames
        udTypeName = delimitedList( identifier, "::", combine=True ).setName("udType")
        # have to use longest match for type, in case a user-defined type name starts with a keyword type, like "stringSeq" or "longArray"
        typeName = ( any_ ^ boolean_ ^ char_ ^ double_ ^ fixed_ ^
                    float_ ^ Optional(unsigned_) + long_long_ ^ Optional(unsigned_) + long_ ^ octet_ ^
                     Optional(unsigned_) + short_ ^ string_ ^
                    wchar_ ^ wstring_ ^ udTypeName ).setName("type")

        # Sequences
        sequenceDef = Forward().setName("seq")
        sequenceDef << Group( sequence_ + langle + ( sequenceDef | typeName ) + rangle )
        sequenceDef.setParseAction(self.sequence_handler)


        # Typedefs
        typeDef = sequenceDef | ( typeName + Optional( lbrack + integer + rbrack ) )
        typedefDef = Group( typedef_ + typeDef + identifier + semi ).setName("typedef")
        typedefDef.setParseAction(self.typedef_handler)

        typedefArrayDef = Group( typedef_ + typeName + identifier + lbrack + integer + rbrack + semi)
        typedefArrayDef.setParseAction(self.typedef_array_handler)

        typedefStringSequenceDef = Group( typedef_ + string_ + langle + (integer | identifier) + rangle + identifier +
                                          semi)
        typedefStringSequenceDef.setParseAction(self.typedef_string_sequence_handler)

        oneOftheTypedefDefs = (typedefDef | typedefArrayDef | typedefStringSequenceDef) + optionalComment

        # Enumerations
        enumDef = Forward()
        enumItem = identifier + comma + optionalComment
        enumItem.setParseAction(self.enumeration_item_handler)
        lastEnumItem = identifier + optionalComment
        lastEnumItem.setParseAction(self.enumeration_item_handler)
        enumDef << enum_ + identifier + lbrace + optionalComment + ZeroOrMore(enumItem) + lastEnumItem + rbrace + \
                   semi + optionalComment
        enumDef.setParseAction(self.enumeration_handler)

        # Constants
        constDef = Group( const_ + typeDef + identifier + equals + ( real | integer | quotedString ) + semi ) #| quotedString )
        constDef.setParseAction(self.constant_handler)

        # Exceptions
        exceptionItem = Group( typeDef + identifier + semi )
        exceptionDef = ( exception_ + identifier + lbrace + ZeroOrMore( exceptionItem ) + rbrace + semi )
        exceptionDef.setParseAction(self.exception_handler)

        # Attributes
        attributeDef = Optional( readonly_ ) + attribute_ + typeDef + identifier + semi
        attributeDef.setParseAction(self.attribute_handler)

        # Operations
        parameter = Group( ( inout_ | in_ | out_ ) + typeName + identifier )
        parameter.setParseAction(self.parameter_handler)
        paramlist = delimitedList( parameter ).setName( "paramlist" )

        operationDef = ( ( void_ ^ typeDef ) + identifier + lparen + Optional( paramlist ) + rparen + \
                        Optional( raises_ + lparen + Group( delimitedList( typeName ) ) + rparen ) + semi )
        operationDef.setParseAction(self.operation_handler)

        # Structures
        struct_item = typeName + identifier + semi + optionalComment
        struct_item.setParseAction(self.struct_item_handler)
        structDef = Group( struct_ + identifier + lbrace + optionalComment + OneOrMore(struct_item) + rbrace + semi)
        structDef.setParseAction(self.struct_handler)

        # Modules and Interfaces
        moduleDef = Forward()
        interfaceDef = Forward()
        interfaceItem = (interfaceDef | exceptionDef | constDef | oneOftheTypedefDefs | moduleDef | enumDef |
                         structDef | operationDef | anyComment)
        interfaceDef = Group( interface_ + identifier  + Optional( colon + delimitedList( typeName ) ) + lbrace +
                        ZeroOrMore( interfaceItem ) + rbrace + semi )
        interfaceDef.setParseAction(self.interface_handler)

        moduleItem = (interfaceDef | exceptionDef | constDef | oneOftheTypedefDefs | moduleDef | enumDef |
                      structDef | anyComment)
        moduleDef << module_ + identifier + lbrace + ZeroOrMore( moduleItem ) + rbrace + semi
        moduleDef.setParseAction(self.module_handler)

        # Full definition
        self.bnf = ZeroOrMore(includeDef) + ( moduleDef | OneOrMore( moduleItem ) )
        #self.bnf.ignore( singleLineComment )
        #self.bnf.ignore( cStyleComment )

    def parse_string(self, idlString):

        parseResult = self.bnf.parseString(idlString)

        if self.tokens is not None:
            self.tokens += parseResult
        else:
            self.tokens = parseResult

        print 'Unhandled definitions: ' + str(self.unhandled_definitions)

    ######### HANDLERS FOR ALL THE TOKENS ###############
    def include_handler(self, s, loc, tokens):
        logger.info('tokens: ' + str(tokens))

        include_file = ''.join(tokens[0][3:6])
        logger.debug('Include file: "' + include_file + '"')

        temp_idl_dir = self.idl_directory

        self.parse_idl_file(os.path.join(self.idl_directory, include_file))
        self.idl_directory = temp_idl_dir

    def module_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.idl_tree.add_parent(ModuleNode(loc, tokens))

    def interface_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.idl_tree.add_parent(InterfaceNode(loc, tokens))

    def typedef_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.idl_tree.add_sibling(TypedefNode(loc, tokens))

    def typedef_string_sequence_handler(self, s, loc, tokens):
        logger.info('tokens: ' + str(tokens))
        self.idl_tree.add_sibling(TypedefStringSequenceNode(loc, tokens))

    def typedef_array_handler(self, s, loc, tokens):
        logger.info('tokens: ' + str(tokens))
        self.idl_tree.add_sibling(TypedefArrayNode(loc, tokens))

    def constant_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.idl_tree.add_sibling(ConstantNode(loc, tokens))

    def enumeration_item_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.idl_tree.add_sibling(EnumerationItemNode(loc, tokens))

    def enumeration_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.idl_tree.add_parent(EnumerationNode(loc, tokens))

    def struct_item_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.idl_tree.add_sibling(StructItemNode(loc, tokens))

    def struct_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.idl_tree.add_parent(StructNode(loc, tokens))

    def param_list_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        #self.idl_tree.add_sibling(ParamListNode(loc, tokens))

    def parameter_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.idl_tree.add_sibling(ParameterNode(loc, tokens))

    def operation_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.idl_tree.add_parent(OperationNode(loc, tokens))

    def cStyle_comment_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.idl_tree.add_sibling(CStyleCommentNode(loc, tokens))

    def single_line_comment_handler(self, s, loc, tokens):
        logger.info('tokens: ' + str(tokens))
        self.idl_tree.add_sibling(SingleLineCommentNode(loc, tokens))


    # Not implemented yet
    def sequence_handler(self, s, loc, tokens):
        logger.info('tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def typedef_sequence_handler(self, s, loc, tokens):
        logger.info('tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def exception_handler(self, s, loc, tokens):
        logger.info('tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def attribute_handler(self, s, loc, tokens):
        logger.info('tokens: ' + str(tokens))
        self.unhandled_definitions += 1


class ConformanceTestGenerator():
    def __init__(self, idl_tree):
        self.idl_tree = idl_tree

    def generate_conformance_tests(self, conformance_gen_dir, template_file, language):
        logger.info(locals())
        if not os.path.isdir(conformance_gen_dir):
            raise IOError('Conformance Gen Directory does not exist: ' + conformance_gen_dir)

        if not os.path.isfile(template_file):
            raise IOError('Template file does not exist: ' + template_file)

        if language not in TestConfiguration.Language.allLanguages:
            raise RuntimeError('Language ' + str(language) + ' not in ' +
                               str(TestConfiguration.Language.allLanguages))

        old_source_files = glob.glob(os.path.join(conformance_gen_dir,
                                           '*' + TestConfiguration.Language.source_suffix[language]))
        for o in old_source_files:
            os.remove(o)

        template = stringtemplate3.StringTemplateGroup(name='conformance', file=open(template_file,'r'))

        test_num = 1

        assertion_file = open(os.path.join(conformance_gen_dir, 'assertions.txt'), 'w')
        all_tests_file = open(os.path.join(conformance_gen_dir,
                                           'all_conformance_tests' + TestConfiguration.Language.source_suffix[language]), 'w')
        for current_node in iter(self.idl_tree):
            if not current_node.conformance_test_node:
                continue

            self.generate_node_conformance_test(current_node, language, test_num, template, conformance_gen_dir,
                                                assertion_file, all_tests_file)

            test_num += 1

        assertion_file.close()
        all_tests_file.close()

    def generate_gold_library_files(
            self, language, gold_definition_filename, gold_source_filename, template_file, **kw):

        logger.info(locals())
        package_name = kw.get('package_name', 'Unset_Package_Name')

        definition_dir = os.path.dirname(gold_definition_filename)
        if not os.path.isdir(definition_dir):
            raise IOError('gold library include/spec directory does not exist: ' + definition_dir)

        source_dir = os.path.dirname(gold_source_filename)
        if not os.path.isdir(source_dir):
            raise IOError('gold library source directory does not exist: ' + source_dir)

        if not os.path.isfile(template_file):
            raise IOError('Template file does not exist: ' + template_file)

        if language not in TestConfiguration.Language.allLanguages:
            raise RuntimeError('Language ' + str(language) + ' not in ' +
                               str(TestConfiguration. Language.allLanguages))



        template = stringtemplate3.StringTemplateGroup(name='gold_lib', file=open(template_file, 'r'))

        gold_lib_definition_file = template.getInstanceOf('goldLibDefinitionFile')
        gold_lib_source_file = template.getInstanceOf('goldLibSourceFile')

        if language is TestConfiguration.Language.Ada:
            old_ada_files = glob.glob(os.path.join(definition_dir, '*.ads'))
            for o in old_ada_files:
                os.remove(o)

            old_ada_files = glob.glob(os.path.join(source_dir, '*.adb'))
            for o in old_ada_files:
                os.remove(o)


            for current_node in iter(self.idl_tree):
                if not isinstance(current_node, OperationNode):
                    continue

                test_code = current_node.generate_gold_lib_definition(template, language)
                logger.debug(str(test_code))
                gold_lib_definition_file['DEFINITIONS'] = str(test_code)
                test_code = current_node.generate_gold_lib_body(template, language)
                gold_lib_source_file['BODIES'] = str(test_code)

        gold_lib_source_file['PACKAGE_NAME'] = package_name
        logger.debug(str(gold_lib_definition_file))
        logger.debug(str(gold_lib_source_file))

        gold_spec_file = open(gold_definition_filename, 'w')
        gold_spec_file.write(set_distro_statement.get_distribution_statement(language,
                                                                        os.path.basename(gold_definition_filename)))
        gold_spec_file.write(str(gold_lib_definition_file))
        gold_spec_file.close()

        gold_body_file = open(gold_source_filename, 'w')
        gold_body_file.write(set_distro_statement.get_distribution_statement(language,
                                                                        os.path.basename(gold_source_filename)))
        gold_body_file.write(str(gold_lib_source_file))
        gold_body_file.close()


    def generate_face_specified_gold_library(
            self, data_type_list, language, spec_dir, src_dir, template_file):

        logger.info(locals())

        if not type(data_type_list) != type(list):
            raise RuntimeError('datatype list needs to be a list')

        if not os.path.isdir(spec_dir):
            raise IOError('Conformance Gen Directory does not exist: ' + spec_dir)

        if not os.path.isdir(src_dir):
            raise IOError('Conformance Gen Directory does not exist: ' + src_dir)

        if not os.path.isfile(template_file):
            raise IOError('Template file does not exist: ' + template_file)

        if language not in TestConfiguration.Language.allLanguages:
            raise RuntimeError('Language ' + str(language) + ' not in ' +
                               str(TestConfiguration.Language.allLanguages))

        template = stringtemplate3.StringTemplateGroup(name='gold_lib', file=open(template_file, 'r'))

        gold_lib_definition_file = template.getInstanceOf('faceSpecifiedTypeDefinitionFile')
        gold_lib_source_file = template.getInstanceOf('faceSpecifiedTypeSourceFile')

        if language == TestConfiguration.Language.Ada:
            gold_tss_spec_filename = os.path.join(spec_dir, constants.ADA_TSS_FILENAME
                                                            + TestConfiguration.Language.header_suffix[language])
            base_tss_spec_filename = gold_tss_spec_filename + constants.ADA_BASE_EXTENSION
            gold_tss_body_filename = os.path.join(src_dir, constants.ADA_TSS_FILENAME
                                                            + TestConfiguration.Language.source_suffix[language])
            base_tss_body_filename = gold_tss_body_filename + constants.ADA_BASE_EXTENSION

            for b in [base_tss_body_filename, base_tss_spec_filename]:
                if not os.path.isfile(b):
                    err_msg = "Ada base TSS file does not exist: " + b
                    logger.error(err_msg)
                    raise RuntimeError(err_msg)
            commonTools.copyFile(base_tss_spec_filename, gold_tss_spec_filename)
            commonTools.copyFile(base_tss_body_filename, gold_tss_body_filename)
            read_callback_sub_package_template = template.getInstanceOf("readCallbackSubPackage")

            for data_type in data_type_list:
                self.idl_tree.overwrite_face_specified_type_nodes(data_type)

                send_event_template = template.getInstanceOf("sendEventAccessSignature")
                send_event_template['MESSAGE_TYPE'] = data_type
                send_event_template['FACE_SPECIFIED_TYPE'] = 'face.dm.' + data_type
                read_callback_sub_package_template['SEND_EVENT_ACCESS_SIGNATURE'] = str(send_event_template)

            gold_lib_definition_file['DEFINITIONS'] = str(read_callback_sub_package_template)

            for data_type in data_type_list:
                self.idl_tree.overwrite_face_specified_type_nodes(data_type)

                for current_node in iter(self.idl_tree):
                    if not isinstance(current_node, OperationNode):
                        continue

                    test_code = current_node.generate_gold_lib_definition(template, language)
                    logger.debug(str(test_code))
                    gold_lib_definition_file['DEFINITIONS'] = str(test_code)
                    test_code = current_node.generate_gold_lib_body(template, language)
                    gold_lib_source_file['BODIES'] = str(test_code)


            logger.debug(str(gold_lib_definition_file))
            logger.debug(str(gold_lib_source_file))


            gold_spec_file = open(gold_tss_spec_filename, 'a')
            gold_spec_file.write(str(gold_lib_definition_file))
            gold_spec_file.close()

            gold_body_file = open(gold_tss_body_filename, 'a')
            gold_body_file.write(str(gold_lib_source_file))
            gold_body_file.close()

    def generate_face_specified_conformance_tests(
            self, data_type_list, language, conformance_gen_dir, template_file):

        logger.info(locals())

        if not type(data_type_list) != type(list):
            raise RuntimeError('datatype list needs to be a list')

        if not os.path.isdir(conformance_gen_dir):
            raise IOError('Conformance Gen Directory does not exist: ' + conformance_gen_dir)

        if not os.path.isfile(template_file):
            raise IOError('Template file does not exist: ' + template_file)

        if language not in TestConfiguration.Language.allLanguages:
            raise RuntimeError('Language ' + str(language) + ' not in ' +
                               str(TestConfiguration.Language.allLanguages))

        face_prefix = 'Test_'
        old_source_files = glob.glob(os.path.join(conformance_gen_dir,
                                           face_prefix + '*' + TestConfiguration.Language.source_suffix[language]))
        for o in old_source_files:
            os.remove(o)

        template = stringtemplate3.StringTemplateGroup(name='conformance', file=open(template_file,'r'))

        test_num = 1

        assertion_file = open(os.path.join(conformance_gen_dir, 'face_specified_assertions.txt'), 'w')
        all_test_filename = 'all_' + language + '_conformance_tests.txt'
        all_tests_file = open(os.path.join(conformance_gen_dir, all_test_filename), 'w')
        distribution_statement = set_distro_statement.get_distribution_statement(language, all_test_filename)

        first_time = False # Was true...don't think I need to do this any more
        for current_node in iter(self.idl_tree):
            if not isinstance(current_node, OperationNode):
                continue

            if language == TestConfiguration.Language.Ada and first_time:
                loc = current_node.loc -1
                tokens = [['interface', 'Type_Safe_myTypes']]
                i_node = InterfaceNode(loc,tokens)
                self.idl_tree.current_node = current_node
                self.idl_tree.add_parent(i_node)
                first_time = False


            for d in data_type_list:
                self.idl_tree.overwrite_face_specified_type_nodes(d)
                specified_name = face_prefix + d + "_" + current_node.name + \
                                 TestConfiguration.Language.source_suffix[language]
                self.generate_node_conformance_test(current_node, language, test_num, template,
                                                    conformance_gen_dir, assertion_file, all_tests_file,
                                                    ['face.dm'], ' (' + d + ')',
                                                    specified_name)
                test_num += 1


        assertion_file.close()
        all_tests_file.close()

    def generate_node_conformance_test(self, current_node, language, test_num, template,
                                       conformance_gen_dir, assertion_file, all_tests_file,
                                       added_face_packages=list(), added_assertion_text='',
                                       specified_filename=None):

        module_nodes = current_node.get_specific_node_type_parents([ModuleNode, InterfaceNode])
        module_names = list()
        for n in module_nodes:
            module_names.append(n.get_fully_specified_name(language))
        conformance_module_names = list()
        for m in module_names:
            conformance_module_names.append('CONFORMANCE_TEST' + TestConfiguration.Language.delimiter[language] + m)
        if language == TestConfiguration.Language.Ada and isinstance(current_node, ConstantNode):
            module_names.append('Interfaces.C')

        module_names.extend(added_face_packages)
        #TODO: add distribution statement to top of each file
        test_code = current_node.get_conformance_test_code(template, language)
        code_template = template.getInstanceOf('conformanceSourceFile')
        code_template['TEST_CODE_SECTION1'] = test_code[0]
        code_template['TEST_CODE_SECTION2'] = test_code[1]
        code_template['FACE_PACKAGES'] = ', '.join(module_names)
        code_template['CONFORMANCE_TEST_FACE_PACKAGES'] = ', '.join(conformance_module_names)
        if specified_filename is not None:
            source_filename = specified_filename
            code_template['TEST_NAME'] = specified_filename.split('.')[0]
        else:
            code_template['TEST_NAME'] = 'Test' + str(test_num)
            source_file_prefix = 'Test'
            if language == Language.Ada:
                source_file_prefix = 'test'
            source_filename = source_file_prefix + str(test_num) + TestConfiguration.Language.source_suffix[language]


        source_file = open(os.path.join(conformance_gen_dir, source_filename), 'w')
        source_file.write(set_distro_statement.get_distribution_statement(language,source_filename))
        source_file.write(str(code_template))
        source_file.close()

        all_tests_file.write(2*'\n' + TestConfiguration.Language.begin_single_line_comment[language] + 80 * '#' +
                             TestConfiguration.Language.end_single_line_comment[language] + '\n')

        all_tests_file.write(TestConfiguration.Language.begin_single_line_comment[language] + ' Test' + str(test_num) +
                             TestConfiguration.Language.source_suffix[language] + ' Test Code ' +
                             TestConfiguration.Language.end_single_line_comment[language] + '\n')

        all_tests_file.write(str(code_template))

        assertion_file.write('AssertionTest' + str(test_num) + ' ' +
                             current_node.get_fully_specified_name(language) + added_assertion_text +
                             ' is conformant\n')


if __name__ == "__main__":

    import commonTools
    import os
    import Dialogs

    logger.info('In IdlParser main')

    initialDir = os.path.join(ConformanceMain.getRootDir(), 'datafiles')

    generate_TSS_2_1_conformance_test = True
    if generate_TSS_2_1_conformance_test:
        idlFilenames = dict()
        idlFilenames['ts common'] = os.path.join(initialDir, 'FACE_TS_common_2.1.idl')



    testing_generated_gold_libs = False
    if testing_generated_gold_libs:
        idlFilenames = dict()
        #idlFilenames['common'] = os.path.join(initialDir, 'FACE_common.idl')
        idlFilenames['ios'] = os.path.join(initialDir, 'FACE_ios.idl')
        idlFilenames['hmfm'] = os.path.join(initialDir, 'FACE_HMFM.idl')
        idlFilenames['tss'] = os.path.join(initialDir, 'FACE_tss.idl')

        package_names = { 'ios':'FACE.IO', 'tss': 'FACE.TS', 'hmfm': 'FACE.HMFM'}
        basenames = { 'ios': 'face-io', 'tss': 'face-ts', 'hmfm':'face-hmfm'}

        template_filename = os.path.join(initialDir, 'Ada_GoldLibTemplates.stg')
        conformance_gen_dir = os.path.join(ConformanceMain.getRootDir(), 'generated_gold_code')

        for gen_dir, idlFilename in idlFilenames.iteritems():
            logger.debug('Dir: ' + gen_dir + ' File: ' + idlFilename)

            if len(idlFilename) > 0:
                idlParser = FaceIdlParser(filename=idlFilename)

                gen_sub_dir = os.path.join(conformance_gen_dir, gen_dir)
                if not os.path.isdir(gen_sub_dir):
                    commonTools.mkdir_p(gen_sub_dir)

                gold_definition_filename = os.path.join(gen_sub_dir, basenames[gen_dir] + '.ads')
                gold_source_filename = os.path.join(gen_sub_dir, basenames[gen_dir] + '.adb')
                conformance_test_generator = ConformanceTestGenerator(idlParser.idl_tree)
                conformance_test_generator.generate_gold_library_files(TestConfiguration.Language.Ada,
                                                                       gold_definition_filename,
                                                                       gold_source_filename,
                                                                       template_filename,
                                                                       package_name=package_names[gen_dir])



    testing_face_specified_gold_lib = False
    if testing_face_specified_gold_lib:
        idlFilename = os.path.join(initialDir, 'FACE_SPECIFIED_TYPE_TSS.idl')
        template_filename = os.path.join(initialDir, 'Ada_GoldLibTemplates.stg')
        conformance_gen_dir = os.path.join(ConformanceMain.getRootDir(), 'generated_code')

        idlParser = FaceIdlParser(filename=idlFilename)
        gen_dir = "tss_face_specified_goldlib"
        gen_sub_dir = os.path.join(conformance_gen_dir, gen_dir)

        if not os.path.isdir(gen_sub_dir):
            commonTools.mkdir_p(gen_sub_dir)

        datatype_list = ['ProcessorStatus', 'AntennaStatus']
        gold_lib_generator = ConformanceTestGenerator(idlParser.idl_tree)
        gold_lib_generator.generate_face_specified_gold_library(datatype_list, TestConfiguration.Language.Ada,
                                                                gen_sub_dir, gen_sub_dir, template_filename)

    testing_face_specified_conformance_test = False
    #testing_face_specified_conformance_test = True
    if testing_face_specified_conformance_test:
        idlFilename = os.path.join(initialDir, 'FACE_SPECIFIED_TYPE_TSS.idl')
        #idlFilename = os.path.join(initialDir, 'DUMMY_SPECIFIED_TYPE_TSS.idl')
        template_filename = os.path.join(initialDir, 'Ada_ConformanceTemplates.stg')
        conformance_gen_dir = os.path.join(ConformanceMain.getRootDir(), 'generated_code')

        idlParser = FaceIdlParser(filename=idlFilename)
        gen_dir = "tss_face_specified"
        gen_sub_dir = os.path.join(conformance_gen_dir, gen_dir)
        if not os.path.isdir(gen_sub_dir):
            commonTools.mkdir_p(gen_sub_dir)

        datatype_list = ['dummy1', 'dummy2']
        conformance_test_generator = ConformanceTestGenerator(idlParser.idl_tree)
        conformance_test_generator.generate_face_specified_conformance_tests(datatype_list, TestConfiguration.Language.Ada,
                                                                             gen_sub_dir, template_filename)

    testing_all_interfaces = False
    #testing_all_interfaces = True
    if testing_all_interfaces:

        idlFilenames = dict()
        idlFilenames['common'] = os.path.join(initialDir, 'FACE_common.idl')
        idlFilenames['ios'] = os.path.join(initialDir, 'FACE_ios.idl')
        idlFilenames['hmfm'] = os.path.join(initialDir, 'FACE_HMFM.idl')
        idlFilenames['tss'] = os.path.join(initialDir, 'FACE_tss.idl')


        template_filename = os.path.join(initialDir, 'Ada_ConformanceTemplates.stg')
        conformance_gen_dir = os.path.join(ConformanceMain.getRootDir(), 'generated_code')
        #idlFilename = Dialogs.osAgnosticOpenFilenameDialog(initialdir=initialDir)

        for gen_dir, idlFilename in idlFilenames.iteritems():
            logger.debug('Dir: ' + gen_dir + ' File: ' + idlFilename)

            if len(idlFilename) > 0:
                idlParser = FaceIdlParser(filename=idlFilename)
                #idlParser.idl_tree.print_tree()
                #pprint.pprint( idlParser.tokens.asList())

                logger.debug('Num unhandled definitions: ' + str(idlParser.unhandled_definitions))

                gen_sub_dir = os.path.join(conformance_gen_dir, gen_dir)
                if not os.path.isdir(gen_sub_dir):
                    commonTools.mkdir_p(gen_sub_dir)

                conformance_test_generator = ConformanceTestGenerator(idlParser.idl_tree)
                conformance_test_generator.generate_conformance_tests(gen_sub_dir, template_filename, TestConfiguration.Language.Ada)




