#!/usr/bin/env python
"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""

__author__ = 'david.fann@vanderbilt.edu'

from os.path import isfile, isdir, relpath, join, dirname, basename, exists, sep
from os import walk, environ
from platform import system
from commonTools import copy_files, getHeaderDefineConstant, getPathSafeObjectName, listToFilePath, isInteractive,\
    mkdir_p, subprocessWithSpecifiedPath, displayFile, getObjectNamesStringFromList, getObjectNamesFromLibraryList, \
    copyFile, safeOpen
import logging
import Dialogs
import sys
import ConformanceTest
import pickle
from ConformanceMain import getRootDir

logger = logging.getLogger(__name__)



class idlToSourceCodeGenerator(object):
    '''
    classdocs
    '''

    def __init__(self, **kw):
    
        '''
        Constructor
        '''

        ''' Required keywords '''
        self.baseConfig = kw['baseConfig']
        self.tsConfig = kw['tsConfig']
        self.datatypeConfig = kw['datatypeConfig']
        ''' Optional keywords '''
        
        idlLocation = listToFilePath(['TSautogeneration', 'IDLCompiler', 'face_idl'], False)
        if system() == 'Windows':
            idlLocation += '.exe'

        self.idlCompiler = kw.get('idlCompiler', join(getRootDir(), idlLocation))
        
        self.idlCompilerFlags = kw.get('idlCompilerFlags', '')
        self.segmentLibs = kw.get('segmentLibs', '')
        self.tsGoldLibName = kw.get('goldLibName', 'tsInterfaceGoldLib.a')
        self.tsConformanceTestName = kw.get('conformanceTestName', 'tsConformanceTest')
        self.createLogFiles = kw.get('createLogFiles', True)
        aceSpecificEnviron = { 'ACE_ROOT' : getRootDir() + sep + 'TSautogeneration' + sep + 'ACE_wrappers',
                          'TAO_ROOT' : getRootDir() + sep + 'TSautogeneration' + sep + 'ACE_wrappers' + sep + 'TAO',
                          'LD_LIBRARY_PATH': getRootDir() + sep + 'TSautogeneration' + sep + 'ACE_wrappers' + sep + 'lib:$LD_LIBRARY_PATH' }
  
        self.environment = kw.get('environment', aceSpecificEnviron)
        
        if not isfile(self.idlCompiler) and self.datatypeConfig.faceSpecifiedType:
            errorMsg = 'IDL compiler does not exist:' + self.idlCompiler
            logger.error(errorMsg)
            if isInteractive():
                Dialogs.ErrorMessageBox('FACE Conformance Test Suite', errorMsg)
            raise IOError(errorMsg)
        
        self.basefiles = list()
        self.relativeDirectories = list()
        self.idlLogs = list()
        self.goldLogs = list()
        self.conformanceLogs = list()
        self.conformanceObjects = list()
        self.sourceCodeFromIDLGenerated = False

    def generateSourceCodeFromIDL(self):
        
        self.basefiles = dict()
        self.relativeDirectories = dict()
        self.idlLogs = list()
        if self.datatypeConfig.faceSpecifiedType:
            if self.datatypeConfig.specifyIdlFiles:
                for i in self.datatypeConfig.specifiedIdlFiles.items():
                    if not self.compileIdlFile(i, None):
                        return False        
            else:
                for rootDir, idlFiles in self.datatypeConfig.idlFiles.iteritems():
                    self.outputDirectory = join(rootDir, 'generated')
                    print 'outputDir: ' + self.outputDirectory
                    for i in idlFiles:
                        if not self.compileIdlFile(i, rootDir):
                            return False
    
        self.sourceCodeFromIDLGenerated = True
        return True
        
    def compileIdlFile(self, idlFile, rootDir):
        if rootDir == None:
            logger.error("Can't Handle specified IDL files yet")
            exit(-1)
        
        if rootDir not in self.relativeDirectories.keys():
            self.relativeDirectories[rootDir] = list()
        
        if rootDir not in self.basefiles.keys():
            self.basefiles[rootDir] = list()
        
        relDir = dirname(idlFile)
                 
        if relDir == './' or relDir == '.' or relDir == '':
            logger.debug('got a . in the relDir')
            relDir = ''
        
        if relDir not in self.relativeDirectories[rootDir]:
            self.relativeDirectories[rootDir].append(relDir)
                
        filename = basename(idlFile)
        basefile = filename[:-4]
        self.basefiles[rootDir].append(join(relDir, basefile).replace('\\', '/'))
        
        logfile = basefile + "_idlcompiler.log"  
        self.idlLogs.append(logfile)

        logDir = self.baseConfig.logDirectory
        outDir = self.outputDirectory

        if relDir != '':
            outDir = join(self.outputDirectory, relDir)
            logDir = join(self.baseConfig.logDirectory, relDir)


        if not isdir(outDir):
            if exists(outDir):
                logger.error('output directory already exists as a file:' + outDir)
                sys.exit()
            mkdir_p(outDir)

        if not isdir(logDir):
            if exists(logDir):
                logger.error('log directory already exists as a file:' + logDir)
                sys.exit()
            mkdir_p(logDir)

        
        commandString = self.idlCompiler 
        for d in self.datatypeConfig.idlIncludePath:
            commandString +=  ' -I' + d
        
        commandString += ' ' + self.idlCompilerFlags + ' -o ' + outDir + '  ' + idlFile
        
        if self.createLogFiles:
            commandString +=  ' > ' + join(self.baseConfig.logDirectory, logfile) + ' 2>&1'
            
        environVar = environ
        if self.environment is not None:
            environVar.update(self.environment)

        #result = subprocess.call(commandString, cwd = rootDir, env = environVar, shell=True)
        result = subprocessWithSpecifiedPath(commandString, self.baseConfig.addedPathDirs, rootDir)
        resultMsg = 'returncode: ' + str(result.returncode) + ' stdout: ' + str(result.stdout) + ' stderr: ' + str(result.stderr)
        logger.debug(resultMsg)


        if result.returncode != 0:
            errorMsg = 'Error in generating code from IDL: ' + idlFile
            if self.createLogFiles:
                errorMsg += ' see ' + join(self.baseConfig.logDirectory, logfile) + ' for more information '
            logger.error(errorMsg)
            
            if isInteractive():
                Dialogs.ErrorMessageBox('Candidate FACE Conformance Test Suite', errorMsg + resultMsg)
                displayFile(join(self.baseConfig.logDirectory, logfile), maxWidth = 150)
                return False
            else: 
                sys.exit()
            
                        
        return True

    def generateTSGoldStandardCode(self, **kw):
        if not self.sourceCodeFromIDLGenerated:
            if not self.generateSourceCodeFromIDL():
                return False

        goldDir = listToFilePath([getRootDir(), 'goldStandardLibs', 'TS', self.baseConfig.language])
        
        self.copyTSGoldStandardCode(goldDir)
        self.generateTSDatatypesHeader(goldDir)
        self.generateTSGoldHeaderFile(goldDir)
        self.generateTSGoldLibMakefileEnv(goldDir, **kw)
            
        return True
    
    def copyTSGoldStandardCode(self, goldDir):
        sourceSuffix = self.baseConfig.getLanguageSuffix()
        headerSuffix = self.baseConfig.getLanguageHeaderSuffix()

        for rootDir, relDirs in self.relativeDirectories.iteritems():
            for relDir in relDirs:
                curDir = join(rootDir + sep + 'generated', relDir)
                copy_files(curDir + sep + '*typed_tss' + headerSuffix, join(goldDir, relDir))
                copy_files(curDir + sep + '*typed_tss' + sourceSuffix, join(goldDir, relDir))
 
        self.copyInterfaceSpecificIncludeFiles()

  
    def generateTSDatatypesHeader(self, goldDir):
        tssHeaderFilename = join(goldDir, 'tss_datatypes' + self.baseConfig.getLanguageHeaderSuffix())
        
        tssHeader = safeOpen(tssHeaderFilename, 'w')
        headerConstant = getHeaderDefineConstant(tssHeaderFilename)
        tssHeader.write('#ifndef ' + headerConstant + '\n')
        tssHeader.write('#define ' + headerConstant + '\n')

        for rootDir, headerFiles in self.datatypeConfig.datatypeHeaderFiles.iteritems():
            for headerFile in headerFiles:
                tssHeader.write('#include "' + headerFile + '"\n')
        
        tssHeader.write('#endif /* #ifndef ' + headerConstant + ' */' )
        tssHeader.close()
        
    def generateTSGoldHeaderFile(self, goldDir):
        headerSuffix = self.baseConfig.getLanguageHeaderSuffix()
        tssGoldHeaderFilename = join(goldDir, 'tsInterfaceGoldLib' + headerSuffix)

        tssGoldHeader = safeOpen(tssGoldHeaderFilename, 'w')
        headerConstant = getHeaderDefineConstant(tssGoldHeaderFilename)
        tssGoldHeader.write('#ifndef ' + headerConstant + '\n')
        tssGoldHeader.write('#define ' + headerConstant + '\n\n')
        tssGoldHeader.write('#include "tss_common' + headerSuffix + '"\n')
        tssGoldHeader.write('#include "untyped_tss' + headerSuffix + '"\n')
        
        for rootDir, basefiles in self.basefiles.iteritems():
            for b in basefiles:
                tssGoldHeader.write('#include "' + b + '_typed_tss' + headerSuffix + '"\n')
        
        tssGoldHeader.write('#endif /* #ifndef ' + headerConstant + ' */' )
  
        tssGoldHeader.close()
   
    
    def generateVerificationTestCasesFile(self, verifDir, **kw):
        
        testCases = safeOpen(join(verifDir, 'testCases.txt'), 'w')

        for rootDir, basefiles in self.basefiles.iteritems():
            for b in basefiles:
                fullName = b + '_typed_tss.' + self.baseConfig.objectSuffix + '\n'
                testCases.write(fullName.replace('\\', '/'))
        testCases.write('untyped_tss.{}\n'.format(self.baseConfig.objectSuffix))
        testCases.write('verifyUntypedTsInterfaceFailLib.{}\n'.format(self.baseConfig.objectSuffix))
        testCases.write('tsInterfaceGoldLib.a\n')
        testCases.close()


    def generateTSGoldLibMakefileEnv(self, goldDir, **kw): 
            
        makeEnvFilename = join(goldDir, 'Makefile.env')
        additionalCompilerFlags = kw.get('additionalCompilerFlags', '')
        makeEnv = safeOpen(makeEnvFilename,'w')
        
        interfaceSpecificFlags = ' -I' + getPathSafeObjectName(getRootDir() + sep + 'include') + ' -I' + getPathSafeObjectName(listToFilePath([getRootDir(), 'goldStandardLibs', 'TS', self.baseConfig.language]))
        
        for rootDir, relDirs in self.relativeDirectories.iteritems():
            for relDir in relDirs:
                interfaceSpecificFlags += ' -I' + getPathSafeObjectName(listToFilePath([getRootDir(),'include', 'interfaceSpecific', relDir]))

        includeFlags = ''        
        for i in self.datatypeConfig.datatypeIncludePath:
            if i != '':
                includeFlags += ' -I' + getPathSafeObjectName(i)

        for rootDir, headerFiles in self.datatypeConfig.datatypeHeaderFiles.iteritems():
            for headerFile in headerFiles:
                relDir = dirname(headerFile)
                if relDir != '':
                    includeFlags += ' -I' + getPathSafeObjectName(join(rootDir, relDir))
                            
        makeEnvVariables = dict()
        makeEnvVariables['compilerFlags'] = additionalCompilerFlags + interfaceSpecificFlags + includeFlags + self.tsConfig.getTestSpecificCompilerFlags()
        
        self.baseConfig.writeMakefileEnvironment(makeEnv, **makeEnvVariables)
        
        
        makeEnv.write('FACE_GOLD_LIB = ' + self.tsGoldLibName + '\n')
        makeEnv.write('FACE_SRCS = untyped_tss' + self.baseConfig.getLanguageSuffix())
        for rootDir, basefiles in self.basefiles.iteritems():
            for b in basefiles:
                fullName = ' ' + b + '_typed_tss' + self.baseConfig.getLanguageSuffix()
                makeEnv.write(fullName.replace('\\', '/'))
        makeEnv.write('\n')
        makeEnv.close()
    
    def generateTSConformanceTestCode(self, **kw):
        if not self.sourceCodeFromIDLGenerated:
            if not self.generateSourceCodeFromIDL():
                return False
        conformanceDir = listToFilePath([getRootDir(), 'conformanceInterfaceTest', 'TS', self.baseConfig.language])
        
        self.copyTSConformanceCode(conformanceDir)
        self.generateTSDatatypesHeader(conformanceDir)
        self.tsConfig.createHeaderFile(join(conformanceDir, 'tss_conformance_test' + self.baseConfig.getLanguageHeaderSuffix()))
        self.generateTSAssertionsFile(conformanceDir)    
        self.tsConfig.getTestAssertions(conformanceDir, self.baseConfig.getLanguageSuffix(), self.baseConfig.logDirectory)
        self.generateTSConformanceMakefileEnv(conformanceDir, **kw)
        return True
    
    def copyTSConformanceCode(self, conformanceDir):
        sourceSuffix = self.baseConfig.getLanguageSuffix()
        headerSuffix = self.baseConfig.getLanguageHeaderSuffix()

        for rootDir, relDirs in self.relativeDirectories.iteritems():
            for relDir in relDirs:
                curDir = join(rootDir + sep + 'generated', relDir)
                copy_files(curDir + sep + '*TsInterfaceConformanceTest' + headerSuffix, join(conformanceDir, relDir))
                copy_files(curDir + sep + '*TsInterfaceConformanceTest' + sourceSuffix, join(conformanceDir, relDir))

        self.copyInterfaceSpecificIncludeFiles()


    def generateTSConformanceMakefileEnv(self, conformanceDir, **kw): 
            
        makeEnvFilename = join(conformanceDir, 'Makefile.env')
        goldLibs = kw.get('goldLibs', self.segmentLibs)
        additionalCompilerFlags = kw.get('additionalCompilerFlags', '')
        makeEnv = safeOpen(makeEnvFilename,'w')
        
        interfaceSpecificFlags = ' -I' + getPathSafeObjectName(getRootDir() + sep + 'include') + ' -I' + getPathSafeObjectName(listToFilePath([getRootDir(), 'conformanceInterfaceTest', 'TS', self.baseConfig.language]))
        
        for rootDir, relDirs in self.relativeDirectories.iteritems():
            for relDir in relDirs:
                interfaceSpecificFlags += ' -I' + getPathSafeObjectName(listToFilePath([getRootDir(),'include', 'interfaceSpecific', relDir]))

        includeFlags = ''        
        for i in self.datatypeConfig.datatypeIncludePath:
            if i != '':
                includeFlags += ' -I' + getPathSafeObjectName(i)
     
        for rootDir, headerFiles in self.datatypeConfig.datatypeHeaderFiles.iteritems():
            for headerFile in headerFiles:
                relDir = dirname(headerFile)
                if relDir != '':
                    includeFlags += ' -I' + getPathSafeObjectName(join(rootDir, relDir))
     
        for i in self.tsConfig.includePath:
            if i != '':
                includeFlags += ' -I' + getPathSafeObjectName(i)
        
        makeEnvVariables = dict()
        makeEnvVariables['compilerFlags'] = additionalCompilerFlags + interfaceSpecificFlags + includeFlags + self.tsConfig.getTestSpecificCompilerFlags()
        if 'environVariables' in kw:
            makeEnvVariables['environVariables'] = kw['environVariables']
        self.baseConfig.writeMakefileEnvironment(makeEnv, **makeEnvVariables)
        
        makeEnv.write('FACE_SRCS = ')
        for t in self.tsConfig.targets:
            makeEnv.write(' ' + t + self.baseConfig.getLanguageSuffix())
        makeEnv.write('\n')
        
        # HERE IS THE STUFF!
        self.tsConfig.object_suffix = self.baseConfig.objectSuffix
        makeEnv.write('FACE_SEGMENT_OBJS = ' + getObjectNamesFromLibraryList(self.tsConfig.libraries, self.tsConfig.objectSuffix, self.tsConfig.recursive_object_search) + '\n')
        makeEnv.write('FACE_GOLD_LIBS = ' + getObjectNamesStringFromList(goldLibs) + '\n')
        
        makeEnv.close()
  
    def generateTSAssertionsFile(self, conformanceDir):
        baseAssertions = safeOpen(join(conformanceDir, 'base_assertions.txt'), 'r')
        assertionFile = safeOpen(join(conformanceDir, 'assertions.txt'), 'w')
    
        baseLines = baseAssertions.readlines()
        assertionFile.writelines(baseLines)

        if self.datatypeConfig.faceSpecifiedType:
            for rootDir, basefiles in self.basefiles.iteritems():
                for b in basefiles:
                    assertionFile.write('Assertion' + b + '_TsInterfaceConformanceTest Testing ' + b + ' has a conformant TS interface\n')
        
        assertionFile.close()
      
    ''' Currently not used...may want to delete'''
    def createLogDirectories(self):
        
        if not isdir(self.baseConfig.logDirectory):
            if exists(self.baseConfig.logDirectory):
                logger.error('Log directory name exist, but is not a directory: ' + self.baseConfig.logDirectory)
                return False
            mkdir_p(self.baseConfig.logDirectory)
        
        for relDir in self.relativeDirectories:
            logDir = join(self.baseConfig.logDirectory, relDir)
            if not isdir(logDir):
                if exists(logDir):
                    logger.error('Log directory name exist, but is not a directory: ' + logDir)
                    return False
                mkdir_p(logDir)
                
        return True
    
    ''' Currently not used...may want to delete  '''
    def createInterfaceSpecificIncludeDirectories(self):
       
        for relDir in self.relativeDirectories:
            incDir = join(getRootDir() + '/include/interfaceSpecific/', relDir)
            if not isdir(incDir):
                if exists(incDir):
                    logger.error('Include directory name exist, but is not a directory: ' + incDir)
                    return False
                mkdir_p(incDir)
                
        return True
 
   
    def copyInterfaceSpecificIncludeFiles(self):
        
        headerSuffix = self.baseConfig.getLanguageHeaderSuffix()
            
        for rootDir, basefiles in self.basefiles.iteritems():
            for b in basefiles:
                filename = b + '_TsInterfaceSpecific' + headerSuffix
                copyFile(listToFilePath([rootDir,'generated', filename], False), join(listToFilePath([getRootDir(),'include','interfaceSpecific',filename], False)), overwrite = False)
    
    
    
if __name__ == '__main__':

    import ConformanceMain
    commandString = 'C:\\Users\\dfann\\Documents\\FACE\\DEVEL\\FACEConformanceTestSuite-1.x-mswinders\\TSautogeneration\\IDLCompiler\\face_idl.exe -IC:\\Users\\dfann\\Documents\\FACE\\DEVEL\\FACEConformanceTestSuite-1.x-mswinders\\verificationTests\\IDL  -o C:\\Users\\dfann\\Documents\\FACE\\DEVEL\\FACEConformanceTestSuite-1.x-mswinders\\verificationTests\\IDL\\generated  EntityHistoryReport_graphics.idl > C:\\Users\\dfann\\Documents\\FACE\\DEVEL\\FACEConformanceTestSuite-1.x-mswinders\\logfiles\\EntityHistoryReport_graphics_idlcompiler.log 2>&1'
    rootDir = 'C:\\Users\\dfann\\Documents\\FACE\\DEVEL\\FACEConformanceTestSuite-1.x-mswinders\\verificationTests\\IDL'
    
    pathList = [u'C:\\\\MinGW\\\\bin', u'C:\\\\MinGW\\\\msys\\\\1.0\\\\bin']
    result = subprocessWithSpecifiedPath(commandString, pathList, rootDir)
    print 'returncode: ' + str(result.returncode) + ' stdout: ' + str(result.stdout) + ' stderr: ' + str(result.stderr)
    sys.exit()

    ConformanceMain.initializeLogger()
    logger = logging.getLogger(__name__)
    logger.info('In idlToSourceCodeGenerator main')   
    config1 = ConformanceTest()

    configurationFile = '/home/dfann-local/FACE/FACEConformanceTestSuite-1.x/config/tsGoldDummy.cfg'

    config = pickle.load( safeOpen(configurationFile, "r"))
    config.datatypeConfig.initializeTest(config.baseConfig.language)
    idlArgs = dict()
    idlArgs['baseConfig'] = config.baseConfig
    idlArgs['tsConfig'] = config.tsConfig
    idlArgs['datatypeConfig'] = config.datatypeConfig

    idlTest = idlToSourceCodeGenerator(**idlArgs)
    idlTest.generateTSGoldStandardCode()
    idlTest.generateTSConformanceTestCode(goldLibs=config.getLinkableLibraries(config.tsConfig.segmentName))


