"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""
"""
---------------------------------------------------------------------------------

idl_parser.py

Parses FACE IDL versions 2.0 and 2.1.

Primary Classes:
FaceIdlParser - Main class in this file.  Utilizes the other classes in the file.
    Usage: face_idl_parser = idl_parser.FaceIdlParser(idl_filename)
      or  face_idl_parser = idl_parser.FaceIdlParser()
          face_idl_parser.parse_idl_file(idl_filename)

IdlTree - Data structure that holds information from the IDL file
Node - base class that is used for holding information in the IDL file.  Has a
    subclass for all types of information that are found in the IDL file.
    Externally used methods:
        get_declaration(language, gold_lib_template) - returns a string of the node's
            declaration in the specified language using the specified template. (i.e. header)
        get_implementation(language, gold_lib_template) - returns a string of the node's
            implementation in the specified language, using the specified template. (i.e. source)
        get_conformance_test(language, conformance_test_template) - returns a string of the node's
            conformance test in the specified language, using the specified template.

        Each of these methods call subclass methods named
        get_LANGUAGE_declaration/implementation/conformance_test(template)
        based on the language specified.

Used in coordination with idl_to_code_generator.py to generate gold standard
library and conformance interface test

Acknowledgement:
FaceIdlParser.initialize_parser modified from an example of using the parsing module
to be able to process a subset of the CORBA IDL grammar. Was copyrighted (c) 2003,
Paul McGuire (MIT License)

"""

__author__ = 'david.fann@vanderbilt.edu'


import os
import sys

import ConformanceMain
if __name__ == "__main__":
    ConformanceMain.initializeLogger()

egg_path = os.path.join(ConformanceMain.getRootDir(), 'face_conformance_app', 'pyparsing-2.0.1-py2.7.egg')
sys.path.append(egg_path)
egg_path = os.path.join(ConformanceMain.getRootDir(), 'face_conformance_app', 'stringtemplate3-3.1-py2.7.egg')
sys.path.append(egg_path)
egg_path = os.path.join(ConformanceMain.getRootDir(), 'face_conformance_app', 'setuptools-0.6c11-py2.7!\\')
sys.path.append(egg_path)

from pyparsing import Literal, CaselessLiteral, Word, Upcase, OneOrMore, ZeroOrMore, \
        Forward, NotAny, delimitedList, oneOf, Group, Optional, Combine, alphas, nums, restOfLine, cStyleComment, \
        alphanums, printables, empty, quotedString, ParseException, ParseResults, Keyword
import collections
import logging
import stringtemplate3
import itertools
import glob


import TestConfiguration
import constants
import commonTools

_C = TestConfiguration.Language.C
_CPP = TestConfiguration.Language.CPP
_ADA = TestConfiguration.Language.Ada
_JAVA = TestConfiguration.Language.Java


logger = logging.getLogger(__name__)


class IdlTree:

    def __init__(self):
        self.root_node = Node(name='ROOT')
        self.root_node.in_namespace = False
        self.face_specified_type_nodes = None
        self.current_node = self.root_node
        self.data_callback_node = None
        self.current_idl_filename = None

    def __iter__(self):
        for v in iter(self.root_node):
            yield v
        #yield v

    def add_sibling(self, new_node):
        parent = self.current_node.parent

        if self.current_node == self.root_node:
           parent = self.root_node

        parent.children.append(new_node)
        new_node.parent = parent

        self.current_node = new_node
        self.current_node.idl_filename = self.current_idl_filename

    def add_child(self, new_node):
        self.current_node.children.append(new_node)
        new_node.parent = self.current_node
        self.current_node = new_node
        self.current_node.idl_filename = self.current_idl_filename


    def add_parent(self, new_node):
        if self.current_node == self.root_node:
            new_node.parent = self.root_node
            self.root_node.children.append(new_node)

        else:
            new_node.parent = self.current_node.parent
            if new_node.parent.loc > new_node.loc:
                raise RuntimeError('new node came before its parent, new_node: ' + new_node.to_string())

            siblings = list()
            for siblingOrChild in self.current_node.parent.children:
                if siblingOrChild.loc < new_node.loc:
                    siblings.append(siblingOrChild)
                else:
                    new_node.children.append(siblingOrChild)
                    siblingOrChild.parent = new_node

            siblings.append(new_node)
            new_node.parent.children = siblings

        self.current_node = new_node
        self.current_node.idl_filename = self.current_idl_filename

    def find_face_specified_type_nodes(self):
        self.face_specified_type_nodes = list()
        for n in iter(self):
            if n.name == "FACE_SPECIFIED_TYPE" or \
                    ('type_name' in n.__dict__.keys() and n.type_name == "FACE_SPECIFIED_TYPE"):
                    self.face_specified_type_nodes.append(n)

            if isinstance(n, OperationNode):
                for c in n.children:
                    if isinstance(c, ParameterNode):
                        if c.type_name == "FACE_SPECIFIED_TYPE":
                            self.face_specified_type_nodes.append(n)
                            break
                        if c.name == 'data_callback' and os.path.basename(c.idl_filename) == 'FACE_TS.idl':
                            # Special case: data_callback in FACE_TS is a face specified node,
                            # in TypeAbstraction, it isn't
                            self.face_specified_type_nodes.append(n)
                            break

            if n.name.lower() == 'data_callback':
                self.data_callback_node = n


    # Hack to replace FACE_SPECIFIED_TYPE in the idl and in the register_read_callback
    def overwrite_face_specified_type_nodes(self, new_name):
        if self.face_specified_type_nodes is None:
            self.find_face_specified_type_nodes()

        for n in self.face_specified_type_nodes:
            n.type_name = 'face.dm.' + new_name
        if self.data_callback_node is not None:
            self.data_callback_node.type_name = 'FACE.TS.Read_Callback.Send_Event_face_dm_' + new_name + '_Access'

    def print_tree(self, indent_text = '-'):
        self.root_node._print_node_tree(0, indent_text)

    def in_tree(self, node):
        return self._search_tree(node, self.root_node)

    def _search_tree(self, node, tree_node):
        if node == tree_node:
            return True
        for n in tree_node.children:
            if self._search_tree(node, n):
                return True

        return False

    def combine_modules(self):
        module_nodes = dict()
        for v in self:
            if isinstance(v, ModuleNode):
                if v.get_fully_specified_name() in module_nodes:  # Same module
                    original_module_node = module_nodes[v.get_fully_specified_name()]
                    if original_module_node.parent != v.parent:
                        raise RuntimeError('modules with same fully specified name should have the same parent: '
                            + str(v.parent) + ' ' + str(original_module_node.parent))

                    # put children of duplicate as children of original
                    # and set parent of duplicate children to be original
                    for c in v.children:
                        original_module_node.children.append(c)
                        c.parent = original_module_node

                    # delete duplicate from it's parent children
                    v.parent.children.remove(v)
                else:
                    module_nodes[v.get_fully_specified_name()] = v

    def get_fully_specified_type_name(self, type_name, language):
        fully_specified_type_name = TestConfiguration.Language.type_name_dict[language].get(type_name, None)
        if fully_specified_type_name is not None:
            return fully_specified_type_name # it is a language specific type

        for t in self:
            if t.name == type_name:
                return t.get_fully_specified_name(language)

        return None


class Node:

    def __init__(self, **kw):
        self.name = kw.get('name', None)
        self.parent = kw.get('parent', None)
        self.children = kw.get('children', list())
        self.loc = kw.get('loc', None)
        self.in_namespace = True
        self.conformance_test_node = False
        self.declaration_node = False
        self.implementation_node = False
        self.idl_filename = None
        self.declaration_language_jump = {
            _C: self.get_c_declaration,
            _CPP: self.get_cpp_declaration,
            _ADA: self.get_ada_declaration,
            _JAVA: self.get_java_declaration
        }
        self.implementation_language_jump = {
            _C: self.get_c_implementation,
            _CPP: self.get_cpp_implementation,
            _ADA: self.get_ada_implementation,
            _JAVA: self.get_java_implementation
        }

        self.conformance_test_language_jump= {
            _C: self.get_c_conformance_test,
            _CPP: self.get_cpp_conformance_test,
            _ADA: self.get_ada_conformance_test,
            _JAVA: self.get_java_conformance_test
        }


    def __iter__(self):
        yield self
        for v in itertools.chain(*itertools.imap(iter, self.children)):
            yield v

    def is_parent(self, possible_parent_node):

        true_parent = self.parent
        while not true_parent is None:
            if possible_parent_node == true_parent:
                return True
            else:
                true_parent = true_parent.parent

        return False



    def get_declaration(self, language, template):
        return self.declaration_language_jump[language](template)

    def get_c_declaration(self, template):
        warning = 'C declaration not implemented for ' + self.__class__.__name__ + ' node'
        logger.warning(warning)
        return warning + '\n'

    def get_cpp_declaration(self, template):
        warning = 'C++ declaration not implemented for ' + self.__class__.__name__ + ' node'
        logger.warning(warning)
        return warning + '\n'

    def get_ada_declaration(self, template):
        warning = 'Ada declaration not implemented for ' + self.__class__.__name__ + ' node'
        logger.warning(warning)
        return warning + '\n'

    def get_java_declaration(self, template):
        warning = 'Java declaration not implemented for ' + self.__class__.__name__ + ' node'
        logger.warning(warning)
        return warning + '\n'

    def get_implementation(self, language, template):
        return self.implementation_language_jump[language](template)

    def get_c_implementation(self, template):
        warning = 'C implementation not implemented for ' + self.__class__.__name__ + ' node'
        logger.warning(warning)
        return warning + '\n'

    def get_cpp_implementation(self, template):
        warning = 'C++ implementation not implemented for ' + self.__class__.__name__ + ' node'
        logger.warning(warning)
        return warning + '\n'

    def get_ada_implementation(self, template):
        warning = 'Ada implementation not implemented for ' + self.__class__.__name__ + ' node'
        logger.warning(warning)
        return warning + '\n'

    def get_java_implementation(self, template):
        warning = 'Java implementation not implemented for ' + self.__class__.__name__ + ' node'
        logger.warning(warning)
        return warning + '\n'

    def get_conformance_test(self, language, template):
        return self.conformance_test_language_jump[language](template)

    def get_c_conformance_test(self, template):
        warning = 'C conformance test not implemented for ' + self.__class__.__name__ + ' node'
        logger.warning(warning)
        return [warning + '\n']

    def get_cpp_conformance_test(self, template):
        warning = 'C++ conformance test not implemented for ' + self.__class__.__name__ + ' node'
        logger.warning(warning)
        return [warning + '\n']

    def get_ada_conformance_test(self, template):
        warning = 'Ada conformance test not implemented for ' + self.__class__.__name__ + ' node'
        logger.warning(warning)
        return [warning + '\n']

    def get_java_conformance_test(self, template):
        warning = 'Java conformance test not implemented for ' + self.__class__.__name__ + ' node'
        logger.warning(warning)
        return [warning + '\n']

    def get_node_description(self, delimiter=':'):
        if self.parent is not None:
            return self.parent.get_nodlllle_description() + delimiter + self.name + ' (' + self.__class__.__name__ + ')'
        else:
            return self.name

    def get_declaration_name(self, language):
        if language in [_CPP, _ADA]:
            # C++ or Ada doesn't want fully specified name in declaration since it is handled by the namespacing
            return self.name
        else:
            return self.get_fully_specified_name(language)

    def get_fully_specified_name(self, language=TestConfiguration.Language.C):

        if language == _ADA:  # Magic needed to get the right name in Ada since it doesn't follow idl file
            idl_file_to_specified_name_dict = {
                'FACE_TS.idl': 'FACE.TS.',
                'FACE_TypeAbstraction_TS.idl': 'FACE.TS.',
                'FACE_TS_common.idl': 'FACE.TS.',
                'FACE_tss.idl': 'FACE.TS.',  # 2.0 Version
                'FACE_IOS.idl': 'FACE.IO.',
                'FACE_ios.idl': 'FACE.IO.', # 2.0 Version
                'FACE_common.idl': 'FACE.',
                'FACE_HMFM.idl': 'FACE.HMFM.'
            }

            spec_name = idl_file_to_specified_name_dict.get(os.path.basename(self.idl_filename),
                                                            'NO MAPPING FOR IDL FILE: ' +
                                                            os.path.basename(self.idl_filename))
            return spec_name + self.name

        parent_node = self.parent
        parent_name = ''

        while parent_node is not None:
            if parent_node.in_namespace:
                parent_name = parent_node.get_fully_specified_name(language)
                break
            else:
                parent_node = parent_node.parent

        if len(parent_name) > 0:
            parent_name += TestConfiguration.Language.delimiter[language]
            return  parent_name + TestConfiguration.Language.format_case(language, self.name)
        else:
            return TestConfiguration.Language.format_case(language, self.name)

    def _find_fully_specified_name(self, language, name):
        if name in TestConfiguration.Language.type_name_dict[language].values():
            return name

        if name in TestConfiguration.Language.type_name_dict[language].keys():
            return TestConfiguration.Language.type_name_dict[language][name]

        named_node = self._find_node_name(name)
        if named_node is None:
            return 'COULD NOT FIND NAME IN TREE: ' + name
        else:
            return named_node.get_fully_specified_name(language)

    def _find_node_name(self, name):
        """ Look for node with name at this node, then at the node's children, and finally recursively at it's parent
        """
        if self.name == name:
            return self
        for c in self.children:
            if c.name == name:
                return c
        if self.parent is not None:
            return self.parent._find_node_name(name)
        else:
            return None

    def get_fully_specified_type_name(self, language):

        if 'type_name' in vars(self).keys():
            return_value = TestConfiguration.Language.type_name_dict[language].get(self.type_name, None)
            if return_value is None and self.parent is not None:
                # TODO: This should be looked up in the idlTree to get the fully specified typename
                type_name_node = self.parent._find_node_name(self.type_name)
                if type_name_node is not None:
                    return_value = type_name_node.get_fully_specified_name(language)
                else:
                    # Couldn't find it in the all the nodes...so my guess is just use the type name
                    return_value = self.type_name
            return return_value
        else:
            return None

    def _print_node_tree(self, indent=0, indent_text='-', name_only=True):
        if name_only:
            print (indent * indent_text) + self.get_fully_specified_name(_C)
        else:
            print (indent * indent_text) + self.to_string()
        indent += 1
        for c in self.children:
            c._print_node_tree(indent, indent_text)

    def to_string(self):
        return_value = self.__class__.__name__ + ' '
        for variable, value in sorted(vars(self).iteritems()):
            if variable == 'children' or variable == 'parent':
                continue
            return_value += '[' + variable + ': ' + str(value) + ']'
        return return_value

    def get_conformance_test_code(self, language, template):
        warning_message = self.__class__.__name__ + ' conformance test code not implemented yet: ' + self.name
        logger.warning(warning_message)
        return None

    def get_specific_node_type_parents(self, node_types):
        parent = self.parent
        specific_parents = list()

        while parent is not None:
            for n in node_types:
                if isinstance(parent, n):
                    specific_parents.append(parent)
                    break
            parent = parent.parent

        return reversed(specific_parents)


class ModuleNode(Node):
    def __init__(self, loc, tokens, **kw):
        Node.__init__(self, loc=loc, **kw)

        if len(tokens) < 2 or tokens[0] != 'module':
            raise RuntimeError("Unexpected tokens, expected ['module','MODULE_NAME', ...]")

        self.name = tokens[1]



class OperationNode(Node):

    def __init__(self, loc, tokens, **kw):
        Node.__init__(self, loc=loc, **kw)
        if len(tokens) < 2:
            raise RuntimeError("Unexpected tokens, expected [['OPERATION_RETURN_TYPE', 'OPERATION_NAME',...]]")
        self.name = tokens[1]
        self.return_type = tokens[0]
        self.conformance_test_node = True    # Need to generate a conformance test for this type of node
        self.declaration_node = True  # Need to generate a definition in gold lib
        self.implementation_node = True     # Need to generate source code in gold lib

    def get_c_declaration(self, template):

        operation_signature_template = template.getInstanceOf('functionDeclaration')

        if self.name.lower() == 'send_event':  # magically should be typedef to a pointer
            operation_signature_template = template.getInstanceOf('functionPtrDeclaration')

        operation_signature_template['FUNCTION_NAME'] = self.get_declaration_name(_C)

        # declaration and implementation are exactly the same except for template instance
        return self._get_c_or_cpp_operation_values(_C, operation_signature_template, template)

    def get_c_implementation(self, template):
        operation_signature_template = template.getInstanceOf('functionImplementation')
        operation_signature_template['FUNCTION_NAME'] = self.get_fully_specified_name(_C)

        if self.name.lower() == 'send_event':
            # VERY SPECIAL CASE...is magically a pointer...so don't need an implementation
            return ''

        # declaration and implementation are exactly the same except for template instance
        return self._get_c_or_cpp_operation_values(_C, operation_signature_template, template)

    def get_cpp_declaration(self, template):
        operation_signature_template = template.getInstanceOf('functionDeclaration')

        if self.name.lower() == 'send_event':  # magically should be typedef to a pointer
            operation_signature_template = template.getInstanceOf('functionPtrDeclaration')

        operation_signature_template['FUNCTION_NAME'] = self.get_declaration_name(_CPP)

        # declaration and implementation are exactly the same except for template instance
        return self._get_c_or_cpp_operation_values(_CPP, operation_signature_template, template)

    def get_cpp_implementation(self, template):
        if self.name.lower() == 'send_event':
            # VERY SPECIAL CASE...is magically a pointer...so don't need an implementation
            return ''

        operation_signature_template = template.getInstanceOf('functionImplementation')
        operation_signature_template['FUNCTION_NAME'] = self.get_fully_specified_name(_CPP)


        # declaration and implementation are exactly the same except for template instance
        return self._get_c_or_cpp_operation_values(_CPP, operation_signature_template, template)

    def _get_c_or_cpp_operation_values(self, language, operation_signature_template, template):

        operation_signature_template['RETURN_TYPE'] = self.return_type

        if self.name == 'Initialize':
            pass

        last_parameter = None
        for n in reversed(self.children):
            if isinstance(n, ParameterNode):
                last_parameter = n
                break

        if last_parameter is None:  # No parameters
            operation_signature_template['PARAMETERS'] = 'void'

        else:  # Get parameter signatures
            for child in self.children:
                if not isinstance(child, ParameterNode):
                    continue

                parameter_signature_template = template.getInstanceOf('parameterDeclaration')
                parameter_signature_template['DIRECTION'] = child.get_equal_spaced_direction()
                type_name = child.get_type_prefix(language) + child.get_fully_specified_type_name(language)

                if child.name == 'data_callback':  # VERY SPECIAL CASE
                    parameter_signature_template['TYPE_NAME'] = type_name + \
                                                                TestConfiguration.Language.delimiter[language] + \
                                                                'send_event_Ptr'
                    parameter_signature_template['NAME'] = child.name
                else:
                    parameter_signature_template['TYPE_NAME'] = type_name
                    parameter_signature_template['NAME'] = child.get_direction_based_reference_name(language)

                if child is last_parameter:
                    parameter_signature_template['LAST_ELEMENT'] = True

                operation_signature_template['PARAMETERS'] = str(parameter_signature_template)


        return str(operation_signature_template)

    def get_c_conformance_test(self, template):
        parameters = list()

        last_element = None
        for n in reversed(self.children):
            if isinstance(n, ParameterNode):
                last_element = n
                break

        """
        CR 477: Invokes method under test instead of setting to function pointers for FACE_HMFM.idl.
                NOTE: IOS and TSS are not affected.
        """
        if 'HMFM.idl' in self.idl_filename:
            parameters_instantiation_list = list()
            for n in self.children:
                if isinstance(n, ParameterNode):
                    parameter_declaration = template.getInstanceOf('parameterDeclaration_2')
                    parameter_instantiation = template.getInstanceOf('parameterInstantiation')
                    if n.direction is not None:
                        parameter_declaration['DIRECTION'] = n.get_equal_spaced_direction()

                    parameter_declaration['NAME'] = n.name
                    parameter_instantiation['TYPE_NAME'] = n.get_type_prefix(_C) + n.get_fully_specified_type_name(_C)
                    parameter_instantiation['NAME'] = n.get_direction_based_reference_name(_C)

                    if n is last_element:
                        parameter_declaration['LAST_ELEMENT'] = True
                    parameters.append(str(parameter_declaration))
                    parameters_instantiation_list.append(str(parameter_instantiation))

            function_signature_test = template.getInstanceOf('functionSignatureTest_2')
            function_signature_test['FUNCTION_NAME'] = self.get_fully_specified_name(_C)
            function_signature_test['PARAMETERS'] = '\n'.join(parameters)
            function_signature_test['PARAMETERS_INSTANTIATION'] = '\n'.join(parameters_instantiation_list)

            return [str(function_signature_test)]
        else:
            for n in self.children:
                if isinstance(n, ParameterNode):
                    parameter_declaration = template.getInstanceOf('parameterDeclaration')
                    if n.direction is not None:
                        parameter_declaration['DIRECTION'] = n.get_equal_spaced_direction()

                    if n.type_name == 'Read_Callback': # SPECIAL CASE
                        parameter_declaration['TYPE_NAME'] = 'CONFORMANCE_TEST_send_event_ptr'
                        parameter_declaration['NAME'] = n.name
                    else:
                        parameter_declaration['TYPE_NAME'] = n.get_type_prefix(_C) + n.get_fully_specified_type_name(_C)
                        parameter_declaration['NAME'] = n.get_direction_based_reference_name(_C)
                    if n is last_element:
                        parameter_declaration['LAST_ELEMENT'] = True
                    parameters.append(str(parameter_declaration))

            function_signature_test = template.getInstanceOf('functionSignatureTest')
            function_signature_test['RETURN_TYPE'] = self.return_type.replace('extern', '')
            function_signature_test['FUNCTION_NAME'] = self.get_fully_specified_name(_C)
            function_signature_test['PARAMETERS'] = '\n'.join(parameters)

            return [str(function_signature_test)]

    def get_cpp_conformance_test(self, template):
        parameters = list()

        last_element = None
        for n in reversed(self.children):
            if isinstance(n, ParameterNode):
                last_element = n
                break

        """
        CR 477: Invokes method under test instead of setting to function pointers for FACE_HMFM.idl.
                NOTE: IOS and TSS are not affected.
        """
        if 'HMFM.idl' in self.idl_filename:
            parameters_instantiation_list = list()
            for n in self.children:
                if isinstance(n, ParameterNode):
                    parameter_declaration = template.getInstanceOf('parameterDeclaration_2')
                    parameter_instantiation = template.getInstanceOf('parameterInstantiation')
                    if n.direction is not None:
                        parameter_declaration['DIRECTION'] = n.get_equal_spaced_direction()

                    parameter_declaration['NAME'] = n.name
                    parameter_instantiation['TYPE_NAME'] = n.get_type_prefix(_CPP) + n.get_fully_specified_type_name(_CPP)
                    parameter_instantiation['NAME'] = n.name

                    if n is last_element:
                        parameter_declaration['LAST_ELEMENT'] = True
                    parameters.append(str(parameter_declaration))
                    parameters_instantiation_list.append(str(parameter_instantiation))

            function_signature_test = template.getInstanceOf('functionSignatureTest_2')
            function_signature_test['FUNCTION_NAME'] = self.get_fully_specified_name(_CPP)
            function_signature_test['PARAMETERS'] = '\n'.join(parameters)
            function_signature_test['PARAMETERS_INSTANTIATION'] = '\n'.join(parameters_instantiation_list)

            return [str(function_signature_test)]

        else:
            for n in self.children:
                if isinstance(n, ParameterNode):
                    parameter_declaration = template.getInstanceOf('parameterDeclaration')
                    if n.direction is not None:
                        parameter_declaration['DIRECTION'] = n.get_equal_spaced_direction()

                    if n.type_name == 'Read_Callback': # SPECIAL CASE
                        parameter_declaration['TYPE_NAME'] = 'CONFORMANCE_TEST_send_event_ptr'
                        parameter_declaration['NAME'] = n.name

                    else:
                        parameter_declaration['TYPE_NAME'] = n.get_type_prefix(_CPP) + n.get_fully_specified_type_name(_CPP)
                        parameter_declaration['NAME'] = n.get_direction_based_reference_name(_CPP)
                    if n is last_element:
                        parameter_declaration['LAST_ELEMENT'] = True
                    parameters.append(str(parameter_declaration))

            function_signature_test = template.getInstanceOf('functionSignatureTest')
            function_signature_test['RETURN_TYPE'] = self.return_type.replace('extern', '')
            function_signature_test['FUNCTION_NAME'] = self.get_fully_specified_name(_CPP)
            function_signature_test['PARAMETERS'] = '\n'.join(parameters)

            return [str(function_signature_test)]

    def get_ada_declaration(self, template):
        logger.info(locals())
        language = _ADA

        if self.name == 'send_event':  # VERY SPECIAL CASE
            operation_signature_template = template.getInstanceOf('accessProcedureDeclaration')
            operation_signature_template['NAME'] = \
                TestConfiguration.Language.format_case(language, self.name + '_Access')
            operation_signature_template['PARAMETERS'] = self.get_ada_parameters_declaration(template)

            package_template_instance = template.getInstanceOf('packageDeclaration')
            package_template_instance['NAME'] = 'Read_Callback'
            package_template_instance['BODY'] = str(operation_signature_template)
            return str(package_template_instance)

        operation_signature_template = template.getInstanceOf('operationDeclaration')
        operation_signature_template['NAME'] = TestConfiguration.Language.format_case(language, self.name)
        operation_signature_template['PARAMETERS'] = self.get_ada_parameters_declaration(template)

        return str(operation_signature_template)

    def get_ada_parameters_declaration(self, template):
        language = _ADA

        parameter_list = list()
        max_name_size = 0
        last_parameter = None
        for child in reversed(self.children):
            if isinstance(child, ParameterNode):
                if last_parameter is None:
                    last_parameter = child
                if len(child.name) > max_name_size:
                    max_name_size = len(child.name)

        for child in self.children:
            if not isinstance(child, ParameterNode):
                    continue

            parameter_signature_template = template.getInstanceOf('parameterDeclaration')
            parameter_signature_template['DIRECTION'] = child.get_equal_spaced_direction()
            type_name = child.get_fully_specified_type_name(language)

            if child.name == 'data_callback':  # VERY SPECIAL CASE
                parameter_signature_template['TYPE_NAME'] = 'Read_Callback.Send_Event_Access'
                parameter_signature_template['NAME'] = child.name + (max_name_size + 1 - len(child.name)) * ' '
            else:
                parameter_signature_template['TYPE_NAME'] = type_name
                parameter_signature_template['NAME'] = child.name + (max_name_size + 1 - len(child.name)) * ' '

            if child is last_parameter:
                parameter_signature_template['LAST_ELEMENT'] = True

            parameter_list.append(str(parameter_signature_template))
            logger.debug(parameter_list[-1])

        return '\n'.join(parameter_list)

    def get_ada_parameters_implementation(self, template):
        language = _ADA

        parameter_list = list()
        max_name_size = 0
        for child in self.children:
            if isinstance(child, ParameterNode):
                if len(child.name) > max_name_size:
                    max_name_size = len(child.name)

        for child in self.children:
            if not isinstance(child, ParameterNode):
                    continue

            parameter_signature_template = template.getInstanceOf('parameterImplementation')
            type_name = child.get_fully_specified_type_name(language)

            if child.name == 'data_callback':  # VERY SPECIAL CASE
                parameter_signature_template['TYPE_NAME'] = 'Read_Callback.Send_Event_Access'
                parameter_signature_template['NAME'] = child.name + (max_name_size + 1 - len(child.name)) * ' '
                parameter_signature_template['DO_NOT_INIT'] = True
            else:
                if type_name in ['System.Address', 'FACE.SYSTEM_ADDRESS_TYPE', 'FACE.HMFM.FAULT_MESSAGE_ADDRESS_TYPE', 'FACE.HMFM.FAULT_HANDLER_ENTRY_TYPE', 'FACE.IO.INTERFACE_HANDLE_TYPE', 'FACE.IO.CALLBACK_ADDRESS_TYPE', 'FACE.IO.MESSAGE_ADDR_TYPE' ]:
                    parameter_signature_template['IS_NULL'] = True 
                    if type_name in ['System.Address', 'FACE.SYSTEM_ADDRESS_TYPE', 'FACE.HMFM.FAULT_MESSAGE_ADDRESS_TYPE', 'FACE.IO.INTERFACE_HANDLE_TYPE', 'FACE.IO.MESSAGE_ADDR_TYPE']:
                        parameter_signature_template['DEFAULT'] = type_name + '(System.Null_Address)'
                    if type_name in ['FACE.HMFM.FAULT_HANDLER_ENTRY_TYPE', 'FACE.IO.CALLBACK_ADDRESS_TYPE']:
                        parameter_signature_template['DEFAULT'] = 'NULL'
                if type_name in ['FACE.TS.TRANSPORT_CONNECTION_STATUS_TYPE', 'FACE.HMFM.FAULT_STATUS_TYPE', 'FACE.IO.STATUS_TYPE']:
                    parameter_signature_template['DO_NOT_INIT'] = True
                if type_name in ['Standard.String', 'FACE.HMFM.FAULT_MESSAGE_TYPE', 'FACE.CONFIGURATION_RESOURCE', 'FACE.TS.CONNECTION_NAME_TYPE', 'FACE.TS.WAITSET_TYPE', 'FACE.IO.INTERFACE_NAME_TYPE']:
                    parameter_signature_template['IS_ARRAY'] = True
                    if type_name in ['Standard.String', 'FACE.HMFM.FAULT_MESSAGE_TYPE', 'FACE.CONFIGURATION_RESOURCE', 'FACE.TS.CONNECTION_NAME_TYPE', 'FACE.IO.INTERFACE_NAME_TYPE']:
                        parameter_signature_template['DEFAULT'] = "'X'"
                    if type_name in ['FACE.TS.WAITSET_TYPE']:
                        parameter_signature_template['DEFAULT'] = "FACE_Basic_Types.Boolean(True)"
                    
                parameter_signature_template['TYPE_NAME'] = type_name
                parameter_signature_template['NAME'] = child.name + (max_name_size + 1 - len(child.name)) * ' '

            parameter_list.append(str(parameter_signature_template))
            logger.debug(parameter_list[-1])

        return '\n'.join(parameter_list)

    def get_ada_parameters_assignment(self, template):
        language = _ADA

        last_parameter = None
        for child in reversed(self.children):
            if isinstance(child, ParameterNode):
                if last_parameter is None:
                    last_parameter = child

        parameter_list = list()

        for child in self.children:
            if not isinstance(child, ParameterNode):
                    continue

            parameter_signature_template = template.getInstanceOf('parameterAssignment')
            parameter_signature_template['NAME'] = child.name

            if child is last_parameter:
                parameter_signature_template['LAST_ELEMENT'] = True

            parameter_list.append(str(parameter_signature_template))

        return '\n'.join(parameter_list)

    def get_ada_conformance_test(self, template):
        language = _ADA
        section1_template = template.getInstanceOf('conformanceTestCodeSection1Operation')
        section2_template = template.getInstanceOf('conformanceTestCodeSection2Operation')

        # Operation Names are FACE.TS.TS.Name in Ada not FACE.TS.Name (magic change)
        fully_specified_name = self.get_fully_specified_name(language)
        if constants.getFACEStandardVersion() != constants.FACE_STANDARD_VERSION_2_0:
            fully_specified_name = fully_specified_name.replace('.TS.', '.TS.TS.')
        section1_template['NAME'] = self.name

        section1_template['FULLY_SPECIFIED_NAME'] = fully_specified_name

        section2_template['NAME'] = self.name
        section2_template['FULLY_SPECIFIED_NAME'] = fully_specified_name

        parameter_code =  self.get_ada_parameters_declaration(template)
        if len(parameter_code) > 0:
            section1_template['SECTION1_PARAMETER_ITEM_CODE'] = parameter_code

        parameter_code = self.get_ada_parameters_implementation(template)
        if len(parameter_code) > 0:
            section1_template['SECTION1A_PARAMETER_ITEM_CODE'] = parameter_code

        parameter_code = self.get_ada_parameters_assignment(template)
        if len(parameter_code) > 0:
            section2_template['SECTION2_PARAMETER_ITEM_CODE'] = parameter_code

        if self.return_type is not None and len(self.return_type) > 0:
            section1_template['RETURN_TYPE'] = self.return_type
            section2_template['RETURN_TYPE'] = self.return_type

        return [str(section1_template), str(section2_template)]

    def get_ada_implementation(self, template):
        logger.info(locals())
        language = _ADA
        operation_implementation_template = template.getInstanceOf('operationImplementation')

        operation_implementation_template['OPERATION_NAME'] = TestConfiguration.Language.format_case(language, self.name)
        operation_implementation_template['PARAMETERS'] = self.get_ada_parameters_declaration(template)

        return str(operation_implementation_template)


class ParameterNode(Node):
    def __init__(self, loc, tokens, **kw):
        Node.__init__(self, loc=loc, **kw)

        if len(tokens) != 1 or len(tokens[0]) < 3 or \
                (tokens[0][0] != 'in' and tokens[0][0] != 'out' and tokens[0][0] != 'inout'):
            raise RuntimeError("Unexpected tokens, expected [['in/out', 'PARAM_TYPE', 'PARAM_NAME', ...]], got " + str(tokens))

        self.direction = tokens[0][0]
        self.type_name = tokens[0][1]
        self.name = tokens[0][2]
        self.const_type_names = ['CONFIGURATION_RESOURCE', 'WAITSET_TYPE', 'CONNECTION_NAME_TYPE',
                                 'INTERFACE_NAME_TYPE', 'CONFIGURATION_FILE_NAME']
        # declaration, implementation, and conformance tests handled in OperationNode
        self.declaration_node = False
        self.implementation_node = False
        self.conformance_test_node = False

    def get_type_prefix(self, language):
        type_prefix = ''

        if language is TestConfiguration.Language.CPP and \
                self.direction == 'in' and \
                self.type_name in self.const_type_names:
            type_prefix = 'const '

        return type_prefix

    def get_equal_spaced_direction(self):
        direction_dict = {'in': 'in    ', 'out': '   out', 'inout': 'in out'}
        return direction_dict[self.direction]

    def get_direction_based_reference_name(self, language):

        if language == TestConfiguration.Language.C:
            if self.direction == 'in':
                return self.name
            else:
                return '*' + self.name

        elif language == TestConfiguration.Language.CPP:
            if self.direction == 'in':
                return self.name
            elif self.get_fully_specified_type_name(language).strip() == 'void':  # void is a special case
                return '*' + self.name
            else:
                return '&' + self.name

        elif language == TestConfiguration.Language.Ada:
            return self.name

        else:
            return 'Do not know how to do direction based reference name ' + self.name + ' in ' + language


class EnumerationNode(Node):

    def __init__(self, loc, tokens, **kw):
        Node.__init__(self, loc=loc, **kw)
        if len(tokens) < 2 or tokens[0] != 'enum':
            raise RuntimeError("Unexpected tokens, expected ['enum','ENUM_NAME', ...]")
        self.name = tokens[1]
        self.conformance_test_node = True
        self.in_namespace = False
        self.declaration_node = True # must generate code for definition header
        self.implementation_node = False # Do not need to implement an enumeration
        self.conformance_test_node = True # must generate code for conformance tests

    def get_c_declaration(self, template):
        return self._get_enumeration_declaration(_C, template)

    def get_cpp_declaration(self, template):
        return self._get_enumeration_declaration(_CPP, template)

    def get_ada_declaration(self, template):
        return self._get_enumeration_declaration(_ADA, template)


    def _get_enumeration_declaration(self, language, template):
        """
        Enumeration is a rare type that uses the exact same code for C,C++, and Ada
        :param language: TestConfiguration.Language.C, CPP or Ada,
        :param template: Associated template to use (will vary based on language and whether generating for a
            conformance test or a gold standard library)
        :return:the string containing the declaration code in the language specified
        """

        enumeration_template = template.getInstanceOf('enumerationDeclaration')
        enumeration_template['NAME'] = self.get_declaration_name(language)

        last_item = None
        for c in reversed(self.children):
            if isinstance(c, EnumerationItemNode):
                last_item = c
                break

        for c in self.children:
            if isinstance(c, EnumerationItemNode):
                enumeration_item_template = template.getInstanceOf('enumerationListDeclaration')
                if language == _C:
                    enumeration_item_template['NAME'] = c.get_fully_specified_name(language)
                else: # C++ doesn't need fully specified name since it is namespaced
                    enumeration_item_template['NAME'] = c.name
                if c is last_item:
                    enumeration_item_template['LAST_ELEMENT'] = True
                enumeration_template['ENUMERATION_LIST'] = str(enumeration_item_template)

        return str(enumeration_template)

    def get_c_conformance_test(self, template):
        return self.get_c_or_cpp_conformance_test(_C, template)

    def get_cpp_conformance_test(self, template):
        return self.get_c_or_cpp_conformance_test(_CPP, template)

    def get_c_or_cpp_conformance_test(self, language, template):
        section_template = template.getInstanceOf('enumerationTest')
        section_template['ENUMERATION_NAME'] = self.get_fully_specified_name(language)
        if language == _CPP:
            section_template['CONFORMANCE_TEST_ENUMERATION_NAME'] = self.get_declaration_name(language)

        section_template['CONFORMANCE_TEST_ENUMERATION'] = self.get_declaration(language, template)
        enumeration_item_code_list = list()
        enumeration_order_list = list()
        previous_child = None
        for enumeration_item_node in self.children:
            if not isinstance(enumeration_item_node, EnumerationItemNode):
                continue

            enumeration_exist_template = template.getInstanceOf('enumerationItemsExistTest')
            enumeration_order_template = template.getInstanceOf('enumerationOrderTest')

            enumeration_exist_template['ENUMERATION_ITEM'] = enumeration_item_node.get_fully_specified_name(language)
            enumeration_item_code_list.append(str(enumeration_exist_template))
            if previous_child is not None:
                enumeration_order_template['ENUMERATION_ITEM1'] = previous_child.get_fully_specified_name(language)
                enumeration_order_template['ENUMERATION_ITEM2'] = enumeration_item_node.get_declaration_name(language)
                enumeration_order_list.append(str(enumeration_order_template))
            previous_child = enumeration_item_node

        section_template['ENUMERATION_ITEM_EXIST_TESTS'] = '\n'.join(enumeration_item_code_list)

        # For now take this out, since it may not be a conformance issue (ordering may not be guarenteed
        #section_template['ENUMERATION_ORDER_TESTS'] = '\n'.join(enumeration_order_list)

        return [str(section_template)]

    def get_ada_conformance_test(self, template):
        language = _ADA
        section1_template = template.getInstanceOf('conformanceTestCodeSection1Enumeration')

        section1_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)
        section1_template['NAME'] = self.name

        enumeration_item_code_list = list()
        for child in self.children:
            if not isinstance(child, EnumerationItemNode):
                continue

            enumerated_item_template = template.getInstanceOf('conformanceTestCodeEnumeratedItem')
            enumerated_item_template['FULLY_SPECIFIED_NAME'] = child.get_fully_specified_name(language)
            enumeration_item_code_list.append(str(enumerated_item_template))

        section2_template = template.getInstanceOf('conformanceTestCodeSection2Enumeration')
        section2_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)
        section2_template['ENUMERATED_ITEM_CODE'] = '\n'.join(enumeration_item_code_list)


        return [self.get_ada_declaration(template) + '\n' + str(section1_template), str(section2_template)]


class EnumerationItemNode(Node):
    def __init__(self, loc, tokens, **kw):
        Node.__init__(self, loc=loc, **kw)
        if len(tokens) < 1:
            raise RuntimeError("Unexpected tokens, expected ['ENUM_ITEM_NAME', ...]")

        self.name = tokens[0]
        # declaration/implementation/conformance test handled in Enumeration Node
        self.declaration_node = False
        self.implementation_node = False
        self.conformance_test_node = False


class StructNode(Node):

    def __init__(self, loc, tokens, **kw):
        Node.__init__(self, loc=loc, **kw)
        if len(tokens) != 1 or len(tokens[0]) < 2 or tokens[0][0] != 'struct':
            raise RuntimeError("Unexpected tokens, expected [['struct', 'STRUCT_NAME',...]]")

        self.name = tokens[0][1]
        self.conformance_test_node = True  # Need to generate a conformance test for this type of node
        self.declaration_node = True # must generate code for definition header
        self.implementation_node = False # Does not need to be implemented

    def get_c_declaration(self, template):
        return self.get_struct_declaration(_C, template)

    def get_cpp_declaration(self, template):
        return self.get_struct_declaration(_CPP, template)

    def get_ada_declaration(self, template):
        return self.get_struct_declaration(_ADA, template)

    def get_struct_declaration(self, language, template):
        struct_template = template.getInstanceOf('structDeclaration')
        struct_template['NAME'] = self.get_declaration_name(language)

        max_name_length = 0
        if language == _ADA:
            for c in self.children:
                if isinstance(c, StructItemNode) and max_name_length < len(c.name):
                    max_name_length = len(c.name)
        max_name_length += 1

        for c in self.children:
            if isinstance(c, StructItemNode):
                struct_item_template = template.getInstanceOf('structItemDeclaration')
                if language == _ADA:
                    struct_item_template['NAME'] = (max_name_length - len(c.name)) * ' ' + c.name
                    type_name = c.get_fully_specified_type_name(language)
                    if 'System.Address' in type_name or 'FACE.SYSTEM_ADDRESS_TYPE' in type_name:
                        struct_item_template['IS_SYS_ADDR'] = True                    
                    if 'Standard.String' in type_name or 'FACE.HMFM.FAULT_MESSAGE_TYPE' in type_name:
                        struct_item_template['IS_STRING'] = True                    
                else:
                    struct_item_template['NAME'] = c.name
                struct_item_template['TYPE_NAME'] = c.get_fully_specified_type_name(language)
                struct_template['STRUCT_ITEM_LIST'] = str(struct_item_template)

        return str(struct_template)

    def get_c_conformance_test(self, template):
        return self.get_c_or_cpp_conformance_test(_C, template)

    def get_cpp_conformance_test(self, template):
        return self.get_c_or_cpp_conformance_test(_CPP, template)

    def get_c_or_cpp_conformance_test(self, language, template):
        section_template = template.getInstanceOf('structTest')

        struct_name = self.get_fully_specified_name(language)
        section_template['NAME'] = struct_name
        if language == _CPP:
            section_template['CONFORMANCE_TEST_NAME'] = self.get_declaration_name(language)

        section_template['CONFORMANCE_TEST_STRUCT'] = self.get_declaration(language, template)

        struct_item_tests = list()
        for struct_item_node in self.children:
            if isinstance(struct_item_node, StructItemNode):
                struct_item_template = template.getInstanceOf('structItemTest')
                struct_item_template['STRUCT_NAME'] = struct_name
                struct_item_template['ITEM_NAME'] = struct_item_node.name
                struct_item_template['TYPE'] = struct_item_node.get_fully_specified_type_name(language)
                if struct_item_node.type_name == 'FAULT_MESSAGE_TYPE':  # VERY SPECIAL CASE...already an array
                    struct_item_template['NO_ASSIGNMENT'] = True  # can't do an assignment test
                struct_item_tests.append(str(struct_item_template))

        section_template['STRUCT_ITEM_TESTS'] = '\n'.join(struct_item_tests)

        return [str(section_template)]

    def get_ada_conformance_test(self, template):
        language = _ADA
        section1_template = template.getInstanceOf('conformanceTestCodeSection1Struct')

        section1_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)
        section1_template['NAME'] = self.name

        section2_template = template.getInstanceOf('conformanceTestCodeSection2Struct')
        section2_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)

        for child in self.children:
            if not isinstance(child, StructItemNode):
                continue

            struct_item_section1_template = template.getInstanceOf('conformanceTestCodeSection1StructItem')
            struct_item_section1_template['NAME'] = child.name
            struct_item_section1_template['TYPE_NAME'] = child.get_fully_specified_type_name(language)

            struct_item_section2_template = template.getInstanceOf('conformanceTestCodeSection2StructItem')
            struct_item_section2_template['NAME'] = child.name
            struct_item_section2_template['TYPE_NAME'] = child.get_fully_specified_type_name(language)

            section1_template['SECTION1_STRUCT_ITEM_CODE'] = str(struct_item_section1_template)
            section2_template['SECTION2_STRUCT_ITEM_CODE'] = str(struct_item_section2_template)


        return [self.get_ada_declaration(template) + str(section1_template), str(section2_template)]


class StructItemNode(Node):
    def __init__(self, loc, tokens, **kw):
        Node.__init__(self, loc=loc, **kw)
        if len(tokens) < 2:
            raise RuntimeError("Unexpected tokens, expected ['VARIABLE_TYPE', 'VARIABLE_NAME',...]")

        self.name = tokens[1]
        self.type_name = tokens[0]

        # declaration/implementation handled in StructNode
        self.declaration_node = False
        self.implementation_node = False
        self.conformance_test_node = False


class TypedefNode(Node):
    def __init__(self, loc, tokens, **kw):
        Node.__init__(self, loc=loc, **kw)

        if len(tokens) != 1 or len(tokens[0]) < 4 or tokens[0][0] != 'typedef' or tokens[0][-1] != ';':
            raise RuntimeError("Unexpected tokens, expected [['typedef','TYPE_NAME', 'NAME', ';']] got " + str(tokens))

        self.type_name = ' '.join(tokens[0][1:-2])
        self.name = tokens[0][-2]
        self.conformance_test_node = True  # Need to generate a conformance test for this type of node
        self.declaration_node = True  # must generate code for definition header

    def get_c_declaration(self, template):
        if self.name == 'CALLBACK_ADDRESS_TYPE':  # VERY SPECIAL CASE
            function_template = template.getInstanceOf('functionDeclaration')
            function_template['RETURN_TYPE'] = 'typedef void'
            function_template['FUNCTION_NAME'] = '(*FACE_CALLBACK_ADDRESS_TYPE)'
            parameters = [['FACE_MESSAGE_LENGTH_TYPE', 'message_length'],
                          ['FACE_MESSAGE_ADDR_TYPE', 'data_buffer_address'],
                          ['FACE_RETURN_CODE_TYPE', '*return_code']]

            for p in parameters:
                parameter_signature_template = template.getInstanceOf('parameterDeclaration')
                if not p is parameters[-1]:
                    parameter_signature_template['DIRECTION'] = 'in   '
                parameter_signature_template['TYPE_NAME'] = p[0]
                parameter_signature_template['NAME'] = p[1]
                if p is parameters[-1]:
                    parameter_signature_template['LAST_ELEMENT'] = True
                    parameter_signature_template['DIRECTION'] = '  out'

                function_template['PARAMETERS'] = str(parameter_signature_template)
            return str(function_template)

        elif self.name == 'FAULT_MESSAGE_ADDRESS_TYPE':
            typedef_template = template.getInstanceOf('typedefDeclaration')
            typedef_template['NAME'] = self.get_declaration_name(_C)
            typedef_template['TYPE_NAME'] = 'void *'
            return str(typedef_template)


        elif self.name == 'FAULT_HANDLER_ENTRY_TYPE':  # VERY SPECIAL CASE

            typedef_template = template.getInstanceOf('typedefDeclaration')

            typedef_template['NAME'] = '(*' + self.get_fully_specified_name(_C) + ') (void)'
            typedef_template['TYPE_NAME'] = 'void'
            return str(typedef_template)

        elif self.name == 'SYSTEM_ADDRESS_TYPE':
            typedef_template = template.getInstanceOf('typedefDeclaration')

            typedef_template['NAME'] = '*' + self.get_fully_specified_name(_C)
            typedef_template['TYPE_NAME'] = 'void'
            return str(typedef_template)

        else:  # not a special case
            typedef_template = template.getInstanceOf('typedefDeclaration')

            typedef_template['NAME'] = self.get_fully_specified_name(_C)
            typedef_template['TYPE_NAME'] = self.get_fully_specified_type_name(_C)
            return str(typedef_template)

    def get_c_conformance_test(self, template):
        return self.get_c_or_cpp_conformance_test(_C, template)

    def get_cpp_conformance_test(self, template):
        return self.get_c_or_cpp_conformance_test(_CPP, template)

    def get_c_or_cpp_conformance_test(self, language, template):


        if self.name == 'CALLBACK_ADDRESS_TYPE':  # VERY SPECIAL CASE
            function_signature_test = template.getInstanceOf('functionSignatureTest')
            function_signature_test['RETURN_TYPE'] = 'void'
            function_signature_test['FUNCTION_NAME'] = '*' + self.get_fully_specified_name(language)

            parameters = [['FACE_MESSAGE_LENGTH_TYPE', 'message_length'],
                          ['FACE_MESSAGE_ADDR_TYPE', 'data_buffer_address'],
                          ['FACE_RETURN_CODE_TYPE', '*return_code']]
            if language == _CPP:
                parameters = [['FACE::MESSAGE_LENGTH_TYPE', 'message_length'],
                              ['FACE::MESSAGE_ADDR_TYPE', 'data_buffer_address'],
                              ['FACE::RETURN_CODE_TYPE', '&return_code']]

            for p in parameters:
                parameter_signature_template = template.getInstanceOf('parameterDeclaration')
                if not p is parameters[-1]:
                    parameter_signature_template['DIRECTION'] = 'in   '
                parameter_signature_template['TYPE_NAME'] = p[0]
                parameter_signature_template['NAME'] = p[1]
                if p is parameters[-1]:
                    parameter_signature_template['LAST_ELEMENT'] = True
                    parameter_signature_template['DIRECTION'] = '  out'

                function_signature_test['PARAMETERS'] = str(parameter_signature_template) + '\n'

            return [str(function_signature_test)]


        section_template = template.getInstanceOf('typedefTest')
        section_template['NAME'] = self.get_fully_specified_name(language)

        typedef_name = self.get_fully_specified_type_name(language)
        if typedef_name.strip() == 'void':
            # use void since "in" parameters would automatically add the * to them, here we need to add the *
            typedef_name = 'void *'
        section_template['TYPEDEF'] = typedef_name

        return [str(section_template)]

    def get_cpp_declaration(self, template):
        if self.name == 'CALLBACK_ADDRESS_TYPE':  # VERY SPECIAL CASE
            function_template = template.getInstanceOf('functionDeclaration')
            function_template['RETURN_TYPE'] = 'typedef void'
            function_template['FUNCTION_NAME'] = '(*CALLBACK_ADDRESS_TYPE)'
            parameters = [['FACE::MESSAGE_LENGTH_TYPE', 'message_length'],
                          ['FACE::MESSAGE_ADDR_TYPE', 'data_buffer_address'],
                          ['FACE::RETURN_CODE_TYPE', '&return_code']]

            for p in parameters:
                parameter_signature_template = template.getInstanceOf('parameterDeclaration')
                if not p is parameters[-1]:
                    parameter_signature_template['DIRECTION'] = 'in   '
                parameter_signature_template['TYPE_NAME'] = p[0]
                parameter_signature_template['NAME'] = p[1]
                if p is parameters[-1]:
                    parameter_signature_template['LAST_ELEMENT'] = True
                    parameter_signature_template['DIRECTION'] = '  out'

                function_template['PARAMETERS'] = str(parameter_signature_template)
            return str(function_template)

        elif self.name == 'FAULT_MESSAGE_ADDRESS_TYPE':
            typedef_template = template.getInstanceOf('typedefDeclaration')
            typedef_template['NAME'] = self.get_declaration_name(_CPP)
            typedef_template['TYPE_NAME'] = 'void *'
            return str(typedef_template)

        elif self.name == 'FAULT_HANDLER_ENTRY_TYPE':  # VERY SPECIAL CASE

            typedef_template = template.getInstanceOf('typedefDeclaration')
            typedef_template['NAME'] = '(*' + self.name + ') (void)'
            typedef_template['TYPE_NAME'] = 'void'
            return str(typedef_template)

        elif self.name == 'SYSTEM_ADDRESS_TYPE':
            typedef_template = template.getInstanceOf('typedefDeclaration')

            typedef_template['NAME'] = '*' + self.name
            typedef_template['TYPE_NAME'] = 'void'
            return str(typedef_template)

        else:  # not a special case
            typedef_template = template.getInstanceOf('typedefDeclaration')

            typedef_template['NAME'] = self.get_declaration_name(_CPP)
            typedef_template['TYPE_NAME'] = self.get_fully_specified_type_name(_CPP)
            return str(typedef_template)

    def get_ada_declaration(self, template):
        if self.name == 'CALLBACK_ADDRESS_TYPE':  # VERY SPECIAL CASE
            procedure_template = template.getInstanceOf('accessProcedureDeclaration')

            procedure_template['NAME'] = self.name
            procedure_template['INCLUDE_PRAGMA'] = True
            parameter_list = [('message_length', 'in   ', 'MESSAGE_LENGTH_TYPE'),
                ('data_buffer_address', 'in   ', 'MESSAGE_ADDR_TYPE'),
                ('return_code','  out', 'RETURN_CODE_TYPE')]
            max_parameter_length = len(parameter_list[1][0]) + 1
            for p in parameter_list:
                parameter_signature_template = template.getInstanceOf('parameterDeclaration')
                parameter_signature_template['NAME'] = p[0] + (max_parameter_length - len(p[0])) * ' '
                parameter_signature_template['DIRECTION'] = p[1]
                parameter_signature_template['TYPE_NAME'] = p[2]
                if p is parameter_list[-1]:
                    parameter_signature_template['LAST_ELEMENT'] = True
                procedure_template['PARAMETERS'] = str(parameter_signature_template)

            return str(procedure_template)
        elif self.name == 'FAULT_MESSAGE_ADDRESS_TYPE':
            typedef_template = template.getInstanceOf('typedefDeclaration')
            typedef_template['NAME'] = self.get_declaration_name(_CPP)
            typedef_template['TYPE_NAME'] = 'System.Address'
            return str(typedef_template)

        elif self.name == 'FAULT_HANDLER_ENTRY_TYPE':  # VERY SPECIAL CASE

            access_procedure_template = template.getInstanceOf('accessProcedureDeclaration')
            access_procedure_template['NAME'] = self.name
            access_procedure_template['INCLUDE_PRAGMA'] = True
            return str(access_procedure_template)

        elif self.name == 'SYSTEM_ADDRESS_TYPE':
            typedef_template = template.getInstanceOf('subTypedefDeclaration')

            typedef_template['NAME'] = self.name
            typedef_template['TYPE_NAME'] = 'System.Address'
            return str(typedef_template)

        else:  # not a special case
            typedef_template = template.getInstanceOf('typedefDeclaration')
            if not self.type_name in TestConfiguration.Language.type_name_dict[_ADA].keys():
                typedef_template = template.getInstanceOf('subTypedefDeclaration')


            typedef_template['NAME'] = self.get_declaration_name(_ADA)
            typedef_template['TYPE_NAME'] = self.get_fully_specified_type_name(_ADA)
            return str(typedef_template)

    def get_ada_conformance_test(self, template):
        language = _ADA
        section1A_template = template.getInstanceOf('conformanceTestCodeSection1ATypedef')

        fully_specified_type_name = self.get_fully_specified_type_name(language)
        section1A_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)
        section1A_template['SHORT_NAME'] = self.get_fully_specified_name(language).rsplit('.',1)[-1]
        section1A_template['TYPE_NAME'] = fully_specified_type_name

        section1B_template = template.getInstanceOf('conformanceTestCodeSection1BTypedef')

        section1B_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)
        section1B_template['SHORT_NAME'] = self.get_fully_specified_name(language).rsplit('.',1)[-1]
        section1B_template['TYPE_NAME'] = fully_specified_type_name
        if 'System.Address' not in fully_specified_type_name and 'FACE.SYSTEM_ADDRESS_TYPE' not in fully_specified_type_name:
            section1B_template['IS_INT_TYPE'] = True
			
        section2_template = template.getInstanceOf('conformanceTestCodeSection2Typedef')
        section2_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)
        section2_template['TYPE_NAME'] = self.get_fully_specified_type_name(language)
        section2_template['SHORT_NAME'] = self.get_fully_specified_name(language).rsplit('.',1)[-1]
        section2_template['CONSTANT_DEFINED'] = True
        if 'System.Address' in fully_specified_type_name or 'FACE.SYSTEM_ADDRESS_TYPE' in fully_specified_type_name:
            section2_template['IS_SYS_ADDR'] = True
			

        return [str(section1A_template) + str(section1B_template), str(section2_template)]


class TypedefStringSequenceNode(Node):
    def __init__(self, loc, tokens, **kw):
        Node.__init__(self, loc=loc, **kw)

        if len(tokens) != 1 or len(tokens[0]) < 6 or tokens[0][0] != 'typedef' or tokens[0][1] != 'string':
            raise RuntimeError("Unexpected tokens, expected [['typedef','string', '<' 'SIZE', '>', 'NAME', ';']] " +
                               "got " + str(tokens))

        self.size = tokens[0][3]
        self.name = tokens[0][5]
        self.conformance_test_node = True  # Need to generate a conformance test for this type of node
        self.declaration_node = True # must generate code for definition header

    def get_c_declaration(self, template):
        return self.get_c_or_cpp_declaration(_C, template)

    def get_cpp_declaration(self, template):
        return self.get_c_or_cpp_declaration(_CPP, template)

    def get_c_or_cpp_declaration(self, language, template):
        sequence_template = template.getInstanceOf('typedefArrayDeclaration')

        sequence_template['NAME'] = self.get_declaration_name(language)
        if self.name == 'CONFIGURATION_FILE_NAME':  # MAJOR HACK TO HANDLE 2.0 Standard bug that defines it in both TSS and IOS
            sequence_template['GUARD'] = '1'

        if language == _C:
            sequence_template['TYPE_NAME'] = 'FACE_char'
        else:
            sequence_template['TYPE_NAME'] = 'FACE::Char'

        if self.size.isdigit():
            sequence_template['SIZE'] = self.size
        else:
            sequence_template['SIZE'] = self._find_fully_specified_name(language, self.size)
        return str(sequence_template)

    def get_c_conformance_test(self, template):
        return self.get_c_or_cpp_conformance_test(_C, template)

    def get_cpp_conformance_test(self, template):
        return self.get_c_or_cpp_conformance_test(_CPP, template)

    def get_c_or_cpp_conformance_test(self, language, template):
        section_template = template.getInstanceOf('typedefSequenceTest')

        section_template['NAME'] = self.get_fully_specified_name(language)
        if language == _C:
            section_template['TYPEDEF'] = 'FACE_char'
        else:
            section_template['TYPEDEF'] = 'FACE::Char'
        if self.size.isdigit():
            section_template['SEQUENCE_SIZE'] = self.size
        else:  # VERY SPECIAL CASE...ONLY HAPPENS IN HMFM
            delimiter = TestConfiguration.Language.delimiter[language]
            section_template['SEQUENCE_SIZE'] = 'FACE' + delimiter + 'HMFM' + delimiter + self.size

        return [str(section_template)]

    def get_ada_declaration(self, template):
        typedef_template = template.getInstanceOf('typedefStringSequenceDeclaration')
        language = _ADA
        typedef_template['NAME'] = self.get_declaration_name(language)
        typedef_template['TYPE_NAME'] = 'Standard.String'
        if self.size.isdigit():
            typedef_template['SIZE'] = '1 .. ' + self.size
        else:  # VERY SPECIAL CASE
            typedef_template['SIZE'] = '1 .. Standard.Integer(FACE.HMFM.' + self.size + ')'
        return str(typedef_template)

    def get_ada_conformance_test(self, template):
        language = _ADA

        section1_template = template.getInstanceOf('conformanceTestCodeSection1TypedefStringSequence')

        section1_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)
        section1_template['TYPE_NAME'] = self.get_fully_specified_type_name(language)
        section1_template['NAME'] = self.name
        if self.size.isdigit():
            section1_template['SIZE'] = '1 .. ' + self.size
        else:  # VERY SPECIAL CASE
            section1_template['SIZE'] = '1 .. Standard.Integer(FACE.HMFM.' + self.size + ')'

        section2_template = template.getInstanceOf('conformanceTestCodeSection2TypedefStringSequence')
        section2_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)
        section2_template['TYPE_NAME'] = self.get_fully_specified_type_name(language)
        section2_template['SIZE'] = section1_template['SIZE']
        if self.size.isdigit():
            section2_template['LAST_INDEX'] = self.size
        else:  # VERY SPECIAL CASE
            section2_template['LAST_INDEX'] = 'Standard.Integer(FACE.HMFM.' + self.size + ')'

        return [str(section1_template), str(section2_template)]


class TypedefArrayNode(Node):
    def __init__(self, loc, tokens, **kw):
        Node.__init__(self, loc=loc, **kw)

        if len(tokens) != 1 or len(tokens[0]) < 7 or tokens[0][0] != 'typedef':
            raise RuntimeError("Unexpected tokens, expected [['typedef','TYPE_NAME', 'NAME', '[' 'SIZE', ']',  ';']]" +
                               "got " + str(tokens))

        self.size = tokens[0][4]
        self.name = tokens[0][2]
        self.type_name = tokens[0][1]
        self.conformance_test_node = True  # Need to generate a conformance test for this type of node
        self.declaration_node = True # must generate code for definition header

    def get_c_declaration(self, template):
        return self.get_c_or_cpp_declaration(_C, template)

    def get_cpp_declaration(self, template):
        return self.get_c_or_cpp_declaration(_CPP, template)

    def get_c_or_cpp_declaration(self, language, template):
        typedef_template = template.getInstanceOf('typedefArrayDeclaration')

        typedef_template['NAME'] = self.get_declaration_name(language)
        typedef_template['TYPE_NAME'] = self.get_fully_specified_type_name(language)
        typedef_template['SIZE'] = self.size

        return str(typedef_template)

    def get_c_conformance_test(self, template):
        return self.get_c_or_cpp_conformance_test(_C, template)

    def get_cpp_conformance_test(self, template):
        return self.get_c_or_cpp_conformance_test(_CPP, template)

    def get_c_or_cpp_conformance_test(self, language, template):
        section_template = template.getInstanceOf('typedefSequenceTest')

        section_template['NAME'] = self.get_fully_specified_name(language)
        section_template['TYPEDEF'] = self.get_fully_specified_type_name(language)
        section_template['SEQUENCE_SIZE'] = self.size

        return [str(section_template)]

    def get_ada_declaration(self, template):
        typedef_template = template.getInstanceOf('typedefArrayDeclaration')
        language = _ADA
        typedef_template['NAME'] = self.get_declaration_name(language)
        typedef_template['TYPE_NAME'] = self.get_fully_specified_type_name(language)
        size_value = self.size
        if size_value.isdigit():
            size_value = str(int(size_value) - 1)
            typedef_template['SIZE'] = '0 .. ' + size_value
        else:
            typedef_template['SIZE'] = '0 .. ' + self.size + ' - 1'
        return str(typedef_template)

    def get_ada_conformance_test(self, template):
        language = _ADA
        ada_face_types_defaults = {
            'FACE_Basic_Types.Boolean': 'True'
        }
        section1_template = template.getInstanceOf('conformanceTestCodeSection1TypedefArray')
        section1_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)
        section1_template['TYPE_NAME'] = self.get_fully_specified_type_name(language)
        section1_template['NAME'] = self.name
        size_value = self.size
        if size_value.isdigit():
            size_value = str(int(size_value) - 1)
            section1_template['SIZE'] = '0 .. ' + size_value
        else:
            section1_template['SIZE'] = '0 .. ' + self.size + ' - 1'
 
        fully_specified_type_name = section1_template['TYPE_NAME']
        dot_found = fully_specified_type_name.rfind('.')
        if dot_found > -1:  # there is something to "use"
            section1_template['NAME_USED'] = fully_specified_type_name[:dot_found]

        if self.get_fully_specified_type_name(language) in ada_face_types_defaults:
            section1_template['DEFAULT'] = ada_face_types_defaults[self.get_fully_specified_type_name(language)]
			
        section2_template = template.getInstanceOf('conformanceTestCodeSection2TypedefArray')
        section2_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)
        section2_template['TYPE_NAME'] = self.get_fully_specified_type_name(language)
        section2_template['SIZE'] = self.size
        section2_template['LAST_INDEX'] = str(int(self.size) - 1)
        return [str(section1_template), str(section2_template)]


class ConstantNode(Node):
    def __init__(self, loc, tokens, **kw):
        Node.__init__(self, loc=loc, **kw)

        if len(tokens) != 1 or len(tokens[0]) < 6 or tokens[0][0] != 'const' or tokens[0][5] != ';':
            raise RuntimeError("Unexpected tokens, expected [['const', TYPE_NAME', 'NAME', '=', 'VALUE', ';']]")

        self.type_name = tokens[0][1]
        self.name = tokens[0][2]
        self.value = tokens[0][4]
        self.conformance_test_node = True  # Need to generate a conformance test for this type of node
        self.declaration_node = True  # must generate code for definition header

    def get_c_declaration(self, template):
        return self.get_constant_declaration(_C, template)

    def get_cpp_declaration(self, template):
        return self.get_constant_declaration(_CPP, template)

    def get_ada_declaration(self, template):
        return self.get_constant_declaration(_ADA, template)

    def get_constant_declaration(self, language, template):
        const_template = template.getInstanceOf('constantDeclaration')
        const_template['NAME'] = self.get_declaration_name(language)
        const_template['TYPE_NAME'] = self.get_fully_specified_type_name(language)
        const_template['VALUE'] = self.value

        return str(const_template)

    def get_c_conformance_test(self, template):
        return self.get_c_or_cpp_conformance_test(_C, template)

    def get_cpp_conformance_test(self, template):
        return self.get_c_or_cpp_conformance_test(_CPP, template)

    def get_c_or_cpp_conformance_test(self, language, template):
        section_template = template.getInstanceOf('constantTest')

        section_template['NAME'] = self.get_fully_specified_name(language)
        section_template['TYPE'] = self.get_fully_specified_type_name(language)
        section_template['VALUE'] = self.value

        return [str(section_template)]

    def get_ada_conformance_test(self, template):
        language = _ADA
        section1_template = template.getInstanceOf('conformanceTestCodeSection1Constant')
        section1_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)
        fully_specified_type_name = self.get_fully_specified_type_name(language)
        section1_template['TYPE_NAME'] = fully_specified_type_name
        dot_found = fully_specified_type_name.rfind('.')
        if dot_found > -1:  # there is something to "use"
            section1_template['NAME_USED'] = fully_specified_type_name[:dot_found]

        section1_template['VALUE'] = self.value

        section2_template = template.getInstanceOf('conformanceTestCodeSection2Constant')
        section2_template['FULLY_SPECIFIED_NAME'] = self.get_fully_specified_name(language)
        section2_template['TYPE_NAME'] = self.get_fully_specified_type_name(language)
        section2_template['VALUE'] = self.value

        return [str(section1_template), str(section2_template)]


class InterfaceNode(Node):
    def __init__(self, loc, tokens, **kw):
        Node.__init__(self, loc=loc, **kw)
        if len(tokens) != 1 or tokens[0][0] != 'interface':
            raise RuntimeError("Unexpected tokens, expected [['interface', 'INTERFACE_NAME',...]]")
        self.name = tokens[0][1]


class CommentNode(Node):
    def __init__(self, loc, **kw):
        Node.__init__(self, loc=loc, **kw)
        self.name = "Comment"

    def get_conformance_test_code(self, language, template):
        return None # Comments do not need conformance tests


class CStyleCommentNode(CommentNode):
    def __init__(self, loc, tokens, **kw):
        CommentNode.__init__(self, loc=loc, **kw)

        if len(tokens) != 1 or not tokens[0].startswith('/*') or not tokens[0].endswith('*/'):
            raise RuntimeError("Unexpected tokens, expected ['/* COMMENT */'], got " + str(tokens))

        self.comment = tokens[0].strip('/*')


class SingleLineCommentNode(CommentNode):
    def __init__(self, loc, tokens, **kw):
        CommentNode.__init__(self, loc=loc, **kw)

        if len(tokens) != 2 or not tokens[0].startswith('//'):
            raise RuntimeError("Unexpected tokens, expected ['//', 'COMMENT...'], got " + str(tokens))

        self.comment = tokens[1].strip()


class FaceIdlParser():

    def __init__(self, **kw):
        # class variables
        self.bnf = None
        self.unhandled_definitions = 0
        self.idl_tree = IdlTree()
        self.tokens = None
        self.initialize_parser()

        idl_filename = kw.get('filename', None)

        if idl_filename is not None:
            self.parse_idl_file(idl_filename)

    def parse_idl_file(self, idl_filename):
        print 'idl filename: ' + idl_filename
        if not os.path.isfile(idl_filename):
            raise IOError('Idl File does not exist: ' + idl_filename)

        self.idl_tree.current_idl_filename = idl_filename
        self.idl_directory = os.path.dirname(idl_filename)

        idl_file = open(idl_filename, 'r')
        lines = '\n'.join(idl_file.readlines())
        self.parse_string(lines)
        self.idl_tree.combine_modules()
        self.idl_tree.find_face_specified_type_nodes()
        self.idl_tree.print_tree()
        pass



    def initialize_parser(self):

        # punctuation
        colon  = Literal(":")
        lbrace = Literal("{")
        rbrace = Literal("}")
        lbrack = Literal("[")
        rbrack = Literal("]")
        lparen = Literal("(")
        rparen = Literal(")")
        equals = Literal("=")
        comma  = Literal(",")
        dot    = Literal(".")
        hashtag= Literal("#")
        slash  = Literal("/")
        bslash = Literal("\\")
        star   = Literal("*")
        semi   = Literal(";")
        langle = Literal("<")
        rangle = Literal(">")
        quote  = Literal('"')

        # keywords
        any_       = Keyword("any")
        attribute_ = Keyword("attribute")
        boolean_   = Keyword("boolean")
        case_      = Keyword("case")
        char_      = Keyword("char")
        const_     = Keyword("const")
        context_   = Keyword("context")
        default_   = Keyword("default")
        double_    = Keyword("double")
        enum_      = Keyword("enum")
        exception_ = Keyword("exception")
        false_     = Keyword("FALSE")
        fixed_     = Keyword("fixed")
        float_     = Keyword("float")
        inout_     = Keyword("inout")
        interface_ = Keyword("interface")
        in_        = Keyword("in")
        include_   = Keyword("include")
        long_      = Keyword("long")
        long_long_ = Keyword("long long")
        module_    = Keyword("module")
        object_    = Keyword("Object")
        octet_     = Keyword("octet")
        oneway_    = Keyword("oneway")
        out_       = Keyword("out")
        raises_    = Keyword("raises")
        readonly_  = Keyword("readonly")
        sequence_  = Keyword("sequence")
        short_     = Keyword("short")
        string_    = Keyword("string")
        struct_    = Keyword("struct")
        switch_    = Keyword("switch")
        true_      = Keyword("TRUE")
        typedef_   = Keyword("typedef")
        unsigned_  = Keyword("unsigned")
        union_     = Keyword("union")
        void_      = Keyword("void")
        wchar_     = Keyword("wchar")
        wstring_   = Keyword("wstring")

        identifier = Word( alphas, alphanums + "_" ).setName("identifier")

        real = Combine( Word(nums+"+-", nums) + dot + Optional( Word(nums) )
                        + Optional( CaselessLiteral("E") + Word(nums+"+-",nums) ) )
        integer = ( Combine( CaselessLiteral("0x") + Word( nums+"abcdefABCDEF" ) ) |
                    Word( nums+"+-", nums ) ).setName("int")

        # include definitions
        includeDef = ( Group( hashtag + include_ + langle + identifier + dot + 'idl' + rangle) |
                       Group( hashtag + include_ + quote + identifier + dot + 'idl' + quote) )
        includeDef.setParseAction(self.include_handler)

        # Comments (either single line or cStyle comments
        singleLineComment = "//" + restOfLine
        singleLineComment.setParseAction(self.single_line_comment_handler)
        cStyleComment.setParseAction(self.cStyle_comment_handler)

        anyComment = (singleLineComment | cStyleComment)
        optionalComment = ZeroOrMore(anyComment)

        # Typenames
        udTypeName = delimitedList( identifier, "::", combine=True ).setName("udType")
        # have to use longest match for type, in case a user-defined type name starts with a keyword type, like "stringSeq" or "longArray"
        typeName = ( any_ ^ boolean_ ^ char_ ^ double_ ^ fixed_ ^
                    float_ ^ Optional(unsigned_) + long_long_ ^ Optional(unsigned_) + long_ ^ octet_ ^
                     Optional(unsigned_) + short_ ^ string_ ^
                    wchar_ ^ wstring_ ^ udTypeName ).setName("type")

        # Sequences
        sequenceDef = Forward().setName("seq")
        sequenceDef << Group( sequence_ + langle + ( sequenceDef | typeName ) + rangle )
        sequenceDef.setParseAction(self.sequence_handler)


        # Typedefs
        typeDef = sequenceDef | ( typeName + Optional( lbrack + integer + rbrack ) )
        typedefDef = Group( typedef_ + typeDef + identifier + semi ).setName("typedef")
        typedefDef.setParseAction(self.typedef_handler)

        typedefArrayDef = Group( typedef_ + typeName + identifier + lbrack + integer + rbrack + semi)
        typedefArrayDef.setParseAction(self.typedef_array_handler)

        typedefStringSequenceDef = Group( typedef_ + string_ + langle + (integer | identifier) + rangle + identifier +
                                          semi)
        typedefStringSequenceDef.setParseAction(self.typedef_string_sequence_handler)

        oneOftheTypedefDefs = (typedefDef | typedefArrayDef | typedefStringSequenceDef) + optionalComment

        # Enumerations
        enumDef = Forward()
        enumItem = identifier + comma + optionalComment
        enumItem.setParseAction(self.enumeration_item_handler)
        lastEnumItem = identifier + optionalComment
        lastEnumItem.setParseAction(self.enumeration_item_handler)
        enumDef << enum_ + identifier + lbrace + optionalComment + ZeroOrMore(enumItem) + lastEnumItem + rbrace + \
                   semi + optionalComment
        enumDef.setParseAction(self.enumeration_handler)

        # Constants
        constDef = Group( const_ + typeDef + identifier + equals + ( real | integer | quotedString ) + semi ) #| quotedString )
        constDef.setParseAction(self.constant_handler)

        # Exceptions
        exceptionItem = Group( typeDef + identifier + semi )
        exceptionDef = ( exception_ + identifier + lbrace + ZeroOrMore( exceptionItem ) + rbrace + semi )
        exceptionDef.setParseAction(self.exception_handler)

        # Attributes
        attributeDef = Optional( readonly_ ) + attribute_ + typeDef + identifier + semi
        attributeDef.setParseAction(self.attribute_handler)

        # Operations
        parameter = Group( ( inout_ | in_ | out_ ) + typeName + identifier )
        parameter.setParseAction(self.parameter_handler)
        paramlist = delimitedList( parameter ).setName( "paramlist" )

        operationDef = ( ( void_ ^ typeDef ) + identifier + lparen + Optional( paramlist ) + rparen + \
                        Optional( raises_ + lparen + Group( delimitedList( typeName ) ) + rparen ) + semi )
        operationDef.setParseAction(self.operation_handler)

        # Structures
        struct_item = typeName + identifier + semi + optionalComment
        struct_item.setParseAction(self.struct_item_handler)
        structDef = Group( struct_ + identifier + lbrace + optionalComment + OneOrMore(struct_item) + rbrace + semi)
        structDef.setParseAction(self.struct_handler)

        # Modules and Interfaces
        moduleDef = Forward()
        interfaceDef = Forward()
        interfaceItem = (interfaceDef | exceptionDef | constDef | oneOftheTypedefDefs | moduleDef | enumDef |
                         structDef | operationDef | anyComment)
        interfaceDef = Group( interface_ + identifier  + Optional( colon + delimitedList( typeName ) ) + lbrace +
                        ZeroOrMore( interfaceItem ) + rbrace + semi )
        interfaceDef.setParseAction(self.interface_handler)

        moduleItem = (interfaceDef | exceptionDef | constDef | oneOftheTypedefDefs | moduleDef | enumDef |
                      structDef | anyComment)
        moduleDef << module_ + identifier + lbrace + ZeroOrMore( moduleItem ) + rbrace + semi
        moduleDef.setParseAction(self.module_handler)

        # Full definition
        self.bnf = ZeroOrMore(includeDef) + ( moduleDef | OneOrMore( moduleItem ) )
        #self.bnf.ignore( singleLineComment )
        #self.bnf.ignore( cStyleComment )

    def parse_string(self, idlString):

        parseResult = self.bnf.parseString(idlString)

        if self.tokens is not None:
            self.tokens += parseResult
        else:
            self.tokens = parseResult

        print 'Unhandled definitions: ' + str(self.unhandled_definitions)

    ######### HANDLERS FOR ALL THE TOKENS ###############
    def include_handler(self, s, loc, tokens):
        logger.info('tokens: ' + str(tokens))

        include_file = ''.join(tokens[0][3:6])
        logger.debug('Include file: "' + include_file + '"')

        temp_idl_dir = self.idl_directory
        temp_idl_file = self.idl_tree.current_idl_filename
        self.parse_idl_file(os.path.join(self.idl_directory, include_file))
        self.idl_directory = temp_idl_dir
        self.idl_tree.current_idl_filename = temp_idl_file

    def module_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.idl_tree.add_parent(ModuleNode(loc, tokens))

    def interface_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.idl_tree.add_parent(InterfaceNode(loc, tokens))

    def typedef_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.idl_tree.add_sibling(TypedefNode(loc, tokens))

    def typedef_string_sequence_handler(self, s, loc, tokens):
        logger.info('tokens: ' + str(tokens))
        self.idl_tree.add_sibling(TypedefStringSequenceNode(loc, tokens))

    def typedef_array_handler(self, s, loc, tokens):
        logger.info('tokens: ' + str(tokens))
        self.idl_tree.add_sibling(TypedefArrayNode(loc, tokens))

    def constant_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.idl_tree.add_sibling(ConstantNode(loc, tokens))

    def enumeration_item_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.idl_tree.add_sibling(EnumerationItemNode(loc, tokens))

    def enumeration_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.idl_tree.add_parent(EnumerationNode(loc, tokens))

    def struct_item_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.idl_tree.add_sibling(StructItemNode(loc, tokens))

    def struct_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.idl_tree.add_parent(StructNode(loc, tokens))

    def param_list_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        #self.idl_tree.add_sibling(ParamListNode(loc, tokens))

    def parameter_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.idl_tree.add_sibling(ParameterNode(loc, tokens))

    def operation_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.idl_tree.add_parent(OperationNode(loc, tokens))

    def cStyle_comment_handler(self, s, loc, tokens):
        logger.info('loc: ' + str(loc) + ' tokens: ' + str(tokens))
        self.idl_tree.add_sibling(CStyleCommentNode(loc, tokens))

    def single_line_comment_handler(self, s, loc, tokens):
        logger.info('tokens: ' + str(tokens))
        self.idl_tree.add_sibling(SingleLineCommentNode(loc, tokens))


    # Not implemented yet
    def sequence_handler(self, s, loc, tokens):
        logger.info('tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def typedef_sequence_handler(self, s, loc, tokens):
        logger.info('tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def exception_handler(self, s, loc, tokens):
        logger.info('tokens: ' + str(tokens))
        self.unhandled_definitions += 1

    def attribute_handler(self, s, loc, tokens):
        logger.info('tokens: ' + str(tokens))
        self.unhandled_definitions += 1

'''
class ConformanceTestGenerator():
    def __init__(self, idl_tree):
        self.idl_tree = idl_tree

    def get_conformance_header(self, template_group, language, **params):

        header_list = list()

        if language in [TestConfiguration.Language.C,  TestConfiguration.Language.CPP]:
            include_files = params.get('include_files', list())
            for i in include_files:
                include_file_template = template_group.getInstanceOf('includeFiles')
                include_file_template['FILENAME'] = i
                header_list.append(str(include_file_template))

        return '\n'.join(header_list)

    def generate_gold_library_source_file(self, source_filename, template_file, language, **params):
        logger.info(locals())

        if not os.path.isdir(os.path.dirname(source_filename)):
            raise IOError('Gold Library directory does not exist: ' + os.path.dirname(source_filename))
        if not os.path.isfile(template_file):
            raise IOError('Template file does not exist: ' + template_file)
        if language not in TestConfiguration.Language.allLanguages:
            raise RuntimeError('Language ' + str(language) + ' not in ' +
                               str(TestConfiguration.Language.allLanguages))

        template = stringtemplate3.StringTemplateGroup(name='gold_lib', file=open(template_file,'r'))

        gold_lib_source_template = template.getInstanceOf('goldLibSourceFile')

        if language in [TestConfiguration.Language.C, TestConfiguration.Language.CPP]:
            include_files = params.get('include_files', list())
            include_file_list = list()
            for i in include_files:
                include_file_template = template.getInstanceOf('includeFiles')
                include_file_template['FILENAME'] = i
                include_file_list.append(str(include_file_template))
            gold_lib_source_template['INCLUDE_FILES'] = '\n'.join(include_file_list)

            registerCallback = params.get('registerCallback', True)
            function_declarations = list()
            for current_node in iter(self.idl_tree):
                if current_node.gold_lib_source_node:
                    if self.face_specified_node(current_node):
                        continue
                    if not registerCallback and current_node.name == 'Register_Callback':
                        continue
                    function_declarations.append(current_node.generate_gold_lib_body(template, language))

            gold_lib_source_template['FUNCTION_DECLARATIONS'] = '\n'.join(function_declarations)

            source_file = open(source_filename, 'w')
            source_file.write(str(gold_lib_source_template))
            source_file.close()

    def face_specified_node(self, node):
        if isinstance(node, OperationNode):
            for c in node.children:
                if c.type_name == 'FACE_SPECIFIED_TYPE':
                    return True

            if node.name.lower().find('send_event') != -1:
                return True

        elif isinstance(node, ParameterNode):
            if node.name == 'FACE_SPECIFIED_TYPE':
                return True

        elif isinstance(node, StructNode):
            if node.name == 'FACE_SPECIFIED_TYPE':
                return True

        return False

    def generate_conformance_tests(self, conformance_gen_dir, template_file, language, **params):
        logger.info(locals())
        base_filename = params.get('basename', 'Test')

        if not os.path.isdir(conformance_gen_dir):
            raise IOError('Conformance Gen Directory does not exist: ' + conformance_gen_dir)

        if not os.path.isfile(template_file):
            raise IOError('Template file does not exist: ' + template_file)

        if language not in TestConfiguration.Language.allLanguages:
            raise RuntimeError('Language ' + str(language) + ' not in ' +
                               str(TestConfiguration.Language.allLanguages))

        old_source_files = glob.glob(os.path.join(conformance_gen_dir,
                                           '*' + TestConfiguration.Language.source_suffix[language]))
        for o in old_source_files:
            os.remove(o)

        template = stringtemplate3.StringTemplateGroup(name='conformance', file=open(template_file,'r'))

        test_num = 1

        assertion_file = open(os.path.join(conformance_gen_dir, 'assertions.txt'), 'w')

        all_conformance_tests_template = template.getInstanceOf('conformanceSourceFile')
        all_conformance_tests_template['TEST_NAME'] = 'all_conformance_tests'
        all_conformance_tests_template['HEADER'] = self.get_conformance_header(template, language, **params)


        fork_test_code = [TestConfiguration.Language.begin_single_line_comment[language] +
                          ' Using all non type specific function calls to test for fork use ' +
                          TestConfiguration.Language.end_single_line_comment[language]]

        registerCallback = params.get('registerCallback', True)

        for current_node in iter(self.idl_tree):
            # Only continue if it is conformance test node and it is not a face specified node.
            # The face specifed nodes are generated separately for each data model
            if not current_node.conformance_test_node or self.face_specified_node(current_node):
                continue

            if not registerCallback and current_node.name == 'Register_Callback':
                continue

            test_name = base_filename + str(test_num) + TestConfiguration.Language.source_suffix[language]
            conformance_test_filename = os.path.join(conformance_gen_dir, test_name)

            conformance_source_file_template = template.getInstanceOf('conformanceSourceFile')
            conformance_source_file_template['TEST_NAME'] = test_name
            conformance_source_file_template['HEADER'] = self.get_conformance_header(template, language, **params)

            test_code = current_node.get_conformance_test_code(template, language)
            conformance_source_file_template['TEST_CODE'] = '\n'.join(test_code)

            conformance_file = open(conformance_test_filename, 'w')
            conformance_file.write(str(conformance_source_file_template))
            conformance_file.close()

            all_conformance_tests_template['TEST_CODE'] = \
                TestConfiguration.Language.begin_single_line_comment[language] + 10 * '-' + test_name + 10 * '-' + \
                TestConfiguration.Language.end_single_line_comment[language] + '\n'.join(test_code)

            if isinstance(current_node, OperationNode):
                fork_test_code.extend(test_code)


            assertion_file.write('AssertionTest' + str(test_num) + ' Testing ' +
                                 current_node.get_fully_specified_name(language) + ' conformance\n')

            test_num += 1

        assertion_file.close()

        test_name = 'ForkConformanceTest'
        conformance_source_file_template = template.getInstanceOf('conformanceSourceFile')
        conformance_source_file_template['TEST_NAME'] = test_name
        conformance_source_file_template['HEADER'] = self.get_conformance_header(template, language, **params)
        conformance_source_file_template['TEST_CODE'] = '\n'.join(fork_test_code)
        segment_fork_test_file = open(
            os.path.join(conformance_gen_dir, test_name + TestConfiguration.Language.source_suffix[language]), 'w')
        segment_fork_test_file.write(str(conformance_source_file_template))
        segment_fork_test_file.close()




        all_tests_file = open(os.path.join(conformance_gen_dir,
                                           'all_conformance_tests.txt'), 'w')
        all_tests_file.write(str(all_conformance_tests_template))
        all_tests_file.close()



    def old_generate_conformance_tests(self, conformance_gen_dir, template_file, language):
        logger.info(locals())
        if not os.path.isdir(conformance_gen_dir):
            raise IOError('Conformance Gen Directory does not exist: ' + conformance_gen_dir)

        if not os.path.isfile(template_file):
            raise IOError('Template file does not exist: ' + template_file)

        if language not in TestConfiguration.Language.allLanguages:
            raise RuntimeError('Language ' + str(language) + ' not in ' +
                               str(TestConfiguration.Language.allLanguages))

        old_source_files = glob.glob(os.path.join(conformance_gen_dir,
                                           '*' + TestConfiguration.Language.source_suffix[language]))
        for o in old_source_files:
            os.remove(o)

        template = stringtemplate3.StringTemplateGroup(name='conformance', file=open(template_file,'r'))

        test_num = 1

        assertion_file = open(os.path.join(conformance_gen_dir, 'assertions.txt'), 'w')
        all_tests_file = open(os.path.join(conformance_gen_dir,
                                           'all_conformance_tests' + TestConfiguration.Language.source_suffix[language]), 'w')
        for current_node in iter(self.idl_tree):
            if not current_node.conformance_test_node:
                continue

            if language == TestConfiguration.Language.Ada:
                self.generate_node_ada_conformance_test(current_node, language, test_num, template, conformance_gen_dir,
                                                    assertion_file, all_tests_file)

            if language == TestConfiguration.Language.C:
                self.generate_node_c_conformance_test(current_node, test_num, template, conformance_gen_dir,
                                                      assertion_file, all_tests_file)


            test_num += 1

        assertion_file.close()
        all_tests_file.close()

    def generate_gold_library_files(
            self, language, gold_definition_filename, gold_source_filename, template_file, **kw):

        logger.info(locals())
        package_name = kw.get('package_name', 'Unset_Package_Name')

        definition_dir = os.path.dirname(gold_definition_filename)
        if not os.path.isdir(definition_dir):
            raise IOError('gold library include/spec directory does not exist: ' + definition_dir)

        source_dir = os.path.dirname(gold_source_filename)
        if not os.path.isdir(source_dir):
            raise IOError('gold library source directory does not exist: ' + source_dir)

        if not os.path.isfile(template_file):
            raise IOError('Template file does not exist: ' + template_file)

        if language not in TestConfiguration.Language.allLanguages:
            raise RuntimeError('Language ' + str(language) + ' not in ' +
                               str(TestConfiguration.Language.allLanguages))



        template = stringtemplate3.StringTemplateGroup(name='gold_lib', file=open(template_file, 'r'))

        gold_lib_definition_file = template.getInstanceOf('goldLibDefinitionFile')
        gold_lib_source_file = template.getInstanceOf('goldLibSourceFile')

        if language is TestConfiguration.Language.Ada:
            old_ada_files = glob.glob(os.path.join(definition_dir, '*.ads'))
            for o in old_ada_files:
                os.remove(o)

            old_ada_files = glob.glob(os.path.join(source_dir, '*.adb'))
            for o in old_ada_files:
                os.remove(o)

            for current_node in iter(self.idl_tree):
                if current_node.gold_lib_definition_node:
                    test_code = current_node.generate_gold_lib_definition(template, language)
                    logger.debug(str(test_code))
                    gold_lib_definition_file['DEFINITIONS'] = str(test_code)

                if current_node.gold_lib_source_node:
                    test_code = current_node.generate_gold_lib_body(template, language)
                    gold_lib_source_file['BODIES'] = str(test_code)

        gold_lib_definition_file['PACKAGE_NAME'] = package_name
        gold_lib_source_file['PACKAGE_NAME'] = package_name
        logger.debug(str(gold_lib_definition_file))
        logger.debug(str(gold_lib_source_file))

        gold_spec_file = open(gold_definition_filename, 'w')
        gold_spec_file.write(TestConfiguration.getDistributionStatement(language,
                                                                        os.path.basename(gold_definition_filename)))
        gold_spec_file.write(str(gold_lib_definition_file))
        gold_spec_file.close()

        gold_body_file = open(gold_source_filename, 'w')
        gold_body_file.write(TestConfiguration.getDistributionStatement(language,
                                                                        os.path.basename(gold_source_filename)))
        gold_body_file.write(str(gold_lib_source_file))
        gold_body_file.close()


    def generate_face_specified_gold_library(
            self, data_type_list, language, spec_dir, src_dir, template_file):

        logger.info(locals())

        if not type(data_type_list) != type(list):
            raise RuntimeError('datatype list needs to be a list')

        if not os.path.isdir(spec_dir):
            raise IOError('Conformance Gen Directory does not exist: ' + spec_dir)

        if not os.path.isdir(src_dir):
            raise IOError('Conformance Gen Directory does not exist: ' + src_dir)

        if not os.path.isfile(template_file):
            raise IOError('Template file does not exist: ' + template_file)

        if language not in TestConfiguration.Language.allLanguages:
            raise RuntimeError('Language ' + str(language) + ' not in ' +
                               str(TestConfiguration.Language.allLanguages))

        template = stringtemplate3.StringTemplateGroup(name='gold_lib', file=open(template_file, 'r'))

        gold_lib_definition_file = template.getInstanceOf('faceSpecifiedTypeDefinitionFile')
        gold_lib_source_file = template.getInstanceOf('faceSpecifiedTypeSourceFile')

        if language == TestConfiguration.Language.Ada:
            gold_tss_spec_filename = os.path.join(spec_dir, constants.ADA_TSS_FILENAME
                                                            + TestConfiguration.Language.header_suffix[language])
            base_tss_spec_filename = gold_tss_spec_filename + constants.ADA_BASE_EXTENSION
            gold_tss_body_filename = os.path.join(src_dir, constants.ADA_TSS_FILENAME
                                                            + TestConfiguration.Language.source_suffix[language])
            base_tss_body_filename = gold_tss_body_filename + constants.ADA_BASE_EXTENSION

            for b in [base_tss_body_filename, base_tss_spec_filename]:
                if not os.path.isfile(b):
                    err_msg = "Ada base TSS file does not exist: " + b
                    logger.error(err_msg)
                    raise RuntimeError(err_msg)
            commonTools.copyFile(base_tss_spec_filename, gold_tss_spec_filename)
            commonTools.copyFile(base_tss_body_filename, gold_tss_body_filename)
            read_callback_sub_package_template = template.getInstanceOf("readCallbackSubPackage")

            for data_type in data_type_list:
                self.idl_tree.overwrite_face_specified_type_nodes(data_type)

                send_event_template = template.getInstanceOf("sendEventAccessSignature")
                send_event_template['MESSAGE_TYPE'] = data_type
                send_event_template['FACE_SPECIFIED_TYPE'] = 'face.dm.' + data_type
                read_callback_sub_package_template['SEND_EVENT_ACCESS_SIGNATURE'] = str(send_event_template)

            gold_lib_definition_file['DEFINITIONS'] = str(read_callback_sub_package_template)

            for data_type in data_type_list:
                self.idl_tree.overwrite_face_specified_type_nodes(data_type)

                for current_node in iter(self.idl_tree):
                    if not isinstance(current_node, OperationNode):
                        continue

                    test_code = current_node.generate_gold_lib_definition(template, language)
                    logger.debug(str(test_code))
                    gold_lib_definition_file['DEFINITIONS'] = str(test_code)
                    test_code = current_node.generate_gold_lib_body(template, language)
                    gold_lib_source_file['BODIES'] = str(test_code)


            logger.debug(str(gold_lib_definition_file))
            logger.debug(str(gold_lib_source_file))


            gold_spec_file = open(gold_tss_spec_filename, 'a')
            gold_spec_file.write(str(gold_lib_definition_file))
            gold_spec_file.close()

            gold_body_file = open(gold_tss_body_filename, 'a')
            gold_body_file.write(str(gold_lib_source_file))
            gold_body_file.close()

    def generate_face_specified_conformance_tests(
            self, data_type_list, language, conformance_gen_dir, template_file):

        logger.info(locals())

        if not type(data_type_list) != type(list):
            raise RuntimeError('datatype list needs to be a list')

        if not os.path.isdir(conformance_gen_dir):
            raise IOError('Conformance Gen Directory does not exist: ' + conformance_gen_dir)

        if not os.path.isfile(template_file):
            raise IOError('Template file does not exist: ' + template_file)

        if language not in TestConfiguration.Language.allLanguages:
            raise RuntimeError('Language ' + str(language) + ' not in ' +
                               str(TestConfiguration.Language.allLanguages))

        face_prefix = 'Test_'
        old_source_files = glob.glob(os.path.join(conformance_gen_dir,
                                           face_prefix + '*' + TestConfiguration.Language.source_suffix[language]))
        for o in old_source_files:
            os.remove(o)

        template = stringtemplate3.StringTemplateGroup(name='conformance', file=open(template_file,'r'))

        test_num = 1

        assertion_file = open(os.path.join(conformance_gen_dir, 'face_specified_assertions.txt'), 'w')
        all_test_filename = 'all_' + language + '_conformance_tests.txt'
        all_tests_file = open(os.path.join(conformance_gen_dir, all_test_filename), 'w')
        distribution_statement = TestConfiguration.getDistributionStatement(language, all_test_filename)

        first_time = False # Was true...don't think I need to do this any more
        for current_node in iter(self.idl_tree):
            if not isinstance(current_node, OperationNode):
                continue

            if language == TestConfiguration.Language.Ada and first_time:
                loc = current_node.loc -1
                tokens = [['interface', 'Type_Safe_myTypes']]
                i_node = InterfaceNode(loc,tokens)
                self.idl_tree.current_node = current_node
                self.idl_tree.add_parent(i_node)
                first_time = False


            for d in data_type_list:
                self.idl_tree.overwrite_face_specified_type_nodes(d)
                specified_name = face_prefix + d + "_" + current_node.name + \
                                 TestConfiguration.Language.source_suffix[language]
                self.generate_node_ada_conformance_test(current_node, language, test_num, template,
                                                    conformance_gen_dir, assertion_file, all_tests_file,
                                                    ['face.dm'], ' (' + d + ')',
                                                    specified_name)
                test_num += 1


        assertion_file.close()
        all_tests_file.close()



    def generate_node_ada_conformance_test(self, current_node, language, test_num, template,
                                       conformance_gen_dir, assertion_file, all_tests_file,
                                       added_face_packages=list(), added_assertion_text='',
                                       specified_filename=None):

        module_nodes = current_node.get_specific_node_type_parents([ModuleNode, InterfaceNode])
        module_names = list()
        for n in module_nodes:
            module_names.append(n.get_fully_specified_name(language))
        conformance_module_names = list()
        for m in module_names:
            conformance_module_names.append('CONFORMANCE_TEST' + TestConfiguration.Language.delimiter[language] + m)
        if language == TestConfiguration.Language.Ada and isinstance(current_node, ConstantNode):
            module_names.append('Interfaces.C')

        module_names.extend(added_face_packages)
        #TODO: add distribution statement to top of each file
        test_code = current_node.get_conformance_test_code(template, language)
        code_template = template.getInstanceOf('conformanceSourceFile')
        code_template['TEST_CODE_SECTION1'] = test_code[0]
        code_template['TEST_CODE_SECTION2'] = test_code[1]
        code_template['FACE_PACKAGES'] = ', '.join(module_names)
        code_template['CONFORMANCE_TEST_FACE_PACKAGES'] = ', '.join(conformance_module_names)
        if specified_filename is not None:
            source_filename = specified_filename
            code_template['TEST_NAME'] = specified_filename.split('.')[0]
        else:
            code_template['TEST_NAME'] = 'Test' + str(test_num)
            source_filename = 'Test' + str(test_num) + TestConfiguration.Language.source_suffix[language]


        source_file = open(os.path.join(conformance_gen_dir, source_filename), 'w')
        source_file.write(TestConfiguration.getDistributionStatement(language, source_filename))
        source_file.write(str(code_template))
        source_file.close()

        all_tests_file.write(2*'\n' + TestConfiguration.Language.begin_single_line_comment[language] + 80 * '#' +
                             TestConfiguration.Language.end_single_line_comment[language] + '\n')

        all_tests_file.write(TestConfiguration.Language.begin_single_line_comment[language] + ' Test' + str(test_num) +
                             TestConfiguration.Language.source_suffix[language] + ' Test Code ' +
                             TestConfiguration.Language.end_single_line_comment[language] + '\n')

        all_tests_file.write(str(code_template))

        assertion_file.write('AssertionTest' + str(test_num) + ' ' +
                             current_node.get_fully_specified_name(language) + added_assertion_text +
                             ' is conformant\n')

def generate_IOS_test_code():
    root_dir = ConformanceMain.getRootDir()
    datafiles_dir = os.path.join(root_dir, 'datafiles')
    idl_filenames = dict()
    idl_filenames['FACE_IOS'] = os.path.join(datafiles_dir, 'FACE_IOS_2.1.idl')
    languages = [TestConfiguration.Language.C, TestConfiguration.Language.CPP]
    conformance_template_filenames = {
        TestConfiguration.Language.C:   os.path.join(datafiles_dir,'C_ConformanceTemplates.stg'),
        TestConfiguration.Language.CPP: os.path.join(datafiles_dir, 'C++_ConformanceTemplates.stg')}

    conformance_test_dirs = {
        TestConfiguration.Language.C:   os.path.join(root_dir, constants.C_CONFORMANCE_TEST_DIR,   'IOS_2.1'),
        TestConfiguration.Language.CPP: os.path.join(root_dir, constants.CPP_CONFORMANCE_TEST_DIR, 'IOS_2.1')}

    conformance_test_include_files = {
        TestConfiguration.Language.C:   ['CONFORMANCE_TEST_IOS.h', 'user_includes.h'],
        TestConfiguration.Language.CPP: ['CONFORMANCE_TEST_IOS.hpp', 'user_includes.hpp']}

    gold_lib_template_filenames = {
        TestConfiguration.Language.C:   os.path.join(datafiles_dir, 'C_GoldLibTemplates.stg'),
        TestConfiguration.Language.CPP: os.path.join(datafiles_dir, 'C++_GoldLibTemplates.stg')}

    gold_lib_dirs = {
        TestConfiguration.Language.C:   os.path.join(root_dir, constants.C_GOLD_LIB_DIR,   'src', 'IOS_2.1'),
        TestConfiguration.Language.CPP: os.path.join(root_dir, constants.CPP_GOLD_LIB_DIR, 'src', 'IOS_2.1')}

    source_files =  {'FACE_IOS': 'ios'}
    include_files = {'FACE_IOS': 'IOS'}

    for idl_key, idl_filename in idl_filenames.iteritems():

        idl_parser = FaceIdlParser(filename=idl_filename)

        for language in languages:
            template = conformance_template_filenames[language]

            conformance_test_generator = ConformanceTestGenerator(idl_parser.idl_tree)

            conformance_test_generator.generate_conformance_tests(
                conformance_test_dirs[language],
                template,
                language,
                include_files=conformance_test_include_files[language])

            source_file = os.path.join(gold_lib_dirs[language],
                                       source_files[idl_key] + TestConfiguration.Language.source_suffix[language])

            template = gold_lib_template_filenames[language]
            gold_lib_include_files = [
                os.path.join('face', include_files[idl_key] + TestConfiguration.Language.header_suffix[language])]

            conformance_test_generator.generate_gold_library_source_file(
                source_file,
                template,
                language,
                include_files=gold_lib_include_files)




def generate_TSS_test_code():
    root_dir = ConformanceMain.getRootDir()
    datafiles_dir = os.path.join(root_dir, 'datafiles')

    idl_filenames = dict()
    idl_filenames['FACE_TS'] = os.path.join(datafiles_dir, 'FACE_TS_2.1.idl')
    idl_filenames['TypeAbstraction_TS'] = os.path.join(datafiles_dir, 'FACE_TypeAbstraction_TS_2.1.idl')

    languages = [TestConfiguration.Language.C, TestConfiguration.Language.CPP]
    conformance_template_filenames = {
        TestConfiguration.Language.C:   os.path.join(datafiles_dir,'C_ConformanceTemplates.stg'),
        TestConfiguration.Language.CPP: os.path.join(datafiles_dir, 'C++_ConformanceTemplates.stg')}

    conformance_test_dirs = {
        TestConfiguration.Language.C:   os.path.join(root_dir, constants.C_CONFORMANCE_TEST_DIR,   'TSS_2.1'),
        TestConfiguration.Language.CPP: os.path.join(root_dir, constants.CPP_CONFORMANCE_TEST_DIR, 'TSS_2.1')}

    conformance_test_include_files = {
        TestConfiguration.Language.C:   ['..' + os.path.sep + 'CONFORMANCE_TEST_TS.h',   'user_includes.h'],
        TestConfiguration.Language.CPP: ['..' + os.path.sep + 'CONFORMANCE_TEST_TS.hpp', 'user_includes.hpp']}

    gold_lib_template_filenames = {
        TestConfiguration.Language.C:   os.path.join(datafiles_dir, 'C_GoldLibTemplates.stg'),
        TestConfiguration.Language.CPP: os.path.join(datafiles_dir, 'C++_GoldLibTemplates.stg')}

    gold_lib_dirs = {
        TestConfiguration.Language.C:   os.path.join(root_dir, constants.C_GOLD_LIB_DIR,   'src', 'TSS_2.1'),
        TestConfiguration.Language.CPP: os.path.join(root_dir, constants.CPP_GOLD_LIB_DIR, 'src', 'TSS_2.1')}


    source_files =  {'FACE_TS': 'ts_base', 'TypeAbstraction_TS': 'typeabstraction_ts'}
    include_files = {'FACE_TS': 'TS',      'TypeAbstraction_TS': 'TypeAbstraction_TS'}

    for sub_dir, idl_filename in idl_filenames.iteritems():

        idl_parser = FaceIdlParser(filename=idl_filename)

        for language in languages:
            template = conformance_template_filenames[language]

            conformance_test_generator = ConformanceTestGenerator(idl_parser.idl_tree)

            registerCallbackFlag = sub_dir == 'TypeAbstraction_TS'

            conformance_test_generator.generate_conformance_tests(
                os.path.join(conformance_test_dirs[language], sub_dir),
                template,
                language,
                include_files=conformance_test_include_files[language],
                registerCallback=registerCallbackFlag)

            source_file = os.path.join(gold_lib_dirs[language],
                                       source_files[sub_dir] + TestConfiguration.Language.source_suffix[language])

            template = gold_lib_template_filenames[language]
            gold_lib_include_files = [
                os.path.join('face', include_files[sub_dir] + TestConfiguration.Language.header_suffix[language])]

            conformance_test_generator.generate_gold_library_source_file(
                source_file,
                template,
                language,
                include_files=gold_lib_include_files,
                registerCallback=registerCallbackFlag)


if __name__ == "__main__":

    import commonTools
    import os
    import Dialogs

    logger.info('In IdlParser main')
    generate_TSS_test_code()

    generate_IOS_test_code()




    initialDir = os.path.join(ConformanceMain.getRootDir(), 'datafiles')

    testing_face_specified_gold_lib = False
    if testing_face_specified_gold_lib:
        idlFilename = os.path.join(initialDir, 'FACE_SPECIFIED_TYPE_TSS.idl')
        template_filename = os.path.join(initialDir, 'Ada_GoldLibTemplates.stg')
        conformance_gen_dir = os.path.join(ConformanceMain.getRootDir(), 'generated_code')

        idlParser = FaceIdlParser(filename=idlFilename)
        sub_dir = "tss_face_specified_goldlib"
        gen_sub_dir = os.path.join(conformance_gen_dir, sub_dir)

        if not os.path.isdir(gen_sub_dir):
            commonTools.mkdir_p(gen_sub_dir)

        datatype_list = ['ProcessorStatus', 'AntennaStatus']
        gold_lib_generator = ConformanceTestGenerator(idlParser.idl_tree)
        gold_lib_generator.generate_face_specified_gold_library(datatype_list, TestConfiguration.Language.Ada,
                                                                gen_sub_dir, gen_sub_dir, template_filename)

    testing_face_specified_conformance_test = False
    #testing_face_specified_conformance_test = True
    if testing_face_specified_conformance_test:
        idlFilename = os.path.join(initialDir, 'FACE_SPECIFIED_TYPE_TSS.idl')
        #idlFilename = os.path.join(initialDir, 'DUMMY_SPECIFIED_TYPE_TSS.idl')
        template_filename = os.path.join(initialDir, 'Ada_ConformanceTemplates.stg')
        conformance_gen_dir = os.path.join(ConformanceMain.getRootDir(), 'generated_code')

        idlParser = FaceIdlParser(filename=idlFilename)
        sub_dir = "tss_face_specified"
        gen_sub_dir = os.path.join(conformance_gen_dir, sub_dir)
        if not os.path.isdir(gen_sub_dir):
            commonTools.mkdir_p(gen_sub_dir)

        datatype_list = ['dummy1', 'dummy2']
        conformance_test_generator = ConformanceTestGenerator(idlParser.idl_tree)
        conformance_test_generator.generate_face_specified_conformance_tests(datatype_list, TestConfiguration.Language.Ada,
                                                                             gen_sub_dir, template_filename)

    testing_all_interfaces = False
    #testing_all_interfaces = True
    if testing_all_interfaces:

        idlFilenames = dict()
        idlFilenames['common'] = os.path.join(initialDir, 'FACE_common.idl')
        idlFilenames['ios'] = os.path.join(initialDir, 'FACE_IOS.idl')
        idlFilenames['hmfm'] = os.path.join(initialDir, 'FACE_HMFM.idl')
        idlFilenames['tss'] = os.path.join(initialDir, 'FACE_TS.idl')


        template_filename = os.path.join(initialDir, 'Ada_ConformanceTemplates.stg')
        conformance_gen_dir = os.path.join(ConformanceMain.getRootDir(), 'generated_code')
        #idlFilename = Dialogs.osAgnosticOpenFilenameDialog(initialdir=initialDir)

        for sub_dir, idlFilename in idlFilenames.iteritems():
            logger.debug('Dir: ' + sub_dir + ' File: ' + idlFilename)

            if len(idlFilename) > 0:
                idlParser = FaceIdlParser(filename=idlFilename)
                #idlParser.idl_tree.print_tree()
                #pprint.pprint( idlParser.tokens.asList())

                logger.debug('Num unhandled definitions: ' + str(idlParser.unhandled_definitions))

                gen_sub_dir = os.path.join(conformance_gen_dir, sub_dir)
                if not os.path.isdir(gen_sub_dir):
                    commonTools.mkdir_p(gen_sub_dir)

                conformance_test_generator = ConformanceTestGenerator(idlParser.idl_tree)
                conformance_test_generator.generate_conformance_tests(gen_sub_dir, template_filename, TestConfiguration.Language.Ada)

'''


