#!/usr/bin/env python
"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""

"""
---------------------------------------------------------------------------------

idl_to_code_generator.py

Contains class IdlCodeGenerator that generates both gold standard libraries and conformance tests for C, C++, and Ada
based on FACE standard IDL files (currently supporting 2.1, will also support 2.0 at some point in future.  When it is
run as main, it generates all non data type specific code.  This is used during make_devel scripts to build conformance
test suite from the repository.  It is also called during test runs to generate conformance tests/gold libs for data
type specific code from a given data model, using IdlCodeGenerator.generate_face_specified_conformance_test_code

"""
__author__ = 'david.fann@vanderbilt.edu'


import os
import ConformanceMain
import sys
egg_path = os.path.join(ConformanceMain.getRootDir(),'face_conformance_app', 'pyparsing-2.0.1-py2.7.egg')
sys.path.append(egg_path)
egg_path = os.path.join(ConformanceMain.getRootDir(),'face_conformance_app', 'stringtemplate3-3.1-py2.7.egg')
sys.path.append(egg_path)
egg_path = os.path.join(ConformanceMain.getRootDir(),'face_conformance_app', 'setuptools-0.6c11-py2.7!\\')
sys.path.append(egg_path)


import logging
import stringtemplate3
import glob

# Conformance Test Suite imports
if __name__ == "__main__":
    ConformanceMain.initializeLogger()

import commonTools
import idl_parser
import constants
import set_distro_statement

from TestConfiguration import Language
from collections import OrderedDict
import re

logger = logging.getLogger(__name__)


class IdlCodeGenerator:
    conformance_templates = {
        Language.Ada: 'Ada_ConformanceTemplates.stg',
        Language.C: 'C_ConformanceTemplates.stg',
        Language.CPP: 'C++_ConformanceTemplates.stg',
        Language.Java: None  # Don't generate Java Conformance Code
    }

    gold_lib_templates = {
        Language.Ada: 'Ada_GoldLibTemplates.stg',
        Language.C: 'C_GoldLibTemplates.stg',
        Language.CPP: 'C++_GoldLibTemplates.stg',
        Language.Java: 'Java_GoldLibTemplates.stg'
    }

    types_conformance_tests = {
        Language.C: 'FACE_types_conformance_test.c',
        Language.CPP: 'FACE_types_conformance_test.cpp',
        Language.Ada: 'FACE_Basic_Types_Conformance_Test.adb'
    }

    def __init__(self, idl_tree):
        logger.info(locals())
        assert isinstance(idl_tree, idl_parser.IdlTree)
        self.idl_tree = idl_tree
        self.included_files = list()
        # VERY SPECIAL CASE...need to save declaration in GSL gen for conformance tests later
        self.send_event_ptr_declaration = None

    def get_included_file_heading(self, language, template, idl_filename, version):
        logger.info(locals())
        if not idl_filename in self.included_files:
            self.included_files.append(idl_filename)

            if language in [Language.C, Language.CPP]:
                header_filename, ext = os.path.splitext(os.path.basename(idl_filename))
                face_replace = 'FACE/'
                if version == constants.FACE_STANDARD_VERSION_2_0:
                    face_replace = 'face/'

                header_filename = header_filename.replace('FACE_', face_replace) + Language.header_suffix[language]
                template['HEADING'] = '\n#include <' + header_filename + '>'

    def generate_gold_library_code(self, language, **params):
        logger.info(locals())

        # self.idl_tree.print_tree()

        # One of these must be set...both can be set
        declaration_file = params['declarationFile']

        # Optional parameters
        implementation_file = params.get('implementationFile', None)  # Don't necessarily have an imple file
        root_dir = params.get('rootDir', ConformanceMain.getRootDir())
        templates_dir = params.get('templatesDir', os.path.join(root_dir, constants.DATAFILES_DIR))
        declaration_footer = params.get('declarationFooter', '')
        declaration_heading = params.get('declarationHeading', '')
        include_files = params.get('includeFiles', None)
        version = params.get('version', None)
        assert language in Language.allLanguages

        gold_lib_template_filename = os.path.join(templates_dir, self.gold_lib_templates[language])

        if not os.path.isfile(gold_lib_template_filename):
            raise IOError('Gold Lib Template File: ' + gold_lib_template_filename + ' does not exist')

        gold_lib_template = stringtemplate3.StringTemplateGroup(
            name='gold_lib', file=open(gold_lib_template_filename, 'r'))

        declaration_filename = declaration_file
        relative_filename = os.path.relpath(declaration_filename, root_dir)
        declaration_template_instance = gold_lib_template.getInstanceOf('declarationFile')
        declaration_template_instance['HEADING'] = set_distro_statement.get_distribution_statement(
            language, relative_filename)

        if language in [Language.C, Language.CPP]:
            declaration_header_constant = commonTools.getHeaderDefineConstant(relative_filename)
            declaration_template_instance['IFNDEF_GUARD_CONSTANT'] = declaration_header_constant

            if not include_files is None:
                for i in include_files:
                    include_template = gold_lib_template.getInstanceOf('includeFiles')
                    include_template['FILENAME'] = i
                    declaration_heading += str(include_template)

        if language == Language.Ada:
            package_name, extension = os.path.splitext(os.path.basename(declaration_file))
            declaration_template_instance['PACKAGE_NAME'] = package_name.replace('-', '.').upper()

            if not include_files is None:
                for i in include_files:
                    with_use_template_instance = gold_lib_template.getInstanceOf('withUseLine')
                    with_use_template_instance['WITH_NAME'] = i
                    if i == 'FACE_Basic_Types':
                        with_use_template_instance['USE_NAME'] = i
                    declaration_heading += str(with_use_template_instance)

        declaration_template_instance['HEADING'] = declaration_heading
        declaration_template_instance['FOOTER'] = declaration_footer

        if implementation_file is not None:
            implementation_filename = implementation_file
            implementation_footer = declaration_footer

            implementation_template_instance = gold_lib_template.getInstanceOf('implementationFile')

            implementation_template_instance['HEADING'] = set_distro_statement.get_distribution_statement(
                language, os.path.relpath(implementation_filename, root_dir))

            if language in [Language.C, Language.CPP]:
                if version == constants.FACE_STANDARD_VERSION_2_0:
                    implementation_template_instance['HEADING'] = \
                        '\n#include <face/' + os.path.basename(declaration_file) + '>'
                else:
                    implementation_template_instance['HEADING'] = \
                        '\n#include <FACE/' + os.path.basename(declaration_file) + '>'

            elif language == Language.Ada:
                package_name, package_extension = os.path.splitext(os.path.basename(declaration_file))
                package_name = package_name.replace('-', '.').upper()
                implementation_template_instance['PACKAGE_NAME'] = package_name

            implementation_template_instance['FOOTER'] = implementation_footer

        namespace_list = list()
        namespace_node_list = list()

        for node in self.idl_tree:
            logger.debug('NODE: ' + node.name)

            if isinstance(node, idl_parser.InterfaceNode) or isinstance(node, idl_parser.ModuleNode):
                if language == Language.CPP:
                    namespace_instance = gold_lib_template.getInstanceOf('namespaceDeclaration')
                    namespace_instance['NAMESPACE'] = node.name

                    while len(namespace_node_list) > 0 and not node.is_parent(namespace_node_list[-1]):
                        if len(namespace_list) > 1:
                            namespace_list[-2]['BODY'] = str(namespace_list[-1])
                            namespace_list.pop()
                            namespace_node_list.pop()
                        else:
                            declaration_template_instance['BODY'] = str(namespace_list[-1])
                            namespace_list.pop()
                            namespace_node_list.pop()

                    namespace_list.append(namespace_instance)
                    namespace_node_list.append(node)
                    continue

            if node in self.idl_tree.face_specified_type_nodes:
                continue  # handle these in face_specified method

            if node is self.idl_tree.root_node:
                continue  # root node doesn't have data...

            if node.idl_filename != self.idl_tree.current_idl_filename:
                self.get_included_file_heading(language, declaration_template_instance, node.idl_filename, version)
                continue

            if node.declaration_node:
                if len(namespace_list) > 0:
                    namespace_list[-1]['BODY'] = node.get_declaration(language, gold_lib_template)

                else:
                    declaration_template_instance['BODY'] = node.get_declaration(language, gold_lib_template)

                if node.implementation_node:
                    implementation_template_instance['BODY'] = node.get_implementation(language, gold_lib_template)

                if node.name == 'send_event':
                    self.send_event_ptr_declaration = node.get_declaration(language, gold_lib_template).replace(
                        'send_event', 'CONFORMANCE_TEST_send_event').replace('FACE_Read_Callback_', '').replace(
                        'Ptr', 'ptr'
                    )

        for i in reversed(range(0, len(namespace_list) -1)):
            namespace_list[i]['BODY'] = str(namespace_list[i+1])

        if len(namespace_list) > 0:
            declaration_template_instance['BODY'] = str(namespace_list[0])

        logger.debug('Writing declaration file: {0}'.format(declaration_filename))
        declaration_file = open(declaration_filename, 'w')
        declaration_file.write(str(declaration_template_instance))
        declaration_file.close()

        if implementation_file is not None:
            logger.debug('Writing implementation file: {0}'.format(implementation_filename))

            implementation_file_dir = os.path.dirname(implementation_file)
            if not os.path.exists(implementation_file_dir):
                os.makedirs(implementation_file_dir)

            implementation_file = open(implementation_filename, 'w')
            implementation_file.write(str(implementation_template_instance))
            implementation_file.close()

    @staticmethod
    def get_conformance_header(template_group, language, **params):

        return_value = 'No conformance header for ' + language + ' language'

        if language in [Language.C,  Language.CPP]:
            header_list = list()
            include_files = params.get('includeFiles', list())
            for i in include_files:
                include_file_template = template_group.getInstanceOf('includeFiles')
                include_file_template['FILENAME'] = i
                header_list.append(str(include_file_template))

            return_value = '\n'.join(header_list)

        elif language == Language.Ada:
            with_files = params.get('includeFiles', list())
            return_value = ''
            for w in with_files:
                with_template = template_group.getInstanceOf('withUseLine')
                with_template['WITH_NAME'] = w
                return_value += str(with_template)

        return return_value

    def generate_conformance_test_code(self, language, conformance_gen_dir, **params):
        logger.info(locals())
        assert language in Language.allLanguages

        send_event_node = None
        # optional parameters
        root_dir = params.get('rootDir', ConformanceMain.getRootDir())
        templates_dir = params.get('templatesDir', os.path.join(root_dir, constants.DATAFILES_DIR))
        template_file = params.get('templateFile', os.path.join(templates_dir, self.conformance_templates[language]))
        
        if language == Language.Ada:
            base_filename = params.get('basename', 'test')
        else:
            base_filename = params.get('basename', 'Test')
        assertion_text = params.get('assertionText', 'Test')
        type_abstraction_ts = params.get('typeAbstractionTS', False)
        fork_test = params.get('forkTest', True)

        if not os.path.isdir(conformance_gen_dir):
            raise IOError('Conformance Gen Directory does not exist: ' + conformance_gen_dir)

        if not os.path.isfile(template_file):
            raise IOError('Template file does not exist: ' + template_file)

        if language not in Language.allLanguages:
            raise KeyError('Language ' + str(language) + ' not in ' + str(Language.allLanguages))

        # delete old source files
        old_source_files = glob.glob(os.path.join(conformance_gen_dir, '*' + Language.source_suffix[language]))
        for o in old_source_files:
            os.remove(o)

        template = stringtemplate3.StringTemplateGroup(name='conformance', file=open(template_file,'r'))

        test_num = 1
        assertion_file = open(os.path.join(conformance_gen_dir, 'assertions.txt'), 'w')

        commonTools.copyFile(
            os.path.join(templates_dir, self.types_conformance_tests[language]),
            os.path.join(conformance_gen_dir, base_filename + str(test_num) + Language.source_suffix[language]))
        assertion_file.write('Assertion' + assertion_text + str(test_num) + ' Testing FACE types are defined correctly\n')
        test_num += 1


        all_conformance_tests_template = template.getInstanceOf('conformanceSourceFile')
        all_conformance_tests_template['TEST_NAME'] = 'all_conformance_tests'
        all_conformance_tests_template['HEADER'] = self.get_conformance_header(template, language, **params)

        if language == Language.Ada:
            all_conformance_tests_template = template.getInstanceOf('allConformanceTest')


        fork_test_code = [Language.begin_single_line_comment[language] +
                          ' Using all non type specific function calls to test for fork use ' +
                          Language.end_single_line_comment[language], '']


        for node in iter(self.idl_tree):
            # Only continue if it is conformance test node and it is not a face specified node.
            # The face specified nodes are generated separately for each data model
            # Not sure how to conformance test CALLBACK_ADDRESS_TYPE yet...
            # send_event (for TS register read callback, is supplied by PCS/PSS, and not by TSS.  Don't
            # need to have it in the TS conformance test.
            if (not node.conformance_test_node) or (node in self.idl_tree.face_specified_type_nodes) \
                    or (node.name in ['CALLBACK_ADDRESS_TYPE', 'FAULT_HANDLER_ENTRY_TYPE', 'send_event']):
                continue


            test_name = base_filename + str(test_num)
            conformance_test_filename = os.path.join(conformance_gen_dir, test_name + Language.source_suffix[language])

            conformance_source_file_template = template.getInstanceOf('conformanceSourceFile')
            conformance_source_file_template['TEST_NAME'] = test_name
            conformance_source_file_template['HEADER'] = self.get_conformance_header(template, language, **params)



            test_code = node.get_conformance_test(language, template)

            if node.name == 'Register_Callback' and self.send_event_ptr_declaration is not None:
                if language in [Language.C, Language.CPP]:   # VERY SPECIAL CASE, NEED TO PUT A SEND_EVENT PTR DECLARATION FIRST
                    test_code.insert(0, self.send_event_ptr_declaration)

                elif language == Language.Ada:  # VERY SPECIAL CASE AS WELL, NEED TO ADD
                    test_code[0] = test_code[0].replace('Read_Callback.Send_Event_Access',
                                                        'FACE.TS.TYPEABSTRACTION_TS.Read_Callback.Send_Event_Access')
                    test_code[1] = test_code[1].replace('Read_Callback.Send_Event_Access',
                                                        'FACE.TS.TYPEABSTRACTION_TS.Read_Callback.Send_Event_Access')


            if language == Language.Ada:
                if type_abstraction_ts:
                    test_code[0] = test_code[0].replace('.TS.TS.', '.TS.TYPEABSTRACTION_TS.')
                    test_code[1] = test_code[1].replace('.TS.TS.', '.TS.TYPEABSTRACTION_TS.')

                conformance_source_file_template['TEST_CODE_SECTION1'] = test_code[0]
                conformance_source_file_template['TEST_CODE_SECTION2'] = test_code[1]

                if node.name == 'SYSTEM_ADDRESS_TYPE' and \
                                constants.getFACEStandardVersion() == constants.FACE_STANDARD_VERSION_2_0:
                    # with system is needed for special case, SYSTEM_ADDRESS_TYPE is a System.Address, only time system
                    # is needed.
                    conformance_source_file_template['HEADER'] = 'with System;'

            else:
                conformance_source_file_template['TEST_CODE'] = '\n'.join(test_code)



            conformance_file = open(conformance_test_filename, 'w')
            conformance_file.write(str(conformance_source_file_template))
            conformance_file.close()

            if language == Language.Ada:
                single_conformance_test_template = template.getInstanceOf('singleConformanceTest')
                single_conformance_test_template['TEST_NAME'] = test_name
                single_conformance_test_template['TEST_CODE_SECTION1'] = test_code[0]
                single_conformance_test_template['TEST_CODE_SECTION2'] = test_code[1]

                all_conformance_tests_template['TEST_CODE'] = str(single_conformance_test_template)

            else:
                single_conformance_test_template = template.getInstanceOf('singleConformanceTest')
                single_conformance_test_template['TEST_NAME'] = test_name
                single_conformance_test_template['TEST_CODE'] = '\n'.join(test_code)
                all_conformance_tests_template['TEST_CODE'] = str(single_conformance_test_template)

            if isinstance(node, idl_parser.OperationNode):
                if language == Language.Ada:
                    fork_test_code[0] = fork_test_code[0] + '\n' + test_code[0]
                    fork_test_code[1] = fork_test_code[1] + '\n' + test_code[1]
                else:
                    fork_test_template = template.getInstanceOf('singleConformanceTest')
                    fork_test_template['TEST_NAME'] = node.name + ' Fork Test'
                    fork_test_template['TEST_CODE'] = test_code
                    fork_test_code.extend([str(fork_test_template)])

            fully_specified_name = node.get_fully_specified_name(language)
            if language == Language.Ada and isinstance(node, idl_parser.OperationNode):
                # In Ada there is an additional FACE.TS.TS.NAME for procedure names...magic!
                if type_abstraction_ts:
                    fully_specified_name = fully_specified_name.replace('.TS.', '.TS.TYPEABSTRACTION_TS.')
                else:
                    if constants.getFACEStandardVersion() != constants.FACE_STANDARD_VERSION_2_0:
                        fully_specified_name = fully_specified_name.replace('.TS.', '.TS.TS.')
            assertion_file.write('Assertion' + assertion_text + str(test_num) + ' Testing ' +
                                 fully_specified_name + ' conformance\n')

            test_num += 1

        assertion_file.close()

        if fork_test:
            if language == Language.Ada:
                test_name = 'forkconformancetest'
            else:
                test_name = 'ForkConformanceTest'
            
            conformance_source_file_template = template.getInstanceOf('conformanceSourceFile')
            conformance_source_file_template['TEST_NAME'] = test_name
            conformance_source_file_template['HEADER'] = self.get_conformance_header(template, language, **params)

            if language == Language.Ada:
                fork_test_code[0] = re.sub(' +', ' ', fork_test_code[0])
                fork_test_code[0] = "\n".join(list(OrderedDict.fromkeys(fork_test_code[0].split("\n"))))
                conformance_source_file_template['TEST_CODE_SECTION1'] = fork_test_code[0]
                conformance_source_file_template['TEST_CODE_SECTION2'] = fork_test_code[1]
            else:
                conformance_source_file_template['TEST_CODE'] = '\n'.join(fork_test_code)

            segment_fork_test_file = open(
                os.path.join(conformance_gen_dir, test_name + Language.source_suffix[language]), 'w')
            segment_fork_test_file.write(str(conformance_source_file_template))
            segment_fork_test_file.close()

        all_tests_file = open(os.path.join(conformance_gen_dir,
                                           'all_conformance_tests.txt'), 'w')
        all_tests_file.write(str(all_conformance_tests_template))
        all_tests_file.close()


def generate_face_specified_templates():
    #TODO:  May want to do this in the future...
    pass

def ada_face_basic_types_conformance_test_generator(template_file, conformance_test_filename, **params):
    logger.debug(locals())

    ada_face_basic_types = {
        'Short': 'Interfaces.Integer_16',
        'Long' : 'Interfaces.Integer_32',
        'Long_Long': 'Interfaces.Integer_64',
        'Unsigned_Short': 'Interfaces.Unsigned_16',
        'Unsigned_Long': 'Interfaces.Unsigned_32',
        'Unsigned_Long_Long': 'Interfaces.Unsigned_64',
        'Float': 'Interfaces.IEEE_Float_32',
        'Double': 'Interfaces.IEEE_Float_64',
        'Long_Double': 'Interfaces.IEEE_Extended_Float',
        'Char': 'Standard.Character',
        'Wchar': 'Standard.Wide_Character',
        'Octet': 'Interfaces.Unsigned_8'
    }
    # NOT SURE WHAT TO DO WITH BOOLEAN:'Boolean': 'Standard.Boolean', size 8
    # NOT SURE WHAT TO DO WITH ENUM SIZE: 'Enum_Size': "Interfaces.C.Int'Size"


    template = stringtemplate3.StringTemplateGroup(name='conformance', file=open(template_file,'r'))
    conformance_source_file_template = template.getInstanceOf('conformanceSourceFile')
    conformance_source_file_template['TEST_NAME'] = 'test1'  # always will be copied into test1 for tests


    conformance_source_file_template['HEADER'] = set_distro_statement.get_distribution_statement(
        Language.Ada,  os.path.basename(conformance_test_filename))


    for with_name in ['FACE_Basic_Types', 'Interfaces']:
        with_template = template.getInstanceOf('withUseLine')
        with_template['WITH_NAME'] = with_name
        conformance_source_file_template['HEADER'] = str(with_template)

    for face_type, standard_type in ada_face_basic_types.iteritems():
        conformance_test_template1A = template.getInstanceOf('conformanceTestCodeSection1ATypedef')
        conformance_test_template1B = template.getInstanceOf('conformanceTestCodeSection1BTypedef')
        conformance_test_template2 = template.getInstanceOf('conformanceTestCodeSection2Typedef')

        conformance_test_template1A['FULLY_SPECIFIED_NAME'] = 'FACE_Basic_Types.' + face_type
        conformance_test_template1A['SHORT_NAME'] = face_type
        conformance_test_template1A['TYPE_NAME'] = standard_type

        conformance_test_template1B['FULLY_SPECIFIED_NAME'] = 'FACE_Basic_Types.' + face_type
        conformance_test_template1B['SHORT_NAME'] = face_type
        conformance_test_template1B['TYPE_NAME'] = standard_type
        if face_type in ['Short', 'Long', 'Long_Long', 'Unsigned_Short', 'Unsigned_Long', 'Unsigned_Long_Long', 'Octet']:
            conformance_test_template1B['IS_INT_TYPE'] = True
        if face_type in ['Float', 'Double', 'Long_Double']:
            conformance_test_template1B['IS_FLOAT_TYPE'] = True

        conformance_test_template2['FULLY_SPECIFIED_NAME'] = 'FACE_Basic_Types.' + face_type
        conformance_test_template2['SHORT_NAME'] = face_type
        conformance_test_template2['TYPE_NAME'] = standard_type
        if not face_type in ['Char', 'Wchar']:
            conformance_test_template2['CONSTANT_DEFINED'] = True			

        conformance_source_file_template['TEST_CODE_SECTION1'] = str(conformance_test_template1A)
        conformance_source_file_template['TEST_CODE_SECTION1A'] = str(conformance_test_template1B)
        conformance_source_file_template['TEST_CODE_SECTION2'] = str(conformance_test_template2)


    conformance_test_file = open(conformance_test_filename, 'w')
    conformance_test_file.write(str(conformance_source_file_template))
    conformance_test_file.close()

if __name__ == "__main__":

    languages = [Language.C, Language.CPP, Language.Ada]
    face_versions = [constants.FACE_STANDARD_VERSION_2_0, constants.FACE_STANDARD_VERSION_2_1]

    FSV_2_0 = constants.FACE_STANDARD_VERSION_2_0
    FSV_2_1 = constants.FACE_STANDARD_VERSION_2_1

    declaration_files = dict()
    implementation_files = dict()
    conformance_test_directories = dict()
    declaration_footer = dict()
    idl_files = dict()

    for f in face_versions:
        declaration_files[f] = dict()
        implementation_files[f] = dict()
        idl_files[f] = list()
        declaration_footer[f] = dict()

        for l in languages:
            declaration_files[f][l] = dict()
            implementation_files[f][l] = dict()
            declaration_footer[f][l] = dict()
    # IOS
    ios_idl_2_0 = 'FACE_ios.idl'
    ios_idl_2_1 = 'FACE_IOS.idl'
    idl_files[FSV_2_0].append(ios_idl_2_0)
    idl_files[FSV_2_1].append(ios_idl_2_1)

    declaration_files[FSV_2_0][Language.C][ios_idl_2_0] = os.path.join('include', 'face', 'ios.h')
    declaration_files[FSV_2_1][Language.C][ios_idl_2_1] = os.path.join('include', 'FACE', 'IOS.h')
    declaration_files[FSV_2_0][Language.CPP][ios_idl_2_0] = os.path.join('include', 'face', 'ios.hpp')
    declaration_files[FSV_2_1][Language.CPP][ios_idl_2_1] = os.path.join('include', 'FACE', 'IOS.hpp')
    declaration_files[FSV_2_0][Language.Ada][ios_idl_2_0] = os.path.join('spec', 'face', 'face-io.ads')
    declaration_files[FSV_2_1][Language.Ada][ios_idl_2_1] = os.path.join('spec', 'FACE', 'face-io.ads')

    implementation_files[FSV_2_0][Language.C][ios_idl_2_0] = os.path.join('src', 'IOS', 'IOSGoldLib.c')
    implementation_files[FSV_2_1][Language.C][ios_idl_2_1] = os.path.join('src', 'IOS', 'IOS.c')
    implementation_files[FSV_2_0][Language.CPP][ios_idl_2_0] = os.path.join('src', 'IOS', 'ios.cpp')
    implementation_files[FSV_2_1][Language.CPP][ios_idl_2_1] = os.path.join('src', 'IOS', 'IOS.cpp')
    implementation_files[FSV_2_0][Language.Ada][ios_idl_2_0] = os.path.join('src', 'IOS', 'face-io.adb')
    implementation_files[FSV_2_1][Language.Ada][ios_idl_2_1] = os.path.join('src', 'IOS', 'face-io.adb')

    conformance_test_directories[ios_idl_2_0] = 'IOS'
    conformance_test_directories[ios_idl_2_1] = 'IOS'

    # TSS

    # Main TS IDL FILE
    tss_idl_2_0 = 'FACE_tss.idl'
    tss_idl_2_1 = 'FACE_TS.idl'
    idl_files[FSV_2_0].append(tss_idl_2_0)
    idl_files[FSV_2_1].append(tss_idl_2_1)

    declaration_files[FSV_2_0][Language.C][tss_idl_2_0] = os.path.join('include', 'face', 'tss.h')
    declaration_files[FSV_2_1][Language.C][tss_idl_2_1] = os.path.join('include', 'FACE', 'TS.h')
    declaration_files[FSV_2_0][Language.CPP][tss_idl_2_0] = os.path.join('include', 'face', 'tss.hpp')
    declaration_files[FSV_2_1][Language.CPP][tss_idl_2_1] = os.path.join('include', 'FACE', 'TS.hpp')
    declaration_files[FSV_2_0][Language.Ada][tss_idl_2_0] = os.path.join('spec', 'face', 'face-ts.ads')
    declaration_files[FSV_2_1][Language.Ada][tss_idl_2_1] = os.path.join('spec', 'FACE', 'face-ts-ts.ads')

    implementation_files[FSV_2_0][Language.C][tss_idl_2_0] = os.path.join('src', 'TSS', 'TSSGoldLib.c')
    implementation_files[FSV_2_1][Language.C][tss_idl_2_1] = os.path.join('src', 'TSS', 'TS.c')
    implementation_files[FSV_2_0][Language.CPP][tss_idl_2_0] = os.path.join('src', 'TSS', 'TSSGoldLib.cpp')
    implementation_files[FSV_2_1][Language.CPP][tss_idl_2_1] = os.path.join('src', 'TSS', 'TS.cpp')
    implementation_files[FSV_2_0][Language.Ada][tss_idl_2_0] = os.path.join('src', 'TSS', 'face-ts.adb')
    implementation_files[FSV_2_1][Language.Ada][tss_idl_2_1] = os.path.join('src', 'TSS', 'face-ts-ts.adb')

    declaration_footer[FSV_2_1][Language.C][tss_idl_2_1] = '#include <FACE/TypeSpecific_TS.h>\n'
    declaration_footer[FSV_2_1][Language.CPP][tss_idl_2_1] = '#include <FACE/TypeSpecific_TS.hpp>\n'

    declaration_footer[FSV_2_0][Language.C][tss_idl_2_0] = '#include <face/TypeSpecific_TS.h>\n'
    declaration_footer[FSV_2_0][Language.CPP][tss_idl_2_0] = '#include <face/TypeSpecific_TS.hpp>\n'

    conformance_test_directories[tss_idl_2_0] = 'TSS'
    conformance_test_directories[tss_idl_2_1] = os.path.join('TSS', 'FACE_TS')

    # 2.1 has TS_common
    ts_common_idl_2_1 = 'FACE_TS_common.idl'
    idl_files[FSV_2_1].append(ts_common_idl_2_1)

    declaration_files[FSV_2_1][Language.C][ts_common_idl_2_1] = os.path.join('include', 'FACE', 'TS_common.h')
    declaration_files[FSV_2_1][Language.CPP][ts_common_idl_2_1] = os.path.join('include', 'FACE', 'TS_common.hpp')
    declaration_files[FSV_2_1][Language.Ada][ts_common_idl_2_1] = os.path.join('spec', 'FACE', 'face-ts.ads')

    # No implementation files needed for TS_common

    # Conformance tests are handled in the FACE_TS.idl and the FACE_TypeAbstraction_TS.idl, since this is not
    # a standalone interface

    # 2.1 has TypeAbstraction_TS
    type_abstraction_ts_idl_2_1 = 'FACE_TypeAbstraction_TS.idl'
    idl_files[FSV_2_1].append(type_abstraction_ts_idl_2_1)

    declaration_files[FSV_2_1][Language.C][type_abstraction_ts_idl_2_1] = \
        os.path.join('include', 'FACE', 'TypeAbstraction_TS.h')
    declaration_files[FSV_2_1][Language.CPP][type_abstraction_ts_idl_2_1] = \
        os.path.join('include', 'FACE', 'TypeAbstraction_TS.hpp')
    declaration_files[FSV_2_1][Language.Ada][type_abstraction_ts_idl_2_1] = \
        os.path.join('spec', 'FACE', 'face-ts-typeabstraction_ts.ads')

    implementation_files[FSV_2_1][Language.C][type_abstraction_ts_idl_2_1] = \
        os.path.join('src', 'TSS', 'TypeAbstraction_TS.c')
    implementation_files[FSV_2_1][Language.CPP][type_abstraction_ts_idl_2_1] = \
        os.path.join('src', 'TSS', 'TypeAbstraction_TS.cpp')
    implementation_files[FSV_2_1][Language.Ada][type_abstraction_ts_idl_2_1] = \
        os.path.join('src', 'TSS', 'face-ts-typeabstraction_ts.adb')

    conformance_test_directories[type_abstraction_ts_idl_2_1] = os.path.join('TSS', 'TypeAbstraction_TS')

    # HMFM
    hmfm_idl_2_0 = 'FACE_HMFM.idl'
    hmfm_idl_2_1 = 'FACE_HMFM.idl'
    idl_files[FSV_2_0].append(hmfm_idl_2_0)
    idl_files[FSV_2_1].append(hmfm_idl_2_1)

    declaration_files[FSV_2_0][Language.C][hmfm_idl_2_0] = os.path.join('include', 'face', 'hmfm.h')
    declaration_files[FSV_2_1][Language.C][hmfm_idl_2_1] = os.path.join('include', 'FACE', 'HMFM.h')
    declaration_files[FSV_2_0][Language.CPP][hmfm_idl_2_0] = os.path.join('include', 'face', 'hmfm.hpp')
    declaration_files[FSV_2_1][Language.CPP][hmfm_idl_2_1] = os.path.join('include', 'FACE', 'HMFM.hpp')
    declaration_files[FSV_2_0][Language.Ada][hmfm_idl_2_0] = os.path.join('spec', 'face', 'face-hmfm.ads')
    declaration_files[FSV_2_1][Language.Ada][hmfm_idl_2_1] = os.path.join('spec', 'FACE', 'face-hmfm.ads')

    implementation_files[FSV_2_0][Language.C][hmfm_idl_2_0] = os.path.join('src', 'OSS', 'HMFM', 'hmfm.c')
    implementation_files[FSV_2_1][Language.C][hmfm_idl_2_1] = os.path.join('src', 'OSS', 'HMFM', 'HMFM.c')
    implementation_files[FSV_2_0][Language.CPP][hmfm_idl_2_0] = os.path.join('src', 'OSS', 'HMFM', 'hmfm.cpp')
    implementation_files[FSV_2_1][Language.CPP][hmfm_idl_2_1] = os.path.join('src', 'OSS', 'HMFM', 'HMFM.cpp')
    implementation_files[FSV_2_0][Language.Ada][hmfm_idl_2_0] = os.path.join('src', 'OSS', 'HMFM', 'face-hmfm.adb')
    implementation_files[FSV_2_1][Language.Ada][hmfm_idl_2_1] = os.path.join('src', 'OSS', 'HMFM', 'face-hmfm.adb')

    conformance_test_directories[hmfm_idl_2_0] = os.path.join('OSS', 'HMFM')
    conformance_test_directories[hmfm_idl_2_1] = os.path.join('OSS', 'HMFM')  # Technically redundant

    # Common
    common_idl_2_0 = 'FACE_common.idl'
    common_idl_2_1 = 'FACE_common.idl'
    idl_files[FSV_2_0].append(common_idl_2_0)
    idl_files[FSV_2_1].append(common_idl_2_1)

    declaration_files[FSV_2_0][Language.C][common_idl_2_0] = os.path.join('include', 'face', 'common.h')
    declaration_files[FSV_2_1][Language.C][common_idl_2_1] = os.path.join('include', 'FACE', 'common.h')
    declaration_files[FSV_2_0][Language.CPP][common_idl_2_0] = os.path.join('include', 'face', 'common.hpp')
    declaration_files[FSV_2_1][Language.CPP][common_idl_2_1] = os.path.join('include', 'FACE', 'common.hpp')
    declaration_files[FSV_2_0][Language.Ada][common_idl_2_0] = os.path.join('spec', 'face', 'face.ads')
    declaration_files[FSV_2_1][Language.Ada][common_idl_2_1] = os.path.join('spec', 'FACE', 'face.ads')

    # No implementation files needed for common


    # Conformance tests are handled in the FACE_TS.idl and the FACE_TypeAbstraction_TS.idl, and FACE_IOS.idl,
    # since this is not a standalone interface

    gold_include_files = {}
    # Most include files can be derived from included idl files, but a few cannot and must be specified
    gold_include_files[constants.FACE_STANDARD_VERSION_2_0] = {
        Language.C: {
            common_idl_2_0: ['face/types.h']
        },
        Language.CPP: {
            common_idl_2_0: ['face/types.hpp']
        },
        Language.Ada: {
            ios_idl_2_0: ['FACE_Basic_Types'],
            hmfm_idl_2_0:  ['FACE_Basic_Types'],
            common_idl_2_0:  ['FACE_Basic_Types', 'System'],
            tss_idl_2_0: ['FACE_Basic_Types', 'face.dm']
        }
    }
    gold_include_files[constants.FACE_STANDARD_VERSION_2_1] = {
        Language.C: {
            common_idl_2_1: ['FACE/types.h']
        },
        Language.CPP: {
            common_idl_2_1: ['FACE/types.hpp']
        },
        Language.Ada: {
            ios_idl_2_1: ['FACE_Basic_Types'],
            ts_common_idl_2_1:  ['FACE_Basic_Types', 'System'],
            hmfm_idl_2_1:  ['FACE_Basic_Types'],
            common_idl_2_1:  ['FACE_Basic_Types', 'System'],
            tss_idl_2_1: ['FACE_Basic_Types', 'FACE', 'FACE.TS', 'FACE.DM']
        }
    }


    conformance_include_files = {
        Language.C: {},
        Language.CPP: {},
        Language.Ada: {
            tss_idl_2_0: ['FACE_Basic_Types', 'FACE', 'FACE.TS','face.dm'],
            tss_idl_2_1: ['FACE_Basic_Types', 'FACE', 'FACE.TS', 'FACE.TS.TS',
                          'FACE.DM', 'Interfaces.C', 'System'],
            type_abstraction_ts_idl_2_1: ['FACE_Basic_Types', 'FACE', 'FACE.TS', 'FACE.TS.TYPEABSTRACTION_TS',
                          'Interfaces.C', 'System'],
            ios_idl_2_1: ['FACE_Basic_Types', 'FACE', 'FACE.IO', 'Interfaces.C', 'System'],
            ios_idl_2_0: ['FACE_Basic_Types', 'FACE', 'FACE.IO', 'Interfaces.C', 'System'],
            hmfm_idl_2_1: ['FACE_Basic_Types', 'FACE', 'FACE.HMFM', 'System']
        }
    }

    fork_test_idl_files = [ios_idl_2_0, ios_idl_2_1, tss_idl_2_0, tss_idl_2_1, type_abstraction_ts_idl_2_1]

    # MAIN LOOP -- iterate through all standards and languages
    ROOT_DIR = ConformanceMain.getRootDir()

    ada_face_basic_types_conformance_filename = \
        os.path.join(ROOT_DIR, constants.ADA_FACE_BASIC_TYPES_CONFORMANCE_TEST_FILE)

    #TODO: take out the Ada only below here
    #languages = [Language.CPP]
    face_version = constants.getFACEStandardVersion()
    if face_version == 'Version 2.1.1':
        face_version = 'Version 2.1'
    versions = [face_version]

    #idl_files[FSV_2_1] = [ios_idl_2_1] # only doing ios for now...
    #idl_files[FSV_2_1] = [common_idl_2_1]

    for version in versions:
        for lang in languages:

            if lang == Language.Ada:
                template_filename = os.path.join(
                    ROOT_DIR, constants.DATAFILES_DIR, IdlCodeGenerator.conformance_templates[Language.Ada])
                ada_face_basic_types_conformance_test_generator(
                    template_filename, ada_face_basic_types_conformance_filename)

            for idl_file in idl_files[version]:  # iterate through all idl files associated with the version
                logger.debug('ROOT DIR: [{0}], datafiles dir: [{1}], idl_file: [{2}]'.format(
                    ROOT_DIR, constants.DATAFILES_DIR, idl_file
                ))

                full_idl_file_path = os.path.join(ROOT_DIR, constants.DATAFILES_DIR, idl_file)
                face_idl_parser = idl_parser.FaceIdlParser(filename=full_idl_file_path)
                # idl_parser.idl_tree.print_tree()

                gold_root_dir = os.path.join(ROOT_DIR, constants.GOLD_LIB_ROOT, lang)
                idl_code_generator = IdlCodeGenerator(face_idl_parser.idl_tree)
                declaration_file = os.path.join(gold_root_dir, declaration_files[version][lang][idl_file])
                implementation_file = None
                if idl_file in implementation_files[version][lang].keys():
                    implementation_file = os.path.join(
                        gold_root_dir, implementation_files[version][lang][idl_file])

                idl_code_generator.generate_gold_library_code(
                    lang,
                    declarationFile=declaration_file,
                    implementationFile=implementation_file,
                    includeFiles=gold_include_files[version][lang].get(idl_file, None),
                    declarationFooter=declaration_footer[version][lang].get(idl_file, None),
                    version=version
                )

                if lang == Language.Ada and idl_file in [tss_idl_2_1, tss_idl_2_0]:  # VERY SPECIAL CASE
                    commonTools.move(declaration_file, declaration_file + '.TYPE_INDEPENDENT')
                    commonTools.move(implementation_file, implementation_file + '.TYPE_INDEPENDENT')
                    print 'DECLARATION FILE: {0}, IMPLEMENTATION FILE: {1}'.format(declaration_file, implementation_file)
                conformance_root_dir = os.path.join(ROOT_DIR, constants.CONFORMANCE_TEST_ROOT, lang)
                if idl_file in conformance_test_directories.keys():
                    conformance_test_directory = os.path.join(
                        conformance_root_dir, conformance_test_directories[idl_file])

                    # TODO:  At some point make this consistent across all tests
                    assertion_text = 'Test'
                    if idl_file in [hmfm_idl_2_0, hmfm_idl_2_1] and lang != Language.Ada:
                        assertion_text = ''

                    is_type_abstraction = idl_file == type_abstraction_ts_idl_2_1
                    if is_type_abstraction:
                        pass

                    idl_code_generator.generate_conformance_test_code(
                        lang,
                        conformance_test_directory,
                        includeFiles=conformance_include_files[lang].get(idl_file, list()),
                        assertionText=assertion_text,
                        typeAbstractionTS=is_type_abstraction,
                        forkTest=idl_file in fork_test_idl_files
                    )

