#!/usr/bin/env python
"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""
"""
---------------------------------------------------------------------------------
Copyright (c) Vanderbilt University, 2019

ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED

This software, authored by Vanderbilt University under a contract awarded to and
managed by Alion Science and Technology, was funded by the U.S. Government under
Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in
this software.  An unlimited rights license means that the U.S. Government can
use, modify, reproduce, release or disclose computer software in whole or in
part, in any manner, and for any purpose whatsoever, and to have or authorize
others to do so.

Vanderbilt University disclaims all warranties with regard to this software,
including all implied warranties of merchantability and fitness. In no event
shall Vanderbilt University be liable for any special, indirect or consequential
damages or any damages whatsoever resulting from loss of use, data or profits,
whether in an action of contract, negligence or other tortious action, arising
out of or in connection with the use or performance of this software.

May contain products produced under DoD SENSIAC contract HC104705D4000 under the
sponsorship of the Defense Technical Information Center,
ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218.
SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical
Information Center.

---------------------------------------------------------------------------------
"""
import re

class ArgLexer():
  """C function argument lexer"""
  TYPE_VOID         = 0
  TYPE_SIMPLE       = 1
  TYPE_ARRAY        = 2
  TYPE_VA_ARGS      = 3
  TYPE_FNPTR1       = 4
  TYPE_FNPTR2       = 5
  TYPE_EMPTY        = 6
  TYPE_REALFLOATING = 7

  ENDCAP1 = "\s*,"
  ENDCAP2 = "\s*\Z"

  VOID         = ("\s*(void)\s*", TYPE_VOID)
  SIMPLE       = ("\s*([\w\s*\*]+[\w\*])", TYPE_SIMPLE)
  ARRAY        = ("\s*(\w[\w\s*\*]+[\w\*]\s*\[[\w\s]*\])", TYPE_ARRAY)
  VA_ARGS      = ("\s*(\.\.\.)", TYPE_VA_ARGS)
  FN_PTR1       = ("([\w\s\*]+\(\s*\*\s*\)\s*\([\w\s,\*]*?\))", TYPE_FNPTR1)
  FN_PTR2       = ("([\w\s\*]+\(\s*\*\s*\)\s*\([\w\s,()\*]*\))", TYPE_FNPTR2)
  EMPTY        = ("\s*()", TYPE_EMPTY)
  REALFLOATING = ("\s*(real-floating\s+\w+)", TYPE_REALFLOATING)

  MINIPATTERNS = (VOID, SIMPLE, ARRAY, VA_ARGS, FN_PTR1, FN_PTR2, EMPTY, REALFLOATING )

  @staticmethod
  def typeToString(type):
    if type == ArgLexer.TYPE_SIMPLE:
      return "simple"
    elif type == ArgLexer.TYPE_ARRAY:
      return "array"
    elif type == ArgLexer.TYPE_VA_ARGS:
      return "va_args"
    elif type == ArgLexer.TYPE_FNPTR1:
      return "function pointer"
    elif type == ArgLexer.TYPE_FNPTR2:
      return "function pointer"
    elif type == ArgLexer.TYPE_VOID:
      return "void"
    elif type == ArgLexer.TYPE_EMPTY:
      return "empty"
    elif type == ArgLexer.TYPE_REALFLOATING:
      return "real-floating"
    else:
      return "unknown"

  def __init__(self):
    self.pattern = "%s%s|%s%s" % (ArgLexer.MINIPATTERNS[0][0], ArgLexer.ENDCAP1, ArgLexer.MINIPATTERNS[0][0], ArgLexer.ENDCAP2)
    for idx in range(1, len(ArgLexer.MINIPATTERNS)):
      self.pattern += "|%s%s|%s%s" % (ArgLexer.MINIPATTERNS[idx][0], ArgLexer.ENDCAP1, ArgLexer.MINIPATTERNS[idx][0], ArgLexer.ENDCAP2)

  def split(self, args):
    """Split a set of C function arguments into each argument.
       Return a two-tuple as the result. The first entry is Boolean
       representing successful parsing. The second entry is the
       parsed result on success or an error message on failure.
       The parsed result is a tuple of tuples containing the parsed
       value and its type.
    """
    idx = 0
    args_len = len(args)

    result = (True, [])

    while idx < args_len:
      remaining_args = args[idx:args_len]
      m = re.match(self.pattern, remaining_args)

      if not m:
        return (False, "unable to parse \'%s\'" % (remaining_args,))

      result[1].append( (m.group(m.lastindex), int(round(m.lastindex / 2.0)) - 1) )

      idx += m.end() + 1

    return result

def test_args_equal(l1, l2):
  """Test args list l1 equals args list l2 using a deep inspection method"""
  if len(l1) != len(l2):
    return (False, "args lists are different lengths")

  for idx in range(len(l1)):
    if l1[idx][0] != l2[idx][0] or l1[idx][1] != l2[idx][1]:
      return (False, "failed on %s (%s) != %s (%s)" % (l1[idx][0], l1[idx][1], l2[idx][0], l2[idx][1]))

  return (True, "Passed")

# Perform some testing on the ArgLexer
if __name__ == "__main__":
  tests = \
    (
      {
        "name" : "simple1",
        "test" : "int, long, char*",
        "result" : [ ("int", ArgLexer.TYPE_SIMPLE), ("long", ArgLexer.TYPE_SIMPLE), ("char*", ArgLexer.TYPE_SIMPLE) ],
      },
      {
        "name" : "simple2",
        "test" : " int, long, char*",
        "result" : [ ("int", ArgLexer.TYPE_SIMPLE), ("long", ArgLexer.TYPE_SIMPLE), ("char*", ArgLexer.TYPE_SIMPLE) ],
      },
      {
        "name" : "array1",
        "test" : "int [ ], long[], char* []",
        "result" : [ ("int [ ]", ArgLexer.TYPE_ARRAY), ("long[]", ArgLexer.TYPE_ARRAY), ("char* []", ArgLexer.TYPE_ARRAY) ],
      },
      {
        "name" : "array2",
        "test" : " int [ ], long[], char* []",
        "result" : [ ("int [ ]", ArgLexer.TYPE_ARRAY), ("long[]", ArgLexer.TYPE_ARRAY), ("char* []", ArgLexer.TYPE_ARRAY) ],
      },
      {
        "name" : "simple_array_mixed1",
        "test" : "int [ ], long *, short* []",
        "result" : [ ("int [ ]", ArgLexer.TYPE_ARRAY), ("long *", ArgLexer.TYPE_SIMPLE), ("short* []", ArgLexer.TYPE_ARRAY) ],
      },
      {
        "name" : "va_args1",
        "test" : "int, ...",
        "result" : [ ("int", ArgLexer.TYPE_SIMPLE), ("...", ArgLexer.TYPE_VA_ARGS), ],
      },
      {
        "name" : "va_args2",
        "test" : "...",
        "result" : [ ("...", ArgLexer.TYPE_VA_ARGS), ],
      },
      {
        "name" : "func_ptr1",
        "test" : "int, void (*) (int*, long)",
        "result" : [ ("int", ArgLexer.TYPE_SIMPLE), ("void (*) (int*, long)", ArgLexer.TYPE_FNPTR1), ],
      },
      {
        "name" : "func_ptr2",
        "test" : "int, void ( * ) (int*, long), char* []",
        "result" : [ ("int", ArgLexer.TYPE_SIMPLE), ("void ( * ) (int*, long)", ArgLexer.TYPE_FNPTR1), ("char* []", ArgLexer.TYPE_ARRAY) ],
      },
      {
        "name" : "func_ptr3",
        "test" : "int, void ( * ) (int*, void(*)(int) ), char* []",
        "result" : [ ("int", ArgLexer.TYPE_SIMPLE), ("void ( * ) (int*, void(*)(int) )", ArgLexer.TYPE_FNPTR2), ("char* []", ArgLexer.TYPE_ARRAY) ],
      },
      {
        "name" : "func_ptr4",
        "test" : "void (*)(void), void (*)(void), void(*)(void)",
        "result" : [ ("void (*)(void)", ArgLexer.TYPE_FNPTR1), ("void (*)(void)", ArgLexer.TYPE_FNPTR1), ("void(*)(void)", ArgLexer.TYPE_FNPTR1) ],
      },
      {
        "name" : "void1",
        "test" : "void",
        "result" : [ ("void", ArgLexer.TYPE_VOID) ],
      },
    )

  lex = ArgLexer()

  all_passed = True
  for test in tests:
    args_result = lex.split(test["test"])
    if not args_result[0]:
      all_passed = False
      print "Test (%s) FAILED: %s" % (test["name"], args_result[1])
      continue
    retval = test_args_equal(test["result"], args_result[1])
    if not retval[0]:
      all_passed = False
      print "Test (%s) FAILED: %s" % (test["name"], retval[1])
      continue
  if all_passed:
     print "All Passed"
