"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""

__author__ = 'david.fann@vanderbilt.edu'

import threading
import SocketServer
import socket
import logging
import struct
import os
import sys
import datetime

import constants
import ConformanceMain
egg_path = os.path.join(ConformanceMain.getRootDir(),'face_conformance_app', 'protobuf-2.5.0-py2.7.egg')
sys.path.append(egg_path)
egg_path = os.path.join(ConformanceMain.getRootDir(),'face_conformance_app', 'setuptools-0.6c11-py2.7!\\')
sys.path.append(egg_path)

import google.protobuf


if __name__ == '__main__':
    ConformanceMain.initializeLogger()

logger = logging.getLogger(__name__)


class ProtocolBufferSocketServer():
    """Creates a TCP SocketServer that handles Google's protocol buffer messages (protobuf)

    ProtocolBufferSocketServer is a class that opens a TCP socket server that expects a binary Protocol Buffer message
    (Google Protobuf) to be sent across the socket.  You pass the protobuf message instance, and the handler calls
    ParseFromString on the data buffer that gets sent across the socket.  The format for the socket message is a
    4 byte signed integer of the data buffer size, followed by the protocol buffer (the result of SerializeToString
    in python...similar methods in other languages).

    Important class data members:

    handler_called - (True/False) True if the socket handler has been called (a message has been sent across the socket)
    lock - (threading.Lock) Lock is acquired by the server instance during initialization and held until the socket
            handler is finished.  Useful to hold up the main thread till the buffer value is parsed.
    ip - ip address returned by the socket server (default is local host)
    port - port being used by socket server

    Important class methods:
    finish - called to release lock and shutdown server (this is automatically called by the handler, but if the handler
             never receives a message, it can be called outside the instance to shut down the server correctly.

    Intended method of operation:

    proto_message_instance = Proto_pb2.message_name()
    protobuf_server = ProtocolBufferSocketServer(proto_message_instance)     #  Main thread instantiates the
                                                                             #  ProtocolBufferSocketServer instance.
    p = subprocess.Popen("command   -p " + str(protobuf_server.port) + ... ) #  Main thread forks off the program that
    stdout, stderr = p.communicate()                                         #  will communicate back a protobuf message
                                                                             # (passing the ip/port via command line)

    if protobuf_server.handler_called:  # program actually communicated back via socket
        proto_server.lock.acquire()     # wait till the handler finishes before continuing

    parameters:

    host : (localhost) Host of the socket server
    port : (0) port for socket server to use (0 means arbitrary unused port)
    timeout : (constants.SOCKET_SERVER_TIMEOUT_IN_SECONDS) timeout to wait in handler for data buffer to be
              finished once the handler has been called.  Using serve_forever on server, with no timeouts on it

    """
    def __init__(self, protocol_buffer_instance, **params):

        self.protocol_buffer_instance = protocol_buffer_instance
        self.time_out = params.pop('timeout', constants.SOCKET_SERVER_TIMEOUT_IN_SECONDS)
        self.save_buffer = params.pop('save_buffer', False)
        self.handler_called = False

        self.lock = threading.Lock()
        self.lock.acquire()

        self.server = self._initialize_socket_server(**params)
        self.ip, self.port = self.server.server_address
        self.server_running = False
        self.data_buffer = None

    def _initialize_socket_server(self, **params):

        host = params.get("host", "localhost")
        port = params.get("port", 0)     # Port 0 means to select an arbitrary unused port

        server = SocketServer.TCPServer((host, port), _ProtocolHandler)
        server.protocol_server_instance = self

        # Start a thread with the server -- that thread will then start one
        # more thread for each request
        server_thread = threading.Thread(target=server.serve_forever)
        # Exit the server thread when the main thread terminates
        server_thread.daemon = True
        server_thread.start()
        logger.debug('Socket Server initialized ' + str(server.server_address))
        self.server_running = True
        return server

    def finish(self):
        try:
            logger.debug('freeing lock')
            self.lock.release()
            logger.debug('shutting down socket server')
            if self.server_running:
                self.server.shutdown()
                self.server_running = False
        except Exception as x:
            err_msg = "Error while shutting down socket server: " + type(x) +  x.message
            logger.error(err_msg)
            raise x


        logger.debug('server shutdown')


class _ProtocolHandler(SocketServer.BaseRequestHandler):
    """
    Internal class to the module.  Set as the TCP handler for the ProtocolBufferSocketServer (PBSS).
    The format for the socket message is a 4 byte signed integer of the data buffer size (Network byte order),
    followed by the protocol buffer (the result of SerializeToString in python...similar methods in other languages).

    Calls the ParseFromString of the protobuf message instance sent as an argument to the PBSS.
    First sets the handler_called to True.  Once finished receiving data buffer and parsing it, releases the PBSS lock
    and shuts down the TCP socket server.
    """

    def handle(self):
        self.handle_lock = threading.Lock()
        self.handle_lock.acquire()

        self.server.protocol_server_instance.handler_called = True
        logger.info('Handle called')
        start_time = datetime.datetime.now()

        # read 4 bytes to get size of buffer
        data_buffer = ''
        while len(data_buffer) < 4:
            data_buffer = self.request.recv(4-len(data_buffer))
            self._check_timeout(start_time)
        # convert them to integer using network byte order
        bytes_to_read = struct.unpack("!i", data_buffer)[0]
        # data_buffer is binary so outputting this can cause issues for editors reading the file
        #logger.debug('data_buffer [' + str(data_buffer) + '] buffer size = ' + str(bytes_to_read))
        logger.debug('buffer size = ' + str(bytes_to_read))


        # read size of buffer
        data_buffer = ''
        while bytes_to_read > 0:
            requested_size = min(1024, bytes_to_read)
            data_received = self.request.recv(requested_size)
            bytes_to_read -= len(data_received)
            data_buffer += data_received
            self._check_timeout(start_time)

        logger.debug('Received ' + str(len(data_buffer)) + ' bytes')
        write_protobuffer_to_string = False
        if write_protobuffer_to_string:
            dummyFile = open(os.path.join(ConformanceMain.getRootDir(), 'protobuffer.string'), 'w')
            dummyFile.write(data_buffer)
            dummyFile.close()

        if self.server.protocol_server_instance.save_buffer:
            self.server.protocol_server_instance.data_buffer = data_buffer

        else:
            logger.debug('Parsing data buffer')
            try:
                self.server.protocol_server_instance.protocol_buffer_instance.ParseFromString(data_buffer)
                logger.debug('Parsing successful')
            except Exception as x:
                logger.error('Parsing failed: ' + x.message)
                raise x

        self.handle_lock.release()

    def finish(self):

        logger.debug('Waiting for handler to complete before finishing server')
        self.handle_lock.acquire()
        self.handle_lock.release()
        logger.debug('finishing server instance')
        self.server.protocol_server_instance.finish()


    def _check_timeout(self, start_time):
        delta_time = datetime.datetime.now() - start_time
        if delta_time.seconds > self.server.protocol_server_instance.time_out:
            err_msg = 'socket server handler timed out waited ' \
                      + str(self.server.protocol_server_instance.time_out) + ' seconds'
            logger.error(err_msg)
            raise RuntimeError(err_msg)


def protocol_buffer_send_by_tcp_socket(host, port, protocol_buffer_instance):
    """
    Used to send a protocol buffer message instance via sockets to the supplied host/port.

    """
    logger.debug(protocol_buffer_instance.__class__.__name__)

    logger.debug(str(protocol_buffer_instance))

    '''
    # something like this should be able to be used to make sure the right argument gets passed, but this isn't
    # quite right.

    if not isinstance(protocol_buffer_instance, google.protobuf.Message):
        err_msg = "Protocol Buffer Instance should be a protocol buffer message instance, but is " + \
                 protocol_buffer_instance.__class__.__name__
        logger.error(err_msg)
        raise TypeError(err_msg)
    '''

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect((host, port))
    buffer = protocol_buffer_instance.SerializeToString()
    total_length = len(buffer)
    logger.debug('total_length: ' + str(total_length))
    total_length_network_int = struct.pack("!i", total_length)
    s.send(total_length_network_int)

    length_sent = 0
    while length_sent < total_length:
        bytes_sent = s.send(buffer[length_sent:])
        if bytes_sent == 0:
            err_msg = "socket connection broken"
            logger.error(err_msg)
            raise RuntimeError(err_msg)
        length_sent += bytes_sent


if __name__ == '__main__':
    '''
    Test program that determines if the ProtocolBufferSocketServer class and protocol_buffer_send_by_tcp_socket method
    work correctly, using an example protocol buffer used elsewhere in the test suite
    '''
    import datamodel_transfer_pb2
    import time

    datamodel_protobuf_received = datamodel_transfer_pb2.DataModel()
    datamodel_protobuf_sent = datamodel_transfer_pb2.DataModel()

    uop = datamodel_protobuf_sent.unitOfPortability.add()
    uop.name = "Dummy UoP"
    message = uop.messageType.add()
    message.name = "Dummy Message"

    logger.debug('Running from protocol_buffer_socket_server')

    protocol_server = ProtocolBufferSocketServer(datamodel_protobuf_received)

    logger.debug('Sending protocol_buffer over socket')
    protocol_buffer_send_by_tcp_socket(protocol_server.ip, protocol_server.port, datamodel_protobuf_sent)

    logger.debug('Waiting for lock to free')
    protocol_server.lock.acquire()

    logger.debug('This is what I got: ' + str(datamodel_protobuf_received))

    # Waiting for the server thread to finish...
    time.sleep(2)
