#!/usr/bin/env python
"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""

__author__ = 'david.fann@vanderbilt.edu'

import os
import datetime
import glob
import logging
import optparse

import ConformanceMain

if __name__ == '__main__':
    ConformanceMain.initializeLogger()

import constants
import TestConfiguration
import commonTools

logger = logging.getLogger(__name__)


class DistroLanguages(TestConfiguration.Language):
    Python = 'Python'
    LicenseFile = 'LicenseFile'

    def __init__(self):
        TestConfiguration.Language.__init__(self)

def get_distribution_statement(file_type, include_this_filename=None, **params):

    generated_file = params.get('generated_file', True)
    include_date = params.get('include_date', True)
    include_version = params.get('include_version', True)

    """
    Read the datafiles/distributionStatement.txt file and format to a comment for the fileType given (makefile, c or c++)
    """

    distro_statement_filename = os.path.join(ConformanceMain.getRootDir(), constants.DISTRIBUTION_FILE)

    if not os.path.isfile(distro_statement_filename):
        raise IOError('Could not find distribution statement file: ' + distro_statement_filename)

    # Read the Distribution file and format to a comment for the fileType given (makefile, c, c++, Ada)
    distro_file = commonTools.safeOpen(distro_statement_filename, 'r')
    distro_lines = distro_file.readlines()

    if not include_this_filename is None:
        distro_lines.extend(['\n', 'Filename: ' + include_this_filename + '\n'])

    if include_version:
        version_info = constants.getVersionedFACETestSuiteName()
        distro_lines.extend(['\n', 'Conformance Test Suite ' + version_info.toolVersion +
                             ' supporting FACE Standard ' + version_info.faceStandardVersion + '\n'])

    if generated_file:
        distro_lines.extend(['\n', 'This file is generated and may be overwritten.\n'])

    if include_date:
        distro_lines.extend(['\n', 'Release Date: ' + datetime.date.today().strftime('%d-%b-%Y') + '\n'])

    max_length = 0
    for line in distro_lines:
        if len(line) > max_length:
            max_length = len(line)

    if file_type == DistroLanguages.Make:
        top_border = (max_length + 3) * '#' + '\n'
        bottom_border = top_border
        left_border = '# '
        right_border = ' #\n'
    elif file_type == DistroLanguages.C or file_type == DistroLanguages.CPP:
        top_border = '/' + (max_length + 3) * '*' + '\n'
        bottom_border = ' ' + (max_length + 3) * '*' + '/\n'
        left_border = ' * '
        right_border = ' *\n'

    elif file_type == DistroLanguages.LicenseFile:
        top_border = '"/' + (max_length + 3) * '*' + '\\n\\\n'
        bottom_border = ' ' + (max_length + 3) * '*' + '/"\n'
        left_border = ' * '
        right_border = ' *\\n\\\n'

    elif file_type == DistroLanguages.Ada:
        top_border = (max_length + 3) * '-' + '\n'
        bottom_border = top_border
        left_border = '-- '
        right_border = '\n'

    elif file_type == DistroLanguages.Python:
        top_border = '"""\n' + (max_length + 3) * '-' + '\n'
        bottom_border = (max_length + 3) * '-' + '\n"""\n'
        left_border = 3 * ' '
        right_border = '\n'

    else:
        top_border = '\n'
        bottom_border = '\n'
        left_border = ''
        right_border = '\n'

    returnString = top_border
    for line in distro_lines:
        returnString += left_border + line[:-1] + (max_length - len(line)) * ' ' + right_border

    return returnString + bottom_border


def set_make_devel_distribution_statements():
    """
    Function sets the distribution statement everywhere it is needed before running make_devel to build
    out the conformance test suite.  (sets the LICENSE for the C and C++ builds, will add more later if needed)
    :return: None
    """
    license_filename = os.path.join(ConformanceMain.getRootDir(), constants.C_CPP_GENERATOR_LICENSE_FILE)

    license_file = open(license_filename, 'w')

    pre_distro_statement = '#ifndef LICENSE_H\n#define LICENSE_H\n\n#define GEN_LICENSE \\\n'
    post_distro_statement = '\n\n#endif // LICENSE_H\n'
    license_file.write(pre_distro_statement)

    distro_statement = get_distribution_statement(DistroLanguages.LicenseFile)

    license_file.write(distro_statement)
    license_file.write(post_distro_statement)
    license_file.close()

def set_make_dist_distribution_statements():
    """
    Function sets the distribution statement everywhere it is needed before running make_dist to build
    a release of the the conformance test suite (Prepends the distribution statement to all scripts and makefiles)
    :return: None
    """

    release_dir = constants.getReleaseDirectory()


    python_script_dir = os.path.join(release_dir, constants.APP_DIR)
    if not os.path.isdir(python_script_dir):
        err_msg = 'Python script dir: ' + python_script_dir + ' does not exist'
        logger.error(err_msg)
        raise RuntimeError(err_msg)

    distro_statement = get_distribution_statement(DistroLanguages.Python, generated_file=False)

    temp_suffix = '_temp'
    ignored_files = ['antlr.py', 'DataModelTestResult_pb2.py', 'datamodel_transfer_pb2.py', 'face_conformance_pb2.py',
                    '__init__.py', 'JavaTypes_pb2.py', 'make.py', 'mkparse.py', 'pkg_resources.py', 'TestResult_pb2.py']

    python_script_filenames = glob.glob(os.path.join(python_script_dir, '*.py'))
    for py_filename in python_script_filenames:
        if os.path.basename(py_filename) in ignored_files:
            continue # skip the ignored file

        source_py_file = open(py_filename, 'r')
        target_py_file = open(py_filename + temp_suffix, 'w')

        first_line_written = False
        first_line = source_py_file.readline()

        if first_line.startswith('#!'):  # shebang...want to keep it as the first line
            target_py_file.write(first_line)
            target_py_file.write(distro_statement)
        else:
            target_py_file.write(distro_statement)
            target_py_file.write(first_line)

        # write the rest of the source file
        for line in source_py_file:
            target_py_file.write(line)

        target_py_file.close()
        source_py_file.close()

        # Now move the file with the distribution statement to the old filename
        os.remove(py_filename)
        os.rename(py_filename + temp_suffix, py_filename)


    # Write a distribution statement in the main directory
    distro_file = open(os.path.join(release_dir, 'DISTRIBUTION_STATEMENT.txt'), 'w')
    distro_file.write(get_distribution_statement(''))
    distro_file.close()



if __name__ == '__main__':

    parser = optparse.OptionParser(usage="usage: %prog [options] ")

    # TODO: add this option to the parser to run changeProjectDirectory before running test suite
    parser.add_option("-d", "--devel", action="store_true",
                      help="Sets the distribution statement from the datafiles/distributionStatement.txt to every where " +
                           "it is needed before running the rest of the make_devel_ready script")

    parser.add_option("-r", "--release", action="store_true",
                      help="Sets the distribution statement from the datafiles/distributionStatement.txt to "
                           "everywhere it is needed in the release directory. " +
                           "Called before creating tarball/zipfile for release.")


    (options, args) = parser.parse_args()

    if options.devel and options.release:
        parser.error("options -d and -r are mutually exclusive")

    if options.devel:
        logger.info('make_devel_ready distribution statements')
        set_make_devel_distribution_statements()

    elif options.release:
        logger.info('release distribution statements')
        set_make_dist_distribution_statements()

    else:
        logger.info('printing distribution statement')
        distro_statement = get_distribution_statement('',
                                                      generated_file=True,
                                                      include_date=True,
                                                      include_version=True,
                                                      )

        print distro_statement









