#!/usr/bin/env python
"""
------------------------------------------------------------------------------------
   Copyright (c) Vanderbilt University, 2020                                       
                                                                                   
   ALL RIGHTS RESERVED, UNLESS OTHERWISE STATED                                    
                                                                                   
   This software, authored by Vanderbilt University under a contract awarded to and
   managed by Alion Science and Technology, was funded by the U.S. Government under
   Contract No. FA8075-14-D-0014 and the U.S. Government has unlimited rights in   
   this software.  An unlimited rights license means that the U.S. Government can  
   use, modify, reproduce, release or disclose computer software in whole or in    
   part, in any manner, and for any purpose whatsoever, and to have or authorize   
   others to do so.                                                                
                                                                                   
   Vanderbilt University disclaims all warranties with regard to this software,    
   including all implied warranties of merchantability and fitness. In no event    
   shall Vanderbilt University be liable for any special, indirect or consequential
   damages or any damages whatsoever resulting from loss of use, data or profits,  
   whether in an action of contract, negligence or other tortious action, arising  
   out of or in connection with the use or performance of this software.           
                                                                                   
   May contain products produced under DoD SENSIAC contract HC104705D4000 under the
   sponsorship of the Defense Technical Information Center,                        
   ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA 22060-6218. 
   SENSIAC is a DoD Information Analysis Center Sponsored by the Defense Technical 
   Information Center.                                                             
                                                                                   
   Conformance Test Suite Version 2.1.1r9 supporting FACE Standard Version 2.1.1   
                                                                                   
   Release Date: 29-Aug-2023                                                       
------------------------------------------------------------------------------------
"""

__author__ = 'John Williams <johnwilliams@isis.vanderbilt.edu>'

import Tkinter
import tkFont

class TabWidget(Tkinter.Frame):
    """Tab Widget for Tkinter"""

    def __init__(self, master=None, cnf={}, **kw):
        Tkinter.Frame.__init__(self, master, cnf, **kw)

        self.__currentIndex = None
        """Index of the tab currently shown or None if no tabs exist"""

        self.__buttons = list()
        """List of buttons used for tab selection"""
        self.__widgets = list()
        """List of widgets which are the content for tabs"""

        self.__tabBar = Tkinter.Frame(self, bd=2, relief=Tkinter.SUNKEN)
        """Container for tab buttons"""
        self.__tabBar.pack(side="top", fill="x")

        self.__contentPane = Tkinter.Frame(self, bd=2, relief=Tkinter.RAISED)
        """Container for the currently shown tab"""
        self.__contentPane.pack(side="bottom", fill="both", expand=True)

    def appendTab(self, **kw):
        """Append a tab to the current list.

        Args:
          text - text on the tab button (optional)
          image - icon on the tab button (optional)

        Returns: Tkinter.Frame that is the new tab. Apply your widgets here.
        """

        text = kw.get("text", None)
        image = kw.get("image", None)
        #width = kw.get("width", None)  # May want to do this at a later date...

        button = Tkinter.Button(self.__tabBar, text=text, image=image, relief=Tkinter.GROOVE)
        button.pack(side="left", padx=2)
        button.bind("<Button-1>", self.__tabBtnClicked)
        self.__buttons.append(button)

        widget = Tkinter.Frame(self.__contentPane)
        self.__widgets.append(widget)

        # If it's the first widget, select it. If not, merely append and
        # do not select.
        if self.__currentIndex is None:
            self.setCurrentIndex(0)

        return widget

    def indexOf(self, widget):
        """Returns the tab index of the widget.

        Raises ValueError if widget cannot be found.
        """
        return self.__widgets.index(widget)

    def setCurrentIndex(self, index):
        """Set the currently shown tab by its index."""
        widget = self.__widgets[index]

        self.__currentIndex = index
        self.__showTabWidget(widget)
        self.__showTabBtn(index)

    def setCurrentWidget(self, widget):
        """Set the currently shown tab by its widget"""
        try:
            index = self.indexOf(widget)

            self.__currentIndex = index
            self.__showTabWidget(widget)
            self.__showTabBtn(index)
        except ValueError:
            return

    def __showTabWidget(self, widget):
        for hideme in self.__widgets:
            hideme.forget()

        widget.pack(fill="both", expand=True)

    def __showTabBtn(self, index):
        for not_selected in self.__buttons:
            font = tkFont.Font(family="Helvetica", size=8, weight=tkFont.NORMAL)
            not_selected.config(font=font)

        font = tkFont.Font(family="Helvetica", size=8, weight=tkFont.BOLD)
        self.__buttons[index].config(font=font)

    def setTabText(self, widget, text):
        try:
            index = self.indexOf(widget)
            self.__buttons[index].config(text=text)
        except ValueError:
            return

    def getTabText(self, widget):
        try:
            index = self.indexOf(widget)
            configs = self.__buttons[index].config()
            if len(configs['text']) > 0:
                return_val = configs['text'][-1]
                if not isinstance(return_val, basestring):
                    # SO WEIRD PYTHON BUG...python 2.7.6 maybe more on Windows
                    # If I put the checkmark at the end of the tab, instead of returning a string,
                    # it returns a tuple, with text in the first element and a UTF-8 string for the
                    # checkmark in the second element.  So I have to do this handling for this special case
                    #
                    new_value = ''
                    for r in return_val:
                        new_value += r.decode('utf-8')
                    return_val = new_value

                return return_val
            else:
                return ''
        except ValueError as x:
            print 'Error: ' + x.message
            return


    def __tabBtnClicked(self, event=None):
        button = event.widget
        index = self.__buttons.index(button)
        self.setCurrentIndex(index)



# Test driver for TabWidget
if __name__ == "__main__":
    root = Tkinter.Tk()

    frame = Tkinter.Frame(root)
    frame.pack(fill="both", expand=True)

    tabs = TabWidget(frame)
    tabs.pack(fill="both", expand=True)

    tab1 = tabs.appendTab(text="tab1")
    btn1 = Tkinter.Button(tab1, text="button1")
    btn1.pack()

    tab2 = tabs.appendTab(text="Tab 2")
    btn2 = Tkinter.Button(tab2, text="Button 2")
    btn2.pack()

    tab3 = tabs.appendTab(text="Tab 3")
    btn3 = Tkinter.Button(tab3, text="Button 3")
    btn3.pack()

    tab4 = tabs.appendTab(text="Tab 4")
    btn4 = Tkinter.Button(tab4, text="Button 4")
    btn4.pack()

    root.mainloop()