with Ada.Unchecked_Conversion;

--------------------
-- Element_Arrays --
--------------------
separate (FACE.Sequences.Unbounded)
package body Element_Arrays is

   procedure Free_Element_Array is new
     Ada.Unchecked_Deallocation (Object => Element_Array,
                                 Name   => Element_Array_Access);

   ---------
   -- "&" --
   ---------
   function "&"
     (Left  : in Sequence;
      Right : in Element_Array)
      return Sequence
   is
      Length : constant Integer := Left.SR.Current_Length;
   begin
      return New_Sequence (Left.SR.Content (1 .. Length) & Sequence_Vector (Right (Right'Range)));
   end "&";

   ---------
   -- "&" --
   ---------
   function "&"
     (Left  : in Element_Array;
      Right : in Sequence)
      return Sequence
   is
      Length : constant Integer := Right.SR.Current_Length;
   begin
      return New_Sequence (Sequence_Vector (Left (Left'Range)) & Right.SR.Content (1 .. Length));
   end "&";

   ---------
   -- "*" --
   ---------
   function "*"
     (Left  : in Length_Range;
      Right : in Element_Array)
      return Sequence
   is
      New_Length : constant Length_Range := Left * Right'Length;
      Seq_Low    : Natural := 1;
      Seq_High   : Natural;
      New_Vector : Sequence_Vector (1 .. New_Length);  -- Allocate on heap if this blows stack.
   begin
      Perform_The_Replication:
      for I in 1 .. Left loop
         Seq_High := Seq_Low + Right'Length - 1;

         -- Store a copy of Right in the new sequence.
         New_Vector (Seq_Low .. Seq_High) := Sequence_Vector (Right (Right'Range));

         Seq_Low := Seq_Low + Right'Length;
      end loop Perform_The_Replication;

      return New_Sequence (New_Vector);
   end "*";

   ---------
   -- "=" --
   ---------
   function "="
     (Left  : in Element_Array;
      Right : in Sequence)
      return Boolean
   is
   begin
      return Boolean ((Left /= Null_Element_Array and Right.SR /= null) and then
                        (Sequence_Vector (Left (Left'First .. Left'Last)) =
                           Right.SR.Content (1 .. Right.SR.Current_Length)));
   end "=";

   ---------
   -- "=" --
   ---------
   function "="
     (Left  : in Sequence;
      Right : in Element_Array)
      return Boolean
   is
   begin
      return Right = Left;
   end "=";

   ------------
   -- Append --
   ------------
   procedure Append
     (Source   : in out Sequence;
      New_Item : in Element_Array)
   is
      Length : constant Integer := Source.SR.Current_Length;
   begin
      New_SVec (Old_SR      => Source.SR,
                New_Content => Source.SR.Content (1 .. Length) & Sequence_Vector (New_Item (New_Item'Range)),
                New_SR      => Source.SR);
   end Append;

   -----------
   -- Count --
   -----------
   function Count
     (Source  : in Sequence;
      Pattern : in Element_Array)
      return Length_Range
   is
      Local_Count   : Length_Range := 0;
      Source_Length : constant Length_Range := Source.SR.Current_Length;
   begin
      if Pattern'Length = 0 then
         raise Pattern_Error;
      elsif Pattern'Length > Source_Length then
         raise Length_Error;
      else
         for Index in 1 .. Source_Length + 1 - Pattern'Length loop
            if Element_Array (Source.SR.Content (Index .. Index + Pattern'Length - 1)) = Pattern then
               Local_Count := Local_Count + 1;
            end if;
         end loop;
         return Local_Count;  -- The pattern was not found (= 0).
      end if;
   end Count;

   ----------
   -- Free --
   ----------
   procedure Free (X : in out Element_Array_Access) is
   begin
      Free_Element_Array (X);
   end Free;

   -----------
   -- Index --
   -----------
   function Index
     (Source  : in Sequence;
      Pattern : in Element_Array;
      Going   : in Direction := Forward)
      return Length_Range
   is
      Source_Length : constant Length_Range := Source.SR.Current_Length;
      The_Pattern_Was_Not_Found : constant Length_Range := Length_Range'First;
   begin
      if Pattern'Length = 0 then
         raise Pattern_Error;
      elsif Pattern'Length > Source_Length then
         raise Length_Error;
      else
         case Going is

            when Forward =>
               for Index in 1 .. Source_Length + 1 - Pattern'Length loop
                  if Element_Array (Source.SR.Content (Index .. Index + Pattern'Length - 1)) = Pattern then
                     return Index;
                  end if;
               end loop;

            when Backward =>

               for Index in reverse 1 .. Source_Length + 1 - Pattern'Length loop
                  if Element_Array (Source.SR.Content (Index .. Index + Pattern'Length - 1)) = Pattern then
                     return Index;
                  end if;
               end loop;
         end case;

         return The_Pattern_Was_Not_Found;
      end if;
   end Index;

   ------------
   -- Insert --
   ------------
   function Insert
     (Source   : in Sequence;
      Before   : in Index_Range;
      New_Item : in Element_Array)
      return Sequence
   is
      Source_Length : constant Integer := Source.SR.Current_Length;
   begin
      if Before not in 1 .. Source.SR.Current_Length + 1 then
         raise Index_Error;
      else
         return New_Sequence (Source.SR.Content (1 .. Before - 1)           &
                                Sequence_Vector (New_Item (New_Item'Range)) &
                                Source.SR.Content (Before .. Source_Length));
      end if;
   end Insert;

   ------------
   -- Insert --
   ------------
   procedure Insert
     (Source   : in out Sequence;
      Before   : in Index_Range;
      New_Item : in Element_Array)
   is
   begin
      Source := Insert (Source, Before, New_Item);
   end Insert;

   ---------------
   -- Overwrite --
   ---------------
   function Overwrite
     (Source   : in Sequence;
      Position : in Index_Range;
      New_Item : in Element_Array)
      return Sequence
   is
   begin
      return Replace_Slice (Source => Source,
                            Low    => Position,
                            High   => Position + New_Item'Length - 1,
                            By     => New_Item);
   end Overwrite;

   ---------------
   -- Overwrite --
   ---------------
   procedure Overwrite
     (Source   : in out Sequence;
      Position : in Index_Range;
      New_Item : in Element_Array)
   is
   begin
      Source := Overwrite (Source, Position, New_Item);
   end Overwrite;

   -------------------
   -- Replace_Slice --
   -------------------
   function Replace_Slice
     (Source : in Sequence;
      Low    : in Index_Range;
      High   : in Index_Range;
      By     : in Element_Array)
      return Sequence
   is
      Source_Length : constant Integer := Source.SR.Current_Length;
   begin
      If Low > Source_Length + 1 or High < 0 then raise Index_Error;
      elsif High >= Low then
         return New_Sequence (Source.SR.Content (1 .. Low - 1)  &
                                Sequence_Vector (By (By'Range)) &
                                Source.SR.Content (High + 1 .. Source_Length));
      else  -- High < Low
         return Insert (Source, Low, By);
      end if;
   end Replace_Slice;

   -------------------
   -- Replace_Slice --
   -------------------
   procedure Replace_Slice
     (Source : in out Sequence;
      Low    : in Index_Range;
      High   : in Index_Range;
      By     : in Element_Array)
   is
   begin
      Source := Replace_Slice (Source, Low, High, By);
   end Replace_Slice;

   ---------
   -- Set --
   ---------
   procedure Set
     (Item   : in out Sequence;
      Source : in Element_Array)
   is
      Length : constant Length_Range := Source'Length;
   begin
      Item := To_Sequence (Source);
   end Set;

   -----------
   -- Slice --
   -----------
   function Slice
     (Source : in Sequence;
      Low    : in Index_Range;
      High   : in Index_Range)
      return Element_Array
   is
      Source_Length : constant Integer := Source.SR.Current_Length;
   begin
      If Low > Source_Length + 1 or High > Source_Length then raise Index_Error; end if;
      Perform_Calculations_And_Allocate_Array:
      declare
         Slice_Length  : constant Integer := High - Low + 1;
         Slice_First   : constant Integer := Low;
         Slice_Last    : constant Integer := High;
         EA_First      : constant Index_Type := Index_Type'First;
         EA_Last       : constant Index_Type := Index_Type (Integer (EA_First) + Slice_Length - 1);
         EA            : constant Element_Array (EA_First .. EA_Last)
           := Element_Array (Source.SR.Content (Low .. High));
      begin
         return EA;
      end Perform_Calculations_And_Allocate_Array;
   end Slice;

   ----------------------
   -- To_Element_Array --
   ----------------------
   function To_Element_Array
     (Source : in Sequence)
      return Element_Array
   is
      Length : constant Integer := Source.SR.Current_Length;
      First  : constant Index_Type := Index_Type'First;
      Last   : constant Index_Type'Base := Index_Type'Base (Integer (First) + Length - 1);
   begin
      if Last = First - 1 or else Last in Index_Type then
         declare
            EA : constant Element_Array (First .. Last)
              := Element_Array (Source.SR.Content (1 .. Length));
         begin
            return EA;
         end;
      else
         raise FACE.Sequences.Length_Error;
      end if;
   end To_Element_Array;

   -----------------
   -- To_Sequence --
   -----------------
   function To_Sequence
     (Source : in Element_Array)
      return Sequence
   is
      New_Sequence : Sequence;
      Length : constant Length_Range := Source'Length;
   begin
      New_Sequence.SR                := new Sequence_Record (Max_Length => Length + Allocation_Delta);
      New_Sequence.SR.Current_Length := Length;
      New_Sequence.SR.Content (1 .. Length) := Sequence_Vector (Source (Source'Range));
      return New_Sequence;
   end To_Sequence;

end Element_Arrays;
