------------------------------------------------------------------------------------
-- DISTRIBUTION STATEMENT A. Approved for public release; distribution is unlimited.
--
-- Product produced under DoD SENSIAC contract HC104705D4000 under
-- the sponsorship of the Defense Technical Information Center,
-- ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA
-- 22060-6218.  SENSIAC is a DoD Information Analysis Center Sponsored
-- by the Defense Technical Information Center.
--
-- HANDLING AND DESTRUCTION NOTICE - Comply with distribution statement
-- and destroy by any method that will prevent disclosure of
-- contents or reconstruction of the document.
------------------------------------------------------------------------------------

with Ada.Finalization;
with FACE.Utilities;

generic
   type Element is private;    -- can be any non-limited type
   Null_Element : in Element;  -- constant value used to represent nul
package FACE.Sequences.Unbounded is

   subtype Index_Range  is Positive;  -- local name for Positive
   subtype Length_Range is Natural;   -- local name for Natural

   type Sequence is private;

   Null_Sequence : constant Sequence;  -- e.g., an empty Sequence
                                       -- initial value of all Sequences

   -- Returns the current value
   function Length (Source : in Sequence) return Length_Range;
   pragma Inline (Length);

   -- NOTE:  The general rule is that functions that return a Sequence
   --        always return a newly allocated Sequence.  Procedures that
   --        modify (mode is "in out") a parameter that is a Sequence
   --        attempt to use the space, if any, that is pre-allocated and
   --        available to be used before allocating new space.  Alloca-
   --        tion can occur if the allocated space for the Sequence is
   --        to small to hold a new Sequence but also if this space is
   --        too large, so that the reduced space requirement of the
   --        used portion is considered too a waste of memory.

   --------------------------------------------------------
   -- Conversion, Concatenation, and Selection Functions --
   --------------------------------------------------------

   -- Allocates a new Sequence and copies the value of Source to the new Sequence.
   -- The Ada assignment (:=) operation copies the access type only.
   -- added by EM (ISIS) as (1) convenience, (2) intuitive
   function Copy
     (Source : in Sequence)
      return Sequence;

   -- Allocates a newSequence of the requested length, initialized to
   -- Null_Element.
   function To_Sequence
     (Length : in Length_Range)
      return Sequence;

   procedure Append
     (Source   : in out Sequence;
      New_Item : in Sequence);

   procedure Append
     (Source   : in out Sequence;
      New_Item : in Element);

   function "&"
     (Left, Right : in Sequence)
      return Sequence;

   function "&"
     (Left  : in Sequence;
      Right : in Element)
      return Sequence;

   function "&"
     (Left  : in Element;
      Right : in Sequence)
      return Sequence;

   function Get_Element
     (Source : in Sequence;
      Index  : in Index_Range)
      return Element;

   procedure Replace_Element
     (Source : in out Sequence;
      Index  : in Index_Range;
      By     : in Element);

   function "="
     (Left, Right : in Sequence) return Boolean;

   function Is_Null (Source : in Sequence) return Boolean;

   -----------------------------------------
   -- Sequence transformation subprograms --
   -----------------------------------------

   function Delete
     (Source  : in Sequence;
      From    : in Index_Range;
      Through : in Index_Range)
      return Sequence;

   procedure Delete
     (Source  : in out Sequence;
      From    : in Index_Range;
      Through : in Index_Range);

   -----------------------------------
   -- Sequence selector subprograms --
   -----------------------------------

   function Head
     (Source : in Sequence;
      Count  : in Length_Range;
      Pad    : in Element := Null_Element)
      return Sequence;

   procedure Head
     (Source : in out Sequence;
      Count  : in Length_Range;
      Pad    : in Element := Null_Element);

   function Tail
     (Source : in Sequence;
      Count  : in Length_Range;
      Pad    : in Element := Null_Element)
      return Sequence;

   procedure Tail
     (Source : in out Sequence;
      Count  : in Length_Range;
      Pad    : in Element := Null_Element);

   --------------------------------------
   -- Sequence constructor subprograms --
   --------------------------------------

   function "*"
     (Left  : in Length_Range;
      Right : in Element)
      return Sequence;

   function "*"
     (Left  : in Length_Range;
      Right : in Sequence)
      return Sequence;

   generic
      type Index_Type is range <>;  -- Can have any range.
                                    -- Indices may be negative.
   package Element_Arrays is

      type Element_Array        is array (Index_Type range <>) of Element;
      type Element_Array_Access is access all Element_Array;
      -- Access type for convenience -- not used here.

      Null_Element_Array : Element_Array (Index_Type'Last .. Index_Type'First - 1);

      procedure Free (X : in out Element_Array_Access);
      pragma Inline (Free);

      --------------------------------------------------------
      -- Conversion, Concatenation, and Selection Functions --
      --------------------------------------------------------

      function To_Sequence
        (Source : in Element_Array)
      return Sequence;

      procedure Set
        (Item   : in out Sequence;
         Source : in Element_Array);

      function To_Element_Array
        (Source : in Sequence)
      return Element_Array;

      procedure Append
        (Source   : in out Sequence;
         New_Item : in Element_Array);

      function Slice
        (Source : in Sequence;
         Low    : in Index_Range;
         High   : in Index_Range)
      return Element_Array;

      function "&"
        (Left  : in Sequence;
         Right : in Element_Array)
      return Sequence;

      function "&"
        (Left  : in Element_Array;
         Right : in Sequence)
      return Sequence;

      function "="
        (Left  : in Element_Array;
         Right : in Sequence)
      return Boolean;

      function "="
        (Left  : in Sequence;
         Right : in Element_Array)
      return Boolean;

      ----------------------
      -- Search functions --
      ----------------------
      function Index
        (Source  : in Sequence;
         Pattern : in Element_Array;
         Going   : in Direction := Forward)
      return Length_Range; -- 0 = not found

      function Count
        (Source  : in Sequence;
         Pattern : in Element_Array)
      return Length_Range;

      -----------------------------------------
      -- Sequence transformation subprograms --
      -----------------------------------------

      function Replace_Slice
        (Source : in Sequence;
         Low    : in Index_Range;
         High   : in Index_Range;
         By     : in Element_Array)
      return Sequence;

      procedure Replace_Slice
        (Source : in out Sequence;
         Low    : in Index_Range;
         High   : in Index_Range;
         By     : in Element_Array);

      function Insert
        (Source   : in Sequence;
         Before   : in Index_Range;
         New_Item : in Element_Array)
      return Sequence;

      procedure Insert
        (Source   : in out Sequence;
         Before   : in Index_Range;
         New_Item : in Element_Array);

      function Overwrite
        (Source   : in Sequence;
         Position : in Index_Range;
         New_Item : in Element_Array)
      return Sequence;

      procedure Overwrite
        (Source   : in out Sequence;
         Position : in Index_Range;
         New_Item : in Element_Array);

      --------------------------------------
      -- Sequence constructor subprograms --
      --------------------------------------

      function "*"
        (Left  : in Length_Range;
         Right : in Element_Array)
      return Sequence;

   end Element_Arrays;

private
   use Ada.Finalization;

   type Sequence_Vector is array (Index_Range range <>) of Element;
   type Sequence_Vector_Access is access all Sequence_Vector;

   type Sequence_Record (Max_Length : Length_Range) is limited record
      Accesses       : FACE.Utilities.Usage;
      Current_Length : Length_Range := 0;
      Content        : Sequence_Vector (1 .. Max_Length) := (others => Null_Element);
   end record;

   type Sequence_Record_Access is access all Sequence_Record;

   Empty_Sequence_Record : aliased Sequence_Record (Max_Length => 0);

   type Sequence is new Controlled with record
      SR : Sequence_Record_Access := Empty_Sequence_Record'Access;
   end record;

   procedure Initialize (Object : in out Sequence);
   procedure Adjust     (Object : in out Sequence);
   procedure Finalize   (Object : in out Sequence);

   Null_Sequence : constant Sequence := (Controlled with SR => Empty_Sequence_Record'Access);

end FACE.Sequences.Unbounded;
