-- --------------------------------------------------------------------------
-- -
-- File System constant and type definitions and management services -
-- -
-- --------------------------------------------------------------------------
-- from package APEX:
-- type APEX_Integer
-- type APEX_Long_Integer
-- type Return_Code_Type
-- type MESSAGE_ADDR_TYPE
-- subtype Message_Size_Type
-- subtype System_Time_Type


package APEX.File_System is
   -- Domain limits
   System_Limit_File_Size : constant := 2 ** 30; -- file system scope
   -- implementation dependent
   -- Constants and types
   Max_File_Size : constant := System_Limit_File_Size;
   Max_File_Name_Length : constant := 512; -- 511 char + null
   subtype File_Name_Type is String (1 .. Max_File_Name_Length);
   Max_Directory_Entry_Length : constant := 64; -- 63 char + null
   subtype Directory_Entry_Type is String (1 .. Max_Directory_Entry_Length);
   type File_Id_Type is private;
   Null_File_Id : constant File_Id_Type;  -- RAD (09/12/2018): Reflects correction in the next version of the ARINC 653 standards.
   type Directory_Id_Type is private;
   Null_Directory_Id : constant Directory_Id_Type;  -- RAD (09/12/2018): Reflects correction in the next version of the ARINC 653 standards. 
   type FILE_SIZE_TYPE is new APEX_INTEGER;
   -- 32-bit or 64-bit (APEX_LONG_INTEGER)
   -- implementation dependent
   type File_Errno_Type is private;
   Null_File_Errno : constant File_Errno_Type;  -- RAD (09/12/2018): Reflects correction in the next version of the ARINC 653 standards. 
   type File_Mode_Type is (Read, Read_Write);
   type File_Seek_Type is (Seek_Set, Seek_Cur, Seek_End);
   type Entry_Kind_Type is (File_Entry,
                            Directory_Entry,
                            Other_Entry,
                            End_Of_Directory);
   type Time_Set_Type is (Unset, Set);
   type Media_Type is (Volatile, Nonvolatile, Remote);
   type Composite_Time_Type is record
      Tm_Sec : APEX_Integer; -- seconds after the minute [0,59]
      Tm_Min : APEX_Integer; -- minutes after the hour [0,59]
      Tm_Hour : APEX_Integer; -- hours since midnight [0,23]
      Tm_Mday : APEX_Integer; -- day of the month [1,31]
      Tm_Mon : APEX_Integer; -- months since January [0,11]
      Tm_Year : APEX_Integer; -- years since 1900
      Tm_Wday : APEX_Integer; -- days since Sunday [0,6]
      Tm_Yday : APEX_Integer; -- days since January 1 [0,365]
      Tm_Isdst : APEX_Integer; -- Daylight Savings Time flag
      Tm_Is_Set : Time_Set_Type; -- time has been set
   end record;
   type File_Status_Type is record
      Creation_Time : Composite_Time_Type;
      Last_Update : Composite_Time_Type;
      Position : File_Size_Type;
      Size : File_Size_Type;
      Nb_Of_Changes : APEX_Integer;
      Nb_Of_Write_Errors : APEX_Integer;
   end record;
   type Volume_Status_Type is record
      Total_Bytes : APEX_Long_Integer;
      -- nb of bytes allocated to a volume
      Used_Bytes : APEX_Long_Integer;
      -- nb of bytes used
      Free_Bytes : APEX_Long_Integer;
      -- nb of bytes available
      Max_Atomic_Size : APEX_Long_Integer;
      -- nb of bytes that can be atomically transferred
      Block_Size : APEX_Integer;
      -- nb of bytes per block
      Access_Rights : File_Mode_Type;
      -- configuration defined access rights
      Media : Media_Type;
      -- configuration defined media type
   end record;
   -- ERRNO Error codes
   -- The values defined by the ERRNO constants are implementation dependent
   EPERM : constant := 1; -- Operation not permitted
   ENOENT : constant := 2; -- No such file or directory
   EIO : constant := 5; -- Input/output error
   EBADF : constant := 9; -- Bad file descriptor
   EACCES : constant := 13; -- Permission denied
   EBUSY : constant := 16; -- Resource busy
   EEXIST : constant := 17; -- File exists
   ENOTDIR : constant := 20; -- Not a volume or directory
   EISDIR : constant := 21; -- Is a directory
   EINVAL : constant := 22; -- Invalid argument
   EMFILE : constant := 24; -- Too many open files by the current
   -- partition
   EFBIG : constant := 27; -- File too large
   ENOSPC : constant := 28; -- No space left on volume
   EROFS : constant := 30; -- Storage device containing file is
   -- currently write protected
   ENAMETOOLONG : constant := 78; -- Filename too long
   EOVERFLOW : constant := 79; -- Current file position beyond end
   -- of file
   ENOTEMPTY : constant := 93; -- Directory not empty
   ESTALE : constant := 151; -- file or directory ID which an
   -- application has been using is
   -- no longer valid
   -- FILE services
   procedure Open_New_File (
                            File_Name : in File_Name_Type;
                            File_Id : out File_Id_Type;
                            Return_Code : out Return_Code_Type;
                            Errno : in out File_Errno_Type );
   procedure Open_File (
                        File_Name : in File_Name_Type;
                        File_Mode : in File_Mode_Type;
                        File_Id : out File_Id_Type;
                        Return_Code : out Return_Code_Type;
                        Errno : in out File_Errno_Type );
   procedure Close_File (
                         File_Id : in File_Id_Type;
                         Return_Code : out Return_Code_Type;
                         Errno : in out File_Errno_Type );
   procedure Read_File (
                        File_Id : in File_Id_Type;
                        Message_Addr : in Message_Addr_Type;
                        -- the message address is passed IN,
                        -- although
                        -- the respective message is passed OUT
                        In_Length : in Message_Size_Type;
                        -- max number of bytes to read
                        Out_Length : out Message_Size_Type;
                        -- current number of bytes read
                        Return_Code : out Return_Code_Type;
                        Errno : in out File_Errno_Type );
   procedure Write_File (
                         File_Id : in File_Id_Type;
                         Message_Addr : in Message_Addr_Type;
                         Length : in Message_Size_Type;
                         Return_Code : out Return_Code_Type;
                         Errno : in out File_Errno_Type );
   procedure Seek_File (
                        File_Id : in File_Id_Type;
                        Offset : in File_Size_Type;
                        Whence : in File_Seek_Type;
                        Position : out File_Size_Type;
                        Return_Code : out Return_Code_Type;
                        Errno : in out File_Errno_Type );
   procedure Resize_File (
                          File_Id : in File_Id_Type;
                          New_Size : in File_Size_Type;
                          Return_Code : out Return_Code_Type;
                          Errno : in out File_Errno_Type );
   procedure Sync_File (
                        File_Id : in File_Id_Type;
                        Return_Code : out Return_Code_Type;
                        Errno : in out File_Errno_Type );
   procedure Remove_File (
                          File_Name : in File_Name_Type;
                          Return_Code : out Return_Code_Type;
                          Errno : in out File_Errno_Type );
   procedure Rename_File (
                          Old_File_Name : in File_Name_Type;
                          New_File_Name : in File_Name_Type;
                          Return_Code : out Return_Code_Type;
                          Errno : in out File_Errno_Type );
   procedure Get_File_Status (
                              File_Id : in File_Id_Type;
                              File_Status : out File_Status_Type;
                              Return_Code : out Return_Code_Type;
                              Errno : in out File_Errno_Type );
   procedure Get_Volume_Status (
                                File_Name : in File_Name_Type;
                                Volume_Status : out Volume_Status_Type;
                                Return_Code : out Return_Code_Type;
                                Errno : in out File_Errno_Type );
   procedure Open_Directory (
                             Directory_Name : in File_Name_Type;
                             Directory_Id : out Directory_Id_Type;
                             Return_Code : out Return_Code_Type;
                             Errno : in out File_Errno_Type );
   procedure Close_Directory (
                              Directory_Id : in Directory_Id_Type;
                              Return_Code : out Return_Code_Type;
                              Errno : in out File_Errno_Type );
   procedure Read_Directory (
                             Directory_Id : in Directory_Id_Type;
                             Entry_Name : out Directory_Entry_Type;
                             Entry_Kind : out Entry_Kind_Type;
                             Return_Code : out Return_Code_Type;
                             Errno : in out File_Errno_Type );
   procedure Rewind_Directory (
                               Directory_Id : in Directory_Id_Type;
                               Return_Code : out Return_Code_Type;
                               Errno : in out File_Errno_Type );
   procedure Make_Directory (
                             Directory_Name : in File_Name_Type;
                             Return_Code : out Return_Code_Type;
                             Errno : in out File_Errno_Type );
   procedure Remove_Directory (
                               Directory_Name : in File_Name_Type;
                               Return_Code : out Return_Code_Type;
                               Errno : in out File_Errno_Type );
   procedure Sync_Directory (
        Directory_Id : in Directory_Id_Type;
        Return_Code : out Return_Code_Type;
        Errno : in out File_Errno_Type );

    procedure Rename_Directory (
        Old_Directory_Name : in File_Name_Type;
        New_Directory_Name : in File_Name_Type;
        Return_Code : out Return_Code_Type;
        Errno : in out File_Errno_Type );

private
   type File_Id_Type is new APEX_Integer;
   Null_File_Id : constant File_Id_Type := 0;
   type Directory_Id_Type is new APEX_Integer;
   Null_Directory_Id : constant Directory_Id_Type := 0;
   type File_Errno_Type is new APEX_Integer;
   Null_File_Errno : constant File_Errno_Type := 0;

     pragma Convention (C, File_Mode_Type);
     pragma Convention (C, File_Seek_Type);
     pragma Convention (C, Entry_Kind_Type);
     pragma Convention (C, Time_Set_Type);
     pragma Convention (C, Composite_Time_Type);
     pragma Convention (C, File_Status_Type);
     pragma Convention (C, Volume_Status_Type);

-- FOR NOW WE ARE GENERATING ADA BODIES FOR THESE...MAY WANT TO REVISIT AND USE C objects instead
--     pragma Import (C, Open_New_File, "OPEN_NEW_FILE");
--     pragma Import (C, Open_File, "OPEN_FILE");
--     pragma Import (C, Close_File, "CLOSE_FILE");
--     pragma Import (C, Read_File, "READ_FILE");
--     pragma Import (C, Write_File, "WRITE_FILE");
--     pragma Import (C, Seek_File, "SEEK_FILE");
--     pragma Import (C, Resize_File, "RESIZE_FILE");
--     pragma Import (C, Sync_File, "SYNC_FILE");
--     pragma Import (C, Remove_File, "REMOVE_FILE");
--     pragma Import (C, Rename_File, "RENAME_FILE");
--     pragma Import (C, Get_File_Status, "GET_FILE_STATUS");
--     pragma Import (C, Get_Volume_Status, "GET_VOLUME_STATUS");
--     pragma Import (C, Open_Directory, "OPEN_DIRECTORY");
--     pragma Import (C, Close_Directory, "CLOSE_DIRECTORY");
--     pragma Import (C, Read_Directory, "READ_DIRECTORY");
--     pragma Import (C, Rewind_Directory, "REWIND_DIRECTORY");
--     pragma Import (C, Make_Directory, "MAKE_DIRECTORY");
--     pragma Import (C, Remove_Directory, "REMOVE_DIRECTORY");
--     pragma Import (C, Sync_Directory, "SYNC_DIRECTORY");
--     pragma Import (C, Rename_Directory, "RENAME_DIRECTORY");
  end APEX.File_System;
-- GNAT DOESN'T LIKE TWO PACKAGES IN ONE SPEC, MAY WANT TO REVISIT AND MAKE SEPARATE ADS FILE FOR THE FOLLOWING
--  -- Package rename for compatibility with Ada 83 binding --
--  with APEX.File_System;
--  package APEX_FILE_SYSTEM renames APEX.File_System;
