/*------------------------------------------------------------------*/
/*                                                                  */
/* File System constant and type definitions and management services*/
/*                                                                  */
/*------------------------------------------------------------------*/
#ifndef APEX_FILE_SYSTEM
#define APEX_FILE_SYSTEM

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#include "APEX_TYPES.h"

/* from APEX_TYPES: */
/* typedef unsigned char APEX_BYTE; */
/* typedef long APEX_INTEGER; */
/* typedef long long APEX_LONG_INTEGER; */
/* typedef enum {NO_ERROR, ...} RETURN_CODE_TYPE; */
/* typedef APEX_BYTE * MESSAGE_ADDR_TYPE; */
/* typedef APEX_INTEGER MESSAGE_SIZE_TYPE; */
/* typedef APEX_LONG_INTEGER SYSTEM_TIME_TYPE; */

#define SYSTEM_LIMIT_FILE_SIZE 0x40000000 /* file system scope */
/* implementation dependent */
#define MAX_FILE_SIZE SYSTEM_LIMIT_FILE_SIZE
#define MAX_FILE_NAME_LENGTH 512 /* 511 char + null */
typedef char FILE_NAME_TYPE [MAX_FILE_NAME_LENGTH];
#define MAX_DIRECTORY_ENTRY_LENGTH 64 /* 63 char + null */
typedef char DIRECTORY_ENTRY_TYPE [MAX_DIRECTORY_ENTRY_LENGTH];
typedef APEX_INTEGER FILE_ID_TYPE;
typedef APEX_INTEGER DIRECTORY_ID_TYPE;
typedef APEX_INTEGER FILE_SIZE_TYPE;
/* 32-bit or 64-bit (APEX_LONG_INTEGER) */
/* implementation dependent */
typedef APEX_INTEGER FILE_ERRNO_TYPE;

/* FOLLOWING ENUMERATION IS A DUPLICATE from the MEMORY_BLOCK header...using typedef instead
typedef enum {READ = 0, READ_WRITE=1} FILE_MODE_TYPE; */
#include "APEX_MEMORY_BLOCK.h"
typedef MEMORY_BLOCK_MODE_TYPE FILE_MODE_TYPE;
typedef enum { SEEK_SET = 0, SEEK_CUR = 1, SEEK_END = 2 } FILE_SEEK_TYPE;
typedef
enum {
  FILE_ENTRY = 0,
  DIRECTORY_ENTRY = 1,
  OTHER_ENTRY = 2,
  END_OF_DIRECTORY = 3
} ENTRY_KIND_TYPE;

typedef enum { UNSET = 0, SET = 1 } TIME_SET_TYPE;
typedef enum { VOLATILE = 0, NONVOLATILE = 1, REMOTE = 2} MEDIA_TYPE;
typedef
struct {
  APEX_INTEGER TM_SEC; /* seconds after the minute [0,59] */
  APEX_INTEGER TM_MIN; /* minutes after the hour [0,59] */
  APEX_INTEGER TM_HOUR; /* hours since midnight [0,23] */
  APEX_INTEGER TM_MDAY; /* day of the month [1,31] */
  APEX_INTEGER TM_MON; /* months since January [0,11] */
  APEX_INTEGER TM_YEAR; /* years since 1900 */
  APEX_INTEGER TM_WDAY; /* days since Sunday [0,6] */
  APEX_INTEGER TM_YDAY; /* days since January 1 [0,365] */
  APEX_INTEGER TM_ISDST; /* Daylight Savings Time flag */
  TIME_SET_TYPE TM_IS_SET; /* time has been set */
} COMPOSITE_TIME_TYPE;

typedef
struct {
  COMPOSITE_TIME_TYPE CREATE_TIME;
  COMPOSITE_TIME_TYPE LAST_UPDATE;
  FILE_SIZE_TYPE POSITION;
  FILE_SIZE_TYPE SIZE;
  APEX_INTEGER NB_OF_CHANGES;
  APEX_INTEGER NB_OF_WRITE_ERRORS;
} FILE_STATUS_TYPE;

typedef
struct {
  APEX_LONG_INTEGER TOTAL_BYTES;
  APEX_LONG_INTEGER USED_BYTES;
  APEX_LONG_INTEGER FREE_BYTES;
  APEX_LONG_INTEGER MAX_ATOMIC_SIZE;
  APEX_INTEGER BLOCK_SIZE;
  FILE_MODE_TYPE ACCESS_RIGHTS;
  MEDIA_TYPE MEDIA;
} VOLUME_STATUS_TYPE;

/* ERRNO Error codes */
/* The values defined by the ERRNO constants are implementation dependent */
#define EPERM 1 /* Operation not permitted */
#define ENOENT 2 /* No such file or directory */
#define EIO 5 /* Input/output error */
#define EBADF 9 /* Bad file descriptor */
#define EACCES 13 /* Permission denied */
#define EBUSY 16 /* Resource busy */
#define EEXIST 17 /* File exists */
#define ENOTDIR 20 /* Not a volume or directory */
#define EISDIR 21 /* Is a directory */
#define EINVAL 22 /* Invalid argument */
#define EMFILE 24 /* Too many open files by the current */
/* partition */
#define EFBIG 27 /* File too large */
#define ENOSPC 28 /* No space left on volume */
#define EROFS 30 /* Storage device containing file is */
/* currently write protected */
#define ENAMETOOLONG 78 /* Filename too long */
#define EOVERFLOW 79 /* Current file position beyond end */
/* of file */
#define ENOTEMPTY 93 /* Directory not empty */
#define ESTALE 151 /* file or directory ID which an */
/* application has been using is no */
/* longer valid */
extern void OPEN_NEW_FILE (
  /*in */ FILE_NAME_TYPE FILE_NAME,
  /*out */ FILE_ID_TYPE *FILE_ID,
  /*out */ RETURN_CODE_TYPE *RETURN_CODE,
  /*inout*/ FILE_ERRNO_TYPE *ERRNO );
extern void OPEN_FILE (
  /*in */ FILE_NAME_TYPE FILE_NAME,
  /*in */ FILE_MODE_TYPE FILE_MODE,
  /*out */ FILE_ID_TYPE *FILE_ID,
  /*out */ RETURN_CODE_TYPE *RETURN_CODE,
  /*inout*/ FILE_ERRNO_TYPE *ERRNO );
extern void CLOSE_FILE (
  /*in */ FILE_ID_TYPE FILE_ID,
  /*out */ RETURN_CODE_TYPE *RETURN_CODE,
  /*inout*/ FILE_ERRNO_TYPE *ERRNO );
extern void READ_FILE (
  /*in */ FILE_ID_TYPE FILE_ID,
  /*in */ MESSAGE_ADDR_TYPE MESSAGE_ADDR,
  /*in */ MESSAGE_SIZE_TYPE IN_LENGTH,
  /*out */ MESSAGE_SIZE_TYPE *OUT_LENGTH,
  /*out */ RETURN_CODE_TYPE *RETURN_CODE,
  /*inout*/ FILE_ERRNO_TYPE *ERRNO );
extern void WRITE_FILE (
  /*in */ FILE_ID_TYPE FILE_ID,
  /*in */ MESSAGE_ADDR_TYPE MESSAGE_ADDR,
  /*in */ MESSAGE_SIZE_TYPE LENGTH,
  /*out */ RETURN_CODE_TYPE *RETURN_CODE,
  /*inout*/ FILE_ERRNO_TYPE *ERRNO );
extern void SEEK_FILE (
  /*in */ FILE_ID_TYPE FILE_ID,
  /*in */ FILE_SIZE_TYPE OFFSET,
  /*in */ FILE_SEEK_TYPE WHENCE,
  /*out */ FILE_SIZE_TYPE *POSITION,
  /*out */ RETURN_CODE_TYPE *RETURN_CODE,
  /*inout*/ FILE_ERRNO_TYPE *ERRNO );
extern void RESIZE_FILE (
  /*in */ FILE_ID_TYPE FILE_ID,
  /*in */ FILE_SIZE_TYPE NEW_SIZE,
  /*out */ RETURN_CODE_TYPE *RETURN_CODE,
  /*inout*/ FILE_ERRNO_TYPE *ERRNO );
extern void SYNC_FILE (
  /*in */ FILE_ID_TYPE FILE_ID,
  /*out */ RETURN_CODE_TYPE *RETURN_CODE,
  /*inout*/ FILE_ERRNO_TYPE *ERRNO );
extern void REMOVE_FILE (
  /*in */ FILE_NAME_TYPE FILE_NAME,
  /*out */ RETURN_CODE_TYPE *RETURN_CODE,
  /*inout*/ FILE_ERRNO_TYPE *ERRNO );
extern void RENAME_FILE (
  /*in */ FILE_NAME_TYPE OLD_FILE_NAME,
  /*in */ FILE_NAME_TYPE NEW_FILE_NAME,
  /*out */ RETURN_CODE_TYPE *RETURN_CODE,
  /*inout*/ FILE_ERRNO_TYPE *ERRNO );
extern void GET_FILE_STATUS (
  /*in */ FILE_ID_TYPE FILE_ID,
  /*out */ FILE_STATUS_TYPE *FILE_STATUS,
  /*out */ RETURN_CODE_TYPE *RETURN_CODE,
  /*inout*/ FILE_ERRNO_TYPE *ERRNO );
extern void GET_VOLUME_STATUS (
  /*in */ FILE_NAME_TYPE FILE_NAME,
  /*out */ VOLUME_STATUS_TYPE *VOLUME_STATUS,
  /*out */ RETURN_CODE_TYPE *RETURN_CODE,
  /*inout*/ FILE_ERRNO_TYPE *ERRNO );
extern void OPEN_DIRECTORY (
  /*in */ FILE_NAME_TYPE DIRECTORY_NAME,
  /*out */ DIRECTORY_ID_TYPE *DIRECTORY_ID,
  /*out */ RETURN_CODE_TYPE *RETURN_CODE,
  /*inout*/ FILE_ERRNO_TYPE *ERRNO );
extern void CLOSE_DIRECTORY (
  /*in */ DIRECTORY_ID_TYPE DIRECTORY_ID,
  /*out */ RETURN_CODE_TYPE *RETURN_CODE,
  /*inout*/ FILE_ERRNO_TYPE *ERRNO );
extern void READ_DIRECTORY (
  /*in */ DIRECTORY_ID_TYPE DIRECTORY_ID,
  /*out */ DIRECTORY_ENTRY_TYPE *ENTRY_NAME,
  /*out */ ENTRY_KIND_TYPE *ENTRY_KIND,
  /*out */ RETURN_CODE_TYPE *RETURN_CODE,
  /*inout*/ FILE_ERRNO_TYPE *ERRNO );
extern void REWIND_DIRECTORY (
  /*in */ DIRECTORY_ID_TYPE DIRECTORY_ID,
  /*out */ RETURN_CODE_TYPE *RETURN_CODE,
  /*inout*/ FILE_ERRNO_TYPE *ERRNO );
extern void MAKE_DIRECTORY (
  /*in */ FILE_NAME_TYPE DIRECTORY_NAME,
  /*out */ RETURN_CODE_TYPE *RETURN_CODE,
  /*inout*/ FILE_ERRNO_TYPE *ERRNO );
extern void REMOVE_DIRECTORY (
  /*in */ FILE_NAME_TYPE DIRECTORY_NAME,
  /*out */ RETURN_CODE_TYPE *RETURN_CODE,
  /*inout*/ FILE_ERRNO_TYPE *ERRNO );
extern void SYNC_DIRECTORY (
  /*in */ DIRECTORY_ID_TYPE DIRECTORY_ID,
  /*out */ RETURN_CODE_TYPE *RETURN_CODE,
  /*inout*/ FILE_ERRNO_TYPE *ERRNO );
extern void RENAME_DIRECTORY (
  /*in */ FILE_NAME_TYPE OLD_DIRECTORY_NAME,
  /*in */ FILE_NAME_TYPE NEW_DIRECTORY_NAME,
  /*out */ RETURN_CODE_TYPE *RETURN_CODE,
  /*inout*/ FILE_ERRNO_TYPE *ERRNO );

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif
