/*
 * Product produced under Contract No. FA8075-14-D-0014 and the U.S. Government
 * has unlimited rights in this software.
 * 
 * Vanderbilt University disclaims all warranties with regard to this
 * software, including all implied warranties of merchantability and fitness.
 * In no event shall Vanderbilt University be liable for any special,
 * indirect or consequential damages or any damages whatsoever resulting
 * from loss of use, data or profits, whether in an action of contract,
 * negligence or other tortious action, arising out of or in connection
 * with the use or performance of this software.
 * 
 * May contain products produced under DoD SENSIAC contract HC104705D4000 under
 * the sponsorship of the Defense Technical Information Center,
 * ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA
 * 22060-6218.  SENSIAC is a DoD Information Analysis Center Sponsored
 * by the Defense Technical Information Center.
 */

package edu.vu.isis.FACE.posix;

import java.nio.channels.ByteChannel;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.NonReadableChannelException;
import java.nio.channels.NonWritableChannelException;
import java.nio.ByteBuffer;
import java.io.IOException;

/**
 * This class is a wrapper for POSIX message queues for use with the FACE Modeling Tools.
 * It is not intended to be comprehensive, and only covers the functionality required by the FACE Modeling Tools.
 */
public class PosixMQ implements ByteChannel
{
	public PosixMQ() throws IOException
	{
		// Ensure native library is loaded
		if (!loadedNativeLibrary)
		{
			System.out.println("Native library not loaded");
			throw new IOException("Native library not loaded");
		}

		instanceID = instanceCounter++;
		init_native(instanceID);

		descriptor = -1;
		oflag = -1;
	}
	private native void init_native(long id);

	
	private long descriptor;
	private int oflag;
	private long instanceID;
	private static long instanceCounter = 0;
	private String queueName;

	protected void finalize()
	{
		try {
			close();
			finalize_native(this.instanceID);
		} catch (IOException e) {}
	}
	private native void finalize_native(long id);


	@Override
	public boolean isOpen()
	{
		if (descriptor != -1L)
			return true;
		else
			return false;
	}


	/*******************************
	* mq_open
	********************************/
	public void open(String name, int oflag) throws IOException
	{
		// Ensure native library is loaded
		if (!loadedNativeLibrary)
			throw new IOException("Native library not loaded");

		this.oflag = oflag;
		long retVal = mq_open_native(this.instanceID, name, oflag);

		if (retVal == -1)
			throw new IOException("Open failed");
		if (retVal == JNI_JSTRING_CSTRING_ERR)
			throw new IOException("Failed converting jstring to C string");

		descriptor = retVal;
		queueName = name;
	}
	private native long mq_open_native(long id, String name, int oflag);

	public void open(String name, int oflag, int mode, int length, int msgsize) throws IOException
	{
		// Ensure native library is loaded
		if (!loadedNativeLibrary)
			throw new IOException("Native library not loaded");

		this.oflag = oflag;
		long retVal = mq_open_native(this.instanceID, name, oflag, mode, length, msgsize);

		if (retVal == -1)
			throw new IOException("Open failed");
		if (retVal == JNI_JSTRING_CSTRING_ERR)
			throw new IOException("Failed converting jstring to C string");

		descriptor = retVal;
		queueName = name;
	}
	private native long mq_open_native(long id, String name, int oflag, int mode, int length, int msgsize);	


	/********************************************************
	* mq_send
	*   - always sends with priority 0
	*   - will either send all of src's bytes, or none at all
	*********************************************************/
	@Override
	public int write(ByteBuffer src) throws IOException
	{
		// Ensure native library is loaded
		if (!loadedNativeLibrary)
			throw new IOException("Native library not loaded");

		// Ensure message queue is open
		if (!this.isOpen())
			throw new ClosedChannelException();

		// Ensure message queue is opened for sending messages
		if ((this.oflag & O_RDONLY) != 0)
			throw new NonWritableChannelException();

		// Attempt to write entire content of buffer
		src.rewind();
		int numBytesToSend = src.remaining();
		byte[] bytes = new byte[numBytesToSend];
		src.get(bytes);
		long retVal = mq_send_native(this.instanceID, this.descriptor, bytes);

		// Check for failure
		if (retVal != 0)	
		{
			// If we failed to send the bytes, we must put them back in the buffer
			src.rewind();
			src.put(bytes);
		}
		if (retVal == -1)
			throw new IOException("Send failed");
		if (retVal == JNI_NULL_ARG)
			throw new IOException("ByteBuffer has no remaining data to send");
		if (retVal == JNI_JSTRING_CSTRING_ERR)	
			throw new IOException("Failed converting jbyteArray to C array");

		return numBytesToSend;
	}
	private native long mq_send_native(long id, long descriptor, byte[] bytes);


	/****************************************
	* mq_receive
	*   - always receives with priority NULL
	*****************************************/
	@Override
	public int read(ByteBuffer dst)	throws IOException
	{
		// Ensure native library is loaded
		if (!loadedNativeLibrary)
			throw new IOException("Native library not loaded");

		// Ensure message queue is open
		if (!this.isOpen())
			throw new ClosedChannelException();

		// Ensure message queue is opened for reading messages
		if ((this.oflag & O_WRONLY) != 0)
			throw new NonReadableChannelException();

		// Attempt to read into entire remaining space in buffer
		byte[] bytes = new byte[dst.remaining()];
		long retVal = mq_receive_native(this.instanceID, this.descriptor, bytes);
		
		// Check for failure
		if (retVal == -1)
			throw new IOException("Receive failed");
		if (retVal == JNI_NULL_ARG)
			throw new IOException("ByteBuffer has no remaining space");
		if (retVal == JNI_JSTRING_CSTRING_ERR)	
			throw new IOException("Failed converting jbyteArray to C array");

		dst.put(bytes);
		dst.rewind();
		return (int)retVal;
	}
	private native long mq_receive_native(long id, long descriptor, byte[] bytes);


	/*******************************
	* mq_close
	********************************/
	@Override
	public void close() throws IOException
	{
		// Ensure native library is loaded
		if (!loadedNativeLibrary)
			throw new IOException("Native library not loaded");

		// "If this channel is already closed then invoking [close] has no effect."
		if (!this.isOpen())
			return;

		if (mq_close_native(this.instanceID, this.descriptor) != 0)
			throw new IOException("Close failed");

		descriptor = -1;
	}
	private native long mq_close_native(long id, long descriptor);

	/*******************************
	* mq_unlink
	********************************/
	public void unlink() throws IOException
	{

		// Ensure native library is loaded
		if (!loadedNativeLibrary)
			throw new IOException("Native library not loaded");

		// "If this channel is already closed then invoking [close] has no effect."
		if (!this.isOpen())
			return;

		if (mq_unlink_native(this.instanceID, this.queueName) == -1)
			throw new IOException("Unlink failed");

	}
	private native long mq_unlink_native(long id, String name);


	/*******************************
	* Native library handling
	********************************/
	private static boolean loadedNativeLibrary = true;
	static
	{
		try {
			System.loadLibrary("PosixMQ");
		} catch (UnsatisfiedLinkError e) {
			System.err.println("Caught Exception while trying to load library \"PosixMQ\": " + e.getMessage());
			loadedNativeLibrary = false;
		}
	}


	/*******************************
	* Native flags/symbols
	********************************/
	// mq_open oflag
	public static final int O_RDONLY   = getO_RDONLY();
	public static final int O_WRONLY   = get0_WRONLY();
	public static final int O_RDWR     = getO_RDWR();
	public static final int O_CREAT    = getO_CREAT();
	public static final int O_EXCL     = getO_EXCL();
	public static final int O_NONBLOCK = getO_NONBLOCK();
	private native static int getO_RDONLY();
	private native static int get0_WRONLY();
	private native static int getO_RDWR();
	private native static int getO_CREAT();
	private native static int getO_EXCL();
	private native static int getO_NONBLOCK();
	// custom JNI error symbols
	public static final int JNI_JSTRING_CSTRING_ERR = getJNI_JSTRING_CSTRING_ERR();
	public static final int JNI_JARRAY_CARRAY_ERR   = getJNI_JARRAY_CARRAY_ERR();
	public static final int JNI_NULL_ARG            = getJNI_NULL_ARG();
	private native static int getJNI_JSTRING_CSTRING_ERR();
	private native static int getJNI_JARRAY_CARRAY_ERR();
	private native static int getJNI_NULL_ARG();


	/*******************************
	* POSIX Error Handling
	********************************/
	public long getErrno()
	{
		return getErrno_native(this.instanceID);
	}
	private native long getErrno_native(long id);

	public String getStrerror()
	{
		return getStrerror_native(this.instanceID);
	}
	private native String getStrerror_native(long id);

	// Error symbols
	public static final int EACCES = getEACCES(); // mq_open
	public static final int EAGAIN = getEAGAIN(); // mq_send, mq_receive
	public static final int EBADF = getEBADF(); // mq_open, mq_close, mq_send, mq_receive
	public static final int EBADMSG = getEBADMSG(); // mq_receive
	public static final int EEXIST = getEEXIST(); // mq_open
	public static final int EINTR  = getEINTR(); // mq_open, mq_send, mq_receive
	public static final int EINVAL = getEINVAL(); // mq_open, mq_send, mq_receive
	public static final int EMFILE = getEMFILE(); // mq_open
	public static final int EMSGSIZE = getEMSGSIZE(); // mq_send, mq_receive
	public static final int ENAMETOOLONG = getENAMETOOLONG(); // mq_open
	public static final int ENFILE = getENFILE(); // mq_open
	public static final int ENOENT = getENOENT(); // mq_open
	public static final int ENOSPC = getENOSPC(); // mq_open
	//public static final int ETIMEDOUT = getETIMEDOUT(); // mq_send, mq_receive
	private native static int getEACCES();
	private native static int getEAGAIN();
	private native static int getEBADF();
	private native static int getEBADMSG();
	private native static int getEEXIST();
	private native static int getEINTR();
	private native static int getEINVAL();
	private native static int getEMFILE();
	private native static int getEMSGSIZE();
	private native static int getENAMETOOLONG();	
	private native static int getENFILE();
	private native static int getENOENT();
	private native static int getENOSPC();
	//private native static int getETIMEDOUT();


	// Simple test main
	public static void main(String[] args)
	{
		PosixMQ pmq = null;
		PosixMQ pmq2 = null;

		try
		{
			pmq = new PosixMQ();
			pmq2 = new PosixMQ();
		} catch (IOException e) { System.err.println("Exception encountered: " + e.getMessage()); }

		try
		{
			pmq.open("/JNIMQ_1", PosixMQ.O_WRONLY | PosixMQ.O_CREAT, 0666, 1, 4);

			int i = 0x04040404;
			ByteBuffer buf = ByteBuffer.allocate(4);
			buf.putInt(i);
			pmq.write(buf);
			pmq.close();

		} catch (IOException e) { System.err.println("Exception encountered: " + e.getMessage() + " - " + pmq.getStrerror()); }

		try
		{
			pmq2.open("/JNIMQ_1", PosixMQ.O_RDONLY | PosixMQ.O_CREAT);
			ByteBuffer buf = ByteBuffer.allocate(4);
			pmq2.read(buf);
			pmq2.close();

		} catch (IOException e) { System.err.println("Exception encountered: " + e.getMessage() + " - " + pmq2.getStrerror()); }
	}
}	
