/*
 * Product produced under Contract No. FA8075-14-D-0014 and the U.S. Government
 * has unlimited rights in this software.
 * 
 * Vanderbilt University disclaims all warranties with regard to this
 * software, including all implied warranties of merchantability and fitness.
 * In no event shall Vanderbilt University be liable for any special,
 * indirect or consequential damages or any damages whatsoever resulting
 * from loss of use, data or profits, whether in an action of contract,
 * negligence or other tortious action, arising out of or in connection
 * with the use or performance of this software.
 * 
 * May contain products produced under DoD SENSIAC contract HC104705D4000 under
 * the sponsorship of the Defense Technical Information Center,
 * ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA
 * 22060-6218.  SENSIAC is a DoD Information Analysis Center Sponsored
 * by the Defense Technical Information Center.
 */

package edu.vu.isis.FACE.posix;

import java.nio.channels.ByteChannel;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.NonReadableChannelException;
import java.nio.channels.NonWritableChannelException;
import java.nio.ByteBuffer;
import java.io.IOException;

/**
 * This class is a wrapper for POSIX shared memory for use with the FACE Modeling Tools.
 * It is not intended to be comprehensive, and only covers the functionality required by the FACE Modeling Tools.
 */
public class PosixSHM implements ByteChannel
{
	public PosixSHM() throws IOException
	{
		// Ensure native library is loaded
		if (!loadedNativeLibrary)
			throw new IOException("Native library not loaded");

		instanceID = instanceCounter++;
		init_native(instanceID);

		descriptor = -1;
		mappedAddress = MAP_FAILED;
		oflag = -1;
		protections = -1;
		length = -1;
	}
	private native void init_native(long id);


	private long descriptor;
	private long mappedAddress;
	private long oflag;
	private long protections;
	private long length;
	private long instanceID;
	private static long instanceCounter = 0;


	protected void finalize()
	{
		try {
			close();
			finalize_native(this.instanceID);
		} catch (IOException e) {}
	}
	private native void finalize_native(long id);


	@Override
	public boolean isOpen()
	{
		if (descriptor != -1L && mappedAddress != MAP_FAILED)
			return true;
		else
			return false;
	}


	/****************************************************************************
	* open (shm_open, ftruncate, mmap)
	*  - always mmaps to full size of shared memory object
	*      (meaning offset is 0 and mmap length is same as ftruncate length)
	*  - always mmaps with addr=0 (implementation can choose where to put memory)
	*****************************************************************************/
	public void open(String name, long oflag, long mode, long length, long protections, long flags) throws IOException
	{
		// Ensure native library is loaded
		if (!loadedNativeLibrary)
			throw new IOException("Native library not loaded");

		this.oflag = oflag;
		this.protections = protections;
		this.length = length;

		// open/create a shared memory object
		long retVal = shm_open_native(this.instanceID, name, oflag, mode);
		if (retVal == -1)
			throw new IOException("Open failed");
		if (retVal == JNI_JSTRING_CSTRING_ERR)
			throw new IOException("Failed converting jstring to C string");
		descriptor = retVal;

		// set size of shared memory object
		if (ftruncate_native(this.instanceID, this.descriptor, length) != 0)
			throw new IOException("File truncation failed");

		// map shared memory object
		retVal = mmap_native(this.instanceID, length, protections, flags, this.descriptor);
		if (retVal == MAP_FAILED)
			throw new IOException("Memory map failed");
		mappedAddress = retVal;
	}
	private native long shm_open_native(long id, String name, long oflag, long mode);
	private native long ftruncate_native(long id, long fildes, long length);
	private native long mmap_native(long id, long len, long prot, long flags, long fildes);


	/*******************************
	* write (memcpy)
	********************************/
	@Override
	public int write(ByteBuffer src) throws IOException
	{
		// Ensure native library is loaded
		if (!loadedNativeLibrary)
			throw new IOException("Native library not loaded");

		// Ensure shared memory is open
		if (!this.isOpen())
			throw new ClosedChannelException();

		// Ensure shared memory is open for writing
		// All channels are open for writing unless PROT_NONE is set
		if ((this.oflag & O_RDONLY) != 0 || (this.protections == PROT_NONE))
			throw new NonWritableChannelException();

		// Attempt to write entire content of buffer
		src.rewind();
		int numBytesToSend = src.remaining();
		if (numBytesToSend > this.length)
			throw new IOException("Buffer contents larger than shared memory allocation.");
		byte[] bytes = new byte[numBytesToSend];
		src.get(bytes);

		memcpy_native(true, this.mappedAddress, bytes);

		return numBytesToSend;
	}


	/*******************************
	* read (memcpy)
	********************************/
	@Override
	public int read(ByteBuffer dst)	throws IOException
	{
		// Ensure native library is loaded
		if (!loadedNativeLibrary)
			throw new IOException("Native library not loaded");

		// Ensure shared memory is open
		if (!this.isOpen())
			throw new ClosedChannelException();

		// Ensure shared memory is open for reading
		// All channels are readable except those with PROT_NONE
		if (this.protections == PROT_NONE)
			throw new NonReadableChannelException();

		// Attempt to read into entire remaining space in buffer
		int numBytesToReceive = dst.remaining();
		if (numBytesToReceive > this.length)
			throw new IOException("Buffer size larger than shared memory allocation.");
		byte[] bytes = new byte[dst.remaining()];

		memcpy_native(false, this.mappedAddress, bytes);

		dst.put(bytes);
		dst.rewind();
		return (int)this.length;
	}

	// (used for read/write above)
	private native long memcpy_native(boolean writeToAddress, long address, byte[] bytes);


	/*******************************
	* close (munmap, close)
	********************************/
	@Override
	public void close() throws IOException
	{
		// Ensure native library is loaded
		if (!loadedNativeLibrary)
			throw new IOException("Native library not loaded");

		// "If this channel is already closed then invoking [close] has no effect."
		if (!this.isOpen())
			return;

		// unmap shared memory object
		if (munmap_native(this.instanceID, this.mappedAddress, this.length) != 0)
			throw new IOException("Memory unmap failed");		
		mappedAddress = MAP_FAILED;

		// close shared memory object's file descriptor
		if (close_native(this.instanceID, this.descriptor) != 0)
			throw new IOException("Close failed");
		descriptor = -1;
	}
	private native long munmap_native(long id, long addr, long len);
	private native long close_native(long id, long fildes);


	/*******************************
	* Native library handling
	********************************/
	private static boolean loadedNativeLibrary = true;
	static
	{
		try {
			System.loadLibrary("PosixSHM");
		} catch (UnsatisfiedLinkError e) {
			System.err.println("Caught Exception while trying to load library \"PosixSHM\": " + e.getMessage());
			loadedNativeLibrary = false;
		}
	}


	/*******************************
	* Native symbols
	********************************/
	// shm_open oflag
	public static final int O_RDONLY = getO_RDONLY();
	public static final int O_RDWR   = getO_RDWR();
	public static final int O_CREAT  = getO_CREAT();
	public static final int O_EXCL   = getO_EXCL();
	public static final int O_TRUNC  = getO_TRUNC();
	private native static int getO_RDONLY();
	private native static int getO_RDWR();
	private native static int getO_CREAT();
	private native static int getO_EXCL();
	private native static int getO_TRUNC();
	// mmap return
	public static final long MAP_FAILED = getMAP_FAILED();
	private native static long getMAP_FAILED();
	// mmap prot
	public static final int PROT_READ  = getPROT_READ();
	public static final int PROT_WRITE = getPROT_WRITE();
	public static final int PROT_EXEC  = getPROT_EXEC();
	public static final int PROT_NONE  = getPROT_NONE();
	private native static int getPROT_READ();
	private native static int getPROT_WRITE();
	private native static int getPROT_EXEC();
	private native static int getPROT_NONE();
	// mmap flags
	public static final int MAP_SHARED  = getMAP_SHARED();
	public static final int MAP_PRIVATE = getMAP_PRIVATE();
	public static final int MAP_FIXED   = getMAP_FIXED();
	private native static int getMAP_SHARED();
	private native static int getMAP_PRIVATE();
	private native static int getMAP_FIXED();
	// custom JNI error symbols
	public static final int JNI_JSTRING_CSTRING_ERR = getJNI_JSTRING_CSTRING_ERR();
	public static final int JNI_JARRAY_CARRAY_ERR   = getJNI_JARRAY_CARRAY_ERR();
	public static final int JNI_NULL_ARG            = getJNI_NULL_ARG();
	private native static int getJNI_JSTRING_CSTRING_ERR();
	private native static int getJNI_JARRAY_CARRAY_ERR();
	private native static int getJNI_NULL_ARG();	


	/*******************************
	* POSIX Error Handling
	********************************/
	public long getErrno()
	{
		return getErrno_native(this.instanceID);
	}
	private native long getErrno_native(long id);

	public String getStrerror()
	{
		return getStrerror_native(this.instanceID);
	}
	private native String getStrerror_native(long id);

	// Error symbols
	public static final int EACCES = getEACCES(); // shm_open, mmap
	public static final int EAGAIN = getEAGAIN(); // mmap
	public static final int EBADF = getEBADF(); // close, mmap
	public static final int EBADMSG = getEBADMSG();
	public static final int EEXIST = getEEXIST(); // shm_open
	public static final int EFBIG = getEFBIG(); // ftruncate
	public static final int EINTR  = getEINTR(); // shm_open, ftruncate, close
	public static final int EINVAL = getEINVAL(); // shm_open, ftruncate, mmap, munmap
	public static final int EIO = getEIO(); // ftruncate, close
	public static final int EMFILE = getEMFILE(); // shm_open, mmap
	public static final int ENAMETOOLONG = getENAMETOOLONG(); // shm_open
	public static final int ENFILE = getENFILE(); // shm_open
	public static final int ENODEV = getENODEV(); // mmap
	public static final int ENOENT = getENOENT(); // shm_open
	public static final int ENOMEM = getENOMEM(); // mmap
	public static final int ENOSPC = getENOSPC(); // shm_open
	public static final int ENOTSUP = getENOTSUP(); // mmap
	public static final int ENXIO = getENXIO(); // mmap
	public static final int EOVERFLOW = getEOVERFLOW(); // mmap
	private native static int getEACCES();
	private native static int getEAGAIN();
	private native static int getEBADF();
	private native static int getEBADMSG();
	private native static int getEEXIST();
	private native static int getEFBIG();
	private native static int getEINTR();
	private native static int getEINVAL();
	private native static int getEIO();
	private native static int getEMFILE();
	private native static int getENAMETOOLONG();	
	private native static int getENFILE();
	private native static int getENODEV();
	private native static int getENOENT();
	private native static int getENOMEM();
	private native static int getENOSPC();
	private native static int getENOTSUP();
	private native static int getENXIO();
	private native static int getEOVERFLOW();


	// Simple test main
	public static void main(String[] args)
	{
		PosixSHM shm = null;
		PosixSHM shm2 = null;

		try
		{
			shm = new PosixSHM();
			shm2 = new PosixSHM();
		} catch (IOException e) { System.err.println("Exception encountered: " + e.getMessage()); }

		try
		{
			shm.open("/JNISHM_1", PosixSHM.O_RDWR | PosixSHM.O_CREAT, 0777, 64, PosixSHM.PROT_WRITE, PosixSHM.MAP_SHARED);

			int i = 0x04040404;
			ByteBuffer buf = ByteBuffer.allocate(4);
			buf.putInt(i);
			System.out.println("Writing " + i + " to shared memory " + "\"/JNISHM_1\"...");
			shm.write(buf);
			shm.close();

		} catch (IOException e) { System.err.println("Exception encountered: " + e.getMessage() + " - " + shm.getStrerror() + " (" + shm.getErrno() + ")"); }

		try
		{
			shm2.open("/JNISHM_1", PosixSHM.O_RDWR | PosixSHM.O_CREAT, 0777, 64, PosixSHM.PROT_READ, PosixSHM.MAP_SHARED);

			ByteBuffer buf = ByteBuffer.allocate(4);
			shm2.read(buf);
			System.out.println("Received " + buf.getInt() + " from shared memory " + "\"/JNISHM_1\"");
			shm2.close();

		} catch (IOException e) { System.err.println("Exception encountered: " + e.getMessage() + " - " + shm2.getStrerror() + " (" + shm.getErrno() + ")"); }
	}
}
