#!/usr/bin/env bash

function usage {
    echo -e "Usage: runConformanceTest.sh [options] [config_file1] [config_file2] ...\n"

    echo -e "If config file(s) is(are) supplied, conformance test will run directly (no"
    echo -e "GUI). If no config_file is supplied, the test suite GUI will launch for"
    echo -e "interactive control of conformance test.\n"

    echo -e "Options:"
    echo -e "  -h, --help            show this help message and exit"
    echo -e "  -d new_project_directory, --projectDir=new_project_directory"
    echo -e "                        Run test suite using config file, but changing the"
    echo -e "                        project directory to new_project_directory. Changing"
    echo -e "                        project directory only affects segment related files,"
    echo -e "                        not build files. A configuration file must be"
    echo -e "                        specified, and the original project directory must be"
    echo -e "                        specified in the configuration file.  Multiple"
    echo -e "                        configuration files may be specified."
    echo -e "  -m, --model           Run data model test only."
    echo -e "  -o new_log_directory, --logDir=new_log_directory"
    echo -e "                        Run test suite using config file, but changing the log"
    echo -e "                        directory to new_log_directory"
    echo -e "  -g, --gui             Run test suite with GUI even if a configuration file"
    echo -e "                        is supplied"
    echo -e "  -v, --version         Echo test suite version and FACE Standard version"
    echo -e "                        conformance is checked against"
    echo -e "  -l generated_library_directory, --gsl=generated_library_directory"
    echo -e "                        Generate Gold Standard Libraries needed by segments"
    echo -e "                        specified in the configuration file and store in the"
    echo -e "                        generated_library_directory. A configuration file must"
    echo -e "                        be specified, but subsequent configuration files will"
    echo -e "                        be ignored"

} 

function pythonCheck {
	# check if python is installed
	echo 'Checking for Python ...'

	PYTHON=$(which python)

	if [ $? -ne 0 ]; then
		echo '------------------- ERROR -------------------'
		echo 'Conformance Test Suite requires Python 2.6+ installed on the host system'
		echo 'Python has not been installed on this system'
		echo 'Python' $PYTHON
		exit
	fi

	echo 'Python found at' $PYTHON


	# Make sure python is 2.7 or later
	echo 'Checking Python version ...'

	PYTHON_OK=`$PYTHON -c 'import sys; print (sys.version_info >= (2, 6) and "1" or "0")'`

	if [ "$PYTHON_OK" = '0' ]; then
		echo '------------------- ERROR -------------------'
		echo 'Conformance Test Suite requires Python 2.6+ installed on the host system'
		echo "Current Python version:" 
		$PYTHON -V
		exit
	fi

	echo 'Python version is useable'
    # If python is useable, then need the ROOT_DIR set to do the rest of the script
	getRootDirectory

}

function tkinterCheck {
	# Make sure Tkinter is installed
	echo 'Checking for Tkinter ...'

	$PYTHON $ROOT_DIR/face_conformance_app/check_tkinter.py
	if [ $? -ne 0 ]; then
		echo '------------------- ERROR -------------------'
		echo 'Conformance Test Suite requires Tkinter installed on the host system'
		echo 'For RHEL6/CentOS6, install the following packages:'
		echo '  tix'
		echo '  tcl'
		echo '  tk'
		echo '  tkinter'
		echo 'For Ubuntu, install the following packages:'
		echo '  python-tk'
		exit
	fi

	echo 'Tkinter found'
}

function javaCheck {
	# Make sure Java is installed (and able to be run from Python)
	echo 'Checking for Java ...'

	$PYTHON $ROOT_DIR/face_conformance_app/check_java.py
	if [ $? -ne 0 ]; then
		echo '------------------- ERROR -------------------'
		echo 'Conformance Test Suite requires Java 1.7 or greater installed on the host system'
		exit
	fi

	echo 'Java found'
}


function conformanceTestLaunch {

	# Run the tool
	echo 'Starting Conformance Test Suite from the face_conformance_app subdirectory...'
	echo $PYTHON 'face_conformance_app/ConformanceMain.py ' "$@"

	#pushd $ROOT_DIR/face_conformance_app/ > /dev/null
	$PYTHON $ROOT_DIR/face_conformance_app/ConformanceMain.py "$@"
	returnCode=$?	
	#popd > /dev/null
	exit $returnCode
}

function getRootDirectory {

    SOURCE="${BASH_SOURCE[0]}"
    while [ -h "$SOURCE" ]; do # resolve $SOURCE until the file is no longer a symlink
        DIR="$( cd -P "$( dirname "$SOURCE" )" && pwd )"
        SOURCE="$(readlink "$SOURCE")"
        [[ $SOURCE != /* ]] && SOURCE="$DIR/$SOURCE" # if $SOURCE was a relative symlink, we need to resolve it relative to the path where the symlink file was located
    done
    ROOT_DIR="$( cd -P "$( dirname "$SOURCE" )" && pwd )"
}


if [ $# -ge 1 ]; then 
	case "$1" in
		-h|--help)
			usage
			;;

		-g|--gui) # need GUI...check for Tkinter
		    pythonCheck
			tkinterCheck
			javaCheck
			conformanceTestLaunch "$@"
			;;

		-m|--model)  # Run only data model test
		    pythonCheck
		    javaCheck
		    conformanceTestLaunch "$@"
		    ;;

		-v|--version) #only checking version, no GUI required
			pythonCheck
			javaCheck
			conformanceTestLaunch "$@"
			;;
			
		*)	# Must be config files...no GUI required
			pythonCheck
			javaCheck
			conformanceTestLaunch "$@"
			;;
	esac
else # no arguments, need GUI...check for Tkinter
	pythonCheck
	tkinterCheck
	javaCheck
	conformanceTestLaunch "$@"
fi
	


