with Ada.Unchecked_Conversion;
--------------------
-- Element_Arrays --
--------------------
separate (FACE.Sequences.Bounded)
package body Element_Arrays is

   ---------
   -- "&" --
   ---------

   function "&"
     (Left  : in Sequence;
      Right : in Element_Array)
      return Sequence
   is
   begin
      return Append (Left, Right, Error);
   end "&";

   ---------
   -- "&" --
   ---------

   function "&"
     (Left  : in Element_Array;
      Right : in Sequence)
      return Sequence
   is
   begin
      return Append (Left, Right, Error);
   end "&";

   ---------
   -- "*" --
   ---------
   --   This function returns a sequence composed of whole copies only
   --   of Right.  No partial copies of Right are returned as part of
   --   the returned value.
   function "*"
     (Left  : in Length_Range;
      Right : in Element_Array)
      return Sequence
   is
      Seq_Length : constant Length_Range := Length_Range'Min (Left * Right'Length,
                                                              Max_Length - (Max_Length rem Right'Length));
      -- Need explicit range for subscript of Right in order to convert
      -- it to a slice for legal comparison of arrays with different
      -- subscript types.

      Seq_Low    : Natural := 1;
      Seq_High   : Natural;
      New_Seq    : Sequence;
   begin
      New_Seq.Current_Length := Seq_Length;

      Perform_The_Replication:
      for I in 1 .. Left loop
         Seq_High := Seq_Low + Right'Length - 1;
         exit Perform_The_Replication when Seq_High > Max_Length;

         -- Store a copy of Right in the new sequence.
         New_Seq.Content (Seq_Low .. Seq_High) := Sequence_Vector_Type (Right (Right'Range));
         -- Need explicit range for subscript of Right in order to
         -- convert it to a slice for legal comparison of arrays
         -- with different subscript types.

         Seq_Low := Seq_Low + Right'Length;
      end loop Perform_The_Replication;

      return New_Seq;
   end "*";

   ---------
   -- "=" --
   ---------

   function "="
     (Left  : in Sequence;
      Right : in Element_Array)
      return Boolean
   is
      Left_Length  : constant Length_Range := Left .Current_Length;
      Right_Length : constant Integer      := Integer (Right'Length);
      Temp_Right   : constant Sequence_Vector_Type (1 .. Right_Length)
        := Sequence_Vector_Type (Right (Right'Range));
      -- Need explicit range for subscript of Right in order to convert
      -- it to a slice for legal comparison of arrays with different
      -- subscript types.
   begin
      return Boolean
        ((Left_Length = 0 and Right_Length = 0) or else
           (Left_Length = Right_Length and then
            Left.Content (1 .. Left_Length) = Temp_Right));
   end "=";

   ---------
   -- "=" --
   ---------

   function "="
     (Left  : in Element_Array;
      Right : in Sequence)
      return Boolean
   is
   begin
      return Right = Left;
   end "=";

   ------------
   -- Append --
   ------------

   function Append
     (Left  : in Sequence;
      Right : in Element_Array;
      Drop  : in Truncation := Error)
      return Sequence
   is
   begin
      return Append (Left  => Left,
                     Right => To_Sequence (Right, Drop),
                     Drop  => Drop);
   end Append;

   ------------
   -- Append --
   ------------

   function Append
     (Left  : in Element_Array;
      Right : in Sequence;
      Drop  : in Truncation := Error)
      return Sequence
   is
   begin
      return Append (Left  => To_Sequence (Left, Drop),
                     Right => Right,
                     Drop  => Drop);
   end Append;

   ------------
   -- Append --
   ------------

   procedure Append
     (Source   : in out Sequence;
      New_Item : in Element_Array;
      Drop     : in Truncation := Error)
   is
   begin
      Source := Append (Source, New_Item, Drop);
   end Append;

   -----------
   -- Count --
   -----------

   function Count
     (Source  : in Sequence;
      Pattern : in Element_Array)
      return Length_Range
   is
      Local_Count : Length_Range := 0;
   begin
      if Pattern'Length = 0 then
         raise Pattern_Error;
      elsif Pattern'Length > Source.Current_Length then
         raise Length_Error;
      else
         for Index in 1 .. Source.Current_Length + 1 - Pattern'Length loop
            if Element_Array (Source.Content (Index .. Index + Pattern'Length - 1)) = Pattern (Pattern'Range) then
               -- Need explicit range for subscript of Pattern in
               -- order to convert it to a slice for legal comparison
               -- of arrays with different subscript types.
               Local_Count := Local_Count + 1;
            end if;
         end loop;
         return Local_Count;  -- The pattern was not found.
      end if;
   end Count;

   -----------
   -- Index --
   -----------

   function Index
     (Source  : in Sequence;
      Pattern : in Element_Array;
      Going   : in Direction := Forward)
      return Length_Range
   is
   begin
      if Pattern'Length = 0 then
         raise Pattern_Error;
      elsif Pattern'Length > Source.Current_Length then
         raise Length_Error;
      else
         if Going = Forward then
            for Index in 1 .. Source.Current_Length + 1 - Pattern'Length loop
               if Element_Array (Source.Content (Index .. Index + Pattern'Length - 1)) = Pattern (Pattern'Range) then
                  -- Need explicit range for subscript of Pattern in
                  -- order to convert it to a slice for legal comparison
                  -- of arrays with different subscript types.
                  return Index;
               end if;
            end loop;
         else  -- Going = Backward
            for Index in reverse 1 .. Source.Current_Length + 1 - Pattern'Length loop
               if Element_Array (Source.Content (Index .. Index + Pattern'Length - 1)) = Pattern (Pattern'Range) then
                  -- Need explicit range for subscript of Pattern in
                  -- order to convert it to a slice for legal comparison
                  -- of arrays with different subscript types.
                  return Index;
               end if;
            end loop;
         end if;
         return 0;  -- The pattern was not found.
      end if;
   end Index;

   ------------
   -- Insert --
   ------------

   function Insert
     (Source   : in Sequence;
      Before   : in Length_Range;
      New_Item : in Element_Array;
      Drop     : in Truncation := Error)
      return Sequence
   is
   begin
      return Replace_Slice (Source, Before, Index_Range'First - 1, New_Item, Drop);
   end Insert;

   ------------
   -- Insert --
   ------------

   procedure Insert
     (Source   : in out Sequence;
      Before   : in Length_Range;
      New_Item : in Element_Array;
      Drop     : in Truncation := Error)
   is
   begin
      Source := Insert (Source, Before, New_Item, Drop);
   end Insert;

   ---------------
   -- Overwrite --
   ---------------

   function Overwrite
     (Source   : in Sequence;
      Position : in Length_Range;
      New_Item : in Element_Array;
      Drop     : in Truncation := Error)
      return Sequence
   is
   begin
      return Replace_Slice (Source, Position, Position + New_Item'Length - 1, New_Item, Drop);
   end Overwrite;

   ---------------
   -- Overwrite --
   ---------------

   procedure Overwrite
     (Source   : in out Sequence;
      Position : in Length_Range;
      New_Item : in Element_Array;
      Drop     : in Truncation := Error)
   is
   begin
      Source := Overwrite (Source, Position, New_Item, Drop);
   end Overwrite;

   -------------------
   -- Replace_Slice --
   -------------------

   function Replace_Slice
     (Source : in Sequence;
      Low    : in Length_Range;
      High   : in Length_Range;
      By     : in Element_Array;
      Drop   : in Truncation := Error)
      return Sequence
   is
   begin
      if Low > Source.Current_Length + 1 or High < 0 then
         raise Index_Error;
      elsif High >= Low then
         Temporary_High_GE_Low:
         declare
            New_Length : Natural := Source.Current_Length - ((High + 1) - Low) + By'Length;
            type New_Vector_Type is array (Positive range <>) of Element;
            subtype New_Vector is New_Vector_Type (1 .. New_Length);
            New_Vec    : New_Vector;
         begin
            New_Vec :=
              New_Vector_Type (Source.Content (1 .. Low - 1)) &
              New_Vector_Type (By (By'Range))                 &
              New_Vector_Type (Source.Content (High + 1 .. Source.Current_Length));
            if New_Length > Max_Length then
               case Drop is
                  when Error =>
                     raise Length_Error;
                  when Seq.Left =>    -- Truncate on left side.
                     return Sequence'
                       (Max_Length,
                        Sequence_Vector (New_Vec ((New_Length - Max_Length) + 1 .. Max_Length + (New_Length - Max_Length))));
                  when Seq.Right =>   -- Truncate on right side.
                     return Sequence'
                       (Max_Length,
                        Sequence_Vector (New_Vec (1 .. Max_Length)));
               end case;
            else  -- New_Length <= Max_Length
               -- No need to drop any elements
               New_Length := Natural'Min (New_Length, Max_Length);
               return Sequence'(New_Length,
                                Sequence_Vector_Type (New_Vec (1 .. New_Length)) &
                                Sequence_Vector_Type'(Index_Range (Max_Length - New_Length) => Null_Element));
            end if;
         end Temporary_High_GE_Low;
      elsif High < Low then
         Temporary_High_LT_Low:
         declare
            New_Length : Natural := Source.Current_Length + By'Length;
            type New_Vector_Type is array (Positive range <>) of Element;
            subtype New_Vector is New_Vector_Type (1 .. New_Length);
            New_Vec    : New_Vector;
         begin
            New_Vec :=
              New_Vector_Type (Source.Content (1 .. Low - 1)) &
              New_Vector_Type (By (By'Range))                 &
              New_Vector_Type (Source.Content (Low .. Source.Current_Length));
            if New_Length > Max_Length then
               case Drop is
               when Error =>
                  raise Length_Error;
                  when Seq.Left =>    -- Truncate on left side.
                     return Sequence'
                       (Max_Length,
                        Sequence_Vector (New_Vec ((New_Length - Max_Length) + 1 .. Max_Length + (New_Length - Max_Length))));
                  when Seq.Right =>   -- Truncate on right side.
                     return Sequence'
                       (Max_Length,
                        Sequence_Vector (New_Vec (1 .. Max_Length)));
               end case;
            else  -- New_Length <= Max_Length
               -- No need to drop any elements
               New_Length := Natural'Min (New_Length, Max_Length);
               return Sequence'(New_Length,
                                Sequence_Vector (New_Vec (1 .. New_Length)));
            end if;
         end Temporary_High_LT_Low;
      else  -- others cases
         raise Index_Error;
      end if;
   end Replace_Slice;

   -------------------
   -- Replace_Slice --
   -------------------

   procedure Replace_Slice
     (Source : in out Sequence;
      Low    : in Length_Range;
      High   : in Length_Range;
      By     : in Element_Array;
      Drop  : in Truncation := Error)
   is
   begin
      Source := Replace_Slice (Source, Low, High, By, Drop);
   end Replace_Slice;

   ---------------
   -- Replicate --
   ---------------

   function Replicate
     (Count : in Length_Range;
      Item  : in Element_Array;
      Drop  : in Truncation := Error)
      return Sequence
   is
   begin
      if (Count * Item'Length) > Max_Length then
         case Drop is
            when Error =>
               raise Length_Error;
            when Seq.Left | Seq.Right =>
               -- Truncate on left or right side.
               -- The result is the same, since "*" only returns whole
               -- multiples of Item.
               return Count * Item;
         end case;
      else
         return Count * Item;
      end if;
   end Replicate;

   ---------
   -- Set --
   ---------

   procedure Set
     (Item   : in out Sequence;
      Source : in Element_Array;
      Drop   : in Truncation := Error)
   is
   begin
      Item := To_Sequence (Source, Drop);
   end Set;

   -----------
   -- Slice --
   -----------

   function Slice
     (Source : in Sequence;
      Low    : in Index_Range;
      High   : in Index_Range)
      return Element_Array
   is
   begin
      if Low > Source.Current_Length or
        High > Source.Current_Length or
        Low  <= 0 or High <= 0 or
        High < Low then
         raise Index_Error;
      else
         return Element_Array (Source.Content (Low .. High));
      end if;
   end Slice;

   ----------------------
   -- To_Element_Array --
   ----------------------

   function To_Element_Array
     (Source : in Sequence)
      return Element_Array
   is
   begin
      if Source.Current_Length = 0 then
         return Null_Element_Array;
      else
         Safe_To_Perform_Calculations:
         declare
            Range_Low       : constant Index_Type := Index_Type'First;
            Range_High      : constant Index_Type := Index_Type (Length_Range (Index_Type'First) + Source.Current_Length - 1);
            Array_To_Return : Element_Array (Index_Type'Range) := (others => Null_Element);
         begin
            Array_To_Return (Range_Low .. Range_High) := Element_Array (Source.Content (1 .. Source.Current_Length));
            -- Need explicit range for subscript of Source.Content in
            -- order to convert it to a slice for legal comparison of
            -- arrays with different subscript types.
            return Array_To_Return;
         end Safe_To_Perform_Calculations;
      end if;
   end To_Element_Array;

   -----------------
   -- To_Sequence --
   -----------------
   function To_Sequence
     (Source : in Element_Array;
      Drop   : in Truncation := Error)
      return Sequence
   is
      Max_Length_Index : constant Index_Type := Index_Type (Max_Length);
   begin
      if Source'Length > Max_Length then
         case Drop is
            when Error =>
               raise Length_Error;
            when Seq.Left =>   -- Truncate on left side.
               return Sequence'(Current_Length => Max_Length,
                                Content        => Sequence_Vector (Source (Index_Type (Source'Last + 1) - Max_Length_Index .. Index_Type (Source'Last))));
            when Seq.Right =>  -- Truncate on right side.
               return Sequence'(Current_Length => Max_Length,
                                Content        => Sequence_Vector (Source (Index_Type'Base (Source'First) .. Index_Type'Base (Source'First + Max_Length_Index - 1))));
         end case;
      elsif Source'Length < Max_Length then
         Set_Up_Temporary_For_Copying:
         declare
            Seq_To_Return : Sequence :=
                              (Current_Length => Source'Length,
                               Content        => (others => Null_Element));
          begin
            Copy_Array_To_Sequence:
            for Array_Index in Source'First .. Source'Last loop
               Seq_To_Return.Content (Index_Range (Array_Index - Source'First + 1)) := Source (Array_Index);
            end loop Copy_Array_To_Sequence;

            return Seq_To_Return;
         end Set_Up_Temporary_For_Copying;
      else  -- New_Length = Max_Length
         return Sequence'(Current_Length => Max_Length,
                          Content        => Sequence_Vector (Source (Source'Range)));
         -- Need explicit range for subscript of Source in order to
         -- convert it to a slice for legal comparison of arrays with
         -- different subscript types.
      end if;
   end To_Sequence;

   function To_Sequence (Length : in Index_Range)
                         return Sequence
   is
   begin
      return Sequence'(Current_Length => Length,
                       Content        => (others => Null_Element));
   end To_Sequence;

end Element_Arrays;
