with Ada.Strings.Fixed;

package body FACE.Sequences.Bounded is

   package Str renames Ada.Strings.Fixed;
   package Seq renames FACE.Sequences;

   ---------
   -- "&" --
   ---------

   function "&"
     (Left, Right : in Sequence)
      return Sequence
   is
   begin
      return Append (Left, Right);
   end "&";

   ---------
   -- "&" --
   ---------

   function "&"
     (Left  : in Sequence;
      Right : in Element)
      return Sequence
   is
   begin
      return Append (Left, Right);
   end "&";

   ---------
   -- "&" --
   ---------

   function "&"
     (Left  : in Element;
      Right : in Sequence)
      return Sequence
   is
   begin
      return Append (Left, Right);
   end "&";

   ---------
   -- "*" --
   ---------

   function "*"
     (Left  : in Length_Range;
      Right : in Element)
      return Sequence
   is
   begin
      return Sequence'(Left, (others => Right));
   end "*";

   ---------
   -- "*" --
   ---------

   function "*"
     (Left  : in Length_Range;
      Right : in Sequence)
      return Sequence
   is
      Len_To_Return : Natural := Left * Right.Current_Length;
      -- Natural so that upper bound of Length_Range will not be violated.
   begin
      if Len_To_Return = 0 then
         return Null_Sequence;
      elsif Len_To_Return > Max_Length then
         raise Length_Error;
      else  -- 0 < Len_To_Return <= Max_Length
         Allocate_Temporary:
         declare
            Seq_To_Return : Sequence;
            Left_Index    : Positive := 1;
            Right_Index   : Positive := Right.Current_Length;
         begin
            Seq_To_Return.Current_Length := Len_To_Return;
            Copy_Sequence_Multiple_Times:
            for Count in 1 .. Left loop
               Seq_To_Return.Content (Left_Index .. Right_Index) := Right.Content (1 .. Right.Current_Length);
               Left_Index  := Left_Index  + Right.Current_Length;
               Right_Index := Right_Index + Right.Current_Length;
            end loop Copy_Sequence_Multiple_Times;
            return Seq_To_Return;
         end Allocate_Temporary;
      end if;
   end "*";

   ---------
   -- "=" --
   ---------

   function "=" (Left, Right : in Sequence) return Boolean
   is
      Left_Length  : constant Length_Range := Left .Current_Length;
      Right_Length : constant Length_Range := Right.Current_Length;
   begin
      return Boolean
        ((Left_Length = 0 and Right_Length = 0) or else
           (Left_Length = Right_Length and then
            Left.Content (1 .. Left_Length) = Right.Content (1 .. Right_Length)));
   end "=";

   ------------
   -- Append --
   ------------

   function Append
     (Left, Right : in Sequence;
      Drop        : in Truncation := Error)
      return Sequence
   is
      New_Length   : Natural := Left.Current_Length + Right.Current_Length;
      -- Natural so that upper bound of Length_Range will not be violated.

      Source_Left_LIndex,
      Source_Right_LIndex,

      Target_Left_LIndex,
      Target_Right_LIndex  : Length_Range := 1;

      Source_Left_RIndex   : Length_Range := Left.Current_Length;
      Source_Right_RIndex  : Length_Range := Right.Current_Length;

      Target_Left_RIndex,
      Target_Right_RIndex  : Length_Range := Max_Length;

      Left_Participates,
      Right_Participates : Boolean := False;
   begin
      -- First work out what's what.
      if New_Length > Max_Length then
         case Drop is
            when Error =>
               raise Length_Error;
            when Seq.Left =>   -- Truncate on left side.
               Right_Participates := True;
               if Right.Current_Length < Max_Length then
                  Left_Participates   := True;
                  Target_Left_RIndex  := Max_Length - Right.Current_Length;
                  Target_Right_LIndex := Target_Left_RIndex + 1;
                  Source_Left_LIndex  := Left.Current_Length + 1 - Target_Left_RIndex;
               end if;
            when Seq.Right =>  -- Truncate on right side.
               Left_Participates := True;
               if Left.Current_Length < Max_Length then
                  Right_Participates  := True;
                  Target_Left_RIndex  := Left.Current_Length;
                  Target_Right_LIndex := Target_Left_RIndex + 1;
                  Source_Right_RIndex := Max_Length - Target_Left_RIndex;
               end if;
         end case;

         New_Length := Max_Length;
      else  -- New_Length <= Max_Length
         Left_Participates  := True;
         Right_Participates := True;
         Target_Left_RIndex        := Left.Current_Length;
         Target_Right_LIndex       := Target_Left_RIndex + 1;
      end if;

      -- Return a new Sequence what's what.
      if Left_Participates and Right_Participates then
         if New_Length = Max_Length then
         return Sequence'
           (Current_Length => New_Length,
            Content        =>
              Left.Content  (Source_Left_LIndex  .. Source_Left_RIndex)  &
              Right.Content (Source_Right_LIndex .. Source_Right_RIndex));  -- no padding
         else  -- New_Length < Max_Length
            return Sequence'
              (Current_Length => New_Length,
               Content        =>
                 Left.Content  (Source_Left_LIndex  .. Source_Left_RIndex)  &
                 Right.Content (Source_Right_LIndex .. Source_Right_RIndex) &
                 Null_Sequence.Content (1 .. Max_Length - New_Length));  -- padding
         end if;
      elsif Left_Participates then
         if New_Length = Max_Length then
            return Sequence'
              (Current_Length => New_Length,
               Content        => Left.Content (Source_Left_LIndex  .. Source_Left_RIndex));  -- no padding
         else  -- New_Length < Max_Length
            return Sequence'
              (Current_Length => New_Length,
               Content        =>
                 Left.Content (Source_Left_LIndex  .. Source_Left_RIndex) &
                 Null_Sequence.Content (1 .. Max_Length - New_Length));  -- padding
         end if;
      else  -- Right_Participates
         if New_Length = Max_Length then
            return Sequence'
              (Current_Length => New_Length,
               Content        => Right.Content (Source_Right_LIndex  .. Source_Right_RIndex));  -- no padding
         else  -- New_Length < Max_Length
            return Sequence'
              (Current_Length => New_Length,
               Content        =>
                 Right.Content (Source_Right_LIndex  .. Source_Right_RIndex) &
                 Null_Sequence.Content (1 .. Max_Length - New_Length));  -- padding
         end if;
      end if;
end Append;

   ------------
   -- Append --
   ------------

   function Append
     (Left  : in Sequence;
      Right : in Element;
      Drop  : in Truncation := Error)
      return Sequence
   is
      New_Length : Natural := Length_Range'Min (Left.Current_Length + 1, Max_Length);
      -- Natural so that upper bound of Length_Range will not be violated.

      LIndex,
      RIndex     : Length_Range;
   begin
      if Left.Current_Length = Max_Length then
         case Drop is
            when Error =>
               raise Length_Error;
            when FACE.Sequences.Left  =>
               LIndex := 2;
               RIndex := Max_Length;
            when FACE.Sequences.Right =>
               LIndex := 1;
               RIndex := Max_Length - 1;
         end case;
      else  -- Left.Current_Length < Max_Length
         LIndex := 1;
         RIndex := Left.Current_Length;
      end if;

      if New_Length = Max_Length then
         return Sequence'(Current_Length => New_Length,
                          Content        => Left.Content (LIndex .. RIndex) & Right);  -- no padding
      else  -- New_Length < Max_Length
         Set_Up_Temporary_For_Copying:
         declare
            Seq_To_Return : Sequence :=
                              (Current_Length => New_Length,
                               Content        => (others => Null_Element));
         begin
            Seq_To_Return.Content (1 .. New_Length) := Left.Content (LIndex .. RIndex) & Right;
            return Seq_To_Return;
         end Set_Up_Temporary_For_Copying;
     end if;
   end Append;

   ------------
   -- Append --
   ------------

   function Append
     (Left : in Element;
      Right : in Sequence;
      Drop  : in Truncation := Error)
      return Sequence
   is
      New_Length : Natural := Length_Range'Min (Right.Current_Length + 1, Max_Length);
      -- Natural so that upper bound of Length_Range will not be violated.

      LIndex,
      RIndex     : Length_Range;
   begin
      if Right.Current_Length = Max_Length then
         case Drop is
            when Error =>
               raise Length_Error;
            when FACE.Sequences.Left  =>
               LIndex := 2;
               RIndex := Max_Length;
            when FACE.Sequences.Right =>
               LIndex := 1;
               RIndex := Max_Length - 1;
         end case;
      else  -- Right.Current_Length < Max_Length
         LIndex := 1;
         RIndex := Right.Current_Length;
      end if;

      if New_Length = Max_Length then
         return Sequence'(Current_Length => New_Length,
                          Content        => Left & Right.Content (LIndex .. RIndex));  -- no padding
      else  -- New_Length < Max_Length
         Set_Up_Temporary_For_Copying:
         declare
            Seq_To_Return : Sequence :=
                              (Current_Length => New_Length,
                               Content        => (others => Null_Element));
         begin
            Seq_To_Return.Content (1 .. New_Length) := Left & Right.Content (LIndex .. RIndex);
            return Seq_To_Return;
         end Set_Up_Temporary_For_Copying;
     end if;
   end Append;

   ------------
   -- Append --
   ------------

   procedure Append
     (Source   : in out Sequence;
      New_Item : in Sequence;
      Drop     : in Truncation := Error)
   is
   begin
      Source := Append (Source, New_Item, Drop);
   end Append;

   ------------
   -- Append --
   ------------

   procedure Append
     (Source   : in out Sequence;
      New_Item : in Element;
      Drop     : in Truncation := Error)
   is
   begin
      Source := Append (Source, New_Item, Drop);
   end Append;

   ------------
   -- Delete --
   ------------

   function Delete
     (Source  : in Sequence;
      From    : in Index_Range;
      Through : in Index_Range)
      return Sequence
   is
      Include_First : constant Boolean := Boolean (From    /= Source.Content'First);
      Include_Last  : constant Boolean := Boolean (Through /= Source.Current_Length);
   begin
      if From > Through or
        From    not in Source.Content'First .. Source.Current_Length or
        Through not in Source.Content'First .. Source.Current_Length then
         raise Index_Error;
      else
         Error_Checks_OK:
         declare
            Delete_Length : constant Length_Range := Through + 1 - From;
            New_Length    : constant Length_Range := Source.Current_Length - Delete_Length;
         begin
            if Include_First and Include_Last then
               return Sequence'
                 (New_Length,
                  Source.Content (Source.Content'First .. From - 1) &
                    Source.Content (Through + 1 .. Source.Current_Length) &
                    Null_Sequence.Content (1 .. (Max_Length - New_Length))
                 );
            elsif Include_First then
               return Sequence'
                 (New_Length,
                  Source.Content (Source.Content'First .. From - 1) &
                    Null_Sequence.Content (1 .. (Max_Length - New_Length))
                 );
            elsif Include_Last then
               return Sequence'
                 (New_Length,
                  Source.Content (Through + 1 .. Source.Current_Length) &
                    Null_Sequence.Content (1 .. (Max_Length - New_Length))
                 );
            else  -- result is null
               return Null_Sequence;
            end if;
         end Error_Checks_OK;
      end if;
   end Delete;

   ------------
   -- Delete --
   ------------

   procedure Delete
     (Source : in out Sequence;
      From    : in Index_Range;
      Through : in Index_Range)
   is
   begin
      Source := Delete (Source, From, Through);
   end Delete;

   -----------------
   -- Get_Element --
   -----------------

   function Get_Element
     (Source : in Sequence;
      Index  : in Index_Range)
      return Element
   is
   begin
      return Source.Content (Index);
   end Get_Element;

   ----------
   -- Head --
   ----------

   function Head
     (Source : in Sequence;
      Count  : in Length_Range;
      Pad    : in Element;
      Drop   : in Truncation := Error)
      return Sequence
   is
   begin
      if Source.Current_Length = Count then
         return Source;
      elsif Source.Current_Length > Count then
         case Drop is
            when Error =>
               raise Length_Error;
            when Seq.Left =>   -- Truncate on left side.
               return Delete (Source,
                              From    => 1,
                              Through => Source.Current_Length - Count);
            when Seq.Right =>  -- Truncate on right side.
               return Delete (Source,
                              From    => Count + 1,
                              Through => Source.Current_Length);
         end case;
      else  -- Source.Current_Length < Count

            -- Note that Count does not need to be checked against Max_Length,
            -- as Count is guaranteed by Ada not to exceed it.
         return Append (Left  => Source,
                        Right => ((Count - Source.Current_Length) * Pad)
           );
      end if;
   end Head;

   ----------
   -- Head --
   ----------

   procedure Head
     (Source : in out Sequence;
      Count  : in Length_Range;
      Pad    : in Element;
      Drop   : in Truncation := Error)
   is
   begin
      Source := Head (Source, Count, Pad, Drop);
   end Head;

   ------------
   -- Length --
   ------------

   function Length (Source : in Sequence) return Length_Range is
   begin
      return Source.Current_Length;
   end Length;

   ---------------------
   -- Replace_Element --
   ---------------------

   procedure Replace_Element
     (Source : in out Sequence;
      Index  : in Index_Range;
      By     : in Element)
   is
   begin
      Source.Content (Index) := By;
   end Replace_Element;

   ---------------
   -- Replicate --
   ---------------

   function Replicate
     (Count : in Length_Range;
      Item  : in Element;
      Drop  : in Truncation := Error)
      return Sequence
   is
   begin
      if Count <= Max_Length then
         return Count * Item;
      else  -- Count > Max_Length
         case Drop is
            when Error =>
               raise Length_Error;
            when Seq.Left | Seq.Right =>   -- Truncate on either side.
               return Sequence'(Max_Length, (others => Item));
         end case;
      end if;
   end Replicate;

   ---------------
   -- Replicate --
   ---------------
   --   Truncation is assumed to take place only for an entire "Item",
   --   so that Drop=Left and Drop=Right have the identical effect
   --   (since there is no pad).
   function Replicate
     (Count : in Length_Range;
      Item  : in Sequence;
      Drop  : in Truncation := Error)
      return Sequence
   is
      Len_To_Return : Natural := Count * Item.Current_Length;
      -- Natural so that upper bound of Length_Range will not be violated.
   begin
      if Len_To_Return <= Max_Length then
         return Count * Item;
      else  -- Len_To_Return > Max_Length
         case Drop is
            when Error =>
               raise Length_Error;
            when Seq.Left | Seq.Right =>   -- Truncate on either side.
               return (Max_Length / Item.Current_Length) * Item;
         end case;
      end if;
   end Replicate;

   ----------
   -- Tail --
   ----------

   function Tail
     (Source : in Sequence;
      Count  : in Length_Range;
      Pad    : in Element;
      Drop   : in Truncation := Error)
      return Sequence
   is
   begin
      if Source.Current_Length = Count then
         return Source;
      elsif Source.Current_Length > Count then
         case Drop is
            when Error =>
               raise Length_Error;
            when Seq.Left =>   -- Truncate on left side.
               return Delete (Source,
                              From    => 1,
                              Through => Source.Current_Length - Count);
            when Seq.Right =>  -- Truncate on right side.
               return Delete (Source,
                              From    => Count + 1,
                              Through => Source.Current_Length);
         end case;
      else  -- Source.Current_Length < Count

            -- Note that Count does not need to be checked against Max_Length,
            -- as Count is guaranteed by Ada not to exceed it.
         return Append (Left  => ((Count - Source.Current_Length) * Pad),
                        Right => Source
           );
      end if;
   end Tail;

   ----------
   -- Tail --
   ----------

   procedure Tail
     (Source : in out Sequence;
      Count  : in Length_Range;
      Pad    : in Element;
      Drop   : in Truncation := Error)
   is
   begin
      Source := Tail (Source, Count, Pad, Drop);
   end Tail;

   -----------------
   -- To_Sequence --
   -----------------

   function To_Sequence
     (Length : in Index_Range)
      return Sequence
   is
   begin
      return Sequence'(Current_Length => Length,
                       Content        => (others => Null_Element));
   end To_Sequence;

   --------------------
   -- Element_Arrays --
   --------------------
   package body Element_Arrays is separate;

end FACE.Sequences.Bounded;
