------------------------------------------------------------------------------------
-- DISTRIBUTION STATEMENT A. Approved for public release; distribution is unlimited.
--
-- Product produced under DoD SENSIAC contract HC104705D4000 under
-- the sponsorship of the Defense Technical Information Center,
-- ATTN: DTIC-AI, 8723 John J. Kingman Rd., Ste 0944, Fort Belvoir, VA
-- 22060-6218.  SENSIAC is a DoD Information Analysis Center Sponsored
-- by the Defense Technical Information Center.
--
-- HANDLING AND DESTRUCTION NOTICE - Comply with distribution statement
-- and destroy by any method that will prevent disclosure of
-- contents or reconstruction of the document.
------------------------------------------------------------------------------------

with Ada.Unchecked_Deallocation;

generic
   type Element is private;    -- can be any non-limited type
   Null_Element : in Element;  -- constant value used to represent nul
   Max_Length   : in Positive; -- max length one of these strings can ever have
                               -- implementation must check in body at run-time
package FACE.Sequences.Bounded is

   subtype Length_Range is Natural      range 0 .. Max_Length;
   subtype Index_Range  is Length_Range range 1 .. Max_Length;

   type Sequence is private;

   Null_Sequence : constant Sequence;  -- e.g., an empty Sequence
                                       -- initial value of all Sequences

   function Length (Source : in Sequence) return Length_Range;
   pragma Inline (Length);

   --------------------------------------------------------
   -- Conversion, Concatenation, and Selection Functions --
   --------------------------------------------------------

   function Append
     (Left, Right : in Sequence;
      Drop        : in Truncation := Error)
      return Sequence;

   function Append
     (Left  : in Sequence;
      Right : in Element;
      Drop  : in Truncation := Error)
      return Sequence;

   function Append
     (Left  : in Element;
      Right : in Sequence;
      Drop  : in Truncation := Error)
      return Sequence;

   procedure Append
     (Source   : in out Sequence;
      New_Item : in Sequence;
      Drop     : in Truncation := Error);

   procedure Append
     (Source   : in out Sequence;
      New_Item : in Element;
      Drop     : in Truncation := Error);

   function "&"
     (Left, Right : in Sequence)
      return Sequence;

   function "&"
     (Left  : in Sequence;
      Right : in Element)
      return Sequence;

   function "&"
     (Left  : in Element;
      Right : in Sequence)
      return Sequence;

   function Get_Element
     (Source : in Sequence;
      Index  : in Index_Range)
      return Element;

   procedure Replace_Element
     (Source : in out Sequence;
      Index  : in Index_Range;
      By     : in Element);

   function "=" (Left, Right : in Sequence) return Boolean;

   -----------------------------------------
   -- Sequence transformation subprograms --
   -----------------------------------------

   function Delete
     (Source  : in Sequence;
      From    : in Index_Range;
      Through : in Index_Range)
      return Sequence;

   procedure Delete
     (Source  : in out Sequence;
      From    : in Index_Range;
      Through : in Index_Range);

   -----------------------------------
   -- Sequence selector subprograms --
   -----------------------------------

   function Head
     (Source : in Sequence;
      Count  : in Length_Range;
      Pad    : in Element;
      Drop   : in Truncation := Error)
      return Sequence;

   procedure Head
     (Source : in out Sequence;
      Count  : in Length_Range;
      Pad    : in Element;
      Drop   : in Truncation := Error);

   function Tail
     (Source : in Sequence;
      Count  : in Length_Range;
      Pad    : in Element;
      Drop   : in Truncation := Error)
      return Sequence;

   procedure Tail
     (Source : in out Sequence;
      Count  : in Length_Range;
      Pad    : in Element;
      Drop   : in Truncation := Error);

   --------------------------------------
   -- Sequence constructor subprograms --
   --------------------------------------

   function "*"
     (Left  : in Length_Range;
      Right : in Element)
      return Sequence;

   function "*"
     (Left  : in Length_Range;
      Right : in Sequence)
      return Sequence;

   function Replicate
     (Count : in Length_Range;
      Item  : in Element;
      Drop  : in Truncation := Error)
      return Sequence;

   function Replicate
     (Count : in Length_Range;
      Item  : in Sequence;
      Drop  : in Truncation := Error)
      return Sequence;

   generic
      type Index_Type is range <>;  -- Can have any range.
                                    -- Indices may be negative.
   package Element_Arrays is

      type Element_Array        is array (Index_Type range <>) of Element;
      type Element_Array_Access is access all Element_Array;

      Null_Element_Array : Element_Array (Index_Type'Last .. Index_Type'First - 1);

      --------------------------------------------------------
      -- Conversion, Concatenation, and Selection Functions --
      --------------------------------------------------------

      function To_Sequence
        (Source : in Element_Array;
         Drop   : in Truncation := Error)
         return Sequence;

      function To_Sequence (Length : in Index_Range)
                            return Sequence;

      procedure Set
        (Item   : in out Sequence;
         Source : in Element_Array;
         Drop   : in Truncation := Error);

      function To_Element_Array
        (Source : in Sequence)
         return Element_Array;

      function Append
        (Left  : in Sequence;
         Right : in Element_Array;
         Drop  : in Truncation := Error)
         return Sequence;

      function Append
        (Left  : in Element_Array;
         Right : in Sequence;
         Drop  : in Truncation := Error)
         return Sequence;

      procedure Append
        (Source   : in out Sequence;
         New_Item : in Element_Array;
         Drop     : in Truncation := Error);

      function "&"
        (Left  : in Sequence;
         Right : in Element_Array)
         return Sequence;

      function "&"
        (Left  : in Element_Array;
         Right : in Sequence)
         return Sequence;

      function Slice
        (Source : in Sequence;
         Low    : in Index_Range;
         High   : in Index_Range)
         return Element_Array;

      function "="
        (Left  : in Sequence;
         Right : in Element_Array)
         return Boolean;

      function "="
        (Left  : in Element_Array;
         Right : in Sequence)
         return Boolean;

      ----------------------
      -- Search functions --
      ----------------------

      function Index
        (Source  : in Sequence;
         Pattern : in Element_Array;
         Going   : in Direction := Forward)
         return Length_Range; -- 0 indicates not found

      function Count
        (Source  : in Sequence;
         Pattern : in Element_Array)
         return Length_Range;

      -----------------------------------------
      -- Sequence transformation subprograms --
      -----------------------------------------

      function Replace_Slice
        (Source : in Sequence;
         Low    : in Length_Range;
         High   : in Length_Range;
         By     : in Element_Array;
         Drop   : in Truncation := Error)
         return Sequence;

      procedure Replace_Slice
        (Source : in out Sequence;
         Low    : in Length_Range;
         High   : in Length_Range;
         By     : in Element_Array;
         Drop   : in Truncation := Error);

      function Insert
        (Source   : in Sequence;
         Before   : in Length_Range;
         New_Item : in Element_Array;
         Drop     : in Truncation := Error)
         return Sequence;

      procedure Insert
        (Source   : in out Sequence;
         Before   : in Length_Range;
         New_Item : in Element_Array;
         Drop     : in Truncation := Error);

      function Overwrite
        (Source   : in Sequence;
         Position : in Length_Range;
         New_Item : in Element_Array;
         Drop     : in Truncation := Error)
         return Sequence;

      procedure Overwrite
        (Source   : in out Sequence;
         Position : in Length_Range;
         New_Item : in Element_Array;
         Drop     : in Truncation := Error);

      --------------------------------------
      -- Sequence constructor subprograms --
      --------------------------------------

      function "*"
        (Left  : in Length_Range;
         Right : in Element_Array)
         return Sequence;

      function Replicate
        (Count : in Length_Range;
         Item  : in Element_Array;
         Drop  : in Truncation := Error)
         return Sequence;

   end Element_Arrays;

private
   type Sequence_Vector_Type   is array (Index_Range range <>) of Element;
   subtype Sequence_Vector     is Sequence_Vector_Type (Index_Range'First .. Max_Length);
   type Sequence_Vector_Access is access all Sequence_Vector;

   type Sequence               is record
      Current_Length : Length_Range    := 0;
      Content        : Sequence_Vector := (others => Null_Element);
   end record;

   Null_Sequence : constant Sequence := (Current_Length => 0,
                                         Content        => (others => Null_Element));

end FACE.Sequences.Bounded;
