with Ada.Unchecked_Deallocation;

------------------------------
-- FACE.Sequences.Unbounded --
------------------------------
package body FACE.Sequences.Unbounded is

   -- Allocation_Delta is the min delta for allocating a new record on
   -- the heap.
   Allocation_Delta   : constant := 128;

   type Allocate_Choice is (No_Allocation, Less_Allocation, More_Allocation);

   --
   -- Specifications for subprograms internal to
   -- FACE.Sequences.Unbounded
   --

   ---------------------
   -- Allocate_New_SR --
   ---------------------
   --   Determines whether a new Sequence_Record should be allocated.
   function Allocate_New_SR (Old_SR     : Sequence_Record_Access;
                             New_Length : Length_Range) return Allocate_Choice;

   ------------------
   -- New_Sequence --
   ------------------
   --   This function is a convenient short-hand for accessing the
   --   New_SVec procedure.
   --   Works correctly when Old_SR and New_SR are the same object.
   function New_Sequence (New_Content : in Sequence_Vector) return Sequence;

   --------------
   -- New_SVec --
   --------------
   --   Works correctly when Old_SR and New_SR are the same object.
   --
   procedure New_SVec (Old_SR      : in out Sequence_Record_Access;
                       New_Content : in     Sequence_Vector;
                       New_SR      :    out Sequence_Record_Access);

   --
   -- Subprogram bodies (alphabetized)
   --

   ---------
   -- "&" --
   ---------
   function "&"
     (Left, Right : in Sequence)
      return Sequence
   is
      Left_Length  : constant Length_Range := Left.SR.Current_Length;
      Right_Length : constant Length_Range := Right.SR.Current_Length;
      New_Seq : Sequence;
   begin
      New_SVec (Old_SR      => New_Seq.SR,
                New_Content =>
                  Left.SR.Content  (1 .. Left_Length) &
                  Right.SR.Content (1 .. Right_Length),
                New_SR      => New_Seq.SR);
      return New_Seq;
   end "&";

   ---------
   -- "&" --
   ---------
   function "&"
     (Left  : in Sequence;
      Right : in Element)
      return Sequence
   is
      Left_Length : constant Length_Range := Left.SR.Current_Length;
      New_Seq : Sequence;
   begin
      New_SVec (Old_SR      => New_Seq.SR,
                New_Content => Left.SR.Content  (1 .. Left_Length) & Right,
                New_SR      => New_Seq.SR);
      return New_Seq;
   end "&";

   ---------
   -- "&" --
   ---------
   function "&"
     (Left  : in Element;
      Right : in Sequence)
      return Sequence
   is
      Right_Length : constant Length_Range := Right.SR.Current_Length;
      New_Seq : Sequence;
   begin
      New_SVec (Old_SR      => New_Seq.SR,
                New_Content => Left & Right.SR.Content  (1 .. Right_Length),
                New_SR      => New_Seq.SR);
      return New_Seq;
   end "&";

   ---------
   -- "*" --
   ---------
   function "*"
     (Left  : in Length_Range;
      Right : in Element)
      return Sequence
   is
   begin
      return New_Sequence (New_Content => Sequence_Vector'(1 .. Left => Right));
   end "*";

   ---------
   -- "*" --
   ---------
   function "*"
     (Left  : in Length_Range;
      Right : in Sequence)
      return Sequence
   is
   begin
      if Left < 0 then
         raise Length_Error;
      elsif Left = 0 then
         return Null_Sequence;
      else  -- Left >= 1 then
         declare
            Right_Length       : Length_Range := Right.SR.Current_Length;
            Result_Length      : Length_Range := Left * Right_Length;
            Sequence_To_Return : Sequence     := To_Sequence (Result_Length);
            First_Index        : Index_Range  := 1;
         begin
            for I in 1 .. Left loop
               Sequence_To_Return.SR.Content
                 (First_Index .. First_Index - 1 + Right_Length) :=
                 Right.SR.Content (1 .. Right_Length);
               First_Index := First_Index + Right_Length;
            end loop;

            Sequence_To_Return.SR.Current_Length := Result_Length;
            return Sequence_To_Return;
         end;
      end if;
   end "*";

   ---------
   -- "=" --
   ---------
   function "="
     (Left, Right : in Sequence) return Boolean
   is
      Left_Length  : constant Length_Range := Left .SR.Current_Length;
      Right_Length : constant Length_Range := Right.SR.Current_Length;
   begin
      return Boolean ((Left_Length = 0 and Right_Length = 0) or else
                      (Left_Length = Right_Length and then
                      Left.SR.Content (1 .. Left_Length) = Right.SR.Content (1 .. Right_Length)));
   end "=";

   ------------
   -- Adjust --
   ------------

   procedure Adjust (Object : in out Sequence)
   is
   begin
      Object.SR.Accesses.Increment;
   end Adjust;

   ---------------------
   -- Allocate_New_SR --
   ---------------------
   --   Determines whether a new Sequence_Record should be allocated.
   function Allocate_New_SR (Old_SR : Sequence_Record_Access;
                             New_Length : Length_Range) return Allocate_Choice
   is
      Allocate_Less : constant Boolean
        := Boolean (New_Length < Old_SR.Max_Length - Allocation_Delta);
      Allocate_More : constant Boolean := Boolean (New_Length > Old_SR.Max_Length);
      Allocate_None : constant Boolean := not Allocate_Less and not Allocate_More;
   begin
      if Allocate_Less then
         return Less_Allocation;
      elsif Allocate_More then
         return More_Allocation;
      else
         return No_Allocation;
      end if;
   end Allocate_New_SR;

   ------------
   -- Append --
   ------------
   procedure Append
     (Source   : in out Sequence;
      New_Item : in     Sequence)
   is
      Source_Length   : constant Length_Range := Source.SR.Current_Length;
      New_Item_Length : constant Length_Range := New_Item.SR.Current_Length;
   begin
      New_SVec (Old_SR      => Source.SR,
                New_Content =>
                  Source.SR.Content   (1 .. Source_Length) &
                  New_Item.SR.Content (1 .. New_Item_Length),
                New_SR      => Source.SR);
   end Append;

   ------------
   -- Append --
   ------------
   procedure Append
     (Source   : in out Sequence;
      New_Item : in     Element)
   is
      Source_Length   : constant Length_Range := Source.SR.Current_Length;
   begin
      New_SVec (Old_SR      => Source.SR,
                New_Content => Source.SR.Content (1 .. Source_Length) & New_Item,
                New_SR      => Source.SR);
   end Append;

   ----------
   -- Copy --
   ----------
   --   This function assists in writing correct code, making the
   --   distinction between assignment of an unbounded Sequence and
   --   making an actual copy of an unbounded Sequence.
   function Copy
     (Source : in Sequence)
      return Sequence
   is
      New_Sequence : Sequence;
   begin
      New_Sequence.SR                := new Sequence_Record (Source.SR.Max_Length);
      New_Sequence.SR.Current_Length := Source.SR.Current_Length;
      New_Sequence.SR.Content        := Source.SR.Content;
      return New_Sequence;
   end Copy;

   ------------
   -- Delete --
   ------------
   function Delete
     (Source  : in Sequence;
      From    : in Index_Range;
      Through : in Index_Range)
      return Sequence
   is
   begin
      if From in 1 .. Source.SR.Current_Length and Through in 1 .. Source.SR.Current_Length then
         if Through < From then
            -- not deleting anything
            return New_Sequence (New_Content => Source.SR.Content (1 .. Source.SR.Current_Length));
         else
            return New_Sequence (New_Content =>
                                   Source.SR.Content (1 .. From - 1) &
                                   Source.SR.Content (Through + 1 .. Source.SR.Current_Length));
         end if;
      else
         raise FACE.Sequences.Index_Error;
      end if;
   end Delete;

   ------------
   -- Delete --
   ------------
   procedure Delete
     (Source  : in out Sequence;
      From    : in Index_Range;
      Through : in Index_Range)
   is
   begin
      if From in 1 .. Source.SR.Current_Length and Through in 1 .. Source.SR.Current_Length then
         if Through < From then
            -- not deleting anything
            return;
         else
            New_SVec (Old_SR      => Source.SR,
                      New_Content =>
                        Source.SR.Content (1 .. From - 1) &
                        Source.SR.Content (Through + 1 .. Source.SR.Current_Length),
                      New_SR      => Source.SR);

            return;
         end if;
      else
         raise FACE.Sequences.Index_Error;
      end if;
   end Delete;

   ----------
   -- Head --
   ----------
   function Head
     (Source : in Sequence;
      Count  : in Length_Range;
      Pad    : in Element := Null_Element)
      return Sequence
   is
      Source_Length : constant Length_Range := Source.SR.Current_Length;
   begin
      if Count <= Source_Length then
         -- not deleting anything
         return New_Sequence (New_Content => Source.SR.Content (1 .. Count));
      else  -- Count > Source_Length
         return New_Sequence (New_Content =>
                                Source.SR.Content (1 .. Source_Length) &
                                (1 .. Count - Source_Length => Pad));
      end if;
   end Head;

   ----------
   -- Head --
   ----------
   procedure Head
     (Source : in out Sequence;
      Count  : in Length_Range;
      Pad    : in Element := Null_Element)
   is
   begin
      Source := Head (Source, Count, Pad);
   end Head;

   --------------
   -- Finalize --
   --------------

   procedure Finalize (Object : in out Sequence)
   is
      Usage_Dropped_To_Zero : Boolean;

      --------------------------
      -- Free_Sequence_Record --
      --------------------------
      procedure Free_Sequence_Record is new Ada.Unchecked_Deallocation
        (Sequence_Record,
         Sequence_Record_Access);
   begin
      if Object.SR /= null then
         Object.SR.Accesses.Decrement (Usage_Dropped_To_Zero);

         if Boolean (Object.SR /= Null_Sequence.SR) and Usage_Dropped_To_Zero then
            Free_Sequence_Record (Object.SR);
         end if;

         Object.SR := null;
      end if;  -- Object.SR /= null
   end Finalize;

   -----------------
   -- Get_Element --
   -----------------
   function Get_Element
     (Source : in Sequence;
      Index  : in Index_Range)
      return Element
   is
   begin
      if Index in 1 .. Source.SR.Current_Length then
         return Source.SR.Content (Index);
      else
         raise FACE.Sequences.Index_Error;
      end if;
   end Get_Element;

   ----------------
   -- Initialize --
   ----------------

   procedure Initialize (Object : in out Sequence)
   is
   begin
      Object.SR.Accesses.Increment;
   end Initialize;

   -------------
   -- Is_Null --
   -------------
   function Is_Null
     (Source : in Sequence) return Boolean
   is
   begin
      return Boolean (Source.SR = null or else Source.SR.Current_Length = 0);
   end Is_Null;

   ------------
   -- Length --
   ------------
   function Length (Source : in Sequence) return Length_Range is
   begin
      return Source.SR.Current_Length;
   end Length;

   ------------------
   -- New_Sequence --
   ------------------
   --   This function is a convenient short-hand for accessing the
   --   New_SVec procedure.
   --   Works correctly when Old_SR and New_SR are the same object.
   function New_Sequence (New_Content : in Sequence_Vector) return Sequence
   is
      Sequence_To_Return : Sequence;
   begin
      New_SVec (Old_SR      => Sequence_To_Return.SR,
                New_Content => New_Content,
                New_SR      => Sequence_To_Return.SR);
      return Sequence_To_Return;
   end New_Sequence;

   --------------
   -- New_SVec --
   --------------
   --   Works correctly when Old_SR and New_SR are the same object.
   --
   procedure New_SVec (Old_SR      : in out Sequence_Record_Access;
                       New_Content : in     Sequence_Vector;
                       New_SR      :    out Sequence_Record_Access)
   is
      Content_Length : constant Length_Range := New_Content'Length;
      New_Max_Length : constant Length_Range := Content_Length + Allocation_Delta;
   begin
      case Allocate_New_SR (Old_SR, Content_Length) is
         when Less_Allocation | More_Allocation =>
            -- Regardless of whether the current space is not large
            -- enough or too large, a new Sequence_Record must be
            -- allocated.
            New_SR := new Sequence_Record (New_Max_Length);
         when No_Allocation =>
            -- We are using space in the old sequence record.
            New_SR := Old_SR;
      end case;

      New_SR.Current_Length := Content_Length;
      New_SR.Content (1 .. New_SR.Current_Length) := New_Content;
   end New_SVec;

   ---------------------
   -- Replace_Element --
   ---------------------
   procedure Replace_Element
     (Source : in out Sequence;
      Index  : in Index_Range;
      By     : in Element)
   is
   begin
      if Index in 1 .. Source.SR.Current_Length then
         Source.SR.Content (Index) := By;
      else
         raise FACE.Sequences.Index_Error;
      end if;
   end Replace_Element;

   ----------
   -- Tail --
   ----------
   function Tail
     (Source : in Sequence;
      Count  : in Length_Range;
      Pad    : in Element := Null_Element)
      return Sequence
   is
      Source_Length : constant Length_Range := Source.SR.Current_Length;
   begin
      if Count <= Source_Length then
         -- not deleting anything
         return New_Sequence
           (New_Content =>
              Source.SR.Content (Source_Length - Count + 1 .. Source_Length));
      else  -- Count <= Source_Length
         return New_Sequence (New_Content =>
                                (1 .. Count - Source_Length => Pad) &
                                Source.SR.Content (1 .. Source_Length));
      end if;
   end Tail;

   ----------
   -- Tail --
   ----------
   procedure Tail
     (Source : in out Sequence;
      Count  : in Length_Range;
      Pad    : in Element := Null_Element)
   is
   begin
      Source := Tail (Source, Count, Pad);
   end Tail;

   -----------------
   -- To_Sequence --
   -----------------
   function To_Sequence
     (Length : in Length_Range)
      return Sequence
   is
      New_Sequence : Sequence;
   begin
      New_Sequence.SR                := new Sequence_Record (Max_Length => Length + Allocation_Delta);
      New_Sequence.SR.Current_Length := 0;
      New_Sequence.SR.Content        := (others => Null_Element);
      return New_Sequence;
   end To_Sequence;

   --------------------
   -- Element_Arrays --
   --------------------
   package body Element_Arrays is separate;

end FACE.Sequences.Unbounded;
